(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
  typeof define === 'function' && define.amd ? define(['exports'], factory) :
  (factory((global.substance = {})));
}(this, (function (exports) { 'use strict';

  var ChangeStore = function ChangeStore(seed) {
    this._changes = seed || {};
  };


  ChangeStore.prototype.getChanges = function getChanges (documentId, sinceVersion, toVersion, cb) {
    if (typeof sinceVersion === 'function') {
      cb = sinceVersion;
      sinceVersion = 0;
    } else if (typeof toVersion === 'function') {
      cb = toVersion;
      toVersion = undefined;
    }
    if (!(documentId && sinceVersion >= 0 && cb)) {
      throw new Error('Invalid arguments')
    }
    var version = this._getVersion(documentId);
    var changes = this._getChanges(documentId);
    changes = changes.slice(sinceVersion, toVersion);
    cb(null, changes, version);
  };


  ChangeStore.prototype.addChange = function addChange (documentId, change, cb) {
    if (!documentId || !change) {
      throw new Error('Invalid arguments')
    }
    this._addChange(documentId, change);
    var newVersion = this._getVersion(documentId);
    cb(null, newVersion);
  };


  ChangeStore.prototype.deleteChanges = function deleteChanges (documentId, cb) {
    var deletedChanges = this._deleteChanges(documentId);
    cb(null, deletedChanges.length);
  };


  ChangeStore.prototype.getVersion = function getVersion (id, cb) {
    cb(null, this._getVersion(id));
  };




  ChangeStore.prototype._deleteChanges = function _deleteChanges (documentId) {
    var changes = this._getChanges(documentId);
    delete this._changes[documentId];
    return changes
  };

  ChangeStore.prototype._getVersion = function _getVersion (documentId) {
    var changes = this._changes[documentId];
    return changes ? changes.length : 0
  };

  ChangeStore.prototype._getChanges = function _getChanges (documentId) {
    return this._changes[documentId] || []
  };

  ChangeStore.prototype._addChange = function _addChange (documentId, change) {
    if (!this._changes[documentId]) {
      this._changes[documentId] = [];
    }
    this._changes[documentId].push(change);
  };

  function forEach(iteratee, func) {
    if (!iteratee) { return }
    if (iteratee.constructor.prototype.forEach) {
      iteratee.forEach(func);
    } else {
      Object.keys(iteratee).forEach(function(key) {
        return func(iteratee[key], key)
      });
    }
  }

  var SubstanceError = (function (Error) {
    function SubstanceError(name, options) {
      Error.call(this, name, options);
      this.name = name;
      this.message = options.message;
      this.info = options.info;
      this.errorCode = options.errorCode;
      this.cause = options.cause;

      if (Error.captureStackTrace) {
        Error.captureStackTrace(this, (SubstanceError));
      }
    }

    if ( Error ) SubstanceError.__proto__ = Error;
    SubstanceError.prototype = Object.create( Error && Error.prototype );
    SubstanceError.prototype.constructor = SubstanceError;

    SubstanceError.prototype.inspect = function inspect () {
      var parts = [];

      
      parts.push(this.stack);

      
      if (this.info) {
        parts.push(this.info + '. ');
      }

      
      if (this.cause) {
        parts.push('\nCaused by: ');

        if (this.cause.inspect) {
          
          parts.push(this.cause.inspect());
        } else {
          
          parts.push(this.cause.toString());
        }
      }
      return parts.join('')
    };

    return SubstanceError;
  }(Error));


  SubstanceError.fromJSON = function(err) {
    if (!err) { return null }
    var error = new SubstanceError(err.name, {
      message: err.message,
      info: err.info,
      errorCode: err.errorCode,
      cause: SubstanceError.fromJSON(err.cause)
    });
    return error
  };

  function isObject(val) {
    var type = typeof val;
    return Boolean(val) && (type === 'object' || type === 'function')
  }

  var DEBUG = false;
  var count = 0;
  var COUNT_MSG = '%s listeners registered in the whole system.';


  var EventEmitter = function EventEmitter () {};

  var prototypeAccessors = { __events__: { configurable: true } };

  EventEmitter.prototype.emit = function emit (event) {
    if (event in this.__events__) {
      
      
      var bindings = this.__events__[event].slice();
      var args = Array.prototype.slice.call(arguments, 1);
      for (var i = 0, len = bindings.length; i < len; i++) {
        var binding = bindings[i];
        
        binding.method.apply(binding.context, args);
      }
      return true
    }
    return false
  };


  EventEmitter.prototype.on = function on (event, method, context) {
    
    _on.call(this, event, method, context);
  };


  EventEmitter.prototype.off = function off (event, method, context) { 
    if (arguments.length === 1 && isObject(arguments[0])) {
      _disconnect.call(this, arguments[0]);
    } else {
      _off.apply(this, arguments);
    }
  };

  EventEmitter.prototype._debugEvents = function _debugEvents () {
    
    console.log('### EventEmitter: ', this);
    forEach(this.__events__, function (handlers, name) {
      console.log("- %s listeners for %s: ", handlers.length, name, handlers);
    });
    
  };

  prototypeAccessors.__events__.get = function () {
    if (!this.___events___) {
      this.___events___ = {};
    }
    return this.___events___
  };

  Object.defineProperties( EventEmitter.prototype, prototypeAccessors );


  function _on(event, method, context) {
    
    var bindings;
    validateMethod( method, context );
    if (this.__events__.hasOwnProperty(event)) {
      bindings = this.__events__[event];
    } else {
      
      bindings = this.__events__[event] = [];
    }
    
    bindings.push({
      method: method,
      context: context || null
    });
    if (DEBUG) {
      count++;
      console.info('_on()', event, method.name, context, this);
      console.info(COUNT_MSG, count);
    }
    return this
    
  }


  function _off(event, method, context) {
    var this$1 = this;

    
    if (arguments.length === 0) {
      if (DEBUG) {
        forEach(this.__events__, function (bindings) {
          bindings.forEach(function (b) {
            console.info('_off()', b.method.name, b.context, this$1);
          });
          count -= bindings.length;
        });
        console.info(COUNT_MSG, count);
      }
      this.___events___ = {};
      return this
    }
    if (arguments.length === 1) {
      
      if (DEBUG) {
        count -= (this.__events__[event] || []).length;
        console.info(COUNT_MSG, count);
      }
      delete this.__events__[event];
      return this
    }
    validateMethod(method, context);
    if (!(event in this.__events__) || !this.__events__[event].length) {
      if (DEBUG) { console.info('NO MATCHING BINDINGS'); }
      
      return this
    }
    
    if (arguments.length < 3) {
      context = null;
    }
    
    var bindings = this.__events__[event];
    for (var i = bindings.length-1; i >= 0; i--) {
      var b = bindings[i];
      if (b.method === method && b.context === context) {
        bindings.splice(i, 1);
        if (DEBUG) { count--; }
      }
    }
    
    if (bindings.length === 0) {
      delete this.__events__[event];
    }
    if (DEBUG) { console.info(COUNT_MSG, count); }
    return this
    
  }


  function _disconnect(context) {
    var this$1 = this;

    
    
    forEach(this.__events__, function (bindings, event) {
      for (var i = bindings.length-1; i>=0; i--) {
        
        
        if (bindings[i] && bindings[i].context === context) {
          _off.call(this$1, event, bindings[i].method, context);
        }
      }
    });
    return this
    
  }

  function validateMethod(method, context) {
    
    if (typeof method === 'string') {
      
      if (context === undefined || context === null) {
        throw new Error( 'Method name "' + method + '" has no context.' )
      }
      if (!(method in context)) {
        
        
        throw new Error( 'Method not found: "' + method + '"' )
      }
      if (typeof context[method] !== 'function') {
        
        
        throw new Error( 'Property "' + method + '" is not a function' )
      }
    } else if (typeof method !== 'function') {
      throw new Error( 'Invalid callback. Function or method name expected.' )
    }
  }

  function uuid(prefix, len) {
    if (prefix && prefix[prefix.length-1] !== "-") {
      prefix = prefix.concat("-");
    }
    var chars = '0123456789abcdefghijklmnopqrstuvwxyz'.split('');
    var uuid = [];
    var radix = 16;
    var idx;
    len = len || 32;
    if (len) {
      
      for (idx = 0; idx < len; idx++) { uuid[idx] = chars[0 | Math.random()*radix]; }
    } else {
      
      var r;
      
      uuid[8] = uuid[13] = uuid[18] = uuid[23] = '-';
      uuid[14] = '4';
      
      
      for (idx = 0; idx < 36; idx++) {
        if (!uuid[idx]) {
          r = 0 | Math.random()*16;
          uuid[idx] = chars[(idx === 19) ? (r & 0x3) | 0x8 : r];
        }
      }
    }
    return (prefix ? prefix : "") + uuid.join('');
  }

  var ServerRequest = function ServerRequest(message, ws) {
    this.message = message;
    this.ws = ws;
    this.isAuthenticated = false;
    this.isAuhorized = false;
  };


  ServerRequest.prototype.setAuthenticated = function setAuthenticated (session) {
    this.isAuthenticated = true;
    this.session = session;
  };


  ServerRequest.prototype.setAuthorized = function setAuthorized (authorizationData) {
    this.isAuthorized = true;
    this.authorizationData = authorizationData;
  };


  ServerRequest.prototype.setEnhanced = function setEnhanced () {
    this.isEnhanced = true;
  };

  var ServerResponse = function ServerResponse() {
    this.isReady = false; 
    this.isEnhanced = false; 
    this.isSent = false; 
    this.err = null;
    this.data = null;
  };


  ServerResponse.prototype.error = function error (err) {
    this.err = err;
    this.isReady = true;
  };


  ServerResponse.prototype.send = function send (data) {
    this.data = data;
    this.isReady = true;
  };


  ServerResponse.prototype.setEnhanced = function setEnhanced () {
    this.isEnhanced = true;
  };

  ServerResponse.prototype.setSent = function setSent () {
    this.isSent = true;
  };

  var Server = (function (EventEmitter$$1) {
    function Server(config) {
      EventEmitter$$1.call(this);

      this.config = config;
      this._onConnection = this._onConnection.bind(this);
    }

    if ( EventEmitter$$1 ) Server.__proto__ = EventEmitter$$1;
    Server.prototype = Object.create( EventEmitter$$1 && EventEmitter$$1.prototype );
    Server.prototype.constructor = Server;

    Server.prototype.bind = function bind (wss) {
      if (this.wss) {
        throw new Error('Server is already bound to a websocket')
      }
      this.wss = wss;
      this._connections = new WeakMap();
      this._collaborators = {};
      this.wss.on('connection', this._onConnection);

      var interval = this.config.heartbeat;
      if (interval) {
        this._heartbeat = setInterval(this._sendHeartbeat.bind(this), interval);
      }
      this._bound = true;
    };

    
    Server.prototype.unbind = function unbind () {
      if (this._bound) {
        this.wss.off('connection', this._onConnection);
      } else {
        throw new Error('Server is not yet bound to a websocket.')
      }
    };

    
    Server.prototype.onConnection = function onConnection () {
      
    };

    
    Server.prototype.onDisconnect = function onDisconnect () {
      
    };

    
    Server.prototype.authenticate = function authenticate (req, res) {
      req.setAuthenticated();
      this.next(req, res);
    };

    
    Server.prototype.authorize = function authorize (req, res) {
      req.setAuthorized();
      this.next(req, res);
    };


    
    Server.prototype.enhanceRequest = function enhanceRequest (req, res) {
      req.setEnhanced();
      this.next(req, res);
    };

    
    Server.prototype.execute = function execute () {
      throw new Error('This method needs to be specified')
    };

    
    Server.prototype.enhanceResponse = function enhanceResponse (req, res) {
      res.setEnhanced();
      this.next(req, res);
    };

    
    Server.prototype._onConnection = function _onConnection (ws) {
      var collaboratorId = uuid();
      var connection = {
        collaboratorId: collaboratorId
      };
      this._connections.set(ws, connection);

      
      this._collaborators[collaboratorId] = {
        connection: ws
      };

      ws.on('message', this._onMessage.bind(this, ws));
      ws.on('close', this._onClose.bind(this, ws));
    };

    
    Server.prototype._onClose = function _onClose (ws) {
      var conn = this._connections.get(ws);
      var collaboratorId = conn.collaboratorId;

      this.onDisconnect(collaboratorId);

      
      delete this._collaborators[collaboratorId];
      this._connections.delete(ws);
    };

    
    Server.prototype.__initial = function __initial (req, res) {
      return !req.isAuthenticated && !req.isAuthorized && !res.isReady
    };

    Server.prototype.__authenticated = function __authenticated (req, res) {
      return req.isAuthenticated && !req.isAuthorized && !res.isReady
    };

    Server.prototype.__authorized = function __authorized (req, res) {
      return req.isAuthenticated && req.isAuthorized && !req.isEnhanced && !res.isReady
    };

    Server.prototype.__requestEnhanced = function __requestEnhanced (req, res) {
      return req.isAuthenticated && req.isAuthorized && req.isEnhanced && !res.isReady
    };

    Server.prototype.__executed = function __executed (req, res) {
      
      return req.isAuthenticated && req.isAuthorized && res.isReady && res.data && !res.isEnhanced
    };

    Server.prototype.__enhanced = function __enhanced (req, res) {
      return res.isReady && res.isEnhanced && !res.isSent
    };

    Server.prototype.__error = function __error (req, res) {
      return res.err && !res.isSent
    };

    Server.prototype.__done = function __done (req, res) {
      return res.isSent
    };

    Server.prototype.next = function next (req, res) {
      if (this.__initial(req, res)) {
        this.authenticate(req, res);
      } else if (this.__authenticated(req, res)) {
        this.authorize(req, res);
      } else if (this.__authorized(req, res)) {
        this.enhanceRequest(req, res);
      } else if (this.__requestEnhanced(req, res)) {
        this.execute(req, res);
      } else if (this.__executed(req, res)) {
        this.enhanceResponse(req, res);
      } else if (this.__enhanced(req, res)) {
        this.sendResponse(req, res);
      } else if (this.__error(req, res)) {
        this.sendError(req, res);
      } else if (this.__done(req,res)) ;
    };

    
    Server.prototype.sendError = function sendError (req, res) {
      var collaboratorId = req.message.collaboratorId;
      var msg = res.err;
      this.send(collaboratorId, msg);
      res.setSent();
      this.next(req, res);
    };

    
    Server.prototype._sendHeartbeat = function _sendHeartbeat () {
      Object.keys(this._collaborators).forEach(function(collaboratorId) {
        this.send(collaboratorId, {
          type: 'highfive',
          scope: '_internal'
        });
      }.bind(this));
    };

    
    Server.prototype.sendResponse = function sendResponse (req, res) {
      var collaboratorId = req.message.collaboratorId;
      this.send(collaboratorId, res.data);
      res.setSent();
      this.next(req, res);
    };

    Server.prototype._isWebsocketOpen = function _isWebsocketOpen (ws) {
      return ws && ws.readyState === 1
    };

    
    Server.prototype.send = function send (collaboratorId, message) {
      if (!message.scope && this.config.scope) {
        message.scope = this.config.scope;
      }

      var ws = this._collaborators[collaboratorId].connection;
      if (this._isWebsocketOpen(ws)) {
        ws.send(this.serializeMessage(message));
      } else {
        console.error('Server#send: Websocket for collaborator', collaboratorId, 'is no longer open', message);
      }
    };

    
    Server.prototype.broadCast = function broadCast (collaborators, message) {
      collaborators.forEach(function(collaboratorId) {
        this.send(collaboratorId, message);
      }.bind(this));
    };

    
    Server.prototype._processRequest = function _processRequest (req) {
      var res = new ServerResponse();
      this.next(req, res);
    };

    
    Server.prototype._onMessage = function _onMessage (ws, msg) {
      
      var conn = this._connections.get(ws);
      msg = this.deserializeMessage(msg);

      if (msg.scope === this.scope) {
        
        msg.collaboratorId = conn.collaboratorId;
        var req = new ServerRequest(msg, ws);
        this._processRequest(req);
      }
    };

    Server.prototype.serializeMessage = function serializeMessage (msg) {
      return JSON.stringify(msg)
    };

    Server.prototype.deserializeMessage = function deserializeMessage (msg) {
      return JSON.parse(msg)
    };

    return Server;
  }(EventEmitter));

  function map(iteratee, func) {
    if (!iteratee) { return [] }
    if (!func) { func = function(item) { return item }; }
    if (Array.isArray(iteratee)) {
      return iteratee.map(func)
    } else {
      return Object.keys(iteratee).map(function(key) {
        return func(iteratee[key], key)
      })
    }
  }

  function isPlainObject(o) {
    return Boolean(o) && o.constructor === {}.constructor
  }

  function isArray(a) {
    return Array.isArray(a)
  }

  function clone(val) {
    if (isArray(val)) {
      return val.slice(0)
    }
    if (isObject(val)) {
      return Object.assign({}, val)
    }
    
    
    return val
  }

  var platform = {

    inBrowser: false,

    inNodeJS: false,

    inElectron: false,

    
    isIE: false,
    

    isFF: false,

    isWebkit: false,

    
    version: -1,

    

    isWindows: false,

    isMac: false,

    devtools: false,

    
    
    _reset: detect,
  };

  function detect() {

    if (typeof window !== 'undefined') {
      platform.inBrowser = true;

      
      var ua = window.navigator.userAgent;
      var msie = ua.indexOf('MSIE ');
      var trident = ua.indexOf('Trident/');
      var edge = ua.indexOf('Edge/');

      if (msie > 0) {
        
        platform.isIE = true;
        platform.version = 10;
        
        
      } else if (trident > 0) {
        
        platform.isIE = true;
        platform.version = 11;
        platform.isTrident = true;
        
        
        
      } else if (edge > 0) {
        
        platform.isIE = true;
        platform.isEdge = true;
        platform.version = 12;
        
        parseInt(ua.substring(edge + 5, ua.indexOf('.', edge)), 10);
      }

      
      platform.isFF = window.navigator.userAgent.toLowerCase().indexOf('firefox') > -1;

      
      platform.isWebkit = !platform.isFF && !platform.isIE;
    } else {
      platform.inBrowser = false;
    }

    if (platform.inBrowser) {
      platform.isWindows = (window.navigator !== undefined && window.navigator.appVersion && window.navigator.appVersion.indexOf("Win") !== -1);
      platform.isMac = (window.navigator !== undefined && window.navigator.platform.indexOf('Mac') >= 0);
    }

    
    if (typeof process !== 'undefined') {
      if (platform.inBrowser) {
        platform.inElectron = true;
      } else {
        platform.inNodeJS = true;
      }
    }

    if (platform.inBrowser) {
      var widthThreshold = window.outerWidth - window.innerWidth > 160;
      var heightThreshold = window.outerHeight - window.innerHeight > 160;
      var orientation = widthThreshold ? 'vertical' : 'horizontal';
      if (!(heightThreshold && widthThreshold) &&
        ((window.Firebug && window.Firebug.chrome && window.Firebug.chrome.isInitialized) || widthThreshold || heightThreshold)) {
        platform.devtools = true;
        platform.devtoolsOrientation = orientation;
      }
    }
  }

  detect();

  function cloneDeep(val) {
    if (isArray(val)) {
      return _cloneArrayDeep(val);
    }
    
    
    
    if (platform.inBrowser && val instanceof window.File) {
      return val
    }
    if (isObject(val)) {
      return _cloneObjectDeep(val)
    }
    
    
    return val
  }

  function _cloneObjectDeep(obj) {
    var res = {};
    forEach(obj, function (val, key) {
      res[key] = cloneDeep(val);
    });
    return res
  }

  function _cloneArrayDeep(arr) {
    return arr.map(cloneDeep)
  }

  function isNumber(n) {
    return typeof n === 'number'
  }

  function isEqual(a, b) {
    if (a === b) { return true }
    if (isArray(a) && isArray(b)) {
      if (a.length !== b.length) { return false }
      for (var i = 0; i < a.length; i++) {
        if (!isEqual(a[i], b[i])) { return false }
      }
      return true
    }
    if (isPlainObject(a) && isPlainObject(b)) {
      var akeys = Object.keys(a).sort();
      var bkeys = Object.keys(b).sort();
      if (!isEqual(akeys, bkeys)) { return false }
      for (var i$1 = 0; i$1 < akeys.length; i$1++) {
        var key = akeys[i$1];
        if (!isEqual(a[key], b[key])) { return false }
      }
      return true
    }
    return false
  }

  function isNil(o) {
    return o === null || o === undefined
  }

  function isString(s) {
    return typeof s === 'string'
  }

  var freeGlobal = typeof global == 'object' && global && global.Object === Object && global;

  var freeSelf = typeof self == 'object' && self && self.Object === Object && self;


  var root = freeGlobal || freeSelf || Function('return this')();

  var Symbol = root.Symbol;

  function arrayMap(array, iteratee) {
    var index = -1,
        length = array == null ? 0 : array.length,
        result = Array(length);

    while (++index < length) {
      result[index] = iteratee(array[index], index, array);
    }
    return result;
  }

  var isArray$1 = Array.isArray;

  var objectProto = Object.prototype;


  var hasOwnProperty = objectProto.hasOwnProperty;


  var nativeObjectToString = objectProto.toString;


  var symToStringTag = Symbol ? Symbol.toStringTag : undefined;


  function getRawTag(value) {
    var isOwn = hasOwnProperty.call(value, symToStringTag),
        tag = value[symToStringTag];

    try {
      value[symToStringTag] = undefined;
      var unmasked = true;
    } catch (e) {}

    var result = nativeObjectToString.call(value);
    if (unmasked) {
      if (isOwn) {
        value[symToStringTag] = tag;
      } else {
        delete value[symToStringTag];
      }
    }
    return result;
  }

  var objectProto$1 = Object.prototype;


  var nativeObjectToString$1 = objectProto$1.toString;


  function objectToString(value) {
    return nativeObjectToString$1.call(value);
  }

  var nullTag = '[object Null]',
      undefinedTag = '[object Undefined]';


  var symToStringTag$1 = Symbol ? Symbol.toStringTag : undefined;


  function baseGetTag(value) {
    if (value == null) {
      return value === undefined ? undefinedTag : nullTag;
    }
    return (symToStringTag$1 && symToStringTag$1 in Object(value))
      ? getRawTag(value)
      : objectToString(value);
  }

  function isObjectLike(value) {
    return value != null && typeof value == 'object';
  }

  var symbolTag = '[object Symbol]';


  function isSymbol(value) {
    return typeof value == 'symbol' ||
      (isObjectLike(value) && baseGetTag(value) == symbolTag);
  }

  var INFINITY = 1 / 0;


  var symbolProto = Symbol ? Symbol.prototype : undefined,
      symbolToString = symbolProto ? symbolProto.toString : undefined;


  function baseToString(value) {
    
    if (typeof value == 'string') {
      return value;
    }
    if (isArray$1(value)) {
      
      return arrayMap(value, baseToString) + '';
    }
    if (isSymbol(value)) {
      return symbolToString ? symbolToString.call(value) : '';
    }
    var result = (value + '');
    return (result == '0' && (1 / value) == -INFINITY) ? '-0' : result;
  }

  function toString(value) {
    return value == null ? '' : baseToString(value);
  }

  function baseSlice(array, start, end) {
    var index = -1,
        length = array.length;

    if (start < 0) {
      start = -start > length ? 0 : (length + start);
    }
    end = end > length ? length : end;
    if (end < 0) {
      end += length;
    }
    length = start > end ? 0 : ((end - start) >>> 0);
    start >>>= 0;

    var result = Array(length);
    while (++index < length) {
      result[index] = array[index + start];
    }
    return result;
  }

  function castSlice(array, start, end) {
    var length = array.length;
    end = end === undefined ? length : end;
    return (!start && end >= length) ? array : baseSlice(array, start, end);
  }

  var rsAstralRange = '\\ud800-\\udfff',
      rsComboMarksRange = '\\u0300-\\u036f',
      reComboHalfMarksRange = '\\ufe20-\\ufe2f',
      rsComboSymbolsRange = '\\u20d0-\\u20ff',
      rsComboRange = rsComboMarksRange + reComboHalfMarksRange + rsComboSymbolsRange,
      rsVarRange = '\\ufe0e\\ufe0f';


  var rsZWJ = '\\u200d';


  var reHasUnicode = RegExp('[' + rsZWJ + rsAstralRange  + rsComboRange + rsVarRange + ']');


  function hasUnicode(string) {
    return reHasUnicode.test(string);
  }

  function asciiToArray(string) {
    return string.split('');
  }

  var rsAstralRange$1 = '\\ud800-\\udfff',
      rsComboMarksRange$1 = '\\u0300-\\u036f',
      reComboHalfMarksRange$1 = '\\ufe20-\\ufe2f',
      rsComboSymbolsRange$1 = '\\u20d0-\\u20ff',
      rsComboRange$1 = rsComboMarksRange$1 + reComboHalfMarksRange$1 + rsComboSymbolsRange$1,
      rsVarRange$1 = '\\ufe0e\\ufe0f';


  var rsAstral = '[' + rsAstralRange$1 + ']',
      rsCombo = '[' + rsComboRange$1 + ']',
      rsFitz = '\\ud83c[\\udffb-\\udfff]',
      rsModifier = '(?:' + rsCombo + '|' + rsFitz + ')',
      rsNonAstral = '[^' + rsAstralRange$1 + ']',
      rsRegional = '(?:\\ud83c[\\udde6-\\uddff]){2}',
      rsSurrPair = '[\\ud800-\\udbff][\\udc00-\\udfff]',
      rsZWJ$1 = '\\u200d';


  var reOptMod = rsModifier + '?',
      rsOptVar = '[' + rsVarRange$1 + ']?',
      rsOptJoin = '(?:' + rsZWJ$1 + '(?:' + [rsNonAstral, rsRegional, rsSurrPair].join('|') + ')' + rsOptVar + reOptMod + ')*',
      rsSeq = rsOptVar + reOptMod + rsOptJoin,
      rsSymbol = '(?:' + [rsNonAstral + rsCombo + '?', rsCombo, rsRegional, rsSurrPair, rsAstral].join('|') + ')';


  var reUnicode = RegExp(rsFitz + '(?=' + rsFitz + ')|' + rsSymbol + rsSeq, 'g');


  function unicodeToArray(string) {
    return string.match(reUnicode) || [];
  }

  function stringToArray(string) {
    return hasUnicode(string)
      ? unicodeToArray(string)
      : asciiToArray(string);
  }

  function createCaseFirst(methodName) {
    return function(string) {
      string = toString(string);

      var strSymbols = hasUnicode(string)
        ? stringToArray(string)
        : undefined;

      var chr = strSymbols
        ? strSymbols[0]
        : string.charAt(0);

      var trailing = strSymbols
        ? castSlice(strSymbols, 1).join('')
        : string.slice(1);

      return chr[methodName]() + trailing;
    };
  }

  var upperFirst = createCaseFirst('toUpperCase');

  function capitalize(string) {
    return upperFirst(toString(string).toLowerCase());
  }

  function arrayReduce(array, iteratee, accumulator, initAccum) {
    var index = -1,
        length = array == null ? 0 : array.length;

    if (initAccum && length) {
      accumulator = array[++index];
    }
    while (++index < length) {
      accumulator = iteratee(accumulator, array[index], index, array);
    }
    return accumulator;
  }

  function basePropertyOf(object) {
    return function(key) {
      return object == null ? undefined : object[key];
    };
  }

  var deburredLetters = {
    
    '\xc0': 'A',  '\xc1': 'A', '\xc2': 'A', '\xc3': 'A', '\xc4': 'A', '\xc5': 'A',
    '\xe0': 'a',  '\xe1': 'a', '\xe2': 'a', '\xe3': 'a', '\xe4': 'a', '\xe5': 'a',
    '\xc7': 'C',  '\xe7': 'c',
    '\xd0': 'D',  '\xf0': 'd',
    '\xc8': 'E',  '\xc9': 'E', '\xca': 'E', '\xcb': 'E',
    '\xe8': 'e',  '\xe9': 'e', '\xea': 'e', '\xeb': 'e',
    '\xcc': 'I',  '\xcd': 'I', '\xce': 'I', '\xcf': 'I',
    '\xec': 'i',  '\xed': 'i', '\xee': 'i', '\xef': 'i',
    '\xd1': 'N',  '\xf1': 'n',
    '\xd2': 'O',  '\xd3': 'O', '\xd4': 'O', '\xd5': 'O', '\xd6': 'O', '\xd8': 'O',
    '\xf2': 'o',  '\xf3': 'o', '\xf4': 'o', '\xf5': 'o', '\xf6': 'o', '\xf8': 'o',
    '\xd9': 'U',  '\xda': 'U', '\xdb': 'U', '\xdc': 'U',
    '\xf9': 'u',  '\xfa': 'u', '\xfb': 'u', '\xfc': 'u',
    '\xdd': 'Y',  '\xfd': 'y', '\xff': 'y',
    '\xc6': 'Ae', '\xe6': 'ae',
    '\xde': 'Th', '\xfe': 'th',
    '\xdf': 'ss',
    
    '\u0100': 'A',  '\u0102': 'A', '\u0104': 'A',
    '\u0101': 'a',  '\u0103': 'a', '\u0105': 'a',
    '\u0106': 'C',  '\u0108': 'C', '\u010a': 'C', '\u010c': 'C',
    '\u0107': 'c',  '\u0109': 'c', '\u010b': 'c', '\u010d': 'c',
    '\u010e': 'D',  '\u0110': 'D', '\u010f': 'd', '\u0111': 'd',
    '\u0112': 'E',  '\u0114': 'E', '\u0116': 'E', '\u0118': 'E', '\u011a': 'E',
    '\u0113': 'e',  '\u0115': 'e', '\u0117': 'e', '\u0119': 'e', '\u011b': 'e',
    '\u011c': 'G',  '\u011e': 'G', '\u0120': 'G', '\u0122': 'G',
    '\u011d': 'g',  '\u011f': 'g', '\u0121': 'g', '\u0123': 'g',
    '\u0124': 'H',  '\u0126': 'H', '\u0125': 'h', '\u0127': 'h',
    '\u0128': 'I',  '\u012a': 'I', '\u012c': 'I', '\u012e': 'I', '\u0130': 'I',
    '\u0129': 'i',  '\u012b': 'i', '\u012d': 'i', '\u012f': 'i', '\u0131': 'i',
    '\u0134': 'J',  '\u0135': 'j',
    '\u0136': 'K',  '\u0137': 'k', '\u0138': 'k',
    '\u0139': 'L',  '\u013b': 'L', '\u013d': 'L', '\u013f': 'L', '\u0141': 'L',
    '\u013a': 'l',  '\u013c': 'l', '\u013e': 'l', '\u0140': 'l', '\u0142': 'l',
    '\u0143': 'N',  '\u0145': 'N', '\u0147': 'N', '\u014a': 'N',
    '\u0144': 'n',  '\u0146': 'n', '\u0148': 'n', '\u014b': 'n',
    '\u014c': 'O',  '\u014e': 'O', '\u0150': 'O',
    '\u014d': 'o',  '\u014f': 'o', '\u0151': 'o',
    '\u0154': 'R',  '\u0156': 'R', '\u0158': 'R',
    '\u0155': 'r',  '\u0157': 'r', '\u0159': 'r',
    '\u015a': 'S',  '\u015c': 'S', '\u015e': 'S', '\u0160': 'S',
    '\u015b': 's',  '\u015d': 's', '\u015f': 's', '\u0161': 's',
    '\u0162': 'T',  '\u0164': 'T', '\u0166': 'T',
    '\u0163': 't',  '\u0165': 't', '\u0167': 't',
    '\u0168': 'U',  '\u016a': 'U', '\u016c': 'U', '\u016e': 'U', '\u0170': 'U', '\u0172': 'U',
    '\u0169': 'u',  '\u016b': 'u', '\u016d': 'u', '\u016f': 'u', '\u0171': 'u', '\u0173': 'u',
    '\u0174': 'W',  '\u0175': 'w',
    '\u0176': 'Y',  '\u0177': 'y', '\u0178': 'Y',
    '\u0179': 'Z',  '\u017b': 'Z', '\u017d': 'Z',
    '\u017a': 'z',  '\u017c': 'z', '\u017e': 'z',
    '\u0132': 'IJ', '\u0133': 'ij',
    '\u0152': 'Oe', '\u0153': 'oe',
    '\u0149': "'n", '\u017f': 's'
  };


  var deburrLetter = basePropertyOf(deburredLetters);

  var reLatin = /[\xc0-\xd6\xd8-\xf6\xf8-\xff\u0100-\u017f]/g;


  var rsComboMarksRange$2 = '\\u0300-\\u036f',
      reComboHalfMarksRange$2 = '\\ufe20-\\ufe2f',
      rsComboSymbolsRange$2 = '\\u20d0-\\u20ff',
      rsComboRange$2 = rsComboMarksRange$2 + reComboHalfMarksRange$2 + rsComboSymbolsRange$2;


  var rsCombo$1 = '[' + rsComboRange$2 + ']';


  var reComboMark = RegExp(rsCombo$1, 'g');


  function deburr(string) {
    string = toString(string);
    return string && string.replace(reLatin, deburrLetter).replace(reComboMark, '');
  }

  var reAsciiWord = /[^\x00-\x2f\x3a-\x40\x5b-\x60\x7b-\x7f]+/g;


  function asciiWords(string) {
    return string.match(reAsciiWord) || [];
  }

  var reHasUnicodeWord = /[a-z][A-Z]|[A-Z]{2,}[a-z]|[0-9][a-zA-Z]|[a-zA-Z][0-9]|[^a-zA-Z0-9 ]/;


  function hasUnicodeWord(string) {
    return reHasUnicodeWord.test(string);
  }

  var rsAstralRange$2 = '\\ud800-\\udfff',
      rsComboMarksRange$3 = '\\u0300-\\u036f',
      reComboHalfMarksRange$3 = '\\ufe20-\\ufe2f',
      rsComboSymbolsRange$3 = '\\u20d0-\\u20ff',
      rsComboRange$3 = rsComboMarksRange$3 + reComboHalfMarksRange$3 + rsComboSymbolsRange$3,
      rsDingbatRange = '\\u2700-\\u27bf',
      rsLowerRange = 'a-z\\xdf-\\xf6\\xf8-\\xff',
      rsMathOpRange = '\\xac\\xb1\\xd7\\xf7',
      rsNonCharRange = '\\x00-\\x2f\\x3a-\\x40\\x5b-\\x60\\x7b-\\xbf',
      rsPunctuationRange = '\\u2000-\\u206f',
      rsSpaceRange = ' \\t\\x0b\\f\\xa0\\ufeff\\n\\r\\u2028\\u2029\\u1680\\u180e\\u2000\\u2001\\u2002\\u2003\\u2004\\u2005\\u2006\\u2007\\u2008\\u2009\\u200a\\u202f\\u205f\\u3000',
      rsUpperRange = 'A-Z\\xc0-\\xd6\\xd8-\\xde',
      rsVarRange$2 = '\\ufe0e\\ufe0f',
      rsBreakRange = rsMathOpRange + rsNonCharRange + rsPunctuationRange + rsSpaceRange;


  var rsApos = "['\u2019]",
      rsBreak = '[' + rsBreakRange + ']',
      rsCombo$2 = '[' + rsComboRange$3 + ']',
      rsDigits = '\\d+',
      rsDingbat = '[' + rsDingbatRange + ']',
      rsLower = '[' + rsLowerRange + ']',
      rsMisc = '[^' + rsAstralRange$2 + rsBreakRange + rsDigits + rsDingbatRange + rsLowerRange + rsUpperRange + ']',
      rsFitz$1 = '\\ud83c[\\udffb-\\udfff]',
      rsModifier$1 = '(?:' + rsCombo$2 + '|' + rsFitz$1 + ')',
      rsNonAstral$1 = '[^' + rsAstralRange$2 + ']',
      rsRegional$1 = '(?:\\ud83c[\\udde6-\\uddff]){2}',
      rsSurrPair$1 = '[\\ud800-\\udbff][\\udc00-\\udfff]',
      rsUpper = '[' + rsUpperRange + ']',
      rsZWJ$2 = '\\u200d';


  var rsMiscLower = '(?:' + rsLower + '|' + rsMisc + ')',
      rsMiscUpper = '(?:' + rsUpper + '|' + rsMisc + ')',
      rsOptContrLower = '(?:' + rsApos + '(?:d|ll|m|re|s|t|ve))?',
      rsOptContrUpper = '(?:' + rsApos + '(?:D|LL|M|RE|S|T|VE))?',
      reOptMod$1 = rsModifier$1 + '?',
      rsOptVar$1 = '[' + rsVarRange$2 + ']?',
      rsOptJoin$1 = '(?:' + rsZWJ$2 + '(?:' + [rsNonAstral$1, rsRegional$1, rsSurrPair$1].join('|') + ')' + rsOptVar$1 + reOptMod$1 + ')*',
      rsOrdLower = '\\d*(?:1st|2nd|3rd|(?![123])\\dth)(?=\\b|[A-Z_])',
      rsOrdUpper = '\\d*(?:1ST|2ND|3RD|(?![123])\\dTH)(?=\\b|[a-z_])',
      rsSeq$1 = rsOptVar$1 + reOptMod$1 + rsOptJoin$1,
      rsEmoji = '(?:' + [rsDingbat, rsRegional$1, rsSurrPair$1].join('|') + ')' + rsSeq$1;


  var reUnicodeWord = RegExp([
    rsUpper + '?' + rsLower + '+' + rsOptContrLower + '(?=' + [rsBreak, rsUpper, '$'].join('|') + ')',
    rsMiscUpper + '+' + rsOptContrUpper + '(?=' + [rsBreak, rsUpper + rsMiscLower, '$'].join('|') + ')',
    rsUpper + '?' + rsMiscLower + '+' + rsOptContrLower,
    rsUpper + '+' + rsOptContrUpper,
    rsOrdUpper,
    rsOrdLower,
    rsDigits,
    rsEmoji
  ].join('|'), 'g');


  function unicodeWords(string) {
    return string.match(reUnicodeWord) || [];
  }

  function words(string, pattern, guard) {
    string = toString(string);
    pattern = guard ? undefined : pattern;

    if (pattern === undefined) {
      return hasUnicodeWord(string) ? unicodeWords(string) : asciiWords(string);
    }
    return string.match(pattern) || [];
  }

  var rsApos$1 = "['\u2019]";


  var reApos = RegExp(rsApos$1, 'g');


  function createCompounder(callback) {
    return function(string) {
      return arrayReduce(words(deburr(string).replace(reApos, '')), callback, '');
    };
  }

  var camelCase = createCompounder(function(result, word, index) {
    word = word.toLowerCase();
    return result + (index ? capitalize(word) : word);
  });

  function isObject$1(value) {
    var type = typeof value;
    return value != null && (type == 'object' || type == 'function');
  }

  var now = function() {
    return root.Date.now();
  };

  var NAN = 0 / 0;


  var reTrim = /^\s+|\s+$/g;


  var reIsBadHex = /^[-+]0x[0-9a-f]+$/i;


  var reIsBinary = /^0b[01]+$/i;


  var reIsOctal = /^0o[0-7]+$/i;


  var freeParseInt = parseInt;


  function toNumber(value) {
    if (typeof value == 'number') {
      return value;
    }
    if (isSymbol(value)) {
      return NAN;
    }
    if (isObject$1(value)) {
      var other = typeof value.valueOf == 'function' ? value.valueOf() : value;
      value = isObject$1(other) ? (other + '') : other;
    }
    if (typeof value != 'string') {
      return value === 0 ? value : +value;
    }
    value = value.replace(reTrim, '');
    var isBinary = reIsBinary.test(value);
    return (isBinary || reIsOctal.test(value))
      ? freeParseInt(value.slice(2), isBinary ? 2 : 8)
      : (reIsBadHex.test(value) ? NAN : +value);
  }

  var FUNC_ERROR_TEXT = 'Expected a function';


  var nativeMax = Math.max,
      nativeMin = Math.min;


  function debounce(func, wait, options) {
    var lastArgs,
        lastThis,
        maxWait,
        result,
        timerId,
        lastCallTime,
        lastInvokeTime = 0,
        leading = false,
        maxing = false,
        trailing = true;

    if (typeof func != 'function') {
      throw new TypeError(FUNC_ERROR_TEXT);
    }
    wait = toNumber(wait) || 0;
    if (isObject$1(options)) {
      leading = !!options.leading;
      maxing = 'maxWait' in options;
      maxWait = maxing ? nativeMax(toNumber(options.maxWait) || 0, wait) : maxWait;
      trailing = 'trailing' in options ? !!options.trailing : trailing;
    }

    function invokeFunc(time) {
      var args = lastArgs,
          thisArg = lastThis;

      lastArgs = lastThis = undefined;
      lastInvokeTime = time;
      result = func.apply(thisArg, args);
      return result;
    }

    function leadingEdge(time) {
      
      lastInvokeTime = time;
      
      timerId = setTimeout(timerExpired, wait);
      
      return leading ? invokeFunc(time) : result;
    }

    function remainingWait(time) {
      var timeSinceLastCall = time - lastCallTime,
          timeSinceLastInvoke = time - lastInvokeTime,
          timeWaiting = wait - timeSinceLastCall;

      return maxing
        ? nativeMin(timeWaiting, maxWait - timeSinceLastInvoke)
        : timeWaiting;
    }

    function shouldInvoke(time) {
      var timeSinceLastCall = time - lastCallTime,
          timeSinceLastInvoke = time - lastInvokeTime;

      
      
      
      return (lastCallTime === undefined || (timeSinceLastCall >= wait) ||
        (timeSinceLastCall < 0) || (maxing && timeSinceLastInvoke >= maxWait));
    }

    function timerExpired() {
      var time = now();
      if (shouldInvoke(time)) {
        return trailingEdge(time);
      }
      
      timerId = setTimeout(timerExpired, remainingWait(time));
    }

    function trailingEdge(time) {
      timerId = undefined;

      
      
      if (trailing && lastArgs) {
        return invokeFunc(time);
      }
      lastArgs = lastThis = undefined;
      return result;
    }

    function cancel() {
      if (timerId !== undefined) {
        clearTimeout(timerId);
      }
      lastInvokeTime = 0;
      lastArgs = lastCallTime = lastThis = timerId = undefined;
    }

    function flush() {
      return timerId === undefined ? result : trailingEdge(now());
    }

    function debounced() {
      var time = now(),
          isInvoking = shouldInvoke(time);

      lastArgs = arguments;
      lastThis = this;
      lastCallTime = time;

      if (isInvoking) {
        if (timerId === undefined) {
          return leadingEdge(lastCallTime);
        }
        if (maxing) {
          
          timerId = setTimeout(timerExpired, wait);
          return invokeFunc(lastCallTime);
        }
      }
      if (timerId === undefined) {
        timerId = setTimeout(timerExpired, wait);
      }
      return result;
    }
    debounced.cancel = cancel;
    debounced.flush = flush;
    return debounced;
  }

  var reIsDeepProp = /\.|\[(?:[^[\]]*|(["'])(?:(?!\1)[^\\]|\\.)*?\1)\]/,
      reIsPlainProp = /^\w*$/;


  function isKey(value, object) {
    if (isArray$1(value)) {
      return false;
    }
    var type = typeof value;
    if (type == 'number' || type == 'symbol' || type == 'boolean' ||
        value == null || isSymbol(value)) {
      return true;
    }
    return reIsPlainProp.test(value) || !reIsDeepProp.test(value) ||
      (object != null && value in Object(object));
  }

  var asyncTag = '[object AsyncFunction]',
      funcTag = '[object Function]',
      genTag = '[object GeneratorFunction]',
      proxyTag = '[object Proxy]';


  function isFunction(value) {
    if (!isObject$1(value)) {
      return false;
    }
    
    
    var tag = baseGetTag(value);
    return tag == funcTag || tag == genTag || tag == asyncTag || tag == proxyTag;
  }

  var coreJsData = root['__core-js_shared__'];

  var maskSrcKey = (function() {
    var uid = /[^.]+$/.exec(coreJsData && coreJsData.keys && coreJsData.keys.IE_PROTO || '');
    return uid ? ('Symbol(src)_1.' + uid) : '';
  }());


  function isMasked(func) {
    return !!maskSrcKey && (maskSrcKey in func);
  }

  var funcProto = Function.prototype;


  var funcToString = funcProto.toString;


  function toSource(func) {
    if (func != null) {
      try {
        return funcToString.call(func);
      } catch (e) {}
      try {
        return (func + '');
      } catch (e) {}
    }
    return '';
  }

  var reRegExpChar = /[\\^$.*+?()[\]{}|]/g;


  var reIsHostCtor = /^\[object .+?Constructor\]$/;


  var funcProto$1 = Function.prototype,
      objectProto$2 = Object.prototype;


  var funcToString$1 = funcProto$1.toString;


  var hasOwnProperty$1 = objectProto$2.hasOwnProperty;


  var reIsNative = RegExp('^' +
    funcToString$1.call(hasOwnProperty$1).replace(reRegExpChar, '\\$&')
    .replace(/hasOwnProperty|(function).*?(?=\\\()| for .+?(?=\\\])/g, '$1.*?') + '$'
  );


  function baseIsNative(value) {
    if (!isObject$1(value) || isMasked(value)) {
      return false;
    }
    var pattern = isFunction(value) ? reIsNative : reIsHostCtor;
    return pattern.test(toSource(value));
  }

  function getValue(object, key) {
    return object == null ? undefined : object[key];
  }

  function getNative(object, key) {
    var value = getValue(object, key);
    return baseIsNative(value) ? value : undefined;
  }

  var nativeCreate = getNative(Object, 'create');

  function hashClear() {
    this.__data__ = nativeCreate ? nativeCreate(null) : {};
    this.size = 0;
  }

  function hashDelete(key) {
    var result = this.has(key) && delete this.__data__[key];
    this.size -= result ? 1 : 0;
    return result;
  }

  var HASH_UNDEFINED = '__lodash_hash_undefined__';


  var objectProto$3 = Object.prototype;


  var hasOwnProperty$2 = objectProto$3.hasOwnProperty;


  function hashGet(key) {
    var data = this.__data__;
    if (nativeCreate) {
      var result = data[key];
      return result === HASH_UNDEFINED ? undefined : result;
    }
    return hasOwnProperty$2.call(data, key) ? data[key] : undefined;
  }

  var objectProto$4 = Object.prototype;


  var hasOwnProperty$3 = objectProto$4.hasOwnProperty;


  function hashHas(key) {
    var data = this.__data__;
    return nativeCreate ? (data[key] !== undefined) : hasOwnProperty$3.call(data, key);
  }

  var HASH_UNDEFINED$1 = '__lodash_hash_undefined__';


  function hashSet(key, value) {
    var data = this.__data__;
    this.size += this.has(key) ? 0 : 1;
    data[key] = (nativeCreate && value === undefined) ? HASH_UNDEFINED$1 : value;
    return this;
  }

  function Hash(entries) {
    var this$1 = this;

    var index = -1,
        length = entries == null ? 0 : entries.length;

    this.clear();
    while (++index < length) {
      var entry = entries[index];
      this$1.set(entry[0], entry[1]);
    }
  }


  Hash.prototype.clear = hashClear;
  Hash.prototype['delete'] = hashDelete;
  Hash.prototype.get = hashGet;
  Hash.prototype.has = hashHas;
  Hash.prototype.set = hashSet;

  function listCacheClear() {
    this.__data__ = [];
    this.size = 0;
  }

  function eq(value, other) {
    return value === other || (value !== value && other !== other);
  }

  function assocIndexOf(array, key) {
    var length = array.length;
    while (length--) {
      if (eq(array[length][0], key)) {
        return length;
      }
    }
    return -1;
  }

  var arrayProto = Array.prototype;


  var splice = arrayProto.splice;


  function listCacheDelete(key) {
    var data = this.__data__,
        index = assocIndexOf(data, key);

    if (index < 0) {
      return false;
    }
    var lastIndex = data.length - 1;
    if (index == lastIndex) {
      data.pop();
    } else {
      splice.call(data, index, 1);
    }
    --this.size;
    return true;
  }

  function listCacheGet(key) {
    var data = this.__data__,
        index = assocIndexOf(data, key);

    return index < 0 ? undefined : data[index][1];
  }

  function listCacheHas(key) {
    return assocIndexOf(this.__data__, key) > -1;
  }

  function listCacheSet(key, value) {
    var data = this.__data__,
        index = assocIndexOf(data, key);

    if (index < 0) {
      ++this.size;
      data.push([key, value]);
    } else {
      data[index][1] = value;
    }
    return this;
  }

  function ListCache(entries) {
    var this$1 = this;

    var index = -1,
        length = entries == null ? 0 : entries.length;

    this.clear();
    while (++index < length) {
      var entry = entries[index];
      this$1.set(entry[0], entry[1]);
    }
  }


  ListCache.prototype.clear = listCacheClear;
  ListCache.prototype['delete'] = listCacheDelete;
  ListCache.prototype.get = listCacheGet;
  ListCache.prototype.has = listCacheHas;
  ListCache.prototype.set = listCacheSet;

  var Map$1 = getNative(root, 'Map');

  function mapCacheClear() {
    this.size = 0;
    this.__data__ = {
      'hash': new Hash,
      'map': new (Map$1 || ListCache),
      'string': new Hash
    };
  }

  function isKeyable(value) {
    var type = typeof value;
    return (type == 'string' || type == 'number' || type == 'symbol' || type == 'boolean')
      ? (value !== '__proto__')
      : (value === null);
  }

  function getMapData(map, key) {
    var data = map.__data__;
    return isKeyable(key)
      ? data[typeof key == 'string' ? 'string' : 'hash']
      : data.map;
  }

  function mapCacheDelete(key) {
    var result = getMapData(this, key)['delete'](key);
    this.size -= result ? 1 : 0;
    return result;
  }

  function mapCacheGet(key) {
    return getMapData(this, key).get(key);
  }

  function mapCacheHas(key) {
    return getMapData(this, key).has(key);
  }

  function mapCacheSet(key, value) {
    var data = getMapData(this, key),
        size = data.size;

    data.set(key, value);
    this.size += data.size == size ? 0 : 1;
    return this;
  }

  function MapCache(entries) {
    var this$1 = this;

    var index = -1,
        length = entries == null ? 0 : entries.length;

    this.clear();
    while (++index < length) {
      var entry = entries[index];
      this$1.set(entry[0], entry[1]);
    }
  }


  MapCache.prototype.clear = mapCacheClear;
  MapCache.prototype['delete'] = mapCacheDelete;
  MapCache.prototype.get = mapCacheGet;
  MapCache.prototype.has = mapCacheHas;
  MapCache.prototype.set = mapCacheSet;

  var FUNC_ERROR_TEXT$1 = 'Expected a function';


  function memoize(func, resolver) {
    if (typeof func != 'function' || (resolver != null && typeof resolver != 'function')) {
      throw new TypeError(FUNC_ERROR_TEXT$1);
    }
    var memoized = function() {
      var args = arguments,
          key = resolver ? resolver.apply(this, args) : args[0],
          cache = memoized.cache;

      if (cache.has(key)) {
        return cache.get(key);
      }
      var result = func.apply(this, args);
      memoized.cache = cache.set(key, result) || cache;
      return result;
    };
    memoized.cache = new (memoize.Cache || MapCache);
    return memoized;
  }


  memoize.Cache = MapCache;

  var MAX_MEMOIZE_SIZE = 500;


  function memoizeCapped(func) {
    var result = memoize(func, function(key) {
      if (cache.size === MAX_MEMOIZE_SIZE) {
        cache.clear();
      }
      return key;
    });

    var cache = result.cache;
    return result;
  }

  var rePropName = /[^.[\]]+|\[(?:(-?\d+(?:\.\d+)?)|(["'])((?:(?!\2)[^\\]|\\.)*?)\2)\]|(?=(?:\.|\[\])(?:\.|\[\]|$))/g;


  var reEscapeChar = /\\(\\)?/g;


  var stringToPath = memoizeCapped(function(string) {
    var result = [];
    if (string.charCodeAt(0) === 46 ) {
      result.push('');
    }
    string.replace(rePropName, function(match, number, quote, subString) {
      result.push(quote ? subString.replace(reEscapeChar, '$1') : (number || match));
    });
    return result;
  });

  function castPath(value, object) {
    if (isArray$1(value)) {
      return value;
    }
    return isKey(value, object) ? [value] : stringToPath(toString(value));
  }

  var INFINITY$1 = 1 / 0;


  function toKey(value) {
    if (typeof value == 'string' || isSymbol(value)) {
      return value;
    }
    var result = (value + '');
    return (result == '0' && (1 / value) == -INFINITY$1) ? '-0' : result;
  }

  function baseGet(object, path) {
    path = castPath(path, object);

    var index = 0,
        length = path.length;

    while (object != null && index < length) {
      object = object[toKey(path[index++])];
    }
    return (index && index == length) ? object : undefined;
  }

  function get(object, path, defaultValue) {
    var result = object == null ? undefined : baseGet(object, path);
    return result === undefined ? defaultValue : result;
  }

  function stackClear() {
    this.__data__ = new ListCache;
    this.size = 0;
  }

  function stackDelete(key) {
    var data = this.__data__,
        result = data['delete'](key);

    this.size = data.size;
    return result;
  }

  function stackGet(key) {
    return this.__data__.get(key);
  }

  function stackHas(key) {
    return this.__data__.has(key);
  }

  var LARGE_ARRAY_SIZE = 200;


  function stackSet(key, value) {
    var data = this.__data__;
    if (data instanceof ListCache) {
      var pairs = data.__data__;
      if (!Map$1 || (pairs.length < LARGE_ARRAY_SIZE - 1)) {
        pairs.push([key, value]);
        this.size = ++data.size;
        return this;
      }
      data = this.__data__ = new MapCache(pairs);
    }
    data.set(key, value);
    this.size = data.size;
    return this;
  }

  function Stack(entries) {
    var data = this.__data__ = new ListCache(entries);
    this.size = data.size;
  }


  Stack.prototype.clear = stackClear;
  Stack.prototype['delete'] = stackDelete;
  Stack.prototype.get = stackGet;
  Stack.prototype.has = stackHas;
  Stack.prototype.set = stackSet;

  var HASH_UNDEFINED$2 = '__lodash_hash_undefined__';


  function setCacheAdd(value) {
    this.__data__.set(value, HASH_UNDEFINED$2);
    return this;
  }

  function setCacheHas(value) {
    return this.__data__.has(value);
  }

  function SetCache(values) {
    var this$1 = this;

    var index = -1,
        length = values == null ? 0 : values.length;

    this.__data__ = new MapCache;
    while (++index < length) {
      this$1.add(values[index]);
    }
  }


  SetCache.prototype.add = SetCache.prototype.push = setCacheAdd;
  SetCache.prototype.has = setCacheHas;

  function arraySome(array, predicate) {
    var index = -1,
        length = array == null ? 0 : array.length;

    while (++index < length) {
      if (predicate(array[index], index, array)) {
        return true;
      }
    }
    return false;
  }

  function cacheHas(cache, key) {
    return cache.has(key);
  }

  var COMPARE_PARTIAL_FLAG = 1,
      COMPARE_UNORDERED_FLAG = 2;


  function equalArrays(array, other, bitmask, customizer, equalFunc, stack) {
    var isPartial = bitmask & COMPARE_PARTIAL_FLAG,
        arrLength = array.length,
        othLength = other.length;

    if (arrLength != othLength && !(isPartial && othLength > arrLength)) {
      return false;
    }
    
    var stacked = stack.get(array);
    if (stacked && stack.get(other)) {
      return stacked == other;
    }
    var index = -1,
        result = true,
        seen = (bitmask & COMPARE_UNORDERED_FLAG) ? new SetCache : undefined;

    stack.set(array, other);
    stack.set(other, array);

    
    while (++index < arrLength) {
      var arrValue = array[index],
          othValue = other[index];

      if (customizer) {
        var compared = isPartial
          ? customizer(othValue, arrValue, index, other, array, stack)
          : customizer(arrValue, othValue, index, array, other, stack);
      }
      if (compared !== undefined) {
        if (compared) {
          continue;
        }
        result = false;
        break;
      }
      
      if (seen) {
        if (!arraySome(other, function(othValue, othIndex) {
              if (!cacheHas(seen, othIndex) &&
                  (arrValue === othValue || equalFunc(arrValue, othValue, bitmask, customizer, stack))) {
                return seen.push(othIndex);
              }
            })) {
          result = false;
          break;
        }
      } else if (!(
            arrValue === othValue ||
              equalFunc(arrValue, othValue, bitmask, customizer, stack)
          )) {
        result = false;
        break;
      }
    }
    stack['delete'](array);
    stack['delete'](other);
    return result;
  }

  var Uint8Array = root.Uint8Array;

  function mapToArray(map) {
    var index = -1,
        result = Array(map.size);

    map.forEach(function(value, key) {
      result[++index] = [key, value];
    });
    return result;
  }

  function setToArray(set) {
    var index = -1,
        result = Array(set.size);

    set.forEach(function(value) {
      result[++index] = value;
    });
    return result;
  }

  var COMPARE_PARTIAL_FLAG$1 = 1,
      COMPARE_UNORDERED_FLAG$1 = 2;


  var boolTag = '[object Boolean]',
      dateTag = '[object Date]',
      errorTag = '[object Error]',
      mapTag = '[object Map]',
      numberTag = '[object Number]',
      regexpTag = '[object RegExp]',
      setTag = '[object Set]',
      stringTag = '[object String]',
      symbolTag$1 = '[object Symbol]';

  var arrayBufferTag = '[object ArrayBuffer]',
      dataViewTag = '[object DataView]';


  var symbolProto$1 = Symbol ? Symbol.prototype : undefined,
      symbolValueOf = symbolProto$1 ? symbolProto$1.valueOf : undefined;


  function equalByTag(object, other, tag, bitmask, customizer, equalFunc, stack) {
    switch (tag) {
      case dataViewTag:
        if ((object.byteLength != other.byteLength) ||
            (object.byteOffset != other.byteOffset)) {
          return false;
        }
        object = object.buffer;
        other = other.buffer;

      case arrayBufferTag:
        if ((object.byteLength != other.byteLength) ||
            !equalFunc(new Uint8Array(object), new Uint8Array(other))) {
          return false;
        }
        return true;

      case boolTag:
      case dateTag:
      case numberTag:
        
        
        return eq(+object, +other);

      case errorTag:
        return object.name == other.name && object.message == other.message;

      case regexpTag:
      case stringTag:
        
        
        
        return object == (other + '');

      case mapTag:
        var convert = mapToArray;

      case setTag:
        var isPartial = bitmask & COMPARE_PARTIAL_FLAG$1;
        convert || (convert = setToArray);

        if (object.size != other.size && !isPartial) {
          return false;
        }
        
        var stacked = stack.get(object);
        if (stacked) {
          return stacked == other;
        }
        bitmask |= COMPARE_UNORDERED_FLAG$1;

        
        stack.set(object, other);
        var result = equalArrays(convert(object), convert(other), bitmask, customizer, equalFunc, stack);
        stack['delete'](object);
        return result;

      case symbolTag$1:
        if (symbolValueOf) {
          return symbolValueOf.call(object) == symbolValueOf.call(other);
        }
    }
    return false;
  }

  function arrayPush(array, values) {
    var index = -1,
        length = values.length,
        offset = array.length;

    while (++index < length) {
      array[offset + index] = values[index];
    }
    return array;
  }

  function baseGetAllKeys(object, keysFunc, symbolsFunc) {
    var result = keysFunc(object);
    return isArray$1(object) ? result : arrayPush(result, symbolsFunc(object));
  }

  function arrayFilter(array, predicate) {
    var index = -1,
        length = array == null ? 0 : array.length,
        resIndex = 0,
        result = [];

    while (++index < length) {
      var value = array[index];
      if (predicate(value, index, array)) {
        result[resIndex++] = value;
      }
    }
    return result;
  }

  function stubArray() {
    return [];
  }

  var objectProto$5 = Object.prototype;


  var propertyIsEnumerable = objectProto$5.propertyIsEnumerable;


  var nativeGetSymbols = Object.getOwnPropertySymbols;


  var getSymbols = !nativeGetSymbols ? stubArray : function(object) {
    if (object == null) {
      return [];
    }
    object = Object(object);
    return arrayFilter(nativeGetSymbols(object), function(symbol) {
      return propertyIsEnumerable.call(object, symbol);
    });
  };

  function baseTimes(n, iteratee) {
    var index = -1,
        result = Array(n);

    while (++index < n) {
      result[index] = iteratee(index);
    }
    return result;
  }

  var argsTag = '[object Arguments]';


  function baseIsArguments(value) {
    return isObjectLike(value) && baseGetTag(value) == argsTag;
  }

  var objectProto$6 = Object.prototype;


  var hasOwnProperty$4 = objectProto$6.hasOwnProperty;


  var propertyIsEnumerable$1 = objectProto$6.propertyIsEnumerable;


  var isArguments = baseIsArguments(function() { return arguments; }()) ? baseIsArguments : function(value) {
    return isObjectLike(value) && hasOwnProperty$4.call(value, 'callee') &&
      !propertyIsEnumerable$1.call(value, 'callee');
  };

  function stubFalse() {
    return false;
  }

  var freeExports = typeof exports == 'object' && exports && !exports.nodeType && exports;


  var freeModule = freeExports && typeof module == 'object' && module && !module.nodeType && module;


  var moduleExports = freeModule && freeModule.exports === freeExports;


  var Buffer = moduleExports ? root.Buffer : undefined;


  var nativeIsBuffer = Buffer ? Buffer.isBuffer : undefined;


  var isBuffer = nativeIsBuffer || stubFalse;

  var MAX_SAFE_INTEGER = 9007199254740991;


  var reIsUint = /^(?:0|[1-9]\d*)$/;


  function isIndex(value, length) {
    var type = typeof value;
    length = length == null ? MAX_SAFE_INTEGER : length;

    return !!length &&
      (type == 'number' ||
        (type != 'symbol' && reIsUint.test(value))) &&
          (value > -1 && value % 1 == 0 && value < length);
  }

  var MAX_SAFE_INTEGER$1 = 9007199254740991;


  function isLength(value) {
    return typeof value == 'number' &&
      value > -1 && value % 1 == 0 && value <= MAX_SAFE_INTEGER$1;
  }

  var argsTag$1 = '[object Arguments]',
      arrayTag = '[object Array]',
      boolTag$1 = '[object Boolean]',
      dateTag$1 = '[object Date]',
      errorTag$1 = '[object Error]',
      funcTag$1 = '[object Function]',
      mapTag$1 = '[object Map]',
      numberTag$1 = '[object Number]',
      objectTag = '[object Object]',
      regexpTag$1 = '[object RegExp]',
      setTag$1 = '[object Set]',
      stringTag$1 = '[object String]',
      weakMapTag = '[object WeakMap]';

  var arrayBufferTag$1 = '[object ArrayBuffer]',
      dataViewTag$1 = '[object DataView]',
      float32Tag = '[object Float32Array]',
      float64Tag = '[object Float64Array]',
      int8Tag = '[object Int8Array]',
      int16Tag = '[object Int16Array]',
      int32Tag = '[object Int32Array]',
      uint8Tag = '[object Uint8Array]',
      uint8ClampedTag = '[object Uint8ClampedArray]',
      uint16Tag = '[object Uint16Array]',
      uint32Tag = '[object Uint32Array]';


  var typedArrayTags = {};
  typedArrayTags[float32Tag] = typedArrayTags[float64Tag] =
  typedArrayTags[int8Tag] = typedArrayTags[int16Tag] =
  typedArrayTags[int32Tag] = typedArrayTags[uint8Tag] =
  typedArrayTags[uint8ClampedTag] = typedArrayTags[uint16Tag] =
  typedArrayTags[uint32Tag] = true;
  typedArrayTags[argsTag$1] = typedArrayTags[arrayTag] =
  typedArrayTags[arrayBufferTag$1] = typedArrayTags[boolTag$1] =
  typedArrayTags[dataViewTag$1] = typedArrayTags[dateTag$1] =
  typedArrayTags[errorTag$1] = typedArrayTags[funcTag$1] =
  typedArrayTags[mapTag$1] = typedArrayTags[numberTag$1] =
  typedArrayTags[objectTag] = typedArrayTags[regexpTag$1] =
  typedArrayTags[setTag$1] = typedArrayTags[stringTag$1] =
  typedArrayTags[weakMapTag] = false;


  function baseIsTypedArray(value) {
    return isObjectLike(value) &&
      isLength(value.length) && !!typedArrayTags[baseGetTag(value)];
  }

  function baseUnary(func) {
    return function(value) {
      return func(value);
    };
  }

  var freeExports$1 = typeof exports == 'object' && exports && !exports.nodeType && exports;


  var freeModule$1 = freeExports$1 && typeof module == 'object' && module && !module.nodeType && module;


  var moduleExports$1 = freeModule$1 && freeModule$1.exports === freeExports$1;


  var freeProcess = moduleExports$1 && freeGlobal.process;


  var nodeUtil = (function() {
    try {
      
      var types = freeModule$1 && freeModule$1.require && freeModule$1.require('util').types;

      if (types) {
        return types;
      }

      
      return freeProcess && freeProcess.binding && freeProcess.binding('util');
    } catch (e) {}
  }());

  var nodeIsTypedArray = nodeUtil && nodeUtil.isTypedArray;


  var isTypedArray = nodeIsTypedArray ? baseUnary(nodeIsTypedArray) : baseIsTypedArray;

  var objectProto$7 = Object.prototype;


  var hasOwnProperty$5 = objectProto$7.hasOwnProperty;


  function arrayLikeKeys(value, inherited) {
    var isArr = isArray$1(value),
        isArg = !isArr && isArguments(value),
        isBuff = !isArr && !isArg && isBuffer(value),
        isType = !isArr && !isArg && !isBuff && isTypedArray(value),
        skipIndexes = isArr || isArg || isBuff || isType,
        result = skipIndexes ? baseTimes(value.length, String) : [],
        length = result.length;

    for (var key in value) {
      if ((inherited || hasOwnProperty$5.call(value, key)) &&
          !(skipIndexes && (
             
             key == 'length' ||
             
             (isBuff && (key == 'offset' || key == 'parent')) ||
             
             (isType && (key == 'buffer' || key == 'byteLength' || key == 'byteOffset')) ||
             
             isIndex(key, length)
          ))) {
        result.push(key);
      }
    }
    return result;
  }

  var objectProto$8 = Object.prototype;


  function isPrototype(value) {
    var Ctor = value && value.constructor,
        proto = (typeof Ctor == 'function' && Ctor.prototype) || objectProto$8;

    return value === proto;
  }

  function overArg(func, transform) {
    return function(arg) {
      return func(transform(arg));
    };
  }

  var nativeKeys = overArg(Object.keys, Object);

  var objectProto$9 = Object.prototype;


  var hasOwnProperty$6 = objectProto$9.hasOwnProperty;


  function baseKeys(object) {
    if (!isPrototype(object)) {
      return nativeKeys(object);
    }
    var result = [];
    for (var key in Object(object)) {
      if (hasOwnProperty$6.call(object, key) && key != 'constructor') {
        result.push(key);
      }
    }
    return result;
  }

  function isArrayLike(value) {
    return value != null && isLength(value.length) && !isFunction(value);
  }

  function keys(object) {
    return isArrayLike(object) ? arrayLikeKeys(object) : baseKeys(object);
  }

  function getAllKeys(object) {
    return baseGetAllKeys(object, keys, getSymbols);
  }

  var COMPARE_PARTIAL_FLAG$2 = 1;


  var objectProto$10 = Object.prototype;


  var hasOwnProperty$7 = objectProto$10.hasOwnProperty;


  function equalObjects(object, other, bitmask, customizer, equalFunc, stack) {
    var isPartial = bitmask & COMPARE_PARTIAL_FLAG$2,
        objProps = getAllKeys(object),
        objLength = objProps.length,
        othProps = getAllKeys(other),
        othLength = othProps.length;

    if (objLength != othLength && !isPartial) {
      return false;
    }
    var index = objLength;
    while (index--) {
      var key = objProps[index];
      if (!(isPartial ? key in other : hasOwnProperty$7.call(other, key))) {
        return false;
      }
    }
    
    var stacked = stack.get(object);
    if (stacked && stack.get(other)) {
      return stacked == other;
    }
    var result = true;
    stack.set(object, other);
    stack.set(other, object);

    var skipCtor = isPartial;
    while (++index < objLength) {
      key = objProps[index];
      var objValue = object[key],
          othValue = other[key];

      if (customizer) {
        var compared = isPartial
          ? customizer(othValue, objValue, key, other, object, stack)
          : customizer(objValue, othValue, key, object, other, stack);
      }
      
      if (!(compared === undefined
            ? (objValue === othValue || equalFunc(objValue, othValue, bitmask, customizer, stack))
            : compared
          )) {
        result = false;
        break;
      }
      skipCtor || (skipCtor = key == 'constructor');
    }
    if (result && !skipCtor) {
      var objCtor = object.constructor,
          othCtor = other.constructor;

      
      if (objCtor != othCtor &&
          ('constructor' in object && 'constructor' in other) &&
          !(typeof objCtor == 'function' && objCtor instanceof objCtor &&
            typeof othCtor == 'function' && othCtor instanceof othCtor)) {
        result = false;
      }
    }
    stack['delete'](object);
    stack['delete'](other);
    return result;
  }

  var DataView = getNative(root, 'DataView');

  var Promise$1 = getNative(root, 'Promise');

  var Set$1 = getNative(root, 'Set');

  var WeakMap$1 = getNative(root, 'WeakMap');

  var mapTag$2 = '[object Map]',
      objectTag$1 = '[object Object]',
      promiseTag = '[object Promise]',
      setTag$2 = '[object Set]',
      weakMapTag$1 = '[object WeakMap]';

  var dataViewTag$2 = '[object DataView]';


  var dataViewCtorString = toSource(DataView),
      mapCtorString = toSource(Map$1),
      promiseCtorString = toSource(Promise$1),
      setCtorString = toSource(Set$1),
      weakMapCtorString = toSource(WeakMap$1);


  var getTag = baseGetTag;


  if ((DataView && getTag(new DataView(new ArrayBuffer(1))) != dataViewTag$2) ||
      (Map$1 && getTag(new Map$1) != mapTag$2) ||
      (Promise$1 && getTag(Promise$1.resolve()) != promiseTag) ||
      (Set$1 && getTag(new Set$1) != setTag$2) ||
      (WeakMap$1 && getTag(new WeakMap$1) != weakMapTag$1)) {
    getTag = function(value) {
      var result = baseGetTag(value),
          Ctor = result == objectTag$1 ? value.constructor : undefined,
          ctorString = Ctor ? toSource(Ctor) : '';

      if (ctorString) {
        switch (ctorString) {
          case dataViewCtorString: return dataViewTag$2;
          case mapCtorString: return mapTag$2;
          case promiseCtorString: return promiseTag;
          case setCtorString: return setTag$2;
          case weakMapCtorString: return weakMapTag$1;
        }
      }
      return result;
    };
  }

  var getTag$1 = getTag;

  var COMPARE_PARTIAL_FLAG$3 = 1;


  var argsTag$2 = '[object Arguments]',
      arrayTag$1 = '[object Array]',
      objectTag$2 = '[object Object]';


  var objectProto$11 = Object.prototype;


  var hasOwnProperty$8 = objectProto$11.hasOwnProperty;


  function baseIsEqualDeep(object, other, bitmask, customizer, equalFunc, stack) {
    var objIsArr = isArray$1(object),
        othIsArr = isArray$1(other),
        objTag = objIsArr ? arrayTag$1 : getTag$1(object),
        othTag = othIsArr ? arrayTag$1 : getTag$1(other);

    objTag = objTag == argsTag$2 ? objectTag$2 : objTag;
    othTag = othTag == argsTag$2 ? objectTag$2 : othTag;

    var objIsObj = objTag == objectTag$2,
        othIsObj = othTag == objectTag$2,
        isSameTag = objTag == othTag;

    if (isSameTag && isBuffer(object)) {
      if (!isBuffer(other)) {
        return false;
      }
      objIsArr = true;
      objIsObj = false;
    }
    if (isSameTag && !objIsObj) {
      stack || (stack = new Stack);
      return (objIsArr || isTypedArray(object))
        ? equalArrays(object, other, bitmask, customizer, equalFunc, stack)
        : equalByTag(object, other, objTag, bitmask, customizer, equalFunc, stack);
    }
    if (!(bitmask & COMPARE_PARTIAL_FLAG$3)) {
      var objIsWrapped = objIsObj && hasOwnProperty$8.call(object, '__wrapped__'),
          othIsWrapped = othIsObj && hasOwnProperty$8.call(other, '__wrapped__');

      if (objIsWrapped || othIsWrapped) {
        var objUnwrapped = objIsWrapped ? object.value() : object,
            othUnwrapped = othIsWrapped ? other.value() : other;

        stack || (stack = new Stack);
        return equalFunc(objUnwrapped, othUnwrapped, bitmask, customizer, stack);
      }
    }
    if (!isSameTag) {
      return false;
    }
    stack || (stack = new Stack);
    return equalObjects(object, other, bitmask, customizer, equalFunc, stack);
  }

  function baseIsEqual(value, other, bitmask, customizer, stack) {
    if (value === other) {
      return true;
    }
    if (value == null || other == null || (!isObjectLike(value) && !isObjectLike(other))) {
      return value !== value && other !== other;
    }
    return baseIsEqualDeep(value, other, bitmask, customizer, baseIsEqual, stack);
  }

  var COMPARE_PARTIAL_FLAG$4 = 1,
      COMPARE_UNORDERED_FLAG$2 = 2;


  function baseIsMatch(object, source, matchData, customizer) {
    var index = matchData.length,
        length = index,
        noCustomizer = !customizer;

    if (object == null) {
      return !length;
    }
    object = Object(object);
    while (index--) {
      var data = matchData[index];
      if ((noCustomizer && data[2])
            ? data[1] !== object[data[0]]
            : !(data[0] in object)
          ) {
        return false;
      }
    }
    while (++index < length) {
      data = matchData[index];
      var key = data[0],
          objValue = object[key],
          srcValue = data[1];

      if (noCustomizer && data[2]) {
        if (objValue === undefined && !(key in object)) {
          return false;
        }
      } else {
        var stack = new Stack;
        if (customizer) {
          var result = customizer(objValue, srcValue, key, object, source, stack);
        }
        if (!(result === undefined
              ? baseIsEqual(srcValue, objValue, COMPARE_PARTIAL_FLAG$4 | COMPARE_UNORDERED_FLAG$2, customizer, stack)
              : result
            )) {
          return false;
        }
      }
    }
    return true;
  }

  function isStrictComparable(value) {
    return value === value && !isObject$1(value);
  }

  function getMatchData(object) {
    var result = keys(object),
        length = result.length;

    while (length--) {
      var key = result[length],
          value = object[key];

      result[length] = [key, value, isStrictComparable(value)];
    }
    return result;
  }

  function matchesStrictComparable(key, srcValue) {
    return function(object) {
      if (object == null) {
        return false;
      }
      return object[key] === srcValue &&
        (srcValue !== undefined || (key in Object(object)));
    };
  }

  function baseMatches(source) {
    var matchData = getMatchData(source);
    if (matchData.length == 1 && matchData[0][2]) {
      return matchesStrictComparable(matchData[0][0], matchData[0][1]);
    }
    return function(object) {
      return object === source || baseIsMatch(object, source, matchData);
    };
  }

  function baseHasIn(object, key) {
    return object != null && key in Object(object);
  }

  function hasPath(object, path, hasFunc) {
    path = castPath(path, object);

    var index = -1,
        length = path.length,
        result = false;

    while (++index < length) {
      var key = toKey(path[index]);
      if (!(result = object != null && hasFunc(object, key))) {
        break;
      }
      object = object[key];
    }
    if (result || ++index != length) {
      return result;
    }
    length = object == null ? 0 : object.length;
    return !!length && isLength(length) && isIndex(key, length) &&
      (isArray$1(object) || isArguments(object));
  }

  function hasIn(object, path) {
    return object != null && hasPath(object, path, baseHasIn);
  }

  var COMPARE_PARTIAL_FLAG$5 = 1,
      COMPARE_UNORDERED_FLAG$3 = 2;


  function baseMatchesProperty(path, srcValue) {
    if (isKey(path) && isStrictComparable(srcValue)) {
      return matchesStrictComparable(toKey(path), srcValue);
    }
    return function(object) {
      var objValue = get(object, path);
      return (objValue === undefined && objValue === srcValue)
        ? hasIn(object, path)
        : baseIsEqual(srcValue, objValue, COMPARE_PARTIAL_FLAG$5 | COMPARE_UNORDERED_FLAG$3);
    };
  }

  function identity(value) {
    return value;
  }

  function baseProperty(key) {
    return function(object) {
      return object == null ? undefined : object[key];
    };
  }

  function basePropertyDeep(path) {
    return function(object) {
      return baseGet(object, path);
    };
  }

  function property(path) {
    return isKey(path) ? baseProperty(toKey(path)) : basePropertyDeep(path);
  }

  function baseIteratee(value) {
    
    
    if (typeof value == 'function') {
      return value;
    }
    if (value == null) {
      return identity;
    }
    if (typeof value == 'object') {
      return isArray$1(value)
        ? baseMatchesProperty(value[0], value[1])
        : baseMatches(value);
    }
    return property(value);
  }

  function createFind(findIndexFunc) {
    return function(collection, predicate, fromIndex) {
      var iterable = Object(collection);
      if (!isArrayLike(collection)) {
        var iteratee = baseIteratee(predicate, 3);
        collection = keys(collection);
        predicate = function(key) { return iteratee(iterable[key], key, iterable); };
      }
      var index = findIndexFunc(collection, predicate, fromIndex);
      return index > -1 ? iterable[iteratee ? collection[index] : index] : undefined;
    };
  }

  function baseFindIndex(array, predicate, fromIndex, fromRight) {
    var length = array.length,
        index = fromIndex + (fromRight ? 1 : -1);

    while ((fromRight ? index-- : ++index < length)) {
      if (predicate(array[index], index, array)) {
        return index;
      }
    }
    return -1;
  }

  var INFINITY$2 = 1 / 0,
      MAX_INTEGER = 1.7976931348623157e+308;


  function toFinite(value) {
    if (!value) {
      return value === 0 ? value : 0;
    }
    value = toNumber(value);
    if (value === INFINITY$2 || value === -INFINITY$2) {
      var sign = (value < 0 ? -1 : 1);
      return sign * MAX_INTEGER;
    }
    return value === value ? value : 0;
  }

  function toInteger(value) {
    var result = toFinite(value),
        remainder = result % 1;

    return result === result ? (remainder ? result - remainder : result) : 0;
  }

  var nativeMax$1 = Math.max;


  function findIndex(array, predicate, fromIndex) {
    var length = array == null ? 0 : array.length;
    if (!length) {
      return -1;
    }
    var index = fromIndex == null ? 0 : toInteger(fromIndex);
    if (index < 0) {
      index = nativeMax$1(length + index, 0);
    }
    return baseFindIndex(array, baseIteratee(predicate, 3), index);
  }

  var find = createFind(findIndex);

  function isMatch(object, source) {
    return object === source || baseIsMatch(object, source, getMatchData(source));
  }

  var defineProperty = (function() {
    try {
      var func = getNative(Object, 'defineProperty');
      func({}, '', {});
      return func;
    } catch (e) {}
  }());

  function baseAssignValue(object, key, value) {
    if (key == '__proto__' && defineProperty) {
      defineProperty(object, key, {
        'configurable': true,
        'enumerable': true,
        'value': value,
        'writable': true
      });
    } else {
      object[key] = value;
    }
  }

  function assignMergeValue(object, key, value) {
    if ((value !== undefined && !eq(object[key], value)) ||
        (value === undefined && !(key in object))) {
      baseAssignValue(object, key, value);
    }
  }

  function createBaseFor(fromRight) {
    return function(object, iteratee, keysFunc) {
      var index = -1,
          iterable = Object(object),
          props = keysFunc(object),
          length = props.length;

      while (length--) {
        var key = props[fromRight ? length : ++index];
        if (iteratee(iterable[key], key, iterable) === false) {
          break;
        }
      }
      return object;
    };
  }

  var baseFor = createBaseFor();

  var freeExports$2 = typeof exports == 'object' && exports && !exports.nodeType && exports;


  var freeModule$2 = freeExports$2 && typeof module == 'object' && module && !module.nodeType && module;


  var moduleExports$2 = freeModule$2 && freeModule$2.exports === freeExports$2;


  var Buffer$1 = moduleExports$2 ? root.Buffer : undefined,
      allocUnsafe = Buffer$1 ? Buffer$1.allocUnsafe : undefined;


  function cloneBuffer(buffer, isDeep) {
    if (isDeep) {
      return buffer.slice();
    }
    var length = buffer.length,
        result = allocUnsafe ? allocUnsafe(length) : new buffer.constructor(length);

    buffer.copy(result);
    return result;
  }

  function cloneArrayBuffer(arrayBuffer) {
    var result = new arrayBuffer.constructor(arrayBuffer.byteLength);
    new Uint8Array(result).set(new Uint8Array(arrayBuffer));
    return result;
  }

  function cloneTypedArray(typedArray, isDeep) {
    var buffer = isDeep ? cloneArrayBuffer(typedArray.buffer) : typedArray.buffer;
    return new typedArray.constructor(buffer, typedArray.byteOffset, typedArray.length);
  }

  function copyArray(source, array) {
    var index = -1,
        length = source.length;

    array || (array = Array(length));
    while (++index < length) {
      array[index] = source[index];
    }
    return array;
  }

  var objectCreate = Object.create;


  var baseCreate = (function() {
    function object() {}
    return function(proto) {
      if (!isObject$1(proto)) {
        return {};
      }
      if (objectCreate) {
        return objectCreate(proto);
      }
      object.prototype = proto;
      var result = new object;
      object.prototype = undefined;
      return result;
    };
  }());

  var getPrototype = overArg(Object.getPrototypeOf, Object);

  function initCloneObject(object) {
    return (typeof object.constructor == 'function' && !isPrototype(object))
      ? baseCreate(getPrototype(object))
      : {};
  }

  function isArrayLikeObject(value) {
    return isObjectLike(value) && isArrayLike(value);
  }

  var objectTag$3 = '[object Object]';


  var funcProto$2 = Function.prototype,
      objectProto$12 = Object.prototype;


  var funcToString$2 = funcProto$2.toString;


  var hasOwnProperty$9 = objectProto$12.hasOwnProperty;


  var objectCtorString = funcToString$2.call(Object);


  function isPlainObject$1(value) {
    if (!isObjectLike(value) || baseGetTag(value) != objectTag$3) {
      return false;
    }
    var proto = getPrototype(value);
    if (proto === null) {
      return true;
    }
    var Ctor = hasOwnProperty$9.call(proto, 'constructor') && proto.constructor;
    return typeof Ctor == 'function' && Ctor instanceof Ctor &&
      funcToString$2.call(Ctor) == objectCtorString;
  }

  function safeGet(object, key) {
    return key == '__proto__'
      ? undefined
      : object[key];
  }

  var objectProto$13 = Object.prototype;


  var hasOwnProperty$10 = objectProto$13.hasOwnProperty;


  function assignValue(object, key, value) {
    var objValue = object[key];
    if (!(hasOwnProperty$10.call(object, key) && eq(objValue, value)) ||
        (value === undefined && !(key in object))) {
      baseAssignValue(object, key, value);
    }
  }

  function copyObject(source, props, object, customizer) {
    var isNew = !object;
    object || (object = {});

    var index = -1,
        length = props.length;

    while (++index < length) {
      var key = props[index];

      var newValue = customizer
        ? customizer(object[key], source[key], key, object, source)
        : undefined;

      if (newValue === undefined) {
        newValue = source[key];
      }
      if (isNew) {
        baseAssignValue(object, key, newValue);
      } else {
        assignValue(object, key, newValue);
      }
    }
    return object;
  }

  function nativeKeysIn(object) {
    var result = [];
    if (object != null) {
      for (var key in Object(object)) {
        result.push(key);
      }
    }
    return result;
  }

  var objectProto$14 = Object.prototype;


  var hasOwnProperty$11 = objectProto$14.hasOwnProperty;


  function baseKeysIn(object) {
    if (!isObject$1(object)) {
      return nativeKeysIn(object);
    }
    var isProto = isPrototype(object),
        result = [];

    for (var key in object) {
      if (!(key == 'constructor' && (isProto || !hasOwnProperty$11.call(object, key)))) {
        result.push(key);
      }
    }
    return result;
  }

  function keysIn(object) {
    return isArrayLike(object) ? arrayLikeKeys(object, true) : baseKeysIn(object);
  }

  function toPlainObject(value) {
    return copyObject(value, keysIn(value));
  }

  function baseMergeDeep(object, source, key, srcIndex, mergeFunc, customizer, stack) {
    var objValue = safeGet(object, key),
        srcValue = safeGet(source, key),
        stacked = stack.get(srcValue);

    if (stacked) {
      assignMergeValue(object, key, stacked);
      return;
    }
    var newValue = customizer
      ? customizer(objValue, srcValue, (key + ''), object, source, stack)
      : undefined;

    var isCommon = newValue === undefined;

    if (isCommon) {
      var isArr = isArray$1(srcValue),
          isBuff = !isArr && isBuffer(srcValue),
          isTyped = !isArr && !isBuff && isTypedArray(srcValue);

      newValue = srcValue;
      if (isArr || isBuff || isTyped) {
        if (isArray$1(objValue)) {
          newValue = objValue;
        }
        else if (isArrayLikeObject(objValue)) {
          newValue = copyArray(objValue);
        }
        else if (isBuff) {
          isCommon = false;
          newValue = cloneBuffer(srcValue, true);
        }
        else if (isTyped) {
          isCommon = false;
          newValue = cloneTypedArray(srcValue, true);
        }
        else {
          newValue = [];
        }
      }
      else if (isPlainObject$1(srcValue) || isArguments(srcValue)) {
        newValue = objValue;
        if (isArguments(objValue)) {
          newValue = toPlainObject(objValue);
        }
        else if (!isObject$1(objValue) || (srcIndex && isFunction(objValue))) {
          newValue = initCloneObject(srcValue);
        }
      }
      else {
        isCommon = false;
      }
    }
    if (isCommon) {
      
      stack.set(srcValue, newValue);
      mergeFunc(newValue, srcValue, srcIndex, customizer, stack);
      stack['delete'](srcValue);
    }
    assignMergeValue(object, key, newValue);
  }

  function baseMerge(object, source, srcIndex, customizer, stack) {
    if (object === source) {
      return;
    }
    baseFor(source, function(srcValue, key) {
      if (isObject$1(srcValue)) {
        stack || (stack = new Stack);
        baseMergeDeep(object, source, key, srcIndex, baseMerge, customizer, stack);
      }
      else {
        var newValue = customizer
          ? customizer(safeGet(object, key), srcValue, (key + ''), object, source, stack)
          : undefined;

        if (newValue === undefined) {
          newValue = srcValue;
        }
        assignMergeValue(object, key, newValue);
      }
    }, keysIn);
  }

  function apply(func, thisArg, args) {
    switch (args.length) {
      case 0: return func.call(thisArg);
      case 1: return func.call(thisArg, args[0]);
      case 2: return func.call(thisArg, args[0], args[1]);
      case 3: return func.call(thisArg, args[0], args[1], args[2]);
    }
    return func.apply(thisArg, args);
  }

  var nativeMax$2 = Math.max;


  function overRest(func, start, transform) {
    start = nativeMax$2(start === undefined ? (func.length - 1) : start, 0);
    return function() {
      var args = arguments,
          index = -1,
          length = nativeMax$2(args.length - start, 0),
          array = Array(length);

      while (++index < length) {
        array[index] = args[start + index];
      }
      index = -1;
      var otherArgs = Array(start + 1);
      while (++index < start) {
        otherArgs[index] = args[index];
      }
      otherArgs[start] = transform(array);
      return apply(func, this, otherArgs);
    };
  }

  function constant(value) {
    return function() {
      return value;
    };
  }

  var baseSetToString = !defineProperty ? identity : function(func, string) {
    return defineProperty(func, 'toString', {
      'configurable': true,
      'enumerable': false,
      'value': constant(string),
      'writable': true
    });
  };

  var HOT_COUNT = 800,
      HOT_SPAN = 16;


  var nativeNow = Date.now;


  function shortOut(func) {
    var count = 0,
        lastCalled = 0;

    return function() {
      var stamp = nativeNow(),
          remaining = HOT_SPAN - (stamp - lastCalled);

      lastCalled = stamp;
      if (remaining > 0) {
        if (++count >= HOT_COUNT) {
          return arguments[0];
        }
      } else {
        count = 0;
      }
      return func.apply(undefined, arguments);
    };
  }

  var setToString = shortOut(baseSetToString);

  function baseRest(func, start) {
    return setToString(overRest(func, start, identity), func + '');
  }

  function isIterateeCall(value, index, object) {
    if (!isObject$1(object)) {
      return false;
    }
    var type = typeof index;
    if (type == 'number'
          ? (isArrayLike(object) && isIndex(index, object.length))
          : (type == 'string' && index in object)
        ) {
      return eq(object[index], value);
    }
    return false;
  }

  function createAssigner(assigner) {
    return baseRest(function(object, sources) {
      var index = -1,
          length = sources.length,
          customizer = length > 1 ? sources[length - 1] : undefined,
          guard = length > 2 ? sources[2] : undefined;

      customizer = (assigner.length > 3 && typeof customizer == 'function')
        ? (length--, customizer)
        : undefined;

      if (guard && isIterateeCall(sources[0], sources[1], guard)) {
        customizer = length < 3 ? undefined : customizer;
        length = 1;
      }
      object = Object(object);
      while (++index < length) {
        var source = sources[index];
        if (source) {
          assigner(object, source, index, customizer);
        }
      }
      return object;
    });
  }

  var merge = createAssigner(function(object, source, srcIndex) {
    baseMerge(object, source, srcIndex);
  });

  var mergeWith = createAssigner(function(object, source, srcIndex, customizer) {
    baseMerge(object, source, srcIndex, customizer);
  });

  function baseSet(object, path, value, customizer) {
    if (!isObject$1(object)) {
      return object;
    }
    path = castPath(path, object);

    var index = -1,
        length = path.length,
        lastIndex = length - 1,
        nested = object;

    while (nested != null && ++index < length) {
      var key = toKey(path[index]),
          newValue = value;

      if (index != lastIndex) {
        var objValue = nested[key];
        newValue = customizer ? customizer(objValue, key, nested) : undefined;
        if (newValue === undefined) {
          newValue = isObject$1(objValue)
            ? objValue
            : (isIndex(path[index + 1]) ? [] : {});
        }
      }
      assignValue(nested, key, newValue);
      nested = nested[key];
    }
    return object;
  }

  function setWith(object, path, value, customizer) {
    customizer = typeof customizer == 'function' ? customizer : undefined;
    return object == null ? object : baseSet(object, path, value, customizer);
  }

  function baseIsNaN(value) {
    return value !== value;
  }

  function strictIndexOf(array, value, fromIndex) {
    var index = fromIndex - 1,
        length = array.length;

    while (++index < length) {
      if (array[index] === value) {
        return index;
      }
    }
    return -1;
  }

  function baseIndexOf(array, value, fromIndex) {
    return value === value
      ? strictIndexOf(array, value, fromIndex)
      : baseFindIndex(array, baseIsNaN, fromIndex);
  }

  function arrayIncludes(array, value) {
    var length = array == null ? 0 : array.length;
    return !!length && baseIndexOf(array, value, 0) > -1;
  }

  function arrayIncludesWith(array, value, comparator) {
    var index = -1,
        length = array == null ? 0 : array.length;

    while (++index < length) {
      if (comparator(value, array[index])) {
        return true;
      }
    }
    return false;
  }

  function noop() {
    
  }

  var INFINITY$3 = 1 / 0;


  var createSet = !(Set$1 && (1 / setToArray(new Set$1([,-0]))[1]) == INFINITY$3) ? noop : function(values) {
    return new Set$1(values);
  };

  var LARGE_ARRAY_SIZE$1 = 200;


  function baseUniq(array, iteratee, comparator) {
    var index = -1,
        includes = arrayIncludes,
        length = array.length,
        isCommon = true,
        result = [],
        seen = result;

    if (comparator) {
      isCommon = false;
      includes = arrayIncludesWith;
    }
    else if (length >= LARGE_ARRAY_SIZE$1) {
      var set = iteratee ? null : createSet(array);
      if (set) {
        return setToArray(set);
      }
      isCommon = false;
      includes = cacheHas;
      seen = new SetCache;
    }
    else {
      seen = iteratee ? [] : result;
    }
    outer:
    while (++index < length) {
      var value = array[index],
          computed = iteratee ? iteratee(value) : value;

      value = (comparator || value !== 0) ? value : 0;
      if (isCommon && computed === computed) {
        var seenIndex = seen.length;
        while (seenIndex--) {
          if (seen[seenIndex] === computed) {
            continue outer;
          }
        }
        if (iteratee) {
          seen.push(computed);
        }
        result.push(value);
      }
      else if (!includes(seen, computed, comparator)) {
        if (seen !== result) {
          seen.push(computed);
        }
        result.push(value);
      }
    }
    return result;
  }

  function uniq(array) {
    return (array && array.length) ? baseUniq(array) : [];
  }

  function last(array) {
    var length = array == null ? 0 : array.length;
    return length ? array[length - 1] : undefined;
  }

  function parent(object, path) {
    return path.length < 2 ? object : baseGet(object, baseSlice(path, 0, -1));
  }

  function baseUnset(object, path) {
    path = castPath(path, object);
    object = parent(object, path);
    return object == null || delete object[toKey(last(path))];
  }

  function unset(object, path) {
    return object == null ? true : baseUnset(object, path);
  }

  var LARGE_ARRAY_SIZE$2 = 200;


  function baseDifference(array, values, iteratee, comparator) {
    var index = -1,
        includes = arrayIncludes,
        isCommon = true,
        length = array.length,
        result = [],
        valuesLength = values.length;

    if (!length) {
      return result;
    }
    if (iteratee) {
      values = arrayMap(values, baseUnary(iteratee));
    }
    if (comparator) {
      includes = arrayIncludesWith;
      isCommon = false;
    }
    else if (values.length >= LARGE_ARRAY_SIZE$2) {
      includes = cacheHas;
      isCommon = false;
      values = new SetCache(values);
    }
    outer:
    while (++index < length) {
      var value = array[index],
          computed = iteratee == null ? value : iteratee(value);

      value = (comparator || value !== 0) ? value : 0;
      if (isCommon && computed === computed) {
        var valuesIndex = valuesLength;
        while (valuesIndex--) {
          if (values[valuesIndex] === computed) {
            continue outer;
          }
        }
        result.push(value);
      }
      else if (!includes(values, computed, comparator)) {
        result.push(value);
      }
    }
    return result;
  }

  var without = baseRest(function(array, values) {
    return isArrayLikeObject(array)
      ? baseDifference(array, values)
      : [];
  });

  function baseForOwn(object, iteratee) {
    return object && baseFor(object, iteratee, keys);
  }

  function createBaseEach(eachFunc, fromRight) {
    return function(collection, iteratee) {
      if (collection == null) {
        return collection;
      }
      if (!isArrayLike(collection)) {
        return eachFunc(collection, iteratee);
      }
      var length = collection.length,
          index = fromRight ? length : -1,
          iterable = Object(collection);

      while ((fromRight ? index-- : ++index < length)) {
        if (iteratee(iterable[index], index, iterable) === false) {
          break;
        }
      }
      return collection;
    };
  }

  var baseEach = createBaseEach(baseForOwn);

  function baseMap(collection, iteratee) {
    var index = -1,
        result = isArrayLike(collection) ? Array(collection.length) : [];

    baseEach(collection, function(value, key, collection) {
      result[++index] = iteratee(value, key, collection);
    });
    return result;
  }

  function baseSortBy(array, comparer) {
    var length = array.length;

    array.sort(comparer);
    while (length--) {
      array[length] = array[length].value;
    }
    return array;
  }

  function compareAscending(value, other) {
    if (value !== other) {
      var valIsDefined = value !== undefined,
          valIsNull = value === null,
          valIsReflexive = value === value,
          valIsSymbol = isSymbol(value);

      var othIsDefined = other !== undefined,
          othIsNull = other === null,
          othIsReflexive = other === other,
          othIsSymbol = isSymbol(other);

      if ((!othIsNull && !othIsSymbol && !valIsSymbol && value > other) ||
          (valIsSymbol && othIsDefined && othIsReflexive && !othIsNull && !othIsSymbol) ||
          (valIsNull && othIsDefined && othIsReflexive) ||
          (!valIsDefined && othIsReflexive) ||
          !valIsReflexive) {
        return 1;
      }
      if ((!valIsNull && !valIsSymbol && !othIsSymbol && value < other) ||
          (othIsSymbol && valIsDefined && valIsReflexive && !valIsNull && !valIsSymbol) ||
          (othIsNull && valIsDefined && valIsReflexive) ||
          (!othIsDefined && valIsReflexive) ||
          !othIsReflexive) {
        return -1;
      }
    }
    return 0;
  }

  function compareMultiple(object, other, orders) {
    var index = -1,
        objCriteria = object.criteria,
        othCriteria = other.criteria,
        length = objCriteria.length,
        ordersLength = orders.length;

    while (++index < length) {
      var result = compareAscending(objCriteria[index], othCriteria[index]);
      if (result) {
        if (index >= ordersLength) {
          return result;
        }
        var order = orders[index];
        return result * (order == 'desc' ? -1 : 1);
      }
    }
    
    
    
    
    
    
    
    return object.index - other.index;
  }

  function baseOrderBy(collection, iteratees, orders) {
    var index = -1;
    iteratees = arrayMap(iteratees.length ? iteratees : [identity], baseUnary(baseIteratee));

    var result = baseMap(collection, function(value, key, collection) {
      var criteria = arrayMap(iteratees, function(iteratee) {
        return iteratee(value);
      });
      return { 'criteria': criteria, 'index': ++index, 'value': value };
    });

    return baseSortBy(result, function(object, other) {
      return compareMultiple(object, other, orders);
    });
  }

  function orderBy(collection, iteratees, orders, guard) {
    if (collection == null) {
      return [];
    }
    if (!isArray$1(iteratees)) {
      iteratees = iteratees == null ? [] : [iteratees];
    }
    orders = guard ? undefined : orders;
    if (!isArray$1(orders)) {
      orders = orders == null ? [] : [orders];
    }
    return baseOrderBy(collection, iteratees, orders);
  }

  var PathObject = function PathObject(root) {
    if (root) {
      this.__root__ = root;
    }
  };

  PathObject.prototype.contains = function contains (id) {
    return Boolean(this.getRoot()[id])
  };

  PathObject.prototype.getRoot = function getRoot () {
    if (this.__root__) {
      return this.__root__
    } else {
      return this
    }
  };


  PathObject.prototype.get = function get$1 (path) {
    if (!path) {
      return undefined
    }
    if (isString(path)) {
      return this.getRoot()[path]
    }
    if (arguments.length > 1) {
      path = Array.prototype.slice(arguments, 0);
    }
    if (!isArray(path)) {
      throw new Error('Illegal argument for PathObject.get()')
    }
    return get(this.getRoot(), path)
  };

  PathObject.prototype.set = function set (path, value) {
    if (!path) {
      throw new Error('Illegal argument: PathObject.set(>path<, value) - path is mandatory.')
    }
    if (isString(path)) {
      this.getRoot()[path] = value;
    } else {
      setWith(this.getRoot(), path, value);
    }
  };

  PathObject.prototype.delete = function delete$1 (path) {
    if (isString(path)) {
      delete this.getRoot()[path];
    } else if (path.length === 1) {
      delete this.getRoot()[path[0]];
    } else {
      var success = unset(this.getRoot(), path);
      if (!success) {
        throw new Error('Could not delete property at path' + path)
      }
    }
  };

  PathObject.prototype.clear = function clear () {
    var root = this.getRoot();
    for (var key in root) {
      if (root.hasOwnProperty(key)) {
        delete root[key];
      }
    }
  };

  PathObject.prototype._isPathObject = true;

  var Conflict = (function (Error) {
    function Conflict(a, b) {
      Error.call(this, "Conflict: " + JSON.stringify(a) +" vs " + JSON.stringify(b));
      this.a = a;
      this.b = b;
    }

    if ( Error ) Conflict.__proto__ = Error;
    Conflict.prototype = Object.create( Error && Error.prototype );
    Conflict.prototype.constructor = Conflict;

    return Conflict;
  }(Error));

  var INSERT = "insert";
  var DELETE = "delete";

  var TextOperation = function TextOperation(data) {
    if (!data || data.type === undefined || data.pos === undefined || data.str === undefined) {
      throw new Error("Illegal argument: insufficient data.")
    }
    
    this.type = data.type;
    
    this.pos = data.pos;
    
    this.str = data.str;
    
    if(!this.isInsert() && !this.isDelete()) {
      throw new Error("Illegal type.")
    }
    if (!isString(this.str)) {
      throw new Error("Illegal argument: expecting string.")
    }
    if (!isNumber(this.pos) || this.pos < 0) {
      throw new Error("Illegal argument: expecting positive number as pos.")
    }
  };

  TextOperation.prototype.apply = function apply (str) {
    if (this.isEmpty()) { return str }
    if (this.type === INSERT) {
      if (str.length < this.pos) {
        throw new Error("Provided string is too short.")
      }
      if (str.splice) {
        return str.splice(this.pos, 0, this.str)
      } else {
        return str.slice(0, this.pos).concat(this.str).concat(str.slice(this.pos))
      }
    }
    else  {
      if (str.length < this.pos + this.str.length) {
        throw new Error("Provided string is too short.")
      }
      if (str.splice) {
        return str.splice(this.pos, this.str.length)
      } else {
        return str.slice(0, this.pos).concat(str.slice(this.pos + this.str.length))
      }
    }
  };

  TextOperation.prototype.clone = function clone () {
    return new TextOperation(this)
  };

  TextOperation.prototype.isNOP = function isNOP () {
    return this.type === "NOP" || this.str.length === 0
  };

  TextOperation.prototype.isInsert = function isInsert () {
    return this.type === INSERT
  };

  TextOperation.prototype.isDelete = function isDelete () {
    return this.type === DELETE
  };

  TextOperation.prototype.getLength = function getLength () {
    return this.str.length
  };

  TextOperation.prototype.invert = function invert () {
    var data = {
      type: this.isInsert() ? DELETE : INSERT,
      pos: this.pos,
      str: this.str
    };
    return new TextOperation(data)
  };

  TextOperation.prototype.hasConflict = function hasConflict (other) {
    return _hasConflict(this, other)
  };

  TextOperation.prototype.isEmpty = function isEmpty () {
    return this.str.length === 0
  };

  TextOperation.prototype.toJSON = function toJSON () {
    return {
      type: this.type,
      pos: this.pos,
      str: this.str
    }
  };

  TextOperation.prototype.toString = function toString () {
    return ["(", (this.isInsert() ? INSERT : DELETE), ",", this.pos, ",'", this.str, "')"].join('')
  };

  TextOperation.prototype._isOperation = true;
  TextOperation.prototype._isTextOperation = true;

  function _hasConflict(a, b) {
    
    
    
    if (a.type === INSERT && b.type === INSERT) { return (a.pos === b.pos) }
    
    
    
    if (a.type === DELETE && b.type === DELETE) {
      
      return !(a.pos >= b.pos + b.str.length || b.pos >= a.pos + a.str.length)
    }
    
    
    
    var del, ins;
    if (a.type === DELETE) {
      del = a; ins = b;
    } else {
      del = b; ins = a;
    }
    return (ins.pos >= del.pos && ins.pos < del.pos + del.str.length)
  }




  function transform_insert_insert(a, b) {
    if (a.pos === b.pos) {
      b.pos += a.str.length;
    }
    else if (a.pos < b.pos) {
      b.pos += a.str.length;
    }
    else {
      a.pos += b.str.length;
    }
  }





  function transform_delete_delete(a, b, first) {
    
    if (a.pos > b.pos) {
      return transform_delete_delete(b, a, !first)
    }
    if (a.pos === b.pos && a.str.length > b.str.length) {
      return transform_delete_delete(b, a, !first)
    }
    
    if (b.pos < a.pos + a.str.length) {
      var s = b.pos - a.pos;
      var s1 = a.str.length - s;
      var s2 = s + b.str.length;
      a.str = a.str.slice(0, s) + a.str.slice(s2);
      b.str = b.str.slice(s1);
      b.pos -= s;
    } else {
      b.pos -= a.str.length;
    }
  }





  function transform_insert_delete(a, b) {
    if (a.type === DELETE) {
      return transform_insert_delete(b, a)
    }
    
    
    if (a.pos <= b.pos) {
      b.pos += a.str.length;
    }
    
    else if (a.pos >= b.pos + b.str.length) {
      a.pos -= b.str.length;
    }
    
    
    
    else {
      var s = a.pos - b.pos;
      b.str = b.str.slice(0, s) + a.str + b.str.slice(s);
      a.str = "";
    }
  }

  function transform(a, b, options) {
    options = options || {};
    if (options["no-conflict"] && _hasConflict(a, b)) {
      throw new Conflict(a, b)
    }
    if (!options.inplace) {
      a = a.clone();
      b = b.clone();
    }
    if (a.type === INSERT && b.type === INSERT) {
      transform_insert_insert(a, b);
    }
    else if (a.type === DELETE && b.type === DELETE) {
      transform_delete_delete(a, b, true);
    }
    else {
      transform_insert_delete(a,b);
    }
    return [a, b]
  }

  TextOperation.transform = function() {
    return transform.apply(null, arguments)
  };



  TextOperation.Insert = function(pos, str) {
    return new TextOperation({ type: INSERT, pos: pos, str: str })
  };

  TextOperation.Delete = function(pos, str) {
    return new TextOperation({ type: DELETE, pos: pos, str: str })
  };

  TextOperation.INSERT = INSERT;
  TextOperation.DELETE = DELETE;

  TextOperation.fromJSON = function(data) {
    return new TextOperation(data)
  };

  var NOP = "NOP";
  var DELETE$1 = "delete";
  var INSERT$1 = "insert";

  var ArrayOperation = function ArrayOperation(data) {
    if (!data || !data.type) {
      throw new Error("Illegal argument: insufficient data.")
    }
    this.type = data.type;
    if (this.type === NOP) { return }

    if (this.type !== INSERT$1 && this.type !== DELETE$1) {
      throw new Error("Illegal type.")
    }
    
    this.pos = data.pos;
    
    this.val = data.val;
    if (!isNumber(this.pos) || this.pos < 0) {
      throw new Error("Illegal argument: expecting positive number as pos.")
    }
  };

  ArrayOperation.prototype.apply = function apply (array) {
    if (this.type === NOP) {
      return array
    }
    if (this.type === INSERT$1) {
      if (array.length < this.pos) {
        throw new Error("Provided array is too small.")
      }
      array.splice(this.pos, 0, this.val);
      return array
    }
    
    else  {
      if (array.length < this.pos) {
        throw new Error("Provided array is too small.")
      }
      if (!isEqual(array[this.pos], this.val)) {
        throw Error("Unexpected value at position " + this.pos + ". Expected " + this.val + ", found " + array[this.pos])
      }
      array.splice(this.pos, 1);
      return array
    }
  };

  ArrayOperation.prototype.clone = function clone () {
    var data = {
      type: this.type,
      pos: this.pos,
      val: cloneDeep(this.val)
    };
    return new ArrayOperation(data)
  };

  ArrayOperation.prototype.invert = function invert () {
    var data = this.toJSON();
    if (this.type === NOP) { data.type = NOP; }
    else if (this.type === INSERT$1) { data.type = DELETE$1; }
    else  { data.type = INSERT$1; }
    return new ArrayOperation(data)
  };

  ArrayOperation.prototype.hasConflict = function hasConflict (other) {
    return ArrayOperation.hasConflict(this, other)
  };

  ArrayOperation.prototype.toJSON = function toJSON () {
    var result = {
      type: this.type,
    };
    if (this.type === NOP) { return result }
    result.pos = this.pos;
    result.val = cloneDeep(this.val);
    return result
  };

  ArrayOperation.prototype.isInsert = function isInsert () {
    return this.type === INSERT$1
  };

  ArrayOperation.prototype.isDelete = function isDelete () {
    return this.type === DELETE$1
  };

  ArrayOperation.prototype.getOffset = function getOffset () {
    return this.pos
  };

  ArrayOperation.prototype.getValue = function getValue () {
    return this.val
  };

  ArrayOperation.prototype.isNOP = function isNOP () {
    return this.type === NOP
  };

  ArrayOperation.prototype.toString = function toString () {
    return ["(", (this.isInsert() ? INSERT$1 : DELETE$1), ",", this.getOffset(), ",'", this.getValue(), "')"].join('')
  };

  ArrayOperation.prototype._isOperation = true;
  ArrayOperation.prototype._isArrayOperation = true;

  function hasConflict(a, b) {
    if (a.type === NOP || b.type === NOP) { return false }
    if (a.type === INSERT$1 && b.type === INSERT$1) {
      return a.pos === b.pos
    } else {
      return false
    }
  }

  function transform_insert_insert$1(a, b) {
    if (a.pos === b.pos) {
      b.pos += 1;
    }
    
    else if (a.pos < b.pos) {
      b.pos += 1;
    }
    
    else {
      a.pos += 1;
    }
  }

  function transform_delete_delete$1(a, b) {
    
    if (a.pos === b.pos) {
      b.type = NOP;
      a.type = NOP;
      return
    }
    if (a.pos < b.pos) {
      b.pos -= 1;
    } else {
      a.pos -= 1;
    }
  }

  function transform_insert_delete$1(a, b) {
    
    if (a.type === DELETE$1) {
      var tmp = a;
      a = b;
      b = tmp;
    }
    if (a.pos <= b.pos) {
      b.pos += 1;
    } else {
      a.pos -= 1;
    }
  }

  var transform$1 = function(a, b, options) {
    options = options || {};
    
    
    if (options['no-conflict'] && hasConflict(a, b)) {
      throw new Conflict(a, b)
    }
    
    if (!options.inplace) {
      a = a.clone();
      b = b.clone();
    }
    if (a.type === NOP || b.type === NOP) ;
    else if (a.type === INSERT$1 && b.type === INSERT$1) {
      transform_insert_insert$1(a, b);
    }
    else if (a.type === DELETE$1 && b.type === DELETE$1) {
      transform_delete_delete$1(a, b);
    }
    else {
      transform_insert_delete$1(a, b);
    }
    return [a, b]
  };

  ArrayOperation.transform = transform$1;
  ArrayOperation.hasConflict = hasConflict;



  ArrayOperation.Insert = function(pos, val) {
    return new ArrayOperation({type:INSERT$1, pos: pos, val: val})
  };

  ArrayOperation.Delete = function(pos, val) {
    return new ArrayOperation({ type:DELETE$1, pos: pos, val: val })
  };

  ArrayOperation.fromJSON = function(data) {
    return new ArrayOperation(data)
  };

  ArrayOperation.NOP = NOP;
  ArrayOperation.DELETE = DELETE$1;
  ArrayOperation.INSERT = INSERT$1;

  var SHIFT = 'shift';

  var CoordinateOperation = function CoordinateOperation(data) {
    if (!data || data.type === undefined) {
      throw new Error("Illegal argument: insufficient data.")
    }
    
    this.type = data.type;
    
    this.val = data.val;
    
    if(!this.isShift()) {
      throw new Error("Illegal type.")
    }
    if (!isNumber(this.val)) {
      throw new Error("Illegal argument: expecting number as shift value.")
    }
  };

  CoordinateOperation.prototype.apply = function apply (coor) {
    coor.offset = coor.offset + this.val;
  };

  CoordinateOperation.prototype.isShift = function isShift () {
    return this.type === SHIFT
  };

  CoordinateOperation.prototype.isNOP = function isNOP () {
    switch (this.type) {
      case SHIFT: {
        return this.val === 0
      }
      default:
        return false
    }
  };

  CoordinateOperation.prototype.clone = function clone () {
    return new CoordinateOperation(this)
  };

  CoordinateOperation.prototype.invert = function invert () {
    var data;
    switch (this.type) {
      case SHIFT:
        data = {
          type: SHIFT,
          val: -this.val
        };
        break
      default:
        throw new Error('Invalid type.')
    }
    return new CoordinateOperation(data)
  };

  CoordinateOperation.prototype.hasConflict = function hasConflict () {
    
    return false
  };

  CoordinateOperation.prototype.toJSON = function toJSON () {
    return {
      type: this.type,
      val: this.val
    }
  };

  CoordinateOperation.prototype.toString = function toString () {
    return ["(", (this.type), ",", this.val, "')"].join('')
  };

  CoordinateOperation.prototype._isOperation = true;
  CoordinateOperation.prototype._isCoordinateOperation = true;

  function transform_shift_shift(a, b) {
    a.val += b.val;
    b.val += a.val;
  }

  function transform$2(a, b, options) {
    options = options || {};
    
    if (!options.inplace) {
      a = a.clone();
      b = b.clone();
    }
    if (a.type === SHIFT && b.type === SHIFT) {
      transform_shift_shift(a, b);
    }
    else {
      throw new Error('Illegal type')
    }
    return [a, b]
  }

  CoordinateOperation.transform = function() {
    var args = [], len = arguments.length;
    while ( len-- ) args[ len ] = arguments[ len ];

    return transform$2.apply(void 0, args)
  };

  CoordinateOperation.fromJSON = function(json) {
    return new CoordinateOperation(json)
  };

  CoordinateOperation.Shift = function(val) {
    return new CoordinateOperation({
      type: SHIFT,
      val: val
    })
  };

  var NOP$1 = "NOP";
  var CREATE = "create";
  var DELETE$2 = 'delete';
  var UPDATE = 'update';
  var SET = 'set';

  var ObjectOperation = function ObjectOperation(data) {
    
    if (!data) {
      throw new Error('Data of ObjectOperation is missing.')
    }
    
    if (!data.type) {
      throw new Error('Invalid data: type is mandatory.')
    }
    this.type = data.type;
    if (data.type === NOP$1) {
      return
    }
    this.path = data.path;
    if (!data.path) {
      throw new Error('Invalid data: path is mandatory.')
    }
    if (this.type === CREATE || this.type === DELETE$2) {
      if (!data.val) {
        throw new Error('Invalid data: value is missing.')
      }
      this.val = data.val;
    }
    else if (this.type === UPDATE) {
      if (data.diff) {
        this.diff = data.diff;
        if (data.diff._isTextOperation) {
          this.propertyType = 'string';
        } else if (data.diff._isArrayOperation) {
          this.propertyType = 'array';
        } else if (data.diff._isCoordinateOperation) {
          this.propertyType = 'coordinate';
        } else {
          throw new Error('Invalid data: diff must be a TextOperation or an ArrayOperation.')
        }
      } else {
        throw new Error("Invalid data: diff is mandatory for update operation.")
      }
    }
    else if (this.type === SET) {
      this.val = data.val;
      this.original = data.original;
    } else {
      throw new Error('Invalid type: '+ data.type)
    }
  };

  ObjectOperation.prototype.apply = function apply (obj) {
    if (this.type === NOP$1) { return obj }
    var adapter;
    if (obj._isPathObject) {
      adapter = obj;
    } else {
      adapter = new PathObject(obj);
    }
    if (this.type === CREATE) {
      adapter.set(this.path, cloneDeep(this.val));
      return obj
    }
    if (this.type === DELETE$2) {
      adapter.delete(this.path, "strict");
    }
    else if (this.type === UPDATE) {
      var diff = this.diff;
      switch (this.propertyType) {
        case 'array': {
          var arr = adapter.get(this.path);
          diff.apply(arr);
          break
        }
        case 'string': {
          var str = adapter.get(this.path);
          if (isNil(str)) { str = ''; }
          str = diff.apply(str);
          adapter.set(this.path, str);
          break
        }
        case 'coordinate': {
          var coor = adapter.get(this.path);
          if (!coor) { throw new Error('No coordinate with path '+this.path) }
          diff.apply(coor);
          break
        }
        default:
          throw new Error('Invalid state.')
      }
    }
    else if (this.type === SET) {
      
      adapter.set(this.path, cloneDeep(this.val));
    }
    else {
      throw new Error('Invalid type.')
    }
    return obj
  };

  ObjectOperation.prototype.clone = function clone () {
    var data = {
      type: this.type,
      path: this.path,
    };
    if (this.val) {
      data.val = cloneDeep(this.val);
    }
    if (this.diff) {
      data.diff = this.diff.clone();
    }
    return new ObjectOperation(data)
  };

  ObjectOperation.prototype.isNOP = function isNOP () {
    if (this.type === NOP$1) { return true }
    else if (this.type === UPDATE) { return this.diff.isNOP() }
  };

  ObjectOperation.prototype.isCreate = function isCreate () {
    return this.type === CREATE
  };

  ObjectOperation.prototype.isDelete = function isDelete () {
    return this.type === DELETE$2
  };

  ObjectOperation.prototype.isUpdate = function isUpdate (propertyType) {
    if (propertyType) {
      return (this.type === UPDATE && this.propertyType === propertyType)
    } else {
      return this.type === UPDATE
    }
  };

  ObjectOperation.prototype.isSet = function isSet () {
    return this.type === SET
  };

  ObjectOperation.prototype.invert = function invert () {
    if (this.type === NOP$1) {
      return new ObjectOperation({ type: NOP$1 })
    }
    var result = new ObjectOperation(this);
    if (this.type === CREATE) {
      result.type = DELETE$2;
    }
    else if (this.type === DELETE$2) {
      result.type = CREATE;
    }
    else if (this.type === UPDATE) {
      var invertedDiff;
      if (this.diff._isTextOperation) {
        invertedDiff = TextOperation.fromJSON(this.diff.toJSON()).invert();
      } else if (this.diff._isArrayOperation) {
        invertedDiff = ArrayOperation.fromJSON(this.diff.toJSON()).invert();
      } else if (this.diff._isCoordinateOperation) {
        invertedDiff = CoordinateOperation.fromJSON(this.diff.toJSON()).invert();
      } else {
        throw new Error('Illegal type')
      }
      result.diff = invertedDiff;
    }
    else  {
      result.val = this.original;
      result.original = this.val;
    }
    return result
  };

  ObjectOperation.prototype.hasConflict = function hasConflict (other) {
    return ObjectOperation.hasConflict(this, other)
  };

  ObjectOperation.prototype.toJSON = function toJSON () {
    if (this.type === NOP$1) {
      return { type: NOP$1 }
    }
    var data = {
      type: this.type,
      path: this.path,
    };
    if (this.type === CREATE || this.type === DELETE$2) {
      data.val = this.val;
    }
    else if (this.type === UPDATE) {
      if (this.diff._isTextOperation) {
        data.propertyType = "string";
      } else if (this.diff._isArrayOperation) {
        data.propertyType = "array";
      } else if (this.diff._isCoordinateOperation) {
        data.propertyType = "coordinate";
      } else {
        throw new Error('Invalid property type.')
      }
      data.diff = this.diff.toJSON();
    }
    else  {
      data.val = this.val;
      data.original = this.original;
    }
    return data
  };

  ObjectOperation.prototype.getType = function getType () {
    return this.type
  };

  ObjectOperation.prototype.getPath = function getPath () {
    return this.path
  };

  ObjectOperation.prototype.getValue = function getValue () {
    return this.val
  };

  ObjectOperation.prototype.getOldValue = function getOldValue () {
    return this.original
  };

  ObjectOperation.prototype.getValueOp = function getValueOp () {
    return this.diff
  };


  ObjectOperation.prototype.toString = function toString () {
    switch (this.type) {
      case CREATE:
        return ["(+,", JSON.stringify(this.path), JSON.stringify(this.val), ")"].join('')
      case DELETE$2:
        return ["(-,", JSON.stringify(this.path), JSON.stringify(this.val), ")"].join('')
      case UPDATE:
        return ["(>>,", JSON.stringify(this.path), this.propertyType, this.diff.toString(), ")"].join('')
      case SET:
        return ["(=,", JSON.stringify(this.path), this.val, this.original, ")"].join('')
      case NOP$1:
        return "NOP"
      default:
        throw new Error('Invalid type')
    }
  };

  ObjectOperation.prototype._isOperation = true;
  ObjectOperation.prototype._isObjectOperation = true;



  function hasConflict$1(a, b) {
    if (a.type === NOP$1 || b.type === NOP$1) { return false }
    return isEqual(a.path, b.path)
  }

  function transform_delete_delete$2(a, b) {
    
    
    a.type = NOP$1;
    b.type = NOP$1;
  }

  function transform_create_create() {
    throw new Error("Can not transform two concurring creates of the same property")
  }

  function transform_delete_create() {
    throw new Error('Illegal state: can not create and delete a value at the same time.')
  }

  function transform_delete_update(a, b, flipped) {
    if (a.type !== DELETE$2) {
      return transform_delete_update(b, a, true)
    }
    var op;
    switch (b.propertyType) {
      case 'string':
        op = TextOperation.fromJSON(b.diff);
        break
      case 'array':
        op = ArrayOperation.fromJSON(b.diff);
        break
      case 'coordinate':
        op = CoordinateOperation.fromJSON(b.diff);
        break
      default:
        throw new Error('Illegal type')
    }
    
    if (!flipped) {
      a.type = NOP$1;
      b.type = CREATE;
      b.val = op.apply(a.val);
    }
    
    else {
      a.val = op.apply(a.val);
      b.type = NOP$1;
    }
  }

  function transform_create_update() {
    
    throw new Error("Can not transform a concurring create and update of the same property")
  }

  function transform_update_update(a, b) {
    
    var op_a, op_b, t;
    switch(b.propertyType) {
      case 'string':
        op_a = TextOperation.fromJSON(a.diff);
        op_b = TextOperation.fromJSON(b.diff);
        t = TextOperation.transform(op_a, op_b, {inplace: true});
        break
      case 'array':
        op_a = ArrayOperation.fromJSON(a.diff);
        op_b = ArrayOperation.fromJSON(b.diff);
        t = ArrayOperation.transform(op_a, op_b, {inplace: true});
        break
      case 'coordinate':
        op_a = CoordinateOperation.fromJSON(a.diff);
        op_b = CoordinateOperation.fromJSON(b.diff);
        t = CoordinateOperation.transform(op_a, op_b, {inplace: true});
        break
      default:
        throw new Error('Illegal type')
    }
    a.diff = t[0];
    b.diff = t[1];
  }

  function transform_create_set() {
    throw new Error('Illegal state: can not create and set a value at the same time.')
  }

  function transform_delete_set(a, b, flipped) {
    if (a.type !== DELETE$2) { return transform_delete_set(b, a, true) }
    if (!flipped) {
      a.type = NOP$1;
      b.type = CREATE;
      b.original = undefined;
    } else {
      a.val = b.val;
      b.type = NOP$1;
    }
  }

  function transform_update_set() {
    throw new Error("Unresolvable conflict: update + set.")
  }

  function transform_set_set(a, b) {
    a.type = NOP$1;
    b.original = a.val;
  }

  var _NOP = 0;
  var _CREATE = 1;
  var _DELETE = 2;
  var _UPDATE = 4;
  var _SET = 8;

  var CODE = (function () {
    var c = {};
    c[NOP$1] =_NOP;
    c[CREATE] = _CREATE;
    c[DELETE$2] = _DELETE;
    c[UPDATE] = _UPDATE;
    c[SET] = _SET;
    return c
  })();

  var __transform__ = (function () {
    
    var t = {};
    t[_DELETE | _DELETE] = transform_delete_delete$2;
    t[_DELETE | _CREATE] = transform_delete_create;
    t[_DELETE | _UPDATE] = transform_delete_update;
    t[_CREATE | _CREATE] = transform_create_create;
    t[_CREATE | _UPDATE] = transform_create_update;
    t[_UPDATE | _UPDATE] = transform_update_update;
    t[_CREATE | _SET   ] = transform_create_set;
    t[_DELETE | _SET   ] = transform_delete_set;
    t[_UPDATE | _SET   ] = transform_update_set;
    t[_SET    | _SET   ] = transform_set_set;
    
    return t
  })();

  function transform$3(a, b, options) {
    options = options || {};
    if (options['no-conflict'] && hasConflict$1(a, b)) {
      throw new Conflict(a, b)
    }
    if (!options.inplace) {
      a = a.clone();
      b = b.clone();
    }
    if (a.isNOP() || b.isNOP()) {
      return [a, b]
    }
    var sameProp = isEqual(a.path, b.path);
    
    if (sameProp) {
      __transform__[CODE[a.type] | CODE[b.type]](a,b);
    }
    return [a, b]
  }

  ObjectOperation.transform = transform$3;
  ObjectOperation.hasConflict = hasConflict$1;



  ObjectOperation.Create = function(idOrPath, val) {
    var path;
    if (isString(idOrPath)) {
      path = [idOrPath];
    } else {
      path = idOrPath;
    }
    return new ObjectOperation({type: CREATE, path: path, val: val})
  };

  ObjectOperation.Delete = function(idOrPath, val) {
    var path;
    if (isString(idOrPath)) {
      path = [idOrPath];
    } else {
      path = idOrPath;
    }
    return new ObjectOperation({type: DELETE$2, path: path, val: val})
  };

  ObjectOperation.Update = function(path, op) {
    return new ObjectOperation({
      type: UPDATE,
      path: path,
      diff: op
    })
  };

  ObjectOperation.Set = function(path, oldVal, newVal) {
    return new ObjectOperation({
      type: SET,
      path: path,
      val: cloneDeep(newVal),
      original: cloneDeep(oldVal)
    })
  };

  ObjectOperation.fromJSON = function(data) {
    data = cloneDeep(data);
    if (data.type === "update") {
      switch (data.propertyType) {
        case "string":
          data.diff = TextOperation.fromJSON(data.diff);
          break
        case "array":
          data.diff = ArrayOperation.fromJSON(data.diff);
          break
        case "coordinate":
          data.diff = CoordinateOperation.fromJSON(data.diff);
          break
        default:
          throw new Error("Unsupported update diff:" + JSON.stringify(data.diff))
      }
    }
    var op = new ObjectOperation(data);
    return op
  };

  ObjectOperation.NOP = NOP$1;
  ObjectOperation.CREATE = CREATE;
  ObjectOperation.DELETE = DELETE$2;
  ObjectOperation.UPDATE = UPDATE;
  ObjectOperation.SET = SET;

  var OperationSerializer = function OperationSerializer() {
    this.SEPARATOR = '\t';
  };

  OperationSerializer.prototype.serialize = function serialize (op) {
    var out = [];
    switch (op.type) {
      case 'create':
        out.push('c');
        out.push(op.val.id);
        out.push(op.val);
        break
      case 'delete':
        out.push('d');
        out.push(op.val.id);
        out.push(op.val);
        break
      case 'set':
        out.push('s');
        out.push(op.path.join('.'));
        out.push(op.val);
        out.push(op.original);
        break
      case 'update':
        out.push('u');
        out.push(op.path.join('.'));
        Array.prototype.push.apply(out, this.serializePrimitiveOp(op.diff));
        break
      default:
        throw new Error('Unsupported operation type.')
    }
    return out
  };

  OperationSerializer.prototype.serializePrimitiveOp = function serializePrimitiveOp (op) {
    var out = [];
    if (op._isTextOperation) {
      if (op.isInsert()) {
        out.push('t+');
      } else if (op.isDelete()) {
        out.push('t-');
      }
      out.push(op.pos);
      out.push(op.str);
    } else if (op._isArrayOperation) {
      if (op.isInsert()) {
        out.push('a+');
      } else if (op.isDelete()) {
        out.push('a-');
      }
      out.push(op.pos);
      out.push(op.val);
    } else if (op._isCoordinateOperation) {
      if (op.isShift()) {
        out.push('c>>');
      } else {
        throw new Error('Unsupported CoordinateOperation type.')
      }
      out.push(op.pos);
      out.push(op.val);
    } else {
      throw new Error('Unsupported operation type.')
    }
    return out
  };

  OperationSerializer.prototype.deserialize = function deserialize (str, tokenizer) {
    if (!tokenizer) {
      tokenizer = new Tokenizer(str, this.SEPARATOR);
    }
    var type = tokenizer.getString();
    var op, path, val, oldVal, diff;
    switch (type) {
      case 'c':
        path = tokenizer.getPath();
        val = tokenizer.getObject();
        op = ObjectOperation.Create(path, val);
        break
      case 'd':
        path = tokenizer.getPath();
        val = tokenizer.getObject();
        op = ObjectOperation.Delete(path, val);
        break
      case 's':
        path = tokenizer.getPath();
        val = tokenizer.getAny();
        oldVal = tokenizer.getAny();
        op = ObjectOperation.Set(path, oldVal, val);
        break
      case 'u':
        path = tokenizer.getPath();
        diff = this.deserializePrimitiveOp(str, tokenizer);
        op = ObjectOperation.Update(path, diff);
        break
      default:
        throw new Error('Illegal type for ObjectOperation: '+ type)
    }
    return op
  };

  OperationSerializer.prototype.deserializePrimitiveOp = function deserializePrimitiveOp (str, tokenizer) {
    if (!tokenizer) {
      tokenizer = new Tokenizer(str, this.SEPARATOR);
    }
    var type = tokenizer.getString();
    var op, pos, val;
    switch (type) {
      case 't+':
        pos = tokenizer.getNumber();
        val = tokenizer.getString();
        op = TextOperation.Insert(pos, val);
        break
      case 't-':
        pos = tokenizer.getNumber();
        val = tokenizer.getString();
        op = TextOperation.Delete(pos, val);
        break
      case 'a+':
        pos = tokenizer.getNumber();
        val = tokenizer.getAny();
        op = ArrayOperation.Insert(pos, val);
        break
      case 'a-':
        pos = tokenizer.getNumber();
        val = tokenizer.getAny();
        op = ArrayOperation.Delete(pos, val);
        break
      case 'c>>':
        val = tokenizer.getNumber();
        op = CoordinateOperation.Shift(val);
        break
      default:
        throw new Error('Unsupported operation type: ' + type)
    }
    return op
  };

  var Tokenizer = function Tokenizer(str, sep) {
    if (isArray(arguments[0])) {
      this.tokens = arguments[0];
    } else {
      this.tokens = str.split(sep);
    }
    this.pos = -1;
  };

  Tokenizer.prototype.error = function error (msg) {
    throw new Error('Parsing error: ' + msg + '\n' + this.tokens[this.pos])
  };

  Tokenizer.prototype.getString = function getString () {
    this.pos++;
    var str = this.tokens[this.pos];
    if (str[0] === '"') {
      str = str.slice(1, -1);
    }
    return str
  };

  Tokenizer.prototype.getNumber = function getNumber () {
    this.pos++;
    var number;
    var token = this.tokens[this.pos];
    try {
      if (isNumber(token)) {
        number = token;
      } else {
        number = parseInt(this.tokens[this.pos], 10);
      }
      return number
    } catch (err) {
      this.error('expected number');
    }
  };

  Tokenizer.prototype.getObject = function getObject () {
    this.pos++;
    var obj;
    var token = this.tokens[this.pos];
    try {
      if (isObject(token)) {
        obj = token;
      } else {
        obj = JSON.parse(this.tokens[this.pos]);
      }
      return obj
    } catch (err) {
      this.error('expected object');
    }
  };

  Tokenizer.prototype.getAny = function getAny () {
    this.pos++;
    var token = this.tokens[this.pos];
    return token
  };

  Tokenizer.prototype.getPath = function getPath () {
    var str = this.getString();
    return str.split('.')
  };

  OperationSerializer.Tokenizer = Tokenizer;

  function last$1(arr) {
    return arr[arr.length-1]
  }

  var Selection = function Selection() {
    
    var _internal = {};
    Object.defineProperty(this, "_internal", {
      enumerable: false,
      value: _internal
    });
    
    _internal.doc = null;
  };

  var prototypeAccessors$1 = { type: { configurable: true } };

  Selection.prototype.clone = function clone () {
    var newSel = this._clone();
    if (this._internal.doc) {
      newSel.attach(this._internal.doc);
    }
    return newSel
  };


  Selection.prototype.getDocument = function getDocument () {
    var doc = this._internal.doc;
    if (!doc) {
      throw new Error('Selection is not attached to a document.')
    }
    return doc
  };

  Selection.prototype.isAttached = function isAttached () {
    return Boolean(this._internal.doc)
  };


  Selection.prototype.attach = function attach (doc) {
    this._internal.doc = doc;
    return this
  };


  Selection.prototype.isNull = function isNull () { return false; };


  Selection.prototype.isPropertySelection = function isPropertySelection () { return false; };


  Selection.prototype.isContainerSelection = function isContainerSelection () { return false; };


  Selection.prototype.isNodeSelection = function isNodeSelection () { return false; };

  Selection.prototype.isCustomSelection = function isCustomSelection () { return false; };


  Selection.prototype.isCollapsed = function isCollapsed () { return true; };


  Selection.prototype.isReverse = function isReverse () { return false; };

  Selection.prototype.getType = function getType () {
    throw new Error('Selection.getType() is abstract.')
  };

  prototypeAccessors$1.type.get = function () {
    return this.getType()
  };


  Selection.prototype.equals = function equals (other) {
    if (this === other) {
      return true
    } else if (!other) {
      return false
    } else if (this.isNull() !== other.isNull()) {
      return false
    } else if (this.getType() !== other.getType()) {
      return false
    } else {
      
      
      return true
    }
  };


  Selection.prototype.toString = function toString () {
    return "null"
  };


  Selection.prototype.toJSON = function toJSON () {
    throw new Error('This method is abstract.')
  };

  Selection.prototype.createWith = function createWith (update) {
    var SelectionClass = this.constructor;
    var data = this.toJSON();
    Object.assign(data, update);
    return SelectionClass.fromJSON(data)
  };

  Object.defineProperties( Selection.prototype, prototypeAccessors$1 );


  Selection.prototype._isSelection = true;


  var NullSelection = (function (Selection) {
    function NullSelection () {
      Selection.apply(this, arguments);
    }

    if ( Selection ) NullSelection.__proto__ = Selection;
    NullSelection.prototype = Object.create( Selection && Selection.prototype );
    NullSelection.prototype.constructor = NullSelection;

    NullSelection.prototype.isNull = function isNull () {
      return true
    };

    NullSelection.prototype.getType = function getType () {
      return 'null'
    };

    NullSelection.prototype.toJSON = function toJSON () {
      return null
    };

    NullSelection.prototype.clone = function clone () {
      return this
    };

    return NullSelection;
  }(Selection));



  Selection.nullSelection = Object.freeze(new NullSelection());

  function isArrayEqual(arr1, arr2) {
    if (arr1 === arr2) { return true }
    if (!isArray(arr1) || !isArray(arr2)) { return false }
    if (arr1.length !== arr2.length) { return false }
    var L = arr1.length;
    for (var i = 0; i < L; i++) {
      if (arr1[i] !== arr2[i]) { return false }
    }
    return true
  }

  var Coordinate = function Coordinate(path, offset) {
    
    if (arguments[0] === 'SKIP') { return }
    if (arguments.length === 1) {
      var data = arguments[0];
      this.path = data.path;
      this.offset = data.offset;
    } else {
      this.path = path;
      this.offset = offset;
    }
    if (!isArray(this.path)) {
      throw new Error('Invalid arguments: path should be an array.')
    }
    if (!isNumber(this.offset) || this.offset < 0) {
      throw new Error('Invalid arguments: offset must be a positive number.')
    }
  };

  Coordinate.prototype.equals = function equals (other) {
    return (other === this ||
      (isArrayEqual(other.path, this.path) && other.offset === this.offset) )
  };

  Coordinate.prototype.withCharPos = function withCharPos (offset) {
    return new Coordinate(this.path, offset)
  };

  Coordinate.prototype.getNodeId = function getNodeId () {
    return this.path[0]
  };

  Coordinate.prototype.getPath = function getPath () {
    return this.path
  };

  Coordinate.prototype.getOffset = function getOffset () {
    return this.offset
  };

  Coordinate.prototype.toJSON = function toJSON () {
    return {
      path: this.path.slice(),
      offset: this.offset
    }
  };

  Coordinate.prototype.toString = function toString () {
    return "(" + this.path.join('.') + ", " + this.offset + ")"
  };

  Coordinate.prototype.isPropertyCoordinate = function isPropertyCoordinate () {
    return this.path.length > 1
  };

  Coordinate.prototype.isNodeCoordinate = function isNodeCoordinate () {
    return this.path.length === 1
  };

  Coordinate.prototype.hasSamePath = function hasSamePath (other) {
    return isArrayEqual(this.path, other.path)
  };

  Coordinate.prototype._isCoordinate = true;

  var PropertySelection = (function (Selection$$1) {
    function PropertySelection(path, startOffset, endOffset, reverse, containerId, surfaceId) {
      Selection$$1.call(this);

      if (arguments.length === 1) {
        var data = arguments[0];
        path = data.path;
        startOffset = data.startOffset;
        endOffset = data.endOffset;
        reverse = data.reverse;
        containerId = data.containerId;
        surfaceId = data.surfaceId;
      }

      if (!path || !isNumber(startOffset)) {
        throw new Error('Invalid arguments: `path` and `startOffset` are mandatory');
      }

      this.start = new Coordinate(path, startOffset);
      this.end = new Coordinate(path, isNumber(endOffset) ? endOffset : startOffset);

      
      this.reverse = Boolean(reverse);

      this.containerId = containerId;

      
      this.surfaceId = surfaceId;
    }

    if ( Selection$$1 ) PropertySelection.__proto__ = Selection$$1;
    PropertySelection.prototype = Object.create( Selection$$1 && Selection$$1.prototype );
    PropertySelection.prototype.constructor = PropertySelection;

    var prototypeAccessors = { path: { configurable: true },startOffset: { configurable: true },endOffset: { configurable: true } };

    prototypeAccessors.path.get = function () {
      return this.start.path
    };

    prototypeAccessors.startOffset.get = function () {
      console.warn('DEPRECATED: Use sel.start.offset instead');
      return this.start.offset
    };

    prototypeAccessors.endOffset.get = function () {
      console.warn('DEPRECATED: Use sel.end.offset instead');
      return this.end.offset
    };

    
    PropertySelection.prototype.toJSON = function toJSON () {
      return {
        type: 'property',
        path: this.start.path,
        startOffset: this.start.offset,
        endOffset: this.end.offset,
        reverse: this.reverse,
        containerId: this.containerId,
        surfaceId: this.surfaceId
      }
    };

    PropertySelection.prototype.isPropertySelection = function isPropertySelection () {
      return true
    };

    PropertySelection.prototype.getType = function getType () {
      return 'property'
    };

    PropertySelection.prototype.isNull = function isNull () {
      return false
    };

    PropertySelection.prototype.isCollapsed = function isCollapsed () {
      return this.start.offset === this.end.offset;
    };

    PropertySelection.prototype.isReverse = function isReverse () {
      return this.reverse
    };

    PropertySelection.prototype.equals = function equals (other) {
      return (
        Selection$$1.prototype.equals.call(this, other) &&
        (this.start.equals(other.start) && this.end.equals(other.end))
      )
    };

    PropertySelection.prototype.toString = function toString () {
      
      return [
        "PropertySelection(", JSON.stringify(this.path), ", ",
        this.start.offset, " -> ", this.end.offset,
        (this.reverse?", reverse":""),
        (this.surfaceId?(", "+this.surfaceId):""),
        ")"
      ].join('')
    };

    
    PropertySelection.prototype.collapse = function collapse (direction) {
      var offset;
      if (direction === 'left') {
        offset = this.start.offset;
      } else {
        offset = this.end.offset;
      }
      return this.createWithNewRange(offset, offset)
    };

    
    

    
    PropertySelection.prototype.getPath = function getPath () {
      return this.start.path;
    };

    PropertySelection.prototype.getNodeId = function getNodeId () {
      return this.start.path[0];
    };

    
    PropertySelection.prototype.isInsideOf = function isInsideOf (other, strict) {
      if (other.isNull()) { return false }
      if (other.isContainerSelection()) {
        return other.contains(this, strict)
      }
      if (strict) {
        return (isArrayEqual(this.path, other.path) &&
          this.start.offset > other.start.offset &&
          this.end.offset < other.end.offset);
      } else {
        return (isArrayEqual(this.path, other.path) &&
          this.start.offset >= other.start.offset &&
          this.end.offset <= other.end.offset);
      }
    };

    
    PropertySelection.prototype.contains = function contains (other, strict) {
      if (other.isNull()) { return false }
      return other.isInsideOf(this, strict)
    };

    
    PropertySelection.prototype.overlaps = function overlaps (other, strict) {
      if (other.isNull()) { return false }
      if (other.isContainerSelection()) {
        
        return other.overlaps(this)
      }
      if (!isArrayEqual(this.path, other.path)) { return false }
      if (strict) {
        return (! (this.start.offset>=other.end.offset||this.end.offset<=other.start.offset) );
      } else {
        return (! (this.start.offset>other.end.offset||this.end.offset<other.start.offset) );
      }
    };

    
    PropertySelection.prototype.isRightAlignedWith = function isRightAlignedWith (other) {
      if (other.isNull()) { return false }
      if (other.isContainerSelection()) {
        
        return other.isRightAlignedWith(this)
      }
      return (isArrayEqual(this.path, other.path) &&
        this.end.offset === other.end.offset);
    };

    
    PropertySelection.prototype.isLeftAlignedWith = function isLeftAlignedWith (other) {
      if (other.isNull()) { return false }
      if (other.isContainerSelection()) {
        
        return other.isLeftAlignedWith(this)
      }
      return (isArrayEqual(this.path, other.path) &&
        this.start.offset === other.start.offset);
    };

    
    PropertySelection.prototype.expand = function expand (other) {
      if (other.isNull()) { return this }

      
      
      
      if (other.isContainerSelection()) {
        return other.expand(this)
      }
      if (!isArrayEqual(this.path, other.path)) {
        throw new Error('Can not expand PropertySelection to a different property.')
      }
      var newStartOffset = Math.min(this.start.offset, other.start.offset);
      var newEndOffset = Math.max(this.end.offset, other.end.offset);
      return this.createWithNewRange(newStartOffset, newEndOffset);
    };

    
    PropertySelection.prototype.truncateWith = function truncateWith (other) {
      if (other.isNull()) { return this }
      if (other.isInsideOf(this, 'strict')) {
        
        throw new Error('Can not truncate with a contained selections')
      }
      if (!this.overlaps(other)) {
        return this
      }
      var otherStartOffset, otherEndOffset;
      if (other.isPropertySelection()) {
        otherStartOffset = other.start.offset;
        otherEndOffset = other.end.offset;
      } else if (other.isContainerSelection()) {
        
        if (isArrayEqual(other.start.path, this.start.path)) {
          otherStartOffset = other.start.offset;
        } else {
          otherStartOffset = this.start.offset;
        }
        if (isArrayEqual(other.end.path, this.start.path)) {
          otherEndOffset = other.end.offset;
        } else {
          otherEndOffset = this.end.offset;
        }
      } else {
        return this
      }

      var newStartOffset;
      var newEndOffset;
      if (this.start.offset > otherStartOffset && this.end.offset > otherEndOffset) {
        newStartOffset = otherEndOffset;
        newEndOffset = this.end.offset;
      } else if (this.start.offset < otherStartOffset && this.end.offset < otherEndOffset) {
        newStartOffset = this.start.offset;
        newEndOffset = otherStartOffset;
      } else if (this.start.offset === otherStartOffset) {
        if (this.end.offset <= otherEndOffset) {
          return Selection$$1.nullSelection;
        } else {
          newStartOffset = otherEndOffset;
          newEndOffset = this.end.offset;
        }
      } else if (this.end.offset === otherEndOffset) {
        if (this.start.offset >= otherStartOffset) {
          return Selection$$1.nullSelection;
        } else {
          newStartOffset = this.start.offset;
          newEndOffset = otherStartOffset;
        }
      } else if (other.contains(this)) {
        return Selection$$1.nullSelection
      } else {
        
        throw new Error('Illegal state.')
      }
      return this.createWithNewRange(newStartOffset, newEndOffset)
    };

    
    PropertySelection.prototype.createWithNewRange = function createWithNewRange (startOffset, endOffset) {
      var sel = new PropertySelection(this.path, startOffset, endOffset, false, this.containerId, this.surfaceId);
      var doc = this._internal.doc;
      if (doc) {
        sel.attach(doc);
      }
      return sel
    };

    PropertySelection.prototype._clone = function _clone () {
      return new PropertySelection(this.start.path, this.start.offset, this.end.offset, this.reverse, this.containerId, this.surfaceId);
    };

    Object.defineProperties( PropertySelection.prototype, prototypeAccessors );

    return PropertySelection;
  }(Selection));

  PropertySelection.fromJSON = function(json) {
    return new PropertySelection(json)
  };

  var ContainerSelection = (function (Selection$$1) {
    function ContainerSelection(containerId, startPath, startOffset, endPath, endOffset, reverse, surfaceId) {
      Selection$$1.call(this);

      if (arguments.length === 1) {
        var data = arguments[0];
        containerId = data.containerId;
        startPath = data.startPath;
        startOffset = data.startOffset;
        endPath = data.endPath;
        endOffset = data.endOffset;
        reverse = data.reverse;
        surfaceId = data.surfaceId;
      }

      
      this.containerId = containerId;
      if (!this.containerId) { throw new Error('Invalid arguments: `containerId` is mandatory') }

      this.start = new Coordinate(startPath, startOffset);
      this.end = new Coordinate(isNil(endPath) ? startPath : endPath, isNil(endOffset) ? startOffset : endOffset);

      this.reverse = Boolean(reverse);

      this.surfaceId = surfaceId;
    }

    if ( Selection$$1 ) ContainerSelection.__proto__ = Selection$$1;
    ContainerSelection.prototype = Object.create( Selection$$1 && Selection$$1.prototype );
    ContainerSelection.prototype.constructor = ContainerSelection;

    var prototypeAccessors = { startPath: { configurable: true },startOffset: { configurable: true },endPath: { configurable: true },endOffset: { configurable: true },path: { configurable: true } };

    

    prototypeAccessors.startPath.get = function () {
      console.warn('DEPRECATED: use sel.start.path instead.');
      return this.start.path
    };

    prototypeAccessors.startOffset.get = function () {
      console.warn('DEPRECATED: use sel.start.offset instead.');
      return this.start.offset
    };

    prototypeAccessors.endPath.get = function () {
      console.warn('DEPRECATED: use sel.end.path instead.');
      return this.end.path
    };

    prototypeAccessors.endOffset.get = function () {
      console.warn('DEPRECATED: use sel.end.offset instead.');
      return this.end.offset
    };

    

    ContainerSelection.prototype.toJSON = function toJSON () {
      return {
        type: 'container',
        containerId: this.containerId,
        startPath: this.start.path,
        startOffset: this.start.offset,
        endPath: this.end.path,
        endOffset: this.end.offset,
        reverse: this.reverse,
        surfaceId: this.surfaceId
      }
    };

    ContainerSelection.prototype.isContainerSelection = function isContainerSelection () {
      return true
    };

    ContainerSelection.prototype.getType = function getType () {
      return 'container'
    };

    ContainerSelection.prototype.isNull = function isNull () {
      return false
    };

    ContainerSelection.prototype.isCollapsed = function isCollapsed () {
      return this.start.equals(this.end)
    };

    ContainerSelection.prototype.isReverse = function isReverse () {
      return this.reverse
    };

    ContainerSelection.prototype.equals = function equals (other) {
      return (
        Selection$$1.prototype.equals.call(this, other) &&
        this.containerId === other.containerId &&
        (this.start.equals(other.start) && this.end.equals(other.end))
      )
    };

    ContainerSelection.prototype.toString = function toString () {
      
      return [
        "ContainerSelection(",
        this.containerId, ", ",
        JSON.stringify(this.start.path), ", ", this.start.offset,
        " -> ",
        JSON.stringify(this.end.path), ", ", this.end.offset,
        (this.reverse?", reverse":""),
        (this.surfaceId?(", "+this.surfaceId):""),
        ")"
      ].join('')
    };

    
    ContainerSelection.prototype.getContainer = function getContainer () {
      if (!this._internal.container) {
        this._internal.container = this.getDocument().get(this.containerId);
      }
      return this._internal.container
    };

    ContainerSelection.prototype.isInsideOf = function isInsideOf (other, strict) {
      
      
      if (other.isNull()) { return false }
      strict = Boolean(strict);
      var r1 = this._range(this);
      var r2 = this._range(other);
      return (r2.start.isBefore(r1.start, strict) &&
        r1.end.isBefore(r2.end, strict))
    };

    ContainerSelection.prototype.contains = function contains (other, strict) {
      
      
      if (other.isNull()) { return false }
      strict = Boolean(strict);
      var r1 = this._range(this);
      var r2 = this._range(other);
      return (r1.start.isBefore(r2.start, strict) &&
        r2.end.isBefore(r1.end, strict))
    };

    ContainerSelection.prototype.containsNode = function containsNode (nodeId, strict) {
      var container = this.getContainer();
      if (!container.contains(nodeId)) { return false }
      var coor = new Coordinate([nodeId], 0);
      var address = container.getAddress(coor);
      var r = this._range(this);
      
      var contained = r.start.isBefore(address, strict);
      if (contained) {
        address.offset = 1;
        contained = r.end.isAfter(address, strict);
      }
      return contained
    };

    ContainerSelection.prototype.overlaps = function overlaps (other) {
      var r1 = this._range(this);
      var r2 = this._range(other);
      
      return !(r1.end.isBefore(r2.start, false) ||
        r2.end.isBefore(r1.start, false))
    };

    ContainerSelection.prototype.isLeftAlignedWith = function isLeftAlignedWith (other) {
      var r1 = this._range(this);
      var r2 = this._range(other);
      return r1.start.isEqual(r2.start)
    };

    ContainerSelection.prototype.isRightAlignedWith = function isRightAlignedWith (other) {
      var r1 = this._range(this);
      var r2 = this._range(other);
      return r1.end.isEqual(r2.end)
    };

    
    ContainerSelection.prototype.collapse = function collapse (direction) {
      var coor;
      if (direction === 'left') {
        coor = this.start;
      } else {
        coor = this.end;
      }
      return _createNewSelection(this, coor, coor)
    };

    ContainerSelection.prototype.expand = function expand (other) {
      var r1 = this._range(this);
      var r2 = this._range(other);
      var start;
      var end;

      if (r1.start.isEqual(r2.start)) {
        start = new Coordinate(this.start.path, Math.min(this.start.offset, other.start.offset));
      } else if (r1.start.isAfter(r2.start)) {
        start = new Coordinate(other.start.path, other.start.offset);
      } else {
        start = this.start;
      }
      if (r1.end.isEqual(r2.end)) {
        end = new Coordinate(this.end.path, Math.max(this.end.offset, other.end.offset));
      } else if (r1.end.isBefore(r2.end, false)) {
        end = new Coordinate(other.end.path, other.end.offset);
      } else {
        end = this.end;
      }

      return _createNewSelection(this, start, end)
    };

    ContainerSelection.prototype.truncateWith = function truncateWith (other) {
      if (other.isInsideOf(this, 'strict')) {
        
        throw new Error('Can not truncate with a contained selections')
      }
      if (!this.overlaps(other)) {
        return this
      }
      var r1 = this._range(this);
      var r2 = this._range(other);
      var start, end;
      if (r2.start.isBefore(r1.start, 'strict') && r2.end.isBefore(r1.end, 'strict')) {
        start = other.end;
        end = this.end;
      } else if (r1.start.isBefore(r2.start, 'strict') && r1.end.isBefore(r2.end, 'strict')) {
        start = this.start;
        end = other.start;
      } else if (r1.start.isEqual(r2.start)) {
        if (r2.end.isBefore(r1.end, 'strict')) {
          start = other.end;
          end = this.end;
        } else {
          
          return Selection$$1.nullSelection
        }
      } else if (r1.end.isEqual(r2.end)) {
        if (r1.start.isBefore(r2.start, 'strict')) {
          start = this.start;
          end = other.start;
        } else {
          
          return Selection$$1.nullSelection
        }
      } else if (this.isInsideOf(other)) {
        return Selection$$1.nullSelection
      } else {
        throw new Error('Could not determine coordinates for truncate. Check input')
      }
      return _createNewSelection(this, start, end)
    };

    
    ContainerSelection.prototype.getNodeIds = function getNodeIds () {
      var container = this.getContainer();
      var startPos = container.getPosition(this.start.path[0]);
      var endPos = container.getPosition(this.end.path[0]);
      return container.getContent().slice(startPos, endPos+1)
    };

    
    ContainerSelection.prototype.splitIntoPropertySelections = function splitIntoPropertySelections () {
      var sels = [];
      var fragments = this.getFragments();
      fragments.forEach(function(fragment) {
        if (fragment instanceof Selection$$1.Fragment) {
          sels.push(
            new PropertySelection(fragment.path, fragment.startOffset,
              fragment.endOffset, false, this.containerId, this.surfaceId)
          );
        }
      }.bind(this));
      return sels
    };

    ContainerSelection.prototype._clone = function _clone () {
      return new ContainerSelection(this)
    };

    ContainerSelection.prototype._range = function _range (sel) {
      
      
      
      if (sel._internal.addressRange) {
        return sel._internal.addressRange
      }

      var container = this.getContainer();
      var startAddress = container.getAddress(sel.start);
      var endAddress;
      if (sel.isCollapsed()) {
        endAddress = startAddress;
      } else {
        endAddress = container.getAddress(sel.end);
      }
      var addressRange = {
        start: startAddress,
        end: endAddress
      };
      if (sel._isContainerSelection) {
        sel._internal.addressRange = addressRange;
      }
      return addressRange
    };

    prototypeAccessors.path.get = function () {
      throw new Error('ContainerSelection has no path property. Use startPath and endPath instead')
    };

    Object.defineProperties( ContainerSelection.prototype, prototypeAccessors );

    return ContainerSelection;
  }(Selection));

  ContainerSelection.prototype._isContainerSelection = true;

  ContainerSelection.fromJSON = function(properties) {
    var sel = new ContainerSelection(properties);
    return sel
  };

  function _createNewSelection(containerSel, start, end) {
    var newSel;

    if (start === end) {
      newSel = new PropertySelection({
        path: start.path,
        startOffset: start.offset,
        endOffset: start.offset,
        containerId: containerSel.containerId,
        surfaceId: containerSel.surfaceId
      });
    } else {
      newSel = new ContainerSelection(
        containerSel.containerId,
        start.path, start.offset, end.path, end.offset,
        false, containerSel.surfaceId
      );
    }
    
    var doc = containerSel._internal.doc;
    if (doc) {
      newSel.attach(doc);
    }
    return newSel
  }

  var NodeSelection = (function (Selection$$1) {
    function NodeSelection(containerId, nodeId, mode, reverse, surfaceId) {
      Selection$$1.call(this);

      if (arguments.length === 1) {
        var data = arguments[0];
        containerId = data.containerId;
        nodeId = data.nodeId;
        mode = data.mode;
        reverse = data.reverse;
        surfaceId = data.surfaceId;
      }

      if (!isString(containerId)) {
        throw new Error("'containerId' is mandatory.");
      }
      if (!isString(nodeId)) {
        throw new Error("'nodeId' is mandatory.");
      }
      mode = mode || "full";

      this.containerId = containerId;
      this.nodeId = nodeId;
      this.mode = mode;
      this.reverse = Boolean(reverse);
      this.surfaceId = surfaceId;

      this.start = new Coordinate([nodeId], 0);
      this.end = new Coordinate([nodeId], 1);
    }

    if ( Selection$$1 ) NodeSelection.__proto__ = Selection$$1;
    NodeSelection.prototype = Object.create( Selection$$1 && Selection$$1.prototype );
    NodeSelection.prototype.constructor = NodeSelection;

    NodeSelection.prototype.equals = function equals (other) {
      return (
        Selection$$1.prototype.equals.call(this, other) &&
        this.nodeId === other.nodeId &&
        this.mode === other.mode
      )
    };

    NodeSelection.prototype.isNodeSelection = function isNodeSelection () {
      return true;
    };

    NodeSelection.prototype.getType = function getType () {
      return 'node';
    };

    NodeSelection.prototype.getNodeId = function getNodeId () {
      return this.nodeId;
    };

    NodeSelection.prototype.isFull = function isFull () {
      return this.mode === 'full';
    };

    NodeSelection.prototype.isBefore = function isBefore () {
      return this.mode === 'before';
    };

    NodeSelection.prototype.isAfter = function isAfter () {
      return this.mode === 'after';
    };

    NodeSelection.prototype.isCollapsed = function isCollapsed () {
      return this.mode !== 'full';
    };

    NodeSelection.prototype.toJSON = function toJSON () {
      return {
        type: 'node',
        nodeId: this.nodeId,
        mode: this.mode,
        reverse: this.reverse,
        containerId: this.containerId,
        surfaceId: this.surfaceId
      };
    };

    NodeSelection.prototype.toString = function toString () {
      
      return [
        "NodeSelection(",
        this.containerId, ".", this.nodeId, ", ",
        this.mode, ", ",
        (this.reverse?", reverse":""),
        (this.surfaceId?(", "+this.surfaceId):""),
        ")"
      ].join('');
    };

    NodeSelection.prototype.collapse = function collapse (direction) {
      if (direction === 'left') {
        if (this.isBefore()) {
          return this;
        } else {
          return new NodeSelection(this.containerId, this.nodeId, 'before', this.reverse, this.surfaceId);
        }
      } else if (direction === 'right') {
        if (this.isAfter()) {
          return this;
        } else {
          return new NodeSelection(this.containerId, this.nodeId, 'after', this.reverse, this.surfaceId);
        }
      } else {
        throw new Error("'direction' must be either 'left' or 'right'");
      }
    };

    NodeSelection.prototype._getCoordinate = function _getCoordinate () {
      if (this.mode === 'before') {
        return new Coordinate([this.nodeId], 0);
      } else if (this.mode === 'after') {
        return new Coordinate([this.nodeId], 1);
      }
    };

    NodeSelection.prototype._clone = function _clone () {
      return new NodeSelection(this);
    };

    return NodeSelection;
  }(Selection));

  NodeSelection.prototype._isNodeSelection = true;

  NodeSelection.fromJSON = function(json) {
    return new NodeSelection(json);
  };


  NodeSelection._createFromCoordinate = function(coor) {
    var containerId = coor.containerId;
    var nodeId = coor.getNodeId();
    var mode = coor.offset === 0 ? 'before' : 'after';
    return new NodeSelection(containerId, nodeId, mode, false);
  };

  var CustomSelection = (function (Selection$$1) {
    function CustomSelection(customType, data, surfaceId) {
      Selection$$1.call(this);

      if (arguments.length === 1) {
        var _data = arguments[0];
        customType = _data.customType;
        data = _data.data;
        surfaceId = _data.surfaceId;
      }

      this.customType = customType;
      this.data = data || {};
      this.surfaceId = surfaceId;
    }

    if ( Selection$$1 ) CustomSelection.__proto__ = Selection$$1;
    CustomSelection.prototype = Object.create( Selection$$1 && Selection$$1.prototype );
    CustomSelection.prototype.constructor = CustomSelection;

    CustomSelection.prototype.isCustomSelection = function isCustomSelection () {
      return true;
    };

    CustomSelection.prototype.getType = function getType () {
      return 'custom';
    };

    CustomSelection.prototype.getCustomType = function getCustomType () {
      return this.customType;
    };

    CustomSelection.prototype.toJSON = function toJSON () {
      return {
        type: 'custom',
        customType: this.customType,
        data: cloneDeep(this.data),
        surfaceId: this.surfaceId
      };
    };

    CustomSelection.prototype.toString = function toString () {
      
      return [
        'CustomSelection(',
        this.customType,', ',
        JSON.stringify(this.data),
        ")"
      ].join('');
    };

    CustomSelection.prototype.equals = function equals (other) {
      return (
        Selection$$1.prototype.equals.call(this, other) &&
        other.isCustomSelection() &&
        isEqual(this.data, other.data)
      );
    };

    CustomSelection.prototype._clone = function _clone () {
      return new CustomSelection(this)
    };

    return CustomSelection;
  }(Selection));

  CustomSelection.prototype._isCustomSelection = true;

  CustomSelection.fromJSON = function(json) {
    return new CustomSelection(json);
  };

  function fromJSON(json) {
    if (!json) { return Selection.nullSelection }
    var type = json.type;
    switch(type) {
      case 'property':
        return PropertySelection.fromJSON(json)
      case 'container':
        return ContainerSelection.fromJSON(json)
      case 'node':
        return NodeSelection.fromJSON(json)
      case 'custom':
        return CustomSelection.fromJSON(json)
      default:
        
        return Selection.nullSelection
    }
  }


  function isFirst(doc, coor) {
    if (coor.isNodeCoordinate() && coor.offset === 0) { return true }
    var node = doc.get(coor.path[0]).getContainerRoot();
    if (node.isText() && coor.offset === 0) { return true }
    if (node.isList()) {
      var itemId = coor.path[0];
      if (node.items[0] === itemId && coor.offset === 0) { return true }
    }
  }


  function isLast(doc, coor) {
    if (coor.isNodeCoordinate() && coor.offset > 0) { return true }
    var node = doc.get(coor.path[0]).getContainerRoot();
    if (node.isText() && coor.offset >= node.getLength()) { return true }
    if (node.isList()) {
      var itemId = coor.path[0];
      var item = doc.get(itemId);
      if (last$1(node.items) === itemId && coor.offset === item.getLength()) { return true }
    }
  }

  function isEntirelySelected(doc, node, start, end) {
    var ref = getRangeInfo(doc, node, start, end);
    var isEntirelySelected = ref.isEntirelySelected;
    return isEntirelySelected
  }

  function getRangeInfo(doc, node, start, end) {
    var isFirst = true;
    var isLast = true;
    if (node.isText()) {
      if (start && start.offset !== 0) { isFirst = false; }
      if (end && end.offset < node.getLength()) { isLast = false; }
    } else if (node.isList()) {
      if (start) {
        var itemId = start.path[0];
        var itemPos = node.getItemPosition(itemId);
        if (itemPos > 0 || start.offset !== 0) { isFirst = false; }
      }
      if (end) {
        var itemId$1 = end.path[0];
        var itemPos$1 = node.getItemPosition(itemId$1);
        var item = doc.get(itemId$1);
        if (itemPos$1 < node.items.length-1 || end.offset < item.getLength()) { isLast = false; }
      }
    }
    var isEntirelySelected = isFirst && isLast;
    return {isFirst: isFirst, isLast: isLast, isEntirelySelected: isEntirelySelected}
  }

  function setCursor(tx, node, containerId, mode) {
    if (node.isText()) {
      var offset = 0;
      if (mode === 'after') {
        var text = node.getText();
        offset = text.length;
      }
      tx.setSelection({
        type: 'property',
        path: node.getPath(),
        startOffset: offset,
        containerId: containerId
      });
    } else if (node.isList()) {
      var item, offset$1;
      if (mode === 'after') {
        item = node.getLastItem();
        offset$1 = item.getLength();
      } else {
        item = node.getFirstItem();
        offset$1 = 0;
      }
      tx.setSelection({
        type: 'property',
        path: item.getPath(),
        startOffset: offset$1,
        containerId: containerId
      });
    } else {
      tx.setSelection({
        type: 'node',
        containerId: containerId,
        nodeId: node.id,
        
        
        
      });
    }
  }

  function selectNode(tx, nodeId, containerId) {
    tx.setSelection(createNodeSelection({ doc: tx, nodeId: nodeId, containerId: containerId }));
  }

  function createNodeSelection(ref) {
    var doc = ref.doc;
    var nodeId = ref.nodeId;
    var containerId = ref.containerId;
    var mode = ref.mode;
    var reverse = ref.reverse;
    var surfaceId = ref.surfaceId;

    var node = doc.get(nodeId);
    if (!node) { return Selection.nullSelection }
    node = node.getContainerRoot();
    if (node.isText()) {
      return new PropertySelection({
        path: node.getPath(),
        startOffset: mode === 'after' ? node.getLength() : 0,
        endOffset: mode === 'before' ? 0 : node.getLength(),
        reverse: reverse,
        containerId: containerId,
        surfaceId: surfaceId
      })
    } else if (node.isList() && node.getLength()>0) {
      var first = node.getFirstItem();
      var last = node.getLastItem();
      var start = {
        path: first.getPath(),
        offset: 0
      };
      var end = {
        path: last.getPath(),
        offset: last.getLength()
      };
      if (mode === 'after') { start = end; }
      else if (mode === 'before') { end = start; }
      return new ContainerSelection({
        startPath: start.path,
        startOffset: start.offset,
        endPath: end.path,
        endOffset: end.offset,
        reverse: reverse,
        containerId: containerId,
        surfaceId: surfaceId
      })
    } else {
      return new NodeSelection({ nodeId: nodeId, mode: mode, reverse: reverse, containerId: containerId, surfaceId: surfaceId })
    }
  }

  function stepIntoIsolatedNode(editorSession, comp) {
    
    
    if (comp.grabFocus()) { return true }

    
    var surface = comp.find('.sc-surface');
    if (surface) {
      
      if (surface._isTextPropertyEditor) {
        var doc = editorSession.getDocument();
        var path = surface.getPath();
        var text = doc.get(path, 'strict');
        editorSession.setSelection({
          type: 'property',
          path: path,
          startOffset: text.length,
          surfaceId: surface.id
        });
        return true
      } else if (surface._isContainerEditor) {
        var container = surface.getContainer();
        if (container.length > 0) {
          var first = container.getChildAt(0);
          setCursor(editorSession, first, container.id, 'after');
        }
        return true
      }
    }
    return false
  }

  function augmentSelection(selData, oldSel) {
    
    if (selData && oldSel && !selData.surfaceId && !oldSel.isNull()) {
      selData.containerId = selData.containerId || oldSel.containerId;
      selData.surfaceId = selData.surfaceId || oldSel.surfaceId;
    }
    return selData
  }

  var selectionHelpers = /*#__PURE__*/Object.freeze({
    fromJSON: fromJSON,
    isFirst: isFirst,
    isLast: isLast,
    isEntirelySelected: isEntirelySelected,
    getRangeInfo: getRangeInfo,
    setCursor: setCursor,
    selectNode: selectNode,
    createNodeSelection: createNodeSelection,
    stepIntoIsolatedNode: stepIntoIsolatedNode,
    augmentSelection: augmentSelection
  });

  var DocumentChange = function DocumentChange(ops, before, after) {
    if (arguments.length === 1 && isPlainObject(arguments[0])) {
      var data = arguments[0];
      
      this.sha = data.sha;
      
      this.timestamp = data.timestamp;
      
      this.before = data.before || {};
      
      this.ops = data.ops;
      this.info = data.info; 
      
      this.after = data.after || {};
    } else if (arguments.length === 3) {
      this.sha = uuid();
      this.info = {};
      this.timestamp = Date.now();
      this.ops = ops.slice(0);
      this.before = before || {};
      this.after = after || {};
    } else {
      throw new Error('Illegal arguments.')
    }
    
    this.updated = null;
    
    this.created = null;
    
    this.deleted = null;
  };


  DocumentChange.prototype._extractInformation = function _extractInformation (doc) {
    
    
    if (this._extracted) { return }

    var ops = this.ops;
    var created = {};
    var deleted = {};
    var updated = {};
    var affectedContainerAnnos = [];

    
    function _checkAnnotation(op) {
      switch (op.type) {
        case "create":
        case "delete": {
          var node = op.val;
          if (node.hasOwnProperty('start')) {
            updated[node.start.path] = true;
          }
          if (node.hasOwnProperty('end')) {
            updated[node.end.path] = true;
          }
          break
        }
        case "update":
        case "set": {
          
          var node$1 = doc.get(op.path[0]);
          if (node$1) {
            if (node$1.isPropertyAnnotation()) {
              updated[node$1.start.path] = true;
            } else if (node$1.isContainerAnnotation()) {
              affectedContainerAnnos.push(node$1);
            }
          }
          break
        }
        default:
          
          throw new Error('Illegal state')
      }
    }

    for (var i = 0; i < ops.length; i++) {
      var op = ops[i];
      if (op.type === "create") {
        created[op.val.id] = op.val;
        delete deleted[op.val.id];
      }
      if (op.type === "delete") {
        delete created[op.val.id];
        deleted[op.val.id] = op.val;
      }
      if (op.type === "set" || op.type === "update") {
        updated[op.path] = true;
        
        updated[op.path[0]] = true;
      }
      _checkAnnotation(op);
    }

    affectedContainerAnnos.forEach(function(anno) {
      var container = doc.get(anno.containerId, 'strict');
      var startPos = container.getPosition(anno.start.path[0]);
      var endPos = container.getPosition(anno.end.path[0]);
      for (var pos = startPos; pos <= endPos; pos++) {
        var node = container.getChildAt(pos);
        var path = (void 0);
        if (node.isText()) {
          path = node.getPath();
        } else {
          path = [node.id];
        }
        if (!deleted[node.id]) {
          updated[path] = true;
        }
      }
    });

    
    if(Object.keys(deleted).length > 0) {
      forEach(updated, function(_, key) {
        var nodeId = key.split(',')[0];
        if (deleted[nodeId]) {
          delete updated[key];
        }
      });
    }

    this.created = created;
    this.deleted = deleted;
    this.updated = updated;

    this._extracted = true;
  };

  DocumentChange.prototype.invert = function invert () {
      var this$1 = this;

    
    var copy = this.toJSON();
    copy.ops = [];
    
    var tmp = copy.before;
    copy.before = copy.after;
    copy.after = tmp;
    var inverted = DocumentChange.fromJSON(copy);
    var ops = [];
    for (var i = this.ops.length - 1; i >= 0; i--) {
      ops.push(this$1.ops[i].invert());
    }
    inverted.ops = ops;
    return inverted
  };


  DocumentChange.prototype.isAffected = function isAffected (path) {
    console.error('DEPRECATED: use change.hasUpdated() instead');
    return this.hasUpdated(path)
  };

  DocumentChange.prototype.isUpdated = function isUpdated (path) {
    console.error('DEPRECATED: use change.hasUpdated() instead');
    return this.hasUpdated(path)
  };


  DocumentChange.prototype.hasUpdated = function hasUpdated (path) {
    return this.updated[path]
  };

  DocumentChange.prototype.hasDeleted = function hasDeleted (id) {
    return this.deleted[id]
  };

  DocumentChange.prototype.serialize = function serialize () {
    
    

    var opSerializer = new OperationSerializer();
    var data = this.toJSON();
    data.ops = this.ops.map(function(op) {
      return opSerializer.serialize(op)
    });
    return JSON.stringify(data)
  };

  DocumentChange.prototype.clone = function clone$$1 () {
    return DocumentChange.fromJSON(this.toJSON())
  };

  DocumentChange.prototype.toJSON = function toJSON () {
    var data = {
      
      sha: this.sha,
      
      before: clone(this.before),
      ops: map(this.ops, function(op) {
        return op.toJSON()
      }),
      info: this.info,
      
      after: clone(this.after),
    };

    
    
    data.after.selection = undefined;
    data.before.selection = undefined;

    var sel = this.before.selection;
    if (sel && sel._isSelection) {
      data.before.selection = sel.toJSON();
    }
    sel = this.after.selection;
    if (sel && sel._isSelection) {
      data.after.selection = sel.toJSON();
    }
    return data
  };

  DocumentChange.deserialize = function(str) {
    var opSerializer = new OperationSerializer();
    var data = JSON.parse(str);
    data.ops = data.ops.map(function(opData) {
      return opSerializer.deserialize(opData)
    });
    if (data.before.selection) {
      data.before.selection = fromJSON(data.before.selection);
    }
    if (data.after.selection) {
      data.after.selection = fromJSON(data.after.selection);
    }
    return new DocumentChange(data)
  };

  DocumentChange.fromJSON = function(data) {
    
    var change = cloneDeep(data);
    change.ops = data.ops.map(function(opData) {
      return ObjectOperation.fromJSON(opData)
    });
    change.before.selection = fromJSON(data.before.selection);
    change.after.selection = fromJSON(data.after.selection);
    return new DocumentChange(change)
  };

  function transformDocumentChange(A, B) {
    _transformInplaceBatch(A, B);
  }

  function transformSelection(sel, a) {
    var newSel = sel.clone();
    var hasChanged = _transformSelectionInplace(newSel, a);
    if (hasChanged) {
      return newSel
    } else {
      return sel
    }
  }

  function _transformInplaceSingle(a, b) {
    for (var i = 0; i < a.ops.length; i++) {
      var a_op = a.ops[i];
      for (var j = 0; j < b.ops.length; j++) {
        var b_op = b.ops[j];
        
        
        ObjectOperation.transform(a_op, b_op, {inplace: true});
      }
    }
    if (a.before) {
      _transformSelectionInplace(a.before.selection, b);
    }
    if (a.after) {
      _transformSelectionInplace(a.after.selection, b);
    }
    if (b.before) {
      _transformSelectionInplace(b.before.selection, a);
    }
    if (b.after) {
      _transformSelectionInplace(b.after.selection, a);
    }
  }

  function _transformInplaceBatch(A, B) {
    if (!isArray(A)) {
      A = [A];
    }
    if (!isArray(B)) {
      B = [B];
    }
    for (var i = 0; i < A.length; i++) {
      var a = A[i];
      for (var j = 0; j < B.length; j++) {
        var b = B[j];
        _transformInplaceSingle(a,b);
      }
    }
  }

  function _transformSelectionInplace(sel, a) {
    if (!sel || (!sel.isPropertySelection() && !sel.isContainerSelection()) ) {
      return false
    }
    var ops = a.ops;
    var hasChanged = false;
    var isCollapsed = sel.isCollapsed();
    for(var i=0; i<ops.length; i++) {
      var op = ops[i];
      hasChanged |= _transformCoordinateInplace(sel.start, op);
      if (!isCollapsed) {
        hasChanged |= _transformCoordinateInplace(sel.end, op);
      } else {
        if (sel.isContainerSelection()) {
          sel.end.path = sel.start.path;
        }
        sel.end.offset = sel.start.offset;
      }
    }
    return hasChanged
  }

  function _transformCoordinateInplace(coor, op) {
    if (!isEqual(op.path, coor.path)) { return false }
    var hasChanged = false;
    if (op.type === 'update' && op.propertyType === 'string') {
      var diff = op.diff;
      var newOffset;
      if (diff.isInsert() && diff.pos <= coor.offset) {
        newOffset = coor.offset + diff.str.length;
        
        coor.offset = newOffset;
        hasChanged = true;
      } else if (diff.isDelete() && diff.pos <= coor.offset) {
        newOffset = Math.max(diff.pos, coor.offset - diff.str.length);
        
        coor.offset = newOffset;
        hasChanged = true;
      }
    }
    return hasChanged
  }

  var operationHelpers = /*#__PURE__*/Object.freeze({
    transformDocumentChange: transformDocumentChange,
    transformSelection: transformSelection
  });

  var CollabEngine = (function (EventEmitter$$1) {
    function CollabEngine(documentEngine) {
      EventEmitter$$1.call(this);
      this.documentEngine = documentEngine;
      
      this._collaborators = {};
    }

    if ( EventEmitter$$1 ) CollabEngine.__proto__ = EventEmitter$$1;
    CollabEngine.prototype = Object.create( EventEmitter$$1 && EventEmitter$$1.prototype );
    CollabEngine.prototype.constructor = CollabEngine;

    
    CollabEngine.prototype._register = function _register (collaboratorId, documentId, collaboratorInfo) {
      var collaborator = this._collaborators[collaboratorId];

      if (!collaborator) {
        collaborator = this._collaborators[collaboratorId] = {
          collaboratorId: collaboratorId,
          documents: {}
        };
      }

      
      collaborator.info = collaboratorInfo;

      
      collaborator.documents[documentId] = {};
    };

    
    CollabEngine.prototype._unregister = function _unregister (collaboratorId, documentId) {
      var collaborator = this._collaborators[collaboratorId];
      delete collaborator.documents[documentId];
      var docCount = Object.keys(collaborator.documents).length;
      
      if (docCount === 0) {
        delete this._collaborators[collaboratorId];
      }
    };

    
    CollabEngine.prototype.getDocumentIds = function getDocumentIds (collaboratorId) {
      var collaborator = this._collaborators[collaboratorId];
      if (!collaborator) {
        
        
        return []
      }
      return Object.keys(collaborator.documents)
    };

    
    CollabEngine.prototype.getCollaborators = function getCollaborators (documentId, collaboratorId) {
      var collaborators = {};
      forEach(this._collaborators, function(collab) {
        var doc = collab.documents[documentId];
        if (doc && collab.collaboratorId !== collaboratorId) {
          var entry = {
            
            collaboratorId: collab.collaboratorId
          };
          entry = Object.assign({}, collab.info, entry);
          collaborators[collab.collaboratorId] = entry;
        }
      });
      return collaborators
    };

    
    CollabEngine.prototype.getCollaboratorIds = function getCollaboratorIds (documentId, collaboratorId) {
      var collaborators = this.getCollaborators(documentId, collaboratorId);
      return map(collaborators, function(c) {
        return c.collaboratorId
      })
    };

    
    CollabEngine.prototype.sync = function sync (ref, cb) {
      var documentId = ref.documentId;
      var version = ref.version;
      var change = ref.change;
      var collaboratorId = ref.collaboratorId;

      this._sync({documentId: documentId, version: version, change: change}, function(err, result) {
        if (err) { return cb(err) }
        
        this._register(collaboratorId, documentId);
        cb(null, result);
      }.bind(this));
    };

    
    CollabEngine.prototype._sync = function _sync (ref, cb) {
      var this$1 = this;
      var documentId = ref.documentId;
      var version = ref.version;
      var change = ref.change;

      this.documentEngine.getVersion(documentId, function (err, serverVersion) {
        if (version > serverVersion) {
          cb(new SubstanceError('InvalidVersionError', {
            message: 'Client version greater than server version'
          }));
        } else if (change && serverVersion === version) {
          this$1._syncFF({documentId: documentId, version: version, change: change}, cb);
        } else if (change && serverVersion > version) {
          this$1._syncRB({documentId: documentId, version: version, change: change}, cb);
        } else if (!change) {
          
          this$1._syncPullOnly({documentId: documentId, version: version, change: change}, cb);
        } else {
          console.warn('Unhandled case');
        }
      });
    };

    CollabEngine.prototype._syncPullOnly = function _syncPullOnly (ref, cb) {
      var this$1 = this;
      var documentId = ref.documentId;
      var version = ref.version;
      var change = ref.change;

      console.warn('This code is not yet tested');
      this.documentEngine.getChanges(documentId, version, function (err, changes) {
        var serverChange;

        
        if (changes.length > 0) {
          var ops = [];
          changes.forEach(function (change) {
            ops = ops.concat(change.ops);
          });
          serverChange = new DocumentChange(ops, {}, {});
          serverChange = this$1.serializeChange(serverChange);
        }
        cb(null, {
          serverChange: serverChange,
          change: change,
          version: version
        });
      });
    };

    
    CollabEngine.prototype._syncFF = function _syncFF (ref, cb) {
      var documentId = ref.documentId;
      var change = ref.change;

      this.documentEngine.addChange(documentId, change, function (err, serverVersion) {
        if (err) { return cb(err) }
        cb(null, {
          change: change, 
          serverChange: null,
          version: serverVersion
        });
      });
    };

    
    CollabEngine.prototype._syncRB = function _syncRB (ref, cb) {
      var documentId = ref.documentId;
      var change = ref.change;
      var version = ref.version;

      this._rebaseChange({documentId: documentId, change: change, version: version}, function(err, rebased) {
        
        if (err) { return cb(err) }
        
        this.documentEngine.addChange(
          documentId,
          change,
          function(err, serverVersion) {
            if (err) { return cb(err) }
            cb(null, {
              change: rebased.change,
              
              serverChange: rebased.serverChange,
              version: serverVersion
            });
          }
        );
      }.bind(this));
    };

    
    CollabEngine.prototype._rebaseChange = function _rebaseChange (ref, cb) {
      var documentId = ref.documentId;
      var change = ref.change;
      var version = ref.version;

      this.documentEngine.getChanges(documentId, version, function(err, result, serverVersion) {
        
        var B = result.map(this.deserializeChange);
        var a = this.deserializeChange(change);
        
        transformDocumentChange(a, B);
        var ops = B.reduce(function(ops, change) {
          return ops.concat(change.ops)
        }, []);
        var serverChange = new DocumentChange(ops, {}, {});

        cb(null, {
          change: this.serializeChange(a),
          serverChange: this.serializeChange(serverChange),
          version: serverVersion
        });
      }.bind(this));
    };

    
    CollabEngine.prototype.disconnect = function disconnect (args) {
      this._unregister(args.collaboratorId, args.documentId);
    };

    
    CollabEngine.prototype.serializeChange = function serializeChange (change) {
      return change.toJSON()
    };

    
    CollabEngine.prototype.deserializeChange = function deserializeChange (serializedChange) {
      var ch = DocumentChange.fromJSON(serializedChange);
      return ch
    };

    return CollabEngine;
  }(EventEmitter));

  var CollabServer = (function (Server$$1) {
    function CollabServer(config) {
      Server$$1.call(this, config);

      this.scope = 'substance/collab';
      this.configurator = config.configurator;
      this.documentEngine = this.configurator.getDocumentEngine();
      this.collabEngine = new CollabEngine(this.documentEngine);
    }

    if ( Server$$1 ) CollabServer.__proto__ = Server$$1;
    CollabServer.prototype = Object.create( Server$$1 && Server$$1.prototype );
    CollabServer.prototype.constructor = CollabServer;

    
    CollabServer.prototype._error = function _error (req, res, err) {
      console.error(err);
      res.error({
        scope: this.scope,
        type: 'error',
        error: {
          name: req.message.type+'Error',
          cause: {
            name: err.name
          }
        },
        documentId: req.message.documentId
      });
      this.next(req, res);
    };

    
    CollabServer.prototype.authenticate = function authenticate (req, res) {
      var this$1 = this;

      if (this.config.authenticate) {
        this.config.authenticate(req, function (err, session) {
          if (err) {
            console.error(err);
            
            this$1._error(req, res, new SubstanceError('AuthenticationError', {cause: err}));
            return
          }
          req.setAuthenticated(session);
          this$1.next(req, res);
        });
      } else {
        Server$$1.prototype.authenticate.apply(this, arguments);
      }
    };

    
    CollabServer.prototype.enhanceRequest = function enhanceRequest (req, res) {
      var this$1 = this;

      if (this.config.enhanceRequest) {
        this.config.enhanceRequest(req, function (err) {
          if (err) {
            console.error('enhanceRequest returned an error', err);
            this$1._error(req, res, err);
            return
          }
          req.setEnhanced();
          this$1.next(req, res);
        });
      } else {
        Server$$1.prototype.enhanceRequest.apply(this, arguments);
      }
    };

    
    CollabServer.prototype.onDisconnect = function onDisconnect (collaboratorId) {
      
      
      var documentIds = this.collabEngine.getDocumentIds(collaboratorId);
      documentIds.forEach(function(documentId) {
        this._disconnectDocument(collaboratorId, documentId);
      }.bind(this));
    };

    
    CollabServer.prototype.execute = function execute (req, res) {
      var msg = req.message;
      var method = this[msg.type];

      if (method) {
        method.call(this, req, res);
      } else {
        console.error('Method', msg.type, 'not implemented for CollabServer');
      }
    };

    
    CollabServer.prototype.sync = function sync (req, res) {
      var this$1 = this;

      var args = req.message;

      
      this.collabEngine.sync(args, function (err, result) {
        
        if (err) {
          this$1._error(req, res, err);
          return
        }

        
        var collaborators = this$1.collabEngine.getCollaborators(args.documentId, args.collaboratorId);

        
        res.send({
          scope: this$1.scope,
          type: 'syncDone',
          documentId: args.documentId,
          version: result.version,
          serverChange: result.serverChange,
        });

        
        forEach(collaborators, function (collaborator) {
          this$1.send(collaborator.collaboratorId, {
            scope: this$1.scope,
            type: 'update',
            documentId: args.documentId,
            version: result.version,
            change: result.change
          });
        });
        this$1.next(req, res);
      });
    };

    
    CollabServer.prototype.disconnect = function disconnect (req, res) {
      var args = req.message;
      var collaboratorId = args.collaboratorId;
      var documentId = args.documentId;
      this._disconnectDocument(collaboratorId, documentId);
      
      res.send({
        scope: this.scope,
        type: 'disconnectDone',
        documentId: args.documentId
      });
      this.next(req, res);
    };

    CollabServer.prototype._disconnectDocument = function _disconnectDocument (collaboratorId, documentId) {
      
      this.collabEngine.disconnect({
        documentId: documentId,
        collaboratorId: collaboratorId
      });
    };

    return CollabServer;
  }(Server));

  function request(method, url, data, cb) {
    var request = new XMLHttpRequest();
    request.open(method, url, true);
    request.setRequestHeader('Content-Type', 'application/json; charset=UTF-8');
    request.onload = function() {
      if (request.status >= 200 && request.status < 400) {
        var res = request.responseText;
        if(isJson(res)) { res = JSON.parse(res); }
        cb(null, res);
      } else {
        return cb(new Error('Request failed. Returned status: ' + request.status))
      }
    };

    if (data) {
      request.send(JSON.stringify(data));
    } else {
      request.send();
    }
  }

  function isJson(str) {
    try {
      JSON.parse(str);
    } catch (e) {
      return false
    }
    return true
  }

  var DocumentClient = function DocumentClient(config) {
    this.config = config;
  };


  DocumentClient.prototype.createDocument = function createDocument (newDocument, cb) {
    request('POST', this.config.httpUrl, newDocument, cb);
  };



  DocumentClient.prototype.getDocument = function getDocument (documentId, cb) {
    request('GET', this.config.httpUrl+documentId, null, cb);
  };



  DocumentClient.prototype.deleteDocument = function deleteDocument (documentId, cb) {
    request('DELETE', this.config.httpUrl+documentId, null, cb);
  };

  var ClientConnection = (function (EventEmitter$$1) {
    function ClientConnection(config) {
      EventEmitter$$1.call(this);

      this.config = config;
      this._onMessage = this._onMessage.bind(this);
      this._onConnectionOpen = this._onConnectionOpen.bind(this);
      this._onConnectionClose = this._onConnectionClose.bind(this);

      
      this._connect();
    }

    if ( EventEmitter$$1 ) ClientConnection.__proto__ = EventEmitter$$1;
    ClientConnection.prototype = Object.create( EventEmitter$$1 && EventEmitter$$1.prototype );
    ClientConnection.prototype.constructor = ClientConnection;

    ClientConnection.prototype._createWebSocket = function _createWebSocket () {
      throw SubstanceError('AbstractMethodError')
    };

    
    ClientConnection.prototype._connect = function _connect () {
      this.ws = this._createWebSocket();
      this.ws.addEventListener('open', this._onConnectionOpen);
      this.ws.addEventListener('close', this._onConnectionClose);
      this.ws.addEventListener('message', this._onMessage);
    };

    
    ClientConnection.prototype._disconnect = function _disconnect () {
      this.ws.removeEventListener('message', this._onMessage);
      this.ws.removeEventListener('open', this._onConnectionOpen);
      this.ws.removeEventListener('close', this._onConnectionClose);
      this.ws = null;
    };

    
    ClientConnection.prototype._onConnectionOpen = function _onConnectionOpen () {
      this.emit('open');
    };

    
    ClientConnection.prototype._onConnectionClose = function _onConnectionClose () {
      this._disconnect();
      this.emit('close');
      console.info('websocket connection closed. Attempting to reconnect in 5s.');
      setTimeout(function() {
        this._connect();
      }.bind(this), 5000);
    };

    
    ClientConnection.prototype._onMessage = function _onMessage (msg) {
      msg = this.deserializeMessage(msg.data);
      this.emit('message', msg);
    };

    
    ClientConnection.prototype.send = function send (msg) {
      if (!this.isOpen()) {
        console.warn('Message could not be sent. Connection is not open.', msg);
        return
      }
      this.ws.send(this.serializeMessage(msg));
    };

    
    ClientConnection.prototype.isOpen = function isOpen () {
      return this.ws && this.ws.readyState === 1
    };

    ClientConnection.prototype.serializeMessage = function serializeMessage (msg) {
      return JSON.stringify(msg)
    };

    ClientConnection.prototype.deserializeMessage = function deserializeMessage (msg) {
      return JSON.parse(msg)
    };

    return ClientConnection;
  }(EventEmitter));

  var WebSocketConnection = (function (ClientConnection$$1) {
    function WebSocketConnection () {
      ClientConnection$$1.apply(this, arguments);
    }

    if ( ClientConnection$$1 ) WebSocketConnection.__proto__ = ClientConnection$$1;
    WebSocketConnection.prototype = Object.create( ClientConnection$$1 && ClientConnection$$1.prototype );
    WebSocketConnection.prototype.constructor = WebSocketConnection;

    WebSocketConnection.prototype._createWebSocket = function _createWebSocket () {
      return new window.WebSocket(this.config.wsUrl);
    };

    return WebSocketConnection;
  }(ClientConnection));

  function computeSnapshot(jsonDoc, changeset) {
    
    jsonDoc = cloneDeep(jsonDoc);
    var nodes = jsonDoc.nodes;
    changeset.forEach(function (change) {
      change.ops.forEach(function (opData) {
        try {
          var op = ObjectOperation.fromJSON(opData);
          op.apply(nodes);
        } catch (err) {
          console.error(err, opData);
        }
      });
    });
    return jsonDoc
  }

  var EMPTY_DOC = { nodes: {} };


  var SnapshotEngine = function SnapshotEngine(config) {
    this.changeStore = config.changeStore;
    this.snapshotStore = config.snapshotStore;
  };


  SnapshotEngine.prototype.getSnapshot = function getSnapshot (documentId, version, cb) {
      var this$1 = this;

    var jsonDoc = EMPTY_DOC;
    this._getClosestSnapshot(documentId, version, function (err, snapshot, closestVersion) {
      if (err) {
        return cb(err)
      }
      if (snapshot && version === closestVersion) {
        
        return cb(null, snapshot, version)
      }
      var knownVersion;
      if (snapshot) {
        knownVersion = closestVersion;
      } else {
        knownVersion = 0; 
      }
      if (snapshot) {
        jsonDoc = snapshot;
      }
      
      this$1.changeStore.getChanges(documentId, knownVersion, version, function (err, changes) {
        if (err) { return cb(err) }
        if (changes.length < (version - knownVersion)) {
          return cb('Changes missing for reconstructing version '+ version)
        }
        jsonDoc = computeSnapshot(jsonDoc, changes);
        cb(null, jsonDoc, version);
      });
    });
  };


  SnapshotEngine.prototype.createSnapshot = function createSnapshot (documentId, version, cb) {
      var this$1 = this;

    this.getSnapshot(documentId, version, function (err, snapshot) {
      if (err) { return cb(err) }
      this$1.snapshotStore.saveSnapshot(documentId, version, snapshot, cb);
    });
  };

  SnapshotEngine.prototype._getClosestSnapshot = function _getClosestSnapshot (documentId, version, cb) {
      var this$1 = this;

    var closestVersion;

    this.snapshotStore.getVersions(documentId, function (err, versions) {
      if (err) { return cb(err) }

      if(versions.length === 0) {
        return cb(null, undefined)
      }

      if (versions.indexOf(version) >= 0) {
        closestVersion = version;
      } else {
        
        var smallerVersions = versions.filter(function(v) {
          return parseInt(v, 10) < version
        });
        
        closestVersion = Math.max.apply(null, smallerVersions);
      }
      if (!closestVersion) {
        return cb(null, undefined)
      }
      this$1.snapshotStore.getSnapshot(documentId, closestVersion, cb);
    });
  };

  var DocumentEngine = (function (EventEmitter$$1) {
    function DocumentEngine(config) {
      EventEmitter$$1.call(this);
      this.changeStore = config.changeStore;
      this.snapshotStore = config.snapshotStore;
      
      
      this.snapshotFrequency = config.snapshotFrequency || 1;
      this.snapshotEngine = new SnapshotEngine({
        changeStore: this.changeStore,
        snapshotStore: this.snapshotStore
      });
    }

    if ( EventEmitter$$1 ) DocumentEngine.__proto__ = EventEmitter$$1;
    DocumentEngine.prototype = Object.create( EventEmitter$$1 && EventEmitter$$1.prototype );
    DocumentEngine.prototype.constructor = DocumentEngine;

    
    DocumentEngine.prototype.createDocument = function createDocument (documentId, initialChange, cb) {
      this.addChange(documentId, initialChange, cb);
    };

    
    DocumentEngine.prototype.getDocument = function getDocument (documentId, version, cb) {
      var this$1 = this;

      if (typeof version === 'function') {
        cb = version;
        version = undefined;
      }
      if (!documentId) {
        throw new Error('Invalid Arguments')
      }
      if (version === undefined) {
        this.getVersion(documentId, function (err, version) {
          if (err) { return cb(err) }
          this$1.snapshotEngine.getSnapshot(documentId, version, cb);
        });
      } else {
        this.snapshotEngine.getSnapshot(documentId, version, cb);
      }
    };

    
    DocumentEngine.prototype.deleteDocument = function deleteDocument (documentId, cb) {
      this.changeStore.deleteChanges(documentId, function (err) {
        if (err) {
          return cb(new Error('Deleting changes failed'))
        }
      });
    };

    
    DocumentEngine.prototype.documentExists = function documentExists (documentId, cb) {
      this.getVersion(documentId, function (err, version) {
        if (version >= 0) {
          cb(null, true);
        } else {
          cb(null, false);
        }
      });
    };

    
    DocumentEngine.prototype.getChanges = function getChanges (documentId, sinceVersion, toVersion, cb) {
      this.changeStore.getChanges(documentId, sinceVersion, toVersion, cb);
    };

    
    DocumentEngine.prototype.getVersion = function getVersion (documentId, cb) {
      this.changeStore.getVersion(documentId, cb);
    };

    
    DocumentEngine.prototype.requestSnapshot = function requestSnapshot (documentId, version, cb) {
      if (version % this.snapshotFrequency === 0) {
        this.snapshotEngine.createSnapshot(documentId, version, cb);
      } else {
        cb(null); 
      }
    };

    
    DocumentEngine.prototype.addChange = function addChange (documentId, change, cb) {
      var this$1 = this;

      this.changeStore.addChange(documentId, change, function (err, newVersion) {
        if (err) { return cb(err) }

        this$1.requestSnapshot(documentId, newVersion, function () {
          
          cb(null, newVersion);
        });
      });
    };

    return DocumentEngine;
  }(EventEmitter));

  var CollabServerConfigurator = function CollabServerConfigurator() {
    this.config = {
      heartbeat: 30*1000,
      documentStore: undefined,
      changeStore: undefined,
      snapshotStore: undefined
    };
  };

  CollabServerConfigurator.prototype.setHost = function setHost (host) {
    this.config.host = host;
  };

  CollabServerConfigurator.prototype.setPort = function setPort (port) {
    this.config.port = port;
  };




  CollabServerConfigurator.prototype.setDocumentStore = function setDocumentStore (documentStore) {
    this.config.documentStore = documentStore;
  };

  CollabServerConfigurator.prototype.setChangeStore = function setChangeStore (changeStore) {
    this.config.changeStore = changeStore;
  };

  CollabServerConfigurator.prototype.setSnapshotStore = function setSnapshotStore (snapshotStore) {
    this.config.snapshotStore = snapshotStore;
  };




  CollabServerConfigurator.prototype.getHost = function getHost () {
    return this.config.host
  };

  CollabServerConfigurator.prototype.getPort = function getPort () {
    return this.config.port
  };

  CollabServerConfigurator.prototype.getDocumentStore = function getDocumentStore () {
    return this.config.documentStore
  };

  CollabServerConfigurator.prototype.getChangeStore = function getChangeStore () {
    return this.config.changeStore
  };

  CollabServerConfigurator.prototype.getSnapshotStore = function getSnapshotStore () {
    return this.config.snapshotStore
  };


  CollabServerConfigurator.prototype.getDocumentEngine = function getDocumentEngine () {
    if (!this.documentEngine) {
      this.documentEngine = new DocumentEngine({
        documentStore: this.config.documentStore,
        changeStore: this.config.changeStore,
        snapshotStore: this.config.snapshotStore
      });
    }
    return this.documentEngine
  };


  CollabServerConfigurator.prototype.import = function import$1 (pkg, options) {
    pkg.configure(this, options || {});
    return this
  };

  var CollabClient = (function (EventEmitter$$1) {
    function CollabClient(config) {
      EventEmitter$$1.call(this);

      this.config = config;
      this.connection = config.connection;

      
      this.scope = 'substance/collab';

      
      this._onMessage = this._onMessage.bind(this);
      this._onConnectionOpen = this._onConnectionOpen.bind(this);
      this._onConnectionClose = this._onConnectionClose.bind(this);

      
      this.connection.on('open', this._onConnectionOpen);
      this.connection.on('close', this._onConnectionClose);
      this.connection.on('message', this._onMessage);
    }

    if ( EventEmitter$$1 ) CollabClient.__proto__ = EventEmitter$$1;
    CollabClient.prototype = Object.create( EventEmitter$$1 && EventEmitter$$1.prototype );
    CollabClient.prototype.constructor = CollabClient;

    CollabClient.prototype._onConnectionClose = function _onConnectionClose () {
      this.emit('disconnected');
    };

    CollabClient.prototype._onConnectionOpen = function _onConnectionOpen () {
      this.emit('connected');
    };

    
    CollabClient.prototype._onMessage = function _onMessage (msg) {
      if (msg.scope === this.scope) {
        this.emit('message', msg);
      } else if (msg.scope !== '_internal') {
        console.info('Message ignored. Not sent in hub scope', msg);
      }
    };

    
    CollabClient.prototype.send = function send (msg) {
      if (!this.connection.isOpen()) {
        console.warn('Message could not be sent. Connection not open.', msg);
        return
      }

      msg.scope = this.scope;
      if (this.config.enhanceMessage) {
        msg = this.config.enhanceMessage(msg);
      }
      this.connection.send(msg);
    };

    
    CollabClient.prototype.isConnected = function isConnected () {
      return this.connection.isOpen()
    };

    CollabClient.prototype.dispose = function dispose () {
      this.connection.off(this);
    };

    return CollabClient;
  }(EventEmitter));

  var SnapshotStore = function SnapshotStore(seed) {
    this._snapshots = seed || {};
  };


  SnapshotStore.prototype.getVersions = function getVersions (documentId, cb) {
    var versions = this._getVersions(documentId);
    cb(null, versions);
  };


  SnapshotStore.prototype.getSnapshot = function getSnapshot (documentId, version, cb) {
    if (!arguments.length === 3) {
      throw new Error('Invalid Arguments')
    }
    var docEntry = this._snapshots[documentId];
    if (!docEntry) { return cb(null, undefined) }
    var snapshot = docEntry[version];
    if (snapshot) {
      cb(null, snapshot, version);
    } else {
      cb(null, undefined);
    }
  };


  SnapshotStore.prototype.saveSnapshot = function saveSnapshot (documentId, version, data, cb) {
    if (!documentId || !version || !data) {
      throw new Error('Invalid arguments')
    }
    var docEntry = this._snapshots[documentId];
    if (!docEntry) {
      docEntry = this._snapshots[documentId] = {};
    }
    docEntry[version] = data;
    cb(null, docEntry[version]);
  };


  SnapshotStore.prototype.deleteSnapshot = function deleteSnapshot (documentId, version, cb) {
    var docEntry = this._snapshots[documentId];
    if (!docEntry || !docEntry[version]) {
      return cb(new Error('Snapshot does not exist and can not be deleted'))
    }
    var snapshot = this._snapshots[documentId][version];
    delete this._snapshots[documentId][version];
    cb(null, snapshot);
  };


  SnapshotStore.prototype._getVersions = function _getVersions (documentId) {
    var docEntry = this._snapshots[documentId];
    if (!docEntry) { return [] } 
    return Object.keys(docEntry)
  };

  var CollabServerPackage = {
    name: 'collab-server',
    configure: function (config) {
      config.setChangeStore(new ChangeStore());
      config.setSnapshotStore(new SnapshotStore());
    }
  }

  var DocumentServer = function DocumentServer(params) {
    this.configurator = params.configurator;
    this.engine = this.configurator.getDocumentEngine();
    
    this.path = '/api/documents';
  };


  DocumentServer.prototype.bind = function bind (app) {
    app.post(this.path, this._createDocument.bind(this));
    app.get(this.path + '/:id', this._getDocument.bind(this));
    app.delete(this.path + '/:id', this._deleteDocument.bind(this));
  };


  DocumentServer.prototype._createDocument = function _createDocument (req, res, next) {
    var ref = req.body;
      var documentId = ref.documentId;
      var change = ref.change;
    this.engine.createDocument(documentId, change, function(err, version) {
      if (err) { return next(err) }
      res.json(version);
    });
  };


  DocumentServer.prototype._getDocument = function _getDocument (req, res, next) {
    var documentId = req.params.id;
    this.engine.getDocument(documentId, function(err, jsonDoc, version) {
      if (err) { return next(err) }
      res.json({
        data: jsonDoc,
        version: version
      });
    });
  };


  DocumentServer.prototype._deleteDocument = function _deleteDocument (req, res, next) {
    var documentId = req.params.id;
    this.engine.deleteDocument(documentId, function(err, result) {
      if (err) { return next(err) }
      res.json(result);
    });
  };

  function isFunction$1(f) {
    return typeof f === 'function'
  }

  function deleteFromArray(array, value) {
    if (!array) { return }
    for (var i = 0; i < array.length; i++) {
      if (array[i] === value) {
        array.splice(i, 1);
        i--;
      }
    }
  }

  var TreeNode = function TreeNode () {};



  var TreeIndex = function TreeIndex () {};

  TreeIndex.prototype.get = function get$1 (path) {
    if (arguments.length > 1) {
      path = Array.prototype.slice(arguments, 0);
    }
    if (isString(path)) {
      path = [path];
    }
    return get(this, path);
  };

  TreeIndex.prototype.getAll = function getAll (path) {
    if (arguments.length > 1) {
      path = Array.prototype.slice(arguments, 0);
    }
    if (isString(path)) {
      path = [path];
    }
    if (!isArray(path)) {
      throw new Error('Illegal argument for TreeIndex.get()');
    }
    var node = get(this, path);
    return this._collectValues(node);
  };

  TreeIndex.prototype.set = function set (path, value) {
    if (isString(path)) {
      path = [path];
    }
    setWith(this, path, value, function(val) {
      if (!val) { return new TreeNode(); }
    });
  };

  TreeIndex.prototype.delete = function delete$1 (path) {
    if (isString(path)) {
      delete this[path];
    } else if(path.length === 1) {
      delete this[path[0]];
    } else {
      var key = path[path.length-1];
      path = path.slice(0, -1);
      var parent = get(this, path);
      if (parent) {
        delete parent[key];
      }
    }
  };

  TreeIndex.prototype.clear = function clear () {
    var root = this;
    for (var key in root) {
      if (root.hasOwnProperty(key)) {
        delete root[key];
      }
    }
  };

  TreeIndex.prototype.traverse = function traverse (fn) {
    this._traverse(this, [], fn);
  };

  TreeIndex.prototype.forEach = function forEach () {
      var ref;

      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];
    (ref = this).traverse.apply(ref, args);
  };

  TreeIndex.prototype._traverse = function _traverse (root, path, fn) {
      var this$1 = this;

    var id;
    for (id in root) {
      if (!root.hasOwnProperty(id)) { continue; }
      var child = root[id];
      var childPath = path.concat([id]);
      if (child instanceof TreeNode) {
        this$1._traverse(child, childPath, fn);
      } else {
        fn(child, childPath);
      }
    }
  };

  TreeIndex.prototype._collectValues = function _collectValues (root) {
    
    
    var vals = {};
    this._traverse(root, [], function(val, path) {
      var key = path[path.length-1];
      vals[key] = val;
    });
    return vals;
  };

  var TreeIndexArrays = (function (TreeIndex) {
    function TreeIndexArrays () {
      TreeIndex.apply(this, arguments);
    }

    if ( TreeIndex ) TreeIndexArrays.__proto__ = TreeIndex;
    TreeIndexArrays.prototype = Object.create( TreeIndex && TreeIndex.prototype );
    TreeIndexArrays.prototype.constructor = TreeIndexArrays;

    TreeIndexArrays.prototype.contains = function contains (path) {
      var val = TreeIndex.prototype.get.call(this, path);
      return Boolean(val)
    };

    TreeIndexArrays.prototype.get = function get$$1 (path) {
      var val = TreeIndex.prototype.get.call(this, path);
      if (val instanceof TreeNode) {
        val = val.__values__ || [];
      }
      return val;
    };

    TreeIndexArrays.prototype.set = function set (path, arr) {
      var val = TreeIndex.prototype.get.call(this, path);
      val.__values__ = arr;
    };

    TreeIndexArrays.prototype.add = function add (path, value) {
      if (isString(path)) {
        path = [path];
      }
      if (!isArray(path)) {
        throw new Error('Illegal arguments.');
      }
      var arr;

      
      
      
      
      
      
      setWith(this, path.concat(['__values__','__dummy__']), undefined, function(val, key) {
        if (key === '__values__') {
          if (!val) { val = []; }
          arr = val;
        } else if (!val) {
          val = new TreeNode();
        }
        return val;
      });
      delete arr.__dummy__;
      arr.push(value);
    };

    TreeIndexArrays.prototype.remove = function remove (path, value) {
      var arr = get(this, path);
      if (arr instanceof TreeNode) {
        if (arguments.length === 1) {
          delete arr.__values__;
        } else {
          deleteFromArray(arr.__values__, value);
        }
      }
    };

    TreeIndexArrays.prototype._collectValues = function _collectValues (root) {
      var vals = [];
      this._traverse(root, [], function(val) {
        vals.push(val);
      });
      vals = Array.prototype.concat.apply([], vals);
      return vals
    };

    return TreeIndexArrays;
  }(TreeIndex));

  TreeIndex.Arrays = TreeIndexArrays;

  function filter(iteratee, fn) {
    if (!iteratee) { return [] }
    if (iteratee.constructor.prototype.filter && isFunction$1(iteratee.constructor.prototype.filter)) {
      return iteratee.filter(fn)
    }
    var result = [];
    forEach(iteratee, function (val, key) {
      if (fn(val, key)) {
        result.push(val);
      }
    });
    return result
  }

  function flatten(arr) {
    return Array.prototype.concat.apply([], arr)
  }

  var NodeIndex = function NodeIndex () {};

  NodeIndex.prototype.select = function select (node) { 
    throw new Error('This method is abstract.')
  };

  NodeIndex.prototype.clear = function clear () {
    throw new Error('This method is abstract')
  };


  NodeIndex.prototype.create = function create (node) { 
    throw new Error('This method is abstract.')
  };


  NodeIndex.prototype.delete = function delete$1 (node) { 
    throw new Error('This method is abstract.')
  };

  NodeIndex.prototype.set = function set (node, path, newValue, oldValue) {
    this.update(node, path, newValue, oldValue);
  };


  NodeIndex.prototype.update = function update (node, path, newValue, oldValue) { 
    throw new Error('This method is abstract.')
  };


  NodeIndex.prototype.reset = function reset (data) {
    this.clear();
    this._initialize(data);
  };


  NodeIndex.prototype.clone = function clone () {
    var NodeIndexClass = this.constructor;
    var clone = new NodeIndexClass();
    return clone
  };

  NodeIndex.prototype._initialize = function _initialize (data) {
    forEach(data.getNodes(), function(node) {
      if (this.select(node)) {
        this.create(node);
      }
    }.bind(this));
  };


  NodeIndex.create = function(prototype) {
    var index = Object.assign(new NodeIndex(), prototype);
    index.clone = function() {
      return NodeIndex.create(prototype)
    };
    return index
  };


  NodeIndex.filterByType = function(type) {
    return function(node) {
      return node.isInstanceOf(type)
    }
  };

  var DocumentIndex = (function (NodeIndex$$1) {
  	function DocumentIndex () {
  		NodeIndex$$1.apply(this, arguments);
  	}if ( NodeIndex$$1 ) DocumentIndex.__proto__ = NodeIndex$$1;
  	DocumentIndex.prototype = Object.create( NodeIndex$$1 && NodeIndex$$1.prototype );
  	DocumentIndex.prototype.constructor = DocumentIndex;

  	

  	return DocumentIndex;
  }(NodeIndex));

  var annotationHelpers = {
    insertedText: insertedText,
    deletedText: deletedText,
    transferAnnotations: transferAnnotations,
    expandAnnotation: expandAnnotation,
    fuseAnnotation: fuseAnnotation,
    truncateAnnotation: truncateAnnotation
  }

  function insertedText(doc, coordinate, length) {
    if (!length) { return; }
    var index = doc.getIndex('annotations');
    var annotations = index.get(coordinate.path);
    for (var i = 0; i < annotations.length; i++) {
      var anno = annotations[i];
      var pos = coordinate.offset;
      var start = anno.start.offset;
      var end = anno.end.offset;
      var newStart = start;
      var newEnd = end;
      if ( (pos < start) ||
           (pos === start) ) {
        newStart += length;
      }
      
      if ( (pos < end) ||
           (pos === end && !anno.isInline()) ) {
        newEnd += length;
      }
      
      if (newStart !== start) {
        doc.set([anno.id, 'start', 'offset'], newStart);
      }
      if (newEnd !== end) {
        doc.set([anno.id, 'end', 'offset'], newEnd);
      }
    }

    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
  }

  function deletedText(doc, path, startOffset, endOffset) {
    if (startOffset === endOffset) { return; }
    var index = doc.getIndex('annotations');
    var annotations = index.get(path);
    var length = endOffset - startOffset;
    for (var i = 0; i < annotations.length; i++) {
      var anno = annotations[i];
      var pos1 = startOffset;
      var pos2 = endOffset;
      var start = anno.start.offset;
      var end = anno.end.offset;
      var newStart = start;
      var newEnd = end;
      if (pos2 <= start) {
        newStart -= length;
        newEnd -= length;
        doc.set([anno.id, 'start', 'offset'], newStart);
        doc.set([anno.id, 'end', 'offset'], newEnd);
      } else {
        if (pos1 <= start) {
          newStart = start - Math.min(pos2-pos1, start-pos1);
        }
        if (pos1 <= end) {
          newEnd = end - Math.min(pos2-pos1, end-pos1);
        }
        
        if (start !== end && newStart === newEnd) {
          doc.delete(anno.id);
        } else {
          
          if (start !== newStart) {
            doc.set([anno.id, 'start', 'offset'], newStart);
          }
          if (end !== newEnd) {
            doc.set([anno.id, 'end', 'offset'], newEnd);
          }
        }
      }
    }
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
  }


  function transferAnnotations(doc, path, offset, newPath, newOffset) {
    var index = doc.getIndex('annotations');
    var annotations = index.get(path, offset);
    for (var i = 0; i < annotations.length; i++) {
      var a = annotations[i];
      var isInside = (offset > a.start.offset && offset < a.end.offset);
      var start = a.start.offset;
      var end = a.end.offset;
      
      if (isInside) {
        
        if (a.canSplit()) {
          var newAnno = a.toJSON();
          newAnno.id = uuid(a.type + "_");
          newAnno.start.path = newPath;
          newAnno.start.offset = newOffset;
          newAnno.end.path = newPath;
          newAnno.end.offset = newOffset + a.end.offset - offset;
          doc.create(newAnno);
        }
        
        var newStartOffset = a.start.offset;
        var newEndOffset = offset;
        
        if (newEndOffset === newStartOffset) {
          doc.delete(a.id);
        }
        
        else {
          
          if (newStartOffset !== start) {
            doc.set([a.id, 'start', 'offset'], newStartOffset);
          }
          if (newEndOffset !== end) {
            doc.set([a.id, 'end', 'offset'], newEndOffset);
          }
        }
      }
      
      else if (a.start.offset >= offset) {
        
        
        
        doc.set([a.id, 'start', 'path'], newPath);
        doc.set([a.id, 'start', 'offset'], newOffset + a.start.offset - offset);
        doc.set([a.id, 'end', 'path'], newPath);
        doc.set([a.id, 'end', 'offset'], newOffset + a.end.offset - offset);
      }
    }

    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
  }


  function truncateAnnotation(tx, anno, sel) {
    if (!sel || !sel._isSelection) { throw new Error('Argument "selection" is required.') }
    if (!anno || !anno.isAnnotation()) { throw new Error('Argument "anno" is required and must be an annotation.') }
    var annoSel = anno.getSelection();
    var newAnnoSel = annoSel.truncateWith(sel);
    anno._updateRange(tx, newAnnoSel);
    return anno
  }


  function expandAnnotation(tx, anno, sel) {
    if (!sel || !sel._isSelection) { throw new Error('Argument "selection" is required.') }
    if (!anno || !anno.isAnnotation()) { throw new Error('Argument "anno" is required and must be an annotation.') }
    var annoSel = anno.getSelection();
    var newAnnoSel = annoSel.expand(sel);
    anno._updateRange(tx, newAnnoSel);
    return anno
  }


  function fuseAnnotation(tx, annos) {
    if (!isArray(annos) || annos.length < 2) {
      throw new Error('fuseAnnotation(): at least two annotations are necessary.')
    }
    var sel, annoType;
    annos.forEach(function(anno, idx) {
      if (idx === 0) {
        sel = anno.getSelection();
        annoType = anno.type;
      } else {
        if (anno.type !== annoType) {
          throw new Error('fuseAnnotation(): all annotations must be of the same type.')
        }
        sel = sel.expand(anno.getSelection());
      }
    });
    
    for (var i = 1; i < annos.length; i++) {
      tx.delete(annos[i].id);
    }
    expandAnnotation(tx, annos[0], sel);
    tx.setSelection(sel);
  }

  function getPropertyAnnotationsForSelection(doc, sel, options) {
    options = options || {};
    if (!sel.isPropertySelection()) {
      return []
    }
    var path = sel.getPath();
    var annotations = doc.getIndex('annotations').get(path, sel.start.offset, sel.end.offset);
    if (options.type) {
      annotations = filter(annotations, DocumentIndex.filterByType(options.type));
    }
    return annotations
  }


  function getContainerAnnotationsForSelection(doc, sel, containerId, options) {
    
    
    
    
    
    if (!containerId) {
      throw new Error("'containerId' is required.")
    }
    options = options || {};
    var index = doc.getIndex('container-annotations');
    var annotations = [];
    if (index) {
      annotations = index.get(containerId, options.type);
      annotations = filter(annotations, function(anno) {
        return sel.overlaps(anno.getSelection())
      });
    }
    return annotations
  }


  function isContainerAnnotation(doc, type) {
    var schema = doc.getSchema();
    return schema.isInstanceOf(type, 'container-annotation')
  }


  function getTextForSelection(doc, sel) {
    if (!sel || sel.isNull()) {
      return ""
    } else if (sel.isPropertySelection()) {
      var text = doc.get(sel.start.path);
      return text.substring(sel.start.offset, sel.end.offset)
    } else if (sel.isContainerSelection()) {
      var result = [];
      var nodeIds = sel.getNodeIds();
      var L = nodeIds.length;
      for (var i = 0; i < L; i++) {
        var id = nodeIds[i];
        var node = doc.get(id);
        if (node.isText()) {
          var text$1 = node.getText();
          if (i === L-1) {
            text$1 = text$1.slice(0, sel.end.offset);
          }
          if (i === 0) {
            text$1 = text$1.slice(sel.start.offset);
          }
          result.push(text$1);
        }
      }
      return result.join('\n')
    }
  }

  function getMarkersForSelection(doc, sel) {
    
    if (!sel || !sel.isPropertySelection()) { return [] }
    var path = sel.getPath();
    
    var markers = doc.getIndex('markers').get(path);
    var filtered = filter(markers, function(m) {
      return m.containsSelection(sel)
    });
    return filtered
  }


  function deleteNode(doc, node) {
    
    if (!node) {
      console.warn('Invalid arguments');
      return
    }
    
    if (node.isText()) {
      
      var annos = doc.getIndex('annotations').get(node.id);
      for (var i = 0; i < annos.length; i++) {
        doc.delete(annos[i].id);
      }
    }
    
    
    
    var nodeSchema = node.getSchema();
    forEach(nodeSchema, function (prop) {
      if ((prop.isReference() && prop.isOwned()) || (prop.type === 'file')) {
        if (prop.isArray()) {
          var ids = node[prop.name];
          ids.forEach(function (id) {
            deleteNode(doc, doc.get(id));
          });
        } else {
          deleteNode(doc, doc.get(node[prop.name]));
        }
      }
    });
    doc.delete(node.id);
  }


  function copyNode(node) {
    var nodes = [];
    
    var nodeSchema = node.getSchema();
    var doc = node.getDocument();
    forEach(nodeSchema, function (prop) {
      
      
      if ((prop.isReference() && prop.isOwned()) || (prop.type === 'file')) {
        var val = node[prop.name];
        nodes.push(_copyChildren(val));
      }
    });
    nodes.push(node.toJSON());
    var annotationIndex = node.getDocument().getIndex('annotations');
    var annotations = annotationIndex.get([node.id]);
    forEach(annotations, function(anno) {
      nodes.push(anno.toJSON());
    });
    var result = flatten(nodes).filter(Boolean);
    
    return result

    function _copyChildren(val) {
      if (!val) { return null }
      if (isArray(val)) {
        return flatten(val.map(_copyChildren))
      } else {
        var id = val;
        if (!id) { return null }
        var child = doc.get(id);
        if (!child) { return }
        return copyNode(child)
      }
    }
  }


  function deleteTextRange(doc, start, end) {
    if (!start) {
      start = {
        path: end.path,
        offset: 0
      };
    }
    var path = start.path;
    var text = doc.get(path);
    if (!end) {
      end = {
        path: start.path,
        offset: text.length
      };
    }
    
    if (!isArrayEqual(start.path, end.path)) {
      throw new Error('start and end must be on one property')
    }
    var startOffset = start.offset;
    if (startOffset < 0) { throw new Error("start offset must be >= 0") }
    var endOffset = end.offset;
    if (endOffset > text.length) { throw new Error("end offset must be smaller than the text length") }

    doc.update(path, { type: 'delete', start: startOffset, end: endOffset });
    
    var annos = doc.getAnnotations(path);
    annos.forEach(function(anno) {
      var annoStart = anno.start.offset;
      var annoEnd = anno.end.offset;
      
      if (annoEnd<=startOffset) {
        return
      }
      
      else if (annoStart>=endOffset) {
        doc.update([anno.id, 'start'], { type: 'shift', value: startOffset-endOffset });
        doc.update([anno.id, 'end'], { type: 'shift', value: startOffset-endOffset });
      }
      
      else if (annoStart>=startOffset && annoEnd<=endOffset) {
        doc.delete(anno.id);
      }
      
      else if (annoStart>=startOffset && annoEnd>=endOffset) {
        if (annoStart>startOffset) {
          doc.update([anno.id, 'start'], { type: 'shift', value: startOffset-annoStart });
        }
        doc.update([anno.id, 'end'], { type: 'shift', value: startOffset-endOffset });
      }
      
      else if (annoStart<=startOffset && annoEnd<=endOffset) {
        doc.update([anno.id, 'end'], { type: 'shift', value: startOffset-annoEnd });
      }
      
      else if (annoStart<startOffset && annoEnd >= endOffset) {
        doc.update([anno.id, 'end'], { type: 'shift', value: startOffset-endOffset });
      }
      else {
        console.warn('TODO: handle annotation update case.');
      }
    });
  }

  function deleteListRange(doc, list, start, end) {
    var assign, assign$1, assign$2;

    if (doc !== list.getDocument()) {
      list = doc.get(list.id);
    }
    var startItem, endItem;
    if (!start) {
      startItem = list.getItemAt(0);
      start = {
        path: startItem.getPath(),
        offset: 0
      };
    } else {
      startItem = doc.get(start.path[0]);
    }
    if (!end) {
      endItem = list.getLastItem();
      end = {
        path: endItem.getPath(),
        offset: endItem.getLength()
      };
    } else {
      endItem = doc.get(end.path[0]);
    }
    var startPos = list.getItemPosition(startItem);
    var endPos = list.getItemPosition(endItem);
    
    if (startPos === endPos) {
      deleteTextRange(doc, start, end);
      return
    }
    
    if (startPos > endPos) {
      (assign = [end, start], start = assign[0], end = assign[1]);
      (assign$1 = [endPos, startPos], startPos = assign$1[0], endPos = assign$1[1]);
      (assign$2 = [endItem, startItem], startItem = assign$2[0], endItem = assign$2[1]);
    }
    var firstEntirelySelected = isEntirelySelected(doc, startItem, start, null);
    var lastEntirelySelected = isEntirelySelected(doc, endItem, null, end);

    
    if (lastEntirelySelected) {
      list.removeItemAt(endPos);
      deleteNode(doc, endItem);
    } else {
      deleteTextRange(doc, null, end);
    }

    
    var items = list.getItems();
    for (var i = endPos-1; i > startPos; i--) {
      var item = items[i];
      list.removeItemAt(i);
      deleteNode(doc, item);
    }

    
    if (firstEntirelySelected) {
      
      
      
      
      deleteTextRange(doc, start, null);
    } else {
      deleteTextRange(doc, start, null);
    }

    if (!firstEntirelySelected && !lastEntirelySelected) {
      mergeListItems(doc, list.id, startPos);
    }
  }

  function mergeListItems(doc, listId, itemPos) {
    
    var list = doc.get(listId);
    var targetItem = list.getItemAt(itemPos);
    var targetPath = targetItem.getPath();
    var targetLength = targetItem.getLength();
    var sourceItem = list.getItemAt(itemPos+1);
    var sourcePath = sourceItem.getPath();
    
    list.removeItemAt(itemPos+1);
    
    doc.update(targetPath, { type: 'insert', start: targetLength, text: sourceItem.getText() });
    
    annotationHelpers.transferAnnotations(doc, sourcePath, 0, targetPath, targetLength);
    deleteNode(doc, sourceItem);
  }

  function getNodes(doc, ids) {
    return ids.map(function (id) {
      return doc.get(id, 'strict')
    })
  }



  var SNIPPET_ID = "snippet";
  var TEXT_SNIPPET_ID = "text-snippet";

  var documentHelpers = /*#__PURE__*/Object.freeze({
    getPropertyAnnotationsForSelection: getPropertyAnnotationsForSelection,
    getContainerAnnotationsForSelection: getContainerAnnotationsForSelection,
    isContainerAnnotation: isContainerAnnotation,
    getTextForSelection: getTextForSelection,
    getMarkersForSelection: getMarkersForSelection,
    deleteNode: deleteNode,
    copyNode: copyNode,
    deleteTextRange: deleteTextRange,
    deleteListRange: deleteListRange,
    mergeListItems: mergeListItems,
    getNodes: getNodes,
    SNIPPET_ID: SNIPPET_ID,
    TEXT_SNIPPET_ID: TEXT_SNIPPET_ID
  });

  var SelectionState = function SelectionState(doc) {
    this.document = doc;

    this.selection = Selection.nullSelection;
    this._state = {};
    this._resetState();
  };

  SelectionState.prototype.setSelection = function setSelection (sel) {
    
    if (!sel) {
      sel = Selection.nullSelection;
    } else {
      sel.attach(this.document);
    }
    
    
    this._deriveState(sel);
    this.selection = sel;
    return true
  };

  SelectionState.prototype.getSelection = function getSelection () {
    return this.selection
  };

  SelectionState.prototype.getAnnotationsForType = function getAnnotationsForType (type) {
    var state = this._state;
    if (state.annosByType) {
      return state.annosByType.get(type) || []
    }
    return []
  };

  SelectionState.prototype.getMarkers = function getMarkers () {
    
    return this._state.markers || []
  };

  SelectionState.prototype.isInlineNodeSelection = function isInlineNodeSelection () {
    return this._state.isInlineNodeSelection
  };

  SelectionState.prototype.getContainer = function getContainer () {
    return this._state.container
  };

  SelectionState.prototype.getPreviousNode = function getPreviousNode () {
    return this._state.previousNode
  };

  SelectionState.prototype.getNextNode = function getNextNode () {
    return this._state.nextNode
  };


  SelectionState.prototype.isFirst = function isFirst$$1 () {
    return Boolean(this._state.isFirst)
  };


  SelectionState.prototype.isLast = function isLast$$1 () {
    return Boolean(this._state.isLast)
  };

  SelectionState.prototype.get = function get (key) {
    return this._state[key]
  };


  SelectionState.prototype.set = function set (key, value) {
    if (this._state[key]) {
      throw new Error(("State " + key + " is already set"))
    }
    this._state[key] = value;
  };

  SelectionState.prototype._deriveState = function _deriveState (sel) {
    this._resetState();
    this._deriveContainerSelectionState(sel);
    this._deriveAnnoState(sel);
    if (this.document.getIndex('markers')) {
      this._deriveMarkerState(sel);
    }
    
  };

  SelectionState.prototype._deriveContainerSelectionState = function _deriveContainerSelectionState (sel) {
    var state = this._state;
    var doc = this.document;
    if (sel.containerId) {
      var container = doc.get(sel.containerId);
      state.container = container;
      var startId = sel.start.getNodeId();
      var endId = sel.end.getNodeId();
      var startNode = doc.get(startId).getContainerRoot();
      var startPos = container.getPosition(startNode);
      if (startPos > 0) {
        state.previousNode = container.getNodeAt(startPos-1);
      }
      state.isFirst = isFirst(doc, sel.start);
      var endNode, endPos;
      if (endId === startId) {
        endNode = startNode;
        endPos = startPos;
      } else {
        endNode = doc.get(endId).getContainerRoot();
        endPos = container.getPosition(endNode);
      }
      if (endPos < container.getLength()-1) {
        state.nextNode = container.getNodeAt(endPos+1);
      }
      state.isLast = isLast(doc, sel.end);
    }
  };

  SelectionState.prototype._deriveAnnoState = function _deriveAnnoState (sel) {
    var doc = this.document;
    var state = this._state;

    
    var annosByType = new TreeIndex.Arrays();
    var propAnnos = getPropertyAnnotationsForSelection(doc, sel);
    propAnnos.forEach(function(anno) {
      annosByType.add(anno.type, anno);
    });

    if (propAnnos.length === 1 && propAnnos[0].isInline()) {
      state.isInlineNodeSelection = propAnnos[0].getSelection().equals(sel);
    }

    var containerId = sel.containerId;
    if (containerId) {
      var containerAnnos = getContainerAnnotationsForSelection(doc, sel, containerId);
      containerAnnos.forEach(function(anno) {
        annosByType.add(anno.type, anno);
      });
    }
    state.annosByType = annosByType;
  };

  SelectionState.prototype._deriveMarkerState = function _deriveMarkerState (sel) {
    var doc = this.document;
    var state = this._state;
    var markers = getMarkersForSelection(doc, sel);
    state.markers = markers;
  };

  SelectionState.prototype._resetState = function _resetState () {
    this._state = {
      
      annosByType: null,
      
      markers: null,
      
      isInlineNodeSelection: false,
      
      container: null,
      previousNode: null,
      nextNode: null,
      
      isFirst: false,
      
      isLast: false
    };
    return this._state
  };

  var ChangeHistory = function ChangeHistory() {
    this.reset();
  };

  ChangeHistory.prototype.reset = function reset () {
    
    this.doneChanges = [];
    
    this.undoneChanges = [];
    
    this.lastChange = null;
  };

  ChangeHistory.prototype.canUndo = function canUndo () {
    return this.doneChanges.length > 0
  };

  ChangeHistory.prototype.canRedo = function canRedo () {
    return this.undoneChanges.length > 0
  };

  ChangeHistory.prototype.push = function push (change) {
    this.doneChanges.push(change);
    this.undoneChanges = [];
  };

  var Transaction = function Transaction(master) {
    
    this.master = master;
    this.stage = master.newInstance().createFromDocument(master);
    
    this.stage._isTransactionDocument = true;

    this.tx = this.stage.createEditingInterface();
    
    this._isTransacting = false;
    this._surface = null;

    
    
    master._ops.length = 0;
  };

  var prototypeAccessors$2 = { ops: { configurable: true } };

  Transaction.prototype.dispose = function dispose () {
    this.stage.dispose();
  };



  prototypeAccessors$2.ops.get = function () {
    return this.stage._ops
  };

  prototypeAccessors$2.ops.set = function (ops) {
    this.stage._ops = ops;
  };

  Transaction.prototype.getSelection = function getSelection () {
    return this.tx.getSelection()
  };

  Transaction.prototype.setSelection = function setSelection (sel) {
    this.tx.setSelection(sel);
  };

  Transaction.prototype._reset = function _reset () {
    this._before = {};
    this._after = {};
    this.stage._ops.length = 0;
    this._info = {};
    this.setSelection(null);
  };


  Transaction.prototype._recordChange = function _recordChange (transformation, selection, info) {
    if (this._isTransacting) { throw new Error('Nested transactions are not supported.') }
    if (!isFunction$1(transformation)) { throw new Error('Document.transaction() requires a transformation function.') }
    var hasFinished = false;
    this._isTransacting = true;
    this._reset();
    var change;
    try {
      var tx = this.tx;
      tx.setSelection(selection);
      var selBefore = tx.getSelection();
      transformation(tx, {
        selection: selBefore
      });
      var ops = this.ops;
      if (ops.length > 0) {
        change = new DocumentChange(ops, tx._before, tx._after);
        change.info = info;
        change.before = { selection: selBefore };
        change.after = { selection: tx.getSelection() };
        
        
        if (this.master._isXMLDocument) {
          if (info && info.action === 'type') ; else {
            var res = this.stage._validateChange(change);
            if (!res.ok) {
              
              throw new Error('Transaction is violating the schema: \n' + res.errors.map(function (err){ return err.msg; }).join('\n'))
            }
          }
        }
      }
      hasFinished = true;
    } finally {
      if (!hasFinished) {
        this._rollback();
      }
      this._isTransacting = false;
    }
    return change
  };

  Transaction.prototype._sync = function _sync () {
    var master = this.master;
    var stage = this.stage;
    var ops = master._ops;
    for (var i = 0; i < ops.length; i++) {
      stage._applyOp(ops[i]);
    }
    ops.length = 0;
  };


  Transaction.prototype.__applyChange__ = function __applyChange__ (change) {
    var stage = this.stage;
    var ops = change.ops;
    for (var i = 0; i < ops.length; i++) {
      stage._applyOp(ops[i]);
    }
  };

  Transaction.prototype._rollback = function _rollback () {
    var stage = this.stage;
    var ops = stage._ops;
    for (var i = ops.length - 1; i >= 0; i--) {
      stage._applyOp(ops[i].invert());
    }
    ops.length = 0;
  };

  Object.defineProperties( Transaction.prototype, prototypeAccessors$2 );

  var EditorSession = (function (EventEmitter$$1) {
    function EditorSession(doc, options) {
      var this$1 = this;

      EventEmitter$$1.call(this);
      options = options || {};

      this.__id__ = uuid();
      this.document = doc;
      var configurator = options.configurator;
      if (!configurator) {
        throw new Error('No configurator provided.')
      }
      this.configurator = configurator;

      this._transaction = new Transaction(doc);
      
      
      _patchTxSetSelection(this._transaction, this);

      this._history = new ChangeHistory();
      
      this._currentChange = null;

      
      
      
      
      this._selectionState = new SelectionState(doc);

      this._commandStates = [];

      
      this._resources = ['document', 'selection', 'commandStates'];
      
      this._dirtyFlags = {};
      
      this._change = null;
      this._info = null;

      this._flowStages = ['update', 'pre-render', 'render', 'post-render', 'position', 'finalize'];
      
      this._postponed = [];
      this._observers = {};

      this._lang = options.lang || this.configurator.getDefaultLanguage();
      this._dir = options.dir || 'ltr';

      
      
      var CommandManager = configurator.getCommandManagerClass();
      var DragManager = configurator.getDragManagerClass();
      var FileManager = configurator.getFileManagerClass();
      var GlobalEventHandler = configurator.getGlobalEventHandlerClass();
      var KeyboardManager = configurator.getKeyboardManagerClass();
      var MacroManager = configurator.getMacroManagerClass();
      var MarkersManager = configurator.getMarkersManagerClass();
      var SurfaceManager = configurator.getSurfaceManagerClass();


      
      
      this.surfaceManager = new SurfaceManager(this);
      
      this._context = {
        editorSession: this,
        
        surfaceManager: this.surfaceManager,
      };
      
      if (options.context) {
        Object.assign(this._context, options.context);
      }

      var commands = configurator.getCommands();
      var dropHandlers = configurator.getDropHandlers();
      var macros = configurator.getMacros();
      var converterRegistry = configurator.getConverterRegistry();
      var editingBehavior = configurator.getEditingBehavior();

      this.fileManager = options.fileManager || new FileManager(this, configurator.getFileAdapters(), this._context);

      
      this._hasUnsavedChanges = false;
      this._isSaving = false;

      if (options.saveHandler) {
        this.saveHandler = options.saveHandler;
      } else {
        this.saveHandler = configurator.getSaveHandler();
      }

      
      this._managers = {};
      forEach(configurator.getManagers(), function (ManagerClass, name) {
        this$1._managers[name] = new ManagerClass(this$1._context);
      });

      
      
      this.commandManager = options.commandManager || new CommandManager(this._context, commands);

      
      
      this.dragManager = new DragManager(dropHandlers, Object.assign({}, this._context, {
        commandManager: this.commandManager
      }));
      
      this.macroManager = new MacroManager(this._context, macros);
      this.globalEventHandler = new GlobalEventHandler(this, this.surfaceManager);
      this.markersManager = new MarkersManager(this);
      this.keyboardManager = new KeyboardManager(this, configurator.getKeyboardShortcuts(), {
        context: this._context
      });

      
      this.converterRegistry = converterRegistry;
      this.editingBehavior = editingBehavior;
    }

    if ( EventEmitter$$1 ) EditorSession.__proto__ = EventEmitter$$1;
    EditorSession.prototype = Object.create( EventEmitter$$1 && EventEmitter$$1.prototype );
    EditorSession.prototype.constructor = EditorSession;

    EditorSession.prototype.dispose = function dispose () {
      this._transaction.dispose();
      this.surfaceManager.dispose();
      this.fileManager.dispose();
      this.commandManager.dispose();
      this.dragManager.dispose();
      this.macroManager.dispose();
      this.globalEventHandler.dispose();
      this.markersManager.dispose();

      forEach(this._managers, function (manager) {
        if (manager.dispose) {
          manager.dispose();
        }
      });
    };


    EditorSession.prototype.hasChanged = function hasChanged (resource) {
      return this._dirtyFlags[resource]
    };

    EditorSession.prototype.hasDocumentChanged = function hasDocumentChanged () {
      return this.hasChanged('document')
    };

    EditorSession.prototype.hasSelectionChanged = function hasSelectionChanged () {
      return this.hasChanged('selection')
    };

    EditorSession.prototype.hasCommandStatesChanged = function hasCommandStatesChanged () {
      return this.hasChanged('commandStates')
    };

    EditorSession.prototype.hasLanguageChanged = function hasLanguageChanged () {
      return this.hasChanged('lang')
    };

    EditorSession.prototype.hasTextDirectionChanged = function hasTextDirectionChanged () {
      return this.hasChanged('dir')
    };

    EditorSession.prototype.get = function get (resourceName) {
      switch(resourceName) {
        case 'document':
          return this.getDocument()
        case 'selection':
          return this.getSelection()
        case 'commandStates':
          return this.getCommandStates()
        case 'change':
          return this.getChange()
        case 'lang':
          return this.getLanguage()
        case 'dir':
          return this.getTextDirection()
        default:
          throw new Error('Unknown resource: ' + resourceName)
      }
    };

    EditorSession.prototype.getConfigurator = function getConfigurator () {
      return this.configurator
    };

    EditorSession.prototype.getContext = function getContext () {
      return this._context
    };

    EditorSession.prototype.getDocument = function getDocument () {
      return this.document
    };

    EditorSession.prototype.getManager = function getManager (name) {
      return this._managers[name]
    };

    EditorSession.prototype.getSelection = function getSelection () {
      return this.getSelectionState().getSelection()
    };

    EditorSession.prototype.getSelectionState = function getSelectionState () {
      return this._selectionState
    };

    EditorSession.prototype.getCommandStates = function getCommandStates () {
      return this._commandStates
    };

    EditorSession.prototype.getChange = function getChange () {
      return this._change
    };

    EditorSession.prototype.getChangeInfo = function getChangeInfo () {
      return this._info
    };

    EditorSession.prototype.getFocusedSurface = function getFocusedSurface () {
      return this.surfaceManager.getFocusedSurface()
    };

    EditorSession.prototype.getSurface = function getSurface (surfaceId) {
      return this.surfaceManager.getSurface(surfaceId)
    };

    EditorSession.prototype.getLanguage = function getLanguage () {
      return this._lang
    };

    EditorSession.prototype.getTextDirection = function getTextDirection () {
      return this._dir
    };

    EditorSession.prototype.canUndo = function canUndo () {
      return this._history.canUndo()
    };

    EditorSession.prototype.canRedo = function canRedo () {
      return this._history.canRedo()
    };

    
    EditorSession.prototype.resetHistory = function resetHistory () {
      this._history.reset();
      this._setDirty('commandStates');
      if (!this._flowing) {
        this.startFlow();
      }
    };

    EditorSession.prototype.executeCommand = function executeCommand () {
      var ref;

      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];
      (ref = this.commandManager).executeCommand.apply(ref, args);
    };

    
    EditorSession.prototype.attachEditor = function attachEditor (editor) {
      this.editor = editor;
    };

    EditorSession.prototype.detachEditor = function detachEditor () {
      this.editor = undefined;
    };

    EditorSession.prototype.getEditor = function getEditor () {
      return this.editor
    };

    EditorSession.prototype.setSelection = function setSelection (sel, skipFlow) {
      
      if (sel && isPlainObject(sel)) {
        sel = this.getDocument().createSelection(sel);
      }
      if (sel && !sel.isNull()) {
        if (!sel.surfaceId) {
          var fs = this.getFocusedSurface();
          if (fs) {
            sel.surfaceId = fs.id;
          }
        }
      }

      _addSurfaceId(sel, this);
      _addContainerId(sel, this);

      if (this._setSelection(sel) && !skipFlow) {
        this.startFlow();
      }
      return sel
    };

    EditorSession.prototype.selectNode = function selectNode (nodeId) {
      var surface = this.getFocusedSurface();
      this.setSelection({
        type: 'node',
        nodeId: nodeId,
        containerId: surface.getContainerId(),
        surfaceId: surface.id
      });
    };

    EditorSession.prototype.setCommandStates = function setCommandStates (commandStates) {
      this._commandStates = commandStates;
      this._setDirty('commandStates');
    };

    EditorSession.prototype.setLanguage = function setLanguage (lang) {
      if (this._lang !== lang) {
        this._lang = lang;
        this._setDirty('lang');
        this.startFlow();
      }
    };

    EditorSession.prototype.setTextDirection = function setTextDirection (dir) {
      if (this._dir !== dir) {
        this._dir = dir;
        this._setDirty('dir');
        this.startFlow();
      }
    };

    EditorSession.prototype.createSelection = function createSelection () {
      var doc = this.getDocument();
      return doc.createSelection.apply(doc, arguments)
    };

    EditorSession.prototype.getCollaborators = function getCollaborators () {
      return null
    };

    
    EditorSession.prototype.setSaveHandler = function setSaveHandler (saveHandler) {
      this.saveHandler = saveHandler;
    };

    
    EditorSession.prototype.transaction = function transaction (transformation, info) {
      var t = this._transaction;
      info = info || {};
      t._sync();
      var change = t._recordChange(transformation, this.getSelection(), info);
      if (change) {
        this._commit(change, info);
      } else {
        
        this._setSelection(this._transaction.getSelection());
        this.startFlow();
      }
      return change
    };

    EditorSession.prototype.undo = function undo () {
      this._undoRedo('undo');
    };

    EditorSession.prototype.redo = function redo () {
      this._undoRedo('redo');
    };

    

    EditorSession.prototype.on = function on () {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      var name = args[0];
      if (this._flowStages.indexOf(name) >= 0) {
        
        args.shift();
        var options = args[2] || {};
        var resource = options.resource;
        if (resource) {
          delete options.resource;
          args.unshift(resource);
        }
        this._registerObserver(name, args);
      } else {
        EventEmitter$$1.prototype.on.apply(this, args);
      }
    };

    EditorSession.prototype.off = function off () {
      var this$1 = this;
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      if (args.length === 1) {
        var observer = args[0];
        EventEmitter$$1.prototype.off.apply(this, args);
        
        
        if (observer[this.__id__]) {
          var records = observer[this.__id__];
          delete observer[this.__id__];
          records.forEach(function (record) {
            this$1.__deregisterObserver(record);
          });
        }
      } else {
        var stage = args[0];
        var method = args[1];
        var observer$1 = args[2];
        this._deregisterObserver(stage, method, observer$1);
      }
    };

    
    EditorSession.prototype.onUpdate = function onUpdate () {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      return this._registerObserver('update', args)
    };

    EditorSession.prototype.onPreRender = function onPreRender () {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      return this._registerObserver('pre-render', args)
    };

    
    EditorSession.prototype.onRender = function onRender () {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      return this._registerObserver('render', args)
    };

    
    EditorSession.prototype.onPostRender = function onPostRender () {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      return this._registerObserver('post-render', args)
    };

    
    EditorSession.prototype.onPosition = function onPosition () {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      return this._registerObserver('position', args)
    };

    EditorSession.prototype.onFinalize = function onFinalize () {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      return this._registerObserver('finalize', args)
    };

    EditorSession.prototype._setSelection = function _setSelection (sel) {
      
      
      
      
      

      var hasChanged = this.getSelectionState().setSelection(sel);
      if (hasChanged) { this._setDirty('selection'); }
      return hasChanged
    };

    EditorSession.prototype._undoRedo = function _undoRedo (which) {
      var doc = this.getDocument();
      var from, to;
      if (which === 'redo') {
        from = this._history.undoneChanges;
        to = this._history.doneChanges;
      } else {
        from = this._history.doneChanges;
        to = this._history.undoneChanges;
      }
      var change = from.pop();
      if (change) {
        this._applyChange(change, {});
        this._transaction.__applyChange__(change);
        
        to.push(doc.invert(change));
        
        var sel = change.after.selection;
        if (sel) { sel.attach(doc); }
        this._setSelection(sel);
        
        this.startFlow();
      } else {
        console.warn('No change can be %s.', (which === 'undo'? 'undone':'redone'));
      }
    };

    EditorSession.prototype._transformLocalChangeHistory = function _transformLocalChangeHistory (externalChange) {
      
      
      
      
      var clone = {
        ops: externalChange.ops.map(function(op) { return op.clone(); })
      };
      transformDocumentChange(clone, this._history.doneChanges);
      transformDocumentChange(clone, this._history.undoneChanges);
    };

    EditorSession.prototype._transformSelection = function _transformSelection (change) {
      var oldSelection = this.getSelection();
      var newSelection = transformSelection(oldSelection, change);
      
      return newSelection
    };

    EditorSession.prototype._commit = function _commit (change, info) {
      this._commitChange(change, info);
      
      this._hasUnsavedChanges = true;
      this.startFlow();
    };

    EditorSession.prototype._commitChange = function _commitChange (change, info) {
      change.timestamp = Date.now();
      this._applyChange(change, info);
      if (info['history'] !== false && !info['hidden']) {
        var inverted = this.getDocument().invert(change);
        this._history.push(inverted);
      }
      var newSelection = change.after.selection || Selection.nullSelection;
      
      
      if (!newSelection.isNull() && !newSelection.surfaceId) {
        newSelection.surfaceId = change.after.surfaceId;
      }
      this._setSelection(newSelection);
      this.emit('commit', change);
    };

    EditorSession.prototype._applyChange = function _applyChange (change, info) {
      if (!change) {
        console.error('FIXME: change is null.');
        return
      }
      var doc = this.getDocument();
      doc._apply(change);
      doc._notifyChangeListeners(change, info);
      this._setDirty('document');
      this._change = change;
      this._info = info;
    };

    EditorSession.prototype._applyRemoteChange = function _applyRemoteChange (change) {
      
      if (change.ops.length > 0) {
        this._applyChange(change, { remote: true });
        
        this._transaction.__applyChange__(change);
        this._transformLocalChangeHistory(change);
        this._setSelection(this._transformSelection(change));
        this.startFlow();
      }
    };

    
    EditorSession.prototype.hasUnsavedChanges = function hasUnsavedChanges () {
      return this._hasUnsavedChanges
    };

    
    EditorSession.prototype.save = function save () {
      var this$1 = this;

      var saveHandler = this.saveHandler;

      if (this._hasUnsavedChanges && !this._isSaving) {
        this._isSaving = true;
        
        if (saveHandler) {
          var saveParams = {
            editorSession: this,
            fileManager: this.fileManager
          };
          return saveHandler.saveDocument(saveParams)
          .then(function () {
            this$1._hasUnsavedChanges = false;
            
            
            
            
            this$1.setSelection(this$1.getSelection());
          })
          .catch(function (err) {
            console.error('Error during save', err);
          }).then(function () { 
            this$1._isSaving = false;
          })
        } else {
          console.error('Document saving is not handled at the moment. Make sure saveHandler instance provided to editorSession');
          return Promise.reject()
        }
      }
    };

    
    EditorSession.prototype.startFlow = function startFlow () {
      if (this._flowing) {
        throw new Error('Already in a flow. You need to postpone the update.')
      }
      this._flowing = true;
      try {
        this.performFlow();
      } finally {
        this._resetFlow();
        this._flowing = false;
      }
      
      
      
      var postponed = this._postponed;
      var self = this;
      this._postponed = [];
      setTimeout(function() {
        postponed.forEach(function(fn) {
          fn(self);
        });
      }, 0);
    };

    
    EditorSession.prototype.performFlow = function performFlow () {
      var this$1 = this;

      this._flowStages.forEach(function (stage) {
        this$1._notifyObservers(stage);
      });
    };

    EditorSession.prototype.postpone = function postpone (fn) {
      this._postponed.push(fn);
    };

    EditorSession.prototype._parseObserverArgs = function _parseObserverArgs (args) {
      var params = { stage: null, resource: null, handler: null, context: null, options: {} };
      
      var idx = 0;
      var arg = args[idx];
      if (isString(arg)) {
        params.resource = arg;
        idx++;
        arg = args[idx];
      }
      if (!arg) {
        throw new Error('Provided handler function was nil.')
      }
      if (!isFunction$1(arg)) {
        throw new Error('Expecting a handler Function.')
      }
      params.handler = arg;
      idx++;
      arg = args[idx];
      if (arg) {
        params.context = arg;
        idx++;
        arg = args[idx];
      }
      if (arg) {
        params.options = arg;
      }
      return params
    };

    

    EditorSession.prototype._registerObserver = function _registerObserver (stage, args) {
      
      
      var record = this._parseObserverArgs(args);
      record.stage = stage;
      this.__registerObserver(stage, record);
    };

    EditorSession.prototype.__registerObserver = function __registerObserver (stage, record) {
      
      
      if (record.context) {
        var observer = record.context;
        if (!observer[this.__id__]) {
          observer[this.__id__] = [];
        }
        observer[this.__id__].push(record);
      }
      var observers = this._observers[stage];
      if (!observers) {
        observers = this._observers[stage] = [];
      }
      observers.push(record);
    };

    
    EditorSession.prototype._deregisterObserver = function _deregisterObserver (stage, method, observer) {
      var self = this; 
      if (arguments.length === 1) {
        
        
        forEach(self._observers, function (observers) {
          for (var i = observers.length-1; i >=0 ; i--) {
            var o = observers[i];
            if (o.context === observer) {
              observers.splice(i, 1);
              o._deregistered = true;
            }
          }
        });
      } else {
        var observers = self._observers[stage];
        
        
        if (!observers) {
          EventEmitter$$1.prototype.off.apply(self, arguments);
        } else {
          for (var i = observers.length-1; i >= 0; i--) {
            var o = observers[i];
            if (o.handler === method && o.context === observer) {
              observers.splice(i, 1);
              o._deregistered = true;
            }
          }
        }
      }
    };

    EditorSession.prototype.__deregisterObserver = function __deregisterObserver (record) {
      var stage = record.stage;
      var observers = this._observers[stage];
      var observer = record.context;
      var method = record.handler;
      for (var i = observers.length-1; i >= 0; i--) {
        var o = observers[i];
        if (o.handler === method && o.context === observer) {
          observers.splice(i, 1);
          o._deregistered = true;
        }
      }
    };

    EditorSession.prototype._notifyObservers = function _notifyObservers (stage) {
      var this$1 = this;

      
      
      
      
      
      
      var _observers = this._observers[stage];
      if (!_observers) { return }
      
      
      
      var observers = _observers.slice();
      for (var i = 0; i < observers.length; i++) {
        
        var o = observers[i];
        if (o._deregistered) { continue }
        if (!o.resource) {
          o.handler.call(o.context, this$1);
        } else if (o.resource === 'document') {
          if (!this$1.hasDocumentChanged()) { continue }
          var change = this$1.getChange();
          var info = this$1.getChangeInfo();
          var path = o.options.path;
          if (!path) {
            o.handler.call(o.context, change, info, this$1);
          } else if (change.hasUpdated(path)) {
            o.handler.call(o.context, change, info, this$1);
          }
        } else {
          if (!this$1.hasChanged(o.resource)) { continue }
          var resource = this$1.get(o.resource);
          o.handler.call(o.context, resource, this$1);
        }
      }
    };

    EditorSession.prototype._setDirty = function _setDirty (resource) {
      this._dirtyFlags[resource] = true;
    };

    EditorSession.prototype._resetFlow = function _resetFlow () {
      var this$1 = this;

      Object.keys(this._dirtyFlags).forEach(function (resource) {
        this$1._dirtyFlags[resource] = false;
      });
      this._change = null;
      this._info = null;
    };

    

    EditorSession.prototype.setBlurred = function setBlurred (blurred) {
      this._blurred = blurred;
      
      this.commandManager._updateCommandStates(this);
      this._setDirty('commandStates');
    };

    EditorSession.prototype.isBlurred = function isBlurred () {
      return Boolean(this._blurred)
    };

    return EditorSession;
  }(EventEmitter));

  function _patchTxSetSelection(tx, editorSession) {
    tx.setSelection = function(sel) {
      sel = Transaction.prototype.setSelection.call(tx, sel);
      _addSurfaceId(sel, editorSession);
      _addContainerId(sel, editorSession);
      return sel
    };
  }


  function _addSurfaceId(sel, editorSession) {
    if (sel && !sel.isNull() && !sel.surfaceId) {
      
      var surface = editorSession.getFocusedSurface();
      if (surface) {
        sel.surfaceId = surface.id;
      } else {
        
        console.warn('No focused surface. Selection will not be rendered.');
      }
    }
  }

  function _addContainerId(sel, editorSession) {
    if (sel && !sel.isNull() && sel.surfaceId && !sel.containerId) {
      var surface = editorSession.getSurface(sel.surfaceId);
      if (surface) {
        var containerId = surface.getContainerId();
        if (containerId) {
          sel.containerId = containerId;
        }
      }
    }
  }

  var CollabSession = (function (EditorSession$$1) {
    function CollabSession(doc, config) {
      EditorSession$$1.call(this, doc, config);
      config = config || {};
      this.config = config;
      this.collabClient = config.collabClient;
      if (config.docVersion) {
        console.warn('config.docVersion is deprecated: Use config.version instead');
      }
      if (config.docVersion) {
        console.warn('config.docId is deprecated: Use config.documentId instead');
      }
      this.version = config.version;
      this.documentId = config.documentId || config.docId;
      if (config.autoSync !== undefined) {
        this.autoSync = config.autoSync;
      } else {
        this.autoSync = true;
      }
      if (!this.documentId) {
        throw new SubstanceError('InvalidArgumentsError', {message: 'documentId is mandatory'})
      }
      if (typeof this.version === undefined) {
        throw new SubstanceError('InvalidArgumentsError', {message: 'version is mandatory'})
      }
      
      this._connected = false; 
      this._nextChange = null; 
      this._pendingChange = null; 
      this._pendingSync = false;
      this._error = null;
      
      this.onUpdate('document', this.afterDocumentChange, this);
      
      this.collabClient.on('connected', this.onCollabClientConnected, this);
      this.collabClient.on('disconnected', this.onCollabClientDisconnected, this);
      this.collabClient.on('message', this._onMessage.bind(this));
      
      
      
      if (this.collabClient.isConnected() && this.autoSync) {
        this.sync();
      }
    }

    if ( EditorSession$$1 ) CollabSession.__proto__ = EditorSession$$1;
    CollabSession.prototype = Object.create( EditorSession$$1 && EditorSession$$1.prototype );
    CollabSession.prototype.constructor = CollabSession;

    
    CollabSession.prototype.dispose = function dispose () {
      this.disconnect();
      this.collabClient.off(this);
    };

    
    CollabSession.prototype.disconnect = function disconnect () {
      
      var msg = {
        type: 'disconnect',
        documentId: this.documentId
      };
      
      this._abortSync();
      this._send(msg);
    };

    
    CollabSession.prototype.sync = function sync () {
      
      if (this.__canSync()) {
        var nextChange = this._nextChange;
        var msg = {
          type: 'sync',
          documentId: this.documentId,
          version: this.version,
          change: nextChange ? this.serializeChange(nextChange) : undefined
        };
        this._send(msg);
        this._pendingSync = true;
        this._pendingChange = nextChange;

        
        
        this.emit('sync');
        this._nextChange = null;
        this._error = null;
      } else {
        console.error('Can not sync. Either collabClient is not connected or already syncing');
      }
    };

    CollabSession.prototype.getCollaborators = function getCollaborators () {
      return this.collaborators
    };

    CollabSession.prototype.isConnected = function isConnected () {
      return this._connected
    };

    CollabSession.prototype.serializeChange = function serializeChange (change) {
      return change.toJSON()
    };

    CollabSession.prototype.deserializeChange = function deserializeChange (serializedChange) {
      return DocumentChange.fromJSON(serializedChange)
    };

    

    
    CollabSession.prototype._onMessage = function _onMessage (msg) {
      
      if (msg.documentId !== this.documentId) {
        return false
      }
      
      msg = cloneDeep(msg);
      switch (msg.type) {
        case 'syncDone':
          this.syncDone(msg);
          break
        case 'syncError':
          this.syncError(msg);
          break
        case 'update':
          this.update(msg);
          break
        case 'disconnectDone':
          this.disconnectDone(msg);
          break
        case 'error':
          this.error(msg);
          break
        default:
          console.error('CollabSession: unsupported message', msg.type, msg);
          return false
      }
      return true
    };

    
    CollabSession.prototype._send = function _send (msg) {
      if (this.collabClient.isConnected()) {
        this.collabClient.send(msg);
        return true
      } else {
        console.warn('Try not to call _send when disconnected. Skipping message', msg);
        return false
      }
    };

    
    CollabSession.prototype.update = function update (args) {
      
      var serverChange = args.change;
      var serverVersion = args.version;

      if (!this._nextChange && !this._pendingSync) {
        if (serverChange) {
          serverChange = this.deserializeChange(serverChange);
          this._applyRemoteChange(serverChange);
        }
        if (serverVersion) {
          this.version = serverVersion;
        }
        this.startFlow();
      } else {
        console.info('skipped remote update. Pending sync or local changes.');
      }
    };

    
    CollabSession.prototype.syncDone = function syncDone (args) {
      
      var serverChange = args.serverChange;
      var serverVersion = args.version;

      if (serverChange) {
        serverChange = this.deserializeChange(serverChange);
        this._applyRemoteChange(serverChange);
      }
      this.version = serverVersion;
      
      
      this._pendingChange = null;
      this._pendingSync = false;
      this._error = null;
      
      this._connected = true;
      this.startFlow();
      this.emit('connected');
      
      this._requestSync();
    };

    
    CollabSession.prototype.syncError = function syncError (error) {
      console.info('SyncError occured. Aborting sync', error);
      this._abortSync();
    };

    CollabSession.prototype.disconnectDone = function disconnectDone () {
      
      
      this._afterDisconnected();
    };

    
    CollabSession.prototype.error = function error (message) {
      var error = message.error;
      var errorFn = this[error.name];
      var err = SubstanceError.fromJSON(error);

      if (!errorFn) {
        error('CollabSession: unsupported error', error.name);
        return false
      }

      this.emit('error', err);
      errorFn = errorFn.bind(this);
      errorFn(err);
    };


    

    CollabSession.prototype.afterDocumentChange = function afterDocumentChange (change, info) {
      
      if (!info.remote) {
        this._recordChange(change);
      }
    };

    
    CollabSession.prototype.onCollabClientConnected = function onCollabClientConnected () {
      
      if (this.autoSync) {
        this.sync();
      }
    };

    
    CollabSession.prototype.onCollabClientDisconnected = function onCollabClientDisconnected () {
      
      this._abortSync();
      if (this._connected) {
        this._afterDisconnected();
      }
    };

    

    CollabSession.prototype._commit = function _commit (change, info) {
      this._commitChange(change, info);
      this.startFlow();
    };

    
    CollabSession.prototype._applyRemoteChange = function _applyRemoteChange (change) {
      
      if (change.ops.length > 0) {
        this._transaction.__applyChange__(change);
        this.getDocument()._apply(change);
        this._setDirty('document');
        
        this._transformLocalChangeHistory(change);
        this._setSelection(this._transformSelection(change));
        this._change = change;
        this._info = { remote: true };
        this.startFlow();
      }
    };

    
    CollabSession.prototype._recordChange = function _recordChange (change) {
      if (!this._nextChange) {
        this._nextChange = change;
      } else {
        
        this._nextChange.ops = this._nextChange.ops.concat(change.ops);
        this._nextChange.after = change.after;
      }
      this._requestSync();
    };

    CollabSession.prototype.__canSync = function __canSync () {
      return this.collabClient.isConnected() && !this._pendingSync
    };

    
    CollabSession.prototype._requestSync = function _requestSync () {
      if (this._nextChange && this.__canSync()) {
        this.sync();
      }
    };

    
    CollabSession.prototype._abortSync = function _abortSync () {
      var newNextChange = this._nextChange;

      if (this._pendingChange) {
        newNextChange = this._pendingChange;
        
        if (this._nextChange) {
          newNextChange.ops = newNextChange.ops.concat(this._nextChange.ops);
          newNextChange.after = this._nextChange.after;
        }
        this._pendingChange = null;
      }
      this._pendingSync = false;
      this._error = null;
      this._nextChange = newNextChange;
    };

    
    CollabSession.prototype._afterDisconnected = function _afterDisconnected () {
      this._connected = false;
      this.emit('disconnected');
    };

    
    CollabSession.prototype._hasLocalChanges = function _hasLocalChanges () {
      return this._nextChange && this._nextChange.ops.length > 0
    };

    return CollabSession;
  }(EditorSession));

  var DocumentArchive = function DocumentArchive(sessions, buffer) {
    this.sessions = sessions;
    this.buffer = buffer;

    if (!sessions.manifest) { throw new Error("'manifest' session is required.") }

    this.init();
  };

  DocumentArchive.prototype.init = function init () {
    

  };

  DocumentArchive.prototype.getManifest = function getManifest () {
    return this.sessions.manifest.getDocument()
  };

  DocumentArchive.prototype.getDocumentEntries = function getDocumentEntries () {
    return this.getEditorSession('manifest').getDocument().getDocumentEntries()
  };

  DocumentArchive.prototype.getEditorSession = function getEditorSession (docId) {
    return this.sessions[docId]
  };

  var PLAINOBJ = {};


  var Registry = function Registry(entries, validator) {
    this.entries = {};
    this.names = [];
    this.validator = validator;

    if (entries) {
      forEach(entries, function(entry, name) {
        this.add(name, entry);
      }.bind(this));
    }
  };


  Registry.prototype.contains = function contains (name) {
    return this.entries.hasOwnProperty(name)
  };


  Registry.prototype.add = function add (name, entry) {
    if (this.validator) {
      this.validator(entry);
    }
    if (PLAINOBJ[name]) {
      throw new Error('Illegal key: "'+name+'" is a property of Object which is thus not allowed as a key.')
    }
    if (this.contains(name)) {
      this.remove(name);
    }
    this.entries[name] = entry;
    this.names.push(name);
  };


  Registry.prototype.remove = function remove (name) {
    var pos = this.names.indexOf(name);
    if (pos >= 0) {
      this.names.splice(pos, 1);
    }
    delete this.entries[name];
  };


  Registry.prototype.clear = function clear () {
    this.names = [];
    this.entries = {};
  };


  Registry.prototype.get = function get (name, strict) {
    var result = this.entries[name];
    if (strict && !result) {
      throw new Error('No entry registered for name '+name)
    }
    return result
  };


  Registry.prototype.forEach = function forEach$$1 (callback) {
      var this$1 = this;

    for (var i = 0; i < this.names.length; i++) {
      var name = this$1.names[i];
      var _continue = callback(this$1.entries[name], name);
      if (_continue === false) {
        break
      }
    }
  };

  Registry.prototype.map = function map (callback) {
    var result = [];
    this.forEach(function (entry, name) {
      result.push(callback(entry, name));
    });
    return result
  };

  Registry.prototype.filter = function filter (callback) {
    var result = [];
    this.forEach(function(entry, name) {
      if (callback(entry, name)) {
        result.push(entry);
      }
    });
    return result
  };

  Registry.prototype.values = function values () {
    return this.filter(function () { return true })
  };

  Registry.prototype._isRegistry = true;

  var NodeRegistry = (function (Registry$$1) {
    function NodeRegistry () {
      Registry$$1.apply(this, arguments);
    }

    if ( Registry$$1 ) NodeRegistry.__proto__ = Registry$$1;
    NodeRegistry.prototype = Object.create( Registry$$1 && Registry$$1.prototype );
    NodeRegistry.prototype.constructor = NodeRegistry;

    NodeRegistry.prototype.register = function register (nodeClazz) {
      var type = nodeClazz.prototype.type;
      if ( typeof type !== 'string' || type === '' ) {
        throw new Error( 'Node names must be strings and must not be empty')
      }
      if (!( nodeClazz.prototype._isNode)) {
        throw new Error( 'Nodes must be subclasses of Substance.Data.Node' )
      }
      if (this.contains(type)) {
        
        console.info('Overriding node type', type);
        this.remove(type);
      }
      this.add(type, nodeClazz);
    };

    return NodeRegistry;
  }(Registry));

  function isBoolean(val) {
    return (val === true || val === false || (val && val.constructor === Boolean) )
  }

  var Property = function Property(definition) {
    this.definition = definition;
  };

  var prototypeAccessors$3 = { type: { configurable: true },name: { configurable: true } };

  Property.prototype.isArray = function isArray$1 () {
    return isArray(this.definition.type)
  };

  Property.prototype.isReference = function isReference () {
    if (this.isArray()) {
      return last$1(this.definition.type) === 'id'
    } else {
      return this.definition.type === 'id'
    }
  };

  Property.prototype.isText = function isText () {
    return Boolean(this.definition._isText)
  };

  Property.prototype.isOwned = function isOwned () {
    return Boolean(this.definition.owned)
  };

  Property.prototype.isOptional = function isOptional () {
    return Boolean(this.definition.optional)
  };

  Property.prototype.isNotNull = function isNotNull () {
    return Boolean(this.definition.notNull)
  };

  Property.prototype.hasDefault = function hasDefault () {
    return this.definition.hasOwnProperty('default')
  };

  Property.prototype.getDefault = function getDefault () {
    return this.definition.default
  };

  Property.prototype.createDefaultValue = function createDefaultValue () {
    if (isArray(this.definition.type)) {
      return []
    }
    switch(this.definition.type) {
      case 'object':
        return {}
      case 'number':
        return -1
      case 'coordinate':
        return new Coordinate([], 0)
      case 'boolean':
        return false
      case 'id':
        return null
      case 'string':
        return ''
      default:
        return null
    }
  };

  prototypeAccessors$3.type.get = function () {
    return this.definition.type
  };

  prototypeAccessors$3.name.get = function () {
    return this.definition.name
  };

  Object.defineProperties( Property.prototype, prototypeAccessors$3 );

  var Node = (function (EventEmitter$$1) {
    function Node() {
      EventEmitter$$1.call(this);

      
      
      this._initialize.apply(this, arguments);
    }

    if ( EventEmitter$$1 ) Node.__proto__ = EventEmitter$$1;
    Node.prototype = Object.create( EventEmitter$$1 && EventEmitter$$1.prototype );
    Node.prototype.constructor = Node;

    var prototypeAccessors = { type: { configurable: true } };

    Node.prototype._initialize = function _initialize (data) {
      var this$1 = this;

      var NodeClass = this.constructor;

      var schema = NodeClass.schema;
      for (var name in schema) {
        if (!schema.hasOwnProperty(name)) { continue }
        var prop = schema[name];
        
        
        var propIsGiven = (data[name] !== undefined);
        var hasDefault = prop.hasDefault();
        var isOptional = prop.isOptional();
        if ( (!isOptional && !hasDefault) && !propIsGiven) {
          throw new Error('Property ' + name + ' is mandatory for node type ' + this$1.type)
        }
        if (propIsGiven) {
          this$1[name] = _checked(prop, data[name]);
        } else if (hasDefault) {
          this$1[name] = cloneDeep(_checked(prop, prop.getDefault()));
        }
      }
    };

    Node.prototype.dispose = function dispose () {
      this._disposed = true;
    };

    Node.prototype.isDisposed = function isDisposed () {
      return Boolean(this._disposed)
    };

    
    Node.prototype.isInstanceOf = function isInstanceOf (typeName) {
      return Node.isInstanceOf(this.constructor, typeName)
    };

    Node.prototype.getSchema = function getSchema () {
      return this.constructor.schema
    };

    
    Node.prototype.getTypeNames = function getTypeNames () {
      var typeNames = [];
      var NodeClass = this.constructor;
      while (NodeClass.type !== "node") {
        typeNames.push(NodeClass.type);
        NodeClass = Object.getPrototypeOf(NodeClass);
      }
      return typeNames
    };

    
    Node.prototype.getPropertyType = function getPropertyType (propertyName) {
      var schema = this.constructor.schema;
      return schema[propertyName].type
    };

    
    Node.prototype.toJSON = function toJSON () {
      var this$1 = this;

      var data = {
        type: this.type
      };
      var schema = this.getSchema();
      forEach(schema, function (prop, name) {
        var val = this$1[name];
        if (prop.isOptional() && val === undefined) { return }
        if (isArray(val) || isObject(val)) {
          val = cloneDeep(val);
        }
        data[prop.name] = val;
      });
      return data
    };

    prototypeAccessors.type.get = function () {
      return this.constructor.type
    };

    Object.defineProperties( Node.prototype, prototypeAccessors );

    return Node;
  }(EventEmitter));

  Node.prototype._isNode = true;


  Object.defineProperty(Node, 'schema', {
    get: function get() { return this._schema },
    set: function set(schema) {
      var NodeClass = this;
      
      
      if (schema.type) {
        NodeClass.type = schema.type;
      }
      
      
      
      NodeClass._schema = compileSchema(NodeClass, schema);
    }
  });

  Node.define = Node.defineSchema = function define(schema) {
    this.schema = schema;
  };

  Node.schema = {
    type: "node",
    id: 'string'
  };


  Node.isInstanceOf = function(NodeClass, typeName) {
    var type = NodeClass.type;
    while (type !== "node") {
      if (type === typeName) { return true }
      var _super = Object.getPrototypeOf(NodeClass.prototype).constructor;
      if (_super && _super.type) {
        NodeClass = _super;
        type = NodeClass.type;
      } else {
        break
      }
    }
    return false
  };



  function compileSchema(NodeClass, schema) {
    var compiledSchema = _compileSchema(schema);
    var schemas = [compiledSchema];
    var clazz = NodeClass;
    while(clazz) {
      var parentProto = Object.getPrototypeOf(clazz.prototype);
      if (!parentProto) {
        break
      }
      clazz = parentProto.constructor;
      if (clazz && clazz._schema) {
        schemas.unshift(clazz._schema);
      }
    }
    schemas.unshift({});
    return Object.assign.apply(null, schemas)
  }

  function _compileSchema(schema) {
    var compiledSchema = {};
    forEach(schema, function(definition, name) {
      
      if (name === 'type') {
        return
      }
      if (isString(definition) || isArray(definition)) {
        definition = { type: definition };
      }
      definition = _compileDefintion(definition);
      definition.name = name;
      compiledSchema[name] = new Property(definition);
    });
    return compiledSchema
  }

  function _compileDefintion(definition) {
    var result = definition;
    if (isArray(definition.type) && definition.type[0] !== "array") {
      definition.targetTypes = definition.type;
      definition.type = [ "array", "id" ];
    } else if (definition.type === 'text') {
      result = {
        type: "string",
        default: '',
        _isText: true
      };
    }
    return result
  }

  function _checked(prop, value) {
    var type;
    var name = prop.name;
    if (prop.isArray()) {
      type = "array";
    } else {
      type = prop.type;
    }
    if (value === null) {
      if (prop.isNotNull()) {
        throw new Error('Value for property ' + name + ' is null.')
      } else {
        return value
      }
    }
    if (value === undefined) {
      throw new Error('Value for property ' + name + ' is undefined.')
    }
    if (type === "string" && !isString(value) ||
        type === "boolean" && !isBoolean(value) ||
        type === "number" && !isNumber(value) ||
        type === "array" && !isArray(value) ||
        type === "id" && !isString(value) ||
        type === "object" && !isObject(value)) {
      throw new Error('Illegal value type for property ' + name + ': expected ' + type + ', was ' + (typeof value))
    }
    return value
  }

  var Schema = function Schema(name, version) {
    if (!name) {
      throw new Error("'name' is mandatory")
    }
    if (!version) {
      throw new Error("'version' is mandatory")
    }

    
    this.name = name;
    
    this.version = version;
    
    this.nodeRegistry = new NodeRegistry();
    
    this.tocTypes = [];

    
    this.addNodes(this.getBuiltIns());
  };


  Schema.prototype.addNodes = function addNodes (nodes) {
    if (!nodes) { return }
    forEach(nodes, function(NodeClass) {
      if (!NodeClass.prototype._isNode) {
        console.error('Illegal node class: ', NodeClass);
      } else {
        this.addNode(NodeClass);
      }
    }.bind(this));
  };

  Schema.prototype.addNode = function addNode (NodeClass) {
    this.nodeRegistry.register(NodeClass);
    if (NodeClass.tocType) {
      this.tocTypes.push(NodeClass.type);
    }
  };


  Schema.prototype.getNodeClass = function getNodeClass (name) {
    return this.nodeRegistry.get(name)
  };


  Schema.prototype.getBuiltIns = function getBuiltIns () {
    return []
  };


  Schema.prototype.isInstanceOf = function isInstanceOf (type, parentType) {
    var NodeClass = this.getNodeClass(type);
    if (NodeClass) {
      return Node.isInstanceOf(NodeClass, parentType)
    }
    return false
  };


  Schema.prototype.each = function each () {
    this.nodeRegistry.each.apply(this.nodeRegistry, arguments);
  };


  Schema.prototype.getTocTypes = function getTocTypes () {
    return this.tocTypes
  };


  Schema.prototype.getDefaultTextType = function getDefaultTextType () {
    throw new Error('Schmema.prototype.getDefaultTextType() must be overridden.')
  };

  Schema.prototype.getNodeSchema = function getNodeSchema (type) {
    var NodeClass = this.getNodeClass(type);
    if (!NodeClass) {
      console.error('Unknown node type ', type);
      return null
    }
    return NodeClass.schema
  };

  var DocumentNode = (function (DataNode) {
    function DocumentNode(doc, props) {
      DataNode.call(this, doc, props);
    }

    if ( DataNode ) DocumentNode.__proto__ = DataNode;
    DocumentNode.prototype = Object.create( DataNode && DataNode.prototype );
    DocumentNode.prototype.constructor = DocumentNode;

    DocumentNode.prototype._initialize = function _initialize (doc, props) {
      this.document = doc;
      DataNode.prototype._initialize.call(this, props);
    };

    
    DocumentNode.prototype.getDocument = function getDocument () {
      return this.document
    };

    
    DocumentNode.prototype.hasParent = function hasParent () {
      return Boolean(this.parent)
    };

    
    DocumentNode.prototype.getParent = function getParent () {
      if (isString(this.parent)) { return this.document.get(this.parent) }
      return this.parent
    };

    
    DocumentNode.prototype.getRoot = function getRoot () {
      var node = this;
      while(node.parent) {
        node = node.parent;
      }
      return node
    };

    DocumentNode.prototype.getContainerRoot = function getContainerRoot () {
      var node = this;
      while(node.parent) {
        
        if (node.parent.isContainer()) { return node }
        
        node = node.parent;
      }
      return node
    };

    
    DocumentNode.prototype.hasChildren = function hasChildren () {
      return false
    };

    
    DocumentNode.prototype.getChildIndex = function getChildIndex (child) { 
      return -1
    };

    
    DocumentNode.prototype.getChildAt = function getChildAt (idx) { 
      return null
    };

    
    DocumentNode.prototype.getChildCount = function getChildCount () {
      return 0
    };

    
    

    

    
    DocumentNode.prototype.isBlock = function isBlock () {
      return Boolean(this.constructor.isBlock)
    };

    
    DocumentNode.prototype.isText = function isText () {
      return Boolean(this.constructor.isText)
    };

    DocumentNode.prototype.isList = function isList () {
      return Boolean(this.constructor.isList)
    };

    DocumentNode.prototype.isListItem = function isListItem () {
      return Boolean(this.constructor.isListItem)
    };

    DocumentNode.prototype.isContainer = function isContainer () {
      return Boolean(this._isContainer)
    };

    

    DocumentNode.prototype.isAnnotation = function isAnnotation () {
      return Boolean(this._isAnnotation)
    };

    DocumentNode.prototype.isPropertyAnnotation = function isPropertyAnnotation () {
      return Boolean(this._isPropertyAnnotation)
    };

    DocumentNode.prototype.isContainerAnnotation = function isContainerAnnotation () {
      return Boolean(this._isContainerAnnotation)
    };

    
    DocumentNode.prototype.isInline = function isInline () {
      return Boolean(this.constructor.isInline)
    };

    return DocumentNode;
  }(Node));

  DocumentNode.prototype._isDocumentNode = true;


  DocumentNode.isBlock = false;


  DocumentNode.isText = false;


  DocumentNode.isPropertyAnnotation = false;


  DocumentNode.isContainerAnnotation = false;


  DocumentNode.isInline = false;

  var ContainerAddress = function ContainerAddress(pos, offset) {
    this.pos = pos;
    this.offset = offset;
  };

  ContainerAddress.prototype.isBefore = function isBefore (other, strict) {
    strict = Boolean(strict);
    if (this.pos < other.pos) {
      return true
    } else if (this.pos > other.pos) {
      return false
    } else if (this.offset < other.offset) {
      return true
    } else if (this.offset > other.offset) {
      return false
    }
    if (strict) {
      return false
    } else {
      return true
    }
  };

  ContainerAddress.prototype.isAfter = function isAfter (other, strict) {
    return other.isBefore(this, strict)
  };

  ContainerAddress.prototype.isEqual = function isEqual (other) {
    return (this.pos === other.pos && this.offset === other.offset)
  };

  ContainerAddress.prototype.toString = function toString () {
    return [this.pos,'.',this.offset].join('')
  };

  function ContainerMixin (DocumentNode) {

    var AbstractContainer = (function (DocumentNode) {
      function AbstractContainer () {
        DocumentNode.apply(this, arguments);
      }

      if ( DocumentNode ) AbstractContainer.__proto__ = DocumentNode;
      AbstractContainer.prototype = Object.create( DocumentNode && DocumentNode.prototype );
      AbstractContainer.prototype.constructor = AbstractContainer;

      var prototypeAccessors = { length: { configurable: true } };

      AbstractContainer.prototype.contains = function contains (nodeId) {
        return this.getPosition(nodeId) >= 0
      };

      AbstractContainer.prototype.getPosition = function getPosition (node, strict) {
        if (isString(node)) {
          node = this.document.get(node);
        }
        if (!node) { return -1 }
        var pos = this._getPosition(node);
        if (strict && pos < 0) {
          throw new Error('Node is not within this container: ' + node.id)
        }
        return pos
      };

      AbstractContainer.prototype.getNodeAt = function getNodeAt (idx) {
        var nodeId = this.getNodeIdAt(idx);
        if (nodeId) {
          return this.getDocument().get(nodeId)
        }
      };

      AbstractContainer.prototype.getNodeIdAt = function getNodeIdAt (idx) {
        var content = this.getContent();
        if (idx < 0 || idx >= content.length) {
          
          return undefined
        } else {
          return content[idx]
        }
      };

      AbstractContainer.prototype.getNodes = function getNodes$$1 () {
        var doc = this.getDocument();
        return this.getContent().map(function (id) { return doc.get(id); }).filter(Boolean)
      };

      AbstractContainer.prototype.show = function show (nodeId, pos) {
        
        var arg1 = arguments[0];
        if (!isString(arg1)) {
          if (arg1._isNode) {
            nodeId = arg1.id;
          }
        }
        if (arguments.length > 1) {
          console.error('DEPRECATED: use container.showAt(pos, nodeId) instead');
        } else {
          pos = this.getLength();
        }
        return this.showAt(pos, nodeId)
      };

      AbstractContainer.prototype.showAt = function showAt (pos, nodeId) {
        var doc = this.getDocument();
        var length = this.getLength();
        if (!isNumber(pos) || pos < 0 || pos > length) {
          throw new Error('Index out of bounds')
        }
        if (!isString(nodeId)) {
          if (nodeId._isNode) {
            nodeId = nodeId.id;
          } else {
            throw new Error('Invalid argument.')
          }
        }
        doc.update(this.getContentPath(), { type: 'insert', pos: pos, value: nodeId });
      };

      AbstractContainer.prototype.hide = function hide (nodeId) {
        var pos = this.getPosition(nodeId);
        this.hideAt(pos);
      };

      AbstractContainer.prototype.hideAt = function hideAt (pos) {
        var length = this.getLength();
        if (pos >= 0 && pos < length) {
          var doc = this.getDocument();
          doc.update(this.getContentPath(), { type: 'delete', pos: pos });
        } else {
          throw new Error('Index out of bounds.')
        }
      };

      AbstractContainer.prototype.getAddress = function getAddress (coor) {
        if (!coor._isCoordinate) {
          
          throw new Error('Illegal argument: Container.getAddress(coor) expects a Coordinate instance.')
        }
        var nodeId = coor.path[0];
        var nodePos = this.getPosition(nodeId);
        var offset;
        if (coor.isNodeCoordinate()) {
          if (coor.offset > 0) {
            offset = Number.MAX_VALUE;
          } else {
            offset = 0;
          }
        } else {
          offset = coor.offset;
        }
        return new ContainerAddress(nodePos, offset)
      };

      AbstractContainer.prototype.getLength = function getLength () {
        return this.getContent().length
      };

      prototypeAccessors.length.get = function () {
        return this.getLength()
      };

      AbstractContainer.prototype._getPosition = function _getPosition (node) {
        if (this._isCaching) {
          return this._getCachedPosition(node)
        } else {
          return this._lookupPosition(node)
        }
      };

      AbstractContainer.prototype._getCachedPosition = function _getCachedPosition (node) {
        var cache = this._cachedPositions || this._fillCache();
        var nodeId = node.id;
        var pos = -1;
        if (cache.hasOwnProperty(nodeId)) {
          pos = cache[nodeId];
        } else {
          pos = this._lookupPosition(node);
          cache[nodeId] = pos;
        }
        return pos
      };

      AbstractContainer.prototype._fillCache = function _fillCache () {
        var positions = {};
        this.nodes.forEach(function (id, pos) {
          positions[id] = pos;
        });
        this._cachedPositions = positions;
        return positions
      };

      AbstractContainer.prototype._invalidateCache = function _invalidateCache () {
        this._cachedPositions = null;
      };

      AbstractContainer.prototype._lookupPosition = function _lookupPosition (node) {
        if (node.hasParent()) {
          node = node.getContainerRoot();
        }
        return this.getContent().indexOf(node.id)
      };

      AbstractContainer.prototype._enableCaching = function _enableCaching () {
        
        if (this.document) {
          this.document.data.on('operation:applied', this._onOperationApplied, this);
          this._isCaching = true;
        }
      };

      AbstractContainer.prototype._onOperationApplied = function _onOperationApplied (op) {
        if (op.type === 'set' || op.type === 'update') {
          if (op.path[0] === this.id) {
            this._invalidateCache();
          }
        }
      };

      AbstractContainer.prototype._onDocumentChange = function _onDocumentChange (change) {
        if (change.hasUpdated(this.getContentPath())) {
          this._invalidateCache();
        }
      };

      
      

      AbstractContainer.prototype.hasChildren = function hasChildren () {
        return this.getContent().length > 0
      };

      AbstractContainer.prototype.getChildIndex = function getChildIndex (child) {
        return this.getContent().indexOf(child.id)
      };

      AbstractContainer.prototype.getChildren = function getChildren () {
        return getNodes(this.getDocument(), this.getContent())
      };

      AbstractContainer.prototype.getChildAt = function getChildAt (idx) {
        var childrenIds = this.getContent();
        if (idx < 0 || idx >= childrenIds.length) {
          throw new Error('Array index out of bounds: ' + idx + ", " + childrenIds.length)
        }
        return this.getDocument().get(childrenIds[idx], 'strict')
      };

      AbstractContainer.prototype.getChildCount = function getChildCount () {
        return this.getContent().length
      };

      Object.defineProperties( AbstractContainer.prototype, prototypeAccessors );

      return AbstractContainer;
    }(DocumentNode));
    return AbstractContainer

  }

  var Container = (function (superclass) {
    function Container() {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      superclass.apply(this, args);

      
      
      
      this._enableCaching();
    }

    if ( superclass ) Container.__proto__ = superclass;
    Container.prototype = Object.create( superclass && superclass.prototype );
    Container.prototype.constructor = Container;

    Container.prototype.dispose = function dispose () {
      this.document.off(this);
    };

    Container.prototype.getContentPath = function getContentPath () {
      return [this.id, 'nodes']
    };

    Container.prototype.getContent = function getContent () {
      return this.nodes
    };

    return Container;
  }(ContainerMixin(DocumentNode)));

  Container.prototype._isContainer = true;

  Container.schema = {
    type: 'container',
    nodes: { type: ['array', 'id'], default: [] }
  };

  function AnnotationMixin(DocumentNode) {

    var AbstractAnnotation = (function (DocumentNode) {
      function AbstractAnnotation(doc, props) {
        DocumentNode.call(this, doc, _normalizedProps(props));

        
        this.start = new Coordinate(this.start);
        this.end = new Coordinate(this.end);
      }

      if ( DocumentNode ) AbstractAnnotation.__proto__ = DocumentNode;
      AbstractAnnotation.prototype = Object.create( DocumentNode && DocumentNode.prototype );
      AbstractAnnotation.prototype.constructor = AbstractAnnotation;

      var prototypeAccessors = { path: { configurable: true },startPath: { configurable: true },startOffset: { configurable: true },endPath: { configurable: true },endOffset: { configurable: true } };

      

      prototypeAccessors.path.get = function () {
        console.warn('DEPRECATED: use annotation.start.path instead');
        return this.start.path
      };

      AbstractAnnotation.prototype.getPath = function getPath () {
        return this.start.path
      };

      prototypeAccessors.startPath.get = function () {
        console.warn('DEPRECATED: use annotation.start.path instead.');
        return this.start.path
      };

      prototypeAccessors.startPath.set = function (path) {
        console.warn('DEPRECATED: use annotation.start.path instead.');
        this.start.path = path;
      };

      prototypeAccessors.startOffset.get = function () {
        console.warn('DEPRECATED: use annotation.start.offset instead.');
        return this.start.offset
      };

      prototypeAccessors.startOffset.set = function (offset) {
        console.warn('DEPRECATED: use annotation.start.offset instead.');
        this.start.offset = offset;
      };

      prototypeAccessors.endPath.get = function () {
        console.warn('DEPRECATED: use annotation.end.path instead.');
        return this.end.path
      };

      prototypeAccessors.endPath.set = function (path) {
        console.warn('DEPRECATED: use annotation.end.path instead.');
        this.end.path = path;
      };

      prototypeAccessors.endOffset.get = function () {
        console.warn('DEPRECATED: use annotation.end.offset instead.');
        return this.end.offset
      };

      prototypeAccessors.endOffset.set = function (offset) {
        console.warn('DEPRECATED: use annotation.end.offset instead.');
        this.end.offset = offset;
      };

      

      
      AbstractAnnotation.prototype.getText = function getText () {
        var doc = this.getDocument();
        if (!doc) {
          console.warn('Trying to use a Annotation which is not attached to the document.');
          return ""
        }
        return getTextForSelection(doc, this.getSelection())
      };

      AbstractAnnotation.prototype.isAnnotation = function isAnnotation () {
        return true
      };

      
      AbstractAnnotation.prototype.canSplit = function canSplit () {
        return true
      };

      
      AbstractAnnotation.prototype.isAnchor = function isAnchor () {
        return false
      };

      
      AbstractAnnotation.prototype.getSelection = function getSelection () {
        var doc = this.getDocument();
        
        if (!doc) {
          console.warn('Trying to use a ContainerAnnotation which is not attached to the document.');
          return Selection.nullSelection()
        }
        if (this._isContainerAnnotation) {
          return doc.createSelection({
            type: "container",
            containerId: this.containerId,
            startPath: this.start.path,
            startOffset: this.start.offset,
            endPath: this.end.path,
            endOffset: this.end.offset
          })
        } else {
          return this.getDocument().createSelection({
            type: 'property',
            path: this.start.path,
            startOffset: this.start.offset,
            endOffset: this.end.offset
          })
        }
      };

      AbstractAnnotation.prototype._updateRange = function _updateRange (tx, sel) {
        if (sel.isContainerSelection()) {
          
          if (!isEqual(this.start.path, sel.start.path)) {
            tx.set([this.id, 'start', 'path'], sel.start.path);
          }
          if (this.start.offset !== sel.start.offset) {
            tx.set([this.id, 'start', 'offset'], sel.start.offset);
          }
          if (!isEqual(this.end.path, sel.end.path)) {
            tx.set([this.id, 'end', 'path'], sel.end.path);
          }
          if (this.end.offset !== sel.end.offset) {
            tx.set([this.id, 'end', 'offset'], sel.end.offset);
          }
        } else if (sel.isPropertySelection()) {
          if (!isArrayEqual(this.start.path, sel.start.path)) {
            tx.set([this.id, 'path'], sel.start.path);
          }
          
          if (this.start.offset !== sel.start.offset) {
            tx.set([this.id, 'start', 'offset'], sel.start.offset);
          }
          if (this.end.offset !== sel.end.offset) {
            tx.set([this.id, 'end', 'offset'], sel.end.offset);
          }
        } else {
          throw new Error('Invalid selection.')
        }
      };

      Object.defineProperties( AbstractAnnotation.prototype, prototypeAccessors );

      return AbstractAnnotation;
    }(DocumentNode));

    AbstractAnnotation.prototype._isAnnotation = true;

    AbstractAnnotation.schema = {
      start: { type: "coordinate", default: { path: [], offset: 0 } },
      end: { type: "coordinate", default: { path: [], offset: 0 } }
    };

    return AbstractAnnotation
  }

  function _normalizedProps(props) {
    
    
    if (!props.hasOwnProperty('start')) {
      
      
      
      var start, end;
      if (props.hasOwnProperty('startPath') || props.hasOwnProperty('path')) {
        start = {
          path: props.startPath || props.path,
          offset: props.startOffset
        };
      }
      if (props.hasOwnProperty('endPath') || props.hasOwnProperty('endOffset')) {
        end = {
          path: props.endPath || props.path,
          offset: props.endOffset
        };
      }
      if (start && !end) {
        end = cloneDeep(start);
      }
      if (start) {
        props = Object.assign({}, props);
        delete props.path;
        delete props.startPath;
        delete props.endPath;
        delete props.startOffset;
        delete props.endOffset;
        props.start = start;
        props.end = end;
      }
    } else if (props.hasOwnProperty('end') && !props.end.path) {
      props.end.path = props.start.path;
    }
    return props
  }

  var PropertyAnnotation = (function (superclass) {
    function PropertyAnnotation () {
      superclass.apply(this, arguments);
    }if ( superclass ) PropertyAnnotation.__proto__ = superclass;
    PropertyAnnotation.prototype = Object.create( superclass && superclass.prototype );
    PropertyAnnotation.prototype.constructor = PropertyAnnotation;

    

    return PropertyAnnotation;
  }(AnnotationMixin(DocumentNode)));

  PropertyAnnotation.prototype._isAnnotation = true;
  PropertyAnnotation.prototype._isPropertyAnnotation = true;

  PropertyAnnotation.isPropertyAnnotation = true;
  PropertyAnnotation.autoExpandRight = true;

  PropertyAnnotation.schema = {
    type: "annotation",
    
    
    _content: { type: "string", optional: true}
  };

  var ContainerAnnotation = (function (superclass) {
    function ContainerAnnotation () {
      superclass.apply(this, arguments);
    }

    if ( superclass ) ContainerAnnotation.__proto__ = superclass;
    ContainerAnnotation.prototype = Object.create( superclass && superclass.prototype );
    ContainerAnnotation.prototype.constructor = ContainerAnnotation;

    ContainerAnnotation.prototype.setHighlighted = function setHighlighted (highlighted, scope) {
      if (this.highlighted !== highlighted) {
        this.highlighted = highlighted;
        this.highlightedScope = scope;
        this.emit('highlighted', highlighted, scope);
        forEach(this.fragments, function(frag) {
          frag.emit('highlighted', highlighted, scope);
        });
      }
    };

    return ContainerAnnotation;
  }(AnnotationMixin(DocumentNode)));

  ContainerAnnotation.schema = {
    type: "container-annotation",
    containerId: "string",
    start: "coordinate",
    end: "coordinate"
  };

  ContainerAnnotation.prototype._isAnnotation = true;
  ContainerAnnotation.prototype._isContainerAnnotation = true;

  var DocumentSchema = (function (Schema$$1) {
    function DocumentSchema(schemaSpec) {
      Schema$$1.call(this, schemaSpec.name, schemaSpec.version);
      
      if (!schemaSpec.DocumentClass) {
        throw new Error('DocumentClass is mandatory')
      }
      Object.assign(this, schemaSpec);
    }

    if ( Schema$$1 ) DocumentSchema.__proto__ = Schema$$1;
    DocumentSchema.prototype = Object.create( Schema$$1 && Schema$$1.prototype );
    DocumentSchema.prototype.constructor = DocumentSchema;

    DocumentSchema.prototype.getDocumentClass = function getDocumentClass () {
      return this.DocumentClass
    };

    
    DocumentSchema.prototype.getDefaultTextType = function getDefaultTextType () {
      return this.defaultTextType
    };

    
    DocumentSchema.prototype.getBuiltIns = function getBuiltIns () {
      return [DocumentNode, PropertyAnnotation, Container, ContainerAnnotation]
    };

    return DocumentSchema;
  }(Schema));

  var EditingBehavior = function EditingBehavior() {
    this._merge = {};
    this._mergeComponents = {};
    this._break = {};
  };

  EditingBehavior.prototype.defineMerge = function defineMerge (firstType, secondType, impl) {
    if (!this._merge[firstType]) {
      this._merge[firstType] = {};
    }
    this._merge[firstType][secondType] = impl;
    return this
  };

  EditingBehavior.prototype.canMerge = function canMerge (firstType, secondType) {
    return (this._merge[firstType] && this._merge[firstType][secondType])
  };

  EditingBehavior.prototype.getMerger = function getMerger (firstType, secondType) {
    return this._merge[firstType][secondType]
  };

  EditingBehavior.prototype.defineComponentMerge = function defineComponentMerge (nodeType, impl) {
    this._mergeComponents[nodeType] = impl;
  };

  EditingBehavior.prototype.canMergeComponents = function canMergeComponents (nodeType) {
    return this._mergeComponents[nodeType]
  };

  EditingBehavior.prototype.getComponentMerger = function getComponentMerger (nodeType) {
    return this._mergeComponents[nodeType]
  };

  EditingBehavior.prototype.defineBreak = function defineBreak (nodeType, impl) {
    this._break[nodeType] = impl;
    return this
  };

  EditingBehavior.prototype.canBreak = function canBreak (nodeType) {
    return this._break[nodeType]
  };

  EditingBehavior.prototype.getBreaker = function getBreaker (nodeType) {
    return this._break[nodeType]
  };

  var ComponentRegistry = (function (Registry$$1) {
    function ComponentRegistry(entries) {
      Registry$$1.call(this, entries, function(ComponentClass) {
        if (!ComponentClass.prototype._isComponent) {
          throw new Error('Component registry: wrong type. Expected a ComponentClass. Was: ' + String(ComponentClass))
        }
      });
    }

    if ( Registry$$1 ) ComponentRegistry.__proto__ = Registry$$1;
    ComponentRegistry.prototype = Object.create( Registry$$1 && Registry$$1.prototype );
    ComponentRegistry.prototype.constructor = ComponentRegistry;

    return ComponentRegistry;
  }(Registry));

  function extend() {
    var args = [], len = arguments.length;
    while ( len-- ) args[ len ] = arguments[ len ];

    return Object.assign.apply(Object, args)
  }

  var ArrayIterator = function ArrayIterator(arr) {
    this.arr = arr;
    this.pos = -1;
  };

  var prototypeAccessors$4 = { _isArrayIterator: { configurable: true } };

  prototypeAccessors$4._isArrayIterator.get = function () {
    return true
  };


  ArrayIterator.prototype.hasNext = function hasNext () {
    return this.pos < this.arr.length - 1
  };


  ArrayIterator.prototype.next = function next () {
    this.pos += 1;
    var next = this.arr[this.pos];
    return next
  };


  ArrayIterator.prototype.back = function back () {
    if (this.pos >= 0) {
      this.pos -= 1;
    }
    return this
  };

  ArrayIterator.prototype.peek = function peek () {
    return this.arr[this.pos+1]
  };

  Object.defineProperties( ArrayIterator.prototype, prototypeAccessors$4 );

  function findIndex$1(arr, predicate) {
    if (!isFunction$1(predicate)) { return arr.indexOf(predicate) }
    for (var i = 0; i < arr.length; i++) {
      if (predicate(arr[i])) { return i }
    }
    return -1
  }

  var DOMEventListener = function DOMEventListener(eventName, handler, options) {
    
    if (!isString(eventName) || !isFunction$1(handler)) {
      throw new Error("Illegal arguments: 'eventName' must be a String, and 'handler' must be a Function.")
    }
    options = options || {};
    var origHandler = handler;
    var context = options.context;
    var capture = Boolean(options.capture);

    if (context) {
      handler = handler.bind(context);
    }
    if (options.once === true) {
      handler = _once(this, handler);
    }

    this.eventName = eventName;
    this.originalHandler = origHandler;
    this.handler = handler;
    this.capture = capture;
    this.context = context;
    this.options = options;
    
    this._el = null;
  };

  DOMEventListener.prototype._isDOMEventListener = true;

  DOMEventListener.findIndex = function(eventListeners, eventName, handler) {
    var idx = -1;
    if (arguments[1]._isDOMEventListener) {
      idx = eventListeners.indexOf(arguments[1]);
    } else {
      idx = findIndex$1(eventListeners,
        _matches.bind(null, {
          eventName: eventName,
          originalHandler: handler
        })
      );
    }
    return idx
  };

  function _matches(l1, l2) {
    return l1.eventName === l2.eventName && l1.originalHandler === l2.originalHandler
  }

  function _once(listener, handler) {
    return function(event) {
      handler(event);
      listener._el.removeEventListener(listener);
    }
  }

  var NOT_IMPLEMENTED = 'This method is not implemented.';


  var DOMElement = function DOMElement () {};

  var prototypeAccessors$5 = { id: { configurable: true },tagName: { configurable: true },nodeName: { configurable: true },nodeType: { configurable: true },className: { configurable: true },textContent: { configurable: true },innerHTML: { configurable: true },outerHTML: { configurable: true },firstChild: { configurable: true },lastChild: { configurable: true },nextSibling: { configurable: true },previousSibling: { configurable: true },parentNode: { configurable: true },height: { configurable: true },width: { configurable: true },value: { configurable: true } };

  DOMElement.prototype.getNativeElement = function getNativeElement () {
    
    throw new Error(NOT_IMPLEMENTED)
  };


  DOMElement.prototype.isTextNode = function isTextNode () {
    
    return false
  };


  DOMElement.prototype.isElementNode = function isElementNode () {
    
    return false
  };


  DOMElement.prototype.isCommentNode = function isCommentNode () {
    
    return false
  };


  DOMElement.prototype.isDocumentNode = function isDocumentNode () {
    
    return false
  };


  DOMElement.prototype.getTagName = function getTagName () {
    
    throw new Error(NOT_IMPLEMENTED)
  };


  DOMElement.prototype.setTagName = function setTagName (tagName) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };


  DOMElement.prototype.getId = function getId () {
    return this.getAttribute('id')
  };


  DOMElement.prototype.setId = function setId (id) {
    this.setAttribute('id', id);
    return this
  };


  DOMElement.prototype.hasClass = function hasClass (className) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };


  DOMElement.prototype.addClass = function addClass (classString) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };


  DOMElement.prototype.removeClass = function removeClass (classString) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };

  DOMElement.prototype.hasAttribute = function hasAttribute (name) {
    return Boolean(this.getAttribute(name))
  };


  DOMElement.prototype.attr = function attr () {
    if (arguments.length === 1) {
      if (isString(arguments[0])) {
        return this.getAttribute(arguments[0])
      } else if (isObject(arguments[0])) {
        forEach(arguments[0], function(value, name) {
          this.setAttribute(name, value);
        }.bind(this));
      }
    } else if (arguments.length === 2) {
      this.setAttribute(arguments[0], arguments[1]);
    }
    return this
  };


  DOMElement.prototype.removeAttr = function removeAttr (name) {
    var names = name.split(/\s+/);
    if (names.length === 1) {
      this.removeAttribute(name);
    } else {
      names.forEach(function(name) {
        this.removeAttribute(name);
      }.bind(this));
    }
    return this
  };


  DOMElement.prototype.getAttribute = function getAttribute (name) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };


  DOMElement.prototype.setAttribute = function setAttribute (name, value) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };

  DOMElement.prototype.removeAttribute = function removeAttribute (name) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };

  DOMElement.prototype.getAttributes = function getAttributes () {
    
    throw new Error(NOT_IMPLEMENTED)
  };


  DOMElement.prototype.htmlProp = function htmlProp () {
    if (arguments.length === 1) {
      if (isString(arguments[0])) {
        return this.getProperty(arguments[0])
      } else if (isObject(arguments[0])) {
        forEach(arguments[0], function(value, name) {
          this.setProperty(name, value);
        }.bind(this));
      }
    } else if (arguments.length === 2) {
      this.setProperty(arguments[0], arguments[1]);
    }
    return this
  };

  DOMElement.prototype.getProperty = function getProperty (name) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };

  DOMElement.prototype.setProperty = function setProperty (name, value) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };


  DOMElement.prototype.val = function val (value) {
    if (arguments.length === 0) {
      return this.getValue()
    } else {
      this.setValue(value);
      return this
    }
  };

  DOMElement.prototype.getValue = function getValue () {
    return this.getProperty('value')
  };

  DOMElement.prototype.setValue = function setValue (value) {
    this.setProperty('value', value);
    return this
  };


  DOMElement.prototype.css = function css () {
    
    if (arguments.length === 1) {
      
      if (isString(arguments[0])) {
        return this.getStyle(arguments[0])
      } else if (isObject(arguments[0])) {
        forEach(arguments[0], function(value, name) {
          this.setStyle(name, value);
        }.bind(this));
      } else {
        throw new Error('Illegal arguments.')
      }
    } else if (arguments.length === 2) {
      this.setStyle(arguments[0], arguments[1]);
    } else {
      throw new Error('Illegal arguments.')
    }
    return this
  };

  DOMElement.prototype.getStyle = function getStyle (name) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };

  DOMElement.prototype.setStyle = function setStyle (name, value) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };


  DOMElement.prototype.text = function text (text$1) {
    if (arguments.length === 0) {
      return this.getTextContent()
    } else {
      this.setTextContent(text$1);
    }
    return this
  };


  DOMElement.prototype.getTextContent = function getTextContent () {
    
    throw new Error(NOT_IMPLEMENTED)
  };


  DOMElement.prototype.setTextContent = function setTextContent (text) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };


  DOMElement.prototype.html = function html (html$1) {
    if (arguments.length === 0) {
      return this.getInnerHTML()
    } else {
      this.setInnerHTML(html$1);
    }
    return this
  };


  DOMElement.prototype.getInnerHTML = function getInnerHTML () {
    
    throw new Error(NOT_IMPLEMENTED)
  };


  DOMElement.prototype.getInnerXML = function getInnerXML () {
    return this.getInnerHTML()
  };


  DOMElement.prototype.setInnerHTML = function setInnerHTML (html) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };

  DOMElement.prototype.setInnerXML = function setInnerXML (xml) {
    return this.setInnerHTML(xml)
  };


  DOMElement.prototype.getOuterHTML = function getOuterHTML () {
    
    throw new Error(NOT_IMPLEMENTED)
  };

  DOMElement.prototype.getOuterXML = function getOuterXML () {
    return this.getOuterHTML()
  };


  DOMElement.prototype.on = function on (eventName, handler, context, options) {
    
    if (!isString(eventName)) {
      throw new Error('Illegal argument: "event" must be a String.')
    }
    options = options || {};
    if (context) {
      options.context = context;
    }
    
    if (!handler || !isFunction$1(handler)) {
      throw new Error('Illegal argument: invalid handler function for event ' + eventName)
    }
    this.addEventListener(eventName, handler, options);
    return this
  };


  DOMElement.prototype.off = function off (eventName, handler) {
    
    if (arguments.length === 1 && !isString(eventName)) {
      var context = arguments[0];
      this.getEventListeners().filter(function(l) {
        return l.context === context
      }).forEach(function(l) {
        this.removeEventListener(l);
      }.bind(this));
    } else {
      this.removeEventListener(eventName, handler);
    }
    return this
  };

  DOMElement.prototype.addEventListener = function addEventListener (eventName, handler, options) {
      if ( options === void 0 ) options = {};

    var listener;
    if (arguments.length === 1 && arguments[0]) {
      listener = arguments[0];
    } else {
      listener = this._createEventListener(eventName, handler, options);
    }
    if (!this.eventListeners) {
      this.eventListeners = [];
    }
    listener._el = this;
    this.eventListeners.push(listener);
    this._addEventListenerNative(listener);
    return this
  };

  DOMElement.prototype._createEventListener = function _createEventListener (eventName, handler, options) {
    return new DOMEventListener(eventName, handler, options)
  };

  DOMElement.prototype._addEventListenerNative = function _addEventListenerNative (listener) {}; 

  DOMElement.prototype.removeEventListener = function removeEventListener (eventName, handler) {
    if (!this.eventListeners) { return }
    
    var listener = null, idx = -1;
    idx = DOMEventListener.findIndex(this.eventListeners, eventName, handler);
    listener = this.eventListeners[idx];
    if (idx > -1) {
      this.eventListeners.splice(idx, 1);
      
      listener._el = null;
      this._removeEventListenerNative(listener);
    }
    return this
  };

  DOMElement.prototype._removeEventListenerNative = function _removeEventListenerNative (listener) {}; 

  DOMElement.prototype.removeAllEventListeners = function removeAllEventListeners () {
      var this$1 = this;

    if (!this.eventListeners) { return }
    for (var i = 0; i < this.eventListeners.length; i++) {
      var listener = this$1.eventListeners[i];
      
      listener._el = null;
      this$1._removeEventListenerNative(listener);
    }
    delete this.eventListeners;
  };

  DOMElement.prototype.getEventListeners = function getEventListeners () {
    return this.eventListeners || []
  };


  DOMElement.prototype.getNodeType = function getNodeType () {
    
    throw new Error(NOT_IMPLEMENTED)
  };

  DOMElement.prototype.getContentType = function getContentType () {
    
    throw new Error(NOT_IMPLEMENTED)
  };

  DOMElement.prototype.getChildCount = function getChildCount () {
    
    throw new Error(NOT_IMPLEMENTED)
  };


  DOMElement.prototype.getChildNodes = function getChildNodes () {
    
    throw new Error(NOT_IMPLEMENTED)
  };


  DOMElement.prototype.getChildren = function getChildren () {
    
    throw new Error(NOT_IMPLEMENTED)
  };

  DOMElement.prototype.getChildAt = function getChildAt (pos) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };

  DOMElement.prototype.getChildIndex = function getChildIndex (child) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };

  DOMElement.prototype.getChildNodeIterator = function getChildNodeIterator () {
    return new ArrayIterator(this.getChildNodes())
  };

  DOMElement.prototype.getLastChild = function getLastChild () {
    
    throw new Error(NOT_IMPLEMENTED)
  };

  DOMElement.prototype.getFirstChild = function getFirstChild () {
    
    throw new Error(NOT_IMPLEMENTED)
  };

  DOMElement.prototype.getNextSibling = function getNextSibling () {
    
    throw new Error(NOT_IMPLEMENTED)
  };

  DOMElement.prototype.getPreviousSibling = function getPreviousSibling () {
    
    throw new Error(NOT_IMPLEMENTED)
  };


  DOMElement.prototype.clone = function clone (deep) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };


  DOMElement.prototype.createElement = function createElement (str) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };

  DOMElement.prototype.createTextNode = function createTextNode (text) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };

  DOMElement.prototype.createComment = function createComment (data) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };

  DOMElement.prototype.createProcessingInstruction = function createProcessingInstruction (name, data) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };

  DOMElement.prototype.createCDATASection = function createCDATASection (data) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };


  DOMElement.prototype.is = function is (cssSelector) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };


  DOMElement.prototype.getParent = function getParent () {
    
    throw new Error(NOT_IMPLEMENTED)
  };


  DOMElement.prototype.getOwnerDocument = function getOwnerDocument () {
    
    throw new Error(NOT_IMPLEMENTED)
  };


  DOMElement.prototype.getDoctype = function getDoctype () {
    
    throw new Error('NOT_IMPLEMENTED')
  };

  DOMElement.prototype.setDoctype = function setDoctype (qualifiedNameStr, publicId, systemId) { 
    throw new Error('NOT_IMPLEMENTED')
  };

  DOMElement.prototype.setDocType = function setDocType () {
      var ref;

      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];
    console.error('DEPRECATED: use setDoctype() instead');
    return (ref = this).setDoctype.apply(ref, args)
  };


  DOMElement.prototype.find = function find (cssSelector) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };


  DOMElement.prototype.findAll = function findAll (cssSelector) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };


  DOMElement.prototype.append = function append (child) {
    var children;
    if (arguments.length === 1) {
      if (isArray(child)) {
        children = child;
      } else {
        this.appendChild(child);
        return this
      }
    } else {
      children = arguments;
    }
    if (children) {
      Array.prototype.forEach.call(children, this.appendChild.bind(this));
    }
    return this
  };

  DOMElement.prototype.appendChild = function appendChild (child) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };


  DOMElement.prototype.insertAt = function insertAt (pos, child) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };

  DOMElement.prototype.insertBefore = function insertBefore (newChild, before) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };


  DOMElement.prototype.removeAt = function removeAt (pos) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };

  DOMElement.prototype.removeChild = function removeChild (child) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };

  DOMElement.prototype.replaceChild = function replaceChild (oldChild, newChild) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };


  DOMElement.prototype.remove = function remove () {
    var parent = this.getParent();
    if (parent) {
      parent.removeChild(this);
    }
  };


  DOMElement.prototype.empty = function empty () {
    
    throw new Error(NOT_IMPLEMENTED)
  };

  DOMElement.prototype.serialize = function serialize () {
    return this.getOuterHTML()
  };

  DOMElement.prototype.isInDocument = function isInDocument () {
    var el = this;
    while(el) {
      if (el.isDocumentNode()) {
        return true
      }
      el = el.getParent();
    }
  };


  DOMElement.prototype.focus = function focus () {
    
    return this
  };


  DOMElement.prototype.select = function select () {
    
    return this
  };


  DOMElement.prototype.blur = function blur () {
    
    return this
  };


  DOMElement.prototype.click = function click () {
    
    return this
  };



  DOMElement.prototype.getWidth = function getWidth () {
    
    return 0
  };

  DOMElement.prototype.getHeight = function getHeight () {
    
    return 0
  };


  DOMElement.prototype.getOuterHeight = function getOuterHeight (withMargin) { 
    
    return 0
  };


  DOMElement.prototype.getOffset = function getOffset () {
    
    return { top: 0, left: 0 }
  };


  DOMElement.prototype.getPosition = function getPosition () {
    
    return { top: 0, left: 0 }
  };


  DOMElement.prototype.getElementFactory = function getElementFactory () {
    return this.createElement.bind(this)
  };


  DOMElement.prototype.emit = function emit (name, data) { 
    
    throw new Error(NOT_IMPLEMENTED)
  };



  prototypeAccessors$5.id.get = function () {
    return this.getId()
  };

  prototypeAccessors$5.id.set = function (id) {
    this.setId(id);
  };

  prototypeAccessors$5.tagName.get = function () {
    return this.getTagName()
  };

  prototypeAccessors$5.tagName.set = function (tagName) {
    this.setTagName(tagName);
  };

  prototypeAccessors$5.nodeName.get = function () {
    return this.getTagName()
  };

  prototypeAccessors$5.nodeType.get = function () {
    return this.getNodeType()
  };

  prototypeAccessors$5.className.get = function () {
    return this.getAttribute('class')
  };

  prototypeAccessors$5.className.set = function (className) {
    this.setAttribute('class', className);
  };

  prototypeAccessors$5.textContent.get = function () {
    return this.getTextContent()
  };

  prototypeAccessors$5.textContent.set = function (text) {
    this.setTextContent(text);
  };

  prototypeAccessors$5.innerHTML.get = function () {
    return this.getInnerHTML()
  };

  prototypeAccessors$5.innerHTML.set = function (html) {
    this.setInnerHTML(html);
  };

  prototypeAccessors$5.outerHTML.get = function () {
    return this.getOuterHTML()
  };

  prototypeAccessors$5.firstChild.get = function () {
    return this.getFirstChild()
  };

  prototypeAccessors$5.lastChild.get = function () {
    return this.getLastChild()
  };

  prototypeAccessors$5.nextSibling.get = function () {
    return this.getNextSibling()
  };

  prototypeAccessors$5.previousSibling.get = function () {
    return this.getPreviousSibling()
  };

  prototypeAccessors$5.parentNode.get = function () {
    return this.getParent()
  };

  prototypeAccessors$5.height.get = function () {
    return this.getHeight()
  };

  prototypeAccessors$5.width.get = function () {
    return this.getWidth()
  };

  prototypeAccessors$5.value.get = function () {
    return this.getValue()
  };

  prototypeAccessors$5.value.set = function (value) {
    return this.setValue(value)
  };

  Object.defineProperties( DOMElement.prototype, prototypeAccessors$5 );

  DOMElement.prototype._isDOMElement = true;

  DOMElement.pxStyles = {
    top: true,
    bottom: true,
    left: true,
    right: true,
    height: true,
    width: true
  };

  DOMElement.EMPTY_HTML = '<html><head></head><body></body></html>';

  var SIGNATURE = uuid('_BrowserDOMElement');

  function _attach(nativeEl, browserDOMElement) {
    nativeEl[SIGNATURE] = browserDOMElement;
  }

  function _detach(nativeEl) {
    delete nativeEl[SIGNATURE];
  }

  function _unwrap(nativeEl) {
    return nativeEl[SIGNATURE]
  }

  var BrowserDOMElement = (function (DOMElement$$1) {
    function BrowserDOMElement(el) {
      DOMElement$$1.call(this);
      console.assert(el instanceof window.Node, "Expecting native DOM node.");
      this.el = el;
      
      
      _attach(el, this);
    }

    if ( DOMElement$$1 ) BrowserDOMElement.__proto__ = DOMElement$$1;
    BrowserDOMElement.prototype = Object.create( DOMElement$$1 && DOMElement$$1.prototype );
    BrowserDOMElement.prototype.constructor = BrowserDOMElement;

    var prototypeAccessors = { childNodes: { configurable: true },children: { configurable: true },ownerDocument: { configurable: true } };

    BrowserDOMElement.prototype.getNativeElement = function getNativeElement () {
      return this.el
    };

    BrowserDOMElement.prototype.getNodeType = function getNodeType () {
      switch(this.el.nodeType) {
        case window.Node.TEXT_NODE:
          return "text"
        case window.Node.ELEMENT_NODE:
          return 'element'
        case window.Node.DOCUMENT_NODE:
          return 'document'
        case window.Node.COMMENT_NODE:
          return 'comment'
        case window.Node.PROCESSING_INSTRUCTION_NODE:
          return 'directive'
        case window.Node.CDATA_SECTION_NODE:
          return 'cdata'
        default:
          
      }
    };

    BrowserDOMElement.prototype.getDoctype = function getDoctype () {
      if (this.isDocumentNode()) {
        return this.el.doctype
      } else {
        return this.getOwnerDocument().getDoctype()
      }
    };

    BrowserDOMElement.prototype.setDoctype = function setDoctype (qualifiedNameStr, publicId, systemId) {
      var ownerDocument = this._getNativeOwnerDocument();
      var oldDocType = ownerDocument.doctype;
      var newDocType = ownerDocument.implementation.createDocumentType(
        qualifiedNameStr, publicId, systemId
      );
      if (oldDocType) {
        oldDocType.parentNode.replaceChild(newDocType, oldDocType);
      } else {
        ownerDocument.insertBefore(newDocType, ownerDocument.firstChild);
      }
    };

    BrowserDOMElement.prototype.isTextNode = function isTextNode () {
      return (this.el.nodeType === window.Node.TEXT_NODE)
    };

    BrowserDOMElement.prototype.isElementNode = function isElementNode () {
      return (this.el.nodeType === window.Node.ELEMENT_NODE)
    };

    BrowserDOMElement.prototype.isCommentNode = function isCommentNode () {
      return (this.el.nodeType === window.Node.COMMENT_NODE)
    };

    BrowserDOMElement.prototype.isDocumentNode = function isDocumentNode () {
      return (this.el.nodeType === window.Node.DOCUMENT_NODE)
    };

    BrowserDOMElement.prototype.hasClass = function hasClass (className) {
      return this.el.classList.contains(className)
    };

    BrowserDOMElement.prototype.addClass = function addClass (className) {
      this.el.classList.add(className);
      return this
    };

    BrowserDOMElement.prototype.removeClass = function removeClass (className) {
      this.el.classList.remove(className);
      return this
    };

    BrowserDOMElement.prototype.getAttribute = function getAttribute (name) {
      return this.el.getAttribute(name)
    };

    BrowserDOMElement.prototype.setAttribute = function setAttribute (name, value) {
      this.el.setAttribute(name, String(value));
      return this
    };

    BrowserDOMElement.prototype.removeAttribute = function removeAttribute (name) {
      this.el.removeAttribute(name);
      return this
    };

    BrowserDOMElement.prototype.getAttributes = function getAttributes () {
      if (!this.el.attributes._mapAdapter) {
        this.el.attributes._mapAdapter = new AttributesMapAdapter(this.el.attributes);
      }
      return this.el.attributes._mapAdapter
    };

    BrowserDOMElement.prototype.getProperty = function getProperty (name) {
      return this.el[name]
    };

    BrowserDOMElement.prototype.setProperty = function setProperty (name, value) {
      
      
      if (this._isXML()) { throw new Error('setProperty() is only supported for HTML elements.') }
      this.el[name] = value;
      return this
    };

    BrowserDOMElement.prototype.getTagName = function getTagName () {
      
      
      if (this._isXML()) {
        return this.el.tagName
      } else if (this.el.tagName) {
        return this.el.tagName.toLowerCase()
      }
    };

    BrowserDOMElement.prototype.setTagName = function setTagName (tagName) {
      var newEl = this.createElement(tagName);
      var attributes = this.el.attributes;
      var l = attributes.length;
      var i;
      for(i = 0; i < l; i++) {
        var attr = attributes.item(i);
        newEl.setAttribute(attr.name, attr.value);
      }
      if (this.eventListeners) {
        this.eventListeners.forEach(function(listener) {
          newEl.addEventListener(listener.eventName, listener.handler, listener.capture);
        });
      }
      newEl.append(this.getChildNodes());

      this._replaceNativeEl(newEl.getNativeElement());
      return this
    };

    BrowserDOMElement.prototype.getId = function getId () {
      return this.el.id
    };

    BrowserDOMElement.prototype.setId = function setId (id) {
      this.el.id = id;
      return this
    };

    BrowserDOMElement.prototype.getStyle = function getStyle (name) {
      
      var style = this.getComputedStyle();
      return style[name] || this.el.style[name]
    };

    BrowserDOMElement.prototype.getComputedStyle = function getComputedStyle () {
      return window.getComputedStyle(this.el)
    };

    BrowserDOMElement.prototype.setStyle = function setStyle (name, value) {
      if (DOMElement$$1.pxStyles[name] && isNumber(value)) { value = value + 'px'; }
      this.el.style[name] = value;
      return this
    };

    BrowserDOMElement.prototype.getTextContent = function getTextContent () {
      return this.el.textContent
    };

    BrowserDOMElement.prototype.setTextContent = function setTextContent (text) {
      this.el.textContent = text;
      return this
    };

    BrowserDOMElement.prototype.getInnerHTML = function getInnerHTML () {
      if (this._isXML()) {
        var xs = new window.XMLSerializer();
        var result = Array.prototype.map.call(this.el.childNodes, function (c) { return xs.serializeToString(c); });
        return result.join('')
      } else {
        return this.el.innerHTML
      }
    };

    BrowserDOMElement.prototype.setInnerHTML = function setInnerHTML (html) {
      
      
      this.el.innerHTML = html;
      return this
    };

    BrowserDOMElement.prototype.getOuterHTML = function getOuterHTML () {
      
      
      if (this._isXML()) {
        var xs = new window.XMLSerializer();
        return xs.serializeToString(this.el)
      } else {
        return this.el.outerHTML
      }
    };

    BrowserDOMElement.prototype._addEventListenerNative = function _addEventListenerNative (listener) {
      this.el.addEventListener(listener.eventName, listener.handler, listener.capture);
    };

    BrowserDOMElement.prototype._removeEventListenerNative = function _removeEventListenerNative (listener) {
      this.el.removeEventListener(listener.eventName, listener.handler);
    };

    BrowserDOMElement.prototype.getEventListeners = function getEventListeners () {
      return this.eventListeners || []
    };

    BrowserDOMElement.prototype.getChildCount = function getChildCount () {
      return this.el.childNodes.length
    };

    BrowserDOMElement.prototype.getChildNodes = function getChildNodes () {
      var childNodes = [];
      for (var node = this.el.firstChild; node; node = node.nextSibling) {
        childNodes.push(BrowserDOMElement.wrap(node));
      }
      return childNodes
    };

    BrowserDOMElement.prototype.getChildNodeIterator = function getChildNodeIterator () {
      return new BrowserChildNodeIterator(this.el)
    };

    prototypeAccessors.childNodes.get = function () {
      return this.getChildNodes()
    };

    BrowserDOMElement.prototype.getChildren = function getChildren () {
      
      
      var children = [];
      for (var node = this.el.firstChild; node; node = node.nextSibling) {
        if (node.nodeType === window.Node.ELEMENT_NODE) {
          children.push(BrowserDOMElement.wrap(node));
        }
      }
      return children
    };

    prototypeAccessors.children.get = function () {
      return this.getChildren()
    };

    BrowserDOMElement.prototype.getChildAt = function getChildAt (pos) {
      return BrowserDOMElement.wrap(this.el.childNodes[pos])
    };

    BrowserDOMElement.prototype.getChildIndex = function getChildIndex (child) {
      
      if (!child._isBrowserDOMElement) {
        throw new Error('Expecting a BrowserDOMElement instance.')
      }
      return Array.prototype.indexOf.call(this.el.childNodes, child.el)
    };

    BrowserDOMElement.prototype.getFirstChild = function getFirstChild () {
      var firstChild = this.el.firstChild;
      
      if (firstChild) {
        return BrowserDOMElement.wrap(firstChild)
      } else {
        return null
      }
    };

    BrowserDOMElement.prototype.getLastChild = function getLastChild () {
      var lastChild = this.el.lastChild;
      
      if (lastChild) {
        return BrowserDOMElement.wrap(lastChild)
      } else {
        return null
      }
    };

    BrowserDOMElement.prototype.getNextSibling = function getNextSibling () {
      var next = this.el.nextSibling;
      
      if (next) {
        return BrowserDOMElement.wrap(next)
      } else {
        return null
      }
    };

    BrowserDOMElement.prototype.getPreviousSibling = function getPreviousSibling () {
      var previous = this.el.previousSibling;
      
      if (previous) {
        return BrowserDOMElement.wrap(previous)
      } else {
        return null
      }
    };

    BrowserDOMElement.prototype.clone = function clone (deep) {
      var clone = this.el.cloneNode(deep);
      return BrowserDOMElement.wrap(clone)
    };

    BrowserDOMElement.prototype.createDocument = function createDocument (format) {
      return BrowserDOMElement.createDocument(format)
    };

    BrowserDOMElement.prototype.createElement = function createElement (tagName) {
      var doc = this._getNativeOwnerDocument();
      var el = doc.createElement(tagName);
      return BrowserDOMElement.wrap(el)
    };

    BrowserDOMElement.prototype.createTextNode = function createTextNode (text) {
      var doc = this._getNativeOwnerDocument();
      var el = doc.createTextNode(text);
      return BrowserDOMElement.wrap(el)
    };

    BrowserDOMElement.prototype.createComment = function createComment (data) {
      var doc = this._getNativeOwnerDocument();
      var el = doc.createComment(data);
      return BrowserDOMElement.wrap(el)
    };

    BrowserDOMElement.prototype.createProcessingInstruction = function createProcessingInstruction (name, data) {
      var doc = this._getNativeOwnerDocument();
      var el = doc.createProcessingInstruction(name, data);
      return BrowserDOMElement.wrap(el)
    };

    BrowserDOMElement.prototype.createCDATASection = function createCDATASection (data) {
      var doc = this._getNativeOwnerDocument();
      var el = doc.createCDATASection(data);
      return BrowserDOMElement.wrap(el)
    };

    BrowserDOMElement.prototype.is = function is (cssSelector) {
      
      
      var el = this.el;
      
      if (this.isElementNode()) {
        return matches(el, cssSelector)
      } else {
        return false
      }
    };

    BrowserDOMElement.prototype.getParent = function getParent () {
      var parent = this.el.parentNode;
      
      if (parent) {
        return BrowserDOMElement.wrap(parent)
      } else {
        return null
      }
    };

    BrowserDOMElement.prototype.getOwnerDocument = function getOwnerDocument () {
      return BrowserDOMElement.wrap(this._getNativeOwnerDocument())
    };

    prototypeAccessors.ownerDocument.get = function () {
      return this.getOwnerDocument()
    };

    BrowserDOMElement.prototype._getNativeOwnerDocument = function _getNativeOwnerDocument () {
      return (this.isDocumentNode() ? this.el : this.el.ownerDocument)
    };

    BrowserDOMElement.prototype.find = function find (cssSelector) {
      var result = null;
      if (this.el.querySelector) {
        result = this.el.querySelector(cssSelector);
      }
      if (result) {
        return BrowserDOMElement.wrap(result)
      } else {
        return null
      }
    };

    BrowserDOMElement.prototype.findAll = function findAll (cssSelector) {
      var result = [];
      if (this.el.querySelectorAll) {
        result = this.el.querySelectorAll(cssSelector);
      }
      return Array.prototype.map.call(result, function(el) {
        return BrowserDOMElement.wrap(el)
      })
    };

    BrowserDOMElement.prototype._normalizeChild = function _normalizeChild (child) {
      if (isNil(child)) { return child }

      if (child instanceof window.Node) {
        child = BrowserDOMElement.wrap(child);
      }
      
      
      
      else if (child._isBrowserDOMElement && ! (child instanceof BrowserDOMElement)) {
        child = BrowserDOMElement.wrap(child);
      } else if (isString(child) || isNumber(child)) {
        child = this.createTextNode(child);
      }
      
      if (!child || !child._isBrowserDOMElement) {
        throw new Error('Illegal child type.')
      }
      console.assert(_unwrap(child.el) === child, "The backlink to the wrapper should be consistent");
      return child.getNativeElement()
    };

    BrowserDOMElement.prototype.appendChild = function appendChild (child) {
      var nativeChild = this._normalizeChild(child);
      if (nativeChild) {
        this.el.appendChild(nativeChild);
      }
      return this
    };

    BrowserDOMElement.prototype.insertAt = function insertAt (pos, child) {
      var nativeChild = this._normalizeChild(child);
      var childNodes = this.el.childNodes;
      if (pos >= childNodes.length) {
        this.el.appendChild(nativeChild);
      } else {
        this.el.insertBefore(nativeChild, childNodes[pos]);
      }
      return this
    };

    BrowserDOMElement.prototype.insertBefore = function insertBefore (child, before) {
      
      if (isNil(before)) {
        return this.appendChild(child)
      }
      if (!before._isBrowserDOMElement) {
        throw new Error('insertBefore(): Illegal arguments. "before" must be a BrowserDOMElement instance.')
      }
      var nativeChild = this._normalizeChild(child);
      if (nativeChild) {
        this.el.insertBefore(nativeChild, before.el);
      }
      return this
    };

    BrowserDOMElement.prototype.removeAt = function removeAt (pos) {
      this.el.removeChild(this.el.childNodes[pos]);
      return this;
    };

    BrowserDOMElement.prototype.removeChild = function removeChild (child) {
      
      if (!child || !child._isBrowserDOMElement) {
        throw new Error('removeChild(): Illegal arguments. Expecting a BrowserDOMElement instance.')
      }
      this.el.removeChild(child.el);
      return this
    };

    BrowserDOMElement.prototype.replaceChild = function replaceChild (oldChild, newChild) {
      
      if (!newChild || !oldChild ||
          !newChild._isBrowserDOMElement || !oldChild._isBrowserDOMElement) {
        throw new Error('replaceChild(): Illegal arguments. Expecting BrowserDOMElement instances.')
      }
      
      this.el.replaceChild(newChild.el, oldChild.el);
      return this
    };

    BrowserDOMElement.prototype.empty = function empty () {
      var el = this.el;
      while (el.lastChild) {
        el.removeChild(el.lastChild);
      }
      return this
    };

    BrowserDOMElement.prototype.remove = function remove () {
      if (this.el.parentNode) {
        this.el.parentNode.removeChild(this.el);
      }
      return this
    };

    BrowserDOMElement.prototype.serialize = function serialize () {
      var outerHTML = this.el.outerHTML;
      if (isString(outerHTML)) {
        return outerHTML
      } else {
        var xs = new window.XMLSerializer();
        return xs.serializeToString(this.el)
      }
    };

    BrowserDOMElement.prototype.isInDocument = function isInDocument () {
      var el = this.el;
      while(el) {
        if (el.nodeType === window.Node.DOCUMENT_NODE) {
          return true
        }
        el = el.parentNode;
      }
    };

    BrowserDOMElement.prototype._replaceNativeEl = function _replaceNativeEl (newEl) {
      console.assert(newEl instanceof window.Node, "Expecting a native element.");
      var oldEl = this.el;
      var parentNode = oldEl.parentNode;
      if (parentNode) {
        parentNode.replaceChild(newEl, oldEl);
      }
      this.el = newEl;
      _detach(oldEl);
      _attach(newEl, this);
    };

    BrowserDOMElement.prototype._getChildNodeCount = function _getChildNodeCount () {
      return this.el.childNodes.length
    };

    BrowserDOMElement.prototype.focus = function focus () {
      this.el.focus();
      return this
    };

    BrowserDOMElement.prototype.select = function select () {
      this.el.select();
      return this
    };

    BrowserDOMElement.prototype.blur = function blur () {
      this.el.blur();
      return this
    };

    BrowserDOMElement.prototype.click = function click () {
      this.el.click();
      return this
    };

    BrowserDOMElement.prototype.getWidth = function getWidth () {
      var rect = this.el.getClientRects()[0];
      if (rect) {
        return rect.width
      } else {
        return 0
      }
    };

    BrowserDOMElement.prototype.getHeight = function getHeight () {
      var rect = this.el.getClientRects()[0];
      if (rect) {
        return rect.height
      } else {
        return 0
      }
    };

    BrowserDOMElement.prototype.getOffset = function getOffset () {
      var rect = this.el.getBoundingClientRect();
      return {
        top: rect.top + document.body.scrollTop,
        left: rect.left + document.body.scrollLeft
      }
    };

    BrowserDOMElement.prototype.getPosition = function getPosition () {
      return {left: this.el.offsetLeft, top: this.el.offsetTop}
    };

    BrowserDOMElement.prototype.getOuterHeight = function getOuterHeight (withMargin) {
      var outerHeight = this.el.offsetHeight;
      if (withMargin) {
        var style = this.getComputedStyle();
        outerHeight += parseInt(style.marginTop, 10) + parseInt(style.marginBottom, 10);
      }
      return outerHeight
    };

    BrowserDOMElement.prototype.getContentType = function getContentType () {
      return this._getNativeOwnerDocument().contentType
    };

    BrowserDOMElement.prototype._isXML = function _isXML () {
      return this.getContentType() === 'application/xml'
    };

    BrowserDOMElement.prototype.emit = function emit (name, data) {
      var event;
      if (data) {
        event = new window.CustomEvent(name, {
          detail: data,
          bubbles: true,
          cancelable: true
        });
      } else {
        event = new window.Event(name, {
          bubbles: true,
          cancelable: true
        });
      }
      this.el.dispatchEvent(event);
    };

    Object.defineProperties( BrowserDOMElement.prototype, prototypeAccessors );

    return BrowserDOMElement;
  }(DOMElement));

  BrowserDOMElement.prototype._isBrowserDOMElement = true;


  BrowserDOMElement.createDocument = function(format) {
    var doc;
    if (format === 'xml') {
      
      doc = window.document.implementation.createDocument(null, 'dummy');
      
      doc.removeChild(doc.firstChild);
    } else {
      doc = (new window.DOMParser()).parseFromString(DOMElement.EMPTY_HTML, 'text/html');
    }
    return BrowserDOMElement.wrap(doc)
  };

  BrowserDOMElement.parseMarkup = function(str, format, options) {
    if ( options === void 0 ) options={};

    if (!str) {
      return BrowserDOMElement.createDocument(format)
    }
    if (options.snippet) {
      str = "<div id='__snippet__'>" + str + "</div>";
    }
    var doc;
    var parser = new window.DOMParser();
    if (format === 'html') {
      doc = BrowserDOMElement.wrap(
        _check(
          parser.parseFromString(str, 'text/html')
        )
      );
    } else if (format === 'xml') {
      doc = BrowserDOMElement.wrap(
        _check(
          parser.parseFromString(str, 'application/xml')
        )
      );
    }
    if (options.snippet) {
      var childNodes = doc.find('#__snippet__').childNodes;
      if (childNodes.length === 1) {
        return childNodes[0]
      } else {
        return childNodes
      }
    } else {
      return doc
    }

    function _check(doc) {
      if (doc) {
        var parserError = doc.querySelector('parsererror');
        if (parserError) {
          throw new Error("ParserError: " + parserError)
        }
      }
      return doc
    }
  };

  BrowserDOMElement.wrap =
  BrowserDOMElement.wrapNativeElement = function(el) {
    if (el) {
      var _el = _unwrap(el);
      if (_el) {
        return _el
      } else if (el instanceof window.Node) {
        return new BrowserDOMElement(el)
      } else if (el._isBrowserDOMElement) {
        return new BrowserDOMElement(el.getNativeElement())
      } else if (el === window) {
        return BrowserDOMElement.getBrowserWindow()
      }
    } else {
      return null
    }
  };

  BrowserDOMElement.unwrap = function(nativeEl) {
    return _unwrap(nativeEl)
  };


  var BrowserWindow = function BrowserWindow() {
    
    this.el = window;
    _attach(window, this);
  };

  BrowserWindow.prototype.on = BrowserDOMElement.prototype.on;
  BrowserWindow.prototype.off = BrowserDOMElement.prototype.off;
  BrowserWindow.prototype.addEventListener = BrowserDOMElement.prototype.addEventListener;
  BrowserWindow.prototype.removeEventListener = BrowserDOMElement.prototype.removeEventListener;
  BrowserWindow.prototype._createEventListener = BrowserDOMElement.prototype._createEventListener;
  BrowserWindow.prototype._addEventListenerNative = BrowserDOMElement.prototype._addEventListenerNative;
  BrowserWindow.prototype._removeEventListenerNative = BrowserDOMElement.prototype._removeEventListenerNative;

  BrowserWindow.prototype.getEventListeners = BrowserDOMElement.prototype.getEventListeners;

  BrowserDOMElement.getBrowserWindow = function() {
    if (window[SIGNATURE]) { return window[SIGNATURE] }
    return new BrowserWindow(window)
  };

  BrowserDOMElement.isReverse = function(anchorNode, anchorOffset, focusNode, focusOffset) {
    
    
    if (focusNode && anchorNode) {
      if (!BrowserDOMElement.isReverse._r1) {
        BrowserDOMElement.isReverse._r1 = window.document.createRange();
        BrowserDOMElement.isReverse._r2 = window.document.createRange();
      }
      var _r1 = BrowserDOMElement.isReverse._r1;
      var _r2 = BrowserDOMElement.isReverse._r2;
      _r1.setStart(anchorNode.getNativeElement(), anchorOffset);
      _r2.setStart(focusNode.getNativeElement(), focusOffset);
      var cmp = _r1.compareBoundaryPoints(window.Range.START_TO_START, _r2);
      if (cmp === 1) {
        return true
      }
    }
    return false
  };

  BrowserDOMElement.getWindowSelection = function() {
    var nativeSel = window.getSelection();
    var result = {
      anchorNode: BrowserDOMElement.wrap(nativeSel.anchorNode),
      anchorOffset: nativeSel.anchorOffset,
      focusNode: BrowserDOMElement.wrap(nativeSel.focusNode),
      focusOffset: nativeSel.focusOffset
    };
    return result
  };


  function matches(el, selector) {
    var elProto = window.Element.prototype;
    var _matches = (
      elProto.matches || elProto.matchesSelector ||
      elProto.msMatchesSelector || elProto.webkitMatchesSelector
    );
    return _matches.call(el, selector)
  }

  var AttributesMapAdapter = function AttributesMapAdapter(attributes) {
    this.attributes = attributes;
  };

  var prototypeAccessors$1$1 = { size: { configurable: true } };

  prototypeAccessors$1$1.size.get = function () {
    return this.attributes.length
  };

  AttributesMapAdapter.prototype.get = function get (name) {
    var item = this.attributes.getNamedItem(name);
    if (item) {
      return item.value
    }
  };

  AttributesMapAdapter.prototype.set = function set (name, value) {
    this.attributes.setNamedItem(name, value);
  };

  AttributesMapAdapter.prototype.forEach = function forEach (fn) {
      var this$1 = this;

    var S = this.size;
    for (var i = 0; i < S; i++) {
      var item = this$1.attributes.item(i);
      fn(item.value, item.name);
    }
  };

  AttributesMapAdapter.prototype.map = function map (fn) {
    var result = [];
    this.forEach(function (val, key){ result.push(fn(val, key)); });
    return result
  };

  AttributesMapAdapter.prototype.keys = function keys () {
    return this.map(function (val, key){ return key })
  };

  AttributesMapAdapter.prototype.values = function values () {
    return this.map(function (val){ return val })
  };

  AttributesMapAdapter.prototype.entries = function entries () {
    return this.map(function (val, key){ return [key, val] })
  };

  Object.defineProperties( AttributesMapAdapter.prototype, prototypeAccessors$1$1 );

  var BrowserChildNodeIterator = function BrowserChildNodeIterator(el) {
    this._next = el.firstChild;
    this._curr = null;
  };

  BrowserChildNodeIterator.prototype.hasNext = function hasNext () {
    return Boolean(this._next)
  };

  BrowserChildNodeIterator.prototype.next = function next () {
    var next = this._next;
    this._curr = next;
    this._next = next.nextSibling;
    return BrowserDOMElement.wrap(next)
  };

  BrowserChildNodeIterator.prototype.back = function back () {
    this._next = this._curr;
    this._curr = this._curr.previousSibling;
  };

  BrowserChildNodeIterator.prototype.peek = function peek () {
    return BrowserDOMElement.wrap(this._curr)
  };

  var inBrowser = platform.inBrowser

  var domelementtype = {
  	Text: "text", 
  	Directive: "directive", 
  	Comment: "comment", 
  	Script: "script", 
  	Style: "style", 
  	Tag: "tag", 
  	CDATA: "cdata", 
  	Doctype: "doctype",

  	isTag: function(elem){
  		return elem.type === "tag" || elem.type === "script" || elem.type === "style";
  	}
  };

  var boolbase = {
  	trueFunc: function trueFunc(){
  		return true;
  	},
  	falseFunc: function falseFunc(){
  		return false;
  	}
  };

  var parse_1 = parse;




  var re_nthElement = /^([+\-]?\d*n)?\s*(?:([+\-]?)\s*(\d+))?$/;


  function parse(formula){
  	formula = formula.trim().toLowerCase();

  	if(formula === "even"){
  		return [2, 0];
  	} else if(formula === "odd"){
  		return [2, 1];
  	} else {
  		var parsed = formula.match(re_nthElement);

  		if(!parsed){
  			throw new SyntaxError("n-th rule couldn't be parsed ('" + formula + "')");
  		}

  		var a;

  		if(parsed[1]){
  			a = parseInt(parsed[1], 10);
  			if(isNaN(a)){
  				if(parsed[1].charAt(0) === "-") { a = -1; }
  				else { a = 1; }
  			}
  		} else { a = 0; }

  		return [
  			a,
  			parsed[3] ? parseInt((parsed[2] || "") + parsed[3], 10) : 0
  		];
  	}
  }

  var compile_1 = compile;

  var trueFunc  = boolbase.trueFunc,
      falseFunc = boolbase.falseFunc;


  function compile(parsed){
  	var a = parsed[0],
  	    b = parsed[1] - 1;

  	
  	
  	if(b < 0 && a <= 0) { return falseFunc; }

  	
  	if(a ===-1) { return function(pos){ return pos <= b; }; }
  	if(a === 0) { return function(pos){ return pos === b; }; }
  	
  	if(a === 1) { return b < 0 ? trueFunc : function(pos){ return pos >= b; }; }

  	
  	var bMod = b % a;
  	if(bMod < 0) { bMod += a; }

  	if(a > 1){
  		return function(pos){
  			return pos >= b && pos % a === bMod;
  		};
  	}

  	a *= -1; 

  	return function(pos){
  		return pos <= b && pos % a === bMod;
  	};
  }

  var nthCheck = function nthCheck(formula){
  	return compile_1(parse_1(formula));
  };

  var parse_1$1 = parse_1;
  var compile_1$1 = compile_1;
  nthCheck.parse = parse_1$1;
  nthCheck.compile = compile_1$1;

  var cssWhat = parse$1;

  var re_name = /^(?:\\.|[\w\-\u00c0-\uFFFF])+/,
      re_escape = /\\([\da-f]{1,6}\s?|(\s)|.)/ig,
      
      re_attr = /^\s*((?:\\.|[\w\u00c0-\uFFFF\-])+)\s*(?:(\S?)=\s*(?:(['"])(.*?)\3|(#?(?:\\.|[\w\u00c0-\uFFFF\-])*)|)|)\s*(i)?\]/;

  var actionTypes = {
  	__proto__: null,
  	"undefined": "exists",
  	"":  "equals",
  	"~": "element",
  	"^": "start",
  	"$": "end",
  	"*": "any",
  	"!": "not",
  	"|": "hyphen"
  };

  var simpleSelectors = {
  	__proto__: null,
  	">": "child",
  	"<": "parent",
  	"~": "sibling",
  	"+": "adjacent"
  };

  var attribSelectors = {
  	__proto__: null,
  	"#": ["id", "equals"],
  	".": ["class", "element"]
  };


  var unpackPseudos = {
  	__proto__: null,
  	"has": true,
  	"not": true,
  	"matches": true
  };

  var stripQuotesFromPseudos = {
  	__proto__: null,
  	"contains": true,
  	"icontains": true
  };

  var quotes = {
  	__proto__: null,
  	"\"": true,
  	"'": true
  };


  function funescape( _, escaped, escapedWhitespace ) {
  	var high = "0x" + escaped - 0x10000;
  	
  	
  	
  	return high !== high || escapedWhitespace ?
  		escaped :
  		
  		high < 0 ?
  			String.fromCharCode( high + 0x10000 ) :
  			
  			String.fromCharCode( high >> 10 | 0xD800, high & 0x3FF | 0xDC00 );
  }

  function unescapeCSS(str){
  	return str.replace(re_escape, funescape);
  }

  function isWhitespace(c){
  	return c === " " || c === "\n" || c === "\t" || c === "\f" || c === "\r";
  }

  function parse$1(selector, options){
  	var subselects = [];

  	selector = parseSelector(subselects, selector + "", options);

  	if(selector !== ""){
  		throw new SyntaxError("Unmatched selector: " + selector);
  	}

  	return subselects;
  }

  function parseSelector(subselects, selector, options){
  	var tokens = [],
  		sawWS = false,
  		data, firstChar, name, quot;

  	function getName(){
  		var sub = selector.match(re_name)[0];
  		selector = selector.substr(sub.length);
  		return unescapeCSS(sub);
  	}

  	function stripWhitespace(start){
  		while(isWhitespace(selector.charAt(start))) { start++; }
  		selector = selector.substr(start);
  	}

  	stripWhitespace(0);

  	while(selector !== ""){
  		firstChar = selector.charAt(0);

  		if(isWhitespace(firstChar)){
  			sawWS = true;
  			stripWhitespace(1);
  		} else if(firstChar in simpleSelectors){
  			tokens.push({type: simpleSelectors[firstChar]});
  			sawWS = false;

  			stripWhitespace(1);
  		} else if(firstChar === ","){
  			if(tokens.length === 0){
  				throw new SyntaxError("empty sub-selector");
  			}
  			subselects.push(tokens);
  			tokens = [];
  			sawWS = false;
  			stripWhitespace(1);
  		} else {
  			if(sawWS){
  				if(tokens.length > 0){
  					tokens.push({type: "descendant"});
  				}
  				sawWS = false;
  			}

  			if(firstChar === "*"){
  				selector = selector.substr(1);
  				tokens.push({type: "universal"});
  			} else if(firstChar in attribSelectors){
  				selector = selector.substr(1);
  				tokens.push({
  					type: "attribute",
  					name: attribSelectors[firstChar][0],
  					action: attribSelectors[firstChar][1],
  					value: getName(),
  					ignoreCase: false
  				});
  			} else if(firstChar === "["){
  				selector = selector.substr(1);
  				data = selector.match(re_attr);
  				if(!data){
  					throw new SyntaxError("Malformed attribute selector: " + selector);
  				}
  				selector = selector.substr(data[0].length);
  				name = unescapeCSS(data[1]);

  				if(
  					!options || (
  						"lowerCaseAttributeNames" in options ?
  							options.lowerCaseAttributeNames :
  							!options.xmlMode
  					)
  				){
  					name = name.toLowerCase();
  				}

  				tokens.push({
  					type: "attribute",
  					name: name,
  					action: actionTypes[data[2]],
  					value: unescapeCSS(data[4] || data[5] || ""),
  					ignoreCase: !!data[6]
  				});

  			} else if(firstChar === ":"){
  				if(selector.charAt(1) === ":"){
  					selector = selector.substr(2);
  					tokens.push({type: "pseudo-element", name: getName().toLowerCase()});
  					continue;
  				}

  				selector = selector.substr(1);

  				name = getName().toLowerCase();
  				data = null;

  				if(selector.charAt(0) === "("){
  					if(name in unpackPseudos){
  						quot = selector.charAt(1);
  						var quoted = quot in quotes;

  						selector = selector.substr(quoted + 1);

  						data = [];
  						selector = parseSelector(data, selector, options);

  						if(quoted){
  							if(selector.charAt(0) !== quot){
  								throw new SyntaxError("unmatched quotes in :" + name);
  							} else {
  								selector = selector.substr(1);
  							}
  						}

  						if(selector.charAt(0) !== ")"){
  							throw new SyntaxError("missing closing parenthesis in :" + name + " " + selector);
  						}

  						selector = selector.substr(1);
  					} else {
  						var pos = 1, counter = 1;

  						for(; counter > 0 && pos < selector.length; pos++){
  							if(selector.charAt(pos) === "(") { counter++; }
  							else if(selector.charAt(pos) === ")") { counter--; }
  						}

  						if(counter){
  							throw new SyntaxError("parenthesis not matched");
  						}

  						data = selector.substr(1, pos - 2);
  						selector = selector.substr(pos);

  						if(name in stripQuotesFromPseudos){
  							quot = data.charAt(0);

  							if(quot === data.slice(-1) && quot in quotes){
  								data = data.slice(1, -1);
  							}

  							data = unescapeCSS(data);
  						}
  					}
  				}

  				tokens.push({type: "pseudo", name: name, data: data});
  			} else if(re_name.test(selector)){
  				name = getName();

  				if(!options || ("lowerCaseTags" in options ? options.lowerCaseTags : !options.xmlMode)){
  					name = name.toLowerCase();
  				}

  				tokens.push({type: "tag", name: name});
  			} else {
  				if(tokens.length && tokens[tokens.length - 1].type === "descendant"){
  					tokens.pop();
  				}
  				addToken(subselects, tokens);
  				return selector;
  			}
  		}
  	}

  	addToken(subselects, tokens);

  	return selector;
  }

  function addToken(subselects, tokens){
  	if(subselects.length > 0 && tokens.length === 0){
  		throw new SyntaxError("empty sub-selector");
  	}

  	subselects.push(tokens);
  }

  var _encodeXMLContent = (function (obj) {
    var invObj = getInverseObj(obj);
    var replacer = getInverseReplacer(invObj);
    return getInverse(invObj, replacer)
  })({
    amp: "&",
    gt: ">",
    lt: "<",
  });

  var _encodeXMLAttr = (function (obj) {
    var invObj = getInverseObj(obj);
    var replacer = getInverseReplacer(invObj);
    return getInverse(invObj, replacer)
  })({
    quot: "\"",
  });

  function getInverseObj(obj){
    return Object.keys(obj).sort().reduce(function(inverse, name){
      inverse[obj[name]] = "&" + name + ";";
      return inverse;
    }, {});
  }

  function getInverseReplacer(inverse){
    var single = [],
      multiple = [];

    Object.keys(inverse).forEach(function(k){
      if(k.length === 1){
        single.push("\\" + k);
      } else {
        multiple.push(k);
      }
    });


    multiple.unshift("[" + single.join("") + "]");

    return new RegExp(multiple.join("|"), "g");
  }

  var re_nonASCII = /[^\0-\x7F]/g;
  var re_astralSymbols = /[\uD800-\uDBFF][\uDC00-\uDFFF]/g;

  function singleCharReplacer(c){
    return "&#x" + c.charCodeAt(0).toString(16).toUpperCase() + ";";
  }

  function astralReplacer(c){

    var high = c.charCodeAt(0);
    var low = c.charCodeAt(1);
    var codePoint = (high - 0xD800) * 0x400 + low - 0xDC00 + 0x10000;
    return "&#x" + codePoint.toString(16).toUpperCase() + ";";
  }

  function getInverse(inverse, re){
    function func(name){
      return inverse[name];
    }
    return function(data){
      return data
        .replace(re, func)
        .replace(re_astralSymbols, astralReplacer)
        .replace(re_nonASCII, singleCharReplacer);
    };
  }


  var booleanAttributes = {
    __proto__: null,
    allowfullscreen: true,
    async: true,
    autofocus: true,
    autoplay: true,
    checked: true,
    controls: true,
    default: true,
    defer: true,
    disabled: true,
    hidden: true,
    ismap: true,
    loop: true,
    multiple: true,
    muted: true,
    open: true,
    readonly: true,
    required: true,
    reversed: true,
    scoped: true,
    seamless: true,
    selected: true,
    typemustmatch: true
  };

  var unencodedElements = {
    __proto__: null,
    style: true,
    script: true,
    xmp: true,
    iframe: true,
    noembed: true,
    noframes: true,
    plaintext: true,
    noscript: true
  };

  var singleTag = {
    __proto__: null,
    area: true,
    base: true,
    basefont: true,
    br: true,
    col: true,
    command: true,
    embed: true,
    frame: true,
    hr: true,
    img: true,
    input: true,
    isindex: true,
    keygen: true,
    link: true,
    meta: true,
    param: true,
    source: true,
    track: true,
    wbr: true,
  };


  var DomUtils = function DomUtils () {};

  DomUtils.prototype.isTag = function isTag (elem) {
    return domelementtype.isTag(elem)
  };

  DomUtils.prototype.removeElement = function removeElement (elem){
    if(elem.prev) { elem.prev.next = elem.next; }
    if(elem.next) { elem.next.prev = elem.prev; }
    if(elem.parent){
      var childs = elem.parent.childNodes;
      var pos = childs.lastIndexOf(elem);
      if (pos < 0) { throw new Error('Invalid state') }
      childs.splice(pos, 1);
      elem.parent = null;
    }
  };

  DomUtils.prototype.replaceElement = function replaceElement (elem, replacement){
    if (replacement.parent) { this.removeElement(replacement); }
    var prev = replacement.prev = elem.prev;
    if(prev){
      prev.next = replacement;
    }

    var next = replacement.next = elem.next;
    if(next){
      next.prev = replacement;
    }

    var parent = replacement.parent = elem.parent;
    if(parent){
      var childs = parent.childNodes;
      var pos = childs.lastIndexOf(elem);
      if (pos < 0) { throw new Error('Invalid state') }
      childs[pos] = replacement;
    }
  };

  DomUtils.prototype.appendChild = function appendChild (elem, child){
    if (child.parent) { this.removeElement(child); }
    child.parent = elem;

    if(elem.childNodes.push(child) !== 1){
      var sibling = elem.childNodes[elem.childNodes.length - 2];
      sibling.next = child;
      child.prev = sibling;
      child.next = null;
    }
  };

  DomUtils.prototype.append = function append (elem, next){
    if (next.parent) { this.removeElement(next); }
    var parent = elem.parent,
      currNext = elem.next;

    next.next = currNext;
    next.prev = elem;
    elem.next = next;
    next.parent = parent;

    if(currNext){
      currNext.prev = next;
      if(parent){
        var childs = parent.childNodes;
        var pos = childs.lastIndexOf(currNext);
        if (pos < 0) { throw new Error('Invalid state') }
        childs.splice(pos, 0, next);
      }
    } else if(parent){
      parent.childNodes.push(next);
    }
  };

  DomUtils.prototype.prepend = function prepend (elem, prev){
    if (prev.parent) { this.removeElement(prev); }
    var parent = elem.parent;
    if(parent){
      var childs = parent.childNodes;
      var pos = childs.lastIndexOf(elem);
      if (pos < 0) { throw new Error('Invalid state') }
      childs.splice(pos, 0, prev);
    }

    if(elem.prev){
      elem.prev.next = prev;
    }

    prev.parent = parent;
    prev.prev = elem.prev;
    prev.next = elem;
    elem.prev = prev;
  };


  DomUtils.prototype.filter = function filter (test, element, recurse, limit){
    if(!Array.isArray(element)) { element = [element]; }

    if(typeof limit !== "number" || !isFinite(limit)){
      limit = Infinity;
    }
    return this.find(test, element, recurse !== false, limit);
  };

  DomUtils.prototype.find = function find (test, elems, recurse, limit){
      var this$1 = this;

    var result = [], childs;

    for(var i = 0, j = elems.length; i < j; i++){
      if(test(elems[i])){
        result.push(elems[i]);
        if(--limit <= 0) { break; }
      }

      childs = this$1.getChildren(elems[i]);
      if(recurse && childs && childs.length > 0){
        childs = this$1.find(test, childs, recurse, limit);
        result = result.concat(childs);
        limit -= childs.length;
        if(limit <= 0) { break; }
      }
    }

    return result;
  };

  DomUtils.prototype.findOneChild = function findOneChild (test, elems){
    for(var i = 0, l = elems.length; i < l; i++){
      if(test(elems[i])) { return elems[i]; }
    }

    return null;
  };

  DomUtils.prototype.findOne = function findOne (test, elems){
      var this$1 = this;

    var elem = null;

    for(var i = 0, l = elems.length; i < l && !elem; i++){
      var child = elems[i];
      if(!this$1.isTag(child)){
        continue;
      } else if(test(child)){
        elem = child;
      } else {
        var childNodes = this$1.getChildren(child);
        if (childNodes.length > 0) {
          elem = this$1.findOne(test, childNodes);
        }
      }
    }

    return elem;
  };

  DomUtils.prototype.existsOne = function existsOne (test, elems){
      var this$1 = this;

    for(var i = 0, l = elems.length; i < l; i++){
      var elem = elems[i];
      
      if (!this$1.isTag(elem)) { continue }
      
      if (test(elem)) { return true }
      
      var childNodes = this$1.getChildren(elem);
      if (childNodes.length > 0 && this$1.existsOne(test, childNodes)) { return true }
    }
    return false;
  };

  DomUtils.prototype.findAll = function findAll (test, elems){
      var this$1 = this;

    var result = [];
    for(var i = 0, j = elems.length; i < j; i++){
      var elem = elems[i];
      if(!this$1.isTag(elem)) { continue; }
      if(test(elem)) { result.push(elem); }
      var childNodes = this$1.getChildren(elem);
      if(childNodes.length > 0){
        result = result.concat(this$1.findAll(test, childNodes));
      }
    }
    return result;
  };

  DomUtils.prototype.getAttributes = function getAttributes (el) {
    var attribs = el.getAttributes();
    
    
    if (attribs instanceof Map) {
      return Array.from(attribs)
    } else if (attribs && attribs.forEach) {
      var res = [];
      attribs.forEach(function (val, key) {
        res.push([key, val]);
      });
      return res
    } else {
      return []
    }
  };

  DomUtils.prototype.formatAttribs = function formatAttribs (el, opts) {
      if ( opts === void 0 ) opts = {};

    var output = [];
    var attributes = this.getAttributes(el);
    attributes.forEach(function (ref) {
        var key = ref[0];
        var value = ref[1];

      if (!value && booleanAttributes[key]) {
        output.push(key);
      } else {
        output.push(key + '="' + (opts.decodeEntities ? _encodeXMLAttr(value) : value) + '"');
      }
    });
    return output.join(' ')
  };

  DomUtils.prototype.render = function render (dom, opts) {
      var this$1 = this;

    if (!Array.isArray(dom)) { dom = [dom]; }
    opts = opts || {};
    var output = [];
    for(var i = 0; i < dom.length; i++){
      var elem = dom[i];
      switch (elem.type) {
        case 'root':
        case 'document': {
          output.push(this$1.render(this$1.getChildren(elem), opts));
          break
        }
        case domelementtype.Tag:
        case domelementtype.Script:
        case domelementtype.Style: {
          output.push(this$1.renderTag(elem, opts));
          break
        }
        case domelementtype.CDATA: {
          output.push(this$1.renderCdata(elem));
          break
        }
        case domelementtype.Comment: {
          output.push(this$1.renderComment(elem));
          break
        }
        case domelementtype.Directive: {
          output.push(this$1.renderDirective(elem));
          break
        }
        case domelementtype.Doctype: {
          output.push(this$1.renderDoctype(elem));
          break
        }
        case domelementtype.Text: {
          output.push(this$1.renderText(elem, opts));
          break
        }
        default:
          throw new Error('Not implement yet: render of element type ' + elem.type)
      }
    }
    return output.join('')
  };

  DomUtils.prototype.renderTag = function renderTag (elem, opts) {
    var name = this.getName(elem);
    if (name === "svg") { opts = {decodeEntities: opts.decodeEntities, xmlMode: true}; }
    var tag = '<' + name;
    var attribs = this.formatAttribs(elem, opts);
    if (attribs) {
      tag += ' ' + attribs;
    }
    var childNodes = this.getChildren(elem);
    if (opts.xmlMode && childNodes.length === 0) {
      tag += '/>';
    } else {
      tag += '>';
      if (childNodes.length > 0) {
        tag += this.render(childNodes, opts);
      }
      if (!singleTag[name] || opts.xmlMode) {
        tag += '</' + name + '>';
      }
    }
    return tag
  };

  DomUtils.prototype.renderDirective = function renderDirective (elem) {
    return '<?' + this.getData(elem) + '?>'
  };

  DomUtils.prototype.renderDoctype = function renderDoctype (elem) {
    var ref = this.getData(elem);
      var name = ref.name;
      var publicId = ref.publicId;
      var systemId = ref.systemId;
    var frags = ['DOCTYPE', name];
    if (publicId) {
      frags.push('PUBLIC');
      frags.push('"'+publicId+'"');
      if (systemId) { frags.push('"'+systemId+'"'); }
    }
    return '<!' + frags.join(" ") + '>'
  };

  DomUtils.prototype.renderText = function renderText (elem, opts) {
    var text = this.getText(elem);
    if (opts.decodeEntities) {
      var parent = this.getParent(elem);
      if (!(parent && this.getName(parent) in unencodedElements)) {
        text = _encodeXMLContent(text);
      }
    }
    return text
  };

  DomUtils.prototype.renderCdata = function renderCdata (elem) {
    return '<![CDATA[' + this.getData(elem) + ']]>'
  };

  DomUtils.prototype.renderComment = function renderComment (elem) {
    return '<!--' + this.getData(elem) + '-->'
  };

  DomUtils.prototype.getInnerHTML = function getInnerHTML (elem, opts){
      var this$1 = this;

    var childNodes = this.getChildren(elem);
    return childNodes.map(function (child) {
      return this$1.render(child, opts);
    }).join("")
  };

  DomUtils.prototype.getOuterHTML = function getOuterHTML (elem, opts) {
    return this.render(elem, opts)
  };

  DomUtils.prototype.getData = function getData (elem) {
    return elem.data
  };

  DomUtils.prototype.getText = function getText (elem, sub){
      var this$1 = this;

    if(Array.isArray(elem)) { return elem.map(function (e) { return this$1.getText(e, sub); }).join(""); }
    switch(elem.type) {
      case domelementtype.Tag:
      case domelementtype.Script:
      case domelementtype.Style:
        return this.getText(this.getChildren(elem), true)
      case domelementtype.Text:
      case domelementtype.CDATA:
        return elem.data
      case domelementtype.Comment:
        
        
        if (sub) {
          return ""
        }
        return elem.data
      default:
        return ""
    }
  };

  DomUtils.prototype.getChildren = function getChildren (elem) {
    return elem.childNodes;
  };

  DomUtils.prototype.getParent = function getParent (elem){
    return elem.parent;
  };

  DomUtils.prototype.getSiblings = function getSiblings (elem){
    var parent = this.getParent(elem);
    return parent ? this.getChildren(parent) : [elem];
  };

  DomUtils.prototype.getAttributeValue = function getAttributeValue (elem, name){
    return elem.getAttribute(name);
  };

  DomUtils.prototype.hasAttrib = function hasAttrib (elem, name){
    return elem.hasAttribute(name);
  };

  DomUtils.prototype.getName = function getName (elem){
    return elem.name
  };

  DomUtils.prototype.getNameWithoutNS = function getNameWithoutNS (elem){
    return elem.nameWithoutNS
  };

  var domUtils = new DomUtils();
  domUtils.DomUtils = DomUtils;

  var universal = 50;
  var tag = 30;
  var attribute = 1;
  var pseudo = 0;
  var descendant = -1;
  var child = -1;
  var parent$1 = -1;
  var sibling = -1;
  var adjacent = -1;
  var procedure = {
  	universal: universal,
  	tag: tag,
  	attribute: attribute,
  	pseudo: pseudo,
  	descendant: descendant,
  	child: child,
  	parent: parent$1,
  	sibling: sibling,
  	adjacent: adjacent
  };

  var procedure$1 = Object.freeze({
    universal: universal,
    tag: tag,
    attribute: attribute,
    pseudo: pseudo,
    descendant: descendant,
    child: child,
    parent: parent$1,
    sibling: sibling,
    adjacent: adjacent,
    default: procedure
  });

  var procedure$2 = ( procedure$1 && procedure ) || procedure$1;

  var sort = sortByProcedure;





  var attributes = {
  	__proto__: null,
  	exists: 10,
  	equals: 8,
  	not: 7,
  	start: 6,
  	end: 6,
  	any: 5,
  	hyphen: 4,
  	element: 4
  };

  function sortByProcedure(arr){
  	var procs = arr.map(getProcedure);
  	for(var i = 1; i < arr.length; i++){
  		var procNew = procs[i];

  		if(procNew < 0) { continue; }

  		for(var j = i - 1; j >= 0 && procNew < procs[j]; j--){
  			var token = arr[j + 1];
  			arr[j + 1] = arr[j];
  			arr[j] = token;
  			procs[j + 1] = procs[j];
  			procs[j] = procNew;
  		}
  	}
  }

  function getProcedure(token){
  	var proc = procedure$2[token.type];

  	if(proc === procedure$2.attribute){
  		proc = attributes[token.action];

  		if(proc === attributes.equals && token.name === "id"){
  			
  			proc = 9;
  		}

  		if(token.ignoreCase){
  			
  			
  			proc >>= 1;
  		}
  	} else if(proc === procedure$2.pseudo){
  		if(!token.data){
  			proc = 3;
  		} else if(token.name === "has" || token.name === "contains"){
  			proc = 0; 
  		} else if(token.name === "matches" || token.name === "not"){
  			proc = 0;
  			for(var i = 0; i < token.data.length; i++){
  				
  				if(token.data[i].length !== 1) { continue; }
  				var cur = getProcedure(token.data[i][0]);
  				
  				if(cur === 0){
  					proc = 0;
  					break;
  				}
  				if(cur > proc) { proc = cur; }
  			}
  			if(token.data.length > 1 && proc > 0) { proc -= 1; }
  		} else {
  			proc = 1;
  		}
  	}
  	return proc;
  }

  var falseFunc$1 = boolbase.falseFunc;


  var reChars = /[-[\]{}()*+?.,\\^$|#\s]/g;

  function factory(adapter){
  	
  	var attributeRules = {
  		__proto__: null,
  		equals: function(next, data){
  			var name  = data.name,
  				value = data.value;

  			if(data.ignoreCase){
  				value = value.toLowerCase();

  				return function equalsIC(elem){
  					var attr = adapter.getAttributeValue(elem, name);
  					return attr != null && attr.toLowerCase() === value && next(elem);
  				};
  			}

  			return function equals(elem){
  				return adapter.getAttributeValue(elem, name) === value && next(elem);
  			};
  		},
  		hyphen: function(next, data){
  			var name  = data.name,
  				value = data.value,
  				len = value.length;

  			if(data.ignoreCase){
  				value = value.toLowerCase();

  				return function hyphenIC(elem){
  					var attr = adapter.getAttributeValue(elem, name);
  					return attr != null &&
  							(attr.length === len || attr.charAt(len) === "-") &&
  							attr.substr(0, len).toLowerCase() === value &&
  							next(elem);
  				};
  			}

  			return function hyphen(elem){
  				var attr = adapter.getAttributeValue(elem, name);
  				return attr != null &&
  						attr.substr(0, len) === value &&
  						(attr.length === len || attr.charAt(len) === "-") &&
  						next(elem);
  			};
  		},
  		element: function(next, data){
  			var name = data.name,
  				value = data.value;
  			if (data.name === 'class') {
  				var value$1 = data.value;
  				if (/\s/.test(value$1)) { return function() { return false } }
  				return function clazz(elem) {
  					var classes = elem.classes;
  					return classes && classes.has(value$1) && next(elem)
  				}
  			} else {
  				if(/\s/.test(value)){
  					return falseFunc$1;
  				}

  				value = value.replace(reChars, "\\$&");

  				var pattern = "(?:^|\\s)" + value + "(?:$|\\s)",
  					flags = data.ignoreCase ? "i" : "",
  					regex = new RegExp(pattern, flags);

  				return function element(elem){
  					var attr = adapter.getAttributeValue(elem, name);
  					return attr != null && regex.test(attr) && next(elem);
  				};
  			}
  		},
  		exists: function(next, data){
  			var name = data.name;
  			return function exists(elem){
  				return adapter.hasAttrib(elem, name) && next(elem);
  			};
  		},
  		start: function(next, data){
  			var name  = data.name,
  				value = data.value,
  				len = value.length;

  			if(len === 0){
  				return falseFunc$1;
  			}

  			if(data.ignoreCase){
  				value = value.toLowerCase();

  				return function startIC(elem){
  					var attr = adapter.getAttributeValue(elem, name);
  					return attr != null && attr.substr(0, len).toLowerCase() === value && next(elem);
  				};
  			}

  			return function start(elem){
  				var attr = adapter.getAttributeValue(elem, name);
  				return attr != null && attr.substr(0, len) === value && next(elem);
  			};
  		},
  		end: function(next, data){
  			var name  = data.name,
  				value = data.value,
  				len   = -value.length;

  			if(len === 0){
  				return falseFunc$1;
  			}

  			if(data.ignoreCase){
  				value = value.toLowerCase();

  				return function endIC(elem){
  					var attr = adapter.getAttributeValue(elem, name);
  					return attr != null && attr.substr(len).toLowerCase() === value && next(elem);
  				};
  			}

  			return function end(elem){
  				var attr = adapter.getAttributeValue(elem, name);
  				return attr != null && attr.substr(len) === value && next(elem);
  			};
  		},
  		any: function(next, data){
  			var name  = data.name,
  				value = data.value;

  			if(value === ""){
  				return falseFunc$1;
  			}

  			if(data.ignoreCase){
  				var regex = new RegExp(value.replace(reChars, "\\$&"), "i");

  				return function anyIC(elem){
  					var attr = adapter.getAttributeValue(elem, name);
  					return attr != null && regex.test(attr) && next(elem);
  				};
  			}

  			return function any(elem){
  				var attr = adapter.getAttributeValue(elem, name);
  				return attr != null && attr.indexOf(value) >= 0 && next(elem);
  			};
  		},
  		not: function(next, data){
  			var name  = data.name,
  				value = data.value;

  			if(value === ""){
  				return function notEmpty(elem){
  					return !!adapter.getAttributeValue(elem, name) && next(elem);
  				};
  			} else if(data.ignoreCase){
  				value = value.toLowerCase();

  				return function notIC(elem){
  					var attr = adapter.getAttributeValue(elem, name);
  					return attr != null && attr.toLowerCase() !== value && next(elem);
  				};
  			}

  			return function not(elem){
  				return adapter.getAttributeValue(elem, name) !== value && next(elem);
  			};
  		}
  	};

  	return {
  		compile: function(next, data, options){
  			if(options && options.strict && (
  				data.ignoreCase || data.action === "not"
  			)) { throw new Error("Unsupported attribute selector"); }
  			return attributeRules[data.action](next, data);
  		},
  		rules: attributeRules
  	};
  }

  var attributes$1 = factory;

  function generalFactory(adapter, Pseudos){
  	
  	return {
  		__proto__: null,

  		attribute: attributes$1(adapter).compile,
  		pseudo: Pseudos.compile,

  		
  		tag: function(next, data){
  			var name = data.name;
  			return function tag(elem){
  				return adapter.getNameWithoutNS(elem) === name && next(elem);
  			}
  		},

  		
  		descendant: function(next){
  			return function descendant(elem){

  				var found = false;

  				while(!found && (elem = adapter.getParent(elem))){
  					found = next(elem);
  				}

  				return found;
  			};
  		},
  		_flexibleDescendant: function(next){
  			
  			return function descendant(elem){

  				var found = next(elem);

  				while(!found && (elem = adapter.getParent(elem))){
  					found = next(elem);
  				}

  				return found;
  			};
  		},
  		parent: function(next, data, options){
  			if(options && options.strict) { throw new Error("Parent selector isn't part of CSS3"); }

  			return function parent(elem){
  				return adapter.getChildren(elem).some(test);
  			};

  			function test(elem){
  				return adapter.isTag(elem) && next(elem);
  			}
  		},
  		child: function(next){
  			return function child(elem){
  				var parent = adapter.getParent(elem);
  				return !!parent && next(parent);
  			};
  		},
  		sibling: function(next){
  			return function sibling(elem){
  				var siblings = adapter.getSiblings(elem);

  				for(var i = 0; i < siblings.length; i++){
  					if(adapter.isTag(siblings[i])){
  						if(siblings[i] === elem) { break; }
  						if(next(siblings[i])) { return true; }
  					}
  				}

  				return false;
  			};
  		},
  		adjacent: function(next){
  			return function adjacent(elem){
  				var siblings = adapter.getSiblings(elem),
  					lastElement;

  				for(var i = 0; i < siblings.length; i++){
  					if(adapter.isTag(siblings[i])){
  						if(siblings[i] === elem) { break; }
  						lastElement = siblings[i];
  					}
  				}

  				return !!lastElement && next(lastElement);
  			};
  		},
  		universal: function(next){
  			return next;
  		}
  	};
  }

  var general = generalFactory;

  var trueFunc$1          = boolbase.trueFunc,
  	falseFunc$1$1         = boolbase.falseFunc;

  function filtersFactory(adapter){
  	var attributes  = attributes$1(adapter),
  		checkAttrib = attributes.rules.equals;

  	
  	function equals(a, b){
  		if(typeof adapter.equals === "function") { return adapter.equals(a, b); }

  		return a === b;
  	}

  	function getAttribFunc(name, value){
  		var data = {name: name, value: value};
  		return function attribFunc(next){
  			return checkAttrib(next, data);
  		};
  	}

  	function getChildFunc(next){
  		return function(elem){
  			return !!adapter.getParent(elem) && next(elem);
  		};
  	}

  	var filters = {
  		contains: function(next, text){
  			return function contains(elem){
  				return next(elem) && adapter.getText(elem).indexOf(text) >= 0;
  			};
  		},
  		icontains: function(next, text){
  			var itext = text.toLowerCase();
  			return function icontains(elem){
  				return next(elem) &&
  					adapter.getText(elem).toLowerCase().indexOf(itext) >= 0;
  			};
  		},

  		
  		"nth-child": function(next, rule){
  			var func = nthCheck(rule);

  			if(func === falseFunc$1$1) { return func; }
  			if(func === trueFunc$1)  { return getChildFunc(next); }

  			return function nthChild(elem){
  				var siblings = adapter.getSiblings(elem);

  				for(var i = 0, pos = 0; i < siblings.length; i++){
  					if(adapter.isTag(siblings[i])){
  						if(siblings[i] === elem) { break; }
  						else { pos++; }
  					}
  				}

  				return func(pos) && next(elem);
  			};
  		},
  		"nth-last-child": function(next, rule){
  			var func = nthCheck(rule);

  			if(func === falseFunc$1$1) { return func; }
  			if(func === trueFunc$1)  { return getChildFunc(next); }

  			return function nthLastChild(elem){
  				var siblings = adapter.getSiblings(elem);

  				for(var pos = 0, i = siblings.length - 1; i >= 0; i--){
  					if(adapter.isTag(siblings[i])){
  						if(siblings[i] === elem) { break; }
  						else { pos++; }
  					}
  				}

  				return func(pos) && next(elem);
  			};
  		},
  		"nth-of-type": function(next, rule){
  			var func = nthCheck(rule);

  			if(func === falseFunc$1$1) { return func; }
  			if(func === trueFunc$1)  { return getChildFunc(next); }

  			return function nthOfType(elem){
  				var siblings = adapter.getSiblings(elem);

  				for(var pos = 0, i = 0; i < siblings.length; i++){
  					if(adapter.isTag(siblings[i])){
  						if(siblings[i] === elem) { break; }
  						if(adapter.getName(siblings[i]) === adapter.getName(elem)) { pos++; }
  					}
  				}

  				return func(pos) && next(elem);
  			};
  		},
  		"nth-last-of-type": function(next, rule){
  			var func = nthCheck(rule);

  			if(func === falseFunc$1$1) { return func; }
  			if(func === trueFunc$1)  { return getChildFunc(next); }

  			return function nthLastOfType(elem){
  				var siblings = adapter.getSiblings(elem);

  				for(var pos = 0, i = siblings.length - 1; i >= 0; i--){
  					if(adapter.isTag(siblings[i])){
  						if(siblings[i] === elem) { break; }
  						if(adapter.getName(siblings[i]) === adapter.getName(elem)) { pos++; }
  					}
  				}

  				return func(pos) && next(elem);
  			};
  		},

  		
  		root: function(next){
  			return function(elem){
  				return !adapter.getParent(elem) && next(elem);
  			};
  		},

  		scope: function(next, rule, options, context){
  			if(!context || context.length === 0){
  				
  				return filters.root(next);
  			}

  			if(context.length === 1){
  				
  				return function(elem){
  					return equals(context[0], elem) && next(elem);
  				};
  			}

  			return function(elem){
  				return context.indexOf(elem) >= 0 && next(elem);
  			};
  		},

  		
  		checkbox: getAttribFunc("type", "checkbox"),
  		file: getAttribFunc("type", "file"),
  		password: getAttribFunc("type", "password"),
  		radio: getAttribFunc("type", "radio"),
  		reset: getAttribFunc("type", "reset"),
  		image: getAttribFunc("type", "image"),
  		submit: getAttribFunc("type", "submit")
  	};
  	return filters;
  }

  function pseudosFactory(adapter){
  	
  	function getFirstElement(elems){
  		for(var i = 0; elems && i < elems.length; i++){
  			if(adapter.isTag(elems[i])) { return elems[i]; }
  		}
  	}

  	
  	var pseudos = {
  		empty: function(elem){
  			return !adapter.getChildren(elem).some(function(elem){
  				return adapter.isTag(elem) || elem.type === "text";
  			});
  		},

  		"first-child": function(elem){
  			return getFirstElement(adapter.getSiblings(elem)) === elem;
  		},
  		"last-child": function(elem){
  			var siblings = adapter.getSiblings(elem);

  			for(var i = siblings.length - 1; i >= 0; i--){
  				if(siblings[i] === elem) { return true; }
  				if(adapter.isTag(siblings[i])) { break; }
  			}

  			return false;
  		},
  		"first-of-type": function(elem){
  			var siblings = adapter.getSiblings(elem);

  			for(var i = 0; i < siblings.length; i++){
  				if(adapter.isTag(siblings[i])){
  					if(siblings[i] === elem) { return true; }
  					if(adapter.getName(siblings[i]) === adapter.getName(elem)) { break; }
  				}
  			}

  			return false;
  		},
  		"last-of-type": function(elem){
  			var siblings = adapter.getSiblings(elem);

  			for(var i = siblings.length - 1; i >= 0; i--){
  				if(adapter.isTag(siblings[i])){
  					if(siblings[i] === elem) { return true; }
  					if(adapter.getName(siblings[i]) === adapter.getName(elem)) { break; }
  				}
  			}

  			return false;
  		},
  		"only-of-type": function(elem){
  			var siblings = adapter.getSiblings(elem);

  			for(var i = 0, j = siblings.length; i < j; i++){
  				if(adapter.isTag(siblings[i])){
  					if(siblings[i] === elem) { continue; }
  					if(adapter.getName(siblings[i]) === adapter.getName(elem)) { return false; }
  				}
  			}

  			return true;
  		},
  		"only-child": function(elem){
  			var siblings = adapter.getSiblings(elem);

  			for(var i = 0; i < siblings.length; i++){
  				if(adapter.isTag(siblings[i]) && siblings[i] !== elem) { return false; }
  			}

  			return true;
  		},

  		
  		link: function(elem){
  			return adapter.hasAttrib(elem, "href");
  		},
  		visited: falseFunc$1$1, 
  		

  		
  		

  		
  		selected: function(elem){
  			if(adapter.hasAttrib(elem, "selected")) { return true; }
  			else if(adapter.getName(elem) !== "option") { return false; }

  			
  			var parent = adapter.getParent(elem);

  			if(
  				!parent ||
  				adapter.getName(parent) !== "select" ||
  				adapter.hasAttrib(parent, "multiple")
  			) { return false; }

  			var siblings = adapter.getChildren(parent),
  				sawElem  = false;

  			for(var i = 0; i < siblings.length; i++){
  				if(adapter.isTag(siblings[i])){
  					if(siblings[i] === elem){
  						sawElem = true;
  					} else if(!sawElem){
  						return false;
  					} else if(adapter.hasAttrib(siblings[i], "selected")){
  						return false;
  					}
  				}
  			}

  			return sawElem;
  		},
  		
  		
  		
  		
  		
  		
  		disabled: function(elem){
  			return adapter.hasAttrib(elem, "disabled");
  		},
  		enabled: function(elem){
  			return !adapter.hasAttrib(elem, "disabled");
  		},
  		
  		checked: function(elem){
  			return adapter.hasAttrib(elem, "checked") || pseudos.selected(elem);
  		},
  		
  		required: function(elem){
  			return adapter.hasAttrib(elem, "required");
  		},
  		
  		optional: function(elem){
  			return !adapter.hasAttrib(elem, "required");
  		},

  		

  		
  		parent: function(elem){
  			return !pseudos.empty(elem);
  		},
  		
  		header: function(elem){
  			var name = adapter.getName(elem);
  			return name === "h1" ||
  					name === "h2" ||
  					name === "h3" ||
  					name === "h4" ||
  					name === "h5" ||
  					name === "h6";
  		},

  		
  		button: function(elem){
  			var name = adapter.getName(elem);
  			return name === "button" ||
  					name === "input" &&
  					adapter.getAttributeValue(elem, "type") === "button";
  		},
  		
  		input: function(elem){
  			var name = adapter.getName(elem);
  			return name === "input" ||
  					name === "textarea" ||
  					name === "select" ||
  					name === "button";
  		},
  		
  		text: function(elem){
  			var attr;
  			return adapter.getName(elem) === "input" && (
  				!(attr = adapter.getAttributeValue(elem, "type")) ||
  				attr.toLowerCase() === "text"
  			);
  		}
  	};

  	return pseudos;
  }

  function verifyArgs(func, name, subselect){
  	if(subselect === null){
  		if(func.length > 1 && name !== "scope"){
  			throw new Error("pseudo-selector :" + name + " requires an argument");
  		}
  	} else {
  		if(func.length === 1){
  			throw new Error("pseudo-selector :" + name + " doesn't have any arguments");
  		}
  	}
  }


  var re_CSS3 = /^(?:(?:nth|last|first|only)-(?:child|of-type)|root|empty|(?:en|dis)abled|checked|not)$/;

  function factory$1(adapter){
  	var pseudos = pseudosFactory(adapter);
  	var filters = filtersFactory(adapter);

  	return {
  		compile: function(next, data, options, context){
  			var name = data.name,
  				subselect = data.data;

  			if(options && options.strict && !re_CSS3.test(name)){
  				throw new Error(":" + name + " isn't part of CSS3");
  			}

  			if(typeof filters[name] === "function"){
  				verifyArgs(filters[name], name,  subselect);
  				return filters[name](next, subselect, options, context);
  			} else if(typeof pseudos[name] === "function"){
  				var func = pseudos[name];
  				verifyArgs(func, name, subselect);

  				if(next === trueFunc$1) { return func; }

  				return function pseudoArgs(elem){
  					return func(elem, subselect) && next(elem);
  				};
  			} else {
  				throw new Error("unmatched pseudo-class :" + name);
  			}
  		},
  		filters: filters,
  		pseudos: pseudos
  	};
  }

  var pseudos = factory$1;

  var compile$1 = compileFactory;

  var trueFunc$1$1       = boolbase.trueFunc,
  	falseFunc$2      = boolbase.falseFunc;

  function compileFactory(adapter){
  	var Pseudos     = pseudos(adapter),
  		filters     = Pseudos.filters,
  		Rules 			= general(adapter, Pseudos);

  	function compile(selector, options, context){
  		var next = compileUnsafe(selector, options, context);
  		return wrap(next);
  	}

  	function wrap(next){
  		return function base(elem){
  			return adapter.isTag(elem) && next(elem);
  		};
  	}

  	function compileUnsafe(selector, options, context){
  		var token = cssWhat(selector, options);
  		return compileToken(token, options, context);
  	}

  	function includesScopePseudo(t){
  		return t.type === "pseudo" && (
  			t.name === "scope" || (
  				Array.isArray(t.data) &&
  				t.data.some(function(data){
  					return data.some(includesScopePseudo);
  				})
  			)
  		);
  	}

  	var DESCENDANT_TOKEN = {type: "descendant"},
  		FLEXIBLE_DESCENDANT_TOKEN = {type: "_flexibleDescendant"},
  		SCOPE_TOKEN = {type: "pseudo", name: "scope"},
  		PLACEHOLDER_ELEMENT = {};

  	
  	
  	function absolutize(token, context){
  		
  		var hasContext = !!context && !!context.length && context.every(function(e){
  			return e === PLACEHOLDER_ELEMENT || !!adapter.getParent(e);
  		});


  		token.forEach(function(t){
  			if(t.length > 0 && isTraversal(t[0]) && t[0].type !== "descendant"); else if(hasContext && !includesScopePseudo(t)){
  				t.unshift(DESCENDANT_TOKEN);
  			} else {
  				return;
  			}

  			t.unshift(SCOPE_TOKEN);
  		});
  	}

  	function compileToken(token, options, context){
  		token = token.filter(function(t){ return t.length > 0; });

  		token.forEach(sort);

  		var isArrayContext = Array.isArray(context);

  		context = (options && options.context) || context;

  		if(context && !isArrayContext) { context = [context]; }

  		absolutize(token, context);

  		var shouldTestNextSiblings = false;

  		var query = token
  			.map(function(rules){
  				if(rules[0] && rules[1] && rules[0].name === "scope"){
  					var ruleType = rules[1].type;
  					if(isArrayContext && ruleType === "descendant") { rules[1] = FLEXIBLE_DESCENDANT_TOKEN; }
  					else if(ruleType === "adjacent" || ruleType === "sibling") { shouldTestNextSiblings = true; }
  				}
  				return compileRules(rules, options, context);
  			})
  			.reduce(reduceRules, falseFunc$2);

  		query.shouldTestNextSiblings = shouldTestNextSiblings;

  		return query;
  	}

  	function isTraversal(t){
  		return procedure$2[t.type] < 0;
  	}

  	function compileRules(rules, options, context){
  		return rules.reduce(function(func, rule){
  			if(func === falseFunc$2) { return func; }
  			return Rules[rule.type](func, rule, options, context);
  		}, options && options.rootFunc || trueFunc$1$1);
  	}

  	function reduceRules(a, b){
  		if(b === falseFunc$2 || a === trueFunc$1$1){
  			return a;
  		}
  		if(a === falseFunc$2 || b === trueFunc$1$1){
  			return b;
  		}

  		return function combine(elem){
  			return a(elem) || b(elem);
  		};
  	}

  	function containsTraversal(t){
  		return t.some(isTraversal);
  	}

  	
  	
  	
  	filters.not = function(next, token, options, context){
  		var opts = {
  			xmlMode: !!(options && options.xmlMode),
  			strict: !!(options && options.strict)
  		};

  		if(opts.strict){
  			if(token.length > 1 || token.some(containsTraversal)){
  				throw new Error("complex selectors in :not aren't allowed in strict mode");
  			}
  		}

  		var func = compileToken(token, opts, context);

  		if(func === falseFunc$2) { return next; }
  		if(func === trueFunc$1$1)  { return falseFunc$2; }

  		return function(elem){
  			return !func(elem) && next(elem);
  		};
  	};

  	filters.has = function(next, token, options){
  		var opts = {
  			xmlMode: !!(options && options.xmlMode),
  			strict: !!(options && options.strict)
  		};

  		
  		var context = token.some(containsTraversal) ? [PLACEHOLDER_ELEMENT] : null;

  		var func = compileToken(token, opts, context);

  		if(func === falseFunc$2) { return falseFunc$2; }
  		if(func === trueFunc$1$1){
  			return function(elem){
  				return adapter.getChildren(elem).some(adapter.isTag) && next(elem);
  			};
  		}

  		func = wrap(func);

  		if(context){
  			return function has(elem){
  				return next(elem) && (
  					(context[0] = elem), adapter.existsOne(func, adapter.getChildren(elem))
  				);
  			};
  		}

  		return function has(elem){
  			return next(elem) && adapter.existsOne(func, adapter.getChildren(elem));
  		};
  	};

  	filters.matches = function(next, token, options, context){
  		var opts = {
  			xmlMode: !!(options && options.xmlMode),
  			strict: !!(options && options.strict),
  			rootFunc: next
  		};

  		return compileToken(token, opts, context);
  	};

  	compile.compileToken = compileToken;
  	compile.compileUnsafe = compileUnsafe;
  	compile.Pseudos = Pseudos;

  	return compile;
  }

  var cssSelect = CSSselect;

  var falseFunc$3      = boolbase.falseFunc,
  	defaultCompile = compile$1(domUtils);

  function adapterCompile(adapter){
  	if(!adapter.__compile__){
  		adapter.__compile__ = compile$1(adapter);
  	}
  	return adapter.__compile__
  }

  function getSelectorFunc(searchFunc){
  	return function select(query, elems, options){
  		options = options || {};
  		options.adapter = options.adapter || domUtils;
  		var compile$$1 = adapterCompile(options.adapter);

  		if(typeof query !== "function") { query = compile$$1.compileUnsafe(query, options, elems); }
  		if(query.shouldTestNextSiblings) { elems = appendNextSiblings((options && options.context) || elems, options.adapter); }
  		if(!Array.isArray(elems)) { elems = options.adapter.getChildren(elems); }
  		else { elems = options.adapter.removeSubsets(elems); }
  		return searchFunc(query, elems, options);
  	};
  }

  function getNextSiblings(elem, adapter){
  	var siblings = adapter.getSiblings(elem);
  	if(!Array.isArray(siblings)) { return []; }
  	siblings = siblings.slice(0);
  	while(siblings.shift() !== elem){ }
  	return siblings;
  }

  function appendNextSiblings(elems, adapter){
  	
  	if(!Array.isArray(elems)) { elems = [elems]; }
  	var newElems = elems.slice(0);

  	for(var i = 0, len = elems.length; i < len; i++){
  		var nextSiblings = getNextSiblings(newElems[i], adapter);
  		newElems.push.apply(newElems, nextSiblings);
  	}
  	return newElems;
  }

  var selectAll = getSelectorFunc(function selectAll(query, elems, options){
  	return (query === falseFunc$3 || !elems || elems.length === 0) ? [] : options.adapter.findAll(query, elems);
  });

  var selectOne = getSelectorFunc(function selectOne(query, elems, options){
  	return (query === falseFunc$3 || !elems || elems.length === 0) ? null : options.adapter.findOne(query, elems);
  });

  function is(elem, query, options){
  	options = options || {};
  	options.adapter = options.adapter || domUtils;
  	var compile$$1 = adapterCompile(options.adapter);
  	return (typeof query === "function" ? query : compile$$1(query, options))(elem);
  }


  function CSSselect(query, elems, options){
  	return selectAll(query, elems, options);
  }

  CSSselect.compile = defaultCompile;
  CSSselect.filters = defaultCompile.Pseudos.filters;
  CSSselect.pseudos = defaultCompile.Pseudos.pseudos;

  CSSselect.selectAll = selectAll;
  CSSselect.selectOne = selectOne;

  CSSselect.is = is;


  CSSselect.parse = defaultCompile;
  CSSselect.iterate = selectAll;


  CSSselect._compileUnsafe = defaultCompile.compileUnsafe;
  CSSselect._compileToken = defaultCompile.compileToken;

  var amp = "&";
  var apos = "'";
  var gt = ">";
  var lt = "<";
  var quot = "\"";
  var xmlJSON = {
  	amp: amp,
  	apos: apos,
  	gt: gt,
  	lt: lt,
  	quot: quot
  };

  var xml = Object.freeze({
  	amp: amp,
  	apos: apos,
  	gt: gt,
  	lt: lt,
  	quot: quot,
  	default: xmlJSON
  });

  var Aacute = "Á";
  var aacute = "á";
  var Abreve = "Ă";
  var abreve = "ă";
  var ac = "∾";
  var acd = "∿";
  var acE = "∾̳";
  var Acirc = "Â";
  var acirc = "â";
  var acute = "´";
  var Acy = "А";
  var acy = "а";
  var AElig = "Æ";
  var aelig = "æ";
  var af = "⁡";
  var Afr = "𝔄";
  var afr = "𝔞";
  var Agrave = "À";
  var agrave = "à";
  var alefsym = "ℵ";
  var aleph = "ℵ";
  var Alpha = "Α";
  var alpha = "α";
  var Amacr = "Ā";
  var amacr = "ā";
  var amalg = "⨿";
  var amp$1 = "&";
  var AMP = "&";
  var andand = "⩕";
  var And = "⩓";
  var and = "∧";
  var andd = "⩜";
  var andslope = "⩘";
  var andv = "⩚";
  var ang = "∠";
  var ange = "⦤";
  var angle = "∠";
  var angmsdaa = "⦨";
  var angmsdab = "⦩";
  var angmsdac = "⦪";
  var angmsdad = "⦫";
  var angmsdae = "⦬";
  var angmsdaf = "⦭";
  var angmsdag = "⦮";
  var angmsdah = "⦯";
  var angmsd = "∡";
  var angrt = "∟";
  var angrtvb = "⊾";
  var angrtvbd = "⦝";
  var angsph = "∢";
  var angst = "Å";
  var angzarr = "⍼";
  var Aogon = "Ą";
  var aogon = "ą";
  var Aopf = "𝔸";
  var aopf = "𝕒";
  var apacir = "⩯";
  var ap = "≈";
  var apE = "⩰";
  var ape = "≊";
  var apid = "≋";
  var apos$1 = "'";
  var ApplyFunction = "⁡";
  var approx = "≈";
  var approxeq = "≊";
  var Aring = "Å";
  var aring = "å";
  var Ascr = "𝒜";
  var ascr = "𝒶";
  var Assign = "≔";
  var ast = "*";
  var asymp = "≈";
  var asympeq = "≍";
  var Atilde = "Ã";
  var atilde = "ã";
  var Auml = "Ä";
  var auml = "ä";
  var awconint = "∳";
  var awint = "⨑";
  var backcong = "≌";
  var backepsilon = "϶";
  var backprime = "‵";
  var backsim = "∽";
  var backsimeq = "⋍";
  var Backslash = "∖";
  var Barv = "⫧";
  var barvee = "⊽";
  var barwed = "⌅";
  var Barwed = "⌆";
  var barwedge = "⌅";
  var bbrk = "⎵";
  var bbrktbrk = "⎶";
  var bcong = "≌";
  var Bcy = "Б";
  var bcy = "б";
  var bdquo = "„";
  var becaus = "∵";
  var because = "∵";
  var Because = "∵";
  var bemptyv = "⦰";
  var bepsi = "϶";
  var bernou = "ℬ";
  var Bernoullis = "ℬ";
  var Beta = "Β";
  var beta = "β";
  var beth = "ℶ";
  var between = "≬";
  var Bfr = "𝔅";
  var bfr = "𝔟";
  var bigcap = "⋂";
  var bigcirc = "◯";
  var bigcup = "⋃";
  var bigodot = "⨀";
  var bigoplus = "⨁";
  var bigotimes = "⨂";
  var bigsqcup = "⨆";
  var bigstar = "★";
  var bigtriangledown = "▽";
  var bigtriangleup = "△";
  var biguplus = "⨄";
  var bigvee = "⋁";
  var bigwedge = "⋀";
  var bkarow = "⤍";
  var blacklozenge = "⧫";
  var blacksquare = "▪";
  var blacktriangle = "▴";
  var blacktriangledown = "▾";
  var blacktriangleleft = "◂";
  var blacktriangleright = "▸";
  var blank = "␣";
  var blk12 = "▒";
  var blk14 = "░";
  var blk34 = "▓";
  var block = "█";
  var bne = "=⃥";
  var bnequiv = "≡⃥";
  var bNot = "⫭";
  var bnot = "⌐";
  var Bopf = "𝔹";
  var bopf = "𝕓";
  var bot = "⊥";
  var bottom = "⊥";
  var bowtie = "⋈";
  var boxbox = "⧉";
  var boxdl = "┐";
  var boxdL = "╕";
  var boxDl = "╖";
  var boxDL = "╗";
  var boxdr = "┌";
  var boxdR = "╒";
  var boxDr = "╓";
  var boxDR = "╔";
  var boxh = "─";
  var boxH = "═";
  var boxhd = "┬";
  var boxHd = "╤";
  var boxhD = "╥";
  var boxHD = "╦";
  var boxhu = "┴";
  var boxHu = "╧";
  var boxhU = "╨";
  var boxHU = "╩";
  var boxminus = "⊟";
  var boxplus = "⊞";
  var boxtimes = "⊠";
  var boxul = "┘";
  var boxuL = "╛";
  var boxUl = "╜";
  var boxUL = "╝";
  var boxur = "└";
  var boxuR = "╘";
  var boxUr = "╙";
  var boxUR = "╚";
  var boxv = "│";
  var boxV = "║";
  var boxvh = "┼";
  var boxvH = "╪";
  var boxVh = "╫";
  var boxVH = "╬";
  var boxvl = "┤";
  var boxvL = "╡";
  var boxVl = "╢";
  var boxVL = "╣";
  var boxvr = "├";
  var boxvR = "╞";
  var boxVr = "╟";
  var boxVR = "╠";
  var bprime = "‵";
  var breve = "˘";
  var Breve = "˘";
  var brvbar = "¦";
  var bscr = "𝒷";
  var Bscr = "ℬ";
  var bsemi = "⁏";
  var bsim = "∽";
  var bsime = "⋍";
  var bsolb = "⧅";
  var bsol = "\\";
  var bsolhsub = "⟈";
  var bull = "•";
  var bullet = "•";
  var bump = "≎";
  var bumpE = "⪮";
  var bumpe = "≏";
  var Bumpeq = "≎";
  var bumpeq = "≏";
  var Cacute = "Ć";
  var cacute = "ć";
  var capand = "⩄";
  var capbrcup = "⩉";
  var capcap = "⩋";
  var cap = "∩";
  var Cap = "⋒";
  var capcup = "⩇";
  var capdot = "⩀";
  var CapitalDifferentialD = "ⅅ";
  var caps = "∩︀";
  var caret = "⁁";
  var caron = "ˇ";
  var Cayleys = "ℭ";
  var ccaps = "⩍";
  var Ccaron = "Č";
  var ccaron = "č";
  var Ccedil = "Ç";
  var ccedil = "ç";
  var Ccirc = "Ĉ";
  var ccirc = "ĉ";
  var Cconint = "∰";
  var ccups = "⩌";
  var ccupssm = "⩐";
  var Cdot = "Ċ";
  var cdot = "ċ";
  var cedil = "¸";
  var Cedilla = "¸";
  var cemptyv = "⦲";
  var cent = "¢";
  var centerdot = "·";
  var CenterDot = "·";
  var cfr = "𝔠";
  var Cfr = "ℭ";
  var CHcy = "Ч";
  var chcy = "ч";
  var check = "✓";
  var checkmark = "✓";
  var Chi = "Χ";
  var chi = "χ";
  var circ = "ˆ";
  var circeq = "≗";
  var circlearrowleft = "↺";
  var circlearrowright = "↻";
  var circledast = "⊛";
  var circledcirc = "⊚";
  var circleddash = "⊝";
  var CircleDot = "⊙";
  var circledR = "®";
  var circledS = "Ⓢ";
  var CircleMinus = "⊖";
  var CirclePlus = "⊕";
  var CircleTimes = "⊗";
  var cir = "○";
  var cirE = "⧃";
  var cire = "≗";
  var cirfnint = "⨐";
  var cirmid = "⫯";
  var cirscir = "⧂";
  var ClockwiseContourIntegral = "∲";
  var CloseCurlyDoubleQuote = "”";
  var CloseCurlyQuote = "’";
  var clubs = "♣";
  var clubsuit = "♣";
  var colon = ":";
  var Colon = "∷";
  var Colone = "⩴";
  var colone = "≔";
  var coloneq = "≔";
  var comma = ",";
  var commat = "@";
  var comp = "∁";
  var compfn = "∘";
  var complement = "∁";
  var complexes = "ℂ";
  var cong = "≅";
  var congdot = "⩭";
  var Congruent = "≡";
  var conint = "∮";
  var Conint = "∯";
  var ContourIntegral = "∮";
  var copf = "𝕔";
  var Copf = "ℂ";
  var coprod = "∐";
  var Coproduct = "∐";
  var copy = "©";
  var COPY = "©";
  var copysr = "℗";
  var CounterClockwiseContourIntegral = "∳";
  var crarr = "↵";
  var cross = "✗";
  var Cross = "⨯";
  var Cscr = "𝒞";
  var cscr = "𝒸";
  var csub = "⫏";
  var csube = "⫑";
  var csup = "⫐";
  var csupe = "⫒";
  var ctdot = "⋯";
  var cudarrl = "⤸";
  var cudarrr = "⤵";
  var cuepr = "⋞";
  var cuesc = "⋟";
  var cularr = "↶";
  var cularrp = "⤽";
  var cupbrcap = "⩈";
  var cupcap = "⩆";
  var CupCap = "≍";
  var cup = "∪";
  var Cup = "⋓";
  var cupcup = "⩊";
  var cupdot = "⊍";
  var cupor = "⩅";
  var cups = "∪︀";
  var curarr = "↷";
  var curarrm = "⤼";
  var curlyeqprec = "⋞";
  var curlyeqsucc = "⋟";
  var curlyvee = "⋎";
  var curlywedge = "⋏";
  var curren = "¤";
  var curvearrowleft = "↶";
  var curvearrowright = "↷";
  var cuvee = "⋎";
  var cuwed = "⋏";
  var cwconint = "∲";
  var cwint = "∱";
  var cylcty = "⌭";
  var dagger = "†";
  var Dagger = "‡";
  var daleth = "ℸ";
  var darr = "↓";
  var Darr = "↡";
  var dArr = "⇓";
  var dash = "‐";
  var Dashv = "⫤";
  var dashv = "⊣";
  var dbkarow = "⤏";
  var dblac = "˝";
  var Dcaron = "Ď";
  var dcaron = "ď";
  var Dcy = "Д";
  var dcy = "д";
  var ddagger = "‡";
  var ddarr = "⇊";
  var DD = "ⅅ";
  var dd = "ⅆ";
  var DDotrahd = "⤑";
  var ddotseq = "⩷";
  var deg = "°";
  var Del = "∇";
  var Delta = "Δ";
  var delta = "δ";
  var demptyv = "⦱";
  var dfisht = "⥿";
  var Dfr = "𝔇";
  var dfr = "𝔡";
  var dHar = "⥥";
  var dharl = "⇃";
  var dharr = "⇂";
  var DiacriticalAcute = "´";
  var DiacriticalDot = "˙";
  var DiacriticalDoubleAcute = "˝";
  var DiacriticalGrave = "`";
  var DiacriticalTilde = "˜";
  var diam = "⋄";
  var diamond = "⋄";
  var Diamond = "⋄";
  var diamondsuit = "♦";
  var diams = "♦";
  var die = "¨";
  var DifferentialD = "ⅆ";
  var digamma = "ϝ";
  var disin = "⋲";
  var div = "÷";
  var divide = "÷";
  var divideontimes = "⋇";
  var divonx = "⋇";
  var DJcy = "Ђ";
  var djcy = "ђ";
  var dlcorn = "⌞";
  var dlcrop = "⌍";
  var dollar = "$";
  var Dopf = "𝔻";
  var dopf = "𝕕";
  var Dot = "¨";
  var dot = "˙";
  var DotDot = "⃜";
  var doteq = "≐";
  var doteqdot = "≑";
  var DotEqual = "≐";
  var dotminus = "∸";
  var dotplus = "∔";
  var dotsquare = "⊡";
  var doublebarwedge = "⌆";
  var DoubleContourIntegral = "∯";
  var DoubleDot = "¨";
  var DoubleDownArrow = "⇓";
  var DoubleLeftArrow = "⇐";
  var DoubleLeftRightArrow = "⇔";
  var DoubleLeftTee = "⫤";
  var DoubleLongLeftArrow = "⟸";
  var DoubleLongLeftRightArrow = "⟺";
  var DoubleLongRightArrow = "⟹";
  var DoubleRightArrow = "⇒";
  var DoubleRightTee = "⊨";
  var DoubleUpArrow = "⇑";
  var DoubleUpDownArrow = "⇕";
  var DoubleVerticalBar = "∥";
  var DownArrowBar = "⤓";
  var downarrow = "↓";
  var DownArrow = "↓";
  var Downarrow = "⇓";
  var DownArrowUpArrow = "⇵";
  var DownBreve = "̑";
  var downdownarrows = "⇊";
  var downharpoonleft = "⇃";
  var downharpoonright = "⇂";
  var DownLeftRightVector = "⥐";
  var DownLeftTeeVector = "⥞";
  var DownLeftVectorBar = "⥖";
  var DownLeftVector = "↽";
  var DownRightTeeVector = "⥟";
  var DownRightVectorBar = "⥗";
  var DownRightVector = "⇁";
  var DownTeeArrow = "↧";
  var DownTee = "⊤";
  var drbkarow = "⤐";
  var drcorn = "⌟";
  var drcrop = "⌌";
  var Dscr = "𝒟";
  var dscr = "𝒹";
  var DScy = "Ѕ";
  var dscy = "ѕ";
  var dsol = "⧶";
  var Dstrok = "Đ";
  var dstrok = "đ";
  var dtdot = "⋱";
  var dtri = "▿";
  var dtrif = "▾";
  var duarr = "⇵";
  var duhar = "⥯";
  var dwangle = "⦦";
  var DZcy = "Џ";
  var dzcy = "џ";
  var dzigrarr = "⟿";
  var Eacute = "É";
  var eacute = "é";
  var easter = "⩮";
  var Ecaron = "Ě";
  var ecaron = "ě";
  var Ecirc = "Ê";
  var ecirc = "ê";
  var ecir = "≖";
  var ecolon = "≕";
  var Ecy = "Э";
  var ecy = "э";
  var eDDot = "⩷";
  var Edot = "Ė";
  var edot = "ė";
  var eDot = "≑";
  var ee = "ⅇ";
  var efDot = "≒";
  var Efr = "𝔈";
  var efr = "𝔢";
  var eg = "⪚";
  var Egrave = "È";
  var egrave = "è";
  var egs = "⪖";
  var egsdot = "⪘";
  var el = "⪙";
  var Element = "∈";
  var elinters = "⏧";
  var ell = "ℓ";
  var els = "⪕";
  var elsdot = "⪗";
  var Emacr = "Ē";
  var emacr = "ē";
  var empty = "∅";
  var emptyset = "∅";
  var EmptySmallSquare = "◻";
  var emptyv = "∅";
  var EmptyVerySmallSquare = "▫";
  var emsp13 = " ";
  var emsp14 = " ";
  var emsp = " ";
  var ENG = "Ŋ";
  var eng = "ŋ";
  var ensp = " ";
  var Eogon = "Ę";
  var eogon = "ę";
  var Eopf = "𝔼";
  var eopf = "𝕖";
  var epar = "⋕";
  var eparsl = "⧣";
  var eplus = "⩱";
  var epsi = "ε";
  var Epsilon = "Ε";
  var epsilon = "ε";
  var epsiv = "ϵ";
  var eqcirc = "≖";
  var eqcolon = "≕";
  var eqsim = "≂";
  var eqslantgtr = "⪖";
  var eqslantless = "⪕";
  var Equal = "⩵";
  var equals = "=";
  var EqualTilde = "≂";
  var equest = "≟";
  var Equilibrium = "⇌";
  var equiv = "≡";
  var equivDD = "⩸";
  var eqvparsl = "⧥";
  var erarr = "⥱";
  var erDot = "≓";
  var escr = "ℯ";
  var Escr = "ℰ";
  var esdot = "≐";
  var Esim = "⩳";
  var esim = "≂";
  var Eta = "Η";
  var eta = "η";
  var ETH = "Ð";
  var eth = "ð";
  var Euml = "Ë";
  var euml = "ë";
  var euro = "€";
  var excl = "!";
  var exist = "∃";
  var Exists = "∃";
  var expectation = "ℰ";
  var exponentiale = "ⅇ";
  var ExponentialE = "ⅇ";
  var fallingdotseq = "≒";
  var Fcy = "Ф";
  var fcy = "ф";
  var female = "♀";
  var ffilig = "ﬃ";
  var fflig = "ﬀ";
  var ffllig = "ﬄ";
  var Ffr = "𝔉";
  var ffr = "𝔣";
  var filig = "ﬁ";
  var FilledSmallSquare = "◼";
  var FilledVerySmallSquare = "▪";
  var fjlig = "fj";
  var flat = "♭";
  var fllig = "ﬂ";
  var fltns = "▱";
  var fnof = "ƒ";
  var Fopf = "𝔽";
  var fopf = "𝕗";
  var forall = "∀";
  var ForAll = "∀";
  var fork = "⋔";
  var forkv = "⫙";
  var Fouriertrf = "ℱ";
  var fpartint = "⨍";
  var frac12 = "½";
  var frac13 = "⅓";
  var frac14 = "¼";
  var frac15 = "⅕";
  var frac16 = "⅙";
  var frac18 = "⅛";
  var frac23 = "⅔";
  var frac25 = "⅖";
  var frac34 = "¾";
  var frac35 = "⅗";
  var frac38 = "⅜";
  var frac45 = "⅘";
  var frac56 = "⅚";
  var frac58 = "⅝";
  var frac78 = "⅞";
  var frasl = "⁄";
  var frown = "⌢";
  var fscr = "𝒻";
  var Fscr = "ℱ";
  var gacute = "ǵ";
  var Gamma = "Γ";
  var gamma = "γ";
  var Gammad = "Ϝ";
  var gammad = "ϝ";
  var gap = "⪆";
  var Gbreve = "Ğ";
  var gbreve = "ğ";
  var Gcedil = "Ģ";
  var Gcirc = "Ĝ";
  var gcirc = "ĝ";
  var Gcy = "Г";
  var gcy = "г";
  var Gdot = "Ġ";
  var gdot = "ġ";
  var ge = "≥";
  var gE = "≧";
  var gEl = "⪌";
  var gel = "⋛";
  var geq = "≥";
  var geqq = "≧";
  var geqslant = "⩾";
  var gescc = "⪩";
  var ges = "⩾";
  var gesdot = "⪀";
  var gesdoto = "⪂";
  var gesdotol = "⪄";
  var gesl = "⋛︀";
  var gesles = "⪔";
  var Gfr = "𝔊";
  var gfr = "𝔤";
  var gg = "≫";
  var Gg = "⋙";
  var ggg = "⋙";
  var gimel = "ℷ";
  var GJcy = "Ѓ";
  var gjcy = "ѓ";
  var gla = "⪥";
  var gl = "≷";
  var glE = "⪒";
  var glj = "⪤";
  var gnap = "⪊";
  var gnapprox = "⪊";
  var gne = "⪈";
  var gnE = "≩";
  var gneq = "⪈";
  var gneqq = "≩";
  var gnsim = "⋧";
  var Gopf = "𝔾";
  var gopf = "𝕘";
  var grave = "`";
  var GreaterEqual = "≥";
  var GreaterEqualLess = "⋛";
  var GreaterFullEqual = "≧";
  var GreaterGreater = "⪢";
  var GreaterLess = "≷";
  var GreaterSlantEqual = "⩾";
  var GreaterTilde = "≳";
  var Gscr = "𝒢";
  var gscr = "ℊ";
  var gsim = "≳";
  var gsime = "⪎";
  var gsiml = "⪐";
  var gtcc = "⪧";
  var gtcir = "⩺";
  var gt$1 = ">";
  var GT = ">";
  var Gt = "≫";
  var gtdot = "⋗";
  var gtlPar = "⦕";
  var gtquest = "⩼";
  var gtrapprox = "⪆";
  var gtrarr = "⥸";
  var gtrdot = "⋗";
  var gtreqless = "⋛";
  var gtreqqless = "⪌";
  var gtrless = "≷";
  var gtrsim = "≳";
  var gvertneqq = "≩︀";
  var gvnE = "≩︀";
  var Hacek = "ˇ";
  var hairsp = " ";
  var half = "½";
  var hamilt = "ℋ";
  var HARDcy = "Ъ";
  var hardcy = "ъ";
  var harrcir = "⥈";
  var harr = "↔";
  var hArr = "⇔";
  var harrw = "↭";
  var Hat = "^";
  var hbar = "ℏ";
  var Hcirc = "Ĥ";
  var hcirc = "ĥ";
  var hearts = "♥";
  var heartsuit = "♥";
  var hellip = "…";
  var hercon = "⊹";
  var hfr = "𝔥";
  var Hfr = "ℌ";
  var HilbertSpace = "ℋ";
  var hksearow = "⤥";
  var hkswarow = "⤦";
  var hoarr = "⇿";
  var homtht = "∻";
  var hookleftarrow = "↩";
  var hookrightarrow = "↪";
  var hopf = "𝕙";
  var Hopf = "ℍ";
  var horbar = "―";
  var HorizontalLine = "─";
  var hscr = "𝒽";
  var Hscr = "ℋ";
  var hslash = "ℏ";
  var Hstrok = "Ħ";
  var hstrok = "ħ";
  var HumpDownHump = "≎";
  var HumpEqual = "≏";
  var hybull = "⁃";
  var hyphen = "‐";
  var Iacute = "Í";
  var iacute = "í";
  var ic = "⁣";
  var Icirc = "Î";
  var icirc = "î";
  var Icy = "И";
  var icy = "и";
  var Idot = "İ";
  var IEcy = "Е";
  var iecy = "е";
  var iexcl = "¡";
  var iff = "⇔";
  var ifr = "𝔦";
  var Ifr = "ℑ";
  var Igrave = "Ì";
  var igrave = "ì";
  var ii = "ⅈ";
  var iiiint = "⨌";
  var iiint = "∭";
  var iinfin = "⧜";
  var iiota = "℩";
  var IJlig = "Ĳ";
  var ijlig = "ĳ";
  var Imacr = "Ī";
  var imacr = "ī";
  var image = "ℑ";
  var ImaginaryI = "ⅈ";
  var imagline = "ℐ";
  var imagpart = "ℑ";
  var imath = "ı";
  var Im = "ℑ";
  var imof = "⊷";
  var imped = "Ƶ";
  var Implies = "⇒";
  var incare = "℅";
  var infin = "∞";
  var infintie = "⧝";
  var inodot = "ı";
  var intcal = "⊺";
  var int = "∫";
  var Int = "∬";
  var integers = "ℤ";
  var Integral = "∫";
  var intercal = "⊺";
  var Intersection = "⋂";
  var intlarhk = "⨗";
  var intprod = "⨼";
  var InvisibleComma = "⁣";
  var InvisibleTimes = "⁢";
  var IOcy = "Ё";
  var iocy = "ё";
  var Iogon = "Į";
  var iogon = "į";
  var Iopf = "𝕀";
  var iopf = "𝕚";
  var Iota = "Ι";
  var iota = "ι";
  var iprod = "⨼";
  var iquest = "¿";
  var iscr = "𝒾";
  var Iscr = "ℐ";
  var isin = "∈";
  var isindot = "⋵";
  var isinE = "⋹";
  var isins = "⋴";
  var isinsv = "⋳";
  var isinv = "∈";
  var it = "⁢";
  var Itilde = "Ĩ";
  var itilde = "ĩ";
  var Iukcy = "І";
  var iukcy = "і";
  var Iuml = "Ï";
  var iuml = "ï";
  var Jcirc = "Ĵ";
  var jcirc = "ĵ";
  var Jcy = "Й";
  var jcy = "й";
  var Jfr = "𝔍";
  var jfr = "𝔧";
  var jmath = "ȷ";
  var Jopf = "𝕁";
  var jopf = "𝕛";
  var Jscr = "𝒥";
  var jscr = "𝒿";
  var Jsercy = "Ј";
  var jsercy = "ј";
  var Jukcy = "Є";
  var jukcy = "є";
  var Kappa = "Κ";
  var kappa = "κ";
  var kappav = "ϰ";
  var Kcedil = "Ķ";
  var kcedil = "ķ";
  var Kcy = "К";
  var kcy = "к";
  var Kfr = "𝔎";
  var kfr = "𝔨";
  var kgreen = "ĸ";
  var KHcy = "Х";
  var khcy = "х";
  var KJcy = "Ќ";
  var kjcy = "ќ";
  var Kopf = "𝕂";
  var kopf = "𝕜";
  var Kscr = "𝒦";
  var kscr = "𝓀";
  var lAarr = "⇚";
  var Lacute = "Ĺ";
  var lacute = "ĺ";
  var laemptyv = "⦴";
  var lagran = "ℒ";
  var Lambda = "Λ";
  var lambda = "λ";
  var lang = "⟨";
  var Lang = "⟪";
  var langd = "⦑";
  var langle = "⟨";
  var lap = "⪅";
  var Laplacetrf = "ℒ";
  var laquo = "«";
  var larrb = "⇤";
  var larrbfs = "⤟";
  var larr = "←";
  var Larr = "↞";
  var lArr = "⇐";
  var larrfs = "⤝";
  var larrhk = "↩";
  var larrlp = "↫";
  var larrpl = "⤹";
  var larrsim = "⥳";
  var larrtl = "↢";
  var latail = "⤙";
  var lAtail = "⤛";
  var lat = "⪫";
  var late = "⪭";
  var lates = "⪭︀";
  var lbarr = "⤌";
  var lBarr = "⤎";
  var lbbrk = "❲";
  var lbrace = "{";
  var lbrack = "[";
  var lbrke = "⦋";
  var lbrksld = "⦏";
  var lbrkslu = "⦍";
  var Lcaron = "Ľ";
  var lcaron = "ľ";
  var Lcedil = "Ļ";
  var lcedil = "ļ";
  var lceil = "⌈";
  var lcub = "{";
  var Lcy = "Л";
  var lcy = "л";
  var ldca = "⤶";
  var ldquo = "“";
  var ldquor = "„";
  var ldrdhar = "⥧";
  var ldrushar = "⥋";
  var ldsh = "↲";
  var le = "≤";
  var lE = "≦";
  var LeftAngleBracket = "⟨";
  var LeftArrowBar = "⇤";
  var leftarrow = "←";
  var LeftArrow = "←";
  var Leftarrow = "⇐";
  var LeftArrowRightArrow = "⇆";
  var leftarrowtail = "↢";
  var LeftCeiling = "⌈";
  var LeftDoubleBracket = "⟦";
  var LeftDownTeeVector = "⥡";
  var LeftDownVectorBar = "⥙";
  var LeftDownVector = "⇃";
  var LeftFloor = "⌊";
  var leftharpoondown = "↽";
  var leftharpoonup = "↼";
  var leftleftarrows = "⇇";
  var leftrightarrow = "↔";
  var LeftRightArrow = "↔";
  var Leftrightarrow = "⇔";
  var leftrightarrows = "⇆";
  var leftrightharpoons = "⇋";
  var leftrightsquigarrow = "↭";
  var LeftRightVector = "⥎";
  var LeftTeeArrow = "↤";
  var LeftTee = "⊣";
  var LeftTeeVector = "⥚";
  var leftthreetimes = "⋋";
  var LeftTriangleBar = "⧏";
  var LeftTriangle = "⊲";
  var LeftTriangleEqual = "⊴";
  var LeftUpDownVector = "⥑";
  var LeftUpTeeVector = "⥠";
  var LeftUpVectorBar = "⥘";
  var LeftUpVector = "↿";
  var LeftVectorBar = "⥒";
  var LeftVector = "↼";
  var lEg = "⪋";
  var leg = "⋚";
  var leq = "≤";
  var leqq = "≦";
  var leqslant = "⩽";
  var lescc = "⪨";
  var les = "⩽";
  var lesdot = "⩿";
  var lesdoto = "⪁";
  var lesdotor = "⪃";
  var lesg = "⋚︀";
  var lesges = "⪓";
  var lessapprox = "⪅";
  var lessdot = "⋖";
  var lesseqgtr = "⋚";
  var lesseqqgtr = "⪋";
  var LessEqualGreater = "⋚";
  var LessFullEqual = "≦";
  var LessGreater = "≶";
  var lessgtr = "≶";
  var LessLess = "⪡";
  var lesssim = "≲";
  var LessSlantEqual = "⩽";
  var LessTilde = "≲";
  var lfisht = "⥼";
  var lfloor = "⌊";
  var Lfr = "𝔏";
  var lfr = "𝔩";
  var lg = "≶";
  var lgE = "⪑";
  var lHar = "⥢";
  var lhard = "↽";
  var lharu = "↼";
  var lharul = "⥪";
  var lhblk = "▄";
  var LJcy = "Љ";
  var ljcy = "љ";
  var llarr = "⇇";
  var ll = "≪";
  var Ll = "⋘";
  var llcorner = "⌞";
  var Lleftarrow = "⇚";
  var llhard = "⥫";
  var lltri = "◺";
  var Lmidot = "Ŀ";
  var lmidot = "ŀ";
  var lmoustache = "⎰";
  var lmoust = "⎰";
  var lnap = "⪉";
  var lnapprox = "⪉";
  var lne = "⪇";
  var lnE = "≨";
  var lneq = "⪇";
  var lneqq = "≨";
  var lnsim = "⋦";
  var loang = "⟬";
  var loarr = "⇽";
  var lobrk = "⟦";
  var longleftarrow = "⟵";
  var LongLeftArrow = "⟵";
  var Longleftarrow = "⟸";
  var longleftrightarrow = "⟷";
  var LongLeftRightArrow = "⟷";
  var Longleftrightarrow = "⟺";
  var longmapsto = "⟼";
  var longrightarrow = "⟶";
  var LongRightArrow = "⟶";
  var Longrightarrow = "⟹";
  var looparrowleft = "↫";
  var looparrowright = "↬";
  var lopar = "⦅";
  var Lopf = "𝕃";
  var lopf = "𝕝";
  var loplus = "⨭";
  var lotimes = "⨴";
  var lowast = "∗";
  var lowbar = "_";
  var LowerLeftArrow = "↙";
  var LowerRightArrow = "↘";
  var loz = "◊";
  var lozenge = "◊";
  var lozf = "⧫";
  var lpar = "(";
  var lparlt = "⦓";
  var lrarr = "⇆";
  var lrcorner = "⌟";
  var lrhar = "⇋";
  var lrhard = "⥭";
  var lrm = "‎";
  var lrtri = "⊿";
  var lsaquo = "‹";
  var lscr = "𝓁";
  var Lscr = "ℒ";
  var lsh = "↰";
  var Lsh = "↰";
  var lsim = "≲";
  var lsime = "⪍";
  var lsimg = "⪏";
  var lsqb = "[";
  var lsquo = "‘";
  var lsquor = "‚";
  var Lstrok = "Ł";
  var lstrok = "ł";
  var ltcc = "⪦";
  var ltcir = "⩹";
  var lt$1 = "<";
  var LT = "<";
  var Lt = "≪";
  var ltdot = "⋖";
  var lthree = "⋋";
  var ltimes = "⋉";
  var ltlarr = "⥶";
  var ltquest = "⩻";
  var ltri = "◃";
  var ltrie = "⊴";
  var ltrif = "◂";
  var ltrPar = "⦖";
  var lurdshar = "⥊";
  var luruhar = "⥦";
  var lvertneqq = "≨︀";
  var lvnE = "≨︀";
  var macr = "¯";
  var male = "♂";
  var malt = "✠";
  var maltese = "✠";
  var map$1 = "↦";
  var mapsto = "↦";
  var mapstodown = "↧";
  var mapstoleft = "↤";
  var mapstoup = "↥";
  var marker = "▮";
  var mcomma = "⨩";
  var Mcy = "М";
  var mcy = "м";
  var mdash = "—";
  var mDDot = "∺";
  var measuredangle = "∡";
  var MediumSpace = " ";
  var Mellintrf = "ℳ";
  var Mfr = "𝔐";
  var mfr = "𝔪";
  var mho = "℧";
  var micro = "µ";
  var midast = "*";
  var midcir = "⫰";
  var mid = "∣";
  var middot = "·";
  var minusb = "⊟";
  var minus = "−";
  var minusd = "∸";
  var minusdu = "⨪";
  var MinusPlus = "∓";
  var mlcp = "⫛";
  var mldr = "…";
  var mnplus = "∓";
  var models = "⊧";
  var Mopf = "𝕄";
  var mopf = "𝕞";
  var mp = "∓";
  var mscr = "𝓂";
  var Mscr = "ℳ";
  var mstpos = "∾";
  var Mu = "Μ";
  var mu = "μ";
  var multimap = "⊸";
  var mumap = "⊸";
  var nabla = "∇";
  var Nacute = "Ń";
  var nacute = "ń";
  var nang = "∠⃒";
  var nap = "≉";
  var napE = "⩰̸";
  var napid = "≋̸";
  var napos = "ŉ";
  var napprox = "≉";
  var natural = "♮";
  var naturals = "ℕ";
  var natur = "♮";
  var nbsp = " ";
  var nbump = "≎̸";
  var nbumpe = "≏̸";
  var ncap = "⩃";
  var Ncaron = "Ň";
  var ncaron = "ň";
  var Ncedil = "Ņ";
  var ncedil = "ņ";
  var ncong = "≇";
  var ncongdot = "⩭̸";
  var ncup = "⩂";
  var Ncy = "Н";
  var ncy = "н";
  var ndash = "–";
  var nearhk = "⤤";
  var nearr = "↗";
  var neArr = "⇗";
  var nearrow = "↗";
  var ne = "≠";
  var nedot = "≐̸";
  var NegativeMediumSpace = "​";
  var NegativeThickSpace = "​";
  var NegativeThinSpace = "​";
  var NegativeVeryThinSpace = "​";
  var nequiv = "≢";
  var nesear = "⤨";
  var nesim = "≂̸";
  var NestedGreaterGreater = "≫";
  var NestedLessLess = "≪";
  var NewLine = "\n";
  var nexist = "∄";
  var nexists = "∄";
  var Nfr = "𝔑";
  var nfr = "𝔫";
  var ngE = "≧̸";
  var nge = "≱";
  var ngeq = "≱";
  var ngeqq = "≧̸";
  var ngeqslant = "⩾̸";
  var nges = "⩾̸";
  var nGg = "⋙̸";
  var ngsim = "≵";
  var nGt = "≫⃒";
  var ngt = "≯";
  var ngtr = "≯";
  var nGtv = "≫̸";
  var nharr = "↮";
  var nhArr = "⇎";
  var nhpar = "⫲";
  var ni = "∋";
  var nis = "⋼";
  var nisd = "⋺";
  var niv = "∋";
  var NJcy = "Њ";
  var njcy = "њ";
  var nlarr = "↚";
  var nlArr = "⇍";
  var nldr = "‥";
  var nlE = "≦̸";
  var nle = "≰";
  var nleftarrow = "↚";
  var nLeftarrow = "⇍";
  var nleftrightarrow = "↮";
  var nLeftrightarrow = "⇎";
  var nleq = "≰";
  var nleqq = "≦̸";
  var nleqslant = "⩽̸";
  var nles = "⩽̸";
  var nless = "≮";
  var nLl = "⋘̸";
  var nlsim = "≴";
  var nLt = "≪⃒";
  var nlt = "≮";
  var nltri = "⋪";
  var nltrie = "⋬";
  var nLtv = "≪̸";
  var nmid = "∤";
  var NoBreak = "⁠";
  var NonBreakingSpace = " ";
  var nopf = "𝕟";
  var Nopf = "ℕ";
  var Not = "⫬";
  var not = "¬";
  var NotCongruent = "≢";
  var NotCupCap = "≭";
  var NotDoubleVerticalBar = "∦";
  var NotElement = "∉";
  var NotEqual = "≠";
  var NotEqualTilde = "≂̸";
  var NotExists = "∄";
  var NotGreater = "≯";
  var NotGreaterEqual = "≱";
  var NotGreaterFullEqual = "≧̸";
  var NotGreaterGreater = "≫̸";
  var NotGreaterLess = "≹";
  var NotGreaterSlantEqual = "⩾̸";
  var NotGreaterTilde = "≵";
  var NotHumpDownHump = "≎̸";
  var NotHumpEqual = "≏̸";
  var notin = "∉";
  var notindot = "⋵̸";
  var notinE = "⋹̸";
  var notinva = "∉";
  var notinvb = "⋷";
  var notinvc = "⋶";
  var NotLeftTriangleBar = "⧏̸";
  var NotLeftTriangle = "⋪";
  var NotLeftTriangleEqual = "⋬";
  var NotLess = "≮";
  var NotLessEqual = "≰";
  var NotLessGreater = "≸";
  var NotLessLess = "≪̸";
  var NotLessSlantEqual = "⩽̸";
  var NotLessTilde = "≴";
  var NotNestedGreaterGreater = "⪢̸";
  var NotNestedLessLess = "⪡̸";
  var notni = "∌";
  var notniva = "∌";
  var notnivb = "⋾";
  var notnivc = "⋽";
  var NotPrecedes = "⊀";
  var NotPrecedesEqual = "⪯̸";
  var NotPrecedesSlantEqual = "⋠";
  var NotReverseElement = "∌";
  var NotRightTriangleBar = "⧐̸";
  var NotRightTriangle = "⋫";
  var NotRightTriangleEqual = "⋭";
  var NotSquareSubset = "⊏̸";
  var NotSquareSubsetEqual = "⋢";
  var NotSquareSuperset = "⊐̸";
  var NotSquareSupersetEqual = "⋣";
  var NotSubset = "⊂⃒";
  var NotSubsetEqual = "⊈";
  var NotSucceeds = "⊁";
  var NotSucceedsEqual = "⪰̸";
  var NotSucceedsSlantEqual = "⋡";
  var NotSucceedsTilde = "≿̸";
  var NotSuperset = "⊃⃒";
  var NotSupersetEqual = "⊉";
  var NotTilde = "≁";
  var NotTildeEqual = "≄";
  var NotTildeFullEqual = "≇";
  var NotTildeTilde = "≉";
  var NotVerticalBar = "∤";
  var nparallel = "∦";
  var npar = "∦";
  var nparsl = "⫽⃥";
  var npart = "∂̸";
  var npolint = "⨔";
  var npr = "⊀";
  var nprcue = "⋠";
  var nprec = "⊀";
  var npreceq = "⪯̸";
  var npre = "⪯̸";
  var nrarrc = "⤳̸";
  var nrarr = "↛";
  var nrArr = "⇏";
  var nrarrw = "↝̸";
  var nrightarrow = "↛";
  var nRightarrow = "⇏";
  var nrtri = "⋫";
  var nrtrie = "⋭";
  var nsc = "⊁";
  var nsccue = "⋡";
  var nsce = "⪰̸";
  var Nscr = "𝒩";
  var nscr = "𝓃";
  var nshortmid = "∤";
  var nshortparallel = "∦";
  var nsim = "≁";
  var nsime = "≄";
  var nsimeq = "≄";
  var nsmid = "∤";
  var nspar = "∦";
  var nsqsube = "⋢";
  var nsqsupe = "⋣";
  var nsub = "⊄";
  var nsubE = "⫅̸";
  var nsube = "⊈";
  var nsubset = "⊂⃒";
  var nsubseteq = "⊈";
  var nsubseteqq = "⫅̸";
  var nsucc = "⊁";
  var nsucceq = "⪰̸";
  var nsup = "⊅";
  var nsupE = "⫆̸";
  var nsupe = "⊉";
  var nsupset = "⊃⃒";
  var nsupseteq = "⊉";
  var nsupseteqq = "⫆̸";
  var ntgl = "≹";
  var Ntilde = "Ñ";
  var ntilde = "ñ";
  var ntlg = "≸";
  var ntriangleleft = "⋪";
  var ntrianglelefteq = "⋬";
  var ntriangleright = "⋫";
  var ntrianglerighteq = "⋭";
  var Nu = "Ν";
  var nu = "ν";
  var num = "#";
  var numero = "№";
  var numsp = " ";
  var nvap = "≍⃒";
  var nvdash = "⊬";
  var nvDash = "⊭";
  var nVdash = "⊮";
  var nVDash = "⊯";
  var nvge = "≥⃒";
  var nvgt = ">⃒";
  var nvHarr = "⤄";
  var nvinfin = "⧞";
  var nvlArr = "⤂";
  var nvle = "≤⃒";
  var nvlt = "<⃒";
  var nvltrie = "⊴⃒";
  var nvrArr = "⤃";
  var nvrtrie = "⊵⃒";
  var nvsim = "∼⃒";
  var nwarhk = "⤣";
  var nwarr = "↖";
  var nwArr = "⇖";
  var nwarrow = "↖";
  var nwnear = "⤧";
  var Oacute = "Ó";
  var oacute = "ó";
  var oast = "⊛";
  var Ocirc = "Ô";
  var ocirc = "ô";
  var ocir = "⊚";
  var Ocy = "О";
  var ocy = "о";
  var odash = "⊝";
  var Odblac = "Ő";
  var odblac = "ő";
  var odiv = "⨸";
  var odot = "⊙";
  var odsold = "⦼";
  var OElig = "Œ";
  var oelig = "œ";
  var ofcir = "⦿";
  var Ofr = "𝔒";
  var ofr = "𝔬";
  var ogon = "˛";
  var Ograve = "Ò";
  var ograve = "ò";
  var ogt = "⧁";
  var ohbar = "⦵";
  var ohm = "Ω";
  var oint = "∮";
  var olarr = "↺";
  var olcir = "⦾";
  var olcross = "⦻";
  var oline = "‾";
  var olt = "⧀";
  var Omacr = "Ō";
  var omacr = "ō";
  var Omega = "Ω";
  var omega = "ω";
  var Omicron = "Ο";
  var omicron = "ο";
  var omid = "⦶";
  var ominus = "⊖";
  var Oopf = "𝕆";
  var oopf = "𝕠";
  var opar = "⦷";
  var OpenCurlyDoubleQuote = "“";
  var OpenCurlyQuote = "‘";
  var operp = "⦹";
  var oplus = "⊕";
  var orarr = "↻";
  var Or = "⩔";
  var or = "∨";
  var ord = "⩝";
  var order = "ℴ";
  var orderof = "ℴ";
  var ordf = "ª";
  var ordm = "º";
  var origof = "⊶";
  var oror = "⩖";
  var orslope = "⩗";
  var orv = "⩛";
  var oS = "Ⓢ";
  var Oscr = "𝒪";
  var oscr = "ℴ";
  var Oslash = "Ø";
  var oslash = "ø";
  var osol = "⊘";
  var Otilde = "Õ";
  var otilde = "õ";
  var otimesas = "⨶";
  var Otimes = "⨷";
  var otimes = "⊗";
  var Ouml = "Ö";
  var ouml = "ö";
  var ovbar = "⌽";
  var OverBar = "‾";
  var OverBrace = "⏞";
  var OverBracket = "⎴";
  var OverParenthesis = "⏜";
  var para = "¶";
  var parallel = "∥";
  var par = "∥";
  var parsim = "⫳";
  var parsl = "⫽";
  var part = "∂";
  var PartialD = "∂";
  var Pcy = "П";
  var pcy = "п";
  var percnt = "%";
  var period = ".";
  var permil = "‰";
  var perp = "⊥";
  var pertenk = "‱";
  var Pfr = "𝔓";
  var pfr = "𝔭";
  var Phi = "Φ";
  var phi = "φ";
  var phiv = "ϕ";
  var phmmat = "ℳ";
  var phone = "☎";
  var Pi = "Π";
  var pi = "π";
  var pitchfork = "⋔";
  var piv = "ϖ";
  var planck = "ℏ";
  var planckh = "ℎ";
  var plankv = "ℏ";
  var plusacir = "⨣";
  var plusb = "⊞";
  var pluscir = "⨢";
  var plus = "+";
  var plusdo = "∔";
  var plusdu = "⨥";
  var pluse = "⩲";
  var PlusMinus = "±";
  var plusmn = "±";
  var plussim = "⨦";
  var plustwo = "⨧";
  var pm = "±";
  var Poincareplane = "ℌ";
  var pointint = "⨕";
  var popf = "𝕡";
  var Popf = "ℙ";
  var pound = "£";
  var prap = "⪷";
  var Pr = "⪻";
  var pr = "≺";
  var prcue = "≼";
  var precapprox = "⪷";
  var prec = "≺";
  var preccurlyeq = "≼";
  var Precedes = "≺";
  var PrecedesEqual = "⪯";
  var PrecedesSlantEqual = "≼";
  var PrecedesTilde = "≾";
  var preceq = "⪯";
  var precnapprox = "⪹";
  var precneqq = "⪵";
  var precnsim = "⋨";
  var pre = "⪯";
  var prE = "⪳";
  var precsim = "≾";
  var prime = "′";
  var Prime = "″";
  var primes = "ℙ";
  var prnap = "⪹";
  var prnE = "⪵";
  var prnsim = "⋨";
  var prod = "∏";
  var Product = "∏";
  var profalar = "⌮";
  var profline = "⌒";
  var profsurf = "⌓";
  var prop = "∝";
  var Proportional = "∝";
  var Proportion = "∷";
  var propto = "∝";
  var prsim = "≾";
  var prurel = "⊰";
  var Pscr = "𝒫";
  var pscr = "𝓅";
  var Psi = "Ψ";
  var psi = "ψ";
  var puncsp = " ";
  var Qfr = "𝔔";
  var qfr = "𝔮";
  var qint = "⨌";
  var qopf = "𝕢";
  var Qopf = "ℚ";
  var qprime = "⁗";
  var Qscr = "𝒬";
  var qscr = "𝓆";
  var quaternions = "ℍ";
  var quatint = "⨖";
  var quest = "?";
  var questeq = "≟";
  var quot$1 = "\"";
  var QUOT = "\"";
  var rAarr = "⇛";
  var race = "∽̱";
  var Racute = "Ŕ";
  var racute = "ŕ";
  var radic = "√";
  var raemptyv = "⦳";
  var rang = "⟩";
  var Rang = "⟫";
  var rangd = "⦒";
  var range = "⦥";
  var rangle = "⟩";
  var raquo = "»";
  var rarrap = "⥵";
  var rarrb = "⇥";
  var rarrbfs = "⤠";
  var rarrc = "⤳";
  var rarr = "→";
  var Rarr = "↠";
  var rArr = "⇒";
  var rarrfs = "⤞";
  var rarrhk = "↪";
  var rarrlp = "↬";
  var rarrpl = "⥅";
  var rarrsim = "⥴";
  var Rarrtl = "⤖";
  var rarrtl = "↣";
  var rarrw = "↝";
  var ratail = "⤚";
  var rAtail = "⤜";
  var ratio = "∶";
  var rationals = "ℚ";
  var rbarr = "⤍";
  var rBarr = "⤏";
  var RBarr = "⤐";
  var rbbrk = "❳";
  var rbrace = "}";
  var rbrack = "]";
  var rbrke = "⦌";
  var rbrksld = "⦎";
  var rbrkslu = "⦐";
  var Rcaron = "Ř";
  var rcaron = "ř";
  var Rcedil = "Ŗ";
  var rcedil = "ŗ";
  var rceil = "⌉";
  var rcub = "}";
  var Rcy = "Р";
  var rcy = "р";
  var rdca = "⤷";
  var rdldhar = "⥩";
  var rdquo = "”";
  var rdquor = "”";
  var rdsh = "↳";
  var real = "ℜ";
  var realine = "ℛ";
  var realpart = "ℜ";
  var reals = "ℝ";
  var Re = "ℜ";
  var rect = "▭";
  var reg = "®";
  var REG = "®";
  var ReverseElement = "∋";
  var ReverseEquilibrium = "⇋";
  var ReverseUpEquilibrium = "⥯";
  var rfisht = "⥽";
  var rfloor = "⌋";
  var rfr = "𝔯";
  var Rfr = "ℜ";
  var rHar = "⥤";
  var rhard = "⇁";
  var rharu = "⇀";
  var rharul = "⥬";
  var Rho = "Ρ";
  var rho = "ρ";
  var rhov = "ϱ";
  var RightAngleBracket = "⟩";
  var RightArrowBar = "⇥";
  var rightarrow = "→";
  var RightArrow = "→";
  var Rightarrow = "⇒";
  var RightArrowLeftArrow = "⇄";
  var rightarrowtail = "↣";
  var RightCeiling = "⌉";
  var RightDoubleBracket = "⟧";
  var RightDownTeeVector = "⥝";
  var RightDownVectorBar = "⥕";
  var RightDownVector = "⇂";
  var RightFloor = "⌋";
  var rightharpoondown = "⇁";
  var rightharpoonup = "⇀";
  var rightleftarrows = "⇄";
  var rightleftharpoons = "⇌";
  var rightrightarrows = "⇉";
  var rightsquigarrow = "↝";
  var RightTeeArrow = "↦";
  var RightTee = "⊢";
  var RightTeeVector = "⥛";
  var rightthreetimes = "⋌";
  var RightTriangleBar = "⧐";
  var RightTriangle = "⊳";
  var RightTriangleEqual = "⊵";
  var RightUpDownVector = "⥏";
  var RightUpTeeVector = "⥜";
  var RightUpVectorBar = "⥔";
  var RightUpVector = "↾";
  var RightVectorBar = "⥓";
  var RightVector = "⇀";
  var ring = "˚";
  var risingdotseq = "≓";
  var rlarr = "⇄";
  var rlhar = "⇌";
  var rlm = "‏";
  var rmoustache = "⎱";
  var rmoust = "⎱";
  var rnmid = "⫮";
  var roang = "⟭";
  var roarr = "⇾";
  var robrk = "⟧";
  var ropar = "⦆";
  var ropf = "𝕣";
  var Ropf = "ℝ";
  var roplus = "⨮";
  var rotimes = "⨵";
  var RoundImplies = "⥰";
  var rpar = ")";
  var rpargt = "⦔";
  var rppolint = "⨒";
  var rrarr = "⇉";
  var Rrightarrow = "⇛";
  var rsaquo = "›";
  var rscr = "𝓇";
  var Rscr = "ℛ";
  var rsh = "↱";
  var Rsh = "↱";
  var rsqb = "]";
  var rsquo = "’";
  var rsquor = "’";
  var rthree = "⋌";
  var rtimes = "⋊";
  var rtri = "▹";
  var rtrie = "⊵";
  var rtrif = "▸";
  var rtriltri = "⧎";
  var RuleDelayed = "⧴";
  var ruluhar = "⥨";
  var rx = "℞";
  var Sacute = "Ś";
  var sacute = "ś";
  var sbquo = "‚";
  var scap = "⪸";
  var Scaron = "Š";
  var scaron = "š";
  var Sc = "⪼";
  var sc = "≻";
  var sccue = "≽";
  var sce = "⪰";
  var scE = "⪴";
  var Scedil = "Ş";
  var scedil = "ş";
  var Scirc = "Ŝ";
  var scirc = "ŝ";
  var scnap = "⪺";
  var scnE = "⪶";
  var scnsim = "⋩";
  var scpolint = "⨓";
  var scsim = "≿";
  var Scy = "С";
  var scy = "с";
  var sdotb = "⊡";
  var sdot = "⋅";
  var sdote = "⩦";
  var searhk = "⤥";
  var searr = "↘";
  var seArr = "⇘";
  var searrow = "↘";
  var sect = "§";
  var semi = ";";
  var seswar = "⤩";
  var setminus = "∖";
  var setmn = "∖";
  var sext = "✶";
  var Sfr = "𝔖";
  var sfr = "𝔰";
  var sfrown = "⌢";
  var sharp = "♯";
  var SHCHcy = "Щ";
  var shchcy = "щ";
  var SHcy = "Ш";
  var shcy = "ш";
  var ShortDownArrow = "↓";
  var ShortLeftArrow = "←";
  var shortmid = "∣";
  var shortparallel = "∥";
  var ShortRightArrow = "→";
  var ShortUpArrow = "↑";
  var shy = "­";
  var Sigma = "Σ";
  var sigma = "σ";
  var sigmaf = "ς";
  var sigmav = "ς";
  var sim = "∼";
  var simdot = "⩪";
  var sime = "≃";
  var simeq = "≃";
  var simg = "⪞";
  var simgE = "⪠";
  var siml = "⪝";
  var simlE = "⪟";
  var simne = "≆";
  var simplus = "⨤";
  var simrarr = "⥲";
  var slarr = "←";
  var SmallCircle = "∘";
  var smallsetminus = "∖";
  var smashp = "⨳";
  var smeparsl = "⧤";
  var smid = "∣";
  var smile = "⌣";
  var smt = "⪪";
  var smte = "⪬";
  var smtes = "⪬︀";
  var SOFTcy = "Ь";
  var softcy = "ь";
  var solbar = "⌿";
  var solb = "⧄";
  var sol = "/";
  var Sopf = "𝕊";
  var sopf = "𝕤";
  var spades = "♠";
  var spadesuit = "♠";
  var spar = "∥";
  var sqcap = "⊓";
  var sqcaps = "⊓︀";
  var sqcup = "⊔";
  var sqcups = "⊔︀";
  var Sqrt = "√";
  var sqsub = "⊏";
  var sqsube = "⊑";
  var sqsubset = "⊏";
  var sqsubseteq = "⊑";
  var sqsup = "⊐";
  var sqsupe = "⊒";
  var sqsupset = "⊐";
  var sqsupseteq = "⊒";
  var square = "□";
  var Square = "□";
  var SquareIntersection = "⊓";
  var SquareSubset = "⊏";
  var SquareSubsetEqual = "⊑";
  var SquareSuperset = "⊐";
  var SquareSupersetEqual = "⊒";
  var SquareUnion = "⊔";
  var squarf = "▪";
  var squ = "□";
  var squf = "▪";
  var srarr = "→";
  var Sscr = "𝒮";
  var sscr = "𝓈";
  var ssetmn = "∖";
  var ssmile = "⌣";
  var sstarf = "⋆";
  var Star = "⋆";
  var star = "☆";
  var starf = "★";
  var straightepsilon = "ϵ";
  var straightphi = "ϕ";
  var strns = "¯";
  var sub = "⊂";
  var Sub = "⋐";
  var subdot = "⪽";
  var subE = "⫅";
  var sube = "⊆";
  var subedot = "⫃";
  var submult = "⫁";
  var subnE = "⫋";
  var subne = "⊊";
  var subplus = "⪿";
  var subrarr = "⥹";
  var subset = "⊂";
  var Subset = "⋐";
  var subseteq = "⊆";
  var subseteqq = "⫅";
  var SubsetEqual = "⊆";
  var subsetneq = "⊊";
  var subsetneqq = "⫋";
  var subsim = "⫇";
  var subsub = "⫕";
  var subsup = "⫓";
  var succapprox = "⪸";
  var succ = "≻";
  var succcurlyeq = "≽";
  var Succeeds = "≻";
  var SucceedsEqual = "⪰";
  var SucceedsSlantEqual = "≽";
  var SucceedsTilde = "≿";
  var succeq = "⪰";
  var succnapprox = "⪺";
  var succneqq = "⪶";
  var succnsim = "⋩";
  var succsim = "≿";
  var SuchThat = "∋";
  var sum = "∑";
  var Sum = "∑";
  var sung = "♪";
  var sup1 = "¹";
  var sup2 = "²";
  var sup3 = "³";
  var sup = "⊃";
  var Sup = "⋑";
  var supdot = "⪾";
  var supdsub = "⫘";
  var supE = "⫆";
  var supe = "⊇";
  var supedot = "⫄";
  var Superset = "⊃";
  var SupersetEqual = "⊇";
  var suphsol = "⟉";
  var suphsub = "⫗";
  var suplarr = "⥻";
  var supmult = "⫂";
  var supnE = "⫌";
  var supne = "⊋";
  var supplus = "⫀";
  var supset = "⊃";
  var Supset = "⋑";
  var supseteq = "⊇";
  var supseteqq = "⫆";
  var supsetneq = "⊋";
  var supsetneqq = "⫌";
  var supsim = "⫈";
  var supsub = "⫔";
  var supsup = "⫖";
  var swarhk = "⤦";
  var swarr = "↙";
  var swArr = "⇙";
  var swarrow = "↙";
  var swnwar = "⤪";
  var szlig = "ß";
  var Tab = "\t";
  var target = "⌖";
  var Tau = "Τ";
  var tau = "τ";
  var tbrk = "⎴";
  var Tcaron = "Ť";
  var tcaron = "ť";
  var Tcedil = "Ţ";
  var tcedil = "ţ";
  var Tcy = "Т";
  var tcy = "т";
  var tdot = "⃛";
  var telrec = "⌕";
  var Tfr = "𝔗";
  var tfr = "𝔱";
  var there4 = "∴";
  var therefore = "∴";
  var Therefore = "∴";
  var Theta = "Θ";
  var theta = "θ";
  var thetasym = "ϑ";
  var thetav = "ϑ";
  var thickapprox = "≈";
  var thicksim = "∼";
  var ThickSpace = "  ";
  var ThinSpace = " ";
  var thinsp = " ";
  var thkap = "≈";
  var thksim = "∼";
  var THORN = "Þ";
  var thorn = "þ";
  var tilde = "˜";
  var Tilde = "∼";
  var TildeEqual = "≃";
  var TildeFullEqual = "≅";
  var TildeTilde = "≈";
  var timesbar = "⨱";
  var timesb = "⊠";
  var times = "×";
  var timesd = "⨰";
  var tint = "∭";
  var toea = "⤨";
  var topbot = "⌶";
  var topcir = "⫱";
  var top = "⊤";
  var Topf = "𝕋";
  var topf = "𝕥";
  var topfork = "⫚";
  var tosa = "⤩";
  var tprime = "‴";
  var trade = "™";
  var TRADE = "™";
  var triangle = "▵";
  var triangledown = "▿";
  var triangleleft = "◃";
  var trianglelefteq = "⊴";
  var triangleq = "≜";
  var triangleright = "▹";
  var trianglerighteq = "⊵";
  var tridot = "◬";
  var trie = "≜";
  var triminus = "⨺";
  var TripleDot = "⃛";
  var triplus = "⨹";
  var trisb = "⧍";
  var tritime = "⨻";
  var trpezium = "⏢";
  var Tscr = "𝒯";
  var tscr = "𝓉";
  var TScy = "Ц";
  var tscy = "ц";
  var TSHcy = "Ћ";
  var tshcy = "ћ";
  var Tstrok = "Ŧ";
  var tstrok = "ŧ";
  var twixt = "≬";
  var twoheadleftarrow = "↞";
  var twoheadrightarrow = "↠";
  var Uacute = "Ú";
  var uacute = "ú";
  var uarr = "↑";
  var Uarr = "↟";
  var uArr = "⇑";
  var Uarrocir = "⥉";
  var Ubrcy = "Ў";
  var ubrcy = "ў";
  var Ubreve = "Ŭ";
  var ubreve = "ŭ";
  var Ucirc = "Û";
  var ucirc = "û";
  var Ucy = "У";
  var ucy = "у";
  var udarr = "⇅";
  var Udblac = "Ű";
  var udblac = "ű";
  var udhar = "⥮";
  var ufisht = "⥾";
  var Ufr = "𝔘";
  var ufr = "𝔲";
  var Ugrave = "Ù";
  var ugrave = "ù";
  var uHar = "⥣";
  var uharl = "↿";
  var uharr = "↾";
  var uhblk = "▀";
  var ulcorn = "⌜";
  var ulcorner = "⌜";
  var ulcrop = "⌏";
  var ultri = "◸";
  var Umacr = "Ū";
  var umacr = "ū";
  var uml = "¨";
  var UnderBar = "_";
  var UnderBrace = "⏟";
  var UnderBracket = "⎵";
  var UnderParenthesis = "⏝";
  var Union = "⋃";
  var UnionPlus = "⊎";
  var Uogon = "Ų";
  var uogon = "ų";
  var Uopf = "𝕌";
  var uopf = "𝕦";
  var UpArrowBar = "⤒";
  var uparrow = "↑";
  var UpArrow = "↑";
  var Uparrow = "⇑";
  var UpArrowDownArrow = "⇅";
  var updownarrow = "↕";
  var UpDownArrow = "↕";
  var Updownarrow = "⇕";
  var UpEquilibrium = "⥮";
  var upharpoonleft = "↿";
  var upharpoonright = "↾";
  var uplus = "⊎";
  var UpperLeftArrow = "↖";
  var UpperRightArrow = "↗";
  var upsi = "υ";
  var Upsi = "ϒ";
  var upsih = "ϒ";
  var Upsilon = "Υ";
  var upsilon = "υ";
  var UpTeeArrow = "↥";
  var UpTee = "⊥";
  var upuparrows = "⇈";
  var urcorn = "⌝";
  var urcorner = "⌝";
  var urcrop = "⌎";
  var Uring = "Ů";
  var uring = "ů";
  var urtri = "◹";
  var Uscr = "𝒰";
  var uscr = "𝓊";
  var utdot = "⋰";
  var Utilde = "Ũ";
  var utilde = "ũ";
  var utri = "▵";
  var utrif = "▴";
  var uuarr = "⇈";
  var Uuml = "Ü";
  var uuml = "ü";
  var uwangle = "⦧";
  var vangrt = "⦜";
  var varepsilon = "ϵ";
  var varkappa = "ϰ";
  var varnothing = "∅";
  var varphi = "ϕ";
  var varpi = "ϖ";
  var varpropto = "∝";
  var varr = "↕";
  var vArr = "⇕";
  var varrho = "ϱ";
  var varsigma = "ς";
  var varsubsetneq = "⊊︀";
  var varsubsetneqq = "⫋︀";
  var varsupsetneq = "⊋︀";
  var varsupsetneqq = "⫌︀";
  var vartheta = "ϑ";
  var vartriangleleft = "⊲";
  var vartriangleright = "⊳";
  var vBar = "⫨";
  var Vbar = "⫫";
  var vBarv = "⫩";
  var Vcy = "В";
  var vcy = "в";
  var vdash = "⊢";
  var vDash = "⊨";
  var Vdash = "⊩";
  var VDash = "⊫";
  var Vdashl = "⫦";
  var veebar = "⊻";
  var vee = "∨";
  var Vee = "⋁";
  var veeeq = "≚";
  var vellip = "⋮";
  var verbar = "|";
  var Verbar = "‖";
  var vert = "|";
  var Vert = "‖";
  var VerticalBar = "∣";
  var VerticalLine = "|";
  var VerticalSeparator = "❘";
  var VerticalTilde = "≀";
  var VeryThinSpace = " ";
  var Vfr = "𝔙";
  var vfr = "𝔳";
  var vltri = "⊲";
  var vnsub = "⊂⃒";
  var vnsup = "⊃⃒";
  var Vopf = "𝕍";
  var vopf = "𝕧";
  var vprop = "∝";
  var vrtri = "⊳";
  var Vscr = "𝒱";
  var vscr = "𝓋";
  var vsubnE = "⫋︀";
  var vsubne = "⊊︀";
  var vsupnE = "⫌︀";
  var vsupne = "⊋︀";
  var Vvdash = "⊪";
  var vzigzag = "⦚";
  var Wcirc = "Ŵ";
  var wcirc = "ŵ";
  var wedbar = "⩟";
  var wedge = "∧";
  var Wedge = "⋀";
  var wedgeq = "≙";
  var weierp = "℘";
  var Wfr = "𝔚";
  var wfr = "𝔴";
  var Wopf = "𝕎";
  var wopf = "𝕨";
  var wp = "℘";
  var wr = "≀";
  var wreath = "≀";
  var Wscr = "𝒲";
  var wscr = "𝓌";
  var xcap = "⋂";
  var xcirc = "◯";
  var xcup = "⋃";
  var xdtri = "▽";
  var Xfr = "𝔛";
  var xfr = "𝔵";
  var xharr = "⟷";
  var xhArr = "⟺";
  var Xi = "Ξ";
  var xi = "ξ";
  var xlarr = "⟵";
  var xlArr = "⟸";
  var xmap = "⟼";
  var xnis = "⋻";
  var xodot = "⨀";
  var Xopf = "𝕏";
  var xopf = "𝕩";
  var xoplus = "⨁";
  var xotime = "⨂";
  var xrarr = "⟶";
  var xrArr = "⟹";
  var Xscr = "𝒳";
  var xscr = "𝓍";
  var xsqcup = "⨆";
  var xuplus = "⨄";
  var xutri = "△";
  var xvee = "⋁";
  var xwedge = "⋀";
  var Yacute = "Ý";
  var yacute = "ý";
  var YAcy = "Я";
  var yacy = "я";
  var Ycirc = "Ŷ";
  var ycirc = "ŷ";
  var Ycy = "Ы";
  var ycy = "ы";
  var yen = "¥";
  var Yfr = "𝔜";
  var yfr = "𝔶";
  var YIcy = "Ї";
  var yicy = "ї";
  var Yopf = "𝕐";
  var yopf = "𝕪";
  var Yscr = "𝒴";
  var yscr = "𝓎";
  var YUcy = "Ю";
  var yucy = "ю";
  var yuml = "ÿ";
  var Yuml = "Ÿ";
  var Zacute = "Ź";
  var zacute = "ź";
  var Zcaron = "Ž";
  var zcaron = "ž";
  var Zcy = "З";
  var zcy = "з";
  var Zdot = "Ż";
  var zdot = "ż";
  var zeetrf = "ℨ";
  var ZeroWidthSpace = "​";
  var Zeta = "Ζ";
  var zeta = "ζ";
  var zfr = "𝔷";
  var Zfr = "ℨ";
  var ZHcy = "Ж";
  var zhcy = "ж";
  var zigrarr = "⇝";
  var zopf = "𝕫";
  var Zopf = "ℤ";
  var Zscr = "𝒵";
  var zscr = "𝓏";
  var zwj = "‍";
  var zwnj = "‌";
  var entitiesJSON = {
  	Aacute: Aacute,
  	aacute: aacute,
  	Abreve: Abreve,
  	abreve: abreve,
  	ac: ac,
  	acd: acd,
  	acE: acE,
  	Acirc: Acirc,
  	acirc: acirc,
  	acute: acute,
  	Acy: Acy,
  	acy: acy,
  	AElig: AElig,
  	aelig: aelig,
  	af: af,
  	Afr: Afr,
  	afr: afr,
  	Agrave: Agrave,
  	agrave: agrave,
  	alefsym: alefsym,
  	aleph: aleph,
  	Alpha: Alpha,
  	alpha: alpha,
  	Amacr: Amacr,
  	amacr: amacr,
  	amalg: amalg,
  	amp: amp$1,
  	AMP: AMP,
  	andand: andand,
  	And: And,
  	and: and,
  	andd: andd,
  	andslope: andslope,
  	andv: andv,
  	ang: ang,
  	ange: ange,
  	angle: angle,
  	angmsdaa: angmsdaa,
  	angmsdab: angmsdab,
  	angmsdac: angmsdac,
  	angmsdad: angmsdad,
  	angmsdae: angmsdae,
  	angmsdaf: angmsdaf,
  	angmsdag: angmsdag,
  	angmsdah: angmsdah,
  	angmsd: angmsd,
  	angrt: angrt,
  	angrtvb: angrtvb,
  	angrtvbd: angrtvbd,
  	angsph: angsph,
  	angst: angst,
  	angzarr: angzarr,
  	Aogon: Aogon,
  	aogon: aogon,
  	Aopf: Aopf,
  	aopf: aopf,
  	apacir: apacir,
  	ap: ap,
  	apE: apE,
  	ape: ape,
  	apid: apid,
  	apos: apos$1,
  	ApplyFunction: ApplyFunction,
  	approx: approx,
  	approxeq: approxeq,
  	Aring: Aring,
  	aring: aring,
  	Ascr: Ascr,
  	ascr: ascr,
  	Assign: Assign,
  	ast: ast,
  	asymp: asymp,
  	asympeq: asympeq,
  	Atilde: Atilde,
  	atilde: atilde,
  	Auml: Auml,
  	auml: auml,
  	awconint: awconint,
  	awint: awint,
  	backcong: backcong,
  	backepsilon: backepsilon,
  	backprime: backprime,
  	backsim: backsim,
  	backsimeq: backsimeq,
  	Backslash: Backslash,
  	Barv: Barv,
  	barvee: barvee,
  	barwed: barwed,
  	Barwed: Barwed,
  	barwedge: barwedge,
  	bbrk: bbrk,
  	bbrktbrk: bbrktbrk,
  	bcong: bcong,
  	Bcy: Bcy,
  	bcy: bcy,
  	bdquo: bdquo,
  	becaus: becaus,
  	because: because,
  	Because: Because,
  	bemptyv: bemptyv,
  	bepsi: bepsi,
  	bernou: bernou,
  	Bernoullis: Bernoullis,
  	Beta: Beta,
  	beta: beta,
  	beth: beth,
  	between: between,
  	Bfr: Bfr,
  	bfr: bfr,
  	bigcap: bigcap,
  	bigcirc: bigcirc,
  	bigcup: bigcup,
  	bigodot: bigodot,
  	bigoplus: bigoplus,
  	bigotimes: bigotimes,
  	bigsqcup: bigsqcup,
  	bigstar: bigstar,
  	bigtriangledown: bigtriangledown,
  	bigtriangleup: bigtriangleup,
  	biguplus: biguplus,
  	bigvee: bigvee,
  	bigwedge: bigwedge,
  	bkarow: bkarow,
  	blacklozenge: blacklozenge,
  	blacksquare: blacksquare,
  	blacktriangle: blacktriangle,
  	blacktriangledown: blacktriangledown,
  	blacktriangleleft: blacktriangleleft,
  	blacktriangleright: blacktriangleright,
  	blank: blank,
  	blk12: blk12,
  	blk14: blk14,
  	blk34: blk34,
  	block: block,
  	bne: bne,
  	bnequiv: bnequiv,
  	bNot: bNot,
  	bnot: bnot,
  	Bopf: Bopf,
  	bopf: bopf,
  	bot: bot,
  	bottom: bottom,
  	bowtie: bowtie,
  	boxbox: boxbox,
  	boxdl: boxdl,
  	boxdL: boxdL,
  	boxDl: boxDl,
  	boxDL: boxDL,
  	boxdr: boxdr,
  	boxdR: boxdR,
  	boxDr: boxDr,
  	boxDR: boxDR,
  	boxh: boxh,
  	boxH: boxH,
  	boxhd: boxhd,
  	boxHd: boxHd,
  	boxhD: boxhD,
  	boxHD: boxHD,
  	boxhu: boxhu,
  	boxHu: boxHu,
  	boxhU: boxhU,
  	boxHU: boxHU,
  	boxminus: boxminus,
  	boxplus: boxplus,
  	boxtimes: boxtimes,
  	boxul: boxul,
  	boxuL: boxuL,
  	boxUl: boxUl,
  	boxUL: boxUL,
  	boxur: boxur,
  	boxuR: boxuR,
  	boxUr: boxUr,
  	boxUR: boxUR,
  	boxv: boxv,
  	boxV: boxV,
  	boxvh: boxvh,
  	boxvH: boxvH,
  	boxVh: boxVh,
  	boxVH: boxVH,
  	boxvl: boxvl,
  	boxvL: boxvL,
  	boxVl: boxVl,
  	boxVL: boxVL,
  	boxvr: boxvr,
  	boxvR: boxvR,
  	boxVr: boxVr,
  	boxVR: boxVR,
  	bprime: bprime,
  	breve: breve,
  	Breve: Breve,
  	brvbar: brvbar,
  	bscr: bscr,
  	Bscr: Bscr,
  	bsemi: bsemi,
  	bsim: bsim,
  	bsime: bsime,
  	bsolb: bsolb,
  	bsol: bsol,
  	bsolhsub: bsolhsub,
  	bull: bull,
  	bullet: bullet,
  	bump: bump,
  	bumpE: bumpE,
  	bumpe: bumpe,
  	Bumpeq: Bumpeq,
  	bumpeq: bumpeq,
  	Cacute: Cacute,
  	cacute: cacute,
  	capand: capand,
  	capbrcup: capbrcup,
  	capcap: capcap,
  	cap: cap,
  	Cap: Cap,
  	capcup: capcup,
  	capdot: capdot,
  	CapitalDifferentialD: CapitalDifferentialD,
  	caps: caps,
  	caret: caret,
  	caron: caron,
  	Cayleys: Cayleys,
  	ccaps: ccaps,
  	Ccaron: Ccaron,
  	ccaron: ccaron,
  	Ccedil: Ccedil,
  	ccedil: ccedil,
  	Ccirc: Ccirc,
  	ccirc: ccirc,
  	Cconint: Cconint,
  	ccups: ccups,
  	ccupssm: ccupssm,
  	Cdot: Cdot,
  	cdot: cdot,
  	cedil: cedil,
  	Cedilla: Cedilla,
  	cemptyv: cemptyv,
  	cent: cent,
  	centerdot: centerdot,
  	CenterDot: CenterDot,
  	cfr: cfr,
  	Cfr: Cfr,
  	CHcy: CHcy,
  	chcy: chcy,
  	check: check,
  	checkmark: checkmark,
  	Chi: Chi,
  	chi: chi,
  	circ: circ,
  	circeq: circeq,
  	circlearrowleft: circlearrowleft,
  	circlearrowright: circlearrowright,
  	circledast: circledast,
  	circledcirc: circledcirc,
  	circleddash: circleddash,
  	CircleDot: CircleDot,
  	circledR: circledR,
  	circledS: circledS,
  	CircleMinus: CircleMinus,
  	CirclePlus: CirclePlus,
  	CircleTimes: CircleTimes,
  	cir: cir,
  	cirE: cirE,
  	cire: cire,
  	cirfnint: cirfnint,
  	cirmid: cirmid,
  	cirscir: cirscir,
  	ClockwiseContourIntegral: ClockwiseContourIntegral,
  	CloseCurlyDoubleQuote: CloseCurlyDoubleQuote,
  	CloseCurlyQuote: CloseCurlyQuote,
  	clubs: clubs,
  	clubsuit: clubsuit,
  	colon: colon,
  	Colon: Colon,
  	Colone: Colone,
  	colone: colone,
  	coloneq: coloneq,
  	comma: comma,
  	commat: commat,
  	comp: comp,
  	compfn: compfn,
  	complement: complement,
  	complexes: complexes,
  	cong: cong,
  	congdot: congdot,
  	Congruent: Congruent,
  	conint: conint,
  	Conint: Conint,
  	ContourIntegral: ContourIntegral,
  	copf: copf,
  	Copf: Copf,
  	coprod: coprod,
  	Coproduct: Coproduct,
  	copy: copy,
  	COPY: COPY,
  	copysr: copysr,
  	CounterClockwiseContourIntegral: CounterClockwiseContourIntegral,
  	crarr: crarr,
  	cross: cross,
  	Cross: Cross,
  	Cscr: Cscr,
  	cscr: cscr,
  	csub: csub,
  	csube: csube,
  	csup: csup,
  	csupe: csupe,
  	ctdot: ctdot,
  	cudarrl: cudarrl,
  	cudarrr: cudarrr,
  	cuepr: cuepr,
  	cuesc: cuesc,
  	cularr: cularr,
  	cularrp: cularrp,
  	cupbrcap: cupbrcap,
  	cupcap: cupcap,
  	CupCap: CupCap,
  	cup: cup,
  	Cup: Cup,
  	cupcup: cupcup,
  	cupdot: cupdot,
  	cupor: cupor,
  	cups: cups,
  	curarr: curarr,
  	curarrm: curarrm,
  	curlyeqprec: curlyeqprec,
  	curlyeqsucc: curlyeqsucc,
  	curlyvee: curlyvee,
  	curlywedge: curlywedge,
  	curren: curren,
  	curvearrowleft: curvearrowleft,
  	curvearrowright: curvearrowright,
  	cuvee: cuvee,
  	cuwed: cuwed,
  	cwconint: cwconint,
  	cwint: cwint,
  	cylcty: cylcty,
  	dagger: dagger,
  	Dagger: Dagger,
  	daleth: daleth,
  	darr: darr,
  	Darr: Darr,
  	dArr: dArr,
  	dash: dash,
  	Dashv: Dashv,
  	dashv: dashv,
  	dbkarow: dbkarow,
  	dblac: dblac,
  	Dcaron: Dcaron,
  	dcaron: dcaron,
  	Dcy: Dcy,
  	dcy: dcy,
  	ddagger: ddagger,
  	ddarr: ddarr,
  	DD: DD,
  	dd: dd,
  	DDotrahd: DDotrahd,
  	ddotseq: ddotseq,
  	deg: deg,
  	Del: Del,
  	Delta: Delta,
  	delta: delta,
  	demptyv: demptyv,
  	dfisht: dfisht,
  	Dfr: Dfr,
  	dfr: dfr,
  	dHar: dHar,
  	dharl: dharl,
  	dharr: dharr,
  	DiacriticalAcute: DiacriticalAcute,
  	DiacriticalDot: DiacriticalDot,
  	DiacriticalDoubleAcute: DiacriticalDoubleAcute,
  	DiacriticalGrave: DiacriticalGrave,
  	DiacriticalTilde: DiacriticalTilde,
  	diam: diam,
  	diamond: diamond,
  	Diamond: Diamond,
  	diamondsuit: diamondsuit,
  	diams: diams,
  	die: die,
  	DifferentialD: DifferentialD,
  	digamma: digamma,
  	disin: disin,
  	div: div,
  	divide: divide,
  	divideontimes: divideontimes,
  	divonx: divonx,
  	DJcy: DJcy,
  	djcy: djcy,
  	dlcorn: dlcorn,
  	dlcrop: dlcrop,
  	dollar: dollar,
  	Dopf: Dopf,
  	dopf: dopf,
  	Dot: Dot,
  	dot: dot,
  	DotDot: DotDot,
  	doteq: doteq,
  	doteqdot: doteqdot,
  	DotEqual: DotEqual,
  	dotminus: dotminus,
  	dotplus: dotplus,
  	dotsquare: dotsquare,
  	doublebarwedge: doublebarwedge,
  	DoubleContourIntegral: DoubleContourIntegral,
  	DoubleDot: DoubleDot,
  	DoubleDownArrow: DoubleDownArrow,
  	DoubleLeftArrow: DoubleLeftArrow,
  	DoubleLeftRightArrow: DoubleLeftRightArrow,
  	DoubleLeftTee: DoubleLeftTee,
  	DoubleLongLeftArrow: DoubleLongLeftArrow,
  	DoubleLongLeftRightArrow: DoubleLongLeftRightArrow,
  	DoubleLongRightArrow: DoubleLongRightArrow,
  	DoubleRightArrow: DoubleRightArrow,
  	DoubleRightTee: DoubleRightTee,
  	DoubleUpArrow: DoubleUpArrow,
  	DoubleUpDownArrow: DoubleUpDownArrow,
  	DoubleVerticalBar: DoubleVerticalBar,
  	DownArrowBar: DownArrowBar,
  	downarrow: downarrow,
  	DownArrow: DownArrow,
  	Downarrow: Downarrow,
  	DownArrowUpArrow: DownArrowUpArrow,
  	DownBreve: DownBreve,
  	downdownarrows: downdownarrows,
  	downharpoonleft: downharpoonleft,
  	downharpoonright: downharpoonright,
  	DownLeftRightVector: DownLeftRightVector,
  	DownLeftTeeVector: DownLeftTeeVector,
  	DownLeftVectorBar: DownLeftVectorBar,
  	DownLeftVector: DownLeftVector,
  	DownRightTeeVector: DownRightTeeVector,
  	DownRightVectorBar: DownRightVectorBar,
  	DownRightVector: DownRightVector,
  	DownTeeArrow: DownTeeArrow,
  	DownTee: DownTee,
  	drbkarow: drbkarow,
  	drcorn: drcorn,
  	drcrop: drcrop,
  	Dscr: Dscr,
  	dscr: dscr,
  	DScy: DScy,
  	dscy: dscy,
  	dsol: dsol,
  	Dstrok: Dstrok,
  	dstrok: dstrok,
  	dtdot: dtdot,
  	dtri: dtri,
  	dtrif: dtrif,
  	duarr: duarr,
  	duhar: duhar,
  	dwangle: dwangle,
  	DZcy: DZcy,
  	dzcy: dzcy,
  	dzigrarr: dzigrarr,
  	Eacute: Eacute,
  	eacute: eacute,
  	easter: easter,
  	Ecaron: Ecaron,
  	ecaron: ecaron,
  	Ecirc: Ecirc,
  	ecirc: ecirc,
  	ecir: ecir,
  	ecolon: ecolon,
  	Ecy: Ecy,
  	ecy: ecy,
  	eDDot: eDDot,
  	Edot: Edot,
  	edot: edot,
  	eDot: eDot,
  	ee: ee,
  	efDot: efDot,
  	Efr: Efr,
  	efr: efr,
  	eg: eg,
  	Egrave: Egrave,
  	egrave: egrave,
  	egs: egs,
  	egsdot: egsdot,
  	el: el,
  	Element: Element,
  	elinters: elinters,
  	ell: ell,
  	els: els,
  	elsdot: elsdot,
  	Emacr: Emacr,
  	emacr: emacr,
  	empty: empty,
  	emptyset: emptyset,
  	EmptySmallSquare: EmptySmallSquare,
  	emptyv: emptyv,
  	EmptyVerySmallSquare: EmptyVerySmallSquare,
  	emsp13: emsp13,
  	emsp14: emsp14,
  	emsp: emsp,
  	ENG: ENG,
  	eng: eng,
  	ensp: ensp,
  	Eogon: Eogon,
  	eogon: eogon,
  	Eopf: Eopf,
  	eopf: eopf,
  	epar: epar,
  	eparsl: eparsl,
  	eplus: eplus,
  	epsi: epsi,
  	Epsilon: Epsilon,
  	epsilon: epsilon,
  	epsiv: epsiv,
  	eqcirc: eqcirc,
  	eqcolon: eqcolon,
  	eqsim: eqsim,
  	eqslantgtr: eqslantgtr,
  	eqslantless: eqslantless,
  	Equal: Equal,
  	equals: equals,
  	EqualTilde: EqualTilde,
  	equest: equest,
  	Equilibrium: Equilibrium,
  	equiv: equiv,
  	equivDD: equivDD,
  	eqvparsl: eqvparsl,
  	erarr: erarr,
  	erDot: erDot,
  	escr: escr,
  	Escr: Escr,
  	esdot: esdot,
  	Esim: Esim,
  	esim: esim,
  	Eta: Eta,
  	eta: eta,
  	ETH: ETH,
  	eth: eth,
  	Euml: Euml,
  	euml: euml,
  	euro: euro,
  	excl: excl,
  	exist: exist,
  	Exists: Exists,
  	expectation: expectation,
  	exponentiale: exponentiale,
  	ExponentialE: ExponentialE,
  	fallingdotseq: fallingdotseq,
  	Fcy: Fcy,
  	fcy: fcy,
  	female: female,
  	ffilig: ffilig,
  	fflig: fflig,
  	ffllig: ffllig,
  	Ffr: Ffr,
  	ffr: ffr,
  	filig: filig,
  	FilledSmallSquare: FilledSmallSquare,
  	FilledVerySmallSquare: FilledVerySmallSquare,
  	fjlig: fjlig,
  	flat: flat,
  	fllig: fllig,
  	fltns: fltns,
  	fnof: fnof,
  	Fopf: Fopf,
  	fopf: fopf,
  	forall: forall,
  	ForAll: ForAll,
  	fork: fork,
  	forkv: forkv,
  	Fouriertrf: Fouriertrf,
  	fpartint: fpartint,
  	frac12: frac12,
  	frac13: frac13,
  	frac14: frac14,
  	frac15: frac15,
  	frac16: frac16,
  	frac18: frac18,
  	frac23: frac23,
  	frac25: frac25,
  	frac34: frac34,
  	frac35: frac35,
  	frac38: frac38,
  	frac45: frac45,
  	frac56: frac56,
  	frac58: frac58,
  	frac78: frac78,
  	frasl: frasl,
  	frown: frown,
  	fscr: fscr,
  	Fscr: Fscr,
  	gacute: gacute,
  	Gamma: Gamma,
  	gamma: gamma,
  	Gammad: Gammad,
  	gammad: gammad,
  	gap: gap,
  	Gbreve: Gbreve,
  	gbreve: gbreve,
  	Gcedil: Gcedil,
  	Gcirc: Gcirc,
  	gcirc: gcirc,
  	Gcy: Gcy,
  	gcy: gcy,
  	Gdot: Gdot,
  	gdot: gdot,
  	ge: ge,
  	gE: gE,
  	gEl: gEl,
  	gel: gel,
  	geq: geq,
  	geqq: geqq,
  	geqslant: geqslant,
  	gescc: gescc,
  	ges: ges,
  	gesdot: gesdot,
  	gesdoto: gesdoto,
  	gesdotol: gesdotol,
  	gesl: gesl,
  	gesles: gesles,
  	Gfr: Gfr,
  	gfr: gfr,
  	gg: gg,
  	Gg: Gg,
  	ggg: ggg,
  	gimel: gimel,
  	GJcy: GJcy,
  	gjcy: gjcy,
  	gla: gla,
  	gl: gl,
  	glE: glE,
  	glj: glj,
  	gnap: gnap,
  	gnapprox: gnapprox,
  	gne: gne,
  	gnE: gnE,
  	gneq: gneq,
  	gneqq: gneqq,
  	gnsim: gnsim,
  	Gopf: Gopf,
  	gopf: gopf,
  	grave: grave,
  	GreaterEqual: GreaterEqual,
  	GreaterEqualLess: GreaterEqualLess,
  	GreaterFullEqual: GreaterFullEqual,
  	GreaterGreater: GreaterGreater,
  	GreaterLess: GreaterLess,
  	GreaterSlantEqual: GreaterSlantEqual,
  	GreaterTilde: GreaterTilde,
  	Gscr: Gscr,
  	gscr: gscr,
  	gsim: gsim,
  	gsime: gsime,
  	gsiml: gsiml,
  	gtcc: gtcc,
  	gtcir: gtcir,
  	gt: gt$1,
  	GT: GT,
  	Gt: Gt,
  	gtdot: gtdot,
  	gtlPar: gtlPar,
  	gtquest: gtquest,
  	gtrapprox: gtrapprox,
  	gtrarr: gtrarr,
  	gtrdot: gtrdot,
  	gtreqless: gtreqless,
  	gtreqqless: gtreqqless,
  	gtrless: gtrless,
  	gtrsim: gtrsim,
  	gvertneqq: gvertneqq,
  	gvnE: gvnE,
  	Hacek: Hacek,
  	hairsp: hairsp,
  	half: half,
  	hamilt: hamilt,
  	HARDcy: HARDcy,
  	hardcy: hardcy,
  	harrcir: harrcir,
  	harr: harr,
  	hArr: hArr,
  	harrw: harrw,
  	Hat: Hat,
  	hbar: hbar,
  	Hcirc: Hcirc,
  	hcirc: hcirc,
  	hearts: hearts,
  	heartsuit: heartsuit,
  	hellip: hellip,
  	hercon: hercon,
  	hfr: hfr,
  	Hfr: Hfr,
  	HilbertSpace: HilbertSpace,
  	hksearow: hksearow,
  	hkswarow: hkswarow,
  	hoarr: hoarr,
  	homtht: homtht,
  	hookleftarrow: hookleftarrow,
  	hookrightarrow: hookrightarrow,
  	hopf: hopf,
  	Hopf: Hopf,
  	horbar: horbar,
  	HorizontalLine: HorizontalLine,
  	hscr: hscr,
  	Hscr: Hscr,
  	hslash: hslash,
  	Hstrok: Hstrok,
  	hstrok: hstrok,
  	HumpDownHump: HumpDownHump,
  	HumpEqual: HumpEqual,
  	hybull: hybull,
  	hyphen: hyphen,
  	Iacute: Iacute,
  	iacute: iacute,
  	ic: ic,
  	Icirc: Icirc,
  	icirc: icirc,
  	Icy: Icy,
  	icy: icy,
  	Idot: Idot,
  	IEcy: IEcy,
  	iecy: iecy,
  	iexcl: iexcl,
  	iff: iff,
  	ifr: ifr,
  	Ifr: Ifr,
  	Igrave: Igrave,
  	igrave: igrave,
  	ii: ii,
  	iiiint: iiiint,
  	iiint: iiint,
  	iinfin: iinfin,
  	iiota: iiota,
  	IJlig: IJlig,
  	ijlig: ijlig,
  	Imacr: Imacr,
  	imacr: imacr,
  	image: image,
  	ImaginaryI: ImaginaryI,
  	imagline: imagline,
  	imagpart: imagpart,
  	imath: imath,
  	Im: Im,
  	imof: imof,
  	imped: imped,
  	Implies: Implies,
  	incare: incare,
  	infin: infin,
  	infintie: infintie,
  	inodot: inodot,
  	intcal: intcal,
  	int: int,
  	Int: Int,
  	integers: integers,
  	Integral: Integral,
  	intercal: intercal,
  	Intersection: Intersection,
  	intlarhk: intlarhk,
  	intprod: intprod,
  	InvisibleComma: InvisibleComma,
  	InvisibleTimes: InvisibleTimes,
  	IOcy: IOcy,
  	iocy: iocy,
  	Iogon: Iogon,
  	iogon: iogon,
  	Iopf: Iopf,
  	iopf: iopf,
  	Iota: Iota,
  	iota: iota,
  	iprod: iprod,
  	iquest: iquest,
  	iscr: iscr,
  	Iscr: Iscr,
  	isin: isin,
  	isindot: isindot,
  	isinE: isinE,
  	isins: isins,
  	isinsv: isinsv,
  	isinv: isinv,
  	it: it,
  	Itilde: Itilde,
  	itilde: itilde,
  	Iukcy: Iukcy,
  	iukcy: iukcy,
  	Iuml: Iuml,
  	iuml: iuml,
  	Jcirc: Jcirc,
  	jcirc: jcirc,
  	Jcy: Jcy,
  	jcy: jcy,
  	Jfr: Jfr,
  	jfr: jfr,
  	jmath: jmath,
  	Jopf: Jopf,
  	jopf: jopf,
  	Jscr: Jscr,
  	jscr: jscr,
  	Jsercy: Jsercy,
  	jsercy: jsercy,
  	Jukcy: Jukcy,
  	jukcy: jukcy,
  	Kappa: Kappa,
  	kappa: kappa,
  	kappav: kappav,
  	Kcedil: Kcedil,
  	kcedil: kcedil,
  	Kcy: Kcy,
  	kcy: kcy,
  	Kfr: Kfr,
  	kfr: kfr,
  	kgreen: kgreen,
  	KHcy: KHcy,
  	khcy: khcy,
  	KJcy: KJcy,
  	kjcy: kjcy,
  	Kopf: Kopf,
  	kopf: kopf,
  	Kscr: Kscr,
  	kscr: kscr,
  	lAarr: lAarr,
  	Lacute: Lacute,
  	lacute: lacute,
  	laemptyv: laemptyv,
  	lagran: lagran,
  	Lambda: Lambda,
  	lambda: lambda,
  	lang: lang,
  	Lang: Lang,
  	langd: langd,
  	langle: langle,
  	lap: lap,
  	Laplacetrf: Laplacetrf,
  	laquo: laquo,
  	larrb: larrb,
  	larrbfs: larrbfs,
  	larr: larr,
  	Larr: Larr,
  	lArr: lArr,
  	larrfs: larrfs,
  	larrhk: larrhk,
  	larrlp: larrlp,
  	larrpl: larrpl,
  	larrsim: larrsim,
  	larrtl: larrtl,
  	latail: latail,
  	lAtail: lAtail,
  	lat: lat,
  	late: late,
  	lates: lates,
  	lbarr: lbarr,
  	lBarr: lBarr,
  	lbbrk: lbbrk,
  	lbrace: lbrace,
  	lbrack: lbrack,
  	lbrke: lbrke,
  	lbrksld: lbrksld,
  	lbrkslu: lbrkslu,
  	Lcaron: Lcaron,
  	lcaron: lcaron,
  	Lcedil: Lcedil,
  	lcedil: lcedil,
  	lceil: lceil,
  	lcub: lcub,
  	Lcy: Lcy,
  	lcy: lcy,
  	ldca: ldca,
  	ldquo: ldquo,
  	ldquor: ldquor,
  	ldrdhar: ldrdhar,
  	ldrushar: ldrushar,
  	ldsh: ldsh,
  	le: le,
  	lE: lE,
  	LeftAngleBracket: LeftAngleBracket,
  	LeftArrowBar: LeftArrowBar,
  	leftarrow: leftarrow,
  	LeftArrow: LeftArrow,
  	Leftarrow: Leftarrow,
  	LeftArrowRightArrow: LeftArrowRightArrow,
  	leftarrowtail: leftarrowtail,
  	LeftCeiling: LeftCeiling,
  	LeftDoubleBracket: LeftDoubleBracket,
  	LeftDownTeeVector: LeftDownTeeVector,
  	LeftDownVectorBar: LeftDownVectorBar,
  	LeftDownVector: LeftDownVector,
  	LeftFloor: LeftFloor,
  	leftharpoondown: leftharpoondown,
  	leftharpoonup: leftharpoonup,
  	leftleftarrows: leftleftarrows,
  	leftrightarrow: leftrightarrow,
  	LeftRightArrow: LeftRightArrow,
  	Leftrightarrow: Leftrightarrow,
  	leftrightarrows: leftrightarrows,
  	leftrightharpoons: leftrightharpoons,
  	leftrightsquigarrow: leftrightsquigarrow,
  	LeftRightVector: LeftRightVector,
  	LeftTeeArrow: LeftTeeArrow,
  	LeftTee: LeftTee,
  	LeftTeeVector: LeftTeeVector,
  	leftthreetimes: leftthreetimes,
  	LeftTriangleBar: LeftTriangleBar,
  	LeftTriangle: LeftTriangle,
  	LeftTriangleEqual: LeftTriangleEqual,
  	LeftUpDownVector: LeftUpDownVector,
  	LeftUpTeeVector: LeftUpTeeVector,
  	LeftUpVectorBar: LeftUpVectorBar,
  	LeftUpVector: LeftUpVector,
  	LeftVectorBar: LeftVectorBar,
  	LeftVector: LeftVector,
  	lEg: lEg,
  	leg: leg,
  	leq: leq,
  	leqq: leqq,
  	leqslant: leqslant,
  	lescc: lescc,
  	les: les,
  	lesdot: lesdot,
  	lesdoto: lesdoto,
  	lesdotor: lesdotor,
  	lesg: lesg,
  	lesges: lesges,
  	lessapprox: lessapprox,
  	lessdot: lessdot,
  	lesseqgtr: lesseqgtr,
  	lesseqqgtr: lesseqqgtr,
  	LessEqualGreater: LessEqualGreater,
  	LessFullEqual: LessFullEqual,
  	LessGreater: LessGreater,
  	lessgtr: lessgtr,
  	LessLess: LessLess,
  	lesssim: lesssim,
  	LessSlantEqual: LessSlantEqual,
  	LessTilde: LessTilde,
  	lfisht: lfisht,
  	lfloor: lfloor,
  	Lfr: Lfr,
  	lfr: lfr,
  	lg: lg,
  	lgE: lgE,
  	lHar: lHar,
  	lhard: lhard,
  	lharu: lharu,
  	lharul: lharul,
  	lhblk: lhblk,
  	LJcy: LJcy,
  	ljcy: ljcy,
  	llarr: llarr,
  	ll: ll,
  	Ll: Ll,
  	llcorner: llcorner,
  	Lleftarrow: Lleftarrow,
  	llhard: llhard,
  	lltri: lltri,
  	Lmidot: Lmidot,
  	lmidot: lmidot,
  	lmoustache: lmoustache,
  	lmoust: lmoust,
  	lnap: lnap,
  	lnapprox: lnapprox,
  	lne: lne,
  	lnE: lnE,
  	lneq: lneq,
  	lneqq: lneqq,
  	lnsim: lnsim,
  	loang: loang,
  	loarr: loarr,
  	lobrk: lobrk,
  	longleftarrow: longleftarrow,
  	LongLeftArrow: LongLeftArrow,
  	Longleftarrow: Longleftarrow,
  	longleftrightarrow: longleftrightarrow,
  	LongLeftRightArrow: LongLeftRightArrow,
  	Longleftrightarrow: Longleftrightarrow,
  	longmapsto: longmapsto,
  	longrightarrow: longrightarrow,
  	LongRightArrow: LongRightArrow,
  	Longrightarrow: Longrightarrow,
  	looparrowleft: looparrowleft,
  	looparrowright: looparrowright,
  	lopar: lopar,
  	Lopf: Lopf,
  	lopf: lopf,
  	loplus: loplus,
  	lotimes: lotimes,
  	lowast: lowast,
  	lowbar: lowbar,
  	LowerLeftArrow: LowerLeftArrow,
  	LowerRightArrow: LowerRightArrow,
  	loz: loz,
  	lozenge: lozenge,
  	lozf: lozf,
  	lpar: lpar,
  	lparlt: lparlt,
  	lrarr: lrarr,
  	lrcorner: lrcorner,
  	lrhar: lrhar,
  	lrhard: lrhard,
  	lrm: lrm,
  	lrtri: lrtri,
  	lsaquo: lsaquo,
  	lscr: lscr,
  	Lscr: Lscr,
  	lsh: lsh,
  	Lsh: Lsh,
  	lsim: lsim,
  	lsime: lsime,
  	lsimg: lsimg,
  	lsqb: lsqb,
  	lsquo: lsquo,
  	lsquor: lsquor,
  	Lstrok: Lstrok,
  	lstrok: lstrok,
  	ltcc: ltcc,
  	ltcir: ltcir,
  	lt: lt$1,
  	LT: LT,
  	Lt: Lt,
  	ltdot: ltdot,
  	lthree: lthree,
  	ltimes: ltimes,
  	ltlarr: ltlarr,
  	ltquest: ltquest,
  	ltri: ltri,
  	ltrie: ltrie,
  	ltrif: ltrif,
  	ltrPar: ltrPar,
  	lurdshar: lurdshar,
  	luruhar: luruhar,
  	lvertneqq: lvertneqq,
  	lvnE: lvnE,
  	macr: macr,
  	male: male,
  	malt: malt,
  	maltese: maltese,
  	map: map$1,
  	mapsto: mapsto,
  	mapstodown: mapstodown,
  	mapstoleft: mapstoleft,
  	mapstoup: mapstoup,
  	marker: marker,
  	mcomma: mcomma,
  	Mcy: Mcy,
  	mcy: mcy,
  	mdash: mdash,
  	mDDot: mDDot,
  	measuredangle: measuredangle,
  	MediumSpace: MediumSpace,
  	Mellintrf: Mellintrf,
  	Mfr: Mfr,
  	mfr: mfr,
  	mho: mho,
  	micro: micro,
  	midast: midast,
  	midcir: midcir,
  	mid: mid,
  	middot: middot,
  	minusb: minusb,
  	minus: minus,
  	minusd: minusd,
  	minusdu: minusdu,
  	MinusPlus: MinusPlus,
  	mlcp: mlcp,
  	mldr: mldr,
  	mnplus: mnplus,
  	models: models,
  	Mopf: Mopf,
  	mopf: mopf,
  	mp: mp,
  	mscr: mscr,
  	Mscr: Mscr,
  	mstpos: mstpos,
  	Mu: Mu,
  	mu: mu,
  	multimap: multimap,
  	mumap: mumap,
  	nabla: nabla,
  	Nacute: Nacute,
  	nacute: nacute,
  	nang: nang,
  	nap: nap,
  	napE: napE,
  	napid: napid,
  	napos: napos,
  	napprox: napprox,
  	natural: natural,
  	naturals: naturals,
  	natur: natur,
  	nbsp: nbsp,
  	nbump: nbump,
  	nbumpe: nbumpe,
  	ncap: ncap,
  	Ncaron: Ncaron,
  	ncaron: ncaron,
  	Ncedil: Ncedil,
  	ncedil: ncedil,
  	ncong: ncong,
  	ncongdot: ncongdot,
  	ncup: ncup,
  	Ncy: Ncy,
  	ncy: ncy,
  	ndash: ndash,
  	nearhk: nearhk,
  	nearr: nearr,
  	neArr: neArr,
  	nearrow: nearrow,
  	ne: ne,
  	nedot: nedot,
  	NegativeMediumSpace: NegativeMediumSpace,
  	NegativeThickSpace: NegativeThickSpace,
  	NegativeThinSpace: NegativeThinSpace,
  	NegativeVeryThinSpace: NegativeVeryThinSpace,
  	nequiv: nequiv,
  	nesear: nesear,
  	nesim: nesim,
  	NestedGreaterGreater: NestedGreaterGreater,
  	NestedLessLess: NestedLessLess,
  	NewLine: NewLine,
  	nexist: nexist,
  	nexists: nexists,
  	Nfr: Nfr,
  	nfr: nfr,
  	ngE: ngE,
  	nge: nge,
  	ngeq: ngeq,
  	ngeqq: ngeqq,
  	ngeqslant: ngeqslant,
  	nges: nges,
  	nGg: nGg,
  	ngsim: ngsim,
  	nGt: nGt,
  	ngt: ngt,
  	ngtr: ngtr,
  	nGtv: nGtv,
  	nharr: nharr,
  	nhArr: nhArr,
  	nhpar: nhpar,
  	ni: ni,
  	nis: nis,
  	nisd: nisd,
  	niv: niv,
  	NJcy: NJcy,
  	njcy: njcy,
  	nlarr: nlarr,
  	nlArr: nlArr,
  	nldr: nldr,
  	nlE: nlE,
  	nle: nle,
  	nleftarrow: nleftarrow,
  	nLeftarrow: nLeftarrow,
  	nleftrightarrow: nleftrightarrow,
  	nLeftrightarrow: nLeftrightarrow,
  	nleq: nleq,
  	nleqq: nleqq,
  	nleqslant: nleqslant,
  	nles: nles,
  	nless: nless,
  	nLl: nLl,
  	nlsim: nlsim,
  	nLt: nLt,
  	nlt: nlt,
  	nltri: nltri,
  	nltrie: nltrie,
  	nLtv: nLtv,
  	nmid: nmid,
  	NoBreak: NoBreak,
  	NonBreakingSpace: NonBreakingSpace,
  	nopf: nopf,
  	Nopf: Nopf,
  	Not: Not,
  	not: not,
  	NotCongruent: NotCongruent,
  	NotCupCap: NotCupCap,
  	NotDoubleVerticalBar: NotDoubleVerticalBar,
  	NotElement: NotElement,
  	NotEqual: NotEqual,
  	NotEqualTilde: NotEqualTilde,
  	NotExists: NotExists,
  	NotGreater: NotGreater,
  	NotGreaterEqual: NotGreaterEqual,
  	NotGreaterFullEqual: NotGreaterFullEqual,
  	NotGreaterGreater: NotGreaterGreater,
  	NotGreaterLess: NotGreaterLess,
  	NotGreaterSlantEqual: NotGreaterSlantEqual,
  	NotGreaterTilde: NotGreaterTilde,
  	NotHumpDownHump: NotHumpDownHump,
  	NotHumpEqual: NotHumpEqual,
  	notin: notin,
  	notindot: notindot,
  	notinE: notinE,
  	notinva: notinva,
  	notinvb: notinvb,
  	notinvc: notinvc,
  	NotLeftTriangleBar: NotLeftTriangleBar,
  	NotLeftTriangle: NotLeftTriangle,
  	NotLeftTriangleEqual: NotLeftTriangleEqual,
  	NotLess: NotLess,
  	NotLessEqual: NotLessEqual,
  	NotLessGreater: NotLessGreater,
  	NotLessLess: NotLessLess,
  	NotLessSlantEqual: NotLessSlantEqual,
  	NotLessTilde: NotLessTilde,
  	NotNestedGreaterGreater: NotNestedGreaterGreater,
  	NotNestedLessLess: NotNestedLessLess,
  	notni: notni,
  	notniva: notniva,
  	notnivb: notnivb,
  	notnivc: notnivc,
  	NotPrecedes: NotPrecedes,
  	NotPrecedesEqual: NotPrecedesEqual,
  	NotPrecedesSlantEqual: NotPrecedesSlantEqual,
  	NotReverseElement: NotReverseElement,
  	NotRightTriangleBar: NotRightTriangleBar,
  	NotRightTriangle: NotRightTriangle,
  	NotRightTriangleEqual: NotRightTriangleEqual,
  	NotSquareSubset: NotSquareSubset,
  	NotSquareSubsetEqual: NotSquareSubsetEqual,
  	NotSquareSuperset: NotSquareSuperset,
  	NotSquareSupersetEqual: NotSquareSupersetEqual,
  	NotSubset: NotSubset,
  	NotSubsetEqual: NotSubsetEqual,
  	NotSucceeds: NotSucceeds,
  	NotSucceedsEqual: NotSucceedsEqual,
  	NotSucceedsSlantEqual: NotSucceedsSlantEqual,
  	NotSucceedsTilde: NotSucceedsTilde,
  	NotSuperset: NotSuperset,
  	NotSupersetEqual: NotSupersetEqual,
  	NotTilde: NotTilde,
  	NotTildeEqual: NotTildeEqual,
  	NotTildeFullEqual: NotTildeFullEqual,
  	NotTildeTilde: NotTildeTilde,
  	NotVerticalBar: NotVerticalBar,
  	nparallel: nparallel,
  	npar: npar,
  	nparsl: nparsl,
  	npart: npart,
  	npolint: npolint,
  	npr: npr,
  	nprcue: nprcue,
  	nprec: nprec,
  	npreceq: npreceq,
  	npre: npre,
  	nrarrc: nrarrc,
  	nrarr: nrarr,
  	nrArr: nrArr,
  	nrarrw: nrarrw,
  	nrightarrow: nrightarrow,
  	nRightarrow: nRightarrow,
  	nrtri: nrtri,
  	nrtrie: nrtrie,
  	nsc: nsc,
  	nsccue: nsccue,
  	nsce: nsce,
  	Nscr: Nscr,
  	nscr: nscr,
  	nshortmid: nshortmid,
  	nshortparallel: nshortparallel,
  	nsim: nsim,
  	nsime: nsime,
  	nsimeq: nsimeq,
  	nsmid: nsmid,
  	nspar: nspar,
  	nsqsube: nsqsube,
  	nsqsupe: nsqsupe,
  	nsub: nsub,
  	nsubE: nsubE,
  	nsube: nsube,
  	nsubset: nsubset,
  	nsubseteq: nsubseteq,
  	nsubseteqq: nsubseteqq,
  	nsucc: nsucc,
  	nsucceq: nsucceq,
  	nsup: nsup,
  	nsupE: nsupE,
  	nsupe: nsupe,
  	nsupset: nsupset,
  	nsupseteq: nsupseteq,
  	nsupseteqq: nsupseteqq,
  	ntgl: ntgl,
  	Ntilde: Ntilde,
  	ntilde: ntilde,
  	ntlg: ntlg,
  	ntriangleleft: ntriangleleft,
  	ntrianglelefteq: ntrianglelefteq,
  	ntriangleright: ntriangleright,
  	ntrianglerighteq: ntrianglerighteq,
  	Nu: Nu,
  	nu: nu,
  	num: num,
  	numero: numero,
  	numsp: numsp,
  	nvap: nvap,
  	nvdash: nvdash,
  	nvDash: nvDash,
  	nVdash: nVdash,
  	nVDash: nVDash,
  	nvge: nvge,
  	nvgt: nvgt,
  	nvHarr: nvHarr,
  	nvinfin: nvinfin,
  	nvlArr: nvlArr,
  	nvle: nvle,
  	nvlt: nvlt,
  	nvltrie: nvltrie,
  	nvrArr: nvrArr,
  	nvrtrie: nvrtrie,
  	nvsim: nvsim,
  	nwarhk: nwarhk,
  	nwarr: nwarr,
  	nwArr: nwArr,
  	nwarrow: nwarrow,
  	nwnear: nwnear,
  	Oacute: Oacute,
  	oacute: oacute,
  	oast: oast,
  	Ocirc: Ocirc,
  	ocirc: ocirc,
  	ocir: ocir,
  	Ocy: Ocy,
  	ocy: ocy,
  	odash: odash,
  	Odblac: Odblac,
  	odblac: odblac,
  	odiv: odiv,
  	odot: odot,
  	odsold: odsold,
  	OElig: OElig,
  	oelig: oelig,
  	ofcir: ofcir,
  	Ofr: Ofr,
  	ofr: ofr,
  	ogon: ogon,
  	Ograve: Ograve,
  	ograve: ograve,
  	ogt: ogt,
  	ohbar: ohbar,
  	ohm: ohm,
  	oint: oint,
  	olarr: olarr,
  	olcir: olcir,
  	olcross: olcross,
  	oline: oline,
  	olt: olt,
  	Omacr: Omacr,
  	omacr: omacr,
  	Omega: Omega,
  	omega: omega,
  	Omicron: Omicron,
  	omicron: omicron,
  	omid: omid,
  	ominus: ominus,
  	Oopf: Oopf,
  	oopf: oopf,
  	opar: opar,
  	OpenCurlyDoubleQuote: OpenCurlyDoubleQuote,
  	OpenCurlyQuote: OpenCurlyQuote,
  	operp: operp,
  	oplus: oplus,
  	orarr: orarr,
  	Or: Or,
  	or: or,
  	ord: ord,
  	order: order,
  	orderof: orderof,
  	ordf: ordf,
  	ordm: ordm,
  	origof: origof,
  	oror: oror,
  	orslope: orslope,
  	orv: orv,
  	oS: oS,
  	Oscr: Oscr,
  	oscr: oscr,
  	Oslash: Oslash,
  	oslash: oslash,
  	osol: osol,
  	Otilde: Otilde,
  	otilde: otilde,
  	otimesas: otimesas,
  	Otimes: Otimes,
  	otimes: otimes,
  	Ouml: Ouml,
  	ouml: ouml,
  	ovbar: ovbar,
  	OverBar: OverBar,
  	OverBrace: OverBrace,
  	OverBracket: OverBracket,
  	OverParenthesis: OverParenthesis,
  	para: para,
  	parallel: parallel,
  	par: par,
  	parsim: parsim,
  	parsl: parsl,
  	part: part,
  	PartialD: PartialD,
  	Pcy: Pcy,
  	pcy: pcy,
  	percnt: percnt,
  	period: period,
  	permil: permil,
  	perp: perp,
  	pertenk: pertenk,
  	Pfr: Pfr,
  	pfr: pfr,
  	Phi: Phi,
  	phi: phi,
  	phiv: phiv,
  	phmmat: phmmat,
  	phone: phone,
  	Pi: Pi,
  	pi: pi,
  	pitchfork: pitchfork,
  	piv: piv,
  	planck: planck,
  	planckh: planckh,
  	plankv: plankv,
  	plusacir: plusacir,
  	plusb: plusb,
  	pluscir: pluscir,
  	plus: plus,
  	plusdo: plusdo,
  	plusdu: plusdu,
  	pluse: pluse,
  	PlusMinus: PlusMinus,
  	plusmn: plusmn,
  	plussim: plussim,
  	plustwo: plustwo,
  	pm: pm,
  	Poincareplane: Poincareplane,
  	pointint: pointint,
  	popf: popf,
  	Popf: Popf,
  	pound: pound,
  	prap: prap,
  	Pr: Pr,
  	pr: pr,
  	prcue: prcue,
  	precapprox: precapprox,
  	prec: prec,
  	preccurlyeq: preccurlyeq,
  	Precedes: Precedes,
  	PrecedesEqual: PrecedesEqual,
  	PrecedesSlantEqual: PrecedesSlantEqual,
  	PrecedesTilde: PrecedesTilde,
  	preceq: preceq,
  	precnapprox: precnapprox,
  	precneqq: precneqq,
  	precnsim: precnsim,
  	pre: pre,
  	prE: prE,
  	precsim: precsim,
  	prime: prime,
  	Prime: Prime,
  	primes: primes,
  	prnap: prnap,
  	prnE: prnE,
  	prnsim: prnsim,
  	prod: prod,
  	Product: Product,
  	profalar: profalar,
  	profline: profline,
  	profsurf: profsurf,
  	prop: prop,
  	Proportional: Proportional,
  	Proportion: Proportion,
  	propto: propto,
  	prsim: prsim,
  	prurel: prurel,
  	Pscr: Pscr,
  	pscr: pscr,
  	Psi: Psi,
  	psi: psi,
  	puncsp: puncsp,
  	Qfr: Qfr,
  	qfr: qfr,
  	qint: qint,
  	qopf: qopf,
  	Qopf: Qopf,
  	qprime: qprime,
  	Qscr: Qscr,
  	qscr: qscr,
  	quaternions: quaternions,
  	quatint: quatint,
  	quest: quest,
  	questeq: questeq,
  	quot: quot$1,
  	QUOT: QUOT,
  	rAarr: rAarr,
  	race: race,
  	Racute: Racute,
  	racute: racute,
  	radic: radic,
  	raemptyv: raemptyv,
  	rang: rang,
  	Rang: Rang,
  	rangd: rangd,
  	range: range,
  	rangle: rangle,
  	raquo: raquo,
  	rarrap: rarrap,
  	rarrb: rarrb,
  	rarrbfs: rarrbfs,
  	rarrc: rarrc,
  	rarr: rarr,
  	Rarr: Rarr,
  	rArr: rArr,
  	rarrfs: rarrfs,
  	rarrhk: rarrhk,
  	rarrlp: rarrlp,
  	rarrpl: rarrpl,
  	rarrsim: rarrsim,
  	Rarrtl: Rarrtl,
  	rarrtl: rarrtl,
  	rarrw: rarrw,
  	ratail: ratail,
  	rAtail: rAtail,
  	ratio: ratio,
  	rationals: rationals,
  	rbarr: rbarr,
  	rBarr: rBarr,
  	RBarr: RBarr,
  	rbbrk: rbbrk,
  	rbrace: rbrace,
  	rbrack: rbrack,
  	rbrke: rbrke,
  	rbrksld: rbrksld,
  	rbrkslu: rbrkslu,
  	Rcaron: Rcaron,
  	rcaron: rcaron,
  	Rcedil: Rcedil,
  	rcedil: rcedil,
  	rceil: rceil,
  	rcub: rcub,
  	Rcy: Rcy,
  	rcy: rcy,
  	rdca: rdca,
  	rdldhar: rdldhar,
  	rdquo: rdquo,
  	rdquor: rdquor,
  	rdsh: rdsh,
  	real: real,
  	realine: realine,
  	realpart: realpart,
  	reals: reals,
  	Re: Re,
  	rect: rect,
  	reg: reg,
  	REG: REG,
  	ReverseElement: ReverseElement,
  	ReverseEquilibrium: ReverseEquilibrium,
  	ReverseUpEquilibrium: ReverseUpEquilibrium,
  	rfisht: rfisht,
  	rfloor: rfloor,
  	rfr: rfr,
  	Rfr: Rfr,
  	rHar: rHar,
  	rhard: rhard,
  	rharu: rharu,
  	rharul: rharul,
  	Rho: Rho,
  	rho: rho,
  	rhov: rhov,
  	RightAngleBracket: RightAngleBracket,
  	RightArrowBar: RightArrowBar,
  	rightarrow: rightarrow,
  	RightArrow: RightArrow,
  	Rightarrow: Rightarrow,
  	RightArrowLeftArrow: RightArrowLeftArrow,
  	rightarrowtail: rightarrowtail,
  	RightCeiling: RightCeiling,
  	RightDoubleBracket: RightDoubleBracket,
  	RightDownTeeVector: RightDownTeeVector,
  	RightDownVectorBar: RightDownVectorBar,
  	RightDownVector: RightDownVector,
  	RightFloor: RightFloor,
  	rightharpoondown: rightharpoondown,
  	rightharpoonup: rightharpoonup,
  	rightleftarrows: rightleftarrows,
  	rightleftharpoons: rightleftharpoons,
  	rightrightarrows: rightrightarrows,
  	rightsquigarrow: rightsquigarrow,
  	RightTeeArrow: RightTeeArrow,
  	RightTee: RightTee,
  	RightTeeVector: RightTeeVector,
  	rightthreetimes: rightthreetimes,
  	RightTriangleBar: RightTriangleBar,
  	RightTriangle: RightTriangle,
  	RightTriangleEqual: RightTriangleEqual,
  	RightUpDownVector: RightUpDownVector,
  	RightUpTeeVector: RightUpTeeVector,
  	RightUpVectorBar: RightUpVectorBar,
  	RightUpVector: RightUpVector,
  	RightVectorBar: RightVectorBar,
  	RightVector: RightVector,
  	ring: ring,
  	risingdotseq: risingdotseq,
  	rlarr: rlarr,
  	rlhar: rlhar,
  	rlm: rlm,
  	rmoustache: rmoustache,
  	rmoust: rmoust,
  	rnmid: rnmid,
  	roang: roang,
  	roarr: roarr,
  	robrk: robrk,
  	ropar: ropar,
  	ropf: ropf,
  	Ropf: Ropf,
  	roplus: roplus,
  	rotimes: rotimes,
  	RoundImplies: RoundImplies,
  	rpar: rpar,
  	rpargt: rpargt,
  	rppolint: rppolint,
  	rrarr: rrarr,
  	Rrightarrow: Rrightarrow,
  	rsaquo: rsaquo,
  	rscr: rscr,
  	Rscr: Rscr,
  	rsh: rsh,
  	Rsh: Rsh,
  	rsqb: rsqb,
  	rsquo: rsquo,
  	rsquor: rsquor,
  	rthree: rthree,
  	rtimes: rtimes,
  	rtri: rtri,
  	rtrie: rtrie,
  	rtrif: rtrif,
  	rtriltri: rtriltri,
  	RuleDelayed: RuleDelayed,
  	ruluhar: ruluhar,
  	rx: rx,
  	Sacute: Sacute,
  	sacute: sacute,
  	sbquo: sbquo,
  	scap: scap,
  	Scaron: Scaron,
  	scaron: scaron,
  	Sc: Sc,
  	sc: sc,
  	sccue: sccue,
  	sce: sce,
  	scE: scE,
  	Scedil: Scedil,
  	scedil: scedil,
  	Scirc: Scirc,
  	scirc: scirc,
  	scnap: scnap,
  	scnE: scnE,
  	scnsim: scnsim,
  	scpolint: scpolint,
  	scsim: scsim,
  	Scy: Scy,
  	scy: scy,
  	sdotb: sdotb,
  	sdot: sdot,
  	sdote: sdote,
  	searhk: searhk,
  	searr: searr,
  	seArr: seArr,
  	searrow: searrow,
  	sect: sect,
  	semi: semi,
  	seswar: seswar,
  	setminus: setminus,
  	setmn: setmn,
  	sext: sext,
  	Sfr: Sfr,
  	sfr: sfr,
  	sfrown: sfrown,
  	sharp: sharp,
  	SHCHcy: SHCHcy,
  	shchcy: shchcy,
  	SHcy: SHcy,
  	shcy: shcy,
  	ShortDownArrow: ShortDownArrow,
  	ShortLeftArrow: ShortLeftArrow,
  	shortmid: shortmid,
  	shortparallel: shortparallel,
  	ShortRightArrow: ShortRightArrow,
  	ShortUpArrow: ShortUpArrow,
  	shy: shy,
  	Sigma: Sigma,
  	sigma: sigma,
  	sigmaf: sigmaf,
  	sigmav: sigmav,
  	sim: sim,
  	simdot: simdot,
  	sime: sime,
  	simeq: simeq,
  	simg: simg,
  	simgE: simgE,
  	siml: siml,
  	simlE: simlE,
  	simne: simne,
  	simplus: simplus,
  	simrarr: simrarr,
  	slarr: slarr,
  	SmallCircle: SmallCircle,
  	smallsetminus: smallsetminus,
  	smashp: smashp,
  	smeparsl: smeparsl,
  	smid: smid,
  	smile: smile,
  	smt: smt,
  	smte: smte,
  	smtes: smtes,
  	SOFTcy: SOFTcy,
  	softcy: softcy,
  	solbar: solbar,
  	solb: solb,
  	sol: sol,
  	Sopf: Sopf,
  	sopf: sopf,
  	spades: spades,
  	spadesuit: spadesuit,
  	spar: spar,
  	sqcap: sqcap,
  	sqcaps: sqcaps,
  	sqcup: sqcup,
  	sqcups: sqcups,
  	Sqrt: Sqrt,
  	sqsub: sqsub,
  	sqsube: sqsube,
  	sqsubset: sqsubset,
  	sqsubseteq: sqsubseteq,
  	sqsup: sqsup,
  	sqsupe: sqsupe,
  	sqsupset: sqsupset,
  	sqsupseteq: sqsupseteq,
  	square: square,
  	Square: Square,
  	SquareIntersection: SquareIntersection,
  	SquareSubset: SquareSubset,
  	SquareSubsetEqual: SquareSubsetEqual,
  	SquareSuperset: SquareSuperset,
  	SquareSupersetEqual: SquareSupersetEqual,
  	SquareUnion: SquareUnion,
  	squarf: squarf,
  	squ: squ,
  	squf: squf,
  	srarr: srarr,
  	Sscr: Sscr,
  	sscr: sscr,
  	ssetmn: ssetmn,
  	ssmile: ssmile,
  	sstarf: sstarf,
  	Star: Star,
  	star: star,
  	starf: starf,
  	straightepsilon: straightepsilon,
  	straightphi: straightphi,
  	strns: strns,
  	sub: sub,
  	Sub: Sub,
  	subdot: subdot,
  	subE: subE,
  	sube: sube,
  	subedot: subedot,
  	submult: submult,
  	subnE: subnE,
  	subne: subne,
  	subplus: subplus,
  	subrarr: subrarr,
  	subset: subset,
  	Subset: Subset,
  	subseteq: subseteq,
  	subseteqq: subseteqq,
  	SubsetEqual: SubsetEqual,
  	subsetneq: subsetneq,
  	subsetneqq: subsetneqq,
  	subsim: subsim,
  	subsub: subsub,
  	subsup: subsup,
  	succapprox: succapprox,
  	succ: succ,
  	succcurlyeq: succcurlyeq,
  	Succeeds: Succeeds,
  	SucceedsEqual: SucceedsEqual,
  	SucceedsSlantEqual: SucceedsSlantEqual,
  	SucceedsTilde: SucceedsTilde,
  	succeq: succeq,
  	succnapprox: succnapprox,
  	succneqq: succneqq,
  	succnsim: succnsim,
  	succsim: succsim,
  	SuchThat: SuchThat,
  	sum: sum,
  	Sum: Sum,
  	sung: sung,
  	sup1: sup1,
  	sup2: sup2,
  	sup3: sup3,
  	sup: sup,
  	Sup: Sup,
  	supdot: supdot,
  	supdsub: supdsub,
  	supE: supE,
  	supe: supe,
  	supedot: supedot,
  	Superset: Superset,
  	SupersetEqual: SupersetEqual,
  	suphsol: suphsol,
  	suphsub: suphsub,
  	suplarr: suplarr,
  	supmult: supmult,
  	supnE: supnE,
  	supne: supne,
  	supplus: supplus,
  	supset: supset,
  	Supset: Supset,
  	supseteq: supseteq,
  	supseteqq: supseteqq,
  	supsetneq: supsetneq,
  	supsetneqq: supsetneqq,
  	supsim: supsim,
  	supsub: supsub,
  	supsup: supsup,
  	swarhk: swarhk,
  	swarr: swarr,
  	swArr: swArr,
  	swarrow: swarrow,
  	swnwar: swnwar,
  	szlig: szlig,
  	Tab: Tab,
  	target: target,
  	Tau: Tau,
  	tau: tau,
  	tbrk: tbrk,
  	Tcaron: Tcaron,
  	tcaron: tcaron,
  	Tcedil: Tcedil,
  	tcedil: tcedil,
  	Tcy: Tcy,
  	tcy: tcy,
  	tdot: tdot,
  	telrec: telrec,
  	Tfr: Tfr,
  	tfr: tfr,
  	there4: there4,
  	therefore: therefore,
  	Therefore: Therefore,
  	Theta: Theta,
  	theta: theta,
  	thetasym: thetasym,
  	thetav: thetav,
  	thickapprox: thickapprox,
  	thicksim: thicksim,
  	ThickSpace: ThickSpace,
  	ThinSpace: ThinSpace,
  	thinsp: thinsp,
  	thkap: thkap,
  	thksim: thksim,
  	THORN: THORN,
  	thorn: thorn,
  	tilde: tilde,
  	Tilde: Tilde,
  	TildeEqual: TildeEqual,
  	TildeFullEqual: TildeFullEqual,
  	TildeTilde: TildeTilde,
  	timesbar: timesbar,
  	timesb: timesb,
  	times: times,
  	timesd: timesd,
  	tint: tint,
  	toea: toea,
  	topbot: topbot,
  	topcir: topcir,
  	top: top,
  	Topf: Topf,
  	topf: topf,
  	topfork: topfork,
  	tosa: tosa,
  	tprime: tprime,
  	trade: trade,
  	TRADE: TRADE,
  	triangle: triangle,
  	triangledown: triangledown,
  	triangleleft: triangleleft,
  	trianglelefteq: trianglelefteq,
  	triangleq: triangleq,
  	triangleright: triangleright,
  	trianglerighteq: trianglerighteq,
  	tridot: tridot,
  	trie: trie,
  	triminus: triminus,
  	TripleDot: TripleDot,
  	triplus: triplus,
  	trisb: trisb,
  	tritime: tritime,
  	trpezium: trpezium,
  	Tscr: Tscr,
  	tscr: tscr,
  	TScy: TScy,
  	tscy: tscy,
  	TSHcy: TSHcy,
  	tshcy: tshcy,
  	Tstrok: Tstrok,
  	tstrok: tstrok,
  	twixt: twixt,
  	twoheadleftarrow: twoheadleftarrow,
  	twoheadrightarrow: twoheadrightarrow,
  	Uacute: Uacute,
  	uacute: uacute,
  	uarr: uarr,
  	Uarr: Uarr,
  	uArr: uArr,
  	Uarrocir: Uarrocir,
  	Ubrcy: Ubrcy,
  	ubrcy: ubrcy,
  	Ubreve: Ubreve,
  	ubreve: ubreve,
  	Ucirc: Ucirc,
  	ucirc: ucirc,
  	Ucy: Ucy,
  	ucy: ucy,
  	udarr: udarr,
  	Udblac: Udblac,
  	udblac: udblac,
  	udhar: udhar,
  	ufisht: ufisht,
  	Ufr: Ufr,
  	ufr: ufr,
  	Ugrave: Ugrave,
  	ugrave: ugrave,
  	uHar: uHar,
  	uharl: uharl,
  	uharr: uharr,
  	uhblk: uhblk,
  	ulcorn: ulcorn,
  	ulcorner: ulcorner,
  	ulcrop: ulcrop,
  	ultri: ultri,
  	Umacr: Umacr,
  	umacr: umacr,
  	uml: uml,
  	UnderBar: UnderBar,
  	UnderBrace: UnderBrace,
  	UnderBracket: UnderBracket,
  	UnderParenthesis: UnderParenthesis,
  	Union: Union,
  	UnionPlus: UnionPlus,
  	Uogon: Uogon,
  	uogon: uogon,
  	Uopf: Uopf,
  	uopf: uopf,
  	UpArrowBar: UpArrowBar,
  	uparrow: uparrow,
  	UpArrow: UpArrow,
  	Uparrow: Uparrow,
  	UpArrowDownArrow: UpArrowDownArrow,
  	updownarrow: updownarrow,
  	UpDownArrow: UpDownArrow,
  	Updownarrow: Updownarrow,
  	UpEquilibrium: UpEquilibrium,
  	upharpoonleft: upharpoonleft,
  	upharpoonright: upharpoonright,
  	uplus: uplus,
  	UpperLeftArrow: UpperLeftArrow,
  	UpperRightArrow: UpperRightArrow,
  	upsi: upsi,
  	Upsi: Upsi,
  	upsih: upsih,
  	Upsilon: Upsilon,
  	upsilon: upsilon,
  	UpTeeArrow: UpTeeArrow,
  	UpTee: UpTee,
  	upuparrows: upuparrows,
  	urcorn: urcorn,
  	urcorner: urcorner,
  	urcrop: urcrop,
  	Uring: Uring,
  	uring: uring,
  	urtri: urtri,
  	Uscr: Uscr,
  	uscr: uscr,
  	utdot: utdot,
  	Utilde: Utilde,
  	utilde: utilde,
  	utri: utri,
  	utrif: utrif,
  	uuarr: uuarr,
  	Uuml: Uuml,
  	uuml: uuml,
  	uwangle: uwangle,
  	vangrt: vangrt,
  	varepsilon: varepsilon,
  	varkappa: varkappa,
  	varnothing: varnothing,
  	varphi: varphi,
  	varpi: varpi,
  	varpropto: varpropto,
  	varr: varr,
  	vArr: vArr,
  	varrho: varrho,
  	varsigma: varsigma,
  	varsubsetneq: varsubsetneq,
  	varsubsetneqq: varsubsetneqq,
  	varsupsetneq: varsupsetneq,
  	varsupsetneqq: varsupsetneqq,
  	vartheta: vartheta,
  	vartriangleleft: vartriangleleft,
  	vartriangleright: vartriangleright,
  	vBar: vBar,
  	Vbar: Vbar,
  	vBarv: vBarv,
  	Vcy: Vcy,
  	vcy: vcy,
  	vdash: vdash,
  	vDash: vDash,
  	Vdash: Vdash,
  	VDash: VDash,
  	Vdashl: Vdashl,
  	veebar: veebar,
  	vee: vee,
  	Vee: Vee,
  	veeeq: veeeq,
  	vellip: vellip,
  	verbar: verbar,
  	Verbar: Verbar,
  	vert: vert,
  	Vert: Vert,
  	VerticalBar: VerticalBar,
  	VerticalLine: VerticalLine,
  	VerticalSeparator: VerticalSeparator,
  	VerticalTilde: VerticalTilde,
  	VeryThinSpace: VeryThinSpace,
  	Vfr: Vfr,
  	vfr: vfr,
  	vltri: vltri,
  	vnsub: vnsub,
  	vnsup: vnsup,
  	Vopf: Vopf,
  	vopf: vopf,
  	vprop: vprop,
  	vrtri: vrtri,
  	Vscr: Vscr,
  	vscr: vscr,
  	vsubnE: vsubnE,
  	vsubne: vsubne,
  	vsupnE: vsupnE,
  	vsupne: vsupne,
  	Vvdash: Vvdash,
  	vzigzag: vzigzag,
  	Wcirc: Wcirc,
  	wcirc: wcirc,
  	wedbar: wedbar,
  	wedge: wedge,
  	Wedge: Wedge,
  	wedgeq: wedgeq,
  	weierp: weierp,
  	Wfr: Wfr,
  	wfr: wfr,
  	Wopf: Wopf,
  	wopf: wopf,
  	wp: wp,
  	wr: wr,
  	wreath: wreath,
  	Wscr: Wscr,
  	wscr: wscr,
  	xcap: xcap,
  	xcirc: xcirc,
  	xcup: xcup,
  	xdtri: xdtri,
  	Xfr: Xfr,
  	xfr: xfr,
  	xharr: xharr,
  	xhArr: xhArr,
  	Xi: Xi,
  	xi: xi,
  	xlarr: xlarr,
  	xlArr: xlArr,
  	xmap: xmap,
  	xnis: xnis,
  	xodot: xodot,
  	Xopf: Xopf,
  	xopf: xopf,
  	xoplus: xoplus,
  	xotime: xotime,
  	xrarr: xrarr,
  	xrArr: xrArr,
  	Xscr: Xscr,
  	xscr: xscr,
  	xsqcup: xsqcup,
  	xuplus: xuplus,
  	xutri: xutri,
  	xvee: xvee,
  	xwedge: xwedge,
  	Yacute: Yacute,
  	yacute: yacute,
  	YAcy: YAcy,
  	yacy: yacy,
  	Ycirc: Ycirc,
  	ycirc: ycirc,
  	Ycy: Ycy,
  	ycy: ycy,
  	yen: yen,
  	Yfr: Yfr,
  	yfr: yfr,
  	YIcy: YIcy,
  	yicy: yicy,
  	Yopf: Yopf,
  	yopf: yopf,
  	Yscr: Yscr,
  	yscr: yscr,
  	YUcy: YUcy,
  	yucy: yucy,
  	yuml: yuml,
  	Yuml: Yuml,
  	Zacute: Zacute,
  	zacute: zacute,
  	Zcaron: Zcaron,
  	zcaron: zcaron,
  	Zcy: Zcy,
  	zcy: zcy,
  	Zdot: Zdot,
  	zdot: zdot,
  	zeetrf: zeetrf,
  	ZeroWidthSpace: ZeroWidthSpace,
  	Zeta: Zeta,
  	zeta: zeta,
  	zfr: zfr,
  	Zfr: Zfr,
  	ZHcy: ZHcy,
  	zhcy: zhcy,
  	zigrarr: zigrarr,
  	zopf: zopf,
  	Zopf: Zopf,
  	Zscr: Zscr,
  	zscr: zscr,
  	zwj: zwj,
  	zwnj: zwnj,
  	"in": "∈",
  	"Map": "⤅"
  };

  var entities = Object.freeze({
  	Aacute: Aacute,
  	aacute: aacute,
  	Abreve: Abreve,
  	abreve: abreve,
  	ac: ac,
  	acd: acd,
  	acE: acE,
  	Acirc: Acirc,
  	acirc: acirc,
  	acute: acute,
  	Acy: Acy,
  	acy: acy,
  	AElig: AElig,
  	aelig: aelig,
  	af: af,
  	Afr: Afr,
  	afr: afr,
  	Agrave: Agrave,
  	agrave: agrave,
  	alefsym: alefsym,
  	aleph: aleph,
  	Alpha: Alpha,
  	alpha: alpha,
  	Amacr: Amacr,
  	amacr: amacr,
  	amalg: amalg,
  	amp: amp$1,
  	AMP: AMP,
  	andand: andand,
  	And: And,
  	and: and,
  	andd: andd,
  	andslope: andslope,
  	andv: andv,
  	ang: ang,
  	ange: ange,
  	angle: angle,
  	angmsdaa: angmsdaa,
  	angmsdab: angmsdab,
  	angmsdac: angmsdac,
  	angmsdad: angmsdad,
  	angmsdae: angmsdae,
  	angmsdaf: angmsdaf,
  	angmsdag: angmsdag,
  	angmsdah: angmsdah,
  	angmsd: angmsd,
  	angrt: angrt,
  	angrtvb: angrtvb,
  	angrtvbd: angrtvbd,
  	angsph: angsph,
  	angst: angst,
  	angzarr: angzarr,
  	Aogon: Aogon,
  	aogon: aogon,
  	Aopf: Aopf,
  	aopf: aopf,
  	apacir: apacir,
  	ap: ap,
  	apE: apE,
  	ape: ape,
  	apid: apid,
  	apos: apos$1,
  	ApplyFunction: ApplyFunction,
  	approx: approx,
  	approxeq: approxeq,
  	Aring: Aring,
  	aring: aring,
  	Ascr: Ascr,
  	ascr: ascr,
  	Assign: Assign,
  	ast: ast,
  	asymp: asymp,
  	asympeq: asympeq,
  	Atilde: Atilde,
  	atilde: atilde,
  	Auml: Auml,
  	auml: auml,
  	awconint: awconint,
  	awint: awint,
  	backcong: backcong,
  	backepsilon: backepsilon,
  	backprime: backprime,
  	backsim: backsim,
  	backsimeq: backsimeq,
  	Backslash: Backslash,
  	Barv: Barv,
  	barvee: barvee,
  	barwed: barwed,
  	Barwed: Barwed,
  	barwedge: barwedge,
  	bbrk: bbrk,
  	bbrktbrk: bbrktbrk,
  	bcong: bcong,
  	Bcy: Bcy,
  	bcy: bcy,
  	bdquo: bdquo,
  	becaus: becaus,
  	because: because,
  	Because: Because,
  	bemptyv: bemptyv,
  	bepsi: bepsi,
  	bernou: bernou,
  	Bernoullis: Bernoullis,
  	Beta: Beta,
  	beta: beta,
  	beth: beth,
  	between: between,
  	Bfr: Bfr,
  	bfr: bfr,
  	bigcap: bigcap,
  	bigcirc: bigcirc,
  	bigcup: bigcup,
  	bigodot: bigodot,
  	bigoplus: bigoplus,
  	bigotimes: bigotimes,
  	bigsqcup: bigsqcup,
  	bigstar: bigstar,
  	bigtriangledown: bigtriangledown,
  	bigtriangleup: bigtriangleup,
  	biguplus: biguplus,
  	bigvee: bigvee,
  	bigwedge: bigwedge,
  	bkarow: bkarow,
  	blacklozenge: blacklozenge,
  	blacksquare: blacksquare,
  	blacktriangle: blacktriangle,
  	blacktriangledown: blacktriangledown,
  	blacktriangleleft: blacktriangleleft,
  	blacktriangleright: blacktriangleright,
  	blank: blank,
  	blk12: blk12,
  	blk14: blk14,
  	blk34: blk34,
  	block: block,
  	bne: bne,
  	bnequiv: bnequiv,
  	bNot: bNot,
  	bnot: bnot,
  	Bopf: Bopf,
  	bopf: bopf,
  	bot: bot,
  	bottom: bottom,
  	bowtie: bowtie,
  	boxbox: boxbox,
  	boxdl: boxdl,
  	boxdL: boxdL,
  	boxDl: boxDl,
  	boxDL: boxDL,
  	boxdr: boxdr,
  	boxdR: boxdR,
  	boxDr: boxDr,
  	boxDR: boxDR,
  	boxh: boxh,
  	boxH: boxH,
  	boxhd: boxhd,
  	boxHd: boxHd,
  	boxhD: boxhD,
  	boxHD: boxHD,
  	boxhu: boxhu,
  	boxHu: boxHu,
  	boxhU: boxhU,
  	boxHU: boxHU,
  	boxminus: boxminus,
  	boxplus: boxplus,
  	boxtimes: boxtimes,
  	boxul: boxul,
  	boxuL: boxuL,
  	boxUl: boxUl,
  	boxUL: boxUL,
  	boxur: boxur,
  	boxuR: boxuR,
  	boxUr: boxUr,
  	boxUR: boxUR,
  	boxv: boxv,
  	boxV: boxV,
  	boxvh: boxvh,
  	boxvH: boxvH,
  	boxVh: boxVh,
  	boxVH: boxVH,
  	boxvl: boxvl,
  	boxvL: boxvL,
  	boxVl: boxVl,
  	boxVL: boxVL,
  	boxvr: boxvr,
  	boxvR: boxvR,
  	boxVr: boxVr,
  	boxVR: boxVR,
  	bprime: bprime,
  	breve: breve,
  	Breve: Breve,
  	brvbar: brvbar,
  	bscr: bscr,
  	Bscr: Bscr,
  	bsemi: bsemi,
  	bsim: bsim,
  	bsime: bsime,
  	bsolb: bsolb,
  	bsol: bsol,
  	bsolhsub: bsolhsub,
  	bull: bull,
  	bullet: bullet,
  	bump: bump,
  	bumpE: bumpE,
  	bumpe: bumpe,
  	Bumpeq: Bumpeq,
  	bumpeq: bumpeq,
  	Cacute: Cacute,
  	cacute: cacute,
  	capand: capand,
  	capbrcup: capbrcup,
  	capcap: capcap,
  	cap: cap,
  	Cap: Cap,
  	capcup: capcup,
  	capdot: capdot,
  	CapitalDifferentialD: CapitalDifferentialD,
  	caps: caps,
  	caret: caret,
  	caron: caron,
  	Cayleys: Cayleys,
  	ccaps: ccaps,
  	Ccaron: Ccaron,
  	ccaron: ccaron,
  	Ccedil: Ccedil,
  	ccedil: ccedil,
  	Ccirc: Ccirc,
  	ccirc: ccirc,
  	Cconint: Cconint,
  	ccups: ccups,
  	ccupssm: ccupssm,
  	Cdot: Cdot,
  	cdot: cdot,
  	cedil: cedil,
  	Cedilla: Cedilla,
  	cemptyv: cemptyv,
  	cent: cent,
  	centerdot: centerdot,
  	CenterDot: CenterDot,
  	cfr: cfr,
  	Cfr: Cfr,
  	CHcy: CHcy,
  	chcy: chcy,
  	check: check,
  	checkmark: checkmark,
  	Chi: Chi,
  	chi: chi,
  	circ: circ,
  	circeq: circeq,
  	circlearrowleft: circlearrowleft,
  	circlearrowright: circlearrowright,
  	circledast: circledast,
  	circledcirc: circledcirc,
  	circleddash: circleddash,
  	CircleDot: CircleDot,
  	circledR: circledR,
  	circledS: circledS,
  	CircleMinus: CircleMinus,
  	CirclePlus: CirclePlus,
  	CircleTimes: CircleTimes,
  	cir: cir,
  	cirE: cirE,
  	cire: cire,
  	cirfnint: cirfnint,
  	cirmid: cirmid,
  	cirscir: cirscir,
  	ClockwiseContourIntegral: ClockwiseContourIntegral,
  	CloseCurlyDoubleQuote: CloseCurlyDoubleQuote,
  	CloseCurlyQuote: CloseCurlyQuote,
  	clubs: clubs,
  	clubsuit: clubsuit,
  	colon: colon,
  	Colon: Colon,
  	Colone: Colone,
  	colone: colone,
  	coloneq: coloneq,
  	comma: comma,
  	commat: commat,
  	comp: comp,
  	compfn: compfn,
  	complement: complement,
  	complexes: complexes,
  	cong: cong,
  	congdot: congdot,
  	Congruent: Congruent,
  	conint: conint,
  	Conint: Conint,
  	ContourIntegral: ContourIntegral,
  	copf: copf,
  	Copf: Copf,
  	coprod: coprod,
  	Coproduct: Coproduct,
  	copy: copy,
  	COPY: COPY,
  	copysr: copysr,
  	CounterClockwiseContourIntegral: CounterClockwiseContourIntegral,
  	crarr: crarr,
  	cross: cross,
  	Cross: Cross,
  	Cscr: Cscr,
  	cscr: cscr,
  	csub: csub,
  	csube: csube,
  	csup: csup,
  	csupe: csupe,
  	ctdot: ctdot,
  	cudarrl: cudarrl,
  	cudarrr: cudarrr,
  	cuepr: cuepr,
  	cuesc: cuesc,
  	cularr: cularr,
  	cularrp: cularrp,
  	cupbrcap: cupbrcap,
  	cupcap: cupcap,
  	CupCap: CupCap,
  	cup: cup,
  	Cup: Cup,
  	cupcup: cupcup,
  	cupdot: cupdot,
  	cupor: cupor,
  	cups: cups,
  	curarr: curarr,
  	curarrm: curarrm,
  	curlyeqprec: curlyeqprec,
  	curlyeqsucc: curlyeqsucc,
  	curlyvee: curlyvee,
  	curlywedge: curlywedge,
  	curren: curren,
  	curvearrowleft: curvearrowleft,
  	curvearrowright: curvearrowright,
  	cuvee: cuvee,
  	cuwed: cuwed,
  	cwconint: cwconint,
  	cwint: cwint,
  	cylcty: cylcty,
  	dagger: dagger,
  	Dagger: Dagger,
  	daleth: daleth,
  	darr: darr,
  	Darr: Darr,
  	dArr: dArr,
  	dash: dash,
  	Dashv: Dashv,
  	dashv: dashv,
  	dbkarow: dbkarow,
  	dblac: dblac,
  	Dcaron: Dcaron,
  	dcaron: dcaron,
  	Dcy: Dcy,
  	dcy: dcy,
  	ddagger: ddagger,
  	ddarr: ddarr,
  	DD: DD,
  	dd: dd,
  	DDotrahd: DDotrahd,
  	ddotseq: ddotseq,
  	deg: deg,
  	Del: Del,
  	Delta: Delta,
  	delta: delta,
  	demptyv: demptyv,
  	dfisht: dfisht,
  	Dfr: Dfr,
  	dfr: dfr,
  	dHar: dHar,
  	dharl: dharl,
  	dharr: dharr,
  	DiacriticalAcute: DiacriticalAcute,
  	DiacriticalDot: DiacriticalDot,
  	DiacriticalDoubleAcute: DiacriticalDoubleAcute,
  	DiacriticalGrave: DiacriticalGrave,
  	DiacriticalTilde: DiacriticalTilde,
  	diam: diam,
  	diamond: diamond,
  	Diamond: Diamond,
  	diamondsuit: diamondsuit,
  	diams: diams,
  	die: die,
  	DifferentialD: DifferentialD,
  	digamma: digamma,
  	disin: disin,
  	div: div,
  	divide: divide,
  	divideontimes: divideontimes,
  	divonx: divonx,
  	DJcy: DJcy,
  	djcy: djcy,
  	dlcorn: dlcorn,
  	dlcrop: dlcrop,
  	dollar: dollar,
  	Dopf: Dopf,
  	dopf: dopf,
  	Dot: Dot,
  	dot: dot,
  	DotDot: DotDot,
  	doteq: doteq,
  	doteqdot: doteqdot,
  	DotEqual: DotEqual,
  	dotminus: dotminus,
  	dotplus: dotplus,
  	dotsquare: dotsquare,
  	doublebarwedge: doublebarwedge,
  	DoubleContourIntegral: DoubleContourIntegral,
  	DoubleDot: DoubleDot,
  	DoubleDownArrow: DoubleDownArrow,
  	DoubleLeftArrow: DoubleLeftArrow,
  	DoubleLeftRightArrow: DoubleLeftRightArrow,
  	DoubleLeftTee: DoubleLeftTee,
  	DoubleLongLeftArrow: DoubleLongLeftArrow,
  	DoubleLongLeftRightArrow: DoubleLongLeftRightArrow,
  	DoubleLongRightArrow: DoubleLongRightArrow,
  	DoubleRightArrow: DoubleRightArrow,
  	DoubleRightTee: DoubleRightTee,
  	DoubleUpArrow: DoubleUpArrow,
  	DoubleUpDownArrow: DoubleUpDownArrow,
  	DoubleVerticalBar: DoubleVerticalBar,
  	DownArrowBar: DownArrowBar,
  	downarrow: downarrow,
  	DownArrow: DownArrow,
  	Downarrow: Downarrow,
  	DownArrowUpArrow: DownArrowUpArrow,
  	DownBreve: DownBreve,
  	downdownarrows: downdownarrows,
  	downharpoonleft: downharpoonleft,
  	downharpoonright: downharpoonright,
  	DownLeftRightVector: DownLeftRightVector,
  	DownLeftTeeVector: DownLeftTeeVector,
  	DownLeftVectorBar: DownLeftVectorBar,
  	DownLeftVector: DownLeftVector,
  	DownRightTeeVector: DownRightTeeVector,
  	DownRightVectorBar: DownRightVectorBar,
  	DownRightVector: DownRightVector,
  	DownTeeArrow: DownTeeArrow,
  	DownTee: DownTee,
  	drbkarow: drbkarow,
  	drcorn: drcorn,
  	drcrop: drcrop,
  	Dscr: Dscr,
  	dscr: dscr,
  	DScy: DScy,
  	dscy: dscy,
  	dsol: dsol,
  	Dstrok: Dstrok,
  	dstrok: dstrok,
  	dtdot: dtdot,
  	dtri: dtri,
  	dtrif: dtrif,
  	duarr: duarr,
  	duhar: duhar,
  	dwangle: dwangle,
  	DZcy: DZcy,
  	dzcy: dzcy,
  	dzigrarr: dzigrarr,
  	Eacute: Eacute,
  	eacute: eacute,
  	easter: easter,
  	Ecaron: Ecaron,
  	ecaron: ecaron,
  	Ecirc: Ecirc,
  	ecirc: ecirc,
  	ecir: ecir,
  	ecolon: ecolon,
  	Ecy: Ecy,
  	ecy: ecy,
  	eDDot: eDDot,
  	Edot: Edot,
  	edot: edot,
  	eDot: eDot,
  	ee: ee,
  	efDot: efDot,
  	Efr: Efr,
  	efr: efr,
  	eg: eg,
  	Egrave: Egrave,
  	egrave: egrave,
  	egs: egs,
  	egsdot: egsdot,
  	el: el,
  	Element: Element,
  	elinters: elinters,
  	ell: ell,
  	els: els,
  	elsdot: elsdot,
  	Emacr: Emacr,
  	emacr: emacr,
  	empty: empty,
  	emptyset: emptyset,
  	EmptySmallSquare: EmptySmallSquare,
  	emptyv: emptyv,
  	EmptyVerySmallSquare: EmptyVerySmallSquare,
  	emsp13: emsp13,
  	emsp14: emsp14,
  	emsp: emsp,
  	ENG: ENG,
  	eng: eng,
  	ensp: ensp,
  	Eogon: Eogon,
  	eogon: eogon,
  	Eopf: Eopf,
  	eopf: eopf,
  	epar: epar,
  	eparsl: eparsl,
  	eplus: eplus,
  	epsi: epsi,
  	Epsilon: Epsilon,
  	epsilon: epsilon,
  	epsiv: epsiv,
  	eqcirc: eqcirc,
  	eqcolon: eqcolon,
  	eqsim: eqsim,
  	eqslantgtr: eqslantgtr,
  	eqslantless: eqslantless,
  	Equal: Equal,
  	equals: equals,
  	EqualTilde: EqualTilde,
  	equest: equest,
  	Equilibrium: Equilibrium,
  	equiv: equiv,
  	equivDD: equivDD,
  	eqvparsl: eqvparsl,
  	erarr: erarr,
  	erDot: erDot,
  	escr: escr,
  	Escr: Escr,
  	esdot: esdot,
  	Esim: Esim,
  	esim: esim,
  	Eta: Eta,
  	eta: eta,
  	ETH: ETH,
  	eth: eth,
  	Euml: Euml,
  	euml: euml,
  	euro: euro,
  	excl: excl,
  	exist: exist,
  	Exists: Exists,
  	expectation: expectation,
  	exponentiale: exponentiale,
  	ExponentialE: ExponentialE,
  	fallingdotseq: fallingdotseq,
  	Fcy: Fcy,
  	fcy: fcy,
  	female: female,
  	ffilig: ffilig,
  	fflig: fflig,
  	ffllig: ffllig,
  	Ffr: Ffr,
  	ffr: ffr,
  	filig: filig,
  	FilledSmallSquare: FilledSmallSquare,
  	FilledVerySmallSquare: FilledVerySmallSquare,
  	fjlig: fjlig,
  	flat: flat,
  	fllig: fllig,
  	fltns: fltns,
  	fnof: fnof,
  	Fopf: Fopf,
  	fopf: fopf,
  	forall: forall,
  	ForAll: ForAll,
  	fork: fork,
  	forkv: forkv,
  	Fouriertrf: Fouriertrf,
  	fpartint: fpartint,
  	frac12: frac12,
  	frac13: frac13,
  	frac14: frac14,
  	frac15: frac15,
  	frac16: frac16,
  	frac18: frac18,
  	frac23: frac23,
  	frac25: frac25,
  	frac34: frac34,
  	frac35: frac35,
  	frac38: frac38,
  	frac45: frac45,
  	frac56: frac56,
  	frac58: frac58,
  	frac78: frac78,
  	frasl: frasl,
  	frown: frown,
  	fscr: fscr,
  	Fscr: Fscr,
  	gacute: gacute,
  	Gamma: Gamma,
  	gamma: gamma,
  	Gammad: Gammad,
  	gammad: gammad,
  	gap: gap,
  	Gbreve: Gbreve,
  	gbreve: gbreve,
  	Gcedil: Gcedil,
  	Gcirc: Gcirc,
  	gcirc: gcirc,
  	Gcy: Gcy,
  	gcy: gcy,
  	Gdot: Gdot,
  	gdot: gdot,
  	ge: ge,
  	gE: gE,
  	gEl: gEl,
  	gel: gel,
  	geq: geq,
  	geqq: geqq,
  	geqslant: geqslant,
  	gescc: gescc,
  	ges: ges,
  	gesdot: gesdot,
  	gesdoto: gesdoto,
  	gesdotol: gesdotol,
  	gesl: gesl,
  	gesles: gesles,
  	Gfr: Gfr,
  	gfr: gfr,
  	gg: gg,
  	Gg: Gg,
  	ggg: ggg,
  	gimel: gimel,
  	GJcy: GJcy,
  	gjcy: gjcy,
  	gla: gla,
  	gl: gl,
  	glE: glE,
  	glj: glj,
  	gnap: gnap,
  	gnapprox: gnapprox,
  	gne: gne,
  	gnE: gnE,
  	gneq: gneq,
  	gneqq: gneqq,
  	gnsim: gnsim,
  	Gopf: Gopf,
  	gopf: gopf,
  	grave: grave,
  	GreaterEqual: GreaterEqual,
  	GreaterEqualLess: GreaterEqualLess,
  	GreaterFullEqual: GreaterFullEqual,
  	GreaterGreater: GreaterGreater,
  	GreaterLess: GreaterLess,
  	GreaterSlantEqual: GreaterSlantEqual,
  	GreaterTilde: GreaterTilde,
  	Gscr: Gscr,
  	gscr: gscr,
  	gsim: gsim,
  	gsime: gsime,
  	gsiml: gsiml,
  	gtcc: gtcc,
  	gtcir: gtcir,
  	gt: gt$1,
  	GT: GT,
  	Gt: Gt,
  	gtdot: gtdot,
  	gtlPar: gtlPar,
  	gtquest: gtquest,
  	gtrapprox: gtrapprox,
  	gtrarr: gtrarr,
  	gtrdot: gtrdot,
  	gtreqless: gtreqless,
  	gtreqqless: gtreqqless,
  	gtrless: gtrless,
  	gtrsim: gtrsim,
  	gvertneqq: gvertneqq,
  	gvnE: gvnE,
  	Hacek: Hacek,
  	hairsp: hairsp,
  	half: half,
  	hamilt: hamilt,
  	HARDcy: HARDcy,
  	hardcy: hardcy,
  	harrcir: harrcir,
  	harr: harr,
  	hArr: hArr,
  	harrw: harrw,
  	Hat: Hat,
  	hbar: hbar,
  	Hcirc: Hcirc,
  	hcirc: hcirc,
  	hearts: hearts,
  	heartsuit: heartsuit,
  	hellip: hellip,
  	hercon: hercon,
  	hfr: hfr,
  	Hfr: Hfr,
  	HilbertSpace: HilbertSpace,
  	hksearow: hksearow,
  	hkswarow: hkswarow,
  	hoarr: hoarr,
  	homtht: homtht,
  	hookleftarrow: hookleftarrow,
  	hookrightarrow: hookrightarrow,
  	hopf: hopf,
  	Hopf: Hopf,
  	horbar: horbar,
  	HorizontalLine: HorizontalLine,
  	hscr: hscr,
  	Hscr: Hscr,
  	hslash: hslash,
  	Hstrok: Hstrok,
  	hstrok: hstrok,
  	HumpDownHump: HumpDownHump,
  	HumpEqual: HumpEqual,
  	hybull: hybull,
  	hyphen: hyphen,
  	Iacute: Iacute,
  	iacute: iacute,
  	ic: ic,
  	Icirc: Icirc,
  	icirc: icirc,
  	Icy: Icy,
  	icy: icy,
  	Idot: Idot,
  	IEcy: IEcy,
  	iecy: iecy,
  	iexcl: iexcl,
  	iff: iff,
  	ifr: ifr,
  	Ifr: Ifr,
  	Igrave: Igrave,
  	igrave: igrave,
  	ii: ii,
  	iiiint: iiiint,
  	iiint: iiint,
  	iinfin: iinfin,
  	iiota: iiota,
  	IJlig: IJlig,
  	ijlig: ijlig,
  	Imacr: Imacr,
  	imacr: imacr,
  	image: image,
  	ImaginaryI: ImaginaryI,
  	imagline: imagline,
  	imagpart: imagpart,
  	imath: imath,
  	Im: Im,
  	imof: imof,
  	imped: imped,
  	Implies: Implies,
  	incare: incare,
  	infin: infin,
  	infintie: infintie,
  	inodot: inodot,
  	intcal: intcal,
  	int: int,
  	Int: Int,
  	integers: integers,
  	Integral: Integral,
  	intercal: intercal,
  	Intersection: Intersection,
  	intlarhk: intlarhk,
  	intprod: intprod,
  	InvisibleComma: InvisibleComma,
  	InvisibleTimes: InvisibleTimes,
  	IOcy: IOcy,
  	iocy: iocy,
  	Iogon: Iogon,
  	iogon: iogon,
  	Iopf: Iopf,
  	iopf: iopf,
  	Iota: Iota,
  	iota: iota,
  	iprod: iprod,
  	iquest: iquest,
  	iscr: iscr,
  	Iscr: Iscr,
  	isin: isin,
  	isindot: isindot,
  	isinE: isinE,
  	isins: isins,
  	isinsv: isinsv,
  	isinv: isinv,
  	it: it,
  	Itilde: Itilde,
  	itilde: itilde,
  	Iukcy: Iukcy,
  	iukcy: iukcy,
  	Iuml: Iuml,
  	iuml: iuml,
  	Jcirc: Jcirc,
  	jcirc: jcirc,
  	Jcy: Jcy,
  	jcy: jcy,
  	Jfr: Jfr,
  	jfr: jfr,
  	jmath: jmath,
  	Jopf: Jopf,
  	jopf: jopf,
  	Jscr: Jscr,
  	jscr: jscr,
  	Jsercy: Jsercy,
  	jsercy: jsercy,
  	Jukcy: Jukcy,
  	jukcy: jukcy,
  	Kappa: Kappa,
  	kappa: kappa,
  	kappav: kappav,
  	Kcedil: Kcedil,
  	kcedil: kcedil,
  	Kcy: Kcy,
  	kcy: kcy,
  	Kfr: Kfr,
  	kfr: kfr,
  	kgreen: kgreen,
  	KHcy: KHcy,
  	khcy: khcy,
  	KJcy: KJcy,
  	kjcy: kjcy,
  	Kopf: Kopf,
  	kopf: kopf,
  	Kscr: Kscr,
  	kscr: kscr,
  	lAarr: lAarr,
  	Lacute: Lacute,
  	lacute: lacute,
  	laemptyv: laemptyv,
  	lagran: lagran,
  	Lambda: Lambda,
  	lambda: lambda,
  	lang: lang,
  	Lang: Lang,
  	langd: langd,
  	langle: langle,
  	lap: lap,
  	Laplacetrf: Laplacetrf,
  	laquo: laquo,
  	larrb: larrb,
  	larrbfs: larrbfs,
  	larr: larr,
  	Larr: Larr,
  	lArr: lArr,
  	larrfs: larrfs,
  	larrhk: larrhk,
  	larrlp: larrlp,
  	larrpl: larrpl,
  	larrsim: larrsim,
  	larrtl: larrtl,
  	latail: latail,
  	lAtail: lAtail,
  	lat: lat,
  	late: late,
  	lates: lates,
  	lbarr: lbarr,
  	lBarr: lBarr,
  	lbbrk: lbbrk,
  	lbrace: lbrace,
  	lbrack: lbrack,
  	lbrke: lbrke,
  	lbrksld: lbrksld,
  	lbrkslu: lbrkslu,
  	Lcaron: Lcaron,
  	lcaron: lcaron,
  	Lcedil: Lcedil,
  	lcedil: lcedil,
  	lceil: lceil,
  	lcub: lcub,
  	Lcy: Lcy,
  	lcy: lcy,
  	ldca: ldca,
  	ldquo: ldquo,
  	ldquor: ldquor,
  	ldrdhar: ldrdhar,
  	ldrushar: ldrushar,
  	ldsh: ldsh,
  	le: le,
  	lE: lE,
  	LeftAngleBracket: LeftAngleBracket,
  	LeftArrowBar: LeftArrowBar,
  	leftarrow: leftarrow,
  	LeftArrow: LeftArrow,
  	Leftarrow: Leftarrow,
  	LeftArrowRightArrow: LeftArrowRightArrow,
  	leftarrowtail: leftarrowtail,
  	LeftCeiling: LeftCeiling,
  	LeftDoubleBracket: LeftDoubleBracket,
  	LeftDownTeeVector: LeftDownTeeVector,
  	LeftDownVectorBar: LeftDownVectorBar,
  	LeftDownVector: LeftDownVector,
  	LeftFloor: LeftFloor,
  	leftharpoondown: leftharpoondown,
  	leftharpoonup: leftharpoonup,
  	leftleftarrows: leftleftarrows,
  	leftrightarrow: leftrightarrow,
  	LeftRightArrow: LeftRightArrow,
  	Leftrightarrow: Leftrightarrow,
  	leftrightarrows: leftrightarrows,
  	leftrightharpoons: leftrightharpoons,
  	leftrightsquigarrow: leftrightsquigarrow,
  	LeftRightVector: LeftRightVector,
  	LeftTeeArrow: LeftTeeArrow,
  	LeftTee: LeftTee,
  	LeftTeeVector: LeftTeeVector,
  	leftthreetimes: leftthreetimes,
  	LeftTriangleBar: LeftTriangleBar,
  	LeftTriangle: LeftTriangle,
  	LeftTriangleEqual: LeftTriangleEqual,
  	LeftUpDownVector: LeftUpDownVector,
  	LeftUpTeeVector: LeftUpTeeVector,
  	LeftUpVectorBar: LeftUpVectorBar,
  	LeftUpVector: LeftUpVector,
  	LeftVectorBar: LeftVectorBar,
  	LeftVector: LeftVector,
  	lEg: lEg,
  	leg: leg,
  	leq: leq,
  	leqq: leqq,
  	leqslant: leqslant,
  	lescc: lescc,
  	les: les,
  	lesdot: lesdot,
  	lesdoto: lesdoto,
  	lesdotor: lesdotor,
  	lesg: lesg,
  	lesges: lesges,
  	lessapprox: lessapprox,
  	lessdot: lessdot,
  	lesseqgtr: lesseqgtr,
  	lesseqqgtr: lesseqqgtr,
  	LessEqualGreater: LessEqualGreater,
  	LessFullEqual: LessFullEqual,
  	LessGreater: LessGreater,
  	lessgtr: lessgtr,
  	LessLess: LessLess,
  	lesssim: lesssim,
  	LessSlantEqual: LessSlantEqual,
  	LessTilde: LessTilde,
  	lfisht: lfisht,
  	lfloor: lfloor,
  	Lfr: Lfr,
  	lfr: lfr,
  	lg: lg,
  	lgE: lgE,
  	lHar: lHar,
  	lhard: lhard,
  	lharu: lharu,
  	lharul: lharul,
  	lhblk: lhblk,
  	LJcy: LJcy,
  	ljcy: ljcy,
  	llarr: llarr,
  	ll: ll,
  	Ll: Ll,
  	llcorner: llcorner,
  	Lleftarrow: Lleftarrow,
  	llhard: llhard,
  	lltri: lltri,
  	Lmidot: Lmidot,
  	lmidot: lmidot,
  	lmoustache: lmoustache,
  	lmoust: lmoust,
  	lnap: lnap,
  	lnapprox: lnapprox,
  	lne: lne,
  	lnE: lnE,
  	lneq: lneq,
  	lneqq: lneqq,
  	lnsim: lnsim,
  	loang: loang,
  	loarr: loarr,
  	lobrk: lobrk,
  	longleftarrow: longleftarrow,
  	LongLeftArrow: LongLeftArrow,
  	Longleftarrow: Longleftarrow,
  	longleftrightarrow: longleftrightarrow,
  	LongLeftRightArrow: LongLeftRightArrow,
  	Longleftrightarrow: Longleftrightarrow,
  	longmapsto: longmapsto,
  	longrightarrow: longrightarrow,
  	LongRightArrow: LongRightArrow,
  	Longrightarrow: Longrightarrow,
  	looparrowleft: looparrowleft,
  	looparrowright: looparrowright,
  	lopar: lopar,
  	Lopf: Lopf,
  	lopf: lopf,
  	loplus: loplus,
  	lotimes: lotimes,
  	lowast: lowast,
  	lowbar: lowbar,
  	LowerLeftArrow: LowerLeftArrow,
  	LowerRightArrow: LowerRightArrow,
  	loz: loz,
  	lozenge: lozenge,
  	lozf: lozf,
  	lpar: lpar,
  	lparlt: lparlt,
  	lrarr: lrarr,
  	lrcorner: lrcorner,
  	lrhar: lrhar,
  	lrhard: lrhard,
  	lrm: lrm,
  	lrtri: lrtri,
  	lsaquo: lsaquo,
  	lscr: lscr,
  	Lscr: Lscr,
  	lsh: lsh,
  	Lsh: Lsh,
  	lsim: lsim,
  	lsime: lsime,
  	lsimg: lsimg,
  	lsqb: lsqb,
  	lsquo: lsquo,
  	lsquor: lsquor,
  	Lstrok: Lstrok,
  	lstrok: lstrok,
  	ltcc: ltcc,
  	ltcir: ltcir,
  	lt: lt$1,
  	LT: LT,
  	Lt: Lt,
  	ltdot: ltdot,
  	lthree: lthree,
  	ltimes: ltimes,
  	ltlarr: ltlarr,
  	ltquest: ltquest,
  	ltri: ltri,
  	ltrie: ltrie,
  	ltrif: ltrif,
  	ltrPar: ltrPar,
  	lurdshar: lurdshar,
  	luruhar: luruhar,
  	lvertneqq: lvertneqq,
  	lvnE: lvnE,
  	macr: macr,
  	male: male,
  	malt: malt,
  	maltese: maltese,
  	map: map$1,
  	mapsto: mapsto,
  	mapstodown: mapstodown,
  	mapstoleft: mapstoleft,
  	mapstoup: mapstoup,
  	marker: marker,
  	mcomma: mcomma,
  	Mcy: Mcy,
  	mcy: mcy,
  	mdash: mdash,
  	mDDot: mDDot,
  	measuredangle: measuredangle,
  	MediumSpace: MediumSpace,
  	Mellintrf: Mellintrf,
  	Mfr: Mfr,
  	mfr: mfr,
  	mho: mho,
  	micro: micro,
  	midast: midast,
  	midcir: midcir,
  	mid: mid,
  	middot: middot,
  	minusb: minusb,
  	minus: minus,
  	minusd: minusd,
  	minusdu: minusdu,
  	MinusPlus: MinusPlus,
  	mlcp: mlcp,
  	mldr: mldr,
  	mnplus: mnplus,
  	models: models,
  	Mopf: Mopf,
  	mopf: mopf,
  	mp: mp,
  	mscr: mscr,
  	Mscr: Mscr,
  	mstpos: mstpos,
  	Mu: Mu,
  	mu: mu,
  	multimap: multimap,
  	mumap: mumap,
  	nabla: nabla,
  	Nacute: Nacute,
  	nacute: nacute,
  	nang: nang,
  	nap: nap,
  	napE: napE,
  	napid: napid,
  	napos: napos,
  	napprox: napprox,
  	natural: natural,
  	naturals: naturals,
  	natur: natur,
  	nbsp: nbsp,
  	nbump: nbump,
  	nbumpe: nbumpe,
  	ncap: ncap,
  	Ncaron: Ncaron,
  	ncaron: ncaron,
  	Ncedil: Ncedil,
  	ncedil: ncedil,
  	ncong: ncong,
  	ncongdot: ncongdot,
  	ncup: ncup,
  	Ncy: Ncy,
  	ncy: ncy,
  	ndash: ndash,
  	nearhk: nearhk,
  	nearr: nearr,
  	neArr: neArr,
  	nearrow: nearrow,
  	ne: ne,
  	nedot: nedot,
  	NegativeMediumSpace: NegativeMediumSpace,
  	NegativeThickSpace: NegativeThickSpace,
  	NegativeThinSpace: NegativeThinSpace,
  	NegativeVeryThinSpace: NegativeVeryThinSpace,
  	nequiv: nequiv,
  	nesear: nesear,
  	nesim: nesim,
  	NestedGreaterGreater: NestedGreaterGreater,
  	NestedLessLess: NestedLessLess,
  	NewLine: NewLine,
  	nexist: nexist,
  	nexists: nexists,
  	Nfr: Nfr,
  	nfr: nfr,
  	ngE: ngE,
  	nge: nge,
  	ngeq: ngeq,
  	ngeqq: ngeqq,
  	ngeqslant: ngeqslant,
  	nges: nges,
  	nGg: nGg,
  	ngsim: ngsim,
  	nGt: nGt,
  	ngt: ngt,
  	ngtr: ngtr,
  	nGtv: nGtv,
  	nharr: nharr,
  	nhArr: nhArr,
  	nhpar: nhpar,
  	ni: ni,
  	nis: nis,
  	nisd: nisd,
  	niv: niv,
  	NJcy: NJcy,
  	njcy: njcy,
  	nlarr: nlarr,
  	nlArr: nlArr,
  	nldr: nldr,
  	nlE: nlE,
  	nle: nle,
  	nleftarrow: nleftarrow,
  	nLeftarrow: nLeftarrow,
  	nleftrightarrow: nleftrightarrow,
  	nLeftrightarrow: nLeftrightarrow,
  	nleq: nleq,
  	nleqq: nleqq,
  	nleqslant: nleqslant,
  	nles: nles,
  	nless: nless,
  	nLl: nLl,
  	nlsim: nlsim,
  	nLt: nLt,
  	nlt: nlt,
  	nltri: nltri,
  	nltrie: nltrie,
  	nLtv: nLtv,
  	nmid: nmid,
  	NoBreak: NoBreak,
  	NonBreakingSpace: NonBreakingSpace,
  	nopf: nopf,
  	Nopf: Nopf,
  	Not: Not,
  	not: not,
  	NotCongruent: NotCongruent,
  	NotCupCap: NotCupCap,
  	NotDoubleVerticalBar: NotDoubleVerticalBar,
  	NotElement: NotElement,
  	NotEqual: NotEqual,
  	NotEqualTilde: NotEqualTilde,
  	NotExists: NotExists,
  	NotGreater: NotGreater,
  	NotGreaterEqual: NotGreaterEqual,
  	NotGreaterFullEqual: NotGreaterFullEqual,
  	NotGreaterGreater: NotGreaterGreater,
  	NotGreaterLess: NotGreaterLess,
  	NotGreaterSlantEqual: NotGreaterSlantEqual,
  	NotGreaterTilde: NotGreaterTilde,
  	NotHumpDownHump: NotHumpDownHump,
  	NotHumpEqual: NotHumpEqual,
  	notin: notin,
  	notindot: notindot,
  	notinE: notinE,
  	notinva: notinva,
  	notinvb: notinvb,
  	notinvc: notinvc,
  	NotLeftTriangleBar: NotLeftTriangleBar,
  	NotLeftTriangle: NotLeftTriangle,
  	NotLeftTriangleEqual: NotLeftTriangleEqual,
  	NotLess: NotLess,
  	NotLessEqual: NotLessEqual,
  	NotLessGreater: NotLessGreater,
  	NotLessLess: NotLessLess,
  	NotLessSlantEqual: NotLessSlantEqual,
  	NotLessTilde: NotLessTilde,
  	NotNestedGreaterGreater: NotNestedGreaterGreater,
  	NotNestedLessLess: NotNestedLessLess,
  	notni: notni,
  	notniva: notniva,
  	notnivb: notnivb,
  	notnivc: notnivc,
  	NotPrecedes: NotPrecedes,
  	NotPrecedesEqual: NotPrecedesEqual,
  	NotPrecedesSlantEqual: NotPrecedesSlantEqual,
  	NotReverseElement: NotReverseElement,
  	NotRightTriangleBar: NotRightTriangleBar,
  	NotRightTriangle: NotRightTriangle,
  	NotRightTriangleEqual: NotRightTriangleEqual,
  	NotSquareSubset: NotSquareSubset,
  	NotSquareSubsetEqual: NotSquareSubsetEqual,
  	NotSquareSuperset: NotSquareSuperset,
  	NotSquareSupersetEqual: NotSquareSupersetEqual,
  	NotSubset: NotSubset,
  	NotSubsetEqual: NotSubsetEqual,
  	NotSucceeds: NotSucceeds,
  	NotSucceedsEqual: NotSucceedsEqual,
  	NotSucceedsSlantEqual: NotSucceedsSlantEqual,
  	NotSucceedsTilde: NotSucceedsTilde,
  	NotSuperset: NotSuperset,
  	NotSupersetEqual: NotSupersetEqual,
  	NotTilde: NotTilde,
  	NotTildeEqual: NotTildeEqual,
  	NotTildeFullEqual: NotTildeFullEqual,
  	NotTildeTilde: NotTildeTilde,
  	NotVerticalBar: NotVerticalBar,
  	nparallel: nparallel,
  	npar: npar,
  	nparsl: nparsl,
  	npart: npart,
  	npolint: npolint,
  	npr: npr,
  	nprcue: nprcue,
  	nprec: nprec,
  	npreceq: npreceq,
  	npre: npre,
  	nrarrc: nrarrc,
  	nrarr: nrarr,
  	nrArr: nrArr,
  	nrarrw: nrarrw,
  	nrightarrow: nrightarrow,
  	nRightarrow: nRightarrow,
  	nrtri: nrtri,
  	nrtrie: nrtrie,
  	nsc: nsc,
  	nsccue: nsccue,
  	nsce: nsce,
  	Nscr: Nscr,
  	nscr: nscr,
  	nshortmid: nshortmid,
  	nshortparallel: nshortparallel,
  	nsim: nsim,
  	nsime: nsime,
  	nsimeq: nsimeq,
  	nsmid: nsmid,
  	nspar: nspar,
  	nsqsube: nsqsube,
  	nsqsupe: nsqsupe,
  	nsub: nsub,
  	nsubE: nsubE,
  	nsube: nsube,
  	nsubset: nsubset,
  	nsubseteq: nsubseteq,
  	nsubseteqq: nsubseteqq,
  	nsucc: nsucc,
  	nsucceq: nsucceq,
  	nsup: nsup,
  	nsupE: nsupE,
  	nsupe: nsupe,
  	nsupset: nsupset,
  	nsupseteq: nsupseteq,
  	nsupseteqq: nsupseteqq,
  	ntgl: ntgl,
  	Ntilde: Ntilde,
  	ntilde: ntilde,
  	ntlg: ntlg,
  	ntriangleleft: ntriangleleft,
  	ntrianglelefteq: ntrianglelefteq,
  	ntriangleright: ntriangleright,
  	ntrianglerighteq: ntrianglerighteq,
  	Nu: Nu,
  	nu: nu,
  	num: num,
  	numero: numero,
  	numsp: numsp,
  	nvap: nvap,
  	nvdash: nvdash,
  	nvDash: nvDash,
  	nVdash: nVdash,
  	nVDash: nVDash,
  	nvge: nvge,
  	nvgt: nvgt,
  	nvHarr: nvHarr,
  	nvinfin: nvinfin,
  	nvlArr: nvlArr,
  	nvle: nvle,
  	nvlt: nvlt,
  	nvltrie: nvltrie,
  	nvrArr: nvrArr,
  	nvrtrie: nvrtrie,
  	nvsim: nvsim,
  	nwarhk: nwarhk,
  	nwarr: nwarr,
  	nwArr: nwArr,
  	nwarrow: nwarrow,
  	nwnear: nwnear,
  	Oacute: Oacute,
  	oacute: oacute,
  	oast: oast,
  	Ocirc: Ocirc,
  	ocirc: ocirc,
  	ocir: ocir,
  	Ocy: Ocy,
  	ocy: ocy,
  	odash: odash,
  	Odblac: Odblac,
  	odblac: odblac,
  	odiv: odiv,
  	odot: odot,
  	odsold: odsold,
  	OElig: OElig,
  	oelig: oelig,
  	ofcir: ofcir,
  	Ofr: Ofr,
  	ofr: ofr,
  	ogon: ogon,
  	Ograve: Ograve,
  	ograve: ograve,
  	ogt: ogt,
  	ohbar: ohbar,
  	ohm: ohm,
  	oint: oint,
  	olarr: olarr,
  	olcir: olcir,
  	olcross: olcross,
  	oline: oline,
  	olt: olt,
  	Omacr: Omacr,
  	omacr: omacr,
  	Omega: Omega,
  	omega: omega,
  	Omicron: Omicron,
  	omicron: omicron,
  	omid: omid,
  	ominus: ominus,
  	Oopf: Oopf,
  	oopf: oopf,
  	opar: opar,
  	OpenCurlyDoubleQuote: OpenCurlyDoubleQuote,
  	OpenCurlyQuote: OpenCurlyQuote,
  	operp: operp,
  	oplus: oplus,
  	orarr: orarr,
  	Or: Or,
  	or: or,
  	ord: ord,
  	order: order,
  	orderof: orderof,
  	ordf: ordf,
  	ordm: ordm,
  	origof: origof,
  	oror: oror,
  	orslope: orslope,
  	orv: orv,
  	oS: oS,
  	Oscr: Oscr,
  	oscr: oscr,
  	Oslash: Oslash,
  	oslash: oslash,
  	osol: osol,
  	Otilde: Otilde,
  	otilde: otilde,
  	otimesas: otimesas,
  	Otimes: Otimes,
  	otimes: otimes,
  	Ouml: Ouml,
  	ouml: ouml,
  	ovbar: ovbar,
  	OverBar: OverBar,
  	OverBrace: OverBrace,
  	OverBracket: OverBracket,
  	OverParenthesis: OverParenthesis,
  	para: para,
  	parallel: parallel,
  	par: par,
  	parsim: parsim,
  	parsl: parsl,
  	part: part,
  	PartialD: PartialD,
  	Pcy: Pcy,
  	pcy: pcy,
  	percnt: percnt,
  	period: period,
  	permil: permil,
  	perp: perp,
  	pertenk: pertenk,
  	Pfr: Pfr,
  	pfr: pfr,
  	Phi: Phi,
  	phi: phi,
  	phiv: phiv,
  	phmmat: phmmat,
  	phone: phone,
  	Pi: Pi,
  	pi: pi,
  	pitchfork: pitchfork,
  	piv: piv,
  	planck: planck,
  	planckh: planckh,
  	plankv: plankv,
  	plusacir: plusacir,
  	plusb: plusb,
  	pluscir: pluscir,
  	plus: plus,
  	plusdo: plusdo,
  	plusdu: plusdu,
  	pluse: pluse,
  	PlusMinus: PlusMinus,
  	plusmn: plusmn,
  	plussim: plussim,
  	plustwo: plustwo,
  	pm: pm,
  	Poincareplane: Poincareplane,
  	pointint: pointint,
  	popf: popf,
  	Popf: Popf,
  	pound: pound,
  	prap: prap,
  	Pr: Pr,
  	pr: pr,
  	prcue: prcue,
  	precapprox: precapprox,
  	prec: prec,
  	preccurlyeq: preccurlyeq,
  	Precedes: Precedes,
  	PrecedesEqual: PrecedesEqual,
  	PrecedesSlantEqual: PrecedesSlantEqual,
  	PrecedesTilde: PrecedesTilde,
  	preceq: preceq,
  	precnapprox: precnapprox,
  	precneqq: precneqq,
  	precnsim: precnsim,
  	pre: pre,
  	prE: prE,
  	precsim: precsim,
  	prime: prime,
  	Prime: Prime,
  	primes: primes,
  	prnap: prnap,
  	prnE: prnE,
  	prnsim: prnsim,
  	prod: prod,
  	Product: Product,
  	profalar: profalar,
  	profline: profline,
  	profsurf: profsurf,
  	prop: prop,
  	Proportional: Proportional,
  	Proportion: Proportion,
  	propto: propto,
  	prsim: prsim,
  	prurel: prurel,
  	Pscr: Pscr,
  	pscr: pscr,
  	Psi: Psi,
  	psi: psi,
  	puncsp: puncsp,
  	Qfr: Qfr,
  	qfr: qfr,
  	qint: qint,
  	qopf: qopf,
  	Qopf: Qopf,
  	qprime: qprime,
  	Qscr: Qscr,
  	qscr: qscr,
  	quaternions: quaternions,
  	quatint: quatint,
  	quest: quest,
  	questeq: questeq,
  	quot: quot$1,
  	QUOT: QUOT,
  	rAarr: rAarr,
  	race: race,
  	Racute: Racute,
  	racute: racute,
  	radic: radic,
  	raemptyv: raemptyv,
  	rang: rang,
  	Rang: Rang,
  	rangd: rangd,
  	range: range,
  	rangle: rangle,
  	raquo: raquo,
  	rarrap: rarrap,
  	rarrb: rarrb,
  	rarrbfs: rarrbfs,
  	rarrc: rarrc,
  	rarr: rarr,
  	Rarr: Rarr,
  	rArr: rArr,
  	rarrfs: rarrfs,
  	rarrhk: rarrhk,
  	rarrlp: rarrlp,
  	rarrpl: rarrpl,
  	rarrsim: rarrsim,
  	Rarrtl: Rarrtl,
  	rarrtl: rarrtl,
  	rarrw: rarrw,
  	ratail: ratail,
  	rAtail: rAtail,
  	ratio: ratio,
  	rationals: rationals,
  	rbarr: rbarr,
  	rBarr: rBarr,
  	RBarr: RBarr,
  	rbbrk: rbbrk,
  	rbrace: rbrace,
  	rbrack: rbrack,
  	rbrke: rbrke,
  	rbrksld: rbrksld,
  	rbrkslu: rbrkslu,
  	Rcaron: Rcaron,
  	rcaron: rcaron,
  	Rcedil: Rcedil,
  	rcedil: rcedil,
  	rceil: rceil,
  	rcub: rcub,
  	Rcy: Rcy,
  	rcy: rcy,
  	rdca: rdca,
  	rdldhar: rdldhar,
  	rdquo: rdquo,
  	rdquor: rdquor,
  	rdsh: rdsh,
  	real: real,
  	realine: realine,
  	realpart: realpart,
  	reals: reals,
  	Re: Re,
  	rect: rect,
  	reg: reg,
  	REG: REG,
  	ReverseElement: ReverseElement,
  	ReverseEquilibrium: ReverseEquilibrium,
  	ReverseUpEquilibrium: ReverseUpEquilibrium,
  	rfisht: rfisht,
  	rfloor: rfloor,
  	rfr: rfr,
  	Rfr: Rfr,
  	rHar: rHar,
  	rhard: rhard,
  	rharu: rharu,
  	rharul: rharul,
  	Rho: Rho,
  	rho: rho,
  	rhov: rhov,
  	RightAngleBracket: RightAngleBracket,
  	RightArrowBar: RightArrowBar,
  	rightarrow: rightarrow,
  	RightArrow: RightArrow,
  	Rightarrow: Rightarrow,
  	RightArrowLeftArrow: RightArrowLeftArrow,
  	rightarrowtail: rightarrowtail,
  	RightCeiling: RightCeiling,
  	RightDoubleBracket: RightDoubleBracket,
  	RightDownTeeVector: RightDownTeeVector,
  	RightDownVectorBar: RightDownVectorBar,
  	RightDownVector: RightDownVector,
  	RightFloor: RightFloor,
  	rightharpoondown: rightharpoondown,
  	rightharpoonup: rightharpoonup,
  	rightleftarrows: rightleftarrows,
  	rightleftharpoons: rightleftharpoons,
  	rightrightarrows: rightrightarrows,
  	rightsquigarrow: rightsquigarrow,
  	RightTeeArrow: RightTeeArrow,
  	RightTee: RightTee,
  	RightTeeVector: RightTeeVector,
  	rightthreetimes: rightthreetimes,
  	RightTriangleBar: RightTriangleBar,
  	RightTriangle: RightTriangle,
  	RightTriangleEqual: RightTriangleEqual,
  	RightUpDownVector: RightUpDownVector,
  	RightUpTeeVector: RightUpTeeVector,
  	RightUpVectorBar: RightUpVectorBar,
  	RightUpVector: RightUpVector,
  	RightVectorBar: RightVectorBar,
  	RightVector: RightVector,
  	ring: ring,
  	risingdotseq: risingdotseq,
  	rlarr: rlarr,
  	rlhar: rlhar,
  	rlm: rlm,
  	rmoustache: rmoustache,
  	rmoust: rmoust,
  	rnmid: rnmid,
  	roang: roang,
  	roarr: roarr,
  	robrk: robrk,
  	ropar: ropar,
  	ropf: ropf,
  	Ropf: Ropf,
  	roplus: roplus,
  	rotimes: rotimes,
  	RoundImplies: RoundImplies,
  	rpar: rpar,
  	rpargt: rpargt,
  	rppolint: rppolint,
  	rrarr: rrarr,
  	Rrightarrow: Rrightarrow,
  	rsaquo: rsaquo,
  	rscr: rscr,
  	Rscr: Rscr,
  	rsh: rsh,
  	Rsh: Rsh,
  	rsqb: rsqb,
  	rsquo: rsquo,
  	rsquor: rsquor,
  	rthree: rthree,
  	rtimes: rtimes,
  	rtri: rtri,
  	rtrie: rtrie,
  	rtrif: rtrif,
  	rtriltri: rtriltri,
  	RuleDelayed: RuleDelayed,
  	ruluhar: ruluhar,
  	rx: rx,
  	Sacute: Sacute,
  	sacute: sacute,
  	sbquo: sbquo,
  	scap: scap,
  	Scaron: Scaron,
  	scaron: scaron,
  	Sc: Sc,
  	sc: sc,
  	sccue: sccue,
  	sce: sce,
  	scE: scE,
  	Scedil: Scedil,
  	scedil: scedil,
  	Scirc: Scirc,
  	scirc: scirc,
  	scnap: scnap,
  	scnE: scnE,
  	scnsim: scnsim,
  	scpolint: scpolint,
  	scsim: scsim,
  	Scy: Scy,
  	scy: scy,
  	sdotb: sdotb,
  	sdot: sdot,
  	sdote: sdote,
  	searhk: searhk,
  	searr: searr,
  	seArr: seArr,
  	searrow: searrow,
  	sect: sect,
  	semi: semi,
  	seswar: seswar,
  	setminus: setminus,
  	setmn: setmn,
  	sext: sext,
  	Sfr: Sfr,
  	sfr: sfr,
  	sfrown: sfrown,
  	sharp: sharp,
  	SHCHcy: SHCHcy,
  	shchcy: shchcy,
  	SHcy: SHcy,
  	shcy: shcy,
  	ShortDownArrow: ShortDownArrow,
  	ShortLeftArrow: ShortLeftArrow,
  	shortmid: shortmid,
  	shortparallel: shortparallel,
  	ShortRightArrow: ShortRightArrow,
  	ShortUpArrow: ShortUpArrow,
  	shy: shy,
  	Sigma: Sigma,
  	sigma: sigma,
  	sigmaf: sigmaf,
  	sigmav: sigmav,
  	sim: sim,
  	simdot: simdot,
  	sime: sime,
  	simeq: simeq,
  	simg: simg,
  	simgE: simgE,
  	siml: siml,
  	simlE: simlE,
  	simne: simne,
  	simplus: simplus,
  	simrarr: simrarr,
  	slarr: slarr,
  	SmallCircle: SmallCircle,
  	smallsetminus: smallsetminus,
  	smashp: smashp,
  	smeparsl: smeparsl,
  	smid: smid,
  	smile: smile,
  	smt: smt,
  	smte: smte,
  	smtes: smtes,
  	SOFTcy: SOFTcy,
  	softcy: softcy,
  	solbar: solbar,
  	solb: solb,
  	sol: sol,
  	Sopf: Sopf,
  	sopf: sopf,
  	spades: spades,
  	spadesuit: spadesuit,
  	spar: spar,
  	sqcap: sqcap,
  	sqcaps: sqcaps,
  	sqcup: sqcup,
  	sqcups: sqcups,
  	Sqrt: Sqrt,
  	sqsub: sqsub,
  	sqsube: sqsube,
  	sqsubset: sqsubset,
  	sqsubseteq: sqsubseteq,
  	sqsup: sqsup,
  	sqsupe: sqsupe,
  	sqsupset: sqsupset,
  	sqsupseteq: sqsupseteq,
  	square: square,
  	Square: Square,
  	SquareIntersection: SquareIntersection,
  	SquareSubset: SquareSubset,
  	SquareSubsetEqual: SquareSubsetEqual,
  	SquareSuperset: SquareSuperset,
  	SquareSupersetEqual: SquareSupersetEqual,
  	SquareUnion: SquareUnion,
  	squarf: squarf,
  	squ: squ,
  	squf: squf,
  	srarr: srarr,
  	Sscr: Sscr,
  	sscr: sscr,
  	ssetmn: ssetmn,
  	ssmile: ssmile,
  	sstarf: sstarf,
  	Star: Star,
  	star: star,
  	starf: starf,
  	straightepsilon: straightepsilon,
  	straightphi: straightphi,
  	strns: strns,
  	sub: sub,
  	Sub: Sub,
  	subdot: subdot,
  	subE: subE,
  	sube: sube,
  	subedot: subedot,
  	submult: submult,
  	subnE: subnE,
  	subne: subne,
  	subplus: subplus,
  	subrarr: subrarr,
  	subset: subset,
  	Subset: Subset,
  	subseteq: subseteq,
  	subseteqq: subseteqq,
  	SubsetEqual: SubsetEqual,
  	subsetneq: subsetneq,
  	subsetneqq: subsetneqq,
  	subsim: subsim,
  	subsub: subsub,
  	subsup: subsup,
  	succapprox: succapprox,
  	succ: succ,
  	succcurlyeq: succcurlyeq,
  	Succeeds: Succeeds,
  	SucceedsEqual: SucceedsEqual,
  	SucceedsSlantEqual: SucceedsSlantEqual,
  	SucceedsTilde: SucceedsTilde,
  	succeq: succeq,
  	succnapprox: succnapprox,
  	succneqq: succneqq,
  	succnsim: succnsim,
  	succsim: succsim,
  	SuchThat: SuchThat,
  	sum: sum,
  	Sum: Sum,
  	sung: sung,
  	sup1: sup1,
  	sup2: sup2,
  	sup3: sup3,
  	sup: sup,
  	Sup: Sup,
  	supdot: supdot,
  	supdsub: supdsub,
  	supE: supE,
  	supe: supe,
  	supedot: supedot,
  	Superset: Superset,
  	SupersetEqual: SupersetEqual,
  	suphsol: suphsol,
  	suphsub: suphsub,
  	suplarr: suplarr,
  	supmult: supmult,
  	supnE: supnE,
  	supne: supne,
  	supplus: supplus,
  	supset: supset,
  	Supset: Supset,
  	supseteq: supseteq,
  	supseteqq: supseteqq,
  	supsetneq: supsetneq,
  	supsetneqq: supsetneqq,
  	supsim: supsim,
  	supsub: supsub,
  	supsup: supsup,
  	swarhk: swarhk,
  	swarr: swarr,
  	swArr: swArr,
  	swarrow: swarrow,
  	swnwar: swnwar,
  	szlig: szlig,
  	Tab: Tab,
  	target: target,
  	Tau: Tau,
  	tau: tau,
  	tbrk: tbrk,
  	Tcaron: Tcaron,
  	tcaron: tcaron,
  	Tcedil: Tcedil,
  	tcedil: tcedil,
  	Tcy: Tcy,
  	tcy: tcy,
  	tdot: tdot,
  	telrec: telrec,
  	Tfr: Tfr,
  	tfr: tfr,
  	there4: there4,
  	therefore: therefore,
  	Therefore: Therefore,
  	Theta: Theta,
  	theta: theta,
  	thetasym: thetasym,
  	thetav: thetav,
  	thickapprox: thickapprox,
  	thicksim: thicksim,
  	ThickSpace: ThickSpace,
  	ThinSpace: ThinSpace,
  	thinsp: thinsp,
  	thkap: thkap,
  	thksim: thksim,
  	THORN: THORN,
  	thorn: thorn,
  	tilde: tilde,
  	Tilde: Tilde,
  	TildeEqual: TildeEqual,
  	TildeFullEqual: TildeFullEqual,
  	TildeTilde: TildeTilde,
  	timesbar: timesbar,
  	timesb: timesb,
  	times: times,
  	timesd: timesd,
  	tint: tint,
  	toea: toea,
  	topbot: topbot,
  	topcir: topcir,
  	top: top,
  	Topf: Topf,
  	topf: topf,
  	topfork: topfork,
  	tosa: tosa,
  	tprime: tprime,
  	trade: trade,
  	TRADE: TRADE,
  	triangle: triangle,
  	triangledown: triangledown,
  	triangleleft: triangleleft,
  	trianglelefteq: trianglelefteq,
  	triangleq: triangleq,
  	triangleright: triangleright,
  	trianglerighteq: trianglerighteq,
  	tridot: tridot,
  	trie: trie,
  	triminus: triminus,
  	TripleDot: TripleDot,
  	triplus: triplus,
  	trisb: trisb,
  	tritime: tritime,
  	trpezium: trpezium,
  	Tscr: Tscr,
  	tscr: tscr,
  	TScy: TScy,
  	tscy: tscy,
  	TSHcy: TSHcy,
  	tshcy: tshcy,
  	Tstrok: Tstrok,
  	tstrok: tstrok,
  	twixt: twixt,
  	twoheadleftarrow: twoheadleftarrow,
  	twoheadrightarrow: twoheadrightarrow,
  	Uacute: Uacute,
  	uacute: uacute,
  	uarr: uarr,
  	Uarr: Uarr,
  	uArr: uArr,
  	Uarrocir: Uarrocir,
  	Ubrcy: Ubrcy,
  	ubrcy: ubrcy,
  	Ubreve: Ubreve,
  	ubreve: ubreve,
  	Ucirc: Ucirc,
  	ucirc: ucirc,
  	Ucy: Ucy,
  	ucy: ucy,
  	udarr: udarr,
  	Udblac: Udblac,
  	udblac: udblac,
  	udhar: udhar,
  	ufisht: ufisht,
  	Ufr: Ufr,
  	ufr: ufr,
  	Ugrave: Ugrave,
  	ugrave: ugrave,
  	uHar: uHar,
  	uharl: uharl,
  	uharr: uharr,
  	uhblk: uhblk,
  	ulcorn: ulcorn,
  	ulcorner: ulcorner,
  	ulcrop: ulcrop,
  	ultri: ultri,
  	Umacr: Umacr,
  	umacr: umacr,
  	uml: uml,
  	UnderBar: UnderBar,
  	UnderBrace: UnderBrace,
  	UnderBracket: UnderBracket,
  	UnderParenthesis: UnderParenthesis,
  	Union: Union,
  	UnionPlus: UnionPlus,
  	Uogon: Uogon,
  	uogon: uogon,
  	Uopf: Uopf,
  	uopf: uopf,
  	UpArrowBar: UpArrowBar,
  	uparrow: uparrow,
  	UpArrow: UpArrow,
  	Uparrow: Uparrow,
  	UpArrowDownArrow: UpArrowDownArrow,
  	updownarrow: updownarrow,
  	UpDownArrow: UpDownArrow,
  	Updownarrow: Updownarrow,
  	UpEquilibrium: UpEquilibrium,
  	upharpoonleft: upharpoonleft,
  	upharpoonright: upharpoonright,
  	uplus: uplus,
  	UpperLeftArrow: UpperLeftArrow,
  	UpperRightArrow: UpperRightArrow,
  	upsi: upsi,
  	Upsi: Upsi,
  	upsih: upsih,
  	Upsilon: Upsilon,
  	upsilon: upsilon,
  	UpTeeArrow: UpTeeArrow,
  	UpTee: UpTee,
  	upuparrows: upuparrows,
  	urcorn: urcorn,
  	urcorner: urcorner,
  	urcrop: urcrop,
  	Uring: Uring,
  	uring: uring,
  	urtri: urtri,
  	Uscr: Uscr,
  	uscr: uscr,
  	utdot: utdot,
  	Utilde: Utilde,
  	utilde: utilde,
  	utri: utri,
  	utrif: utrif,
  	uuarr: uuarr,
  	Uuml: Uuml,
  	uuml: uuml,
  	uwangle: uwangle,
  	vangrt: vangrt,
  	varepsilon: varepsilon,
  	varkappa: varkappa,
  	varnothing: varnothing,
  	varphi: varphi,
  	varpi: varpi,
  	varpropto: varpropto,
  	varr: varr,
  	vArr: vArr,
  	varrho: varrho,
  	varsigma: varsigma,
  	varsubsetneq: varsubsetneq,
  	varsubsetneqq: varsubsetneqq,
  	varsupsetneq: varsupsetneq,
  	varsupsetneqq: varsupsetneqq,
  	vartheta: vartheta,
  	vartriangleleft: vartriangleleft,
  	vartriangleright: vartriangleright,
  	vBar: vBar,
  	Vbar: Vbar,
  	vBarv: vBarv,
  	Vcy: Vcy,
  	vcy: vcy,
  	vdash: vdash,
  	vDash: vDash,
  	Vdash: Vdash,
  	VDash: VDash,
  	Vdashl: Vdashl,
  	veebar: veebar,
  	vee: vee,
  	Vee: Vee,
  	veeeq: veeeq,
  	vellip: vellip,
  	verbar: verbar,
  	Verbar: Verbar,
  	vert: vert,
  	Vert: Vert,
  	VerticalBar: VerticalBar,
  	VerticalLine: VerticalLine,
  	VerticalSeparator: VerticalSeparator,
  	VerticalTilde: VerticalTilde,
  	VeryThinSpace: VeryThinSpace,
  	Vfr: Vfr,
  	vfr: vfr,
  	vltri: vltri,
  	vnsub: vnsub,
  	vnsup: vnsup,
  	Vopf: Vopf,
  	vopf: vopf,
  	vprop: vprop,
  	vrtri: vrtri,
  	Vscr: Vscr,
  	vscr: vscr,
  	vsubnE: vsubnE,
  	vsubne: vsubne,
  	vsupnE: vsupnE,
  	vsupne: vsupne,
  	Vvdash: Vvdash,
  	vzigzag: vzigzag,
  	Wcirc: Wcirc,
  	wcirc: wcirc,
  	wedbar: wedbar,
  	wedge: wedge,
  	Wedge: Wedge,
  	wedgeq: wedgeq,
  	weierp: weierp,
  	Wfr: Wfr,
  	wfr: wfr,
  	Wopf: Wopf,
  	wopf: wopf,
  	wp: wp,
  	wr: wr,
  	wreath: wreath,
  	Wscr: Wscr,
  	wscr: wscr,
  	xcap: xcap,
  	xcirc: xcirc,
  	xcup: xcup,
  	xdtri: xdtri,
  	Xfr: Xfr,
  	xfr: xfr,
  	xharr: xharr,
  	xhArr: xhArr,
  	Xi: Xi,
  	xi: xi,
  	xlarr: xlarr,
  	xlArr: xlArr,
  	xmap: xmap,
  	xnis: xnis,
  	xodot: xodot,
  	Xopf: Xopf,
  	xopf: xopf,
  	xoplus: xoplus,
  	xotime: xotime,
  	xrarr: xrarr,
  	xrArr: xrArr,
  	Xscr: Xscr,
  	xscr: xscr,
  	xsqcup: xsqcup,
  	xuplus: xuplus,
  	xutri: xutri,
  	xvee: xvee,
  	xwedge: xwedge,
  	Yacute: Yacute,
  	yacute: yacute,
  	YAcy: YAcy,
  	yacy: yacy,
  	Ycirc: Ycirc,
  	ycirc: ycirc,
  	Ycy: Ycy,
  	ycy: ycy,
  	yen: yen,
  	Yfr: Yfr,
  	yfr: yfr,
  	YIcy: YIcy,
  	yicy: yicy,
  	Yopf: Yopf,
  	yopf: yopf,
  	Yscr: Yscr,
  	yscr: yscr,
  	YUcy: YUcy,
  	yucy: yucy,
  	yuml: yuml,
  	Yuml: Yuml,
  	Zacute: Zacute,
  	zacute: zacute,
  	Zcaron: Zcaron,
  	zcaron: zcaron,
  	Zcy: Zcy,
  	zcy: zcy,
  	Zdot: Zdot,
  	zdot: zdot,
  	zeetrf: zeetrf,
  	ZeroWidthSpace: ZeroWidthSpace,
  	Zeta: Zeta,
  	zeta: zeta,
  	zfr: zfr,
  	Zfr: Zfr,
  	ZHcy: ZHcy,
  	zhcy: zhcy,
  	zigrarr: zigrarr,
  	zopf: zopf,
  	Zopf: Zopf,
  	Zscr: Zscr,
  	zscr: zscr,
  	zwj: zwj,
  	zwnj: zwnj,
  	default: entitiesJSON
  });

  var require$$0 = ( xml && xmlJSON ) || xml;

  var require$$1 = ( entities && entitiesJSON ) || entities;

  var inverseXML = getInverseObj$1(require$$0);
  var xmlReplacer = getInverseReplacer$1(inverseXML);

  var XML = getInverse$1(inverseXML, xmlReplacer);

  var inverseHTML = getInverseObj$1(require$$1);
  var htmlReplacer = getInverseReplacer$1(inverseHTML);

  var HTML = getInverse$1(inverseHTML, htmlReplacer);

  function getInverseObj$1(obj){
  	return Object.keys(obj).sort().reduce(function(inverse, name){
  		inverse[obj[name]] = "&" + name + ";";
  		return inverse;
  	}, {});
  }

  function getInverseReplacer$1(inverse){
  	var single = [],
  	    multiple = [];

  	Object.keys(inverse).forEach(function(k){
  		if(k.length === 1){
  			single.push("\\" + k);
  		} else {
  			multiple.push(k);
  		}
  	});

  	
  	multiple.unshift("[" + single.join("") + "]");

  	return new RegExp(multiple.join("|"), "g");
  }

  var re_nonASCII$1 = /[^\0-\x7F]/g;
  var re_astralSymbols$1 = /[\uD800-\uDBFF][\uDC00-\uDFFF]/g;

  function singleCharReplacer$1(c){
  	return "&#x" + c.charCodeAt(0).toString(16).toUpperCase() + ";";
  }

  function astralReplacer$1(c){
  	
  	var high = c.charCodeAt(0);
  	var low  = c.charCodeAt(1);
  	var codePoint = (high - 0xD800) * 0x400 + low - 0xDC00 + 0x10000;
  	return "&#x" + codePoint.toString(16).toUpperCase() + ";";
  }

  function getInverse$1(inverse, re){
  	function func(name){
  		return inverse[name];
  	}

  	return function(data){
  		return data
  				.replace(re, func)
  				.replace(re_astralSymbols$1, astralReplacer$1)
  				.replace(re_nonASCII$1, singleCharReplacer$1);
  	};
  }

  var re_xmlChars = getInverseReplacer$1(inverseXML);

  function escapeXML(data){
  	return data
  			.replace(re_xmlChars, singleCharReplacer$1)
  			.replace(re_astralSymbols$1, astralReplacer$1)
  			.replace(re_nonASCII$1, singleCharReplacer$1);
  }

  var escape = escapeXML;

  var encode = {
  	XML: XML,
  	HTML: HTML,
  	escape: escape
  };

  var decode = {
  	"0": 65533,
  	"128": 8364,
  	"130": 8218,
  	"131": 402,
  	"132": 8222,
  	"133": 8230,
  	"134": 8224,
  	"135": 8225,
  	"136": 710,
  	"137": 8240,
  	"138": 352,
  	"139": 8249,
  	"140": 338,
  	"142": 381,
  	"145": 8216,
  	"146": 8217,
  	"147": 8220,
  	"148": 8221,
  	"149": 8226,
  	"150": 8211,
  	"151": 8212,
  	"152": 732,
  	"153": 8482,
  	"154": 353,
  	"155": 8250,
  	"156": 339,
  	"158": 382,
  	"159": 376
  };

  var decode$1 = Object.freeze({
  	default: decode
  });

  var decodeMap = ( decode$1 && decode ) || decode$1;

  var decode_codepoint = decodeCodePoint;


  function decodeCodePoint(codePoint){

  	if((codePoint >= 0xD800 && codePoint <= 0xDFFF) || codePoint > 0x10FFFF){
  		return "\uFFFD";
  	}

  	if(codePoint in decodeMap){
  		codePoint = decodeMap[codePoint];
  	}

  	var output = "";

  	if(codePoint > 0xFFFF){
  		codePoint -= 0x10000;
  		output += String.fromCharCode(codePoint >>> 10 & 0x3FF | 0xD800);
  		codePoint = 0xDC00 | codePoint & 0x3FF;
  	}

  	output += String.fromCharCode(codePoint);
  	return output;
  }

  var Aacute$1 = "Á";
  var aacute$1 = "á";
  var Acirc$1 = "Â";
  var acirc$1 = "â";
  var acute$1 = "´";
  var AElig$1 = "Æ";
  var aelig$1 = "æ";
  var Agrave$1 = "À";
  var agrave$1 = "à";
  var amp$2 = "&";
  var AMP$1 = "&";
  var Aring$1 = "Å";
  var aring$1 = "å";
  var Atilde$1 = "Ã";
  var atilde$1 = "ã";
  var Auml$1 = "Ä";
  var auml$1 = "ä";
  var brvbar$1 = "¦";
  var Ccedil$1 = "Ç";
  var ccedil$1 = "ç";
  var cedil$1 = "¸";
  var cent$1 = "¢";
  var copy$1 = "©";
  var COPY$1 = "©";
  var curren$1 = "¤";
  var deg$1 = "°";
  var divide$1 = "÷";
  var Eacute$1 = "É";
  var eacute$1 = "é";
  var Ecirc$1 = "Ê";
  var ecirc$1 = "ê";
  var Egrave$1 = "È";
  var egrave$1 = "è";
  var ETH$1 = "Ð";
  var eth$1 = "ð";
  var Euml$1 = "Ë";
  var euml$1 = "ë";
  var frac12$1 = "½";
  var frac14$1 = "¼";
  var frac34$1 = "¾";
  var gt$2 = ">";
  var GT$1 = ">";
  var Iacute$1 = "Í";
  var iacute$1 = "í";
  var Icirc$1 = "Î";
  var icirc$1 = "î";
  var iexcl$1 = "¡";
  var Igrave$1 = "Ì";
  var igrave$1 = "ì";
  var iquest$1 = "¿";
  var Iuml$1 = "Ï";
  var iuml$1 = "ï";
  var laquo$1 = "«";
  var lt$2 = "<";
  var LT$1 = "<";
  var macr$1 = "¯";
  var micro$1 = "µ";
  var middot$1 = "·";
  var nbsp$1 = " ";
  var not$1 = "¬";
  var Ntilde$1 = "Ñ";
  var ntilde$1 = "ñ";
  var Oacute$1 = "Ó";
  var oacute$1 = "ó";
  var Ocirc$1 = "Ô";
  var ocirc$1 = "ô";
  var Ograve$1 = "Ò";
  var ograve$1 = "ò";
  var ordf$1 = "ª";
  var ordm$1 = "º";
  var Oslash$1 = "Ø";
  var oslash$1 = "ø";
  var Otilde$1 = "Õ";
  var otilde$1 = "õ";
  var Ouml$1 = "Ö";
  var ouml$1 = "ö";
  var para$1 = "¶";
  var plusmn$1 = "±";
  var pound$1 = "£";
  var quot$2 = "\"";
  var QUOT$1 = "\"";
  var raquo$1 = "»";
  var reg$1 = "®";
  var REG$1 = "®";
  var sect$1 = "§";
  var shy$1 = "­";
  var sup1$1 = "¹";
  var sup2$1 = "²";
  var sup3$1 = "³";
  var szlig$1 = "ß";
  var THORN$1 = "Þ";
  var thorn$1 = "þ";
  var times$1 = "×";
  var Uacute$1 = "Ú";
  var uacute$1 = "ú";
  var Ucirc$1 = "Û";
  var ucirc$1 = "û";
  var Ugrave$1 = "Ù";
  var ugrave$1 = "ù";
  var uml$1 = "¨";
  var Uuml$1 = "Ü";
  var uuml$1 = "ü";
  var Yacute$1 = "Ý";
  var yacute$1 = "ý";
  var yen$1 = "¥";
  var yuml$1 = "ÿ";
  var legacyJSON = {
  	Aacute: Aacute$1,
  	aacute: aacute$1,
  	Acirc: Acirc$1,
  	acirc: acirc$1,
  	acute: acute$1,
  	AElig: AElig$1,
  	aelig: aelig$1,
  	Agrave: Agrave$1,
  	agrave: agrave$1,
  	amp: amp$2,
  	AMP: AMP$1,
  	Aring: Aring$1,
  	aring: aring$1,
  	Atilde: Atilde$1,
  	atilde: atilde$1,
  	Auml: Auml$1,
  	auml: auml$1,
  	brvbar: brvbar$1,
  	Ccedil: Ccedil$1,
  	ccedil: ccedil$1,
  	cedil: cedil$1,
  	cent: cent$1,
  	copy: copy$1,
  	COPY: COPY$1,
  	curren: curren$1,
  	deg: deg$1,
  	divide: divide$1,
  	Eacute: Eacute$1,
  	eacute: eacute$1,
  	Ecirc: Ecirc$1,
  	ecirc: ecirc$1,
  	Egrave: Egrave$1,
  	egrave: egrave$1,
  	ETH: ETH$1,
  	eth: eth$1,
  	Euml: Euml$1,
  	euml: euml$1,
  	frac12: frac12$1,
  	frac14: frac14$1,
  	frac34: frac34$1,
  	gt: gt$2,
  	GT: GT$1,
  	Iacute: Iacute$1,
  	iacute: iacute$1,
  	Icirc: Icirc$1,
  	icirc: icirc$1,
  	iexcl: iexcl$1,
  	Igrave: Igrave$1,
  	igrave: igrave$1,
  	iquest: iquest$1,
  	Iuml: Iuml$1,
  	iuml: iuml$1,
  	laquo: laquo$1,
  	lt: lt$2,
  	LT: LT$1,
  	macr: macr$1,
  	micro: micro$1,
  	middot: middot$1,
  	nbsp: nbsp$1,
  	not: not$1,
  	Ntilde: Ntilde$1,
  	ntilde: ntilde$1,
  	Oacute: Oacute$1,
  	oacute: oacute$1,
  	Ocirc: Ocirc$1,
  	ocirc: ocirc$1,
  	Ograve: Ograve$1,
  	ograve: ograve$1,
  	ordf: ordf$1,
  	ordm: ordm$1,
  	Oslash: Oslash$1,
  	oslash: oslash$1,
  	Otilde: Otilde$1,
  	otilde: otilde$1,
  	Ouml: Ouml$1,
  	ouml: ouml$1,
  	para: para$1,
  	plusmn: plusmn$1,
  	pound: pound$1,
  	quot: quot$2,
  	QUOT: QUOT$1,
  	raquo: raquo$1,
  	reg: reg$1,
  	REG: REG$1,
  	sect: sect$1,
  	shy: shy$1,
  	sup1: sup1$1,
  	sup2: sup2$1,
  	sup3: sup3$1,
  	szlig: szlig$1,
  	THORN: THORN$1,
  	thorn: thorn$1,
  	times: times$1,
  	Uacute: Uacute$1,
  	uacute: uacute$1,
  	Ucirc: Ucirc$1,
  	ucirc: ucirc$1,
  	Ugrave: Ugrave$1,
  	ugrave: ugrave$1,
  	uml: uml$1,
  	Uuml: Uuml$1,
  	uuml: uuml$1,
  	Yacute: Yacute$1,
  	yacute: yacute$1,
  	yen: yen$1,
  	yuml: yuml$1
  };

  var _entities = {
    encodeXML: encode.XML,
    decodeCodepoint: decode_codepoint,
    entitiesJSON: entitiesJSON,
    legacyJSON: legacyJSON,
    xmlJSON: xmlJSON
  };

  var _entities_decodeCodepoint = _entities.decodeCodepoint;

  var _entities_decodeCodepoint$1 = Object.freeze({
  	default: _entities_decodeCodepoint
  });

  var _entities_entitiesJSON = _entities.entitiesJSON;


  var _entities_entitiesJSON$1 = Object.freeze({
  	default: _entities_entitiesJSON
  });

  var _entities_legacyJSON = _entities.legacyJSON;

  var _entities_legacyJSON$1 = Object.freeze({
  	default: _entities_legacyJSON
  });

  var _entities_xmlJSON = _entities.xmlJSON;

  var _entities_xmlJSON$1 = Object.freeze({
  	default: _entities_xmlJSON
  });

  var decodeCodePoint$1 = ( _entities_decodeCodepoint$1 && _entities_decodeCodepoint ) || _entities_decodeCodepoint$1;

  var entityMap = ( _entities_entitiesJSON$1 && _entities_entitiesJSON ) || _entities_entitiesJSON$1;

  var legacyMap = ( _entities_legacyJSON$1 && _entities_legacyJSON ) || _entities_legacyJSON$1;

  var xmlMap = ( _entities_xmlJSON$1 && _entities_xmlJSON ) || _entities_xmlJSON$1;

  var Tokenizer_1 = Tokenizer$1;

  var i = 0;
  var TEXT                      = i++;
  var BEFORE_TAG_NAME           = i++;
  var IN_TAG_NAME               = i++;
  var IN_SELF_CLOSING_TAG       = i++;
  var BEFORE_CLOSING_TAG_NAME   = i++;
  var IN_CLOSING_TAG_NAME       = i++;
  var AFTER_CLOSING_TAG_NAME    = i++;
  var BEFORE_ATTRIBUTE_NAME     = i++;
  var IN_ATTRIBUTE_NAME         = i++;
  var AFTER_ATTRIBUTE_NAME      = i++;
  var BEFORE_ATTRIBUTE_VALUE    = i++;
  var IN_ATTRIBUTE_VALUE_DQ     = i++;
  var IN_ATTRIBUTE_VALUE_SQ     = i++;
  var IN_ATTRIBUTE_VALUE_NQ     = i++;
  var BEFORE_DECLARATION        = i++;
  var IN_DECLARATION            = i++;
  var IN_PROCESSING_INSTRUCTION = i++;
  var BEFORE_COMMENT            = i++;
  var IN_COMMENT                = i++;
  var AFTER_COMMENT_1           = i++;
  var AFTER_COMMENT_2           = i++;
  var BEFORE_CDATA_1            = i++;
  var BEFORE_CDATA_2            = i++;
  var BEFORE_CDATA_3            = i++;
  var BEFORE_CDATA_4            = i++;
  var BEFORE_CDATA_5            = i++;
  var BEFORE_CDATA_6            = i++;
  var IN_CDATA                  = i++;
  var AFTER_CDATA_1             = i++;
  var AFTER_CDATA_2             = i++;
  var BEFORE_SPECIAL            = i++;
  var BEFORE_SPECIAL_END        = i++;
  var BEFORE_SCRIPT_1           = i++;
  var BEFORE_SCRIPT_2           = i++;
  var BEFORE_SCRIPT_3           = i++;
  var BEFORE_SCRIPT_4           = i++;
  var BEFORE_SCRIPT_5           = i++;
  var AFTER_SCRIPT_1            = i++;
  var AFTER_SCRIPT_2            = i++;
  var AFTER_SCRIPT_3            = i++;
  var AFTER_SCRIPT_4            = i++;
  var AFTER_SCRIPT_5            = i++;
  var BEFORE_STYLE_1            = i++;
  var BEFORE_STYLE_2            = i++;
  var BEFORE_STYLE_3            = i++;
  var BEFORE_STYLE_4            = i++;
  var AFTER_STYLE_1             = i++;
  var AFTER_STYLE_2             = i++;
  var AFTER_STYLE_3             = i++;
  var AFTER_STYLE_4             = i++;
  var BEFORE_ENTITY             = i++;
  var BEFORE_NUMERIC_ENTITY     = i++;
  var IN_NAMED_ENTITY           = i++;
  var IN_NUMERIC_ENTITY         = i++;
  var IN_HEX_ENTITY             = i++;
  var j = 0;
  var SPECIAL_NONE              = j++;
  var SPECIAL_SCRIPT            = j++;
  var SPECIAL_STYLE             = j++;

  function whitespace(c){
  	return c === " " || c === "\n" || c === "\t" || c === "\f" || c === "\r";
  }

  function characterState(char, SUCCESS){
  	return function(c){
  		if(c === char) { this._state = SUCCESS; }
  	};
  }

  function ifElseState(upper, SUCCESS, FAILURE){
  	var lower = upper.toLowerCase();

  	if(upper === lower){
  		return function(c){
  			if(c === lower){
  				this._state = SUCCESS;
  			} else {
  				this._state = FAILURE;
  				this._index--;
  			}
  		};
  	} else {
  		return function(c){
  			if(c === lower || c === upper){
  				this._state = SUCCESS;
  			} else {
  				this._state = FAILURE;
  				this._index--;
  			}
  		};
  	}
  }

  function consumeSpecialNameChar(upper, NEXT_STATE){
  	var lower = upper.toLowerCase();

  	return function(c){
  		if(c === lower || c === upper){
  			this._state = NEXT_STATE;
  		} else {
  			this._state = IN_TAG_NAME;
  			this._index--; 
  		}
  	};
  }

  function Tokenizer$1(options, cbs){
  	this._state = TEXT;
  	this._buffer = "";
  	this._sectionStart = 0;
  	this._index = 0;
  	this._bufferOffset = 0; 
  	this._baseState = TEXT;
  	this._special = SPECIAL_NONE;
  	this._cbs = cbs;
  	this._running = true;
  	this._ended = false;
  	this._xmlMode = !!(options && options.xmlMode);
  	this._decodeEntities = !!(options && options.decodeEntities);
  }

  Tokenizer$1.prototype._stateText = function(c){
  	if(c === "<"){
  		if(this._index > this._sectionStart){
  			this._cbs.ontext(this._getSection());
  		}
  		this._state = BEFORE_TAG_NAME;
  		this._sectionStart = this._index;
  	} else if(this._decodeEntities && this._special === SPECIAL_NONE && c === "&"){
  		if(this._index > this._sectionStart){
  			this._cbs.ontext(this._getSection());
  		}
  		this._baseState = TEXT;
  		this._state = BEFORE_ENTITY;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1.prototype._stateBeforeTagName = function(c){
  	if(c === "/"){
  		this._state = BEFORE_CLOSING_TAG_NAME;
  	} else if(c === "<"){
  		this._cbs.ontext(this._getSection());
  		this._sectionStart = this._index;
  	} else if(c === ">" || this._special !== SPECIAL_NONE || whitespace(c)) {
  		this._state = TEXT;
  	} else if(c === "!"){
  		this._state = BEFORE_DECLARATION;
  		this._sectionStart = this._index + 1;
  	} else if(c === "?"){
  		this._state = IN_PROCESSING_INSTRUCTION;
  		this._sectionStart = this._index + 1;
  	} else {
  		this._state = (!this._xmlMode && (c === "s" || c === "S")) ?
  						BEFORE_SPECIAL : IN_TAG_NAME;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1.prototype._stateInTagName = function(c){
  	if(c === "/" || c === ">" || whitespace(c)){
  		this._emitToken("onopentagname");
  		this._state = BEFORE_ATTRIBUTE_NAME;
  		this._index--;
  	}
  };

  Tokenizer$1.prototype._stateBeforeCloseingTagName = function(c){
  	if(whitespace(c));
  	else if(c === ">"){
  		this._state = TEXT;
  	} else if(this._special !== SPECIAL_NONE){
  		if(c === "s" || c === "S"){
  			this._state = BEFORE_SPECIAL_END;
  		} else {
  			this._state = TEXT;
  			this._index--;
  		}
  	} else {
  		this._state = IN_CLOSING_TAG_NAME;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1.prototype._stateInCloseingTagName = function(c){
  	if(c === ">" || whitespace(c)){
  		this._emitToken("onclosetag");
  		this._state = AFTER_CLOSING_TAG_NAME;
  		this._index--;
  	}
  };

  Tokenizer$1.prototype._stateAfterCloseingTagName = function(c){
  	
  	if(c === ">"){
  		this._state = TEXT;
  		this._sectionStart = this._index + 1;
  	}
  };

  Tokenizer$1.prototype._stateBeforeAttributeName = function(c){
  	if(c === ">"){
  		this._cbs.onopentagend();
  		this._state = TEXT;
  		this._sectionStart = this._index + 1;
  	} else if(c === "/"){
  		this._state = IN_SELF_CLOSING_TAG;
  	} else if(!whitespace(c)){
  		this._state = IN_ATTRIBUTE_NAME;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1.prototype._stateInSelfClosingTag = function(c){
  	if(c === ">"){
  		this._cbs.onselfclosingtag();
  		this._state = TEXT;
  		this._sectionStart = this._index + 1;
  	} else if(!whitespace(c)){
  		this._state = BEFORE_ATTRIBUTE_NAME;
  		this._index--;
  	}
  };

  Tokenizer$1.prototype._stateInAttributeName = function(c){
  	if(c === "=" || c === "/" || c === ">" || whitespace(c)){
  		this._cbs.onattribname(this._getSection());
  		this._sectionStart = -1;
  		this._state = AFTER_ATTRIBUTE_NAME;
  		this._index--;
  	}
  };

  Tokenizer$1.prototype._stateAfterAttributeName = function(c){
  	if(c === "="){
  		this._state = BEFORE_ATTRIBUTE_VALUE;
  	} else if(c === "/" || c === ">"){
  		this._cbs.onattribend();
  		this._state = BEFORE_ATTRIBUTE_NAME;
  		this._index--;
  	} else if(!whitespace(c)){
  		this._cbs.onattribend();
  		this._state = IN_ATTRIBUTE_NAME;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1.prototype._stateBeforeAttributeValue = function(c){
  	if(c === "\""){
  		this._state = IN_ATTRIBUTE_VALUE_DQ;
  		this._sectionStart = this._index + 1;
  	} else if(c === "'"){
  		this._state = IN_ATTRIBUTE_VALUE_SQ;
  		this._sectionStart = this._index + 1;
  	} else if(!whitespace(c)){
  		this._state = IN_ATTRIBUTE_VALUE_NQ;
  		this._sectionStart = this._index;
  		this._index--; 
  	}
  };

  Tokenizer$1.prototype._stateInAttributeValueDoubleQuotes = function(c){
  	if(c === "\""){
  		this._emitToken("onattribdata");
  		this._cbs.onattribend();
  		this._state = BEFORE_ATTRIBUTE_NAME;
  	} else if(this._decodeEntities && c === "&"){
  		this._emitToken("onattribdata");
  		this._baseState = this._state;
  		this._state = BEFORE_ENTITY;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1.prototype._stateInAttributeValueSingleQuotes = function(c){
  	if(c === "'"){
  		this._emitToken("onattribdata");
  		this._cbs.onattribend();
  		this._state = BEFORE_ATTRIBUTE_NAME;
  	} else if(this._decodeEntities && c === "&"){
  		this._emitToken("onattribdata");
  		this._baseState = this._state;
  		this._state = BEFORE_ENTITY;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1.prototype._stateInAttributeValueNoQuotes = function(c){
  	if(whitespace(c) || c === ">"){
  		this._emitToken("onattribdata");
  		this._cbs.onattribend();
  		this._state = BEFORE_ATTRIBUTE_NAME;
  		this._index--;
  	} else if(this._decodeEntities && c === "&"){
  		this._emitToken("onattribdata");
  		this._baseState = this._state;
  		this._state = BEFORE_ENTITY;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1.prototype._stateBeforeDeclaration = function(c){
  	this._state = c === "[" ? BEFORE_CDATA_1 :
  					c === "-" ? BEFORE_COMMENT :
  						IN_DECLARATION;
  };

  Tokenizer$1.prototype._stateInDeclaration = function(c){
  	if(c === ">"){
  		this._cbs.ondeclaration(this._getSection());
  		this._state = TEXT;
  		this._sectionStart = this._index + 1;
  	}
  };

  Tokenizer$1.prototype._stateInProcessingInstruction = function(c){
  	if(c === ">"){
  		this._cbs.onprocessinginstruction(this._getSection());
  		this._state = TEXT;
  		this._sectionStart = this._index + 1;
  	}
  };

  Tokenizer$1.prototype._stateBeforeComment = function(c){
  	if(c === "-"){
  		this._state = IN_COMMENT;
  		this._sectionStart = this._index + 1;
  	} else {
  		this._state = IN_DECLARATION;
  	}
  };

  Tokenizer$1.prototype._stateInComment = function(c){
  	if(c === "-") { this._state = AFTER_COMMENT_1; }
  };

  Tokenizer$1.prototype._stateAfterComment1 = function(c){
  	if(c === "-"){
  		this._state = AFTER_COMMENT_2;
  	} else {
  		this._state = IN_COMMENT;
  	}
  };

  Tokenizer$1.prototype._stateAfterComment2 = function(c){
  	if(c === ">"){
  		
  		this._cbs.oncomment(this._buffer.substring(this._sectionStart, this._index - 2));
  		this._state = TEXT;
  		this._sectionStart = this._index + 1;
  	} else if(c !== "-"){
  		this._state = IN_COMMENT;
  	}
  	
  };

  Tokenizer$1.prototype._stateBeforeCdata1 = ifElseState("C", BEFORE_CDATA_2, IN_DECLARATION);
  Tokenizer$1.prototype._stateBeforeCdata2 = ifElseState("D", BEFORE_CDATA_3, IN_DECLARATION);
  Tokenizer$1.prototype._stateBeforeCdata3 = ifElseState("A", BEFORE_CDATA_4, IN_DECLARATION);
  Tokenizer$1.prototype._stateBeforeCdata4 = ifElseState("T", BEFORE_CDATA_5, IN_DECLARATION);
  Tokenizer$1.prototype._stateBeforeCdata5 = ifElseState("A", BEFORE_CDATA_6, IN_DECLARATION);

  Tokenizer$1.prototype._stateBeforeCdata6 = function(c){
  	if(c === "["){
  		this._state = IN_CDATA;
  		this._sectionStart = this._index + 1;
  	} else {
  		this._state = IN_DECLARATION;
  		this._index--;
  	}
  };

  Tokenizer$1.prototype._stateInCdata = function(c){
  	if(c === "]") { this._state = AFTER_CDATA_1; }
  };

  Tokenizer$1.prototype._stateAfterCdata1 = characterState("]", AFTER_CDATA_2);

  Tokenizer$1.prototype._stateAfterCdata2 = function(c){
  	if(c === ">"){
  		
  		this._cbs.oncdata(this._buffer.substring(this._sectionStart, this._index - 2));
  		this._state = TEXT;
  		this._sectionStart = this._index + 1;
  	} else if(c !== "]") {
  		this._state = IN_CDATA;
  	}
  	
  };

  Tokenizer$1.prototype._stateBeforeSpecial = function(c){
  	if(c === "c" || c === "C"){
  		this._state = BEFORE_SCRIPT_1;
  	} else if(c === "t" || c === "T"){
  		this._state = BEFORE_STYLE_1;
  	} else {
  		this._state = IN_TAG_NAME;
  		this._index--; 
  	}
  };

  Tokenizer$1.prototype._stateBeforeSpecialEnd = function(c){
  	if(this._special === SPECIAL_SCRIPT && (c === "c" || c === "C")){
  		this._state = AFTER_SCRIPT_1;
  	} else if(this._special === SPECIAL_STYLE && (c === "t" || c === "T")){
  		this._state = AFTER_STYLE_1;
  	}
  	else { this._state = TEXT; }
  };

  Tokenizer$1.prototype._stateBeforeScript1 = consumeSpecialNameChar("R", BEFORE_SCRIPT_2);
  Tokenizer$1.prototype._stateBeforeScript2 = consumeSpecialNameChar("I", BEFORE_SCRIPT_3);
  Tokenizer$1.prototype._stateBeforeScript3 = consumeSpecialNameChar("P", BEFORE_SCRIPT_4);
  Tokenizer$1.prototype._stateBeforeScript4 = consumeSpecialNameChar("T", BEFORE_SCRIPT_5);

  Tokenizer$1.prototype._stateBeforeScript5 = function(c){
  	if(c === "/" || c === ">" || whitespace(c)){
  		this._special = SPECIAL_SCRIPT;
  	}
  	this._state = IN_TAG_NAME;
  	this._index--; 
  };

  Tokenizer$1.prototype._stateAfterScript1 = ifElseState("R", AFTER_SCRIPT_2, TEXT);
  Tokenizer$1.prototype._stateAfterScript2 = ifElseState("I", AFTER_SCRIPT_3, TEXT);
  Tokenizer$1.prototype._stateAfterScript3 = ifElseState("P", AFTER_SCRIPT_4, TEXT);
  Tokenizer$1.prototype._stateAfterScript4 = ifElseState("T", AFTER_SCRIPT_5, TEXT);

  Tokenizer$1.prototype._stateAfterScript5 = function(c){
  	if(c === ">" || whitespace(c)){
  		this._special = SPECIAL_NONE;
  		this._state = IN_CLOSING_TAG_NAME;
  		this._sectionStart = this._index - 6;
  		this._index--; 
  	}
  	else { this._state = TEXT; }
  };

  Tokenizer$1.prototype._stateBeforeStyle1 = consumeSpecialNameChar("Y", BEFORE_STYLE_2);
  Tokenizer$1.prototype._stateBeforeStyle2 = consumeSpecialNameChar("L", BEFORE_STYLE_3);
  Tokenizer$1.prototype._stateBeforeStyle3 = consumeSpecialNameChar("E", BEFORE_STYLE_4);

  Tokenizer$1.prototype._stateBeforeStyle4 = function(c){
  	if(c === "/" || c === ">" || whitespace(c)){
  		this._special = SPECIAL_STYLE;
  	}
  	this._state = IN_TAG_NAME;
  	this._index--; 
  };

  Tokenizer$1.prototype._stateAfterStyle1 = ifElseState("Y", AFTER_STYLE_2, TEXT);
  Tokenizer$1.prototype._stateAfterStyle2 = ifElseState("L", AFTER_STYLE_3, TEXT);
  Tokenizer$1.prototype._stateAfterStyle3 = ifElseState("E", AFTER_STYLE_4, TEXT);

  Tokenizer$1.prototype._stateAfterStyle4 = function(c){
  	if(c === ">" || whitespace(c)){
  		this._special = SPECIAL_NONE;
  		this._state = IN_CLOSING_TAG_NAME;
  		this._sectionStart = this._index - 5;
  		this._index--; 
  	}
  	else { this._state = TEXT; }
  };

  Tokenizer$1.prototype._stateBeforeEntity = ifElseState("#", BEFORE_NUMERIC_ENTITY, IN_NAMED_ENTITY);
  Tokenizer$1.prototype._stateBeforeNumericEntity = ifElseState("X", IN_HEX_ENTITY, IN_NUMERIC_ENTITY);


  Tokenizer$1.prototype._parseNamedEntityStrict = function(){
  	
  	if(this._sectionStart + 1 < this._index){
  		var entity = this._buffer.substring(this._sectionStart + 1, this._index),
  		    map = this._xmlMode ? xmlMap : entityMap;

  		if(map.hasOwnProperty(entity)){
  			this._emitPartial(map[entity]);
  			this._sectionStart = this._index + 1;
  		}
  	}
  };



  Tokenizer$1.prototype._parseLegacyEntity = function(){
  	var this$1 = this;

  	var start = this._sectionStart + 1,
  	    limit = this._index - start;

  	if(limit > 6) { limit = 6; } 

  	while(limit >= 2){ 
  		var entity = this$1._buffer.substr(start, limit);

  		if(legacyMap.hasOwnProperty(entity)){
  			this$1._emitPartial(legacyMap[entity]);
  			this$1._sectionStart += limit + 1;
  			return;
  		} else {
  			limit--;
  		}
  	}
  };

  Tokenizer$1.prototype._stateInNamedEntity = function(c){
  	if(c === ";"){
  		this._parseNamedEntityStrict();
  		if(this._sectionStart + 1 < this._index && !this._xmlMode){
  			this._parseLegacyEntity();
  		}
  		this._state = this._baseState;
  	} else if((c < "a" || c > "z") && (c < "A" || c > "Z") && (c < "0" || c > "9")){
  		if(this._xmlMode);
  		else if(this._sectionStart + 1 === this._index);
  		else if(this._baseState !== TEXT){
  			if(c !== "="){
  				this._parseNamedEntityStrict();
  			}
  		} else {
  			this._parseLegacyEntity();
  		}

  		this._state = this._baseState;
  		this._index--;
  	}
  };

  Tokenizer$1.prototype._decodeNumericEntity = function(offset, base){
  	var sectionStart = this._sectionStart + offset;

  	if(sectionStart !== this._index){
  		
  		var entity = this._buffer.substring(sectionStart, this._index);
  		var parsed = parseInt(entity, base);

  		this._emitPartial(decodeCodePoint$1(parsed));
  		this._sectionStart = this._index;
  	} else {
  		this._sectionStart--;
  	}

  	this._state = this._baseState;
  };

  Tokenizer$1.prototype._stateInNumericEntity = function(c){
  	if(c === ";"){
  		this._decodeNumericEntity(2, 10);
  		this._sectionStart++;
  	} else if(c < "0" || c > "9"){
  		if(!this._xmlMode){
  			this._decodeNumericEntity(2, 10);
  		} else {
  			this._state = this._baseState;
  		}
  		this._index--;
  	}
  };

  Tokenizer$1.prototype._stateInHexEntity = function(c){
  	if(c === ";"){
  		this._decodeNumericEntity(3, 16);
  		this._sectionStart++;
  	} else if((c < "a" || c > "f") && (c < "A" || c > "F") && (c < "0" || c > "9")){
  		if(!this._xmlMode){
  			this._decodeNumericEntity(3, 16);
  		} else {
  			this._state = this._baseState;
  		}
  		this._index--;
  	}
  };

  Tokenizer$1.prototype._cleanup = function (){
  	if(this._sectionStart < 0){
  		this._buffer = "";
  		this._bufferOffset += this._index;
  		this._index = 0;
  	} else if(this._running){
  		if(this._state === TEXT){
  			if(this._sectionStart !== this._index){
  				this._cbs.ontext(this._buffer.substr(this._sectionStart));
  			}
  			this._buffer = "";
  			this._bufferOffset += this._index;
  			this._index = 0;
  		} else if(this._sectionStart === this._index){
  			
  			this._buffer = "";
  			this._bufferOffset += this._index;
  			this._index = 0;
  		} else {
  			
  			this._buffer = this._buffer.substr(this._sectionStart);
  			this._index -= this._sectionStart;
  			this._bufferOffset += this._sectionStart;
  		}

  		this._sectionStart = 0;
  	}
  };


  Tokenizer$1.prototype.write = function(chunk){
  	if(this._ended) { this._cbs.onerror(Error(".write() after done!")); }

  	this._buffer += chunk;
  	this._parse();
  };

  Tokenizer$1.prototype._parse = function(){
  	var this$1 = this;

  	while(this._index < this._buffer.length && this._running){
  		var c = this$1._buffer.charAt(this$1._index);
  		if(this$1._state === TEXT) {
  			this$1._stateText(c);
  		} else if(this$1._state === BEFORE_TAG_NAME){
  			this$1._stateBeforeTagName(c);
  		} else if(this$1._state === IN_TAG_NAME) {
  			this$1._stateInTagName(c);
  		} else if(this$1._state === BEFORE_CLOSING_TAG_NAME){
  			this$1._stateBeforeCloseingTagName(c);
  		} else if(this$1._state === IN_CLOSING_TAG_NAME){
  			this$1._stateInCloseingTagName(c);
  		} else if(this$1._state === AFTER_CLOSING_TAG_NAME){
  			this$1._stateAfterCloseingTagName(c);
  		} else if(this$1._state === IN_SELF_CLOSING_TAG){
  			this$1._stateInSelfClosingTag(c);
  		}

  		
  		else if(this$1._state === BEFORE_ATTRIBUTE_NAME){
  			this$1._stateBeforeAttributeName(c);
  		} else if(this$1._state === IN_ATTRIBUTE_NAME){
  			this$1._stateInAttributeName(c);
  		} else if(this$1._state === AFTER_ATTRIBUTE_NAME){
  			this$1._stateAfterAttributeName(c);
  		} else if(this$1._state === BEFORE_ATTRIBUTE_VALUE){
  			this$1._stateBeforeAttributeValue(c);
  		} else if(this$1._state === IN_ATTRIBUTE_VALUE_DQ){
  			this$1._stateInAttributeValueDoubleQuotes(c);
  		} else if(this$1._state === IN_ATTRIBUTE_VALUE_SQ){
  			this$1._stateInAttributeValueSingleQuotes(c);
  		} else if(this$1._state === IN_ATTRIBUTE_VALUE_NQ){
  			this$1._stateInAttributeValueNoQuotes(c);
  		}

  		
  		else if(this$1._state === BEFORE_DECLARATION){
  			this$1._stateBeforeDeclaration(c);
  		} else if(this$1._state === IN_DECLARATION){
  			this$1._stateInDeclaration(c);
  		}

  		
  		else if(this$1._state === IN_PROCESSING_INSTRUCTION){
  			this$1._stateInProcessingInstruction(c);
  		}

  		
  		else if(this$1._state === BEFORE_COMMENT){
  			this$1._stateBeforeComment(c);
  		} else if(this$1._state === IN_COMMENT){
  			this$1._stateInComment(c);
  		} else if(this$1._state === AFTER_COMMENT_1){
  			this$1._stateAfterComment1(c);
  		} else if(this$1._state === AFTER_COMMENT_2){
  			this$1._stateAfterComment2(c);
  		}

  		
  		else if(this$1._state === BEFORE_CDATA_1){
  			this$1._stateBeforeCdata1(c);
  		} else if(this$1._state === BEFORE_CDATA_2){
  			this$1._stateBeforeCdata2(c);
  		} else if(this$1._state === BEFORE_CDATA_3){
  			this$1._stateBeforeCdata3(c);
  		} else if(this$1._state === BEFORE_CDATA_4){
  			this$1._stateBeforeCdata4(c);
  		} else if(this$1._state === BEFORE_CDATA_5){
  			this$1._stateBeforeCdata5(c);
  		} else if(this$1._state === BEFORE_CDATA_6){
  			this$1._stateBeforeCdata6(c);
  		} else if(this$1._state === IN_CDATA){
  			this$1._stateInCdata(c);
  		} else if(this$1._state === AFTER_CDATA_1){
  			this$1._stateAfterCdata1(c);
  		} else if(this$1._state === AFTER_CDATA_2){
  			this$1._stateAfterCdata2(c);
  		}

  		
  		else if(this$1._state === BEFORE_SPECIAL){
  			this$1._stateBeforeSpecial(c);
  		} else if(this$1._state === BEFORE_SPECIAL_END){
  			this$1._stateBeforeSpecialEnd(c);
  		}

  		
  		else if(this$1._state === BEFORE_SCRIPT_1){
  			this$1._stateBeforeScript1(c);
  		} else if(this$1._state === BEFORE_SCRIPT_2){
  			this$1._stateBeforeScript2(c);
  		} else if(this$1._state === BEFORE_SCRIPT_3){
  			this$1._stateBeforeScript3(c);
  		} else if(this$1._state === BEFORE_SCRIPT_4){
  			this$1._stateBeforeScript4(c);
  		} else if(this$1._state === BEFORE_SCRIPT_5){
  			this$1._stateBeforeScript5(c);
  		}

  		else if(this$1._state === AFTER_SCRIPT_1){
  			this$1._stateAfterScript1(c);
  		} else if(this$1._state === AFTER_SCRIPT_2){
  			this$1._stateAfterScript2(c);
  		} else if(this$1._state === AFTER_SCRIPT_3){
  			this$1._stateAfterScript3(c);
  		} else if(this$1._state === AFTER_SCRIPT_4){
  			this$1._stateAfterScript4(c);
  		} else if(this$1._state === AFTER_SCRIPT_5){
  			this$1._stateAfterScript5(c);
  		}

  		
  		else if(this$1._state === BEFORE_STYLE_1){
  			this$1._stateBeforeStyle1(c);
  		} else if(this$1._state === BEFORE_STYLE_2){
  			this$1._stateBeforeStyle2(c);
  		} else if(this$1._state === BEFORE_STYLE_3){
  			this$1._stateBeforeStyle3(c);
  		} else if(this$1._state === BEFORE_STYLE_4){
  			this$1._stateBeforeStyle4(c);
  		}

  		else if(this$1._state === AFTER_STYLE_1){
  			this$1._stateAfterStyle1(c);
  		} else if(this$1._state === AFTER_STYLE_2){
  			this$1._stateAfterStyle2(c);
  		} else if(this$1._state === AFTER_STYLE_3){
  			this$1._stateAfterStyle3(c);
  		} else if(this$1._state === AFTER_STYLE_4){
  			this$1._stateAfterStyle4(c);
  		}

  		
  		else if(this$1._state === BEFORE_ENTITY){
  			this$1._stateBeforeEntity(c);
  		} else if(this$1._state === BEFORE_NUMERIC_ENTITY){
  			this$1._stateBeforeNumericEntity(c);
  		} else if(this$1._state === IN_NAMED_ENTITY){
  			this$1._stateInNamedEntity(c);
  		} else if(this$1._state === IN_NUMERIC_ENTITY){
  			this$1._stateInNumericEntity(c);
  		} else if(this$1._state === IN_HEX_ENTITY){
  			this$1._stateInHexEntity(c);
  		}

  		else {
  			this$1._cbs.onerror(Error("unknown _state"), this$1._state);
  		}

  		this$1._index++;
  	}

  	this._cleanup();
  };

  Tokenizer$1.prototype.pause = function(){
  	this._running = false;
  };
  Tokenizer$1.prototype.resume = function(){
  	this._running = true;

  	if(this._index < this._buffer.length){
  		this._parse();
  	}
  	if(this._ended){
  		this._finish();
  	}
  };

  Tokenizer$1.prototype.end = function(chunk){
  	if(this._ended) { this._cbs.onerror(Error(".end() after done!")); }
  	if(chunk) { this.write(chunk); }

  	this._ended = true;

  	if(this._running) { this._finish(); }
  };

  Tokenizer$1.prototype._finish = function(){
  	
  	if(this._sectionStart < this._index){
  		this._handleTrailingData();
  	}

  	this._cbs.onend();
  };

  Tokenizer$1.prototype._handleTrailingData = function(){
  	var data = this._buffer.substr(this._sectionStart);

  	if(this._state === IN_CDATA || this._state === AFTER_CDATA_1 || this._state === AFTER_CDATA_2){
  		this._cbs.oncdata(data);
  	} else if(this._state === IN_COMMENT || this._state === AFTER_COMMENT_1 || this._state === AFTER_COMMENT_2){
  		this._cbs.oncomment(data);
  	} else if(this._state === IN_NAMED_ENTITY && !this._xmlMode){
  		this._parseLegacyEntity();
  		if(this._sectionStart < this._index){
  			this._state = this._baseState;
  			this._handleTrailingData();
  		}
  	} else if(this._state === IN_NUMERIC_ENTITY && !this._xmlMode){
  		this._decodeNumericEntity(2, 10);
  		if(this._sectionStart < this._index){
  			this._state = this._baseState;
  			this._handleTrailingData();
  		}
  	} else if(this._state === IN_HEX_ENTITY && !this._xmlMode){
  		this._decodeNumericEntity(3, 16);
  		if(this._sectionStart < this._index){
  			this._state = this._baseState;
  			this._handleTrailingData();
  		}
  	} else if(
  		this._state !== IN_TAG_NAME &&
  		this._state !== BEFORE_ATTRIBUTE_NAME &&
  		this._state !== BEFORE_ATTRIBUTE_VALUE &&
  		this._state !== AFTER_ATTRIBUTE_NAME &&
  		this._state !== IN_ATTRIBUTE_NAME &&
  		this._state !== IN_ATTRIBUTE_VALUE_SQ &&
  		this._state !== IN_ATTRIBUTE_VALUE_DQ &&
  		this._state !== IN_ATTRIBUTE_VALUE_NQ &&
  		this._state !== IN_CLOSING_TAG_NAME
  	){
  		this._cbs.ontext(data);
  	}
  	
  	
  };

  Tokenizer$1.prototype.reset = function(){
  	Tokenizer$1.call(this, {xmlMode: this._xmlMode, decodeEntities: this._decodeEntities}, this._cbs);
  };

  Tokenizer$1.prototype.getAbsoluteIndex = function(){
  	return this._bufferOffset + this._index;
  };

  Tokenizer$1.prototype._getSection = function(){
  	return this._buffer.substring(this._sectionStart, this._index);
  };

  Tokenizer$1.prototype._emitToken = function(name){
  	this._cbs[name](this._getSection());
  	this._sectionStart = -1;
  };

  Tokenizer$1.prototype._emitPartial = function(value){
  	if(this._baseState !== TEXT){
  		this._cbs.onattribdata(value); 
  	} else {
  		this._cbs.ontext(value);
  	}
  };

  var _inherits = function() {};

  var _inherits$1 = Object.freeze({
  	default: _inherits
  });

  var _resolve_empty = {};

  var _resolve_empty$1 = Object.freeze({
  	default: _resolve_empty
  });

  var require$$1$1 = ( _inherits$1 && _inherits ) || _inherits$1;

  var require$$2 = ( _resolve_empty$1 && _resolve_empty ) || _resolve_empty$1;

  var Tokenizer$2 = Tokenizer_1;





  var formTags = {
  	input: true,
  	option: true,
  	optgroup: true,
  	select: true,
  	button: true,
  	datalist: true,
  	textarea: true
  };

  var openImpliesClose = {
  	tr      : { tr:true, th:true, td:true },
  	th      : { th:true },
  	td      : { thead:true, th:true, td:true },
  	body    : { head:true, link:true, script:true },
  	li      : { li:true },
  	p       : { p:true },
  	h1      : { p:true },
  	h2      : { p:true },
  	h3      : { p:true },
  	h4      : { p:true },
  	h5      : { p:true },
  	h6      : { p:true },
  	select  : formTags,
  	input   : formTags,
  	output  : formTags,
  	button  : formTags,
  	datalist: formTags,
  	textarea: formTags,
  	option  : { option:true },
  	optgroup: { optgroup:true }
  };

  var voidElements = {
  	__proto__: null,
  	area: true,
  	base: true,
  	basefont: true,
  	br: true,
  	col: true,
  	command: true,
  	embed: true,
  	frame: true,
  	hr: true,
  	img: true,
  	input: true,
  	isindex: true,
  	keygen: true,
  	link: true,
  	meta: true,
  	param: true,
  	source: true,
  	track: true,
  	wbr: true,

  	
  	path: true,
  	circle: true,
  	ellipse: true,
  	line: true,
  	rect: true,
  	use: true,
  	stop: true,
  	polyline: true,
  	polygon: true
  };

  var re_nameEnd = /\s|\//;

  function Parser$1(cbs, options){
  	this._options = options || {};
  	this._cbs = cbs || {};

  	this._tagname = "";
  	this._attribname = "";
  	this._attribvalue = "";
  	this._attribs = null;
  	this._stack = [];

  	this.startIndex = 0;
  	this.endIndex = null;

  	this._lowerCaseTagNames = "lowerCaseTags" in this._options ?
  									!!this._options.lowerCaseTags :
  									!this._options.xmlMode;
  	this._lowerCaseAttributeNames = "lowerCaseAttributeNames" in this._options ?
  									!!this._options.lowerCaseAttributeNames :
  									!this._options.xmlMode;

  	if(this._options.Tokenizer) {
  		Tokenizer$2 = this._options.Tokenizer;
  	}
  	this._tokenizer = new Tokenizer$2(this._options, this);

  	if(this._cbs.onparserinit) { this._cbs.onparserinit(this); }
  }

  require$$1$1(Parser$1, require$$2.EventEmitter);

  Parser$1.prototype._updatePosition = function(initialOffset){
  	if(this.endIndex === null){
  		if(this._tokenizer._sectionStart <= initialOffset){
  			this.startIndex = 0;
  		} else {
  			this.startIndex = this._tokenizer._sectionStart - initialOffset;
  		}
  	}
  	else { this.startIndex = this.endIndex + 1; }
  	this.endIndex = this._tokenizer.getAbsoluteIndex();
  };


  Parser$1.prototype.ontext = function(data){
  	this._updatePosition(1);
  	this.endIndex--;

  	if(this._cbs.ontext) { this._cbs.ontext(data); }
  };

  Parser$1.prototype.onopentagname = function(name){
  	if(this._lowerCaseTagNames){
  		name = name.toLowerCase();
  	}

  	this._tagname = name;

  	if(!this._options.xmlMode && name in openImpliesClose) {
  		for(
  			var el;
  			(el = this._stack[this._stack.length - 1]) in openImpliesClose[name];
  			this.onclosetag(el)
  		){ }
  	}

  	if(this._options.xmlMode || !(name in voidElements)){
  		this._stack.push(name);
  	}

  	if(this._cbs.onopentagname) { this._cbs.onopentagname(name); }
  	if(this._cbs.onopentag) { this._attribs = {}; }
  };

  Parser$1.prototype.onopentagend = function(){
  	this._updatePosition(1);

  	if(this._attribs){
  		if(this._cbs.onopentag) { this._cbs.onopentag(this._tagname, this._attribs); }
  		this._attribs = null;
  	}

  	if(!this._options.xmlMode && this._cbs.onclosetag && this._tagname in voidElements){
  		this._cbs.onclosetag(this._tagname);
  	}

  	this._tagname = "";
  };

  Parser$1.prototype.onclosetag = function(name){
  	var this$1 = this;

  	this._updatePosition(1);
  	if(this._lowerCaseTagNames){
  		name = name.toLowerCase();
  	}
  	
  	
  	
  	if(this._options.xmlMode) {
  		var stack = this._stack;
  		var last = stack.pop();
  		while(last !== name) {
  			if(this$1._cbs.onerror) {
  				this$1._cbs.onerror("Unclosed tag <"+last+">");
  			}
  			last = stack.pop();
  		}
  		this.onopentagend();
  		if(this._cbs.onclosetag) {
  			this._cbs.onclosetag(last);
  		}
  	} else {
  		if(this._stack.length && (!(name in voidElements))) {
  			var pos = this._stack.lastIndexOf(name);
  			if(pos !== -1){
  				if(this._cbs.onclosetag){
  					pos = this._stack.length - pos;
  					while(pos--) { this$1._cbs.onclosetag(this$1._stack.pop()); }
  				}
  				else { this._stack.length = pos; }
  			} else if(name === "p"){
  				this.onopentagname(name);
  				this._closeCurrentTag();
  			}
  		} else if(name === "br" || name === "p"){
  			this.onopentagname(name);
  			this._closeCurrentTag();
  		}
  	}
  };

  Parser$1.prototype.onselfclosingtag = function(){
  	if(this._options.xmlMode || this._options.recognizeSelfClosing){
  		this._closeCurrentTag();
  	} else {
  		this.onopentagend();
  	}
  };

  Parser$1.prototype._closeCurrentTag = function(){
  	var name = this._tagname;

  	this.onopentagend();

  	
  	
  	if(this._stack[this._stack.length - 1] === name){
  		if(this._cbs.onclosetag){
  			this._cbs.onclosetag(name);
  		}
  		this._stack.pop();
  	}
  };

  Parser$1.prototype.onattribname = function(name){
  	if(this._lowerCaseAttributeNames){
  		name = name.toLowerCase();
  	}
  	this._attribname = name;
  };

  Parser$1.prototype.onattribdata = function(value){
  	this._attribvalue += value;
  };

  Parser$1.prototype.onattribend = function(){
  	if(this._cbs.onattribute) { this._cbs.onattribute(this._attribname, this._attribvalue); }
  	if(
  		this._attribs &&
  		!Object.prototype.hasOwnProperty.call(this._attribs, this._attribname)
  	){
  		this._attribs[this._attribname] = this._attribvalue;
  	}
  	this._attribname = "";
  	this._attribvalue = "";
  };

  Parser$1.prototype._getInstructionName = function(value){
  	var idx = value.search(re_nameEnd),
  	    name = idx < 0 ? value : value.substr(0, idx);

  	if(this._lowerCaseTagNames){
  		name = name.toLowerCase();
  	}

  	return name;
  };

  Parser$1.prototype.ondeclaration = function(value){
  	if(this._cbs.onprocessinginstruction){
  		var name = this._getInstructionName(value);
  		this._cbs.onprocessinginstruction("!" + name, "!" + value);
  	}
  };

  Parser$1.prototype.onprocessinginstruction = function(value){
  	if(this._cbs.onprocessinginstruction){
  		var name = this._getInstructionName(value);
  		this._cbs.onprocessinginstruction("?" + name, "?" + value);
  	}
  };

  Parser$1.prototype.oncomment = function(value){
  	this._updatePosition(4);

  	if(this._cbs.oncomment) { this._cbs.oncomment(value); }
  	if(this._cbs.oncommentend) { this._cbs.oncommentend(); }
  };

  Parser$1.prototype.oncdata = function(value){
  	this._updatePosition(1);

  	if(this._options.xmlMode || this._options.recognizeCDATA){
  		if(this._cbs.oncdatastart) { this._cbs.oncdatastart(); }
  		if(this._cbs.ontext) { this._cbs.ontext(value); }
  		if(this._cbs.oncdataend) { this._cbs.oncdataend(); }
  	} else {
  		this.oncomment("[CDATA[" + value + "]]");
  	}
  };

  Parser$1.prototype.onerror = function(err){
  	if(this._cbs.onerror) { this._cbs.onerror(err); }
  };

  Parser$1.prototype.onend = function(){
  	if(this._cbs.onclosetag){
  		for(
  			var i = this._stack.length;
  			i > 0;
  			this._cbs.onclosetag(this._stack[--i])
  		){ }
  	}
  	if(this._cbs.onend) { this._cbs.onend(); }
  };



  Parser$1.prototype.reset = function(){
  	if(this._cbs.onreset) { this._cbs.onreset(); }
  	this._tokenizer.reset();

  	this._tagname = "";
  	this._attribname = "";
  	this._attribs = null;
  	this._stack = [];

  	if(this._cbs.onparserinit) { this._cbs.onparserinit(this); }
  };


  Parser$1.prototype.parseComplete = function(data){
  	this.reset();
  	this.end(data);
  };

  Parser$1.prototype.write = function(chunk){
  	this._tokenizer.write(chunk);
  };

  Parser$1.prototype.end = function(chunk){
  	this._tokenizer.end(chunk);
  };

  Parser$1.prototype.pause = function(){
  	this._tokenizer.pause();
  };

  Parser$1.prototype.resume = function(){
  	this._tokenizer.resume();
  };


  Parser$1.prototype.parseChunk = Parser$1.prototype.write;
  Parser$1.prototype.done = Parser$1.prototype.end;

  var Parser_1 = Parser$1;

  Parser_1.prototype.ondeclaration = function(value){
    if(this._cbs.ondeclaration){
      this._cbs.ondeclaration(value);
    } else if(this._cbs.onprocessinginstruction){
      var name = this._getInstructionName(value);
      this._cbs.onprocessinginstruction("!" + name, "!" + value);
    }
  };


  Parser_1.prototype.oncdata = function(value){
    this._updatePosition(1);

    if(this._options.xmlMode || this._options.recognizeCDATA){
      if(this._cbs.oncdatastart) { this._cbs.oncdatastart(value); }
      
      if(this._cbs.oncdataend) { this._cbs.oncdataend(); }
    } else {
      this.oncomment("[CDATA[" + value + "]]");
    }
  };

  function parseMarkup(markup, options) {
    var format = options.ownerDocument ? options.ownerDocument.format : options.format;
    
    if (!format) {
      throw new Error("Either 'ownerDocument' or 'format' must be set.")
    }
    var parserOptions = Object.assign({}, options, {
      xmlMode : (format === 'xml'),
    });
    var handler = new DomHandler({ format: format, elementFactory: options.elementFactory });
    var parser = new Parser_1(handler, parserOptions);
    parser.end(markup);
    return handler.document
  }


  var re_whitespace = /\s+/g;
  var re_doctype = /^DOCTYPE\s+([^\s]+)(?:\s+PUBLIC\s+["]([^"]+)["](?:\s+["]([^"]+)["])?)\s*$/;


  var DomHandler = function DomHandler(options) {
    if ( options === void 0 ) options = {};

    this.elementFactory = options.elementFactory;
    if (!this.elementFactory) { throw new Error("'elementFactory' is required") }
    this.options = options;
    this.document = null;
    this._tagStack = [];
  };


  DomHandler.prototype.onparserinit = function onparserinit (){
    this.document = this.elementFactory('document', { format: this.options.format });
    this._tagStack = [this.document];
  };

  DomHandler.prototype.onend = function onend (){
    
    if (this._tagStack.length>1) {
      throw new Error("Unexpected EOF. Tag was opened but not closed.")
    }
  };

  DomHandler.prototype.onerror = function onerror (error) {
    throw new Error(error)
  };

  DomHandler.prototype.onclosetag = function onclosetag () {
    this._tagStack.pop();
  };

  DomHandler.prototype._addDomElement = function _addDomElement (element) {
    var parent = this._tagStack[this._tagStack.length - 1];
    if (!parent.childNodes) { parent.childNodes = []; }
    var siblings = parent.childNodes;

    var previousSibling = siblings[siblings.length - 1];
    
    element.next = null;
    if(previousSibling){
      element.prev = previousSibling;
      previousSibling.next = element;
    } else {
      element.prev = null;
    }
    
    siblings.push(element);
    element.parent = parent || null;
  };

  DomHandler.prototype.onopentag = function onopentag (name, attributes) {
    var element = this.document.createElement(name);
    forEach(attributes, function (val, key) {
      element.setAttribute(key, val);
    });
    this._addDomElement(element);
    this._tagStack.push(element);
  };

  DomHandler.prototype.ontext = function ontext (text) {
    if (this.options.normalizeWhitespace) {
      text = text.replace(re_whitespace, " ");
    }
    var lastTag;
    var _top = this._tagStack[this._tagStack.length - 1];
    if (_top && _top.childNodes) { lastTag = _top.childNodes[_top.childNodes.length - 1]; }
    if (lastTag && lastTag.type === domelementtype.Text) {
      lastTag.data += text;
    } else {
      var element = this.document.createTextNode(text);
      this._addDomElement(element);
    }
  };

  DomHandler.prototype.oncomment = function oncomment (data) {
    var lastTag = this._tagStack[this._tagStack.length - 1];
    if(lastTag && lastTag.type === domelementtype.Comment){
      lastTag.data += data;
    } else {
      var element = this.document.createComment(data);
      this._addDomElement(element);
      this._tagStack.push(element);
    }
  };

  DomHandler.prototype.oncommentend = function oncommentend () {
    this._tagStack.pop();
  };

  DomHandler.prototype.oncdatastart = function oncdatastart (data) {
    var element = this.document.createCDATASection(data);
    this._addDomElement(element);
    this._tagStack.push(element);
  };

  DomHandler.prototype.oncdataend = function oncdataend () {
    this._tagStack.pop();
  };

  DomHandler.prototype.onprocessinginstruction = function onprocessinginstruction (name, data) {
    var element = this.document.createProcessingInstruction(name, data);
    this._addDomElement(element);
  };

  DomHandler.prototype.ondeclaration = function ondeclaration (data) {
    if (data.startsWith('DOCTYPE')) {
      var m = re_doctype.exec(data);
      if (!m) { throw new Error('Could not parse DOCTYPE element: ' + data) }
      this.document.setDoctype(m[1], m[2], m[3]);
    } else {
      throw new Error('Not implemented: parse declaration '+data)
    }
  };

  var MemoryDOMElementFactory = function (type, data) {
    return new MemoryDOMElement(type, data)
  };

  var MemoryDOMElement = (function (DOMElement$$1) {
    function MemoryDOMElement(type, args) {
      if ( args === void 0 ) args = {};

      DOMElement$$1.call(this);

      this.type = type;
      if (!type) { throw new Error("'type' is mandatory") }

      this.ownerDocument = args.ownerDocument;
      
      if (type !== 'document' && !this.ownerDocument) {
        throw new Error("'ownerDocument' is mandatory")
      }

      
      
      

      switch(type) {
        case domelementtype.Tag: {
          if (!args.name) { throw new Error("'name' is mandatory.") }
          this.name = this._normalizeName(args.name);
          this.nameWithoutNS = nameWithoutNS(this.name);
          this.properties = new Map();
          this.attributes = new Map();
          this.classes = new Set();
          this.styles = new Map();
          this.eventListeners = [];
          this.childNodes = args.children || args.childNodes || [];
          this._assign(args);
          break
        }
        case domelementtype.Text:
        case domelementtype.Comment: {
          this.data = args.data || '';
          break
        }
        case domelementtype.CDATA: {
          this.data = args.data || '';
          break
        }
        case domelementtype.Directive: {
          if (!args.name) { throw new Error("'name' is mandatory.") }
          this.name = this._normalizeName(args.name);
          this.nameWithoutNS = nameWithoutNS(this.name);
          this.data = args.data;
          break
        }
        case domelementtype.Doctype: {
          this.data = args.data;
          break
        }
        case 'document': {
          var format = args.format;
          this.format = format;
          if (!format) { throw new Error("'format' is mandatory.") }
          this.childNodes = args.children || args.childNodes || [];
          switch(format) {
            case 'xml':
              this.contentType = 'application/xml';
              break
            case 'html':
              this.contentType = 'text/html';
              break
            default:
              throw new Error('Unsupported format ' + format)
          }
          break
        }
        default:
          this.name = null;
          this.properties = new Map();
          this.attributes = new Map();
          this.classes = new Set();
          this.styles = new Map();
          this.eventListeners = [];
          this.childNodes = args.children || args.childNodes || [];
      }
    }

    if ( DOMElement$$1 ) MemoryDOMElement.__proto__ = DOMElement$$1;
    MemoryDOMElement.prototype = Object.create( DOMElement$$1 && DOMElement$$1.prototype );
    MemoryDOMElement.prototype.constructor = MemoryDOMElement;

    var prototypeAccessors = { tagName: { configurable: true },children: { configurable: true } };

    MemoryDOMElement.prototype.getNativeElement = function getNativeElement () {
      return this
    };

    MemoryDOMElement.prototype.getNodeType = function getNodeType () {
      switch(this.type) {
        case domelementtype.Tag:
        case domelementtype.Script:
        case domelementtype.Style:
          return 'element'
        default:
          return this.type
      }
    };

    MemoryDOMElement.prototype.isTextNode = function isTextNode () {
      return this.type === "text"
    };

    MemoryDOMElement.prototype.isElementNode = function isElementNode () {
      return this.type === "tag" || this.type === "script"
    };

    MemoryDOMElement.prototype.isCommentNode = function isCommentNode () {
      return this.type === "comment"
    };

    MemoryDOMElement.prototype.isDocumentNode = function isDocumentNode () {
      return this.type === "document"
    };

    MemoryDOMElement.prototype.isComponentNode = function isComponentNode () {
      return this.type === "component"
    };

    MemoryDOMElement.prototype.clone = function clone (deep) {
      var clone = new MemoryDOMElement(this.type, this);
      if (this.childNodes) {
        clone.childNodes.length = 0;
        if (deep) {
          this.childNodes.forEach(function (child) {
            clone.appendChild(child.clone(deep));
          });
        }
      }
      return clone
    };

    prototypeAccessors.tagName.get = function () {
      return this.getTagName()
    };

    prototypeAccessors.tagName.set = function (tagName) {
      this.setTagName(tagName);
    };

    MemoryDOMElement.prototype.getTagName = function getTagName () {
      return this.name
    };

    MemoryDOMElement.prototype.setTagName = function setTagName (tagName) {
      if (this._isXML()) {
        this.name = String(tagName);
      } else {
        this.name = String(tagName).toLowerCase();
      }
      this.nameWithoutNS = nameWithoutNS(this.name);
      return this
    };

    MemoryDOMElement.prototype.hasAttribute = function hasAttribute (name) {
      return this.attributes.has(name)
    };

    MemoryDOMElement.prototype.getAttribute = function getAttribute (name) {
      return this.attributes.get(name)
    };

    MemoryDOMElement.prototype.setAttribute = function setAttribute (name, value) {
      value = String(value);
      
      switch(name) {
        case 'class':
          this.classes = new Set();
          parseClasses(this.classes, value);
          break
        case 'style':
          this.styles = new Map();
          parseStyles(this.styles, value);
          break
        default:
          
      }
      this.attributes.set(name, value);
      if (this._isHTML()) {
        deriveHTMLPropertyFromAttribute(this, name, value);
      }
      return this
    };

    MemoryDOMElement.prototype.removeAttribute = function removeAttribute (name) {
      switch(name) {
        case 'class':
          this.classes = new Set();
          break
        case 'style':
          this.styles = new Map();
          break
        default:
          
      }
      this.attributes.delete(name);
      return this
    };

    MemoryDOMElement.prototype.getAttributes = function getAttributes () {
      return this.attributes
    };

    MemoryDOMElement.prototype.getProperty = function getProperty (name) {
      if (this.properties) {
        return this.properties.get(name)
      }
    };

    MemoryDOMElement.prototype.setProperty = function setProperty (name, value) {
      if (this.properties) {
        if (this._isXML()) {
          throw new Error('setProperty() is only be used on HTML elements')
        }
        _setHTMLPropertyValue(this, name, value);
      }
      return this
    };

    MemoryDOMElement.prototype.hasClass = function hasClass (name) {
      if (this.classes) {
        return this.classes.has(name)
      }
    };

    MemoryDOMElement.prototype.addClass = function addClass (name) {
      this.classes.add(name);
      this.attributes.set('class', stringifyClasses(this.classes));
      return this
    };

    MemoryDOMElement.prototype.removeClass = function removeClass (name) {
      if (this.classes && this.classes.has(name)) {
        this.classes.delete(name);
        this.attributes.set('class', stringifyClasses(this.classes));
      }
      return this
    };

    MemoryDOMElement.prototype.getContentType = function getContentType () {
      return this.getOwnerDocument().contentType
    };

    MemoryDOMElement.prototype.getDoctype = function getDoctype () {
      if (this.isDocumentNode()) {
        return _findDocTypeElement(this)
      } else {
        return this.getOwnerDocument().getDoctype()
      }
    };

    MemoryDOMElement.prototype.setDoctype = function setDoctype (qualifiedNameStr, publicId, systemId) {
      
      var doc = this.getOwnerDocument();
      var oldDocType = _findDocTypeElement(doc);
      var newDocType = this.createDocumentType(qualifiedNameStr, publicId, systemId);
      if (oldDocType) {
        doc.replaceChild(newDocType, oldDocType);
      } else {
        
        doc.insertBefore(newDocType, doc.getChildren()[0]);
      }
      doc.doctype = newDocType;
    };

    MemoryDOMElement.prototype.getInnerHTML = function getInnerHTML () {
      return domUtils.getInnerHTML(this, { decodeEntities: true })
    };

    
    
    MemoryDOMElement.prototype.setInnerHTML = function setInnerHTML (html) {
      var this$1 = this;

      if (this.childNodes) {
        var _doc = parseMarkup(html, {
          ownerDocument: this.getOwnerDocument(),
          decodeEntities: true,
          elementFactory: MemoryDOMElementFactory
        });
        this.empty();
        
        
        _doc.childNodes.slice(0).forEach(function (child) {
          this$1.appendChild(child);
        });
      }
      return this
    };

    MemoryDOMElement.prototype.getOuterHTML = function getOuterHTML () {
      return domUtils.getOuterHTML(this, { xmlMode: this._isXML(), decodeEntities: true })
    };

    MemoryDOMElement.prototype.getTextContent = function getTextContent () {
      return domUtils.getText(this)
    };

    MemoryDOMElement.prototype.setTextContent = function setTextContent (text) {
      switch(this.type) {
        case domelementtype.Text:
        case domelementtype.Comment:
        case domelementtype.CDATA: {
          this.data = text;
          break
        }
        default: {
          if (this.childNodes) {
            var child = this.createTextNode(text);
            this.empty();
            this.appendChild(child);
          }
        }
      }
      return this
    };

    MemoryDOMElement.prototype.getStyle = function getStyle (name) {
      if (this.styles) {
        return this.styles.get(name)
      }
    };

    MemoryDOMElement.prototype.setStyle = function setStyle (name, value) {
      if (this.styles) {
        if (DOMElement$$1.pxStyles[name] && isNumber(value)) {
          value = value + "px";
        }
        this.styles.set(name, value);
        this.attributes.set('style', stringifyStyles(this.styles));
      }
      return this
    };

    MemoryDOMElement.prototype.is = function is (cssSelector) {
      return cssSelect.is(this, cssSelector, { xmlMode: this._isXML() })
    };

    MemoryDOMElement.prototype.find = function find (cssSelector) {
      return cssSelect.selectOne(cssSelector, this, { xmlMode: this._isXML() })
    };

    MemoryDOMElement.prototype.findAll = function findAll (cssSelector) {
      return cssSelect.selectAll(cssSelector, this, { xmlMode: this._isXML() })
    };

    MemoryDOMElement.prototype.getChildCount = function getChildCount () {
      if (this.childNodes) {
        return this.childNodes.length
      } else {
        return 0
      }
    };

    MemoryDOMElement.prototype.getChildNodes = function getChildNodes () {
      return this.childNodes.slice(0)
    };

    MemoryDOMElement.prototype.getChildren = function getChildren () {
      return this.childNodes.filter(function(node) {
        return node.type === "tag"
      })
    };

    prototypeAccessors.children.get = function () {
      return this.getChildren()
    };

    MemoryDOMElement.prototype.getChildAt = function getChildAt (pos) {
      if (this.childNodes) {
        return this.childNodes[pos]
      }
    };

    MemoryDOMElement.prototype.getChildIndex = function getChildIndex (child) {
      if (this.childNodes) {
        return this.childNodes.indexOf(child)
      }
    };

    MemoryDOMElement.prototype.getLastChild = function getLastChild () {
      if (this.childNodes) {
        return last$1(this.childNodes)
      }
    };

    MemoryDOMElement.prototype.getFirstChild = function getFirstChild () {
      if (this.childNodes) {
        return this.childNodes[0]
      }
    };

    MemoryDOMElement.prototype.getNextSibling = function getNextSibling () {
      return this.next
    };

    MemoryDOMElement.prototype.getPreviousSibling = function getPreviousSibling () {
      return this.prev
    };

    MemoryDOMElement.prototype.getParent = function getParent () {
      return this.parent
    };

    MemoryDOMElement.prototype.getOwnerDocument = function getOwnerDocument () {
      return (this.type === 'document') ? this : this.ownerDocument
    };

    MemoryDOMElement.prototype.getFormat = function getFormat () {
      return this.getOwnerDocument().format
    };

    MemoryDOMElement.prototype.createDocument = function createDocument (format) {
      return MemoryDOMElement.createDocument(format)
    };

    MemoryDOMElement.prototype.createElement = function createElement (tagName) {
      return new MemoryDOMElement(domelementtype.Tag, { name: tagName, ownerDocument: this.getOwnerDocument() })
    };

    MemoryDOMElement.prototype.createTextNode = function createTextNode (text) {
      return new MemoryDOMElement(domelementtype.Text, { data: text, ownerDocument: this.getOwnerDocument() })
    };

    MemoryDOMElement.prototype.createComment = function createComment (data) {
      return new MemoryDOMElement(domelementtype.Comment, { data: data, ownerDocument: this.getOwnerDocument() })
    };

    MemoryDOMElement.prototype.createProcessingInstruction = function createProcessingInstruction (name, data) {
      return new MemoryDOMElement(domelementtype.Directive, { name: name, data: data, ownerDocument: this.getOwnerDocument() })
    };

    MemoryDOMElement.prototype.createDocumentType = function createDocumentType (qualifiedNameStr, publicId, systemId) {
      return new MemoryDOMDoctype(domelementtype.Doctype, { data: { name: qualifiedNameStr, publicId: publicId, systemId: systemId }, ownerDocument: this.getOwnerDocument() })
    };

    MemoryDOMElement.prototype.createCDATASection = function createCDATASection (data) {
      return new MemoryDOMElement(domelementtype.CDATA, { data: data, ownerDocument: this.getOwnerDocument() })
    };

    MemoryDOMElement.prototype.appendChild = function appendChild (child) {
      if (this.childNodes && !isNil(child)) {
        child = this._normalizeChild(child);
        if (!child) { return this }
        domUtils.appendChild(this, child);
        child.ownerDocument = this.getOwnerDocument();
      }
      return this
    };

    MemoryDOMElement.prototype.removeChild = function removeChild (child) {
      if (child.parentNode === this) {
        child.remove();
      }
    };

    MemoryDOMElement.prototype.insertAt = function insertAt (pos, child) {
      child = this._normalizeChild(child);
      if (!child) { return this }
      var childNodes = this.childNodes;
      if (childNodes) {
        
        if (pos >= childNodes.length) {
          domUtils.appendChild(this, child);
        } else {
          domUtils.prepend(childNodes[pos], child);
        }
        child.ownerDocument = this.getOwnerDocument();
      }
      return this
    };

    MemoryDOMElement.prototype.insertBefore = function insertBefore (newChild, before) {
      if (isNil(before)) {
        return this.appendChild(newChild)
      } else if (this.childNodes) {
        var pos = this.childNodes.indexOf(before);
        if (pos > -1) {
          domUtils.prepend(before, newChild);
          newChild.ownerDocument = this.getOwnerDocument();
        } else {
          throw new Error('insertBefore(): reference node is not a child of this element.')
        }
      }
      return this
    };

    MemoryDOMElement.prototype.removeAt = function removeAt (pos) {
      var childNodes = this.childNodes;
      if (childNodes) {
        var child = childNodes[pos];
        child.remove();
      }
      return this
    };

    MemoryDOMElement.prototype.empty = function empty () {
      var childNodes = this.childNodes;
      if (childNodes) {
        childNodes.forEach(function (child) {
          child.next = child.prev = child.parent = null;
        });
        childNodes.length = 0;
      }
      return this
    };

    MemoryDOMElement.prototype.remove = function remove () {
      domUtils.removeElement(this);
      return this
    };

    MemoryDOMElement.prototype.replaceChild = function replaceChild (oldChild, newChild) {
      if (oldChild.parent === this) {
        oldChild.replaceWith(newChild);
      }
      return this
    };

    MemoryDOMElement.prototype.replaceWith = function replaceWith (newEl) {
      newEl = this._normalizeChild(newEl);
      domUtils.replaceElement(this, newEl);
      newEl.ownerDocument = this.getOwnerDocument();
      return this
    };

    MemoryDOMElement.prototype.getEventListeners = function getEventListeners () {
      return this.eventListeners || []
    };

    MemoryDOMElement.prototype.click = function click () {
      this.emit('click', { target: this });
      return this
    };

    MemoryDOMElement.prototype.emit = function emit (name, data) {
      this._propagateEvent(new MemoryDOMElementEvent(name, this, data));
    };

    MemoryDOMElement.prototype._propagateEvent = function _propagateEvent (event) {
      var listeners = this.eventListeners;
      if (listeners) {
        var listener = listeners.find(function (l) {
          return l.eventName === event.type
        });
        if (listener) { listener.handler(event); }
        if (event.stopped) { return }
        var p = this.parentNode;
        if (p) { p._propagateEvent(event); }
      }
    };

    MemoryDOMElement.prototype.removeAllEventListeners = function removeAllEventListeners () {
      this.eventListeners = [];
      return this
    };

    MemoryDOMElement.prototype._assign = function _assign (other) {
      var this$1 = this;

      if (other.name) { this.name = other.name; }
      if (this.classes && other.classes) {
        other.classes.forEach(function (val) {
          this$1.classes.add(val);
        });
      }
      if (this.styles && other.styles) {
        forEach(other.styles, function (val, name) {
          this$1.styles.set(name, val);
        });
      }
      
      
      var otherAttributes = other.attributes || other.attribs;
      if (this.attributes && otherAttributes) {
        forEach(otherAttributes, function (val, name) {
          switch (name) {
            case 'class': {
              parseClasses(this$1.classes, val);
              break
            }
            case 'style': {
              parseStyles(this$1.styles, val);
              break
            }
            default:
              
          }
          this$1.attributes.set(name, val);
        });
      }
      if (this.eventListeners && other.eventListeners) {
        this.eventListeners = this.eventListeners.concat(other.eventListeners);
      }
    };

    MemoryDOMElement.prototype._normalizeChild = function _normalizeChild (child) {
      if (isNil(child)) { return }

      if (isString(child)) {
        child = this.createTextNode(child);
      }
      
      if (!child || !child._isMemoryDOMElement) {
        throw new Error('Illegal argument: only String and MemoryDOMElement instances are valid.')
      }
      return child
    };

    MemoryDOMElement.prototype._normalizeName = function _normalizeName (name) {
      if (this._isXML()) {
        return name
      } else {
        return name.toLowerCase()
      }
    };

    MemoryDOMElement.prototype._isHTML = function _isHTML () {
      return this.getFormat() === 'html'
    };

    MemoryDOMElement.prototype._isXML = function _isXML () {
      return this.getFormat() === 'xml'
    };

    Object.defineProperties( MemoryDOMElement.prototype, prototypeAccessors );

    return MemoryDOMElement;
  }(DOMElement));

  MemoryDOMElement.prototype._isMemoryDOMElement = true;

  MemoryDOMElement.createDocument = function(format) {
    if (format === 'xml') {
      return new MemoryDOMElement('document', { format: format })
    } else {
      return MemoryDOMElement.parseMarkup(DOMElement.EMPTY_HTML, 'html')
    }
  };

  MemoryDOMElement.parseMarkup = function(str, format, options) {
    if ( options === void 0 ) options={};

    if (!str) {
      return MemoryDOMElement.createDocument(format)
    }
    var parserOpts = Object.assign({
      format: format,
      decodeEntities: true,
      elementFactory: MemoryDOMElementFactory
    }, options);
    
    if (options.raw) {
      return parseMarkup(str, parserOpts)
    }
    if (options.snippet) {
      str = "<__snippet__>" + str + "</__snippet__>";
    }
    var doc;
    if (format === 'html') {
      doc = parseMarkup(str, parserOpts);
      _sanitizeHTMLStructure(doc);
    } else if (format === 'xml') {
      doc = parseMarkup(str, parserOpts);
    }
    if (options.snippet) {
      var childNodes = doc.find('__snippet__').childNodes;
      if (childNodes.length === 1) {
        return childNodes[0]
      } else {
        return childNodes
      }
    } else {
      return doc
    }
  };

  MemoryDOMElement.wrap =
  MemoryDOMElement.wrapNativeElement = function(el) {
    if (inBrowser) {
      
      
      
      
      if (el === window || el === window.document) {
        return new DOMElementStub()
      }
      
      
      
      else if (el instanceof window.Node || el._isBrowserDOMElement) ;
    }
    
    if (!el._isMemoryDOMElement) {
      throw new Error('Illegal argument: expected MemoryDOMElement instance')
    }
    return el
  };

  MemoryDOMElement.unwrap = function(el) {
    
    if (!el._isMemoryDOMElement) {
      throw new Error('Illegal argument: expected MemoryDOMElement instance')
    }
    return el
  };




  MemoryDOMElement.isReverse = function() {
    return false
  };


  var _browserWindowStub;
  MemoryDOMElement.getBrowserWindow = function() {
    
    if (!_browserWindowStub) {
      _browserWindowStub = MemoryDOMElement.createDocument('html');
    }
    return _browserWindowStub
  };

  var MemoryDOMDoctype = (function (MemoryDOMElement) {
    function MemoryDOMDoctype () {
      MemoryDOMElement.apply(this, arguments);
    }

    if ( MemoryDOMElement ) MemoryDOMDoctype.__proto__ = MemoryDOMElement;
    MemoryDOMDoctype.prototype = Object.create( MemoryDOMElement && MemoryDOMElement.prototype );
    MemoryDOMDoctype.prototype.constructor = MemoryDOMDoctype;

    var prototypeAccessors$1 = { name: { configurable: true },publicId: { configurable: true },systemId: { configurable: true } };

    prototypeAccessors$1.name.get = function () { return this.data.name };
    prototypeAccessors$1.publicId.get = function () { return this.data.publicId };
    prototypeAccessors$1.systemId.get = function () { return this.data.systemId };

    Object.defineProperties( MemoryDOMDoctype.prototype, prototypeAccessors$1 );

    return MemoryDOMDoctype;
  }(MemoryDOMElement));

  function parseClasses(classes, classStr) {
    classStr.split(/\s+/).forEach(function (name) {
      classes.add(name);
    });
  }

  function stringifyClasses(classes) {
    return Array.from(classes).join(' ')
  }

  function parseStyles(styles, styleStr) {
    styleStr = (styleStr || '').trim();
    if (!styleStr) { return }
    styleStr.split(';').forEach(function (style) {
      var n = style.indexOf(':');
      
      if (n < 1 || n === style.length-1) { return }
      var name = style.slice(0,n).trim();
      var val = style.slice(n+1).trim();
      styles.set(name, val);
    });
  }

  function stringifyStyles(styles) {
    if (!styles) { return '' }
    var str = Object.keys(styles).map(function (name) {
      return name + ':' + styles[name]
    }).join(';');
    if (str.length > 0) { str += ';'; }
    return str
  }

  var BUILTIN_EVENTS = [
    'keydown', 'keyup', 'keypress',
    'mousedown', 'mouseup', 'mouseover', 'click', 'dblclick'
  ].reduce(function (m, k){m[k]=true;return m}, {});

  var MemoryDOMElementEvent = function MemoryDOMElementEvent(type, target, detail) {
    this.type = type;
    this.timeStamp = Date.now();
    this.target = target;

    if (BUILTIN_EVENTS[type]) {
      
      if (detail) {
        Object.assign(this, detail);
      }
    } else {
      this.detail = detail;
    }
  };

  MemoryDOMElementEvent.prototype.stopPropagation = function stopPropagation () {
    this.stopped = true;
  };

  MemoryDOMElementEvent.prototype.preventDefault = function preventDefault () {
    this.defaultPrevented = true;
  };

  var DOMElementStub = function DOMElementStub () {};

  DOMElementStub.prototype.on = function on () {};
  DOMElementStub.prototype.off = function off (){};

  function nameWithoutNS(name) {
    var idx = name.indexOf(':');
    if (idx > 0) {
      return name.slice(idx+1)
    } else {
      return name
    }
  }



  var ATTR_TO_PROPS = {
    "input": {
      "value": true,
      "checked": function (el, name, value) {
        var checked = (value !== 'off');
        el.setProperty('checked', checked);
      }
    }
  };

  function deriveHTMLPropertyFromAttribute(el, name, value) {
    var mappings = ATTR_TO_PROPS[el.tagName];
    if (mappings) {
      var mapper = mappings[name];
      if (mapper === true) {
        el.setProperty(name, value);
      } else if (mapper) {
        mapper(el, name, value);
      }
    }
  }

  var PROPERTY_TRANSFORMATIONS = {
    "input": {
      "checked": function (el, name, value) {
        if (value === true) {
          el.properties.set(name, true);
          el.properties.set('value', 'on');
        } else {
          el.properties.set(name, false);
          el.properties.set('value', 'off');
        }
      },
      "value": function (el, name, value) {
        var type = el.getAttribute('type');
        switch(type) {
          case 'checkbox':
            if (value === 'on') {
              el.properties.set(name, true);
              el.properties.set('value', 'on');
            } else {
              el.properties.set(name, false);
              el.properties.set('value', 'off');
            }
            break
          default:
            _setProperty(el, name, value);
        }
      }
    }
  };

  function _setProperty(el, name, value) {
    if (value === undefined) {
      el.properties.delete(name);
    } else {
      el.properties.set(name, String(value));
    }
  }

  function _setHTMLPropertyValue(el, name, value) {
    var trafos = PROPERTY_TRANSFORMATIONS[el.tagName];
    if (trafos) {
      var mapper = trafos[name];
      if (mapper) {
        mapper(el, name, value);
        return
      }
    }
    _setProperty(el, name, value);
  }

  function _sanitizeHTMLStructure(doc) {
    
    
    
    
    var htmlEl = doc.find('html');
    if (!htmlEl) {
      
      
      var headEl = doc.find('head');
      var titleEl = doc.find('title');
      var metaEls = doc.findAll('meta');
      var bodyEl = doc.find('body');
      if (headEl) { headEl.remove(); }
      if (titleEl) { titleEl.remove(); }
      metaEls.forEach(function (e) { return e.remove(); });
      if (bodyEl) { bodyEl.remove(); }

      
      
      var contentNodes = doc.childNodes.slice();
      contentNodes.forEach(function (c){c.parent = null;});
      doc.childNodes.length = 0;

      htmlEl = doc.createElement('html');
      
      
      
      if (!headEl) {
        headEl = doc.createElement('head');
        headEl.appendChild(titleEl);
        headEl.append(metaEls);
        htmlEl.appendChild(headEl);
      }
      if (!bodyEl) {
        bodyEl = doc.createElement('body');
        bodyEl.append(contentNodes);
      }
      htmlEl.appendChild(bodyEl);

      doc.append(htmlEl);
    }
  }

  function _findDocTypeElement(doc) {
    
    if (doc.doctype) { return doc.doctype }
    var childNodes = doc.childNodes;
    for (var i = 0; i < childNodes.length; i++) {
      var child = childNodes[i];
      if (child.type === domelementtype.Doctype) {
        doc.doctype = child;
        return child
      }
    }
  }

  var DefaultDOMElement = {};

  DefaultDOMElement.createDocument = function(format) {
    return _getDefaultImpl().createDocument(format)
  };


  DefaultDOMElement.createElement = function(tagName) {
    console.error("DEPRECATED: every element should have an ownerDocument. Use DefaultDOMElement.createDocument() to create a document first");
    var doc = DefaultDOMElement.createDocument('html');
    return doc.createElement(tagName)
  };


  DefaultDOMElement.createTextNode = function(text) {
    console.error("DEPRECATED: every element should have a ownerDocument. Use DefaultDOMElement.createDocument() to create a document first");
    var doc = DefaultDOMElement.createDocument('html');
    return doc.createTextNode(text)
  };


  DefaultDOMElement.getBrowserWindow = function() {
    return _getDefaultImpl().getBrowserWindow()
  };


  DefaultDOMElement.parseHTML = function(html, options) {
    return _getDefaultImpl().parseMarkup(html, 'html', options)
  };


  DefaultDOMElement.parseXML = function(xml, options) {
    return _getDefaultImpl().parseMarkup(xml, 'xml', options)
  };

  DefaultDOMElement.parseSnippet = function(str, format) {
    return _getDefaultImpl().parseMarkup(str, format, {snippet: true})
  };

  DefaultDOMElement.wrap =
  DefaultDOMElement.wrapNativeElement = function(nativeEl) {
    if (!nativeEl) { throw new Error('Illegal argument') }
    return _getDefaultImpl().wrap(nativeEl)
  };

  DefaultDOMElement.unwrap = function(nativeEl) {
    if (!nativeEl) { throw new Error('Illegal argument') }
    return _getDefaultImpl().unwrap(nativeEl)
  };



  DefaultDOMElement.isReverse = function(anchorNode, anchorOffset, focusNode, focusOffset) {
    return _getDefaultImpl().isReverse(anchorNode, anchorOffset, focusNode, focusOffset)
  };

  function _getDefaultImpl() {
    
    if (platform.inBrowser || platform.inElectron) {
      return BrowserDOMElement
    } else {
      return MemoryDOMElement
    }
  }

  var _global = (typeof global !== 'undefined') ? global : window;
  var substanceGlobals = _global.hasOwnProperty('Substance') ? _global.Substance : _global.Substance = {
    DEBUG_RENDERING: false
  };

  function flattenOften(arr, max) {
    if (!(max > 0)) { throw new Error("'max' must be a positive number") }
    var l = arr.length;
    arr = flatten(arr);
    var round = 1;
    while (round < max && l < arr.length) {
      l = arr.length;
      arr = flatten(arr);
      round++;
    }
    return arr
  }

  var VirtualElement = (function (DOMElement$$1) {
    function VirtualElement(owner) {
      DOMElement$$1.call(this);

      
      this.parent = null;
      
      this._owner = owner;
      
      this._ref = null;
    }

    if ( DOMElement$$1 ) VirtualElement.__proto__ = DOMElement$$1;
    VirtualElement.prototype = Object.create( DOMElement$$1 && DOMElement$$1.prototype );
    VirtualElement.prototype.constructor = VirtualElement;

    var prototypeAccessors = { childNodes: { configurable: true } };

    VirtualElement.prototype.getParent = function getParent () {
      return this.parent
    };

    prototypeAccessors.childNodes.get = function () {
      return this.getChildNodes()
    };

    VirtualElement.prototype.getChildCount = function getChildCount () {
      return this.children.length
    };

    VirtualElement.prototype.getChildAt = function getChildAt (idx) {
      return this.children[idx]
    };

    
    VirtualElement.prototype.getComponent = function getComponent () {
      return this._comp
    };

    
    VirtualElement.prototype.ref = function ref (ref$1) {
      if (!ref$1) { throw new Error('Illegal argument') }
      
      if (this._ref) { throw new Error('A VirtualElement can only be referenced once.') }
      this._ref = ref$1;
      if (this._context) {
        var refs = this._context.refs;
        if(refs[ref$1]) {
          throw new Error('An item with reference "'+ref$1+'" already exists.')
        }
        refs[ref$1] = this;
      }
      return this
    };

    VirtualElement.prototype.isInDocument = function isInDocument () {
      return false
    };

    Object.defineProperties( VirtualElement.prototype, prototypeAccessors );

    return VirtualElement;
  }(DOMElement));

  VirtualElement.prototype._isVirtualElement = true;


  var VirtualHTMLElement = (function (VirtualElement) {
    function VirtualHTMLElement(tagName) {
      VirtualElement.call(this);

      this._tagName = tagName;
      this.classNames = null;
      this.attributes = null;
      this.htmlProps = null;
      this.style = null;
      this.eventListeners = null;

      
      this.children = [];
    }

    if ( VirtualElement ) VirtualHTMLElement.__proto__ = VirtualElement;
    VirtualHTMLElement.prototype = Object.create( VirtualElement && VirtualElement.prototype );
    VirtualHTMLElement.prototype.constructor = VirtualHTMLElement;

    VirtualHTMLElement.prototype.getTagName = function getTagName () {
      return this._tagName
    };

    VirtualHTMLElement.prototype.setTagName = function setTagName (tagName) {
      this._tagName = tagName;
      return this
    };

    VirtualHTMLElement.prototype.hasClass = function hasClass (className) {
      if (this.classNames) {
        return this.classNames.indexOf(className) > -1
      }
      return false
    };

    VirtualHTMLElement.prototype.addClass = function addClass (className) {
      if (!this.classNames) {
        this.classNames = [];
      }
      this.classNames.push(className);
      return this
    };

    VirtualHTMLElement.prototype.removeClass = function removeClass (className) {
      if (this.classNames) {
        this.classNames = without(this.classNames, className);
      }
      return this
    };

    VirtualHTMLElement.prototype.removeAttribute = function removeAttribute (name) {
      if (this.attributes) {
        delete this.attributes[name];
      }
      return this
    };

    VirtualHTMLElement.prototype.getAttribute = function getAttribute (name) {
      if (this.attributes) {
        return this.attributes[name]
      }
    };

    VirtualHTMLElement.prototype.setAttribute = function setAttribute (name, value) {
      if (!this.attributes) {
        this.attributes = {};
      }
      this.attributes[name] = value;
      return this
    };

    VirtualHTMLElement.prototype.getAttributes = function getAttributes () {
      
      
      
      var attributes = {};
      if (this.attributes) {
        Object.assign(attributes, this.attributes);
      }
      if (this.classNames) {
        attributes.class = this.classNames.join(' ');
      }
      if (this.style) {
        attributes.style = map(this.style, function(val, key) {
          return key + ":" + val
        }).join(';');
      }
      return attributes
    };

    VirtualHTMLElement.prototype.getId = function getId () {
      return this.getAttribute('id')
    };

    VirtualHTMLElement.prototype.setId = function setId (id) {
      this.setAttribute('id', id);
      return this
    };

    VirtualHTMLElement.prototype.setTextContent = function setTextContent (text) {
      text = String(text || '');
      this.empty();
      this.appendChild(text);
      return this
    };

    VirtualHTMLElement.prototype.setInnerHTML = function setInnerHTML (html) {
      html = html || '';
      this.empty();
      this._innerHTMLString = html;
      return this
    };

    VirtualHTMLElement.prototype.getInnerHTML = function getInnerHTML () {
      if (!this.hasOwnProperty('_innerHTMLString')) {
        throw new Error('Not supported.')
      } else {
        return this._innerHTMLString
      }
    };

    VirtualHTMLElement.prototype.getValue = function getValue () {
      return this.htmlProp('value')
    };

    VirtualHTMLElement.prototype.setValue = function setValue (value) {
      this.htmlProp('value', value);
      return this
    };

    VirtualHTMLElement.prototype.getChildNodes = function getChildNodes () {
      return this.children
    };

    VirtualHTMLElement.prototype.getChildren = function getChildren () {
      return this.children.filter(function(child) {
        return child.getNodeType() !== "text"
      })
    };

    VirtualHTMLElement.prototype.isTextNode = function isTextNode () {
      return false
    };

    VirtualHTMLElement.prototype.isElementNode = function isElementNode () {
      return true
    };

    VirtualHTMLElement.prototype.isCommentNode = function isCommentNode () {
      return false
    };

    VirtualHTMLElement.prototype.isDocumentNode = function isDocumentNode () {
      return false
    };

    VirtualHTMLElement.prototype.append = function append () {
      if (this._innerHTMLString) {
        throw Error('It is not possible to mix $$.html() with $$.append(). You can call $$.empty() to reset this virtual element.')
      }
      this._append(this.children, arguments);
      return this
    };

    VirtualHTMLElement.prototype.appendChild = function appendChild (child) {
      if (this._innerHTMLString) {
        throw Error('It is not possible to mix $$.html() with $$.append(). You can call $$.empty() to reset this virtual element.')
      }
      this._appendChild(this.children, child);
      return this
    };

    VirtualHTMLElement.prototype.insertAt = function insertAt (pos, child) {
      child = this._normalizeChild(child);
      if (!child) {
        throw new Error('Illegal child: ' + child)
      }
      if (!child._isVirtualElement) {
        throw new Error('Illegal argument for $$.insertAt():' + child)
      }
      if (pos < 0 || pos > this.children.length) {
        throw new Error('insertAt(): index out of bounds.')
      }
      this._insertAt(this.children, pos, child);
      return this
    };

    VirtualHTMLElement.prototype.insertBefore = function insertBefore (child, before) {
      var pos = this.children.indexOf(before);
      if (pos > -1) {
        this.insertAt(pos, child);
      } else {
        throw new Error('insertBefore(): reference node is not a child of this element.')
      }
      return this
    };

    VirtualHTMLElement.prototype.removeAt = function removeAt (pos) {
      if (pos < 0 || pos >= this.children.length) {
        throw new Error('removeAt(): Index out of bounds.')
      }
      this._removeAt(pos);
      return this
    };

    VirtualHTMLElement.prototype.removeChild = function removeChild (child) {
      if (!child || !child._isVirtualElement) {
        throw new Error('removeChild(): Illegal arguments. Expecting a CheerioDOMElement instance.')
      }
      var idx = this.children.indexOf(child);
      if (idx < 0) {
        throw new Error('removeChild(): element is not a child.')
      }
      this.removeAt(idx);
      return this
    };

    VirtualHTMLElement.prototype.replaceChild = function replaceChild (oldChild, newChild) {
      if (!newChild || !oldChild ||
          !newChild._isVirtualElement || !oldChild._isVirtualElement) {
        throw new Error('replaceChild(): Illegal arguments. Expecting BrowserDOMElement instances.')
      }
      var idx = this.children.indexOf(oldChild);
      if (idx < 0) {
        throw new Error('replaceChild(): element is not a child.')
      }
      this.removeAt(idx);
      this.insertAt(idx, newChild);
      return this
    };

    VirtualHTMLElement.prototype.empty = function empty () {
      var children = this.children;
      while (children.length) {
        var child = children.pop();
        child.parent = null;
      }
      delete this._innerHTMLString;
      return this
    };

    VirtualHTMLElement.prototype.getProperty = function getProperty (name) {
      if (this.htmlProps) {
        return this.htmlProps[name]
      }
    };

    VirtualHTMLElement.prototype.setProperty = function setProperty (name, value) {
      if (!this.htmlProps) {
        this.htmlProps = {};
      }
      this.htmlProps[name] = value;
      return this
    };

    VirtualHTMLElement.prototype.removeProperty = function removeProperty (name) {
      if (this.htmlProps) {
        delete this.htmlProps[name];
      }
      return this
    };

    VirtualHTMLElement.prototype.getStyle = function getStyle (name) {
      if (this.style) {
        return this.style[name]
      }
    };

    VirtualHTMLElement.prototype.setStyle = function setStyle (name, value) {
      if (!this.style) {
        this.style = {};
      }
      if (DOMElement.pxStyles[name] && isNumber(value)) { value = value + 'px'; }
      this.style[name] = value;
      return this
    };

    VirtualHTMLElement.prototype._createEventListener = function _createEventListener (eventName, handler, options) {
      options.context = options.context || this._owner._comp;
      return VirtualElement.prototype._createEventListener.call(this, eventName, handler, options)
    };

    VirtualHTMLElement.prototype.getNodeType = function getNodeType () {
      return "element"
    };

    VirtualHTMLElement.prototype.hasInnerHTML = function hasInnerHTML () {
      return Boolean(this._innerHTMLString)
    };

    VirtualHTMLElement.prototype._normalizeChild = function _normalizeChild (child) {
      if (isNil(child)) {
        return
      } else if (child._isVirtualElement) {
        return child
      } else if (isString(child) || isBoolean(child) || isNumber(child)) {
        return new VirtualTextNode(String(child))
      } else {
        throw new Error('Unsupported child type')
      }
    };

    VirtualHTMLElement.prototype._append = function _append (outlet, args) {
      if (args.length === 1 && !isArray(args[0])) {
        this._appendChild(outlet, args[0]);
        return
      }
      var children;
      if (isArray(args[0])) {
        children = args[0];
      } else if (arguments.length > 1) {
        children = Array.prototype.slice.call(args,0);
      } else {
        return
      }
      children.forEach(this._appendChild.bind(this, outlet));
    };

    VirtualHTMLElement.prototype._appendChild = function _appendChild (outlet, child) {
      child = this._normalizeChild(child);
      
      
      if (!child) { return }
      outlet.push(child);
      this._attach(child);
      return child
    };

    VirtualHTMLElement.prototype._insertAt = function _insertAt (outlet, pos, child) {
      if (!child) { return }
      outlet.splice(pos, 0, child);
      this._attach(child);
    };

    VirtualHTMLElement.prototype._removeAt = function _removeAt (outlet, pos) {
      var child = outlet[pos];
      outlet.splice(pos, 1);
      this._detach(child);
    };

    VirtualHTMLElement.prototype._attach = function _attach (child) {
      child.parent = this;
      if (this._context && child._owner !== this._owner && child._ref) {
        this._context.foreignRefs[child._ref] = child;
      }
    };

    VirtualHTMLElement.prototype._detach = function _detach (child) {
      child.parent = null;
      if (this._context && child._owner !== this._owner && child._ref) {
        delete this.context.foreignRefs[child._ref];
      }
    };

    VirtualHTMLElement.prototype._mergeHTMLConfig = function _mergeHTMLConfig (other) {
      if (other.classNames) {
        if (!this.classNames) {
          this.classNames = [];
        }
        this.classNames = this.classNames.concat(other.classNames);
      }
      if (other.attributes) {
        if (!this.attributes) {
          this.attributes = {};
        }
        Object.assign(this.attributes, other.attributes);
      }
      if (other.htmlProps) {
        if (!this.htmlProps) {
          this.htmlProps = {};
        }
        Object.assign(this.htmlProps, other.htmlProps);
      }
      if (other.style) {
        if (!this.style) {
          this.style = {};
        }
        Object.assign(this.style, other.style);
      }
      if (other.eventListeners) {
        if (!this.eventListeners) {
          this.eventListeners = [];
        }
        this.eventListeners = this.eventListeners.concat(other.eventListeners);
      }
    };

    return VirtualHTMLElement;
  }(VirtualElement));

  VirtualHTMLElement.prototype._isVirtualHTMLElement = true;



  var VirtualComponent = (function (VirtualHTMLElement) {
    function VirtualComponent(ComponentClass, props) {
      VirtualHTMLElement.call(this);

      props = props || {};

      this.ComponentClass = ComponentClass;
      this.props = props;
      if (!props.children) {
        props.children = [];
      }
      this.children = props.children;
    }

    if ( VirtualHTMLElement ) VirtualComponent.__proto__ = VirtualHTMLElement;
    VirtualComponent.prototype = Object.create( VirtualHTMLElement && VirtualHTMLElement.prototype );
    VirtualComponent.prototype.constructor = VirtualComponent;

    var prototypeAccessors$1 = { _isVirtualComponent: { configurable: true } };

    prototypeAccessors$1._isVirtualComponent.get = function () { return true; };

    VirtualComponent.prototype.getComponent = function getComponent () {
      return this._comp
    };

    
    
    VirtualComponent.prototype.getChildren = function getChildren () {
      return this.props.children
    };

    VirtualComponent.prototype.getNodeType = function getNodeType () {
      return 'component'
    };

    VirtualComponent.prototype.outlet = function outlet (name) {
      return new Outlet(this, name)
    };

    VirtualComponent.prototype._attach = function _attach (child) {
      child._preliminaryParent = this;
    };

    VirtualComponent.prototype._detach = function _detach (child) {
      child._preliminaryParent = null;
    };

    VirtualComponent.prototype._copyHTMLConfig = function _copyHTMLConfig () {
      return {
        classNames: clone(this.classNames),
        attributes: clone(this.attributes),
        htmlProps: clone(this.htmlProps),
        style: clone(this.style),
        eventListeners: clone(this.eventListeners)
      }
    };

    Object.defineProperties( VirtualComponent.prototype, prototypeAccessors$1 );

    return VirtualComponent;
  }(VirtualHTMLElement));

  var Outlet = function Outlet(virtualEl, name) {
    this.virtualEl = virtualEl;
    this.name = name;
    Object.freeze(this);
  };

  Outlet.prototype._getOutlet = function _getOutlet () {
    var outlet = this.virtualEl.props[this.name];
    if (!outlet) {
      outlet = [];
      this.virtualEl.props[this.name] = outlet;
    }
    return outlet
  };

  Outlet.prototype.append = function append () {
    var outlet = this._getOutlet();
    this.virtualEl._append(outlet, arguments);
    return this
  };

  Outlet.prototype.empty = function empty () {
    var arr = this.virtualEl.props[this.name];
    arr.forEach(function(el) {
      this._detach(el);
    }.bind(this));
    arr.splice(0, arr.length);
    return this
  };


  var VirtualTextNode = (function (VirtualElement) {
    function VirtualTextNode(text) {
      VirtualElement.call(this);
      this.text = text;
    }

    if ( VirtualElement ) VirtualTextNode.__proto__ = VirtualElement;
    VirtualTextNode.prototype = Object.create( VirtualElement && VirtualElement.prototype );
    VirtualTextNode.prototype.constructor = VirtualTextNode;

    var prototypeAccessors$2 = { _isVirtualTextNode: { configurable: true } };

    prototypeAccessors$2._isVirtualTextNode.get = function () { return true; };

    Object.defineProperties( VirtualTextNode.prototype, prototypeAccessors$2 );

    return VirtualTextNode;
  }(VirtualElement));

  VirtualElement.Component = VirtualComponent;
  VirtualElement.TextNode = VirtualTextNode;


  VirtualElement.createElement = function() {
    var content;
    var _first = arguments[0];
    var _second = arguments[1];
    var type;
    if (isString(_first)) {
      type = "element";
    } else if (isFunction$1(_first) && _first.prototype._isComponent) {
      type = "component";
    } else if (isNil(_first)) {
      throw new Error('$$(null): provided argument was null or undefined.')
    } else {
      throw new Error('Illegal usage of $$()')
    }
    
    var props = {};
    var classNames, ref;
    var eventHandlers = [];
    for(var key in _second) {
      if (!_second.hasOwnProperty(key)) { continue }
      var val = _second[key];
      switch(key) {
        case 'class':
          classNames = val;
          break
        case 'ref':
          ref = val;
          break
        default:
          props[key] = val;
      }
    }
    if (type === 'element') {
      content = new VirtualHTMLElement(_first);
      
      
      content.attr(props);
    } else {
      content = new VirtualComponent(_first, props);
    }
    
    
    content._owner = this.owner;
    if (classNames) {
      content.addClass(classNames);
    }
    if (ref) {
      content.ref(ref);
    }
    eventHandlers.forEach(function(h) {
      if (isFunction$1(h.handler)) {
        content.on(h.name, h.handler);
      } else if (isPlainObject(h.handler)) {
        var params = h.handler;
        content.on(h.name, params.handler, params.context, params);
      } else {
        throw new Error('Illegal arguments for $$(_,{ on'+h.name+'})')
      }
    });
    
    
    if (arguments.length > 2) {
      content.append(flattenOften(Array.prototype.slice.call(arguments, 2), 3));
    }
    return content
  };

  var RenderingEngine = function RenderingEngine(options) {
    if ( options === void 0 ) options = {};

    this.componentFactory = options.componentFactory;
    if (!this.componentFactory) { throw new Error("'componentFactory' is mandatory") }
    this.elementFactory = options.elementFactory || DefaultDOMElement.createDocument('html');
    if (!this.elementFactory) { throw new Error("'elementFactory' is mandatory") }
  };

  RenderingEngine.prototype._render = function _render$1 (comp, oldProps, oldState) {
    
    var vel = _createWrappingVirtualComponent(comp);
    var state = this._createState();
    if (oldProps) {
      state.setOldProps(vel, oldProps);
    }
    if (oldState) {
      state.setOldState(vel, oldState);
    }
    try {
      
      
      
      _capture(state, vel, 'forceCapture');
      

      
      
      _render(state, vel);
      

      _triggerUpdate(state, vel);

    } finally {
      state.dispose();
    }
    
  };


  RenderingEngine.prototype._renderChild = function _renderChild (comp, vel) {
    
    
    var state = this._createState();
    vel.parent = { _comp: comp };
    try {
      _capture(state, vel);
      _render(state, vel);
      return vel._comp
    } finally {
      state.dispose();
    }
  };

  RenderingEngine.prototype._createState = function _createState () {
    return new RenderingState(this.componentFactory, this.elementFactory)
  };


  function _create(state, vel) {
    var comp = vel._comp;
    console.assert(!comp, "Component instance should not exist when this method is used.");
    var parent = vel.parent._comp;
    
    if (!parent) {
      parent = _create(state, vel.parent);
    }
    if (vel._isVirtualComponent) {
      console.assert(parent, "A Component should have a parent.");
      comp = state.componentFactory.createComponent(vel.ComponentClass, parent, vel.props);
      
      vel.props = comp.props;
      comp.__htmlConfig__ = vel._copyHTMLConfig();
    } else if (vel._isVirtualHTMLElement) {
      comp = state.componentFactory.createElementComponent(parent, vel);
    } else if (vel._isVirtualTextNode) {
      comp = state.componentFactory.createTextNodeComponent(parent, vel);
    }
    if (vel._ref) {
      comp._ref = vel._ref;
    }
    if (vel._owner) {
      comp._owner = vel._owner._comp;
    }
    vel._comp = comp;
    return comp
  }

  function _capture(state, vel, forceCapture) {
    if (state.isCaptured(vel)) {
      return vel
    }
    
    var comp = vel._comp;
    if (!comp) {
      comp = _create(state, vel);
      state.setNew(vel);
    }
    if (vel._isVirtualComponent) {
      var needRerender;
      
      
      
      if (forceCapture) {
        needRerender = true;
      } else {
        
        needRerender = !comp.el || comp.shouldRerender(vel.props, comp.state);
        comp.__htmlConfig__ = vel._copyHTMLConfig();
        state.setOldProps(vel, comp.props);
        state.setOldState(vel, comp.state);
        
        comp._setProps(vel.props);
        if (!state.isNew(vel)) {
          state.setUpdated(vel);
        }
      }
      if (needRerender) {
        var context = new CaptureContext(vel);
        var content = comp.render(context.$$);
        if (!content || !content._isVirtualHTMLElement) {
          throw new Error("Component.render must return VirtualHTMLElement")
        }

        if (comp.__htmlConfig__) {
          content._mergeHTMLConfig(comp.__htmlConfig__);
        }
        content._comp = comp;
        vel._content = content;
        if (!state.isNew(vel) && comp.isMounted()) {
          state.setUpdated(vel);
        }
        
        _prepareVirtualComponent(state, comp, content);
        
        
        if (substanceGlobals.DEBUG_RENDERING) {
          
          
          
          
          var stack = content.children.slice(0);
          while (stack.length) {
            var child = stack.shift();
            if (state.isCaptured(child)) { continue }
            
            if (child._isVirtualComponent) { continue }
            if (!child._comp) {
              _create(state, child);
            }
            if (child._isVirtualHTMLElement && child.children.length > 0) {
              stack = stack.concat(child.children);
            }
            state.setCaptured(child);
          }
          state.setCaptured(content);
          
          
          var descendingContext = new DescendingContext(state, context);
          while (descendingContext.hasPendingCaptures()) {
            descendingContext.reset();
            comp.render(descendingContext.$$);
          }
        } else {
          
          
          _capture(state, vel._content);
        }
      } else {
        state.setSkipped(vel);
      }
    } else if (vel._isVirtualHTMLElement) {
      for (var i = 0; i < vel.children.length; i++) {
        _capture(state, vel.children[i]);
      }
    }
    state.setCaptured(vel);
    return vel
  }

  function _render(state, vel) {
    if (state.isSkipped(vel)) { return }
    

    
    
    
    
    
    

    var comp = vel._comp;
    console.assert(comp && comp._isComponent, "A captured VirtualElement must have a component instance attached.");

    
    if (vel._isVirtualComponent) {
      _render(state, vel._content);

      
      var context = vel._content._context;
      var refs = {};
      var foreignRefs = {};
      forEach(context.refs, function (vel, ref) {
        refs[ref] = vel._comp;
      });
      forEach(context.foreignRefs, function (vel, ref) {
        foreignRefs[ref] = vel._comp;
      });
      comp.refs = refs;
      comp.__foreignRefs__ = foreignRefs;
      return
    }

    
    if (!comp.el) {
      comp.el = _createElement(state, vel);
      comp.el._comp = comp;
    }
    _updateElement(comp, vel);

    
    if (vel._isVirtualHTMLElement && !vel.hasInnerHTML()) {
      var newChildren = vel.children;
      var oldComp, virtualComp, newComp;
      var pos1 = 0; var pos2 = 0;

      
      
      
      var oldChildren = [];
      comp.el.getChildNodes().forEach(function(node) {
        var childComp = node._comp;

        
        
        
        

        
        if (!childComp || state.isRelocated(childComp)) {
          comp.el.removeChild(node);
        } else {
          oldChildren.push(childComp);
        }
      });

      while(pos1 < oldChildren.length || pos2 < newChildren.length) {
        
        
        
        
        do {
          oldComp = oldChildren[pos1++];
        } while (oldComp && (state.isDetached(oldComp)))

        virtualComp = newChildren[pos2++];
        
        if (oldComp && !virtualComp) {
          while (oldComp) {
            _removeChild(state, comp, oldComp);
            oldComp = oldChildren[pos1++];
          }
          break
        }

        
        if (oldComp && oldComp.el.isTextNode() &&
            virtualComp && virtualComp._isVirtualTextNode &&
            oldComp.el.textContent === virtualComp.text ) {
          continue
        }

        
        if (!state.isRendered(virtualComp)) {
          _render(state, virtualComp);
        }

        newComp = virtualComp._comp;

        
        
        if (state.isRelocated(newComp)) {
          newComp._setParent(comp);
        }

        console.assert(newComp, 'Component instance should now be available.');

        
        if (virtualComp && !oldComp) {
          _appendChild(state, comp, newComp);
          continue
        }
        
        else if (state.isMapped(virtualComp)) {
          
          if (newComp === oldComp) ; else if (state.isMapped(oldComp)) {
            
            state.setDetached(oldComp);
            _removeChild(state, comp, oldComp);
            pos2--;
          }
          
          else {
            _removeChild(state, comp, oldComp);
            pos2--;
          }
        }
        else if (state.isMapped(oldComp)) {
          _insertChildBefore(state, comp, newComp, oldComp);
          pos1--;
        }
        else {
          
          
          
          
          _replaceChild(state, comp, oldComp, newComp);
        }
      }
    }

    state.setRendered(vel);
  }

  function _triggerUpdate(state, vel) {
    if (vel._isVirtualComponent) {
      if (!state.isSkipped(vel)) {
        vel._content.children.forEach(_triggerUpdate.bind(null, state));
      }
      if (state.isUpdated(vel)) {
        vel._comp.didUpdate(state.getOldProps(vel), state.getOldState(vel));
      }
    } else if (vel._isVirtualHTMLElement) {
      vel.children.forEach(_triggerUpdate.bind(null, state));
    }
  }

  function _appendChild(state, parent, child) {
    parent.el.appendChild(child.el);
    _triggerDidMount(state, parent, child);
  }

  function _replaceChild(state, parent, oldChild, newChild) {
    parent.el.replaceChild(oldChild.el, newChild.el);
    if (!state.isDetached(oldChild)) {
      oldChild.triggerDispose();
    }
    _triggerDidMount(state, parent, newChild);
  }

  function _insertChildBefore(state, parent, child, before) {
    parent.el.insertBefore(child.el, before.el);
    _triggerDidMount(state, parent, child);
  }

  function _removeChild(state, parent, child) {
    parent.el.removeChild(child.el);
    if (!state.isDetached(child)) {
      child.triggerDispose();
    }
  }

  function _triggerDidMount(state, parent, child) {
    if (!state.isDetached(child) &&
        parent.isMounted() && !child.isMounted()) {
      child.triggerDidMount(true);
    }
  }


  function _prepareVirtualComponent(state, comp, vc) {
    var newRefs = {};
    var foreignRefs = {};
    
    
    
    
    if (vc._context) {
      newRefs = vc._context.refs;
      foreignRefs = vc._context.foreignRefs;
    }
    var oldRefs = comp.refs;
    var oldForeignRefs = comp.__foreignRefs__;
    
    forEach(newRefs, function(vc, ref) {
      var comp = oldRefs[ref];
      if (comp) { _mapComponents(state, comp, vc); }
    });
    forEach(foreignRefs, function(vc, ref) {
      var comp = oldForeignRefs[ref];
      if (comp) { _mapComponents(state, comp, vc); }
    });
  }



  function _mapComponents(state, comp, vc) {
    if (!comp && !vc) { return true }
    if (!comp || !vc) { return false }
    
    
    
    
    
    if (state.isMapped(vc) || state.isMapped(comp)) {
      return vc._comp === comp
    }
    if (vc._comp) {
      if (vc._comp === comp) {
        state.setMapped(vc);
        state.setMapped(comp);
        return true
      } else {
        return false
      }
    }
    if (!_isOfSameType(comp, vc)) {
      return false
    }

    vc._comp = comp;
    state.setMapped(vc);
    state.setMapped(comp);

    var canMapParent;
    var parent = comp.getParent();
    if (vc.parent) {
      canMapParent = _mapComponents(state, parent, vc.parent);
    }
    
    
    
    
    else if (vc._preliminaryParent) {
      while (parent && parent._isElementComponent) {
        parent = parent.getParent();
      }
      canMapParent = _mapComponents(state, parent, vc._preliminaryParent);
    }
    if (!canMapParent) {
      state.setRelocated(vc);
      state.setRelocated(comp);
    }
    return canMapParent
  }

  function _isOfSameType(comp, vc) {
    return (
      (comp._isElementComponent && vc._isVirtualHTMLElement) ||
      (comp._isComponent && vc._isVirtualComponent && comp.constructor === vc.ComponentClass) ||
      (comp._isTextNodeComponent && vc._isVirtualTextNode)
    )
  }

  function _createElement(state, vel) {
    var el;
    if (vel._isVirtualTextNode) {
      el = state.elementFactory.createTextNode(vel.text);
    } else {
      el = state.elementFactory.createElement(vel.tagName);
    }
    return el
  }

  function _updateElement(comp, vel) {
    if (comp._isTextNodeComponent) {
      comp.setTextContent(vel.text);
      return
    }
    var el = comp.el;
    console.assert(el, "Component's element should exist at this point.");
    var tagName = el.getTagName();
    if (vel.tagName.toLowerCase() !== tagName) {
      el.setTagName(vel.tagName);
    }
    _updateHash({
      oldHash: el.getAttributes(),
      newHash: vel.getAttributes(),
      update: function(key, val) {
        el.setAttribute(key, val);
      },
      remove: function(key) {
        el.removeAttribute(key);
      }
    });
    _updateHash({
      oldHash: el.htmlProps,
      newHash: vel.htmlProps,
      update: function(key, val) {
        el.setProperty(key, val);
      },
      remove: function(key) {
        el.removeProperty(key);
      }
    });
    _updateListeners({
      el: el,
      oldListeners: el.getEventListeners(),
      newListeners: vel.getEventListeners()
    });

    
    if (vel.hasInnerHTML()) {
      if (!el._hasInnerHTML) {
        el.empty();
        el.setInnerHTML(vel.getInnerHTML());
      } else {
        var oldInnerHTML = el.getInnerHTML();
        var newInnerHTML = vel.getInnerHTML();
        if (oldInnerHTML !== newInnerHTML) {
          el.setInnerHTML(newInnerHTML);
        }
      }
      el._hasInnerHTML = true;
    }
  }

  function _updateHash(args) {
    var newHash = args.newHash;
    var oldHash = args.oldHash || {};
    var update = args.update;
    var remove = args.remove;
    var updatedKeys = {};
    for (var key in newHash) {
      if (newHash.hasOwnProperty(key)) {
        var oldVal = oldHash[key];
        var newVal = newHash[key];
        updatedKeys[key] = true;
        if (oldVal !== newVal) {
          update(key, newVal);
        }
      }
    }
    
    
    if (isFunction$1(oldHash.keys) && oldHash.size > 0) {
      var keys = Array.from(oldHash.keys());
      keys.forEach(function (key) {
        if (!updatedKeys[key]) {
          remove(key);
        }
      });
    } else {
      for (var key$1 in oldHash) {
        if (oldHash.hasOwnProperty(key$1) && !updatedKeys[key$1]) {
          remove(key$1);
        }
      }
    }
  }

  function _updateListeners(args) {
    var el = args.el;
    
    
    
    var newListeners = args.newListeners || [];
    el.removeAllEventListeners();
    for (var i=0; i<newListeners.length;i++) {
      el.addEventListener(newListeners[i]);
    }
  }




  var DescendingContext = function DescendingContext(state, captureContext) {
    this.state = state;
    this.owner = captureContext.owner;
    this.refs = {};
    this.foreignRefs = {};
    this.elements = captureContext.elements;
    this.pos = 0;
    this.updates = captureContext.components.length;
    this.remaining = this.updates;

    this.$$ = this._createComponent.bind(this);
  };

  DescendingContext.prototype._createComponent = function _createComponent () {
    var state = this.state;
    var vel = this.elements[this.pos++];
    
    
    
    if (!state.isCaptured(vel) && vel._isVirtualComponent &&
         vel.parent && state.isCaptured(vel.parent)) {
      _capture(state, vel);
      this.updates++;
      this.remaining--;
    }
    
    
    
    
    vel = VirtualElement.createElement.apply(this, arguments);
    
    vel._context = this;
    vel._owner = this.owner;
    
    
    vel._attach = function() {};
    vel._detach = function() {};
    return vel
  };

  DescendingContext.prototype.hasPendingCaptures = function hasPendingCaptures () {
    return this.updates > 0 && this.remaining > 0
  };

  DescendingContext.prototype.reset = function reset () {
    this.pos = 0;
    this.updates = 0;
    this.refs = {};
  };

  DescendingContext.prototype._ancestorsReady = function _ancestorsReady (vel) {
      var this$1 = this;

    while (vel) {
      if (this$1.state.isCaptured(vel) ||
          
          vel === this$1.owner || vel === this$1.owner._content) {
        return true
      }
      vel = vel.parent;
    }
    return false
  };

  RenderingEngine._internal = {
    _capture: _capture,
    _wrap: _createWrappingVirtualComponent,
  };

  var CaptureContext = function CaptureContext(owner) {
    this.owner = owner;
    this.refs = {};
    this.foreignRefs = {};
    this.elements = [];
    this.components = [];
    this.$$ = this._createComponent.bind(this);
    this.$$.capturing = true;
  };

  CaptureContext.prototype._createComponent = function _createComponent () {
    var vel = VirtualElement.createElement.apply(this, arguments);
    vel._context = this;
    vel._owner = this.owner;
    if (vel._isVirtualComponent) {
      
      this.components.push(vel);
    }
    this.elements.push(vel);
    return vel
  };


  function _createWrappingVirtualComponent(comp) {
    var vel = new VirtualElement.Component(comp.constructor);
    vel._comp = comp;
    if (comp.__htmlConfig__) {
      vel._mergeHTMLConfig(comp.__htmlConfig__);
    }
    return vel
  }

  RenderingEngine.createContext = function(comp) {
    var vel = _createWrappingVirtualComponent(comp);
    return new CaptureContext(vel)
  };


  var RenderingState = function RenderingState(componentFactory, elementFactory) {
    this.componentFactory = componentFactory;
    this.elementFactory = elementFactory;
    this.poluted = [];
    this.id = "__"+uuid();
  };

  RenderingState.prototype.dispose = function dispose () {
    var id = this.id;
    this.poluted.forEach(function(obj) {
      delete obj[id];
    });
  };

  RenderingState.prototype.set = function set (obj, key, val) {
    var info = obj[this.id];
    if (!info) {
      info = {};
      obj[this.id] = info;
      this.poluted.push(obj);
    }
    info[key] = val;
  };

  RenderingState.prototype.get = function get (obj, key) {
    var info = obj[this.id];
    if (info) {
      return info[key]
    }
  };

  RenderingState.prototype.setMapped = function setMapped (c) {
    this.set(c, 'mapped', true);
  };


  RenderingState.prototype.isMapped = function isMapped (c) {
    return Boolean(this.get(c, 'mapped'))
  };



  RenderingState.prototype.setRelocated = function setRelocated (c) {
    this.set(c, 'relocated', true);
  };

  RenderingState.prototype.isRelocated = function isRelocated (c) {
    return Boolean(this.get(c, 'relocated'))
  };

  RenderingState.prototype.setDetached = function setDetached (c) {
    this.set(c, 'detached', true);
  };

  RenderingState.prototype.isDetached = function isDetached (c) {
    return Boolean(this.get(c, 'detached'))
  };

  RenderingState.prototype.setCaptured = function setCaptured (vc) {
    this.set(vc, 'captured', true);
  };

  RenderingState.prototype.isCaptured = function isCaptured (vc) {
    return Boolean(this.get(vc, 'captured'))
  };

  RenderingState.prototype.setNew = function setNew (vc) {
    this.set(vc, 'created', true);
  };

  RenderingState.prototype.isNew = function isNew (vc) {
    return Boolean(this.get(vc, 'created'))
  };

  RenderingState.prototype.setUpdated = function setUpdated (vc) {
    this.set(vc, 'updated', true);
  };

  RenderingState.prototype.isUpdated = function isUpdated (vc) {
    return Boolean(this.get(vc, 'updated'))
  };

  RenderingState.prototype.setSkipped = function setSkipped (vc) {
    this.set(vc, 'skipped', true);
  };

  RenderingState.prototype.isSkipped = function isSkipped (vc) {
    return Boolean(this.get(vc, 'skipped'))
  };

  RenderingState.prototype.setRendered = function setRendered (vc) {
    this.set(vc, 'rendered', true);
  };

  RenderingState.prototype.isRendered = function isRendered (vc) {
    return Boolean(this.get(vc, 'rendered'))
  };

  RenderingState.prototype.setOldProps = function setOldProps (vc, oldProps) {
    this.set(vc, 'oldProps', oldProps);
  };

  RenderingState.prototype.getOldProps = function getOldProps (vc) {
    return this.get(vc, 'oldProps')
  };

  RenderingState.prototype.setOldState = function setOldState (vc, oldState) {
    this.set(vc, 'oldState', oldState);
  };

  RenderingState.prototype.getOldState = function getOldState (vc) {
    return this.get(vc, 'oldState')
  };

  var COMPONENT_FACTORY = {
    createComponent: function createComponent(ComponentClass, parent, props) {
      return new ComponentClass(parent, props)
    },
    createElementComponent: function createElementComponent(parent, virtualElement) {
      return new ElementComponent(parent, virtualElement)
    },
    createTextNodeComponent: function createTextNodeComponent(parent, virtualElement) {
      return new TextNodeComponent(parent, virtualElement)
    }
  };

  var Component = (function (EventEmitter$$1) {
    function Component(parent, props, options) {
      if ( props === void 0 ) props = {};
      if ( options === void 0 ) options = {};

      EventEmitter$$1.call(this);

      
      
      
      
      

      this.parent = (parent && parent._isComponent) ? parent : null;

      
      this.el = options.el;

      
      
      var context = options.context ? options.context : this._getContext() || {};
      this.context = context;
      Object.freeze(this.context);

      
      
      
      this.renderingEngine = (parent && parent.renderingEngine) || context.renderingEngine || options.renderingEngine || new RenderingEngine({
        componentFactory: COMPONENT_FACTORY
      });

      
      
      if (this._SKIP_COMPONENT_INIT) { return }

      this.__id__ = uuid();

      
      this.refs = {};
      
      
      
      this.__foreignRefs__ = {};

      
      this._actionHandlers = {};

      
      this.props = props;
      Object.freeze(this.props);

      
      this.state = this.getInitialState() || {};
      Object.freeze(this.state);
    }

    if ( EventEmitter$$1 ) Component.__proto__ = EventEmitter$$1;
    Component.prototype = Object.create( EventEmitter$$1 && EventEmitter$$1.prototype );
    Component.prototype.constructor = Component;

    var prototypeAccessors = { textContent: { configurable: true },innerHTML: { configurable: true },outerHTML: { configurable: true },childNodes: { configurable: true } };

    Component.prototype.getId = function getId () {
      return this.__id__
    };

    Component.prototype.setId = function setId () {
      throw new Error("'id' is readonly")
    };

    
    Component.prototype.getChildContext = function getChildContext () {
      return this.childContext || {}
    };

    
    Component.prototype.getInitialState = function getInitialState () {
      return {}
    };

    
    Component.prototype.getParent = function getParent () {
      return this.parent
    };

    
    Component.prototype.getRoot = function getRoot () {
      var comp = this;
      var parent = comp;
      while (parent) {
        comp = parent;
        parent = comp.getParent();
      }
      return comp
    };

    Component.prototype.getElement = function getElement () {
      return this.el
    };

    Component.prototype.getNativeElement = function getNativeElement () {
      return this.el.getNativeElement()
    };

    
    Component.prototype.getLabel = function getLabel (name) {
      var args = [], len = arguments.length - 1;
      while ( len-- > 0 ) args[ len ] = arguments[ len + 1 ];

      var labelProvider = this.getLabelProvider();
      if (!labelProvider) { throw new Error('Missing labelProvider.') }
      return labelProvider.getLabel.apply(labelProvider, [ name ].concat( args ))
    };

    Component.prototype.getLabelProvider = function getLabelProvider () {
      return this.context.labelProvider
    };

    
    Component.prototype.getComponent = function getComponent (componentName, maybe) {
      var componentRegistry = this.getComponentRegistry();
      if (!componentRegistry) { throw new Error('Missing componentRegistry.') }
      var ComponentClass = componentRegistry.get(componentName);
      if (!maybe && !ComponentClass) {
        throw new Error('No Component registered with name ' + componentName)
      }
      return ComponentClass
    };

    Component.prototype.getComponentRegistry = function getComponentRegistry () {
      return this.props.componentRegistry || this.context.componentRegistry
    };

    Component.prototype.getFlow = function getFlow () {
      return this.context.flow
    };

    
    Component.prototype.render = function render ($$) {
      
      return $$('div')
    };

    
    Component.prototype.mount = function mount (el) {
      if (!el) {
        throw new Error('Element is required.')
      }
      el = DefaultDOMElement.wrap(el);
      
      this.el = null;
      this.renderingEngine = Component.createRenderingEngine(el.getOwnerDocument());
      this._render();
      el.appendChild(this.el);
      if (el.isInDocument()) {
        this.triggerDidMount(true);
      }
      return this
    };

    
    Component.prototype.shouldRerender = function shouldRerender (newProps, newState) { 
      return true
    };

    
    Component.prototype.rerender = function rerender () {
      this._rerender(this.props, this.state);
    };

    Component.prototype._rerender = function _rerender (oldProps, oldState) {
      this._render(oldProps, oldState);
      
      if (!this.isMounted()) {
        this.didUpdate(oldProps, oldState);
      }
    };

    Component.prototype._render = function _render (oldProps, oldState) {
      if (this.__isRendering__) {
        throw new Error('Component is rendering already.')
      }
      this.__isRendering__ = true;
      try {
        this.renderingEngine._render(this, oldProps, oldState);
      } finally {
        delete this.__isRendering__;
      }
    };

    
    Component.prototype.triggerDidMount = function triggerDidMount () {
      
      
      

      
      
      if (!this.__isMounted__) {
        this.__isMounted__ = true;
        this.didMount();
      }
      
      this.getChildren().forEach(function(child) {
        
        
        child.triggerDidMount(true);
      });
    };

    
    Component.prototype.didMount = function didMount () {};

    
    Component.prototype.didUpdate = function didUpdate () {};

    
    Component.prototype.isMounted = function isMounted () {
      return this.__isMounted__
    };

    
    Component.prototype.triggerDispose = function triggerDispose () {
      this.getChildren().forEach(function(child) {
        child.triggerDispose();
      });
      this.dispose();
      this.__isMounted__ = false;
    };

    
    Component.prototype.dispose = function dispose () {};

    
    Component.prototype._setParent = function _setParent (newParent) {
      this.parent = newParent;
      this.context = this._getContext() || {};
      Object.freeze(this.context);
    };

    
    Component.prototype.send = function send (action) {
      var arguments$1 = arguments;

      
      var comp = this.getParent();
      while(comp) {
        if (comp._actionHandlers && comp._actionHandlers[action]) {
          comp._actionHandlers[action].apply(comp, Array.prototype.slice.call(arguments$1, 1));
          return true
        }
        comp = comp.getParent();
      }
      console.warn('Action', action, 'was not handled.');
      return false
    };

    
    Component.prototype.handleActions = function handleActions (actionHandlers) {
      forEach(actionHandlers, function(method, actionName) {
        this.handleAction(actionName, method);
      }.bind(this));
      return this
    };

    
    Component.prototype.handleAction = function handleAction (name, handler) {
      if (!name || !handler || !isFunction$1(handler)) {
        throw new Error('Illegal arguments.')
      }
      handler = handler.bind(this);
      this._actionHandlers[name] = handler;
    };

    
    Component.prototype.getState = function getState () {
      return this.state
    };

    
    Component.prototype.setState = function setState (newState) {
      var oldProps = this.props;
      var oldState = this.state;
      
      
      var needRerender = !this.__isSettingProps__ &&
        this.shouldRerender(this.getProps(), newState);
      
      this.willUpdateState(newState);
      this.state = newState || {};
      Object.freeze(this.state);
      if (needRerender) {
        this._rerender(oldProps, oldState);
      } else if (!this.__isSettingProps__) {
        this.didUpdate(oldProps, oldState);
      }
    };

    
    Component.prototype.extendState = function extendState (newState) {
      newState = extend({}, this.state, newState);
      this.setState(newState);
    };

    
    Component.prototype.willUpdateState = function willUpdateState (newState) { 
    };

    
    Component.prototype.getProps = function getProps () {
      return this.props
    };

    
    Component.prototype.setProps = function setProps (newProps) {
      var oldProps = this.props;
      var oldState = this.state;
      var needRerender = this.shouldRerender(newProps, this.state);
      this._setProps(newProps);
      if (needRerender) {
        this._rerender(oldProps, oldState);
      } else {
        this.didUpdate(oldProps, oldState);
      }
    };

    Component.prototype._setProps = function _setProps (newProps) {
      newProps = newProps || {};
      
      this.__isSettingProps__ = true;
      try {
        this.willReceiveProps(newProps);
        this.props = newProps || {};
        Object.freeze(newProps);
      } finally {
        delete this.__isSettingProps__;
      }
    };

    
    Component.prototype.extendProps = function extendProps (updatedProps) {
      var newProps = extend({}, this.props, updatedProps);
      this.setProps(newProps);
    };

    
    Component.prototype.willReceiveProps = function willReceiveProps (newProps) { 
    };

    Component.prototype.getTextContent = function getTextContent () {
      if (this.el) {
        return this.el.textContent
      }
    };

    prototypeAccessors.textContent.get = function () {
      return this.getTextContent()
    };

    Component.prototype.getInnerHTML = function getInnerHTML () {
      if (this.el) {
        return this.el.getInnerHTML()
      }
    };

    prototypeAccessors.innerHTML.get = function () {
      return this.getInnerHTML()
    };

    Component.prototype.getOuterHTML = function getOuterHTML () {
      if (this.el) {
        return this.el.getOuterHTML()
      }
    };

    prototypeAccessors.outerHTML.get = function () {
      return this.getOuterHTML()
    };

    Component.prototype.getAttribute = function getAttribute (name) {
      if (this.el) {
        return this.el.getAttribute(name)
      }
    };

    Component.prototype.setAttribute = function setAttribute (name, val) {
      if (this.el) {
        this.el.setAttribute(name, val);
      }
      return this
    };

    Component.prototype.getProperty = function getProperty (name) {
      if (this.el) {
        return this.el.getProperty(name)
      }
    };

    Component.prototype.setProperty = function setProperty (name, val) {
      if (this.el) {
        this.el.setProperty(name, val);
      }
      return this
    };

    Component.prototype.hasClass = function hasClass (name) {
      if (this.el) {
        return this.el.hasClass(name)
      }
    };

    Component.prototype.addClass = function addClass (name) {
      if (this.el) {
        this.el.addClass(name);
      }
      return this
    };

    Component.prototype.removeClass = function removeClass (name) {
      if (this.el) {
        this.el.removeClass(name);
      }
      return this
    };

    Component.prototype.getStyle = function getStyle (name) {
      if (this.el) {
        return this.el.getStyle(name)
      }
    };

    Component.prototype.setStyle = function setStyle (name, val) {
      if (this.el) {
        return this.el.setStyle(name, val)
      }
      return this
    };

    Component.prototype.getValue = function getValue () {
      if (this.el) {
        return this.el.getValue()
      }
    };

    Component.prototype.setValue = function setValue (val) {
      if (this.el) {
        this.el.setValue(val);
      }
      return this
    };

    Component.prototype.getChildCount = function getChildCount () {
      if (!this.el) { return 0 }
      return this.el.getChildCount()
    };

    prototypeAccessors.childNodes.get = function () {
      return this.getChildNodes()
    };

    Component.prototype.getChildNodes = function getChildNodes () {
      if (!this.el) { return [] }
      var childNodes = this.el.getChildNodes();
      childNodes = childNodes.map(_unwrapComp).filter(Boolean);
      return childNodes
    };

    Component.prototype.getChildren = function getChildren () {
      if (!this.el) { return [] }
      var children = this.el.getChildren();
      children = children.map(_unwrapComp).filter(Boolean);
      return children
    };

    Component.prototype.getChildAt = function getChildAt (pos) {
      var node = this.el.getChildAt(pos);
      if (node) {
        return _unwrapCompStrict(node)
      }
    };

    Component.prototype.find = function find (cssSelector) {
      var el = this.el.find(cssSelector);
      return _unwrapComp(el)
    };

    Component.prototype.findAll = function findAll (cssSelector) {
      var els = this.el.findAll(cssSelector);
      return els.map(_unwrapComp).filter(Boolean)
    };

    Component.prototype.appendChild = function appendChild (child) {
      this.insertAt(this.getChildCount(), child);
    };

    Component.prototype.insertAt = function insertAt (pos, childEl) {
      if (isString(childEl)) {
        childEl = new VirtualElement.TextNode(childEl);
      }
      if (!childEl._isVirtualElement) {
        throw new Error('Invalid argument: "child" must be a VirtualElement.')
      }
      var child = this.renderingEngine._renderChild(this, childEl);
      this.el.insertAt(pos, child.el);
      _mountChild(this, child);
    };

    Component.prototype.removeAt = function removeAt (pos) {
      var childEl = this.el.getChildAt(pos);
      if (childEl) {
        var child = _unwrapCompStrict(childEl);
        _disposeChild(child);
        this.el.removeAt(pos);
      }
    };

    Component.prototype.removeChild = function removeChild (child) {
      if (!child || !child._isComponent) {
        throw new Error('removeChild(): Illegal arguments. Expecting a Component instance.')
      }
      
      _disposeChild(child);
      this.el.removeChild(child.el);
    };

    Component.prototype.replaceChild = function replaceChild (oldChild, newChild) {
      if (!newChild || !oldChild ||
          !newChild._isComponent || !oldChild._isComponent) {
        throw new Error('replaceChild(): Illegal arguments. Expecting BrowserDOMElement instances.')
      }
      
      _disposeChild(oldChild);
      this.el.replaceChild(newChild.el, oldChild.el);
      if (this.isMounted()) {
        newChild.triggerDidMount(true);
      }
    };

    
    
    
    
    
    
    Component.prototype.empty = function empty () {
      this._clear();
      return this
    };

    Component.prototype._clear = function _clear () {
      var el = this.el;
      if (el) {
        this.getChildNodes().forEach(function(child) {
          _disposeChild(child);
        });
        el.empty();
      }
      this.refs = {};
      this.__foreignRefs__ = {};
    };

    Component.prototype.remove = function remove () {
      _disposeChild(this);
      this.el.remove();
    };

    Component.prototype.addEventListener = function addEventListener () {
      throw new Error("Not supported.")
    };

    Component.prototype.removeEventListener = function removeEventListener () {
      throw new Error("Not supported.")
    };

    Component.prototype.insertBefore = function insertBefore () {
      throw new Error("Not supported.")
    };

    Component.prototype.click = function click () {
      if (this.el) {
        this.el.click();
      }
    };

    Component.prototype.getComponentPath = function getComponentPath () {
      var path = [];
      var comp = this;
      while (comp) {
        path.unshift(comp);
        comp = comp.getParent();
      }
      return path
    };

    Component.prototype._getContext = function _getContext () {
      var context = {};
      var parent = this.getParent();
      if (parent) {
        context = extend(context, parent.context);
        if (parent.getChildContext) {
          return extend(context, parent.getChildContext())
        }
      }
      return context
    };

    Object.defineProperties( Component.prototype, prototypeAccessors );

    return Component;
  }(EventEmitter));

  Component.prototype._isComponent = true;

  Component.prototype.attr = DOMElement.prototype.attr;

  Component.prototype.htmlProp = DOMElement.prototype.htmlProp;

  Component.prototype.val = DOMElement.prototype.val;

  Component.prototype.css = DOMElement.prototype.css;

  Component.prototype.text = DOMElement.prototype.text;

  Component.prototype.append = DOMElement.prototype.append;

  Component.unwrap = _unwrapComp;

  Component.render = function(props) {
    props = props || {};
    var ComponentClass = this;
    var comp = new ComponentClass(null, props);
    comp._render();
    return comp
  };

  Component.mount = function(props, el) {
    if (arguments.length === 1) {
      el = props;
      props = {};
    }
    if (!el) { throw new Error("'el' is required.") }
    if (isString(el)) {
      var selector = el;
      if (platform.inBrowser) {
        el = window.document.querySelector(selector);
      } else {
        throw new Error("This selector is not supported on server side.")
      }
    }
    el = new DefaultDOMElement.wrap(el);
    var ComponentClass = this;
    var comp = new ComponentClass(null, props);
    comp.mount(el);
    return comp
  };

  Component.getComponentForDOMElement = function(el) {
    return _unwrapComp(el)
  };

  Component.unwrapDOMElement = function(el) {
    console.warn('DEPRECATED: Use Component.getComponentForDOMElement');
    return Component.getComponentForDOMElement(el)
  };

  Component.getComponentFromNativeElement = function(nativeEl) {
    
    
    
    return _unwrapComp(DefaultDOMElement.wrap(nativeEl))
  };

  Component.createRenderingEngine = function(elementFactory) {
    return new RenderingEngine({
      componentFactory: COMPONENT_FACTORY,
      elementFactory: elementFactory
    })
  };


  function _disposeChild(child) {
    child.triggerDispose();
    if (child._owner && child._ref) {
      console.assert(child._owner.refs[child._ref] === child, "Owner's ref should point to this child instance.");
      delete child._owner.refs[child._ref];
    }
  }


  function _mountChild(parent, child) {
    if (parent.isMounted()) {
      child.triggerDidMount(true);
    }
    if (child._owner && child._ref) {
      child._owner.refs[child._ref] = child;
    }
  }


  function _unwrapComp(el) {
    if (el) {
      if (!el._isDOMElement) { el = DefaultDOMElement.unwrap(el); }
      if (el) { return el._comp }
    }
  }

  function _unwrapCompStrict(el) {
    var comp = _unwrapComp(el);
    if (!comp) { throw new Error("Expecting a back-link to the component instance.") }
    return comp
  }


  var ElementComponent = (function (Component) {
    function ElementComponent(parent) {
      Component.call(this, parent);
    }

    if ( Component ) ElementComponent.__proto__ = Component;
    ElementComponent.prototype = Object.create( Component && Component.prototype );
    ElementComponent.prototype.constructor = ElementComponent;

    return ElementComponent;
  }(Component));

  ElementComponent.prototype._isElementComponent = true;
  ElementComponent.prototype._SKIP_COMPONENT_INIT = true;

  var TextNodeComponent = (function (Component) {
    function TextNodeComponent(parent) {
      Component.call(this, parent);
    }

    if ( Component ) TextNodeComponent.__proto__ = Component;
    TextNodeComponent.prototype = Object.create( Component && Component.prototype );
    TextNodeComponent.prototype.constructor = TextNodeComponent;

    TextNodeComponent.prototype.setTextContent = function setTextContent (text) {
      if (!this.el) {
        throw new Error('Component must be rendered first.')
      }
      if (this.el.textContent !== text) {
        this.el.textContent = text;
      }
    };

    TextNodeComponent.prototype.getChildNodes = function getChildNodes () {
      return []
    };

    TextNodeComponent.prototype.getChildren = function getChildren () {
      return []
    };

    return TextNodeComponent;
  }(Component));

  TextNodeComponent.prototype._isTextNodeComponent = true;
  TextNodeComponent.prototype._SKIP_COMPONENT_INIT = true;

  Component.Element = ElementComponent;
  Component.TextNode = TextNodeComponent;

  var FontAwesomeIcon = (function (Component$$1) {
    function FontAwesomeIcon() {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      Component$$1.apply(this, args);
    }

    if ( Component$$1 ) FontAwesomeIcon.__proto__ = Component$$1;
    FontAwesomeIcon.prototype = Object.create( Component$$1 && Component$$1.prototype );
    FontAwesomeIcon.prototype.constructor = FontAwesomeIcon;

    FontAwesomeIcon.prototype.render = function render ($$) {
      if (this.props.stack) {
        return $$('span').addClass('fa-stack')
          .append(this.props.stack.map(function (faClass) {
            return $$('i').addClass('fa ' +faClass+' fa-stack')
          }))
      } else {
        return $$('i').addClass('fa ' + this.props.icon)
      }
    };

    return FontAwesomeIcon;
  }(Component));

  var FontAwesomeIconProvider = function FontAwesomeIconProvider(icons) {
    var this$1 = this;

    this.faMap = {};
    this.textMap = {};
    forEach(icons, function (config, name) {
      var faConfig = config['fontawesome'];
      if (faConfig) {
        if (isString(faConfig)) {
          faConfig = { icon: faConfig };
        }
        this$1.addFAIcon(name, faConfig);
      }
      var text = config['text'];
      if (text) {
        this$1.addTextIcon(name, text);
      }
    });
  };

  FontAwesomeIconProvider.prototype.renderIcon = function renderIcon ($$, name) {
    var faProps = this.faMap[name];
    var text = this.textMap[name];
    if (faProps) {
      return $$(FontAwesomeIcon, faProps)
    } else if (text) {
      return text
    }
  };

  FontAwesomeIconProvider.prototype.addFAIcon = function addFAIcon (name, faClass) {
    this.faMap[name] = faClass;
  };

  FontAwesomeIconProvider.prototype.addTextIcon = function addTextIcon (name, text) {
    this.textMap[name] = text;
  };

  var CommandManager = function CommandManager(context, commands) {
    var editorSession = context.editorSession;
    if (!editorSession) {
      throw new Error('EditorSession required.')
    }
    this.editorSession = context.editorSession;
    
    this.commands = commands;

    
    
    this.context = Object.assign({}, context, {
      
      doc: this.editorSession.getDocument()
    });

    
    this._initialize();

    
    this.editorSession.on('pre-render', this._onSessionUpdate, this);

    
    
    this._updateCommandStates(this.editorSession);
  };

  CommandManager.prototype.dispose = function dispose () {
    this.editorSession.off(this);
  };


  CommandManager.prototype.executeCommand = function executeCommand (commandName, userParams, cb) {
      var this$1 = this;

    var cmd = this._getCommand(commandName);
    if (!cmd) {
      console.warn('command', commandName, 'not registered');
      return
    }
    var commandStates = this.editorSession.getCommandStates();
    var commandState = commandStates[commandName];
    var params = Object.assign(this._getCommandParams(), userParams, {
      commandState: commandState
    });

    if (cmd.isAsync) {
      
      this.editorSession.lock();
      cmd.execute(params, this._getCommandContext(), function (err, info) {
        if (err) {
          if (cb) {
            cb(err);
          } else {
            console.error(err);
          }
        } else {
          if (cb) { cb(null, info); }
        }
        this$1.editorSession.unlock();
      });
    } else {
      var info = cmd.execute(params, this._getCommandContext());
      return info
    }
  };

  CommandManager.prototype._initialize = function _initialize () {
      var this$1 = this;

    this.commandRegistry = new Registry();
    forEach(this.commands, function (command) {
      this$1.commandRegistry.add(command.name, command);
    });
  };

  CommandManager.prototype._getCommand = function _getCommand (commandName) {
    return this.commandRegistry.get(commandName)
  };


  CommandManager.prototype._updateCommandStates = function _updateCommandStates (editorSession) {
    var commandContext = this._getCommandContext();
    var params = this._getCommandParams();
    var surface = params.surface;
    var commandRegistry = this.commandRegistry;

    
    
    var commandStates = {};
    var commandNames = commandRegistry.names.slice();
    
    commandNames.forEach(function (name) {
      commandStates[name] = { disabled: true };
    });
    
    if (surface) {
      var included = surface.props.commands;
      var excluded = surface.props.excludedCommands;
      if (included) {
        commandNames = included.filter(function (name) {
          return commandRegistry.contains(name)
        });
      } else if (excluded) {
        commandNames = without.apply(void 0, [ commandNames ].concat( excluded ));
      }
    }
    var commands = commandNames.map(function (name) { return commandRegistry.get(name); });
    commands.forEach(function (cmd) {
      if (cmd) {
        commandStates[cmd.getName()] = cmd.getCommandState(params, commandContext);
      }
    });
    
    
    
    
    this.commandStates = commandStates;
    editorSession.setCommandStates(commandStates);
  };

  CommandManager.prototype._onSessionUpdate = function _onSessionUpdate (editorSession) {
    
    
    if (
      editorSession.hasChanged('change') ||
      editorSession.hasChanged('selection') ||
      editorSession.hasChanged('commandStates')
    ) {
      this._updateCommandStates(editorSession);
    }
  };

  CommandManager.prototype._getCommandContext = function _getCommandContext () {
    return this.context
  };

  CommandManager.prototype._getCommandParams = function _getCommandParams () {
    var editorSession = this.context.editorSession;
    var selectionState = editorSession.getSelectionState();
    var sel = selectionState.getSelection();
    var surface = this.context.surfaceManager.getFocusedSurface();
    return {
      editorSession: editorSession,
      selectionState: selectionState,
      surface: surface,
      selection: sel,
    }
  };

  function getDOMRangeFromEvent(evt) {
    var range, x = evt.clientX, y = evt.clientY;
    
    if (document.body.createTextRange) {
      range = document.body.createTextRange();
      range.moveToPoint(x, y);
    }
    else if (!isNil(document.createRange)) {
      
      
      if (!isNil(evt.rangeParent)) {
        range = document.createRange();
        range.setStart(evt.rangeParent, evt.rangeOffset);
        range.collapse(true);
      }
      
      else if (document.caretPositionFromPoint) {
        var pos = document.caretPositionFromPoint(x, y);
        range = document.createRange();
        range.setStart(pos.offsetNode, pos.offset);
        range.collapse(true);
      }
      
      else if (document.caretRangeFromPoint) {
        range = document.caretRangeFromPoint(x, y);
      }
    }
    return range
  }


  function getSelectionRect(parentRect) {
    if (platform.inBrowser) {
      var wsel = window.getSelection();
      if (wsel.rangeCount === 0) { return }
      var wrange = wsel.getRangeAt(0);
      var contentRect = parentRect;
      var selectionRect = wrange.getBoundingClientRect();
      if (selectionRect.top === 0 && selectionRect.bottom === 0) {
        selectionRect = _fixForCursorRectBug();
      }
      return getRelativeRect(contentRect, selectionRect)
    }
  }

  function _fixForCursorRectBug() {
    var wsel = window.getSelection();
    var el = wsel.anchorNode;
    if (!el) { return }
    while (el && el.nodeType !== 1) {
      el = el.parentNode;
    }
    var rects = el.getClientRects();
    var rect = rects[0];
    return {
      left: rect.left,
      top: rect.top,
      width: 0,
      height: rect.height,
      right: rect.width,
      bottom: rect.bottom
    }
  }

  function getRelativeRect(parentRect, childRect) {
    var left = childRect.left - parentRect.left;
    var top = childRect.top - parentRect.top;
    return {
      left: left,
      top: top,
      right: parentRect.width - left - childRect.width,
      bottom: parentRect.height - top - childRect.height,
      width: childRect.width,
      height: childRect.height
    }
  }

  function isMouseInsideDOMSelection(e) {
    var wsel = window.getSelection();
    if (wsel.rangeCount === 0) {
      return false
    }
    var wrange = wsel.getRangeAt(0);
    var selectionRect = wrange.getBoundingClientRect();
    return e.clientX >= selectionRect.left &&
           e.clientX <= selectionRect.right &&
           e.clientY >= selectionRect.top &&
           e.clientY <= selectionRect.bottom
  }

  function setDOMSelection(startNode, startOffset, endNode, endOffset) {
    var wsel = window.getSelection();
    var wrange = window.document.createRange();
    if (startNode._isDOMElement) {
      startNode = startNode.getNativeElement();
    }
    if (!endNode) {
      endNode = startNode;
      endOffset = startOffset;
    }
    if (endNode._isDOMElement) {
      endNode = endNode.getNativeElement();
    }
    wrange.setStart(startNode, startOffset);
    wrange.setEnd(endNode, endOffset);
    wsel.removeAllRanges();
    wsel.addRange(wrange);
  }

  function getQueryStringParam(param, url) {
    if (typeof window === 'undefined') { return null }
    var href = url ? url : window.location.href;
    var reg = new RegExp( '[?&]' + param + '=([^&#]*)', 'i' );
    var string = reg.exec(href);
    return string ? decodeURIComponent(string[1]) : null;
  }

  var DragAndDropHandler = function DragAndDropHandler () {};

  var prototypeAccessors$6 = { _isDragAndDropHandler: { configurable: true } };

  DragAndDropHandler.prototype.match = function match (dragState, context) { 
    return false
  };

  DragAndDropHandler.prototype.drop = function drop (dragState, context) { 
    
  };

  prototypeAccessors$6._isDragAndDropHandler.get = function () {
    return true
  };

  Object.defineProperties( DragAndDropHandler.prototype, prototypeAccessors$6 );

  var DragManager = (function (EventEmitter$$1) {
    function DragManager(customDropHandlers, context) {
      EventEmitter$$1.call(this);

      this.context = context;

      var dropAssetHandlers = [];
      var moveInlineHandlers = [];
      customDropHandlers.forEach(function (h) {
        
        var type = h.type || 'drop-asset';
        switch (type) {
          case 'drop-asset': {
            dropAssetHandlers.push(h);
            break
          }
          case 'move-inline': {
            moveInlineHandlers.push(h);
            break
          }
          default:
            console.warn('Unknown type of drop handler.', h);
        }
      });

      
      this.dropHandlers = [
        
        new MoveInline(moveInlineHandlers),
        
        new MoveBlockNode(),
        
        new InsertNodes(dropAssetHandlers, this.context),
        
        
        new CustomHandler() ];
      if (platform.inBrowser) {
        this.el = DefaultDOMElement.wrapNativeElement(document);
        this.el.on('dragstart', this.onDragStart, this);
        
        this.el.on('drop', this.onDragEnd, this);
        this.el.on('dragenter', this.onDragEnter, this);
        this.el.on('dragexit', this.onDragExit, this);
        this.el.on('mousedown', this.onMousedown, this);
      }
    }

    if ( EventEmitter$$1 ) DragManager.__proto__ = EventEmitter$$1;
    DragManager.prototype = Object.create( EventEmitter$$1 && EventEmitter$$1.prototype );
    DragManager.prototype.constructor = DragManager;

    DragManager.prototype.dispose = function dispose () {
      if (this.el) {
        this.el.off(this);
      }
    };

    DragManager.prototype.onDragStart = function onDragStart (e) {
      
      this._initDrag(e, { external: false });
      
      
      var img = document.createElement("img");
      img.src = "data:image/gif;base64,R0lGODlhAQABAAAAACH5BAEKAAEALAAAAAABAAEAAAICTAEAOw==";
      e.dataTransfer.setDragImage(img, 0, 0);
      
      
      
      
      
      
      if (this.dragState.mode === 'inline') {
        e.dataTransfer.setData('text/html', img.outerHTML);
      } else {
        
        
        e.dataTransfer.setData('text/html', '<div></div>');
      }
      
    };

    
    DragManager.prototype.onDragEnter = function onDragEnter (e) {
      if (!this.dragState) {
        
        this._initDrag(e, {external: true});
      }
    };

    DragManager.prototype.onDragEnd = function onDragEnd (event) {
      if (event.__reserved__) { return }
      
      if (this.dragState) {
        event.stopPropagation();
        event.preventDefault();
        
        
        this._onDragEnd(event);
      }
    };

    DragManager.prototype.onDragExit = function onDragExit (e) {
      if (platform.isFF) ; else {
        
        
        
        this._onDragEnd(e);
      }
    };

    DragManager.prototype.extendDragState = function extendDragState (extState) {
      Object.assign(this.dragState, extState);
    };

    
    
    DragManager.prototype.onMousedown = function onMousedown (event) {
      if (this.dragState) {
        this.dragState = null;
        this._onDragEnd(event);
      }
    };

    DragManager.prototype._onDragEnd = function _onDragEnd (event) {
      if (!this.dragState) {
        
        
        
        console.warn('Not in a valid drag state.');
      } else {
        this._handleDrop(event);
      }
      this.emit('drag:finished');
      this.dragState = null;
    };

    
    DragManager.prototype._handleDrop = function _handleDrop (e) {
      var this$1 = this;

      var dragState = this.dragState;
      var i, handler;
      var match = false;
      dragState.event = e;
      dragState.data = this._getData(e);
      
      for (i = 0; i < this.dropHandlers.length && !match; i++) {
        handler = this$1.dropHandlers[i];
        match = handler.match(dragState);
      }
      if (match) {
        var editorSession = this.context.editorSession;
        editorSession.transaction(function (tx) {
          handler.drop(tx, dragState);
        });
      } else {
        console.error('No drop handler could be found.');
      }
    };

    
    DragManager.prototype._initDrag = function _initDrag (event, options) {
      
      
      
      

      
      var sel = this._getSelection();
      var dragState = Object.assign({ startEvent: event }, options);
      this.dragState = dragState;

      
      
      if (dragState.external) {
        dragState.selectionDrag = false;
        dragState.sourceSelection = null;
        dragState.scrollPanes = this._getSurfacesGroupedByScrollPane();
        this.emit('drag:started', dragState);
        return
      }

      
      
      var isSelectionDrag = (
        (sel.isPropertySelection() || sel.isContainerSelection()) &&
        isMouseInsideDOMSelection(event)
      );
      if (isSelectionDrag) {
        
        if (sel.isContainerSelection()) {
          console.warn('Dragging of ContainerSelection is not supported yet.');
          return _stop()
        }
        
        dragState.inline = true;
        dragState.selectionDrag = true;
        dragState.sourceSelection = sel;
        
        return
      }
      var comp = Component.unwrap(event.target);
      if (!comp) { return _stop() }
      var isolatedNodeComponent;
      if (comp._isInlineNodeComponent) {
        isolatedNodeComponent = comp;
        dragState.inline = true;
        dragState.sourceNode = comp.props.node;
      } else {
        isolatedNodeComponent = comp.context.isolatedNodeComponent;
      }
      if (!isolatedNodeComponent) { return _stop() }
      var surface = isolatedNodeComponent.context.surface;
      
      if(isolatedNodeComponent._isInlineNodeComponent) {
        var inlineNode = isolatedNodeComponent.props.node;
        dragState.inline = true;
        dragState.selectionDrag = true;
        dragState.sourceSelection = {
          type: 'property',
          path: inlineNode.start.path,
          startOffset: inlineNode.start.offset,
          endOffset: inlineNode.end.offset,
          containerId: surface.getContainerId(),
          surfaceId: surface.id
        };
        return
      }
      
      
      dragState.selectionDrag = false;
      dragState.nodeDrag = true;
      dragState.sourceSelection = {
        type: 'node',
        nodeId: isolatedNodeComponent.props.node.id,
        containerId: surface.getContainerId(),
        surfaceId: surface.id
      };
      
      
      
      dragState.scrollPanes = this._getSurfacesGroupedByScrollPane();
      
      this.emit('drag:started', dragState);

      function _stop() {
        
        event.preventDefault();
        event.stopPropagation();
      }
    };

    DragManager.prototype._getSurfacesGroupedByScrollPane = function _getSurfacesGroupedByScrollPane () {
      
      
      var surfaces = this.context.surfaceManager.getSurfaces();
      var scrollPanes = {};
      surfaces.forEach(function (surface) {
        
        if (!surface.isContainerEditor()) { return }
        var scrollPane = surface.context.scrollPane;
        var scrollPaneName = scrollPane.getName();
        var surfaceName = surface.getName();
        if (!scrollPanes[scrollPaneName]) {
          var surfaces = {};
          surfaces[surfaceName] = surface;
          scrollPanes[scrollPaneName] = { scrollPane: scrollPane, surfaces: surfaces };
        } else {
          scrollPanes[scrollPaneName].surfaces[surfaceName] = surface;
        }
      });
      return scrollPanes
    };

    DragManager.prototype._getSelection = function _getSelection () {
      return this.context.editorSession.getSelection()
    };

    DragManager.prototype._getComponents = function _getComponents (targetEl) {
      var res = [];
      var curr = targetEl;
      while (curr) {
        var comp = Component.getComponentForDOMElement(curr);
        if (comp) {
          res.unshift(comp);
          if(comp._isSurface) {
            return res
          }
        }
        curr = curr.parentNode;
      }
      return null
    };

    DragManager.prototype._getIsolatedNodeOrContainerChild = function _getIsolatedNodeOrContainerChild (targetEl) {
      var parent, current;
      current = targetEl;
      parent = current.parentNode;
      while(parent) {
        if (parent._comp && parent._comp._isContainerEditor) {
          return current._comp
        } else if (current._comp && current._comp._isIsolatedNode) {
          return current._comp
        }
        current = parent;
        parent = current.parentNode;
      }
    };

    
    DragManager.prototype._extractUris = function _extractUris (dataTransfer) {
      var uris = [];
      var rawUriList = dataTransfer.getData('text/uri-list');
      if (rawUriList) {
        uris = rawUriList.split('\n').filter(function(item) {
          return !item.startsWith('#')
        });
      }
      return uris
    };

    
    DragManager.prototype._getData = function _getData (e) {
      var dataTransfer = e.dataTransfer;
      if (dataTransfer) {
        return {
          files: Array.prototype.slice.call(dataTransfer.files),
          uris: this._extractUris(dataTransfer),
          text: dataTransfer.getData('text/plain'),
          html: dataTransfer.getData('text/html')
        }
      }
    };

    return DragManager;
  }(EventEmitter));


  var MoveBlockNode = (function (DragAndDropHandler$$1) {
    function MoveBlockNode () {
      DragAndDropHandler$$1.apply(this, arguments);
    }

    if ( DragAndDropHandler$$1 ) MoveBlockNode.__proto__ = DragAndDropHandler$$1;
    MoveBlockNode.prototype = Object.create( DragAndDropHandler$$1 && DragAndDropHandler$$1.prototype );
    MoveBlockNode.prototype.constructor = MoveBlockNode;

    MoveBlockNode.prototype.match = function match (dragState) {
      var ref = dragState.dropParams;
      var insertPos = ref.insertPos;
      
      return (!dragState.external && dragState.nodeDrag &&
        dragState.dropType === 'place' && insertPos >= 0
      )
    };

    MoveBlockNode.prototype.drop = function drop (tx, dragState) {
      
      
      
      
      var ref = dragState.dropParams;
      var insertPos = ref.insertPos;
      tx.setSelection(dragState.sourceSelection);
      var copy = tx.copySelection();
      
      tx.deleteSelection({ clear: true });
      var containerId = dragState.targetSurface.getContainerId();
      var surfaceId = dragState.targetSurface.getName();
      var container = tx.get(containerId);
      var targetNodeId = container.getNodeIdAt(insertPos);
      var insertMode = 'before';
      if (!targetNodeId) {
        targetNodeId = container.getNodeIdAt(insertPos-1);
        insertMode = 'after';
      }
      tx.setSelection({
        type: 'node',
        nodeId: targetNodeId,
        mode: insertMode,
        containerId: containerId,
        surfaceId: surfaceId
      });
      tx.paste(copy);
    };

    return MoveBlockNode;
  }(DragAndDropHandler));

  var MoveInline = (function (DragAndDropHandler$$1) {
    function MoveInline () {
      DragAndDropHandler$$1.apply(this, arguments);
    }

    if ( DragAndDropHandler$$1 ) MoveInline.__proto__ = DragAndDropHandler$$1;
    MoveInline.prototype = Object.create( DragAndDropHandler$$1 && DragAndDropHandler$$1.prototype );
    MoveInline.prototype.constructor = MoveInline;

    MoveInline.prototype.match = function match (dragState) {
      return !dragState.external && dragState.inline
    };

    MoveInline.prototype.drop = function drop (tx, dragState) {
      var event = dragState.event;
      var sourceSel = dragState.sourceSelection;
      var wrange = getDOMRangeFromEvent(event);
      if (!wrange) { return }
      var comp = Component.unwrap(event.target);
      if (!comp) { return }
      var domSelection = comp.context.domSelection;
      if (!domSelection) { return }
      var range = domSelection.mapDOMRange(wrange);
      if (!range) { return }
      var targetSel = tx.getDocument()._createSelectionFromRange(range);

      
      tx.selection = sourceSel;
      var snippet = tx.copySelection();
      tx.deleteSelection();
      tx.selection = transformSelection(targetSel, tx);
      tx.paste(snippet);
    };

    return MoveInline;
  }(DragAndDropHandler));

  var InsertNodes = (function (DragAndDropHandler$$1) {
    function InsertNodes(assetHandlers, context) {
      DragAndDropHandler$$1.call(this);
      this.assetHandlers = assetHandlers;
      this.context = context;
    }

    if ( DragAndDropHandler$$1 ) InsertNodes.__proto__ = DragAndDropHandler$$1;
    InsertNodes.prototype = Object.create( DragAndDropHandler$$1 && DragAndDropHandler$$1.prototype );
    InsertNodes.prototype.constructor = InsertNodes;

    InsertNodes.prototype.match = function match (dragState) {
      return dragState.dropType === 'place' && dragState.external
    };

    InsertNodes.prototype.drop = function drop (tx, dragState) {
      var this$1 = this;

      var ref = dragState.dropParams;
      var insertPos = ref.insertPos;
      var files = dragState.data.files;
      var uris = dragState.data.uris;
      var containerId = dragState.targetSurface.getContainerId();
      var surfaceId = dragState.targetSurface.id;
      var container = tx.get(containerId);
      var targetNode = container.getNodeIdAt(insertPos);
      var insertMode = 'before';
      if (!targetNode) {
        targetNode = container.getNodeIdAt(insertPos-1);
        insertMode = 'after';
      }
      tx.setSelection({
        type: 'node',
        nodeId: targetNode,
        mode: insertMode,
        containerId: containerId,
        surfaceId: surfaceId
      });
      if (files.length > 0) {
        files.forEach(function (file) {
          this$1._callHandlers(tx, {
            file: file,
            type: 'file'
          });
        });
      } else if (uris.length > 0) {
        uris.forEach(function (uri) {
          this$1._callHandlers(tx, {
            uri: uri,
            type: 'uri'
          });
        });
      } else {
        console.info('TODO: implement html/text drop here');
      }
    };

    InsertNodes.prototype._callHandlers = function _callHandlers (tx, params) {
      var this$1 = this;

      var i, handler;
      for (i = 0; i < this.assetHandlers.length; i++) {
        handler = this$1.assetHandlers[i];

        var match = handler.match(params, this$1.context);
        if (match) {
          handler.drop(tx, params, this$1.context);
          break
        }
      }
    };

    return InsertNodes;
  }(DragAndDropHandler));


  var CustomHandler = (function (DragAndDropHandler$$1) {
    function CustomHandler () {
      DragAndDropHandler$$1.apply(this, arguments);
    }

    if ( DragAndDropHandler$$1 ) CustomHandler.__proto__ = DragAndDropHandler$$1;
    CustomHandler.prototype = Object.create( DragAndDropHandler$$1 && DragAndDropHandler$$1.prototype );
    CustomHandler.prototype.constructor = CustomHandler;

    CustomHandler.prototype.match = function match (dragState) {
      return dragState.dropType === 'custom'
    };

    CustomHandler.prototype.drop = function drop (tx, dragState) {
      
      dragState.component.handleDrop(tx, dragState);
    };

    return CustomHandler;
  }(DragAndDropHandler));

  var FileManager = function FileManager(editorSession, extensions, context) {
    var this$1 = this;

    this.editorSession = editorSession;
    this.extensions = extensions;
    this.proxies = {};
    this.context = context;

    
    forEach(editorSession.getDocument().getNodes(), function (node) {
      if (node._isFileNode) { this$1.storeFile(node); }
    });

    this.editorSession.onUpdate('document', this._onDocumentChange, this);
  };

  FileManager.prototype.dispose = function dispose () {
    this.editorSession.off(this);
  };

  FileManager.prototype.storeFile = function storeFile (fileNode) {
    var proxy = this.proxies[fileNode.id];
    
    if (!proxy) {
      proxy = this.createFileProxy(fileNode);
      if (proxy) {
        this.proxies[fileNode.id] = proxy;
      }
    }
    fileNode.proxy = proxy;
    return proxy
  };

  FileManager.prototype.createFileProxy = function createFileProxy (fileNode) {
      var this$1 = this;
   
    var context = this.context;
    for (var i = 0; i < this.extensions.length; i++) {
      var ExtClass = this$1.extensions[i];
      if (ExtClass.match(fileNode, context)) {
        return new ExtClass(fileNode, context)
      }
    }
    console.error('No file adapter found for ', fileNode);
  };

  FileManager.prototype.getProxy = function getProxy (fileNode) {
    return this.proxies[fileNode.id]
  };

  FileManager.prototype.sync = function sync () {
    
    
    var promises = map(this.proxies, function (proxy) {
      return proxy.sync()
    });
    return Promise.all(promises)
  };

  FileManager.prototype._onDocumentChange = function _onDocumentChange (change) {
      var this$1 = this;

    var doc = this.editorSession.getDocument();
    forEach(change.created, function (_, id) {
      var node = doc.get(id);
      if (node._isFileNode) {
        this$1.storeFile(node);
      }
    });
  };

  var events = [ 'keydown', 'keyup', 'keypress', 'mousedown', 'mouseup' , 'copy' ];

  var GlobalEventHandler = function GlobalEventHandler(editorSession, surfaceManager) {
    this.editorSession = editorSession;
    this.surfaceManager = surfaceManager;
    this.listeners = [];
    this.initialize();
  };

  GlobalEventHandler.prototype.initialize = function initialize () {
    if (platform.inBrowser) {
      var document = DefaultDOMElement.wrapNativeElement(window.document);
      events.forEach(function(name) {
        document.on(name, this._dispatch.bind(this, name), this);
      }.bind(this));
    }
  };

  GlobalEventHandler.prototype.dispose = function dispose () {
    if (platform.inBrowser) {
      var document = DefaultDOMElement.wrapNativeElement(window.document);
      document.off(this);
    }
  };

  GlobalEventHandler.prototype.addEventListener = function addEventListener (eventName, handler, options) {
    if (!options.id) {
      throw new Error("GlobalEventHandler can only be used with option 'id'")
    }
    var listener = new DOMEventListener(eventName, handler, options);
    this.listeners.push(listener);
  };

  GlobalEventHandler.prototype.removeEventListener = function removeEventListener (listener) {
    var idx = this.listeners.indexOf(listener);
    if (idx > -1) {
      this.listeners.splice(idx, 1);
    }
  };

  GlobalEventHandler.prototype.getEventListeners = function getEventListeners () {
    return this.listeners
  };

  GlobalEventHandler.prototype._getActiveListener = function _getActiveListener (eventName) {
      var this$1 = this;

    var editorSession = this.editorSession;
    var sel = editorSession.getSelection();
    if (sel) {
      var surfaceId = sel.surfaceId;
      for (var i = 0; i < this.listeners.length; i++) {
        var listener = this$1.listeners[i];
        if (listener.eventName === eventName && listener.options.id === surfaceId) {
          return listener
        }
      }
    }
  };

  GlobalEventHandler.prototype._dispatch = function _dispatch (eventName, e) {
    var listener = this._getActiveListener(eventName);
    if (listener) {
      listener.handler(e);
    }
  };

  GlobalEventHandler.prototype.on = DOMElement.prototype.on;
  GlobalEventHandler.prototype.off = DOMElement.prototype.off;

  var keys$1 = {
    UNDEFINED: 0,
    BACKSPACE: 8,
    DELETE: 46,
    INSERT: 45,
    LEFT: 37,
    RIGHT: 39,
    UP: 38,
    DOWN: 40,
    ENTER: 13,
    RETURN: 13,
    END: 35,
    HOME: 36,
    TAB: 9,
    PAGEUP: 33,
    PAGEDOWN: 34,
    ESCAPE: 27,
    ESC: 27,
    SHIFT: 16,
    SPACE: 32,
    PLUS: 171,
    VOLUMEUP: 183,
    VOLUMEDOWN: 182,
    VOLUMEMUTE: 181,
    PRINTSCREEN: 44
  };


  for(var i$1=1;i$1<=24;i$1++) {
    keys$1['F'+i$1] = 111 + i$1;
  }

  function parseKeyEvent(event, onlyModifiers) {
    var frags = [];
    if (event.altKey) {
      if (event.code === 'AltRight') {
        frags.push('ALTGR');
      } else {
        frags.push('ALT');
      }
    }
    if (event.ctrlKey) { frags.push('CTRL'); }
    if (event.metaKey) { frags.push('META'); }
    if (event.shiftKey) { frags.push('SHIFT'); }
    if (!onlyModifiers) {
      frags.push(event.keyCode);
    }
    return frags.join('+')
  }

  var ExecuteCommandHandler = function ExecuteCommandHandler(editorSession, commandName) {
    this.editorSession = editorSession;
    this.commandName = commandName;
  };

  ExecuteCommandHandler.prototype.execute = function execute (params) {
    var commandState = params.editorSession.getCommandStates()[this.commandName];
    if (!commandState || commandState.disabled) { return false }
    this.editorSession.executeCommand(this.commandName, params);
    return true
  };

  var KeyboardManager = function KeyboardManager(editorSession, bindings, options) {
    var this$1 = this;

    this.editorSession = editorSession;
    this.context = options.context || {};
    this.keydownBindings = {};
    this.textinputBindings = {};

    bindings.forEach(function (ref) {
      var key = ref.key;
      var spec = ref.spec;

      
      var type = spec.type || 'keydown';
      if(spec.command) {
        var handler = new ExecuteCommandHandler(editorSession, spec.command);
        var hook = handler.execute.bind(handler);
        if (type === 'keydown') {
          key = parseCombo(key);
          if (!this$1.keydownBindings[key]) { this$1.keydownBindings[key] = []; }
          this$1.keydownBindings[key].push(hook);
        } else if (type === 'textinput') {
          
          this$1.textinputBindings[key] = hook;
        }
      } else {
        throw new Error('Keyboard binding not supported', spec)
      }
    });
  };

  KeyboardManager.prototype.onKeydown = function onKeydown (event) {
      var this$1 = this;

    var key = parseKeyEvent(event);
    var hooks = this.keydownBindings[key];
    if (hooks) {
      var params = this._getParams();
      var hasExecuted = false;
      for (var i = 0; i < hooks.length && !hasExecuted; i++) {
        var hook = hooks[i];
        hasExecuted = hook(params, this$1.context);
      }
      if (hasExecuted) {
        event.preventDefault();
        event.stopPropagation();
      }
      return hasExecuted
    }
  };


  KeyboardManager.prototype.onTextInput = function onTextInput (text) {
    var hook = this.textinputBindings[text];
    if (hook) {
      var params = this._getParams();
      return hook(params, this.context)
    }
  };

  KeyboardManager.prototype._getParams = function _getParams () {
    var editorSession = this.editorSession;
    var selectionState = editorSession.getSelectionState();
    var sel = selectionState.getSelection();
    var surface = this.context.surfaceManager.getFocusedSurface();
    return {
      editorSession: editorSession,
      selectionState: selectionState,
      surface: surface,
      selection: sel,
    }
  };

  function parseCombo(combo) {
    var frags = combo.split('+');
    var data = {
      keyCode: -1
    };
    for (var i = 0; i < frags.length; i++) {
      var frag = frags[i].toUpperCase();
      switch(frag) {
        case 'ALT': {
          data.altKey = true;
          break
        }
        case 'ALTGR': {
          data.altKey = true;
          data.code = 'AltRight';
          break
        }
        case 'CMD': {
          data.metaKey = true;
          break
        }
        case 'CTRL': {
          data.ctrlKey = true;
          break
        }
        case 'COMMANDORCONTROL': {
          if (platform.isMac) {
            data.metaKey = true;
          } else {
            data.ctrlKey = true;
          }
          break
        }
        case 'MEDIANEXTTRACK': {
          data.code = 'MediaTrackNext';
          break
        }
        case 'MEDIAPLAYPAUSE': {
          data.code = 'MediaPlayPause';
          break
        }
        case 'MEDIAPREVIOUSTRACK': {
          data.code = 'MediaPreviousTrack';
          break
        }
        case 'MEDIASTOP': {
          data.code = 'MediaStop';
          break
        }
        case 'SHIFT': {
          data.shiftKey = true;
          break
        }
        case 'SUPER': {
          data.metaKey = true;
          break
        }
        default:
          if (frag.length === 1) {
            data.keyCode = frag.charCodeAt(0);
          } else if (keys$1.hasOwnProperty(frag)) {
            data.keyCode = keys$1[frag];
          } else {
            throw new Error('Unsupported keyboard command: '+ combo)
          }
      }
    }
    return parseKeyEvent(data)
  }

  KeyboardManager.parseCombo = parseCombo;

  var MacroManager = function MacroManager(context, macros) {
    this.context = context;
    this.macros = macros;
    this.context.editorSession.onFinalize('document', this.onDocumentChanged, this);
  };

  MacroManager.prototype.dispose = function dispose () {
    this.context.editorSession.off(this);
  };

  MacroManager.prototype.onDocumentChanged = function onDocumentChanged (change, info) {
    this.executeMacros(change, info);
  };

  MacroManager.prototype.executeMacros = function executeMacros (change, info) {
      var this$1 = this;

    var doc = this.context.editorSession.getDocument();
    var nodeId, node, text, start, end;
    var path;
    
    
    switch(info.action) {
      case 'type': {
        var op = change.ops[0];
        if (op.type === 'update' && op.diff._isTextOperation) {
          path = op.path;
          nodeId = path[0];
          node = doc.get(nodeId);
          text = doc.get(path);
          start = op.diff.pos;
          end = start+op.diff.getLength();
        }
        break
      }
      case 'break': {
        
        
        

        
        
        
        
        var sel = change.before.selection;
        if (!sel.isPropertySelection()) { return }
        path = sel.path;
        nodeId = path[0];
        node = doc.get(nodeId);
        
        if (!node || !node.isText()) { return }
        text = node.getText();
        start = sel.start.offset;
        end = start;
        break
      }
      case 'paste': {
        
        if (change.ops.length === 1) {
          var op$1 = change.ops[0];
          if (op$1.type === 'update' && op$1.propertyType === 'string') {
            path = op$1.path;
            nodeId = path[0];
            node = doc.get(nodeId);
            if (!node.isText()) { return }
            text = node.getText();
            start = op$1.diff.pos;
            end = start+op$1.diff.getLength();
          }
        }
        break
      }
      default:
        return
    }

    var props = {
      action: info.action,
      node: node,
      path: path,
      text: text,
      start: start,
      end: end,
      editorSession: this.context.editorSession,
      selection: this.context.editorSession.getSelection()
    };

    setTimeout(function () {
      for (var i = 0; i < this$1.macros.length; i++) {
        var macro = this$1.macros[i];
        var executed = macro.execute(props, this$1.context);
        if (executed) {
          break
        }
      }
    });

  };

  var ArrayTree = function ArrayTree () {};

  ArrayTree.prototype.add = function add (path, val) {
    if (!this[path]) {
      this[path] = [];
    }
    this[path].push(val);
  };
  ArrayTree.prototype.remove = function remove (path, val) {
    if (this[path]) {
      deleteFromArray(this[path], val);
    }
  };
  ArrayTree.prototype.get = function get (path) {
    return this[path] || []
  };

  var Marker = (function (PropertyAnnotation$$1) {
    function Marker () {
      PropertyAnnotation$$1.apply(this, arguments);
    }

    if ( PropertyAnnotation$$1 ) Marker.__proto__ = PropertyAnnotation$$1;
    Marker.prototype = Object.create( PropertyAnnotation$$1 && PropertyAnnotation$$1.prototype );
    Marker.prototype.constructor = Marker;

    var prototypeAccessors = { type: { configurable: true } };

    Marker.prototype._initialize = function _initialize (doc, props) {
      this.document = doc;
      this.type = props.type;
      if (!props.type) {
        throw new Error("'type' is mandatory")
      }
      if (!props.start) {
        throw new Error("'start' is mandatory")
      }
      if (!props.end) {
        throw new Error("'end' is mandatory")
      }
      Object.assign(this, props);
    };

    
    Marker.prototype.containsSelection = function containsSelection (sel) {
      if (sel.isNull()) { return false; }
      if (sel.isPropertySelection()) {
        return (isArrayEqual(this.start.path, sel.start.path) &&
          this.start.offset <= sel.start.offset &&
          this.end.offset >= sel.end.offset)
      } else {
        console.warn('Marker.contains() does not support other selection types.');
      }
    };

    prototypeAccessors.type.get = function () {
      return this._type
    };

    prototypeAccessors.type.set = function (type) {
      this._type = type;
    };

    Object.defineProperties( Marker.prototype, prototypeAccessors );

    return Marker;
  }(PropertyAnnotation));


  Marker.prototype._isPropertyAnnotation = false;
  Marker.prototype._isMarker = true;

  var MarkersManager = function MarkersManager(editorSession) {
    this.editorSession = editorSession;

    
    this._textProperties = {};
    this._dirtyProps = {};

    this._markers = new MarkersIndex(this);

    
    
    editorSession.onUpdate(this._onChange, this);

    
    editorSession.onRender(this._updateProperties, this);
  };

  MarkersManager.prototype.dispose = function dispose () {
    this.editorSession.off(this);
    this._markers.dispose();
  };

  MarkersManager.prototype.setMarkers = function setMarkers (key, markers) {
      var this$1 = this;

    this.clearMarkers(key);
    markers.forEach(function (m) { return this$1.addMarker(key, m); });
  };

  MarkersManager.prototype.addMarker = function addMarker (key, marker) {
    marker._key = key;
    if (!marker._isMarker) {
      marker = new Marker(this.editorSession.getDocument(), marker);
    }
    this._markers.add(marker);
  };

  MarkersManager.prototype.clearMarkers = function clearMarkers (key) {
    this._markers.clear(key);
  };


  MarkersManager.prototype.register = function register (textProperyComponent) {
    var path = String(textProperyComponent.getPath());
    
    var textProperties = this._textProperties[path];
    if (!textProperties) {
      textProperties = this._textProperties[path] = [];
    }
    textProperties.push(textProperyComponent);
  };


  MarkersManager.prototype.deregister = function deregister (textProperyComponent) {
    var path = String(textProperyComponent.getPath());
    
    var textProperties = this._textProperties[path];
    if (!textProperties) {
      
      return
    }
    deleteFromArray(this._textProperties[path], textProperyComponent);
    if (textProperties.length === 0) {
      delete this._textProperties[path];
    }
  };

  MarkersManager.prototype.getMarkers = function getMarkers (path, opts) {
    opts = opts || {};
    var doc = this.editorSession.getDocument();
    var annos = doc.getAnnotations(path) || [];
    var markers = this._markers.get(path, opts.surfaceId, opts.containerId);
    return annos.concat(markers)
  };

  MarkersManager.prototype._onChange = function _onChange (editorSession) {
    if (editorSession.hasDocumentChanged()) {
      var change = editorSession.getChange();
      this._markers._onDocumentChange(change);
      this._recordDirtyTextProperties(change);
    }
  };

  MarkersManager.prototype._recordDirtyTextProperties = function _recordDirtyTextProperties (change) {
      var this$1 = this;

    
    forEach(change.updated, function (val, id) {
      this$1._dirtyProps[id] = true;
    });
  };


  MarkersManager.prototype._updateProperties = function _updateProperties () {
      var this$1 = this;

    
    Object.keys(this._dirtyProps).forEach(function (path) {
      var textProperties = this$1._textProperties[path];
      if (textProperties) {
        textProperties.forEach(this$1._updateTextProperty.bind(this$1));
      }
    });
    this._dirtyProps = {};
  };


  MarkersManager.prototype._updateTextProperty = function _updateTextProperty (textPropertyComponent) {
    var path = textPropertyComponent.getPath();
    var markers = this.getMarkers(path, {
      surfaceId: textPropertyComponent.getSurfaceId(),
      containerId: textPropertyComponent.getContainerId()
    });
    
    textPropertyComponent.setState({
      markers: markers
    });
  };


  var MarkersIndex = function MarkersIndex(manager) {
    this._manager = manager;

    this._byKey = new ArrayTree();
    this._documentMarkers = new ArrayTree();
    this._surfaceMarkers = {};
    this._containerMarkers = {};
  };

  MarkersIndex.prototype.get = function get (path, surfaceId) {
    var markers = this._documentMarkers[path] || [];
    if (surfaceId && this._surfaceMarkers[surfaceId]) {
      var surfaceMarkers = this._surfaceMarkers[surfaceId][path];
      if (surfaceMarkers) { markers = markers.concat(surfaceMarkers); }
    }
    
    return markers
  };

  MarkersIndex.prototype.add = function add (marker) {
    var key = marker._key;
    this._byKey.add(key, marker);
    this._add(marker);
  };


  MarkersIndex.prototype.remove = function remove (marker) {
    var key = marker._key;
    this._byKey.remove(key, marker);
    this._remove(marker);
  };


  MarkersIndex.prototype.clear = function clear (key) {
      var this$1 = this;

    var markers = this._byKey.get(key);
    markers.forEach(function (marker) {
      this$1._remove(marker);
    });
  };

  MarkersIndex.prototype._add = function _add (marker) {
    var dirtyProps = this._manager._dirtyProps;
    
    var scope = marker.scope || 'document';
    switch (scope) {
      case 'document': {
        var path = marker.start.path;
        
        dirtyProps[path] = true;
        this._documentMarkers.add(path, marker);
        break
      }
      case 'surface': {
        if (!this._surfaceMarkers[marker.surfaceId]) {
          this._surfaceMarkers[marker.surfaceId] = new ArrayTree();
        }
        var path$1 = marker.start.path;
        dirtyProps[path$1] = true;
        this._surfaceMarkers[marker.surfaceId].add(path$1, marker);
        break
      }
      case 'container': {
        console.warn('Container scoped markers are not supported yet');
        break
      }
      default:
        console.error('Invalid marker scope.');
    }
  };

  MarkersIndex.prototype._remove = function _remove (marker) {
    var dirtyProps = this._manager._dirtyProps;
    var scope = marker.scope || 'document';
    switch (scope) {
      case 'document': {
        var path = marker.start.path;
        dirtyProps[path] = true;
        this._documentMarkers.remove(path, marker);
        break
      }
      case 'surface': {
        if (!this._surfaceMarkers[marker.surfaceId]) {
          this._surfaceMarkers[marker.surfaceId] = new ArrayTree();
        }
        var path$1 = marker.start.path;
        dirtyProps[path$1] = true;
        this._surfaceMarkers[marker.surfaceId].remove(path$1, marker);
        break
      }
      case 'container': {
        console.warn('Container scoped markers are not supported yet');
        break
      }
      default:
        console.error('Invalid marker scope.');
    }
  };


  MarkersIndex.prototype._getAllCustomMarkers = function _getAllCustomMarkers (path) {
      var this$1 = this;

    var markers = this._documentMarkers[path] || [];
    for(var surfaceId in this$1._surfaceMarkers) {
      if (!this$1._surfaceMarkers.hasOwnProperty(surfaceId)) { continue }
      var surfaceMarkers = this$1._surfaceMarkers[surfaceId][path];
      if (surfaceMarkers) { markers = markers.concat(surfaceMarkers); }
    }
    
    return markers
  };

  MarkersIndex.prototype._onDocumentChange = function _onDocumentChange (change) {
      var this$1 = this;

    change.ops.forEach(function (op) {
      if (op.type === 'update' && op.diff._isTextOperation) {
        var markers = this$1._getAllCustomMarkers(op.path);
        var diff = op.diff;
        switch (diff.type) {
          case 'insert':
            this$1._transformInsert(markers, diff);
            break
          case 'delete':
            this$1._transformDelete(markers, diff);
            break
          default:
            
        }
      }
    });
  };

  MarkersIndex.prototype._transformInsert = function _transformInsert (markers, op) {
      var this$1 = this;

    var pos = op.pos;
    var length = op.str.length;
    if (length === 0) { return }
    markers.forEach(function (marker) {
      
      var start = marker.start.offset;
      var end = marker.end.offset;
      var newStart = start;
      var newEnd = end;
      if (pos >= end) { return }
      if (pos <= start) {
        newStart += length;
        newEnd += length;
        marker.start.offset = newStart;
        marker.end.offset = newEnd;
        return
      }
      if (pos < end) {
        newEnd += length;
        marker.end.offset = newEnd;
        
        
        
        this$1._remove(marker);
      }
    });
  };

  MarkersIndex.prototype._transformDelete = function _transformDelete (markers, op) {
      var this$1 = this;

    var pos1 = op.pos;
    var length = op.str.length;
    var pos2 = pos1 + length;
    if (pos1 === pos2) { return }
    markers.forEach(function (marker) {
      var start = marker.start.offset;
      var end = marker.end.offset;
      var newStart = start;
      var newEnd = end;
      if (pos2 <= start) {
        newStart -= length;
        newEnd -= length;
        marker.start.offset = newStart;
        marker.end.offset = newEnd;
      } else if (pos1 >= end) ;
      
      
      else {
        if (pos1 <= start) {
          newStart = start - Math.min(pos2-pos1, start-pos1);
        }
        if (pos1 <= end) {
          newEnd = end - Math.min(pos2-pos1, end-pos1);
        }
        
        if (start !== end && newStart === newEnd) {
          this$1._remove(marker);
          return
        }
        if (start !== newStart) {
          marker.start.offset = newStart;
        }
        if (end !== newEnd) {
          marker.end.offset = newEnd;
        }
        this$1._remove(marker);
      }
    });
  };

  var SurfaceManager = function SurfaceManager(editorSession) {
    this.editorSession = editorSession;
    this.surfaces = {};
    this._state = {
      selection: null
    };
    editorSession.onUpdate('selection', this._onSelectionChanged, this);
    editorSession.onPostRender(this._recoverDOMSelection, this);
  };

  SurfaceManager.prototype.dispose = function dispose () {
    this.editorSession.off(this);
  };


  SurfaceManager.prototype.getSurface = function getSurface (name) {
    if (name) {
      
      
      
      
      return this.surfaces[name]
    }
  };


  SurfaceManager.prototype.getFocusedSurface = function getFocusedSurface () {
    var sel = this._state.selection;
    if (sel && sel.surfaceId) {
      return this.getSurface(sel.surfaceId)
    }
  };

  SurfaceManager.prototype.getSurfaces = function getSurfaces () {
      var this$1 = this;

    
    
    return Object.keys(this.surfaces).map(function (key) { return this$1.surfaces[key]; })
  };


  SurfaceManager.prototype.registerSurface = function registerSurface (surface) {
    var id = surface.getId();
    if (this.surfaces[id]) {
      console.error(("A surface with id " + id + " has already been registered."));
    }
    this.surfaces[id] = surface;
  };


  SurfaceManager.prototype.unregisterSurface = function unregisterSurface (surface) {
    surface.off(this);
    var surfaceId = surface.getId();
    
    
    
    
    
    
    
    
    
    
    
    var registeredSurface = this.surfaces[surfaceId];
    if (registeredSurface === surface) {
      delete this.surfaces[surfaceId];
    }
  };

  SurfaceManager.prototype._onSelectionChanged = function _onSelectionChanged (selection) {
    var state = this._state;
    state.selection = selection;
    
    
    
    if (selection && selection.isCustomSelection() && platform.inBrowser) {
      window.getSelection().removeAllRanges();
      window.document.activeElement.blur();
    }
  };


  SurfaceManager.prototype._recoverDOMSelection = function _recoverDOMSelection () {
    
    
    
    if (this.editorSession._blurred) { return }
    var focusedSurface = this.getFocusedSurface();
    
    if (focusedSurface && !focusedSurface.isDisabled()) {
      
      focusedSurface._focus();
      focusedSurface.rerenderDOMSelection();
    }
  };

  var SaveHandlerStub = function SaveHandlerStub () {};

  SaveHandlerStub.prototype.saveDocument = function saveDocument (ref) {
      var fileManager = ref.fileManager;

    console.info('Simulating save ...');

    return fileManager.sync()
    .then(function () {
      
      
      console.info('Creating document snapshot...');
    })

  };

  var DefaultLabelProvider = function DefaultLabelProvider(labels, lang) {
    this.lang = lang || 'en';
    this.labels = labels;
  };

  DefaultLabelProvider.prototype.getLabel = function getLabel (name, params) {
    var labels = this.labels[this.lang];
    if (!labels) { return name }
    var rawLabel = labels[name] || name;
    
    if (params) {
      return this._evalTemplate(rawLabel, params)
    } else {
      return rawLabel
    }
  };

  DefaultLabelProvider.prototype.setLanguage = function setLanguage (lang) {
    this.lang = lang || 'en';
  };

  DefaultLabelProvider.prototype._evalTemplate = function _evalTemplate (label, params) {
    var vars = this._extractVariables(label);
    vars.forEach(function (varName) {
      var searchExp = new RegExp(("\\${" + varName + "}"), 'g');
      var replaceStr = params[varName];
      label = label.replace(searchExp, replaceStr);
    });
    return label
  };

  DefaultLabelProvider.prototype._extractVariables = function _extractVariables (rawLabel) {
    var qualityRegex = /\${(\w+)}/g;
    var matches;
    var vars = [];

    while (matches = qualityRegex.exec(rawLabel)) { 
      vars.push(matches[1]);
    }
    return vars
  };

  var Configurator = function Configurator() {
    this.config = {
      schema: {},
      nodes: {},
      tools: {},
      components: {},
      converters: {},
      importers: {},
      exporters: {},
      fileProxies: [],
      commands: {},
      commandGroups: {},
      toolPanels: {},
      editingBehaviors: [],
      macros: [],
      managers: {},
      dropHandlers: [],
      keyboardShortcuts: [],
      icons: {},
      labels: {},
      lang: 'en',
      editorOptions: [],
      CommandManagerClass: CommandManager,
      DragManagerClass: DragManager,
      SaveHandlerClass: null,
    };
  };





  Configurator.prototype.defineSchema = function defineSchema (schema) {
    if (schema.ArticleClass) {
      console.warn('DEPRECATED: schema.ArticleClass is now called schema.DocumentClass');
      schema.DocumentClass = schema.ArticleClass;
    }
    if (!schema.DocumentClass) {
      throw new Error('schema.DocumentClass is mandatory')
    }
    this.config.schema = schema;
  };

  Configurator.prototype.addEditorOption = function addEditorOption (option) {
    if (!option.key) {
      throw new Error('An option key must be defined')
    }
    if (!option.value) {
      throw new Error('An option value must be defined')
    }
    this.config.editorOptions[option.key] = option.value;
  };

  Configurator.prototype.getEditorOptions = function getEditorOptions () {
    return this.config.editorOptions
  };


  Configurator.prototype.addNode = function addNode (NodeClass, override) {
    var type = NodeClass.type;
    if (!type) {
      throw new Error('A NodeClass must have a type.')
    }
    if (this.config.nodes[type] && !override) {
      throw new Error('NodeClass with this type name is already registered: ' + type)
    }
    this.config.nodes[type] = NodeClass;
  };


  Configurator.prototype.addConverter = function addConverter (type, converter) {
    var converters = this.config.converters[type];
    if (!converters) {
      converters = {};
      this.config.converters[type] = converters;
    }
    if (isFunction$1(converter)) {
      var ConverterClass = converter;
      converter = new ConverterClass();
    }
    if (!converter.type) {
      throw new Error('A converter needs an associated type.')
    }
    converters[converter.type] = converter;
  };


  Configurator.prototype.addImporter = function addImporter (type, ImporterClass) {
    this.config.importers[type] = ImporterClass;
  };


  Configurator.prototype.addExporter = function addExporter (type, ExporterClass) {
    this.config.exporters[type] = ExporterClass;
  };


  Configurator.prototype.addComponent = function addComponent (nodeType, ComponentClass, force) {
    if (!force && this.config.components[nodeType]) {
      throw new Error(nodeType+' already registered')
    }
    if (!ComponentClass) {
      throw new Error('Provided nil for component '+nodeType)
    }
    if (!ComponentClass.prototype._isComponent) {
      throw new Error('ComponentClass must be a subclass of ui/Component.')
    }
    this.config.components[nodeType] = ComponentClass;
  };

  Configurator.prototype.addCommand = function addCommand (name, CommandClass, options) {
    if (!isString(name)) {
      throw new Error("Expecting 'name' to be a String")
    }
    if (!CommandClass) {
      throw new Error('Provided nil for command '+name)
    }
    if (!CommandClass.prototype._isCommand) {
      throw new Error("Expecting 'CommandClass' to be of type ui/Command.")
    }
    options = options || {};
    if (this.config.commands[name] && !options.force) {
      throw new Error(("Another command with name " + name + " has already been registered. Use 'options.force' if this is intentional."))
    }
    this.config.commands[name] = {
      name: name,
      CommandClass: CommandClass,
      options: options
    };

    
    var commandGroup = options.commandGroup;
    if (commandGroup) {
      if (!this.config.commandGroups[commandGroup]) {
        this.config.commandGroups[commandGroup] = [];
      }
      this.config.commandGroups[commandGroup].push(name);
    }
  };

  Configurator.prototype.addTool = function addTool (name, ToolClass) {
    if (!isString(name)) {
      throw new Error("Expecting 'name' to be a String")
    }
    if (!ToolClass) {
      throw new Error('Provided nil for tool '+name)
    }
    if (!ToolClass || !ToolClass.prototype._isTool) {
      throw new Error("Expecting 'ToolClass' to be of type ui/Tool. name:", name)
    }

    this.config.tools[name] = ToolClass;
  };

  Configurator.prototype.getTools = function getTools () {
    return this.config.tools
  };

  Configurator.prototype.addToolPanel = function addToolPanel (name, spec) {
    this.config.toolPanels[name] = spec;
  };

  Configurator.prototype.getToolPanel = function getToolPanel (name) {
    return this.config.toolPanels[name]
  };

  Configurator.prototype.addManager = function addManager (name, ManagerClass) {
    this.config.managers[name] = ManagerClass;
  };

  Configurator.prototype.getManagers = function getManagers () {
    return this.config.managers
  };


  Configurator.prototype.addIcon = function addIcon (iconName, options) {
    var iconConfig = this.config.icons[iconName];
    if (!iconConfig) {
      iconConfig = {};
      this.config.icons[iconName] = iconConfig;
    }
    Object.assign(iconConfig, options);
  };


  Configurator.prototype.addLabel = function addLabel (labelName, label) {
    if (isString(label)) {
      if(!this.config.labels['en']) {
        this.config.labels['en'] = {};
      }
      this.config.labels['en'][labelName] = label;
    } else {
      forEach(label, function(label, lang) {
        if (!this.config.labels[lang]) {
          this.config.labels[lang] = {};
        }
        this.config.labels[lang][labelName] = label;
      }.bind(this));
    }
  };


  Configurator.prototype.addSeed = function addSeed (seed) {
    this.config.seed = seed;
  };


  Configurator.prototype.addEditingBehavior = function addEditingBehavior (editingBehavior) {
    this.config.editingBehaviors.push(editingBehavior);
  };

  Configurator.prototype.addMacro = function addMacro (macro) {
    this.config.macros.push(macro);
  };

  Configurator.prototype.addDragAndDrop = function addDragAndDrop (DragAndDropHandlerClass) {
    
    
    console.warn('DEPRECATED: Use addDropHandler() instead');
    if (!DragAndDropHandlerClass.prototype._isDragAndDropHandler) {
      throw new Error('Only instances of DragAndDropHandler are allowed.')
    }
    this.addDropHandler(new DragAndDropHandlerClass());
  };

  Configurator.prototype.addDropHandler = function addDropHandler (dropHandler) {
    
    if (dropHandler._isDragAndDropHandler) {
      dropHandler.type = dropHandler.type || 'drop-asset';
    }
    this.config.dropHandlers.push(dropHandler);
  };

  Configurator.prototype.addKeyboardShortcut = function addKeyboardShortcut (combo, spec) {
    var entry = {
      key: combo,
      spec: spec
    };
    this.config.keyboardShortcuts.push(entry);
  };

  Configurator.prototype.addFileProxy = function addFileProxy (FileProxyClass) {
    this.config.fileProxies.push(FileProxyClass);
  };

  Configurator.prototype.getFileAdapters = function getFileAdapters () {
    return this.config.fileProxies.slice(0)
  };


  Configurator.prototype.import = function import$1 (pkg, options) {
    pkg.configure(this, options || {});
    return this
  };




  Configurator.prototype.getConfig = function getConfig () {
    return this.config
  };

  Configurator.prototype.getStyles = function getStyles () {
    return this.config.styles
  };

  Configurator.prototype.getSchema = function getSchema () {
    if (!this.schema) {
      this.schema = new DocumentSchema(this.config.schema);
      this.schema.addNodes(this.config.nodes);
    }
    return this.schema
  };

  Configurator.prototype.getDocumentClass = function getDocumentClass () {
    return this.config.schema.DocumentClass
  };

  Configurator.prototype.createArticle = function createArticle (seed) {
    console.warn('DEPRECATED: createArticle is now called createDocument');
    return this.createDocument(seed)
  };

  Configurator.prototype.createDocument = function createDocument (seed) {
    var schema = this.getSchema();
    var DocumentClass = schema.getDocumentClass();
    var doc = new DocumentClass(schema);
    if (seed) {
      seed(doc);
    }
    return doc
  };

  Configurator.prototype.createImporter = function createImporter (type, context, options) {
      if ( options === void 0 ) options = {};

    var ImporterClass = this.config.importers[type];
    var config = Object.assign({
      schema: this.getSchema(),
      converters: this.getConverterRegistry().get(type).values(),
    }, options);
    return new ImporterClass(config, context)
  };

  Configurator.prototype.createExporter = function createExporter (type, context, options) {
      if ( options === void 0 ) options = {};

    var ExporterClass = this.config.exporters[type];
    var config = Object.assign({
      schema: this.getSchema(),
      converters: this.getConverterRegistry().get(type).values()
    }, options);
    return new ExporterClass(config, context)
  };

  Configurator.prototype.getCommandGroups = function getCommandGroups () {
    return this.config.commandGroups
  };

  Configurator.prototype.getComponentRegistry = function getComponentRegistry () {
    var componentRegistry = new ComponentRegistry();
    forEach(this.config.components, function(ComponentClass, name) {
      componentRegistry.add(name, ComponentClass);
    });
    return componentRegistry
  };

  Configurator.prototype.getCommands = function getCommands () {
    return map(this.config.commands, function(item, name) {
      return new item.CommandClass(Object.assign({name: name}, item.options))
    })
  };

  Configurator.prototype.getSurfaceCommandNames = function getSurfaceCommandNames () {
    var commands = this.getCommands();
    var commandNames = commands.map(function(C) {
      return C.name
    });
    return commandNames
  };


  Configurator.prototype.getConverterRegistry = function getConverterRegistry () {
    if (!this.converterRegistry) {
      var converterRegistry = new Registry();
      forEach(this.config.converters, function(converters, name) {
        converterRegistry.add(name, new Registry(converters));
      });
      this.converterRegistry = converterRegistry;
    }
    return this.converterRegistry
  };

  Configurator.prototype.getDropHandlers = function getDropHandlers () {
    return this.config.dropHandlers.slice(0)
  };

  Configurator.prototype.getSeed = function getSeed () {
    return this.config.seed
  };

  Configurator.prototype.getIconProvider = function getIconProvider () {
    return new FontAwesomeIconProvider(this.config.icons)
  };

  Configurator.prototype.getEditingBehavior = function getEditingBehavior () {
    var editingBehavior = new EditingBehavior();
    this.config.editingBehaviors.forEach(function(behavior) {
      behavior.register(editingBehavior);
    });
    return editingBehavior
  };

  Configurator.prototype.getMacros = function getMacros () {
    return this.config.macros
  };

  Configurator.prototype.getKeyboardShortcuts = function getKeyboardShortcuts () {
    return this.config.keyboardShortcuts
  };

  Configurator.prototype.getFindAndReplaceConfig = function getFindAndReplaceConfig () {
    return this.config.findAndReplace
  };

  Configurator.prototype.setFindAndReplaceConfig = function setFindAndReplaceConfig (config) {
    this.config.findAndReplace = config;
  };


  Configurator.prototype.getKeyboardShortcutsByCommand = function getKeyboardShortcutsByCommand () {
    var keyboardShortcuts = {};
    this.config.keyboardShortcuts.forEach(function (entry) {
      if (entry.spec.command) {
        var shortcut = entry.key.toUpperCase();

        if (platform.isMac) {
          shortcut = shortcut.replace(/CommandOrControl/i, '⌘');
          shortcut = shortcut.replace(/Ctrl/i, '^');
          shortcut = shortcut.replace(/Shift/i, '⇧');
          shortcut = shortcut.replace(/Enter/i, '↵');
          shortcut = shortcut.replace(/Alt/i, '⌥');
          shortcut = shortcut.replace(/\+/g, '');
        } else {
          shortcut = shortcut.replace(/CommandOrControl/i, 'Ctrl');
        }

        keyboardShortcuts[entry.spec.command] = shortcut;
      }
    });
    return keyboardShortcuts
  };

  Configurator.prototype.setDefaultLanguage = function setDefaultLanguage (lang) {
    this.config.lang = lang;
  };

  Configurator.prototype.getDefaultLanguage = function getDefaultLanguage () {
    return this.config.lang || 'en'
  };



  Configurator.prototype.setCommandManagerClass = function setCommandManagerClass (CommandManagerClass) {
    this.config.CommandManagerClass = CommandManagerClass;
  };

  Configurator.prototype.getCommandManagerClass = function getCommandManagerClass () {
    return this.config.CommandManagerClass || CommandManager
  };

  Configurator.prototype.setDragManagerClass = function setDragManagerClass (DragManagerClass) {
    this.config.DragManagerClass = DragManagerClass;
  };

  Configurator.prototype.getDragManagerClass = function getDragManagerClass () {
    return this.config.DragManagerClass || DragManager
  };

  Configurator.prototype.setFileManagerClass = function setFileManagerClass (FileManagerClass) {
    this.config.FileManagerClass = FileManagerClass;
  };

  Configurator.prototype.getFileManagerClass = function getFileManagerClass () {
    return this.config.FileManagerClass || FileManager
  };

  Configurator.prototype.setGlobalEventHandlerClass = function setGlobalEventHandlerClass (GlobalEventHandlerClass) {
    this.config.GlobalEventHandlerClass = GlobalEventHandlerClass;
  };

  Configurator.prototype.getGlobalEventHandlerClass = function getGlobalEventHandlerClass () {
    return this.config.GlobalEventHandlerClass || GlobalEventHandler
  };

  Configurator.prototype.setKeyboardManagerClass = function setKeyboardManagerClass (KeyboardManagerClass) {
    this.config.KeyboardManagerClass = KeyboardManagerClass;
  };

  Configurator.prototype.getKeyboardManagerClass = function getKeyboardManagerClass () {
    return this.config.KeyboardManagerClass || KeyboardManager
  };

  Configurator.prototype.setMacroManagerClass = function setMacroManagerClass (MacroManagerClass) {
    this.config.MacroManagerClass = MacroManagerClass;
  };

  Configurator.prototype.getMacroManagerClass = function getMacroManagerClass () {
    return this.config.MacroManagerClass || MacroManager
  };

  Configurator.prototype.setMarkersManagerClass = function setMarkersManagerClass (MarkersManagerClass) {
    this.config.MarkersManagerClass = MarkersManagerClass;
  };

  Configurator.prototype.getMarkersManagerClass = function getMarkersManagerClass () {
    return this.config.MarkersManagerClass || MarkersManager
  };

  Configurator.prototype.setSurfaceManagerClass = function setSurfaceManagerClass (SurfaceManagerClass) {
    this.config.SurfaceManagerClass = SurfaceManagerClass;
  };

  Configurator.prototype.getSurfaceManagerClass = function getSurfaceManagerClass () {
    return this.config.SurfaceManagerClass || SurfaceManager
  };

  Configurator.prototype.setSaveHandlerClass = function setSaveHandlerClass (SaveHandlerClass) {
    this.config.SaveHandlerClass = SaveHandlerClass;
  };

  Configurator.prototype.getSaveHandler = function getSaveHandler () {
    var SaveHandler = this.config.SaveHandlerClass || SaveHandlerStub;
    return new SaveHandler()
  };

  Configurator.prototype.getLabelProviderClass = function getLabelProviderClass () {
    return this.config.LabelProviderClass || DefaultLabelProvider
  };

  Configurator.prototype.setLabelProviderClass = function setLabelProviderClass (LabelProviderClass) {
    this.config.LabelProviderClass = LabelProviderClass;
  };

  Configurator.prototype.getLabelProvider = function getLabelProvider () {
    var LabelProvider = this.getLabelProviderClass();
    return new LabelProvider(this.config.labels)
  };

  function TextNodeMixin(SuperClass) {
    var TextNodeMixin = (function (SuperClass) {
      function TextNodeMixin () {
        SuperClass.apply(this, arguments);
      }

      if ( SuperClass ) TextNodeMixin.__proto__ = SuperClass;
      TextNodeMixin.prototype = Object.create( SuperClass && SuperClass.prototype );
      TextNodeMixin.prototype.constructor = TextNodeMixin;

      TextNodeMixin.prototype.getTextPath = function getTextPath () {
        
        console.warn('DEPRECATED: use node.getPath()');
        return this.getPath()
      };

      TextNodeMixin.prototype.getText = function getText () {
        return this.content
      };

      TextNodeMixin.prototype.isEmpty = function isEmpty () {
        return !this.getText()
      };

      TextNodeMixin.prototype.getLength = function getLength () {
        return this.getText().length
      };

      TextNodeMixin.prototype.getAnnotations = function getAnnotations () {
        return this.getDocument().getIndex('annotations').get(this.getPath())
      };

      return TextNodeMixin;
    }(SuperClass));
    return TextNodeMixin
  }

  var ENTER = 1;
  var EXIT = -1;
  var ANCHOR = -2;






























  var Fragmenter = function Fragmenter(options) {
    Object.assign(this, options);
  };

  Fragmenter.prototype.start = function start (rootContext, text, annotations) {
    if (!isString(text)) {
      throw new Error("Illegal argument: 'text' must be a String, but was " + text)
    }
    this._start(rootContext, text, annotations);
  };

  Fragmenter.prototype.onText = function onText (context, text, entry) { 
  };


  Fragmenter.prototype.onEnter = function onEnter (entry, parentContext) { 
    return null
  };

  Fragmenter.prototype.onExit = function onExit (entry, context, parentContext) { 
  };

  Fragmenter.prototype._enter = function _enter (entry, parentContext) {
    entry.counter++;
    return this.onEnter(entry, parentContext)
  };

  Fragmenter.prototype._exit = function _exit (entry, context, parentContext) {
    this.onExit(entry, context, parentContext);
  };

  Fragmenter.prototype._createText = function _createText (context, text, entry) {
    this.onText(context, text, entry);
  };

  Fragmenter.prototype._start = function _start (rootContext, text, annotations) {
      var this$1 = this;

    var entries = _extractEntries.call(this, annotations);
    var stack = [{context: rootContext, entry: null}];

    var pos = 0;
    for (var i = 0; i < entries.length; i++) {
      var entry = entries[i];
      var textFragment = text.substring(pos, entry.pos);
      if (textFragment) {
        
        this$1._createText(stack[stack.length-1].context, textFragment, entry);
      }

      pos = entry.pos;
      var stackLevel, idx, _entry;
      if (entry.mode === ENTER || entry.mode === ANCHOR) {
        
        for (stackLevel = 1; stackLevel < stack.length; stackLevel++) {
          if (entry.level < stack[stackLevel].entry.level) {
            break
          }
        }
        
        
        for (idx = stack.length-1; idx >= stackLevel; idx--) {
          _entry = stack[idx].entry;
          
          _entry.length = pos - _entry.pos;
          this$1._exit(_entry, stack[idx].context, stack[idx-1].context);
        }
        stack.splice(stackLevel, 0, {entry: entry});
        
        for (idx = stackLevel; idx < stack.length; idx++) {
          _entry = stack[idx].entry;
          
          _entry.pos = pos;
          stack[idx].context = this$1._enter(_entry, stack[idx-1].context);
        }
      }
      if (entry.mode === EXIT || entry.mode === ANCHOR) {
        
        for (stackLevel = 1; stackLevel < stack.length; stackLevel++) {
          if (stack[stackLevel].entry.node === entry.node) {
            break
          }
        }
        for (idx = stack.length-1; idx >= stackLevel; idx--) {
          _entry = stack[idx].entry;
          
          _entry.length = pos - _entry.pos;
          this$1._exit(_entry, stack[idx].context, stack[idx-1].context);
        }
        stack.splice(stackLevel, 1);
        
        for (idx = stackLevel; idx < stack.length; idx++) {
          _entry = stack[idx].entry;
          
          _entry.pos = pos;
          stack[idx].context = this$1._enter(_entry, stack[idx-1].context);
        }
      }
    }

    
    var trailingText = text.substring(pos);
    if (trailingText) {
      this._createText(rootContext, trailingText);
    }
  };

  Fragmenter.SHOULD_NOT_SPLIT = 0;
  Fragmenter.NORMAL = 10;
  Fragmenter.ANY = 100;
  Fragmenter.ALWAYS_ON_TOP = Number.MAX_VALUE;







































  function _extractEntries(annotations) {
    var openers = [];
    var closers = [];
    forEach(annotations, function(a) {
      var isAnchor = (a.isAnchor ? a.isAnchor() : false);
      
      if (isAnchor) {
        openers.push({
          mode: ANCHOR,
          pos: a.offset,
          id: a.id,
          level: Fragmenter.ALWAYS_ON_TOP,
          type: 'anchor',
          node: a,
          counter: -1,
          length: 0
        });
      } else {
        
        
        
        
        
        
        

        
        var l = Fragmenter.NORMAL;
        var isInline = (a.isInline ? a.isInline() : false);
        if (isInline) {
          l = Number.MAX_VALUE;
        } else if (a.constructor.hasOwnProperty('fragmentation')) {
          l = a.constructor.fragmentation;
        } else if (a.hasOwnProperty('fragmentationHint')) {
          l = a.fragmentationHint;
        }
        var startOffset = Math.min(a.start.offset, a.end.offset);
        var endOffset = Math.max(a.start.offset, a.end.offset);
        var opener = {
          pos: startOffset,
          mode: ENTER,
          level: l,
          id: a.id,
          type: a.type,
          node: a,
          length: 0,
          counter: -1,
        };
        openers.push(opener);
        closers.push({
          pos: endOffset,
          mode: EXIT,
          level: l,
          id: a.id,
          type: a.type,
          node: a,
          opener: opener
        });
      }
    });

    
    openers.sort(_compareOpeners);
    
    for (var i = openers.length - 1; i >= 0; i--) {
      openers[i].idx = i;
    }
    closers.sort(_compareClosers);
    
    var entries = new Array(openers.length+closers.length);
    var idx = 0;
    var idx1 = 0;
    var idx2 = 0;
    var opener = openers[idx1];
    var closer = closers[idx2];
    while(opener || closer) {
      if (opener && closer) {
        
        if (closer.pos <= opener.pos && closer.opener !== opener) {
          entries[idx] = closer;
          idx2++;
        } else {
          entries[idx] = opener;
          idx1++;
        }
      } else if (opener) {
        entries[idx] = opener;
        idx1++;
      } else if (closer) {
        entries[idx] = closer;
        idx2++;
      }
      opener = openers[idx1];
      closer = closers[idx2];
      idx++;
    }
    return entries
  }

  function _compareOpeners(a, b) {
    if (a.pos < b.pos) { return -1 }
    if (a.pos > b.pos) { return 1 }
    if (a.mode < b.mode) { return -1 }
    if (a.mode > b.mode) { return 1 }
    if (a.mode === b.mode) {
      if (a.level < b.level) { return -1 }
      if (a.level > b.level) { return 1 }
    }
    return 0
  }


  function _compareClosers(a, b) {
    if (a.pos < b.pos) { return -1 }
    if (a.pos > b.pos) { return 1 }
    
    
    
    
    
    if (a.pos === a.opener.pos && b.pos === b.opener.pos) {
      if (a.opener.idx < b.opener.idx) {
        return -1
      } else {
        return 1
      }
    }
    if (a.opener.idx > b.opener.idx) { return -1 }
    if (a.opener.idx < b.opener.idx) { return 1 }
    return 0
  }

  function appendChild(xmlNode, child) {
    insertAt(xmlNode, xmlNode._childNodes.length, child);
    return xmlNode
  }

  function getChildPos(xmlNode, child) {
    return xmlNode._childNodes.indexOf(child.id)
  }

  function removeChild(xmlNode, child) {
    var childPos = getChildPos(xmlNode, child);
    if (childPos >= 0) {
      removeAt(xmlNode, childPos);
    } else {
      throw new Error(("Node " + (child.id) + " is not a child of " + (xmlNode.id)))
    }
    return xmlNode
  }

  function insertBefore(xmlNode, newChild, ref) {
    if (!ref) {
      appendChild(xmlNode, newChild);
    } else {
      var pos = getChildPos(xmlNode, ref);
      if (pos < 0) {
        throw new Error(("Node " + (newChild.id) + " is not a child of " + (xmlNode.id)))
      }
      insertAt(xmlNode, pos, newChild);
    }
    return xmlNode
  }

  function insertAt(xmlNode, pos, child) {
    var length = xmlNode._childNodes.length;
    if (pos >= 0 && pos <= length) {
      var doc = xmlNode.getDocument();
      doc.update([xmlNode.id, '_childNodes'], { type: 'insert', pos: pos, value: child.id });
    } else {
      throw new Error('Index out of bounds.')
    }
    return xmlNode
  }

  function removeAt(xmlNode, pos) {
    var length = xmlNode._childNodes.length;
    if (pos >= 0 && pos < length) {
      var doc = xmlNode.getDocument();
      doc.update([xmlNode.id, '_childNodes'], { type: 'delete', pos: pos });
    } else {
      throw new Error('Index out of bounds.')
    }
    return xmlNode
  }

  function getInnerXML(xmlNode) {
    if (xmlNode._childNodes) {
      return xmlNode.getChildNodes().map(function (child) {
        return child.toXML().serialize()
      }).join('')
    } else if (xmlNode.isText()) {
      return xmlNode.toXML().getInnerXML()
    }
    return ''
  }

  function getChildAt(xmlNode, idx) {
    var childId = xmlNode._childNodes[idx];
    if (childId) {
      return xmlNode.getDocument().get(childId)
    }
  }


  function node2element(node) {
    
    var dom = DefaultDOMElement.createDocument('xml');
    var el = _node2element(dom, node);
    return el
  }

  function _node2element(dom, node) {
    var el;
    switch(node._elementType) {
      case 'text': {
        el = _renderTextNode(dom, node);
        break
      }
      case 'element':
      case 'inline-element':
      case 'container': {
        el = _renderElementNode(dom, node);
        break
      }
      case 'anchor':
      case 'annotation': {
        el = _createElement$1(dom, node);
        break
      }
      case 'external': {
        el = DefaultDOMElement.parseSnippet(node.xml, 'xml');
        break
      }
      default:
        throw new Error('Invalid element type.')
    }
    return el
  }

  function _createElement$1(dom, node) {
    var el = dom.createElement(node.type);
    el.attr('id', node.id);
    el.attr(node.attributes);
    return el
  }

  function _renderElementNode(dom, node) {
    var el = _createElement$1(dom, node);
    el.append(node.getChildren().map(function (child) { return _node2element(dom, child); }));
    return el
  }

  function _renderTextNode(dom, node) {
    var annos = node.getAnnotations();
    var text = node.getText();
    var el = _createElement$1(dom, node);
    if (annos && annos.length > 0) {
      var fragmenter = new Fragmenter({
        onText: function (context, text) {
          var node = context.node;
          if (node.isText() || (node.isAnnotation() && !node._isInlineNode)) {
            context.el.append(text);
          }
        },
        onEnter: function (fragment) {
          return {
            el: _node2element(dom, fragment.node),
            node: fragment.node
          }
        },
        onExit: function (fragment, context, parentContext) {
          parentContext.el.append(context.el);
        }
      });
      fragmenter.start({ el: el, node: node }, text, annos);
    } else {
      el.append(text);
    }
    return el
  }

  function nameWithoutNS$1(name) {
    var idx = name.indexOf(':');
    if (idx > 0) {
      return name.slice(idx+1)
    } else {
      return name
    }
  }

  var Adapter = (function (superclass) {
    function Adapter () {
      superclass.apply(this, arguments);
    }

    if ( superclass ) Adapter.__proto__ = superclass;
    Adapter.prototype = Object.create( superclass && superclass.prototype );
    Adapter.prototype.constructor = Adapter;

    Adapter.prototype.isTag = function isTag () {
      return true
    };

    Adapter.prototype.getChildren = function getChildren (elem){
      return elem.getChildren() || []
    };

    Adapter.prototype.getAttributeValue = function getAttributeValue (elem, name){
      return elem.getAttribute(name)
    };

    Adapter.prototype.getAttributes = function getAttributes (elem) {
      return ['id', elem.id].concat(map(elem.attributes, function (val, key) { return [key,val] }))
    };

    Adapter.prototype.hasAttrib = function hasAttrib (elem, name){
      return name === 'id' || elem.attributes.hasOwnProperty(name)
    };

    Adapter.prototype.getName = function getName (elem){
      return elem.type
    };

    Adapter.prototype.getNameWithoutNS = function getNameWithoutNS (elem){
      return nameWithoutNS$1(this.getName(elem))
    };

    Adapter.prototype.getText = function getText (elem) {
      if (elem._elementType === 'text') {
        return elem.getText()
      }
      
      return ''
    };

    return Adapter;
  }(domUtils.DomUtils));

  var cssSelectAdapter = new Adapter()

  var XMLDocumentNode = (function (DocumentNode$$1) {
    function XMLDocumentNode () {
      DocumentNode$$1.apply(this, arguments);
    }

    if ( DocumentNode$$1 ) XMLDocumentNode.__proto__ = DocumentNode$$1;
    XMLDocumentNode.prototype = Object.create( DocumentNode$$1 && DocumentNode$$1.prototype );
    XMLDocumentNode.prototype.constructor = XMLDocumentNode;

    var prototypeAccessors = { children: { configurable: true },tagName: { configurable: true },parent: { configurable: true } };

    XMLDocumentNode.prototype._initialize = function _initialize (doc, props) {
      
      
      
      if (props.attributes) {
        delete props.attributes.id;
      }
      DocumentNode$$1.prototype._initialize.call(this, doc, props);
    };

    XMLDocumentNode.prototype.toXML = function toXML () {
      return node2element(this)
    };

    
    XMLDocumentNode.prototype.findChild = function findChild (tagName) {
      var children = this.getChildren();
      for (var i = 0; i < children.length; i++) {
        var child = children[i];
        if (child.type === tagName) { return child }
      }
    };

    XMLDocumentNode.prototype.find = function find (cssSelector) {
      return cssSelect.selectOne(cssSelector, this, { xmlMode: true, adapter: cssSelectAdapter })
    };

    XMLDocumentNode.prototype.findAll = function findAll (cssSelector) {
      return cssSelect.selectAll(cssSelector, this, { xmlMode: true, adapter: cssSelectAdapter })
    };

    XMLDocumentNode.prototype.isContainer = function isContainer () {
      return false
    };

    
    XMLDocumentNode.prototype.isBlock = function isBlock () {
      var parentNode = this.parentNode;
      return (parentNode && parentNode.isContainer())
    };

    prototypeAccessors.children.get = function () {
      return this.getChildren()
    };

    XMLDocumentNode.prototype.getChildren = function getChildren () {
      
      return this.getChildNodes()
    };

    XMLDocumentNode.prototype.getChildNodes = function getChildNodes () {
      if (this._childNodes) {
        return getNodes(this.getDocument(), this._childNodes)
      } else {
        return []
      }
    };

    XMLDocumentNode.prototype.getChildCount = function getChildCount () {
      if (this._childNodes) {
        return this._childNodes.length
      } else {
        return 0
      }
    };

    XMLDocumentNode.prototype.getChildPosition = function getChildPosition (child) {
      if (!child || !this._childNodes) { return -1 }
      return this._childNodes.indexOf(child.id)
    };

    XMLDocumentNode.prototype.getChildNodeIterator = function getChildNodeIterator () {
      return new ArrayIterator(this.getChildNodes())
    };

    XMLDocumentNode.prototype.getFirstChild = function getFirstChild () {
      if (this._childNodes) {
        return this.getDocument().get(this._childNodes[0])
      }
    };

    XMLDocumentNode.prototype.getLastChild = function getLastChild () {
      if (this._childNodes) {
        return this.getDocument().get(last$1(this._childNodes))
      }
    };

    XMLDocumentNode.prototype.getParent = function getParent () {
      return this.parentNode
    };

    prototypeAccessors.tagName.get = function () {
      return this.type
    };

    
    prototypeAccessors.parent.get = function () {
      return this.parentNode
    };

    XMLDocumentNode.prototype.setAttribute = function setAttribute (name, val) {
      if (name === 'id') {
        throw new Error("'id' is read-only and can not be changed")
      }
      this.getDocument().set([this.id, 'attributes', name], val);
      return this
    };

    XMLDocumentNode.prototype.getAttribute = function getAttribute (name) {
      if (name === 'id') { return this.id }
      return this.attributes[name]
    };

    XMLDocumentNode.prototype.removeAttribute = function removeAttribute (name) {
      if (this.attributes.hasOwnProperty(name)) {
        this.getDocument().set([this.id, 'attributes', name], undefined);
        delete this.attributes[name];
      }
      return this
    };

    XMLDocumentNode.prototype.getAttributes = function getAttributes () {
      return clone(this.attributes)
    };

    XMLDocumentNode.prototype.getElementSchema = function getElementSchema () {
      return this.getDocument().getElementSchema(this.type)
    };

    XMLDocumentNode.prototype.serialize = function serialize () {
      return this.toXML()
    };

    XMLDocumentNode.prototype.isTextNode = function isTextNode () {
      return false
    };

    XMLDocumentNode.prototype.isElementNode = function isElementNode () {
      return false
    };

    XMLDocumentNode.prototype.getInnerXML = function getInnerXML$1 () {
      return getInnerXML(this)
    };

    Object.defineProperties( XMLDocumentNode.prototype, prototypeAccessors );

    return XMLDocumentNode;
  }(DocumentNode));

  XMLDocumentNode.prototype.attr = DOMElement.prototype.attr;

  XMLDocumentNode.schema = {
    attributes: { type: 'object', default: {} }
  };

  var XMLTextElement = (function (superclass) {
    function XMLTextElement () {
      superclass.apply(this, arguments);
    }

    if ( superclass ) XMLTextElement.__proto__ = superclass;
    XMLTextElement.prototype = Object.create( superclass && superclass.prototype );
    XMLTextElement.prototype.constructor = XMLTextElement;

    var prototypeAccessors = { textContent: { configurable: true } };

    XMLTextElement.prototype.getPath = function getPath () {
      return [this.id, 'content']
    };

    XMLTextElement.prototype.getText = function getText () {
      return this.content
    };

    
    XMLTextElement.prototype.getChildren = function getChildren () {
      var annos = this.getAnnotations();
      
      annos.sort(_byStartOffset);
      return annos
    };

    XMLTextElement.prototype.setText = function setText (text) {
      var doc = this.getDocument();
      var path = this.getPath();
      var oldText = this.getText();
      
      if (oldText.length > 0) {
        doc.update(path, { type: 'delete', start: 0, end: oldText.length });
      }
      doc.update(path, { type: 'insert', start: 0, text: text });
      return this
    };

    

    XMLTextElement.prototype.getTextContent = function getTextContent () {
      return this.getText()
    };

    XMLTextElement.prototype.setTextContent = function setTextContent (text) {
      return this.setText(text)
    };

    prototypeAccessors.textContent.get = function () {
      return this.getText()
    };

    prototypeAccessors.textContent.set = function (text) {
      this.setText(text);
    };

    XMLTextElement.prototype.appendChild = function appendChild (child) { 
      
      
      
      
      
      
      
      throw new Error('This is not implemented yet.')
    };

    XMLTextElement.prototype.removeChild = function removeChild (child) { 
      
      throw new Error('This is not implemented yet.')
    };

    
    XMLTextElement.prototype.isElementNode = function isElementNode () {
      return true
    };

    Object.defineProperties( XMLTextElement.prototype, prototypeAccessors );

    return XMLTextElement;
  }(TextNodeMixin(XMLDocumentNode)));

  XMLTextElement.prototype._isXMLTextElement = true;

  XMLTextElement.prototype.text = DOMElement.prototype.text;

  XMLTextElement.prototype._elementType = 'text';


  XMLTextElement.isText = true;
  XMLTextElement.isBlock = true;

  XMLTextElement.type = 'text';

  XMLTextElement.schema = {
    content: "text"
  };

  function _byStartOffset(a,b) {
    return a.start.offset - b.start.offset
  }

  var XMLNodeConverter = function XMLNodeConverter(type) {
    this.type = type;
    this.tagName = nameWithoutNS$1(type);
    this.tagNameNS = type;
  };

  XMLNodeConverter.prototype.matchElement = function matchElement (el) {
    return (el.tagName === this.tagNameNS)
  };

  XMLNodeConverter.prototype.export = function export$1 (node, el) {
    el.tagName = this.tagNameNS;
    el.setAttributes(node.attributes);
  };

  var XMLTextElementConverter = (function (XMLNodeConverter$$1) {
    function XMLTextElementConverter () {
      XMLNodeConverter$$1.apply(this, arguments);
    }

    if ( XMLNodeConverter$$1 ) XMLTextElementConverter.__proto__ = XMLNodeConverter$$1;
    XMLTextElementConverter.prototype = Object.create( XMLNodeConverter$$1 && XMLNodeConverter$$1.prototype );
    XMLTextElementConverter.prototype.constructor = XMLTextElementConverter;

    XMLTextElementConverter.prototype.import = function import$1 (el, node, converter) {
      node.content = converter.annotatedText(el, [node.id, 'content'], { preserveWhitespace: true });
    };

    XMLTextElementConverter.prototype.export = function export$1 (node, el, converter) {
      el.tagName = this.tagNameNS;
      el.setAttributes(node.attributes);
      el.append(converter.annotatedText([node.id, 'content']));
    };

    return XMLTextElementConverter;
  }(XMLNodeConverter));

  var XMLElementNode = (function (XMLDocumentNode$$1) {
    function XMLElementNode () {
      XMLDocumentNode$$1.apply(this, arguments);
    }

    if ( XMLDocumentNode$$1 ) XMLElementNode.__proto__ = XMLDocumentNode$$1;
    XMLElementNode.prototype = Object.create( XMLDocumentNode$$1 && XMLDocumentNode$$1.prototype );
    XMLElementNode.prototype.constructor = XMLElementNode;

    XMLElementNode.prototype.appendChild = function appendChild$$1 (child) {
      return appendChild(this, child)
    };

    XMLElementNode.prototype.removeChild = function removeChild$$1 (child) {
      return removeChild(this, child)
    };

    XMLElementNode.prototype.insertBefore = function insertBefore$$1 (newChild, ref) {
      return insertBefore(this, newChild, ref)
    };

    XMLElementNode.prototype.insertAt = function insertAt$$1 (pos, child) {
      return insertAt(this, pos, child)
    };

    XMLElementNode.prototype.removeAt = function removeAt$$1 (pos) {
      return removeAt(this, pos)
    };

    XMLElementNode.prototype.getChildAt = function getChildAt$$1 (idx) {
      return getChildAt(this, idx)
    };

    XMLElementNode.prototype.isElementNode = function isElementNode () {
      return true
    };

    return XMLElementNode;
  }(XMLDocumentNode));

  XMLElementNode.prototype.append = DOMElement.prototype.append;

  XMLElementNode.prototype._elementType = 'element';

  XMLElementNode.type = 'element';

  XMLElementNode.schema = {
    _childNodes: { type: ['array', 'id'], default: [], owned: true}
  };

  XMLElementNode.isBlock = true;

  var ElementNodeConverter = (function (XMLNodeConverter$$1) {
    function ElementNodeConverter () {
      XMLNodeConverter$$1.apply(this, arguments);
    }

    if ( XMLNodeConverter$$1 ) ElementNodeConverter.__proto__ = XMLNodeConverter$$1;
    ElementNodeConverter.prototype = Object.create( XMLNodeConverter$$1 && XMLNodeConverter$$1.prototype );
    ElementNodeConverter.prototype.constructor = ElementNodeConverter;

    ElementNodeConverter.prototype.import = function import$1 (el, node, converter) {
      var it = converter.getChildNodeIterator(el);
      var childNodeIds = [];
      while(it.hasNext()) {
        var childEl = it.next();
        if (childEl.isElementNode()) {
          var childNode = converter.convertElement(childEl);
          childNodeIds.push(childNode.id);
        }
      }
      node._childNodes = childNodeIds;
    };

    ElementNodeConverter.prototype.export = function export$1 (node, el, converter) {
      el.tagName = this.tagNameNS;
      el.setAttributes(node.attributes);
      el.childNodes.forEach(function (childNode) {
        var childEl = converter.convertNode(childNode);
        el.appendChild(childEl);
      });
    };

    return ElementNodeConverter;
  }(XMLNodeConverter));

  var XMLAnnotationNode = (function (superclass) {
    function XMLAnnotationNode () {
      superclass.apply(this, arguments);
    }

    if ( superclass ) XMLAnnotationNode.__proto__ = superclass;
    XMLAnnotationNode.prototype = Object.create( superclass && superclass.prototype );
    XMLAnnotationNode.prototype.constructor = XMLAnnotationNode;

    var prototypeAccessors = { parentNode: { configurable: true } };

    prototypeAccessors.parentNode.get = function () {
      var path = this.start.path;
      var doc = this.getDocument();
      return doc.get(path[0])
    };

    XMLAnnotationNode.prototype.isPropertyAnnotation = function isPropertyAnnotation () {
      return true
    };

    Object.defineProperties( XMLAnnotationNode.prototype, prototypeAccessors );

    return XMLAnnotationNode;
  }(AnnotationMixin(XMLDocumentNode)));

  XMLAnnotationNode.prototype._elementType = 'annotation';


  XMLAnnotationNode.prototype._isPropertyAnnotation = true;

  XMLAnnotationNode.type = 'annotation';


  XMLAnnotationNode.schema = {};

  var XMLAnchorNode = (function (XMLDocumentNode$$1) {
    function XMLAnchorNode () {
      XMLDocumentNode$$1.apply(this, arguments);
    }

    if ( XMLDocumentNode$$1 ) XMLAnchorNode.__proto__ = XMLDocumentNode$$1;
    XMLAnchorNode.prototype = Object.create( XMLDocumentNode$$1 && XMLDocumentNode$$1.prototype );
    XMLAnchorNode.prototype.constructor = XMLAnchorNode;

    var prototypeAccessors = { parentNode: { configurable: true } };

    prototypeAccessors.parentNode.get = function () {
      var path = this.coor.start.path;
      var doc = this.getDocument();
      return doc.get(path[0])
    };

    Object.defineProperties( XMLAnchorNode.prototype, prototypeAccessors );

    return XMLAnchorNode;
  }(XMLDocumentNode));

  XMLAnchorNode.prototype._elementType = 'anchor';

  XMLAnchorNode.type = 'anchor';

  XMLAnchorNode.schema = {
    coor: { type: "coordinate", optional: true }
  };

  var XMLInlineElementNode = (function (XMLAnnotationNode$$1) {
    function XMLInlineElementNode () {
      XMLAnnotationNode$$1.apply(this, arguments);
    }

    if ( XMLAnnotationNode$$1 ) XMLInlineElementNode.__proto__ = XMLAnnotationNode$$1;
    XMLInlineElementNode.prototype = Object.create( XMLAnnotationNode$$1 && XMLAnnotationNode$$1.prototype );
    XMLInlineElementNode.prototype.constructor = XMLInlineElementNode;

    var prototypeAccessors = { parentNode: { configurable: true } };

    prototypeAccessors.parentNode.get = function () {
      var path = this.start.path;
      if (path[0]) {
        var doc = this.getDocument();
        return doc.get(path[0])
      }
      return this._parentNode
    };

    prototypeAccessors.parentNode.set = function (parent) {
      var path = this.start.path;
      if (path[0]) {
        throw new Error('parent of inline-element is implicitly given')
      }
      this._parentNode = parent;
    };

    XMLInlineElementNode.prototype.getChildAt = function getChildAt$$1 (idx) {
      return getChildAt(this, idx)
    };

    XMLInlineElementNode.prototype.appendChild = function appendChild$$1 (child) {
      return appendChild(this, child)
    };

    XMLInlineElementNode.prototype.removeChild = function removeChild$$1 (child) {
      return removeChild(this, child)
    };

    XMLInlineElementNode.prototype.insertBefore = function insertBefore$$1 (newChild, ref) {
      return insertBefore(this, newChild, ref)
    };

    XMLInlineElementNode.prototype.insertAt = function insertAt$$1 (pos, child) {
      return insertAt(this, pos, child)
    };

    XMLInlineElementNode.prototype.removeAt = function removeAt$$1 (pos) {
      return removeAt(this, pos)
    };

    XMLInlineElementNode.prototype.getInnerXML = function getInnerXML$$1 () {
      return this.getChildren().map(function (child) {
        return child.toXML().outerHTML
      }).join('')
    };

    Object.defineProperties( XMLInlineElementNode.prototype, prototypeAccessors );

    return XMLInlineElementNode;
  }(XMLAnnotationNode));

  XMLInlineElementNode.prototype._elementType = 'inline-element';



  XMLInlineElementNode.prototype._isInlineNode = true;
  XMLInlineElementNode.isInline = true;


  XMLInlineElementNode.type = 'inline-element';

  XMLInlineElementNode.schema = {
    _childNodes: { type: ['array', 'id'], default: [], owned: true},
  };

  var XMLExternalNode = (function (XMLDocumentNode$$1) {
    function XMLExternalNode () {
      XMLDocumentNode$$1.apply(this, arguments);
    }if ( XMLDocumentNode$$1 ) XMLExternalNode.__proto__ = XMLDocumentNode$$1;
    XMLExternalNode.prototype = Object.create( XMLDocumentNode$$1 && XMLDocumentNode$$1.prototype );
    XMLExternalNode.prototype.constructor = XMLExternalNode;

    

    return XMLExternalNode;
  }(XMLDocumentNode));

  XMLExternalNode.prototype._elementType = 'external';

  XMLExternalNode.type = 'external';

  XMLExternalNode.schema = {
    xml: { type: 'string', default: ''}
  };

  XMLExternalNode.isBlock = true;

  var ExternalNodeConverter = (function (XMLNodeConverter$$1) {
    function ExternalNodeConverter () {
      XMLNodeConverter$$1.apply(this, arguments);
    }

    if ( XMLNodeConverter$$1 ) ExternalNodeConverter.__proto__ = XMLNodeConverter$$1;
    ExternalNodeConverter.prototype = Object.create( XMLNodeConverter$$1 && XMLNodeConverter$$1.prototype );
    ExternalNodeConverter.prototype.constructor = ExternalNodeConverter;

    ExternalNodeConverter.prototype.import = function import$1 (el, node) {
      node.xml = el.innerHTML;
    };

    ExternalNodeConverter.prototype.export = function export$1 (node, el) {
      el.tagName = this.tagNameNS;
      el.setAttributes(node.attributes);
      el.innerHTML = node.xml;
    };

    return ExternalNodeConverter;
  }(XMLNodeConverter));

  var XMLContainerNode = (function (superclass) {
    function XMLContainerNode () {
      superclass.apply(this, arguments);
    }

    if ( superclass ) XMLContainerNode.__proto__ = superclass;
    XMLContainerNode.prototype = Object.create( superclass && superclass.prototype );
    XMLContainerNode.prototype.constructor = XMLContainerNode;

    XMLContainerNode.prototype.getContentPath = function getContentPath () {
      return [this.id, '_childNodes']
    };

    XMLContainerNode.prototype.getContent = function getContent () {
      return this._childNodes
    };

    XMLContainerNode.prototype.isContainer = function isContainer () {
      return true
    };

    XMLContainerNode.prototype.appendChild = function appendChild (child) {
      superclass.prototype.show.call(this, child.id);
    };

    return XMLContainerNode;
  }(ContainerMixin(XMLElementNode)));

  XMLContainerNode.prototype._elementType = 'container';

  XMLContainerNode.type = 'container';

  XMLContainerNode.schema = {};

  XMLContainerNode.isBlock = true;

  function createCountingIdGenerator() {
    var counters = {};
    return function uuid(prefix) {
      if (!counters.hasOwnProperty(prefix)) {
        counters[prefix] = 1;
      }
      var result = [prefix, '-', counters[prefix]++].join('');
      return result
    }
  }

  var WS_LEFT = /^\s+/g;


  var WS_RIGHT = /\s+$/g;
  var WS_ALL = /\s+/g;


  var SPACE = " ";
  var TABS_OR_NL = /[\t\n\r]+/g;

  var INVISIBLE_CHARACTER = "\u200B";


  var DOMImporter = function DOMImporter(config, context) {
    this.context = context || {};

    if (!config.schema) {
      throw new Error('"config.schema" is mandatory')
    }
    if (!config.converters) {
      throw new Error('"config.converters" is mandatory')
    }

    this.config = Object.assign({ idAttribute: 'id' }, config);
    this.schema = config.schema;
    this.converters = config.converters;
    this.state = null;

    this._defaultBlockConverter = null;
    this._allConverters = [];
    this._blockConverters = [];
    this._propertyAnnotationConverters = [];

    this.state = new DOMImporter.State();

    this._initialize();
  };


  DOMImporter.prototype._initialize = function _initialize () {
      var this$1 = this;

    var schema = this.schema;
    var defaultTextType = schema.getDefaultTextType();
    var converters = this.converters;
    for (var i = 0; i < converters.length; i++) {
      var converter = (void 0);
      if (typeof converters[i] === 'function') {
        var Converter = converters[i];
        converter = new Converter();
      } else {
        converter = converters[i];
      }
      if (!converter.type) {
        throw new Error('Converter must provide the type of the associated node.')
      }
      if (!converter.matchElement && !converter.tagName) {
        throw new Error('Converter must provide a matchElement function or a tagName property.')
      }
      if (!converter.matchElement) {
        converter.matchElement = this$1._defaultElementMatcher.bind(converter);
      }
      var NodeClass = schema.getNodeClass(converter.type);
      if (!NodeClass) {
        throw new Error('No node type defined for converter')
      }
      if (!this$1._defaultBlockConverter && defaultTextType === converter.type) {
        this$1._defaultBlockConverter = converter;
      }
      this$1._allConverters.push(converter);
      
      if (NodeClass.prototype._isPropertyAnnotation) {
        this$1._propertyAnnotationConverters.push(converter);
      } else {
        this$1._blockConverters.push(converter);
      }
    }
  };

  DOMImporter.prototype.dispose = function dispose () {
    if (this.state.doc) {
      this.state.doc.dispose();
    }
  };


  DOMImporter.prototype.reset = function reset () {
    if (this.state.doc) {
      this.state.doc.dispose();
    }
    this.state.reset();
    this.state.doc = this._createDocument();
  };

  DOMImporter.prototype.getDocument = function getDocument () {
    return this.state.doc
  };


  DOMImporter.prototype.convertContainer = function convertContainer (elements, containerId) {
      var this$1 = this;

    if (!this.state.doc) { this.reset(); }
    var state = this.state;
    var iterator = new ArrayIterator(elements);
    var nodeIds = [];
    while(iterator.hasNext()) {
      var el = iterator.next();
      var node = (void 0);
      var blockTypeConverter = this$1._getConverterForElement(el, 'block');
      if (blockTypeConverter) {
        state.pushContext(el.tagName, blockTypeConverter);
        var nodeData = this$1._createNodeData(el, blockTypeConverter.type);
        nodeData = blockTypeConverter.import(el, nodeData, this$1) || nodeData;
        node = this$1._createNode(nodeData);
        var context = state.popContext();
        context.annos.forEach(function (a) {
          this$1._createNode(a);
        });
      } else if (el.isCommentNode()) {
        continue
      } else {
        
        if (el.isTextNode() && /^\s*$/.exec(el.textContent)) { continue }
        
        
        iterator.back();
        node = this$1._wrapInlineElementsIntoBlockElement(iterator);
      }
      if (node) {
        nodeIds.push(node.id);
      }
    }
    return this._createNode({
      type: 'container',
      id: containerId,
      nodes: nodeIds
    })
  };


  DOMImporter.prototype.convertElement = function convertElement (el) {
      var this$1 = this;

    if (!this.state.doc) { this.reset(); }
    var isTopLevel = !this.state.isConverting;
    if (isTopLevel) {
      this.state.isConverting = true;
    }

    var nodeData, annos;
    var converter = this._getConverterForElement(el);
    if (converter) {
      var NodeClass = this.schema.getNodeClass(converter.type);
      nodeData = this._createNodeData(el, converter.type);
      this.state.pushContext(el.tagName, converter);
      
      
      
      
      
      if (NodeClass.isInline) {
        nodeData = this._convertInlineNode(el, nodeData, converter);
      }
      else if (NodeClass.prototype._isPropertyAnnotation) {
        nodeData = this._convertPropertyAnnotation(el, nodeData);
      } else {
        nodeData = converter.import(el, nodeData, this) || nodeData;
      }
      var context = this.state.popContext();
      annos = context.annos;
    } else {
      throw new Error('No converter found for '+el.tagName)
    }
    
    var node = this._createNode(nodeData);
    
    annos.forEach(function (a) {
      this$1._createNode(a);
    });

    
    
    if (this.config["stand-alone"] && isTopLevel) {
      this.state.isConverting = false;
      this.reset();
    }
    return node
  };


  DOMImporter.prototype.annotatedText = function annotatedText (el, path, options) {
      if ( options === void 0 ) options={};

    if (!path) {
      throw new Error('path is mandatory')
    }
    var state = this.state;
    var context = last$1(state.contexts);
    
    
    if (!context) {
      throw new Error('This should be called from within an element converter.')
    }
    
    var oldPreserveWhitespace = state.preserveWhitespace;
    if (options.preserveWhitespace) {
      state.preserveWhitespace = true;
    }
    state.stack.push({ path: path, offset: 0, text: "", annos: []});
    
    
    
    this.state.lastChar = '';
    var iterator = this.getChildNodeIterator(el);
    var text = this._annotatedText(iterator);
    
    
    var top = state.stack.pop();
    context.annos = context.annos.concat(top.annos);

    
    state.preserveWhitespace = oldPreserveWhitespace;

    return text
  };


  DOMImporter.prototype.plainText = function plainText (el) {
    var state = this.state;
    var text = el.textContent;
    if (state.stack.length > 0) {
      var context = last$1(state.stack);
      context.offset += text.length;
      context.text += context.text.concat(text);
    }
    return text
  };


  DOMImporter.prototype._customText = function _customText (text) {
    var state = this.state;
    if (state.stack.length > 0) {
      var context = last$1(state.stack);
      context.offset += text.length;
      context.text += context.text.concat(text);
    }
    return text
  };


  DOMImporter.prototype.nextId = function nextId (prefix) {
    
    
    
    
    return this.state.uuid(prefix)
  };

  DOMImporter.prototype._getNextId = function _getNextId (dom, type) {
      var this$1 = this;

    var id = this.nextId(type);
    while (this.state.ids[id] || dom.find('#'+id)) {
      id = this$1.nextId(type);
    }
    return id
  };

  DOMImporter.prototype._getIdForElement = function _getIdForElement (el, type) {
    var id = el.getAttribute(this.config.idAttribute);
    if (id && !this.state.ids[id]) { return id }
    return this._getNextId(el.getOwnerDocument(), type)
  };




  DOMImporter.prototype._createDocument = function _createDocument () {
    
    var schema = this.config.schema;
    var DocumentClass = schema.getDocumentClass();
    return new DocumentClass(schema)
  };

  DOMImporter.prototype._convertPropertyAnnotation = function _convertPropertyAnnotation (el, nodeData) {
    var path = [nodeData.id, '_content'];
    
    
    
    nodeData._content = this.annotatedText(el, path);
    nodeData.start = { path: path, offset: 0 };
    nodeData.end = { offset: nodeData._content.length };
    return nodeData
  };

  DOMImporter.prototype._convertInlineNode = function _convertInlineNode (el, nodeData, converter) {
    var path = [nodeData.id, '_content'];
    if (converter.import) {
      nodeData = converter.import(el, nodeData, this) || nodeData;
    }
    nodeData._content = '$';
    nodeData.start = { path: path, offset: 0 };
    nodeData.end = { offset: 1 };
    return nodeData
  };

  DOMImporter.prototype._createNodeData = function _createNodeData (el, type) {
    if (!type) {
      throw new Error('type is mandatory.')
    }
    var nodeData = {
      type: type,
      id: this._getIdForElement(el, type)
    };
    this.state.ids[nodeData.id] = true;
    return nodeData
  };

  DOMImporter.prototype._createNode = function _createNode (nodeData) {
    var doc = this.state.doc;
    
    
    
    var node = doc.get(nodeData.id);
    if (node) {
      
      doc.delete(node.id);
    }
    return doc.create(nodeData)
  };

  DOMImporter.prototype.getChildNodeIterator = function getChildNodeIterator (el) {
    return el.getChildNodeIterator()
  };

  DOMImporter.prototype._defaultElementMatcher = function _defaultElementMatcher (el) {
    return el.is(this.tagName)
  };


  DOMImporter.prototype._annotatedText = function _annotatedText (iterator) {
      var this$1 = this;

    var state = this.state;
    var context = last$1(state.stack);
    
    if (!context) {
      throw new Error('Illegal state: context is null.')
    }
    while(iterator.hasNext()) {
      var el = iterator.next();
      var text = "";
      
      
      if (el.isTextNode()) {
        text = this$1._prepareText(el.textContent);
        if (text.length) {
          
          
          context.text = context.text.concat(text);
          context.offset += text.length;
        }
      } else if (el.isCommentNode()) {
        
        continue
      } else if (el.isElementNode()) {
        var annoConverter = this$1._getConverterForElement(el, 'inline');
        
        if (!annoConverter) {
          
          if (!this$1.IGNORE_DEFAULT_WARNINGS) {
            console.warn('Unsupported inline element. We will not create an annotation for it, but process its children to extract annotated text.', el.outerHTML);
          }
          
          
          var iterator$1 = this$1.getChildNodeIterator(el);
          this$1._annotatedText(iterator$1);
          continue
        }
        
        
        
        var startOffset = context.offset;
        var annoType = annoConverter.type;
        var AnnoClass = this$1.schema.getNodeClass(annoType);
        var annoData = this$1._createNodeData(el, annoType);
        
        var stackFrame = {
          path: context.path,
          offset: startOffset,
          text: "",
          annos: []
        };
        state.stack.push(stackFrame);
        
        if (annoConverter.import) {
          state.pushContext(el.tagName, annoConverter);
          annoData = annoConverter.import(el, annoData, this$1) || annoData;
          state.popContext();
        }
        
        
        
        
        if (AnnoClass.isInline) {
          this$1._customText(INVISIBLE_CHARACTER);
          
          
          
          state.lastChar = '';
        } else {
          
          
          
          var iterator$2 = this$1.getChildNodeIterator(el);
          this$1._annotatedText(iterator$2);
        }
        
        state.stack.pop();
        context.offset = stackFrame.offset;
        context.text = context.text.concat(stackFrame.text);
        
        var endOffset = context.offset;
        annoData.start = {
          path: context.path.slice(0),
          offset: startOffset
        };
        annoData.end = {
          offset: endOffset
        };
        
        var parentFrame = last$1(state.stack);
        parentFrame.annos = parentFrame.annos.concat(stackFrame.annos, annoData);
      } else {
        console.warn('Unknown element type. Taking plain text.', el.outerHTML);
        text = this$1._prepareText(el.textContent);
        context.text = context.text.concat(text);
        context.offset += text.length;
      }
    }
    
    return context.text
  };

  DOMImporter.prototype._getConverterForElement = function _getConverterForElement (el, mode) {
      var this$1 = this;

    var converters;
    if (mode === "block") {
      if (!el.tagName) { return null }
      converters = this._blockConverters;
    } else if (mode === "inline") {
      converters = this._propertyAnnotationConverters;
    } else {
      converters = this._allConverters;
    }
    var converter = null;
    for (var i = 0; i < converters.length; i++) {
      if (this$1._converterCanBeApplied(converters[i], el)) {
        converter = converters[i];
        break
      }
    }
    return converter
  };

  DOMImporter.prototype._converterCanBeApplied = function _converterCanBeApplied (converter, el) {
    return converter.matchElement(el, this)
  };


  DOMImporter.prototype._wrapInlineElementsIntoBlockElement = function _wrapInlineElementsIntoBlockElement (childIterator) {
      var this$1 = this;

    if (!childIterator.hasNext()) { return }

    var converter = this._defaultBlockConverter;
    if (!converter) {
      throw new Error('Wrapping inline elements automatically is not supported in this schema.')
    }

    var dom = childIterator.peek().getOwnerDocument();
    var wrapper = dom.createElement('wrapper');
    while(childIterator.hasNext()) {
      var el = childIterator.next();
      
      var blockTypeConverter = this$1._getConverterForElement(el, 'block');
      if (blockTypeConverter) {
        childIterator.back();
        break
      }
      wrapper.append(el.clone());
    }
    var type = this.schema.getDefaultTextType();
    var id = this._getNextId(dom, type);
    var nodeData = { type: type, id: id };
    this.state.pushContext('wrapper', converter);
    nodeData = converter.import(wrapper, nodeData, this) || nodeData;
    var context = this.state.popContext();
    var annos = context.annos;
    
    var node = this._createNode(nodeData);
    
    annos.forEach(function (a) {
      this$1._createNode(a);
    });
    return node
  };





  DOMImporter.prototype._prepareText = function _prepareText (text) {
    var state = this.state;
    if (state.preserveWhitespace) {
      return text
    }
    var repl = SPACE;
    
    text = text.replace(TABS_OR_NL, '');
    
    
    
    
    
    if (state.lastChar === SPACE) {
      
      text = text.replace(WS_LEFT, '');
    } else {
      text = text.replace(WS_LEFT, repl);
    }
    text = text.replace(WS_RIGHT, repl);
    
    
    
    if (this.config.REMOVE_INNER_WS || state.removeInnerWhitespace) {
      text = text.replace(WS_ALL, SPACE);
    }
    state.lastChar = text[text.length-1] || state.lastChar;
    return text
  };


  DOMImporter.prototype._trimTextContent = function _trimTextContent (el) {
    var nodes = el.getChildNodes();
    var firstNode = nodes[0];
    var lastNode = last$1(nodes);
    var text, trimmed;
    
    if (firstNode && firstNode.isTextNode()) {
      text = firstNode.textContent;
      trimmed = this._trimLeft(text);
      firstNode.textContent = trimmed;
    }
    if (lastNode && lastNode.isTextNode()) {
      text = lastNode.textContent;
      trimmed = this._trimRight(text);
      lastNode.textContent = trimmed;
    }
    return el
  };

  DOMImporter.prototype._trimLeft = function _trimLeft (text) {
    return text.replace(WS_LEFT, "")
  };

  DOMImporter.prototype._trimRight = function _trimRight (text) {
    return text.replace(WS_RIGHT, "")
  };

  var DOMImporterState = function DOMImporterState() {
    this.reset();
  };

  DOMImporterState.prototype.reset = function reset () {
    this.preserveWhitespace = false;
    this.nodes = [];
    this.annotations = [];
    this.containerId = null;
    this.container = [];
    this.ids = {};
    
    this.contexts = [];
    
    this.stack = [];
    this.lastChar = "";
    this.skipTypes = {};
    this.ignoreAnnotations = false;
    this.isConverting = false;

    
    
    this.uuid = createCountingIdGenerator();
  };

  DOMImporterState.prototype.pushContext = function pushContext (tagName, converter) {
    this.contexts.push({ tagName: tagName, converter: converter, annos: []});
  };

  DOMImporterState.prototype.popContext = function popContext () {
    return this.contexts.pop()
  };

  DOMImporterState.prototype.getCurrentContext = function getCurrentContext () {
    return last$1(this.contexts)
  };

  DOMImporter.State = DOMImporterState;

  DOMImporter.INVISIBLE_CHARACTER = INVISIBLE_CHARACTER;

  var START = 'START';
  var END = 'END';
  var EPSILON = 'EPSILON';
  var TEXT$1 = 'TEXT';

  var DFA = function DFA(transitions) {
    if (!transitions || Object.keys(transitions).length === 0) {
      transitions = { START: { EPSILON: END } };
    }
    this.transitions = transitions;
  };

  DFA.prototype.consume = function consume (state, id) {
    var T = this.transitions;
    
    
    if (!T[state]) { return -1 }
    var nextState = T[state][id];
    if (nextState !== undefined) {
      return nextState
    }
    while(T[state][EPSILON] !== undefined) {
      state = T[state][EPSILON];
      if (state === END) {
        return -1
      }
      nextState = T[state][id];
      if (nextState !== undefined) {
        return nextState
      }
    }
    return -1
  };

  DFA.prototype.canConsume = function canConsume (state, id) {
    var nextState = this.consume(state, id);
    return (nextState !== -1)
  };

  DFA.prototype.isFinished = function isFinished (state) {
    var T = this.transitions;
    if (state === 'END') { return true }
    
    if (!T[state]) { return false }
    while(T[state][EPSILON] !== undefined) {
      state = T[state][EPSILON];
      if (state === 'END') { return true }
    }
    return false
  };




  DFA.prototype._tokensByPath = function _tokensByPath () {
    var result = [];
    var transitions = this.transitions;
    if (!transitions) { return [] }

    
    var first = {};
    forEach(transitions[START], function (to, token) {
      if (!first[to]) { first[to] = []; }
      first[to].push(token);
    });

    var visited = {START: true, END: true};
    forEach(first, function (tokens, state) {
      
      
      
      
      var _siblings = {};
      tokens.forEach(function (t) {
        if (t !== EPSILON) {
          _siblings[t] = true;
        }
      });
      var stack = [state];
      while(stack.length > 0) {
        var from = stack.pop();
        if (state === END) { continue }
        visited[from] = true;
        var T = transitions[from];
        if (!T) { throw new Error(("Internal Error: no transition from state " + from)) }
        var tokens$1 = Object.keys(T);
        for (var i = 0; i < tokens$1.length; i++) {
          var token = tokens$1[i];
          var to = T[token];
          if (!visited[to]) { stack.push(to); }
          if (token !== EPSILON) {
            _siblings[token] = true;
          }
        }
      }
      var _siblingTokens = Object.keys(_siblings);
      if (_siblingTokens.length > 0) {
        result.push(_siblingTokens);
      }
    });
    return result
  };

  DFA.START = START;
  DFA.END = END;
  DFA.EPSILON = EPSILON;
  DFA.TEXT = TEXT$1;

  function _isTextNodeEmpty(el) {
    return Boolean(/^\s*$/.exec(el.textContent))
  }

  var TEXT$2 = DFA.TEXT;

  var ValidatingChildNodeIterator = function ValidatingChildNodeIterator(el, it, expr) {
    this.el = el;
    this.it = it;
    this.expr = expr;
    this.state = expr.getInitialState();
    this._oldStates = [];
  };

  ValidatingChildNodeIterator.prototype.hasNext = function hasNext () {
    return this.it.hasNext()
  };

  ValidatingChildNodeIterator.prototype.next = function next () {
    var state = this.state;
    var expr = this.expr;
    var next = this.it.next();
    var oldState = cloneDeep(this.state);
    var ok;
    if (next.isTextNode()) {
      ok = expr.consume(state, TEXT$2);
    } else if (next.isElementNode()) {
      ok = expr.consume(state, next.tagName);
    }
    if (!ok) {
      if (next.isTextNode()) {
        if (!_isTextNodeEmpty(next)) {
          console.error(("TEXT is invalid within <" + (expr.name) + ">. Skipping."), next.textContent);
        }
      } else if (next.isElementNode()) {
        var error = last$1(state.errors);
        console.error(error.msg, this.el.getNativeElement());
      }
      
      this.state = oldState;
      return next.createComment(next.outerHTML)
    } else {
      this._oldStates.push(oldState);
      return next
    }
  };

  ValidatingChildNodeIterator.prototype.back = function back () {
    this.it.back();
    this.state = this._oldStates.pop();
    return this
  };

  ValidatingChildNodeIterator.prototype.peek = function peek () {
    return this.it.peek()
  };

  var XMLDocumentImporter = (function (DOMImporter$$1) {
    function XMLDocumentImporter(config, context) {
      
      DOMImporter$$1.call(this, {
        
        schema: config.schema,
        
        converters: config.converters,
        idAttribute: config.schema.xmlSchema.getIdAttribute(),
      }, context);
      this.xmlSchema = config.schema.xmlSchema;
    }

    if ( DOMImporter$$1 ) XMLDocumentImporter.__proto__ = DOMImporter$$1;
    XMLDocumentImporter.prototype = Object.create( DOMImporter$$1 && DOMImporter$$1.prototype );
    XMLDocumentImporter.prototype.constructor = XMLDocumentImporter;

    XMLDocumentImporter.prototype.importDocument = function importDocument (dom) {
      this.reset();
      var doc = this.state.doc;
      if (isString(dom)) {
        dom = DefaultDOMElement.parseXML(dom);
      }
      var startTag = this.xmlSchema.getStartElement();
      var rootEl;
      if (dom.is(startTag)) {
        rootEl = dom;
      } else {
        rootEl = dom.find(startTag);
      }
      if (!rootEl) { throw new Error(("Could not find <" + startTag + "> element.")) }
      
      
      doc.root = this.convertElement(rootEl);
      return this.state.doc
    };

    XMLDocumentImporter.prototype._initialize = function _initialize () {
      var this$1 = this;

      var schema = this.schema;
      var defaultTextType = schema.getDefaultTextType();
      var converters = this.converters;

      this._allConverters = [];
      this._propertyAnnotationConverters = [];
      this._blockConverters = [];

      for (var i = 0; i < converters.length; i++) {
        var converter = (void 0);
        if (typeof converters[i] === 'function') {
          var Converter = converters[i];
          converter = new Converter();
        } else {
          converter = converters[i];
        }
        if (!converter.type) {
          throw new Error('Converter must provide the type of the associated node.')
        }
        if (!converter.matchElement && !converter.tagName) {
          throw new Error('Converter must provide a matchElement function or a tagName property.')
        }
        if (!converter.matchElement) {
          converter.matchElement = this$1._defaultElementMatcher.bind(converter);
        }
        var NodeClass = schema.getNodeClass(converter.type);
        if (!NodeClass) {
          throw new Error('No node type defined for converter')
        }
        if (!this$1._defaultBlockConverter && defaultTextType === converter.type) {
          this$1._defaultBlockConverter = converter;
        }

        
        if (NodeClass.prototype._isAnnotation) {
          this$1._propertyAnnotationConverters.push(converter);
        } else {
          this$1._blockConverters.push(converter);
        }
      }
      this._allConverters = this._blockConverters.concat(this._propertyAnnotationConverters);
    };

    XMLDocumentImporter.prototype._createNodeData = function _createNodeData (el, type) {
      var nodeData = DOMImporter$$1.prototype._createNodeData.call(this, el, type);
      var attributes = {};
      el.getAttributes().forEach(function (value, key) {
        attributes[key] = value;
      });
      nodeData.attributes = attributes;
      return nodeData
    };

    XMLDocumentImporter.prototype.getChildNodeIterator = function getChildNodeIterator (el) {
      
      var schema = this.xmlSchema.getElementSchema(el.tagName);
      var it = el.getChildNodeIterator();
      return new ValidatingChildNodeIterator(el, it, schema.expr)
    };

    XMLDocumentImporter.prototype._convertPropertyAnnotation = function _convertPropertyAnnotation () {
      throw new Error('stand-alone annotations are not supported.')
    };

    XMLDocumentImporter.prototype._convertInlineNode = function _convertInlineNode (el, nodeData, converter) {
      var path = [];
      if (converter.import) {
        nodeData = converter.import(el, nodeData, this) || nodeData;
      }
      nodeData.start = { path: path, offset: 0 };
      nodeData.end = { offset: 0 };
      return nodeData
    };

    return XMLDocumentImporter;
  }(DOMImporter));

  function registerSchema(config, xmlSchema, DocumentClass, options) {
    if ( options === void 0 ) options = {};

    var schemaName = xmlSchema.getName();
    var defaultTextType;
    
    
    if (xmlSchema.getDefaultTextType) {
      defaultTextType = xmlSchema.getDefaultTextType();
    }
    
    config.defineSchema({
      name: schemaName,
      version: xmlSchema.getVersion(),
      defaultTextType: defaultTextType,
      DocumentClass: DocumentClass,
      
      xmlSchema: xmlSchema
    });
    var tagNames = xmlSchema.getTagNames();
    
    tagNames.forEach(function (tagName) {
      var elementSchema = xmlSchema.getElementSchema(tagName);
      var name = elementSchema.name;
      var NodeClass, ConverterClass;
      switch (elementSchema.type) {
        case 'element': {
          NodeClass = XMLElementNode;
          ConverterClass = ElementNodeConverter;
          break
        }
        case 'hybrid': {
          throw new Error('Mixed element types are not supported yet.')
        }
        case 'text': {
          NodeClass = XMLTextElement;
          ConverterClass = XMLTextElementConverter;
          break
        }
        case 'annotation': {
          NodeClass = XMLAnnotationNode;
          ConverterClass = XMLNodeConverter;
          break
        }
        case 'anchor': {
          NodeClass = XMLAnchorNode;
          ConverterClass = XMLNodeConverter;
          break
        }
        case 'inline-element': {
          NodeClass = XMLInlineElementNode;
          
          ConverterClass = ElementNodeConverter;
          break
        }
        case 'external': {
          NodeClass = XMLExternalNode;
          ConverterClass = ExternalNodeConverter;
          break
        }
        case 'container': {
          NodeClass = XMLContainerNode;
          ConverterClass = ElementNodeConverter;
          break
        }
        default:
          throw new Error('Illegal state')
      }
      
      var Node = (function (NodeClass) {
        function Node () {
          NodeClass.apply(this, arguments);
        }if ( NodeClass ) Node.__proto__ = NodeClass;
        Node.prototype = Object.create( NodeClass && NodeClass.prototype );
        Node.prototype.constructor = Node;

        

        return Node;
      }(NodeClass));
      Node.type = name;

      
      var attributes = elementSchema.attributes;
      forEach(attributes, function (spec, name) {
        _defineAttribute(Node, name, spec);
      });

      config.addNode(Node);
      var converter = new ConverterClass(name);
      config.addConverter(schemaName, converter);

      var ImporterClass = options.ImporterClass || XMLDocumentImporter;
      config.addImporter(schemaName, ImporterClass);
    });
  }

  var BUILTIN_ATTRS = ['id', 'type', 'attributes', '_childNodes', '_content'];

  function _defineAttribute(Node, attributeName) {
    var name = attributeName.replace(':', '_');
    name = camelCase(name);
    if (BUILTIN_ATTRS.indexOf(name) >= 0) {
      
      return
    }
    Object.defineProperty(Node.prototype, name, {
      get: function get() {
        return this.getAttribute(attributeName)
      },
      set: function set(val) {
        this.setAttribute(attributeName, val);
        return this
      }
    });
  }

  var START$1 = DFA.START;
  var END$1 = DFA.END;
  var EPSILON$1 = DFA.EPSILON;


  var DFABuilder = function DFABuilder(transitions) {
    this.transitions = transitions;
  };

  DFABuilder.prototype.addTransition = function addTransition (from, to, tokens) {
      var this$1 = this;

    if (!this.transitions) { this.transitions = {}; }
    if (!isArray(tokens)) { tokens = [tokens]; }
    tokens.forEach(function (token) { return _addTransition(this$1.transitions, from, to, token); });
    return this
  };


  DFABuilder.prototype.append = function append (other) {
    if (this.transitions && other.transitions) {
      var t1 = cloneDeep(this.transitions);
      var t2 = cloneDeep(other.transitions);
      
      
      var firstIsOptional = Boolean(t1[START$1][EPSILON$1]);
      var secondIsOptional = Boolean(t2[START$1][EPSILON$1]);

      if (firstIsOptional) {
        
        
        delete t1[START$1][EPSILON$1];
      }
      
      
      var newState = uuid();
      
      
      forEach(t1, function (T) {
        forEach(T, function (to, token) {
          if (to === END$1) {
            T[token] = newState;
          }
        });
      });
      
      
      if (firstIsOptional) {
        forEach(t2[START$1], function (to, token) {
          _addTransition(t1, START$1, to, token);
        });
      }
      
      
      t2[newState] = t2[START$1];
      forEach(t2, function (T) {
        forEach(T, function (to, token) {
          if (to === START$1) {
            T[token] = newState;
          }
        });
      });
      delete t2[START$1];
      
      forEach(t2, function (T, from) {
        forEach(T, function (to, token) {
          _addTransition(t1, from, to, token);
        });
      });
      
      
      if (firstIsOptional && secondIsOptional) {
        _addTransition(t1, START$1, END$1, EPSILON$1);
      }
      this.transitions = t1;
    } else if (other.transitions) {
      this.transitions = cloneDeep(other.transitions);
    }
    return this
  };


  DFABuilder.prototype.merge = function merge (other) {
    if (this.transitions && other.transitions) {
      var t1 = this.transitions;
      var t2 = other.transitions;
      forEach(t2, function (T, from) {
        forEach(T, function (to, token) {
          _addTransition(t1, from, to, token);
        });
      });
    } else if (other.transitions) {
      this.transitions = cloneDeep(other.transitions);
    }
    return this
  };


  DFABuilder.prototype.optional = function optional () {
    var dfa = new DFABuilder(cloneDeep(this.transitions));
    if (this.transitions) {
      dfa.addTransition(START$1, END$1, EPSILON$1);
    }
    return dfa
  };


  DFABuilder.prototype.kleene = function kleene () {
    var dfa = this.plus();
    return dfa.optional()
  };


  DFABuilder.prototype.plus = function plus () {
    var dfa;
    if (this.transitions) {
      var t1 = cloneDeep(this.transitions);
      
      
      
      
      var isOptional = Boolean(t1[START$1][EPSILON$1]);
      delete t1[START$1][EPSILON$1];
      
      
      var newState = uuid();
      forEach(t1, function (T) {
        forEach(T, function (to, token) {
          if (to === END$1) {
            T[token] = newState;
          }
        });
      });
      
      _addTransition(t1, newState, END$1, EPSILON$1);
      
      forEach(t1[START$1], function (to, token) {
        _addTransition(t1, newState, to, token);
      });
      
      if (isOptional) {
        _addTransition(t1, START$1, END$1, EPSILON$1);
      }
      dfa = new DFABuilder(t1);
    } else {
      dfa = new DFABuilder(cloneDeep(this.transitions));
    }
    return dfa
  };

  DFABuilder.prototype.toJSON = function toJSON () {
    return cloneDeep(this.transitions)
  };


  DFABuilder.prototype.copy = function copy () {
    var t = cloneDeep(this.transitions);
    if (this.transitions) {
      var states = Object.keys(t);
      var map = { START: START$1, END: END$1 };
      states.forEach(function (id) {
        if (id === START$1 || id === END$1) { return }
        map[id] = uuid();
      });
      forEach(t, function (T, from) {
        if (from !== START$1 && from !== END$1) {
          t[map[from]] = T;
          delete t[from];
        }
        forEach(T, function (to, token) {
          if (to !== START$1 && to !== END$1) {
            T[token] = map[to];
          }
        });
      });
    }
    return new DFABuilder(t)
  };

  DFABuilder.singleToken = function(token) {
    var dfa = new DFABuilder();
    dfa.addTransition(START$1, END$1, token);
    return dfa
  };

  function _addTransition(transitions, from, to, token) {
    var T = transitions[from];
    if (!T) {
      transitions[from] = T = {};
    }
    if (token === EPSILON$1 && from === START$1 && to !== END$1) {
      throw new Error('The only EPSILON transition from START must be START->END')
    }
    if (T[token] && T[token] !== to) {
      console.error('Token %s already used. Ignoring this transition.', token);
      return
      
    }
    T[token] = to;
  }

  var START$2 = DFA.START;
  var END$2 = DFA.END;
  var TEXT$3 = DFA.TEXT;
  var EPSILON$2 = DFA.EPSILON;



  var Expression = function Expression(name, root) {
    this.name = name;
    this.root = root;

    this._initialize();
  };

  Expression.prototype._initialize = function _initialize () {
    this._compile();
  };

  Expression.prototype.toString = function toString () {
    return this.root.toString()
  };

  Expression.prototype.copy = function copy () {
    return this.root.copy()
  };

  Expression.prototype.toJSON = function toJSON () {
    return {
      name: this.name,
      content: this.root.toJSON()
    }
  };

  Expression.prototype.isAllowed = function isAllowed (tagName) {
    return Boolean(this._allowedChildren[tagName])
  };


  Expression.prototype._normalize = function _normalize () {
    this.root._normalize();
  };


  Expression.prototype._compile = function _compile () {
    
    this.root._compile();
  };

  Expression.prototype._describeError = function _describeError (state, token) {
    var msg = [];
    if (token !== TEXT$3) {
      if (!this.isAllowed(token)) {
        msg.push(("<" + token + "> is not valid in <" + (this.name) + ">\nSchema: " + (this.toString())));
      } else {
        
        msg.push(("<" + token + "> is not allowed at the current position in <" + (this.name) + ">.\n" + (this.toString())));
        
        
        
      }
    } else {
      msg.push(("TEXT is not allowed at the current position: " + (state.trace.join(',')) + "\n" + (this.toString())));
    }
    return msg.join('')
  };

  Expression.fromJSON = function(data) {
    var name = data.name;
    var root = _fromJSON(data.content);
    return createExpression(name, root)
  };

  function createExpression(name, root) {
    if (root instanceof Interleave) {
      return new InterleaveExpr(name, root)
    } else {
      return new DFAExpr(name, root)
    }
  }

  var DFAExpr = (function (Expression) {
    function DFAExpr () {
      Expression.apply(this, arguments);
    }

    if ( Expression ) DFAExpr.__proto__ = Expression;
    DFAExpr.prototype = Object.create( Expression && Expression.prototype );
    DFAExpr.prototype.constructor = DFAExpr;

    DFAExpr.prototype.getInitialState = function getInitialState () {
      return {
        dfaState: START$2,
        errors: [],
        trace: [],
      }
    };

    DFAExpr.prototype.canConsume = function canConsume (state, token) {
      return this.dfa.canConsume(state.dfaState, token)
    };

    DFAExpr.prototype.consume = function consume (state, token) {
      var dfa = this.dfa;
      var oldState = state.dfaState;
      var newState = dfa.consume(oldState, token);
      state.dfaState = newState;
      if (newState === -1) {
        state.errors.push({
          msg: this._describeError(state, token),
          
          
          el: state.el
        });
        return false
      } else {
        state.trace.push(token);
        return true
      }
    };

    DFAExpr.prototype.isFinished = function isFinished (state) {
      return this.dfa.isFinished(state.dfaState)
    };

    DFAExpr.prototype._initialize = function _initialize () {
      Expression.prototype._initialize.call(this);

      this._computeAllowedChildren();
    };

    DFAExpr.prototype._compile = function _compile () {
      Expression.prototype._compile.call(this);
      this.dfa = new DFA(this.root.dfa.transitions);
    };


    DFAExpr.prototype._computeAllowedChildren = function _computeAllowedChildren () {
      this._allowedChildren = _collectAllTokensFromDFA(this.dfa);
    };

    DFAExpr.prototype._findInsertPos = function _findInsertPos (el, newTag, mode) {
      var root = this.root;
      if (root instanceof Sequence) {
        return this._findInsertPosInSequence(el, newTag, mode)
      } else if (root instanceof Plus || root instanceof Kleene) {
        if (mode === 'first') {
          return 0
        } else {
          return el.childNodes.length
        }
      }
    };

    DFAExpr.prototype._isValid = function _isValid (_tokens) {
      var this$1 = this;

      var state = this.getInitialState();
      for (var i = 0; i < _tokens.length; i++) {
        var token = _tokens[i];
        
        
        
        if (!token) { continue }
        if (!this$1.consume(state, token)) {
          return false
        }
      }
      return this.isFinished(state)
    };

    DFAExpr.prototype._findInsertPosInSequence = function _findInsertPosInSequence (el, newTag, mode) {
      var childNodes = el.getChildNodes();
      
      
      
      var tokens = [];
      childNodes.forEach(function (child){
        
        
        
        var tagName = child.tagName;
        if (!tagName) {
          if (child._isDOMElement && child.isTextNode() && !_isTextNodeEmpty(child)) {
            tokens.push(TEXT$3);
          } else {
            tokens.push(null);
          }
        } else {
          tokens.push(tagName);
        }
      });
      var L = tokens.length;
      var self = this;
      function _isValid(pos) {
        var _tokens = tokens.slice(0);
        _tokens.splice(pos, 0, newTag);
        return self._isValid(_tokens)
      }
      if (mode === 'first') {
        for (var pos = 0; pos <= L; pos++) {
          if (_isValid(pos)) {
            return pos
          }
        }
      } else {
        for (var pos$1 = L; pos$1 >= 0; pos$1--) {
          if (_isValid(pos$1)) {
            return pos$1
          }
        }
      }
      return -1
    };

    return DFAExpr;
  }(Expression));

  function _collectAllTokensFromDFA(dfa) {
    
    var children = {};
    if (dfa.transitions) {
      forEach(dfa.transitions, function (T) {
        Object.keys(T).forEach(function (tagName) {
          if (tagName === EPSILON$2) { return }
          children[tagName] = true;
        });
      });
    }
    return children
  }

  var InterleaveExpr = (function (Expression) {
    function InterleaveExpr(name, root) {
      Expression.call(this, name, root);
    }

    if ( Expression ) InterleaveExpr.__proto__ = Expression;
    InterleaveExpr.prototype = Object.create( Expression && Expression.prototype );
    InterleaveExpr.prototype.constructor = InterleaveExpr;

    InterleaveExpr.prototype.getInitialState = function getInitialState () {
      var dfas = this.dfas;
      var dfaStates = new Array(dfas.length);
      dfaStates.fill(START$2);
      return {
        dfaStates: dfaStates,
        errors: [],
        trace: [],
        
        lastDFA: 0,
      }
    };

    InterleaveExpr.prototype.canConsume = function canConsume (state, token) {
      return (this._findNextDFA(state, token) >= 0)
    };

    InterleaveExpr.prototype.consume = function consume (state, token) {
      var idx = this._findNextDFA(state, token);
      if (idx < 0) {
        state.errors.push({
          msg: this._describeError(state, token),
        });
        return false
      } else {
        var dfa = this.dfas[idx];
        var oldState = state.dfaStates[idx];
        var newState = dfa.consume(oldState, token);
        state.dfaStates[idx] = newState;
        state.trace.push(token);
        return true
      }
    };

    InterleaveExpr.prototype.isFinished = function isFinished (state) {
      var dfas = this.dfas;
      for (var i = 0; i < dfas.length; i++) {
        var dfa = dfas[i];
        var dfaState = state.dfaStates[i];
        if (!dfa.isFinished(dfaState)) {
          return false
        }
      }
      return true
    };


    InterleaveExpr.prototype._initialize = function _initialize () {
      Expression.prototype._initialize.call(this);

      this._computeAllowedChildren();
    };

    InterleaveExpr.prototype._compile = function _compile () {
      Expression.prototype._compile.call(this);

      this.blocks = this.root.blocks;
      this.dfas = this.blocks.map(function (b){ return new DFA(b.dfa.transitions); });
    };

    InterleaveExpr.prototype._computeAllowedChildren = function _computeAllowedChildren () {
      this._allowedChildren = Object.assign.apply(Object, this.blocks.map(function (block) {
        return _collectAllTokensFromDFA(block.dfa)
      }));
    };

    InterleaveExpr.prototype._findNextDFA = function _findNextDFA (state, token) {
      console.assert(state.dfaStates.length === this.dfas.length);
      var dfas = this.dfas;
      for (var i = 0; i < state.dfaStates.length; i++) {
        var dfa = dfas[i];
        var dfaState = state.dfaStates[i];
        if (dfa.canConsume(dfaState, token)) {
          return i
        }
      }
      return -1
    };

    InterleaveExpr.prototype._findInsertPos = function _findInsertPos (el, newTag, mode) { 
      
      return el.childNodes.length
    };

    return InterleaveExpr;
  }(Expression));

  var Token = function Token(name) {
    this.name = name;
  };

  Token.prototype.toString = function toString () {
    return this.name
  };

  Token.prototype.toJSON = function toJSON () {
    return this.name
  };

  Token.prototype.copy = function copy () {
    return new Token(this.name)
  };

  Token.prototype._normalize = function _normalize () {};

  Token.prototype._compile = function _compile () {
    this.dfa = DFABuilder.singleToken(this.name);
  };

  Token.fromJSON = function (data) {
    return new Token(data)
  };


  var Choice = function Choice(blocks) {
    this.blocks = blocks;
  };

  Choice.prototype.copy = function copy () {
    return new Choice(this.blocks.map(function (b){ return b.copy(); }))
  };

  Choice.prototype.toJSON = function toJSON () {
    return {
      type: '|',
      blocks: this.blocks.map(function (b){ return b.toJSON(); })
    }
  };

  Choice.prototype._normalize = function _normalize () {
    var blocks = this.blocks;
    for (var i = blocks.length - 1; i >= 0; i--) {
      var block = blocks[i];
      block._normalize();
      
      if (block instanceof Choice) {
        blocks.splice.apply(blocks, [ i,1 ].concat( (block.blocks) ));
      }
    }
  };


  Choice.prototype._compile = function _compile () {
    var dfa = new DFABuilder();
    this.blocks.forEach(function (block) {
      if (block instanceof Token) {
        dfa.addTransition(START$2, END$2, block.name);
      } else if (block instanceof Interleave) {
        throw new Error('Nested interleave blocks are not supported.')
      } else {
        if (!block.dfa) {
          block._compile();
        }
        dfa.merge(block.dfa);
      }
    });
    this.dfa = dfa;
    return dfa
  };

  Choice.prototype.toString = function toString () {
    return '('+ this.blocks.map(function (b){ return b.toString(); }).join('|') + ')'
  };

  Choice.fromJSON = function(data) {
    return new Choice(data.blocks.map(function (block) {
      return _fromJSON(block)
    }))
  };


  var Sequence = function Sequence(blocks) {
    this.blocks = blocks;
  };

  Sequence.prototype.copy = function copy () {
    return new Sequence(this.blocks.map(function (b){ return b.copy(); }))
  };

  Sequence.prototype.toJSON = function toJSON () {
    return {
      type: ',',
      blocks: this.blocks.map(function (b){ return b.toJSON(); })
    }
  };

  Sequence.prototype._compile = function _compile () {
    var dfa = new DFABuilder();
    this.blocks.forEach(function (block) {
      if (block instanceof Token) {
        dfa.append(DFABuilder.singleToken(block.name));
      } else if (block instanceof Interleave) {
        throw new Error('Nested interleave blocks are not supported.')
      } else {
        if (!block.dfa) {
          block._compile();
        }
        dfa.append(block.dfa);
      }
    });
    this.dfa = dfa;
    return dfa
  };

  Sequence.prototype._normalize = function _normalize () {
    var blocks = this.blocks;
    for (var i = blocks.length - 1; i >= 0; i--) {
      var block = blocks[i];
      block._normalize();
      
      if (block instanceof Sequence) {
        blocks.splice.apply(blocks, [ i, 1 ].concat( (block.blocks) ));
      }
    }
  };

  Sequence.prototype.toString = function toString () {
    return '('+ this.blocks.map(function (b){ return b.toString(); }).join(',') + ')'
  };

  Sequence.fromJSON = function(data) {
    return new Sequence(data.blocks.map(function (block) {
      return _fromJSON(block)
    }))
  };


  var Interleave = function Interleave(blocks) {
    this.blocks = blocks;
  };

  Interleave.prototype.copy = function copy () {
    return new Interleave(this.blocks.map(function (b){ return b.copy(); }))
  };

  Interleave.prototype.toString = function toString () {
    return '('+ this.blocks.map(function (b){ return b.toString(); }).join('~') + ')'
  };

  Interleave.prototype.toJSON = function toJSON () {
    return {
      type: '~',
      blocks: this.blocks.map(function (b){ return b.toJSON(); })
    }
  };

  Interleave.prototype._normalize = function _normalize () {
    
  };

  Interleave.prototype._compile = function _compile () {
    this.blocks.forEach(function (block) { return block._compile(); });
  };

  Interleave.fromJSON = function(data) {
    return new Interleave(data.blocks.map(function (block) {
      return _fromJSON(block)
    }))
  };



  var Optional = function Optional(block) {
    this.block = block;
  };

  Optional.prototype.copy = function copy () {
    return new Optional(this.block.copy())
  };

  Optional.prototype.toJSON = function toJSON () {
    return {
      type: '?',
      block: this.block.toJSON()
    }
  };

  Optional.prototype._compile = function _compile () {
    var block = this.block;
    if (block instanceof Interleave) {
      throw new Error('Nested interleave blocks are not supported.')
    }
    if (!block.dfa) {
      block._compile();
    }
    this.dfa = block.dfa.optional();
    return this.dfa
  };

  Optional.prototype._normalize = function _normalize () {
    var block = this.block;
    block._normalize();
    if (block instanceof Optional) {
      this.block = block.block;
    } else if (block instanceof Kleene) {
      console.error('FIXME -  <optional> is useless here', this.toString());
    }
  };

  Optional.prototype.toString = function toString () {
    return this.block.toString() + '?'
  };

  Optional.fromJSON = function(data) {
    return new Optional(_fromJSON(data.block))
  };



  var Kleene = function Kleene(block) {
    this.block = block;
  };

  Kleene.prototype.copy = function copy () {
    return new Kleene(this.block.copy())
  };

  Kleene.prototype.toJSON = function toJSON () {
    return {
      type: '*',
      block: this.block.toJSON()
    }
  };

  Kleene.prototype._compile = function _compile () {
    var block = this.block;
    if (block instanceof Interleave) {
      throw new Error('Nested interleave blocks are not supported.')
    }
    if (!block.dfa) {
      block._compile();
    }
    this.dfa = block.dfa.kleene();
    return this.dfa
  };

  Kleene.prototype._normalize = function _normalize () {
    var block = this.block;
    block._normalize();
    if (block instanceof Optional || block instanceof Kleene) {
      this.block = block.block;
    } else if (block instanceof Plus) {
      throw new Error('This does not make sense:' + this.toString())
    }
  };

  Kleene.prototype.toString = function toString () {
    return this.block.toString() + '*'
  };

  Kleene.fromJSON = function(data) {
    return new Kleene(_fromJSON(data.block))
  };


  var Plus = function Plus(block) {
    this.block = block;
  };

  Plus.prototype.copy = function copy () {
    return new Plus(this.block.copy())
  };

  Plus.prototype.toJSON = function toJSON () {
    return {
      type: '+',
      block: this.block.toJSON()
    }
  };

  Plus.prototype._compile = function _compile () {
    var block = this.block;
    if (block instanceof Interleave) {
      throw new Error('Nested interleave blocks are not supported.')
    }
    if (!block.dfa) {
      block._compile();
    }
    this.dfa = block.dfa.plus();
    return this.dfa
  };

  Plus.prototype._normalize = function _normalize () {
    var block = this.block;
    block._normalize();
    if (block instanceof Optional || block instanceof Kleene) {
      throw new Error('This does not make sense:' + this.toString())
    } else if (block instanceof Plus) {
      this.block = block.block;
    }
  };

  Plus.prototype.toString = function toString () {
    return this.block.toString() + '+'
  };

  Plus.fromJSON = function(data) {
    return new Plus(_fromJSON(data.block))
  };

  function _fromJSON(data) {
    switch(data.type) {
      case ',':
        return Sequence.fromJSON(data)
      case '~':
        return Interleave.fromJSON(data)
      case '|':
        return Choice.fromJSON(data)
      case '?':
        return Optional.fromJSON(data)
      case '+':
        return Plus.fromJSON(data)
      case '*':
        return Kleene.fromJSON(data)
      default:
        if (isString(data)) {
          return new Token(data)
        }
        throw new Error('Unsupported data.')
    }
  }

  var TEXT$4 = DFA.TEXT;

  var XMLSchema = function XMLSchema(elementSchemas, startElement) {
    var this$1 = this;

    if (!elementSchemas[startElement]) {
      throw new Error('startElement must be a valid element.')
    }
    this._elementSchemas = {};
    this.startElement = startElement;
    
    forEach(elementSchemas, function (spec, name) {
      this$1._elementSchemas[name] = new ElementSchema(spec.name, spec.type, spec.attributes, spec.expr);
    });
  };

  XMLSchema.prototype.getIdAttribute = function getIdAttribute () {
    return 'id'
  };

  XMLSchema.prototype.getTagNames = function getTagNames () {
    return Object.keys(this._elementSchemas)
  };

  XMLSchema.prototype.getElementSchema = function getElementSchema (name) {
    return this._elementSchemas[name]
  };

  XMLSchema.prototype.getStartElement = function getStartElement () {
    return this.startElement
  };

  XMLSchema.prototype.toJSON = function toJSON () {
    var result = {
      start: this.getStartElement(),
      elements: {}
    };
    forEach(this._elementSchemas, function (schema, name) {
      result.elements[name] = schema.toJSON();
    });
    return result
  };


  XMLSchema.prototype.toMD = function toMD () {
      var this$1 = this;

    var result = [];
    var elementNames = Object.keys(this._elementSchemas);
    elementNames.sort();
    elementNames.forEach(function (name) {
      var elementSchema = this$1._elementSchemas[name];
      result.push(("# <" + (elementSchema.name) + ">"));
      result.push('');
      result.push(("type: " + (elementSchema.type)));
      result.push('attributes: '+ map(elementSchema.attributes, function (_, name) { return name }).join(', '));
      result.push('children:');
      result.push('  '+elementSchema.expr.toString());
      result.push('');
    });
    return result.join('\n')
  };

  XMLSchema.prototype.validateElement = function validateElement (el) {
    var tagName = el.tagName;
    var elementSchema = this.getElementSchema(tagName);
    return _validateElement(elementSchema, el)
  };

  XMLSchema.fromJSON = function(data) {
    var elementSchemas = {};
    forEach(data.elements, function (elData) {
      var elSchema = ElementSchema.fromJSON(elData);
      elementSchemas[elSchema.name] = elSchema;
    });
    return new XMLSchema(elementSchemas, data.start)
  };

  var ElementSchema = function ElementSchema(name, type, attributes, expr) {
    this.name = name;
    this.type = type;
    this.attributes = attributes;
    this.expr = expr;

    if (!name) {
      throw new Error("'name' is mandatory")
    }
    if (!type) {
      throw new Error("'type' is mandatory")
    }
    if (!attributes) {
      throw new Error("'attributes' is mandatory")
    }
    if (!expr) {
      throw new Error("'expr' is mandatory")
    }
  };

  ElementSchema.prototype.toJSON = function toJSON () {
    return {
      name: this.name,
      type: this.type,
      attributes: this.attributes,
      elements: this.expr.toJSON()
    }
  };

  ElementSchema.prototype.isAllowed = function isAllowed (tagName) {
    return this.expr.isAllowed(tagName)
  };

  ElementSchema.prototype.isTextAllowed = function isTextAllowed () {
    return this.expr.isAllowed(TEXT$4)
  };

  ElementSchema.prototype.printStructure = function printStructure () {
    return ((this.name) + " ::= " + (this.expr.toString()))
  };

  ElementSchema.prototype.findFirstValidPos = function findFirstValidPos (el, newTag) {
    return this.expr._findInsertPos(el, newTag, 'first')
  };

  ElementSchema.prototype.findLastValidPos = function findLastValidPos (el, newTag) {
    return this.expr._findInsertPos(el, newTag, 'last')
  };

  ElementSchema.fromJSON = function(data) {
    return new ElementSchema(
      data.name,
      data.type,
      data.attributes,
      Expression.fromJSON(data.elements)
    )
  };

  function _validateElement(elementSchema, el) {
    var errors = [];
    var valid = true;
    { 
      var res = _checkAttributes(elementSchema, el);
      if (!res.ok) {
        errors = errors.concat(res.errors);
        valid = false;
      }
    }
    { 
      if (elementSchema.type === 'external') ; else {
        
        var res$1;
        if (el._isXMLTextElement) {
          res$1 = _checkChildren(elementSchema, el.toXML());
        } else {
          res$1 = _checkChildren(elementSchema, el);
        }
        if (!res$1.ok) {
          errors = errors.concat(res$1.errors);
          valid = false;
        }
      }
    }
    return {
      errors: errors,
      ok: valid
    }
  }

  function _checkAttributes(elementSchema, el) { 
    return { ok: true }
  }

  function _checkChildren(elementSchema, el) {
    
    
    if (elementSchema.type === 'external') {
      return true
    }
    var expr = elementSchema.expr;
    var state = expr.getInitialState();
    var iterator = el.getChildNodeIterator();
    var valid = true;
    while (valid && iterator.hasNext()) {
      var childEl = iterator.next();
      var token = (void 0);
      if (childEl.isTextNode()) {
        
        if (elementSchema.type !== 'text' && _isTextNodeEmpty(childEl)) {
          continue
        }
        token = TEXT$4;
      } else if (childEl.isElementNode()) {
        token = childEl.tagName;
      } else {
        continue
      }
      if (!expr.consume(state, token)) {
        valid = false;
      }
    }
    
    if (state.errors.length > 0) {
      state.errors.forEach(function (err) {
        err.el = el;
      });
    }
    if (valid && !expr.isFinished(state)) {
      state.errors.push({
        msg: ("<" + (el.tagName) + "> is incomplete.\nSchema: " + (expr.toString())),
        el: el
      });
      valid = false;
    }
    if (valid) {
      state.ok = true;
    }
    return state
  }

  var ManifestSchemaData = {"start":"dar","elements":{"dar":{"name":"dar","type":"element","attributes":{},"elements":{"name":"dar","content":{"type":",","blocks":["documents","assets"]}}},"documents":{"name":"documents","type":"element","attributes":{},"elements":{"name":"documents","content":{"type":"*","block":"document"}}},"assets":{"name":"assets","type":"element","attributes":{},"elements":{"name":"assets","content":{"type":"*","block":"asset"}}},"document":{"name":"document","type":"element","attributes":{"id":{"name":"id"},"name":{"name":"name"},"type":{"name":"type"},"path":{"name":"path"}},"elements":{"name":"document","content":{"type":",","blocks":[]}}},"asset":{"name":"asset","type":"element","attributes":{"id":{"name":"id"},"name":{"name":"name"},"type":{"name":"type"},"path":{"name":"path"}},"elements":{"name":"asset","content":{"type":",","blocks":[]}}}}}

  var ManifestSchema = XMLSchema.fromJSON(ManifestSchemaData);


  ManifestSchema.getName = function() {
    return 'RDC-Manifest'
  };

  ManifestSchema.getVersion = function() {
    return '1.0'
  };

  ManifestSchema.getDocTypeParams = function() {
    return ['manifest', 'RDC-Manifest 1.0', ManifestSchema.uri]
  };


  ManifestSchema.getDefaultTextType = function () {
    return 'text'
  };

  ManifestSchema.uri = '//Manifest-1.0.dtd';

  var Data = (function (EventEmitter$$1) {
    function Data(schema, nodeFactory) {
      EventEmitter$$1.call(this);

      
      if (!schema) {
        throw new Error('schema is mandatory')
      }
      if (!nodeFactory) {
        throw new Error('nodeFactory is mandatory')
      }
      

      this.schema = schema;
      this.nodeFactory = nodeFactory;
      this.nodes = {};
      this.indexes = {};

      
      
      this.__QUEUE_INDEXING__ = false;
      this.queue = [];
    }

    if ( EventEmitter$$1 ) Data.__proto__ = EventEmitter$$1;
    Data.prototype = Object.create( EventEmitter$$1 && EventEmitter$$1.prototype );
    Data.prototype.constructor = Data;

    
    Data.prototype.contains = function contains (id) {
      return Boolean(this.nodes[id])
    };

    
    Data.prototype.get = function get (path, strict) {
      var result = this._get(path);
      if (strict && result === undefined) {
        if (isString(path)) {
          throw new Error("Could not find node with id '"+path+"'.")
        } else if (!this.contains(path[0])) {
          throw new Error("Could not find node with id '"+path[0]+"'.")
        } else {
          throw new Error("Property for path '"+path+"' us undefined.")
        }
      }
      return result
    };

    Data.prototype._get = function _get (path) {
      if (!path) { return undefined }
      var result;
      if (isString(path)) {
        result = this.nodes[path];
      } else if (path.length === 1) {
        result = this.nodes[path[0]];
      } else if (path.length > 1) {
        var context = this.nodes[path[0]];
        for (var i = 1; i < path.length-1; i++) {
          if (!context) { return undefined }
          context = context[path[i]];
        }
        if (!context) { return undefined }
        result = context[path[path.length-1]];
      }
      return result
    };

    
    Data.prototype.getNodes = function getNodes () {
      return this.nodes
    };

    
    Data.prototype.create = function create (nodeData) {
      var node = this.nodeFactory.create(nodeData.type, nodeData);
      if (!node) {
        throw new Error('Illegal argument: could not create node for data:', nodeData)
      }
      if (this.contains(node.id)) {
        throw new Error("Node already exists: " + node.id)
      }
      if (!node.id || !node.type) {
        throw new Error("Node id and type are mandatory.")
      }
      this.nodes[node.id] = node;

      var change = {
        type: 'create',
        node: node
      };

      if (this.__QUEUE_INDEXING__) {
        this.queue.push(change);
      } else {
        this._updateIndexes(change);
      }

      return node
    };

    
    Data.prototype.delete = function delete$1 (nodeId) {
      var node = this.nodes[nodeId];
      if (!node) { return }
      node.dispose();
      delete this.nodes[nodeId];

      var change = {
        type: 'delete',
        node: node,
      };

      if (this.__QUEUE_INDEXING__) {
        this.queue.push(change);
      } else {
        this._updateIndexes(change);
      }

      return node
    };

    
    Data.prototype.set = function set (path, newValue) {
      var node = this.get(path[0]);
      var oldValue = this._set(path, newValue);
      var change = {
        type: 'set',
        node: node,
        path: path,
        newValue: newValue,
        oldValue: oldValue
      };
      if (this.__QUEUE_INDEXING__) {
        this.queue.push(change);
      } else {
        this._updateIndexes(change);
      }
      return oldValue
    };

    Data.prototype._set = function _set (path, newValue) {
      var oldValue = _setValue(this.nodes, path, newValue);
      return oldValue
    };

    
    Data.prototype.update = function update (path, diff) {
      var realPath = this.getRealPath(path);
      if (!realPath) {
        console.error('Could not resolve path', path);
        return
      }
      var node = this.get(realPath[0]);
      var oldValue = this._get(realPath);
      var newValue;
      if (diff.isOperation) {
        newValue = diff.apply(oldValue);
      } else {
        diff = this._normalizeDiff(oldValue, diff);
        if (isString(oldValue)) {
          switch (diff.type) {
            case 'delete': {
              newValue = oldValue.split('').splice(diff.start, diff.end-diff.start).join('');
              break
            }
            case 'insert': {
              newValue = [oldValue.substring(0, diff.start), diff.text, oldValue.substring(diff.start)].join('');
              break
            }
            default:
              throw new Error('Unknown diff type')
          }
        } else if (isArray(oldValue)) {
          newValue = oldValue.slice(0);
          switch (diff.type) {
            case 'delete': {
              newValue.splice(diff.pos, 1);
              break
            }
            case 'insert': {
              newValue.splice(diff.pos, 0, diff.value);
              break
            }
            default:
              throw new Error('Unknown diff type')
          }
        } else if (oldValue._isCoordinate) {
          switch (diff.type) {
            case 'shift': {
              
              oldValue = { path: oldValue.path, offset: oldValue.offset };
              newValue = oldValue;
              newValue.offset += diff.value;
              break
            }
            default:
              throw new Error('Unknown diff type')
          }
        } else {
          throw new Error('Diff is not supported:', JSON.stringify(diff))
        }
      }
      this._set(realPath, newValue);

      var change = {
        type: 'update',
        node: node,
        path: realPath,
        newValue: newValue,
        oldValue: oldValue
      };

      if (this.__QUEUE_INDEXING__) {
        this.queue.push(change);
      } else {
        this._updateIndexes(change);
      }

      return oldValue
    };

    
    Data.prototype._normalizeDiff = function _normalizeDiff (value, diff) {
      if (isString(value)) {
        
        if (diff['delete']) {
          console.warn('DEPRECATED: use doc.update(path, {type:"delete", start:s, end: e}) instead');
          diff = {
            type: 'delete',
            start: diff['delete'].start,
            end: diff['delete'].end
          };
        } else if (diff['insert']) {
          console.warn('DEPRECATED: use doc.update(path, {type:"insert", start:s, text: t}) instead');
          diff = {
            type: 'insert',
            start: diff['insert'].offset,
            text: diff['insert'].value
          };
        }
      } else if (isArray(value)) {
        
        if (diff['delete']) {
          console.warn('DEPRECATED: use doc.update(path, {type:"delete", pos:1}) instead');
          diff = {
            type: 'delete',
            pos: diff['delete'].offset
          };
        } else if (diff['insert']) {
          console.warn('DEPRECATED: use doc.update(path, {type:"insert", pos:1, value: "foo"}) instead');
          diff = {
            type: 'insert',
            pos: diff['insert'].offset,
            value: diff['insert'].value
          };
        }
      } else if (value._isCoordinate) {
        if (diff.hasOwnProperty('shift')) {
          console.warn('DEPRECATED: use doc.update(path, {type:"shift", value:2}) instead');
          diff = {
            type: 'shift',
            value: diff['shift']
          };
        }
      }
      return diff
    };

    
    Data.prototype.toJSON = function toJSON () {
      var nodes = {};
      forEach(this.nodes, function (node){
        nodes[node.id] = node.toJSON();
      });
      return {
        schema: [this.schema.id, this.schema.version],
        nodes: nodes
      }
    };

    Data.prototype.reset = function reset () {
      this.clear();
    };

    
    Data.prototype.clear = function clear () {
      this.nodes = {};
      forEach(this.indexes, function (index) { return index.clear(); });
    };

    
    Data.prototype.addIndex = function addIndex (name, index) {
      if (this.indexes[name]) {
        console.error('Index with name %s already exists.', name);
      }
      index.reset(this);
      this.indexes[name] = index;
      return index
    };

    
    Data.prototype.getIndex = function getIndex (name) {
      return this.indexes[name]
    };

    
    Data.prototype._updateIndexes = function _updateIndexes (change) {
      if (!change || this.__QUEUE_INDEXING__) { return }
      forEach(this.indexes, function(index) {
        if (index.select(change.node)) {
          if (!index[change.type]) {
            console.error('Contract: every NodeIndex must implement ' + change.type);
          }
          index[change.type](change.node, change.path, change.newValue, change.oldValue);
        }
      });
    };

    
    Data.prototype._stopIndexing = function _stopIndexing () {
      this.__QUEUE_INDEXING__ = true;
    };

    
    Data.prototype._startIndexing = function _startIndexing () {
      var this$1 = this;

      this.__QUEUE_INDEXING__ = false;
      while(this.queue.length >0) {
        var change = this$1.queue.shift();
        this$1._updateIndexes(change);
      }
    };

    return Data;
  }(EventEmitter));

  function _setValue(root, path, newValue) {
    var ctx = root;
    var L = path.length;
    for (var i = 0; i < L-1; i++) {
      ctx = ctx[path[i]];
      if (!ctx) { throw new Error('Can not set value.') }
    }
    var oldValue = ctx[path[L-1]];
    ctx[path[L-1]] = newValue;
    return oldValue
  }

  var IncrementalData = (function (Data$$1) {
    function IncrementalData () {
      Data$$1.apply(this, arguments);
    }

    if ( Data$$1 ) IncrementalData.__proto__ = Data$$1;
    IncrementalData.prototype = Object.create( Data$$1 && Data$$1.prototype );
    IncrementalData.prototype.constructor = IncrementalData;

    IncrementalData.prototype.create = function create (nodeData) {
      if (nodeData._isNode) {
        nodeData = nodeData.toJSON();
      }
      var op = ObjectOperation.Create([nodeData.id], nodeData);
      this.apply(op);
      return op
    };

    
    IncrementalData.prototype.delete = function delete$1 (nodeId) {
      var op = null;
      var node = this.get(nodeId);
      if (node) {
        var nodeData = node.toJSON();
        op = ObjectOperation.Delete([nodeId], nodeData);
        this.apply(op);
      }
      return op
    };

    
    IncrementalData.prototype.update = function update (path, diff) {
      var diffOp = this._getDiffOp(path, diff);
      var op = ObjectOperation.Update(path, diffOp);
      this.apply(op);
      return op
    };

    
    IncrementalData.prototype.set = function set (path, newValue) {
      var oldValue = this.get(path);
      var op = ObjectOperation.Set(path, oldValue, newValue);
      this.apply(op);
      return op
    };

    
    IncrementalData.prototype.apply = function apply (op) {
      if (op.type === ObjectOperation.NOP) { return }
      else if (op.type === ObjectOperation.CREATE) {
        
        Data$$1.prototype.create.call(this, cloneDeep(op.val));
      } else if (op.type === ObjectOperation.DELETE) {
        Data$$1.prototype.delete.call(this, op.val.id);
      } else if (op.type === ObjectOperation.UPDATE) {
        var oldVal = this.get(op.path);
        var diff = op.diff;
        if (op.propertyType === 'array') {
          if (! (diff._isArrayOperation) ) {
            diff = ArrayOperation.fromJSON(diff);
          }
          
          diff.apply(oldVal);
        } else if (op.propertyType === 'string') {
          if (!(diff._isTextOperation) ) {
            diff = TextOperation.fromJSON(diff);
          }
          var newVal = diff.apply(oldVal);
          Data$$1.prototype.set.call(this, op.path, newVal);
        } else if (op.propertyType === 'coordinate') {
          if (!(diff._isCoordinateOperation) ) {
            diff = CoordinateOperation.fromJSON(diff);
          }
          diff.apply(oldVal);
        } else {
          throw new Error("Unsupported type for operational update.")
        }
      } else if (op.type === ObjectOperation.SET) {
        Data$$1.prototype.set.call(this, op.path, op.val);
      } else {
        throw new Error("Illegal state.")
      }
      this.emit('operation:applied', op, this);
    };

    
    IncrementalData.prototype._getDiffOp = function _getDiffOp (path, diff) {
      var diffOp = null;
      if (diff.isOperation) {
        diffOp = diff;
      } else {
        var value = this.get(path);
        diff = this._normalizeDiff(value, diff);
        if (value === null || value === undefined) {
          throw new Error('Property has not been initialized: ' + JSON.stringify(path))
        } else if (isString(value)) {
          switch (diff.type) {
            case 'delete': {
              diffOp = TextOperation.Delete(diff.start, value.substring(diff.start, diff.end));
              break
            }
            case 'insert': {
              diffOp = TextOperation.Insert(diff.start, diff.text);
              break
            }
            default:
              throw new Error('Unknown diff type')
          }
        } else if (isArray(value)) {
          switch (diff.type) {
            case 'delete': {
              diffOp = ArrayOperation.Delete(diff.pos, value[diff.pos]);
              break
            }
            case 'insert': {
              diffOp = ArrayOperation.Insert(diff.pos, diff.value);
              break
            }
            default:
              throw new Error('Unknown diff type')
          }
        } else if (value._isCoordinate) {
          switch (diff.type) {
            case 'shift': {
              diffOp = CoordinateOperation.Shift(diff.value);
              break
            }
            default:
              throw new Error('Unknown diff type')
          }
        }
      }
      if (!diffOp) {
        throw new Error('Unsupported diff: ' + JSON.stringify(diff))
      }
      return diffOp
    };

    return IncrementalData;
  }(Data));

  var PropertyIndex = (function (NodeIndex$$1) {
    function PropertyIndex(property) {
      NodeIndex$$1.call(this);

      this._property = property || 'id';
      this.index = new TreeIndex();
    }

    if ( NodeIndex$$1 ) PropertyIndex.__proto__ = NodeIndex$$1;
    PropertyIndex.prototype = Object.create( NodeIndex$$1 && NodeIndex$$1.prototype );
    PropertyIndex.prototype.constructor = PropertyIndex;

    
    PropertyIndex.prototype.get = function get (path) {
      return this.index.get(path) || {}
    };

    
    PropertyIndex.prototype.getAll = function getAll (path) {
      return this.index.getAll(path)
    };

    PropertyIndex.prototype.clear = function clear () {
      this.index.clear();
    };

    
    PropertyIndex.prototype.select = function select (node) { 
      return true
    };

    
    PropertyIndex.prototype.create = function create (node) {
      var values = node[this._property];
      if (!isArray(values)) {
        values = [values];
      }
      forEach(values, function(value) {
        this.index.set([value, node.id], node);
      }.bind(this));
    };

    
    PropertyIndex.prototype.delete = function delete$1 (node) {
      var values = node[this._property];
      if (!isArray(values)) {
        values = [values];
      }
      forEach(values, function(value) {
        this.index.delete([value, node.id]);
      }.bind(this));
    };

    
    PropertyIndex.prototype.update = function update (node, path, newValue, oldValue) {
      if (!this.select(node) || path[1] !== this._property) { return }
      var values = oldValue;
      if (!isArray(values)) {
        values = [values];
      }
      forEach(values, function(value) {
        this.index.delete([value, node.id]);
      }.bind(this));
      values = newValue;
      if (!isArray(values)) {
        values = [values];
      }
      forEach(values, function(value) {
        this.index.set([value, node.id], node);
      }.bind(this));
    };

    PropertyIndex.prototype.set = function set (node, path, newValue, oldValue) {
      this.update(node, path, newValue, oldValue);
    };

    PropertyIndex.prototype._initialize = function _initialize (data) {
      forEach(data.getNodes(), function(node) {
        if (this.select(node)) {
          this.create(node);
        }
      }.bind(this));
    };

    return PropertyIndex;
  }(NodeIndex));

  var AnnotationIndex = (function (DocumentIndex$$1) {
    function AnnotationIndex() {
      DocumentIndex$$1.call(this);

      this.byPath = new TreeIndex();
      this.byType = new TreeIndex();
    }

    if ( DocumentIndex$$1 ) AnnotationIndex.__proto__ = DocumentIndex$$1;
    AnnotationIndex.prototype = Object.create( DocumentIndex$$1 && DocumentIndex$$1.prototype );
    AnnotationIndex.prototype.constructor = AnnotationIndex;

    AnnotationIndex.prototype.select = function select (node) {
      return node.isPropertyAnnotation()
    };

    AnnotationIndex.prototype.clear = function clear () {
      this.byPath.clear();
      this.byType.clear();
    };

    
    AnnotationIndex.prototype.get = function get (path, start, end, type) {
      var annotations;
      if (isString(path) || path.length === 1) {
        annotations = this.byPath.getAll(path) || {};
      } else {
        annotations = this.byPath.get(path);
      }
      annotations = map(annotations);
      if (isNumber(start)) {
        annotations = filter(annotations, AnnotationIndex.filterByRange(start, end));
      }
      if (type) {
        annotations = filter(annotations, DocumentIndex$$1.filterByType(type));
      }
      return annotations
    };

    AnnotationIndex.prototype.create = function create (anno) {
      var path = anno.start.path;
      this.byType.set([anno.type, anno.id], anno);
      if (path && path.length > 0) {
        this.byPath.set(anno.start.path.concat([anno.id]), anno);
      }
    };

    AnnotationIndex.prototype.delete = function delete$1 (anno) {
      this._delete(anno.type, anno.id, anno.start.path);
    };

    AnnotationIndex.prototype._delete = function _delete (type, id, path) {
      this.byType.delete([type, id]);
      if (path && path.length > 0) {
        this.byPath.delete(path.concat([id]));
      }
    };

    AnnotationIndex.prototype.update = function update (node, path, newValue, oldValue) {
      
      if (this.select(node) && path[1] === 'start' && path[2] === "path") {
        this._delete(node.type, node.id, oldValue);
        this.create(node);
      }
    };

    return AnnotationIndex;
  }(DocumentIndex));

  AnnotationIndex.filterByRange = function(start, end) {
    return function(anno) {
      var aStart = anno.start.offset;
      var aEnd = anno.end.offset;
      var overlap = (aEnd >= start);
      
      if (isNumber(end)) {
        overlap = overlap && (aStart <= end);
      }
      return overlap
    }
  };

  var ContainerAnnotationIndex = (function (DocumentIndex$$1) {
    function ContainerAnnotationIndex() {
      DocumentIndex$$1.call(this);
      this.byId = new TreeIndex();
    }

    if ( DocumentIndex$$1 ) ContainerAnnotationIndex.__proto__ = DocumentIndex$$1;
    ContainerAnnotationIndex.prototype = Object.create( DocumentIndex$$1 && DocumentIndex$$1.prototype );
    ContainerAnnotationIndex.prototype.constructor = ContainerAnnotationIndex;

    ContainerAnnotationIndex.prototype.select = function select (node) {
      return node.isContainerAnnotation()
    };

    ContainerAnnotationIndex.prototype.clear = function clear () {
      this.byId.clear();
    };

    ContainerAnnotationIndex.prototype.get = function get (containerId, type) {
      var annotations = map(this.byId.get(containerId));
      if (isString(type)) {
        annotations = filter(annotations, DocumentIndex$$1.filterByType);
      }
      return annotations
    };

    ContainerAnnotationIndex.prototype.create = function create (anno) {
      this.byId.set([anno.containerId, anno.id], anno);
    };

    ContainerAnnotationIndex.prototype.delete = function delete$1 (anno) {
      this.byId.delete([anno.containerId, anno.id]);
    };

    ContainerAnnotationIndex.prototype.update = function update (node, path, newValue, oldValue) { 
      
    };

    return ContainerAnnotationIndex;
  }(DocumentIndex));

  var DocumentNodeFactory = function DocumentNodeFactory(doc) {
    this.doc = doc;
  };

  DocumentNodeFactory.prototype.create = function create (nodeType, nodeData) {
    var NodeClass = this.doc.schema.getNodeClass(nodeType);
    if (!NodeClass) {
      throw new Error('No node registered by that name: ' + nodeType)
    }
    return new NodeClass(this.doc, nodeData)
  };

  function copySelection(doc, selection) {
    if (!selection) { throw new Error("'selection' is mandatory.") }
    var copy = null;
    if (!selection.isNull() && !selection.isCollapsed()) {
      
      if (selection.isPropertySelection()) {
        copy = _copyPropertySelection(doc, selection);
      }
      else if (selection.isContainerSelection()) {
        copy = _copyContainerSelection(doc, selection);
      }
      else if (selection.isNodeSelection()) {
        copy = _copyNodeSelection(doc, selection);
      }
      else {
        console.error('Copy is not yet supported for selection type.');
      }
    }
    return copy
  }

  function _copyPropertySelection(doc, selection) {
    var path = selection.start.path;
    var offset = selection.start.offset;
    var endOffset = selection.end.offset;
    var text = doc.get(path);
    var snippet = doc.createSnippet();
    var containerNode = snippet.getContainer();
    snippet.create({
      type: doc.schema.getDefaultTextType(),
      id: TEXT_SNIPPET_ID,
      content: text.substring(offset, endOffset)
    });
    containerNode.show(TEXT_SNIPPET_ID);
    var annotations = doc.getIndex('annotations').get(path, offset, endOffset);
    forEach(annotations, function(anno) {
      var data = cloneDeep(anno.toJSON());
      var path = [TEXT_SNIPPET_ID, 'content'];
      data.start = {
        path: path,
        offset: Math.max(offset, anno.start.offset)-offset
      };
      data.end = {
        path: path,
        offset: Math.min(endOffset, anno.end.offset)-offset
      };
      snippet.create(data);
    });
    return snippet
  }

  function _copyContainerSelection(tx, sel) {
    var snippet = tx.createSnippet();
    var container = snippet.getContainer();

    var nodeIds = sel.getNodeIds();
    var L = nodeIds.length;
    if (L === 0) { return snippet }

    var start = sel.start;
    var end = sel.end;

    var skippedFirst = false;
    var skippedLast = false;

    
    var created = {};
    for(var i = 0; i<L; i++) {
      var id = nodeIds[i];
      var node = tx.get(id);
      
      if (i===0 && isLast(tx, start)) {
        skippedFirst = true;
        continue
      }
      if (i===L-1 && isFirst(tx, end)) {
        skippedLast = true;
        continue
      }
      if (!created[id]) {
        copyNode(node).forEach(function (nodeData) {
          var copy = snippet.create(nodeData);
          created[copy.id] = true;
        });
        container.show(id);
      }
    }
    if (!skippedFirst) {
      
      var startNode = snippet.get(start.getNodeId()).getContainerRoot();
      if (startNode.isText()) {
        deleteTextRange(snippet, null, start);
      } else if (startNode.isList()) {
        deleteListRange(snippet, startNode, null, start);
      }
    }
    if (!skippedLast) {
      
      var endNode = snippet.get(end.getNodeId()).getContainerRoot();
      if (endNode.isText()) {
        deleteTextRange(snippet, end, null);
      } else if (endNode.isList()) {
        deleteListRange(snippet, endNode, end, null);
      }
    }
    return snippet
  }

  function _copyNodeSelection(doc, selection) {
    var snippet = doc.createSnippet();
    var containerNode = snippet.getContainer();
    var nodeId = selection.getNodeId();
    var node = doc.get(nodeId);
    copyNode(node).forEach(function (nodeData) {
      snippet.create(nodeData);
    });
    containerNode.show(node.id);
    return snippet
  }

  function paste(tx, args) {
    var sel = tx.selection;
    if (!sel || sel.isNull() || sel.isCustomSelection()) {
      throw new Error("Can not paste, without selection or a custom selection.")
    }
    args = args || {};
    args.text = args.text || '';
    var pasteDoc = args.doc;
    
    
    var inContainer = Boolean(sel.containerId);

    
    
    
    if (!pasteDoc && !inContainer) {
      tx.insertText(args.text);
      return
    }
    if (!pasteDoc) {
      pasteDoc = _convertPlainTextToDocument(tx, args);
    }
    if (!sel.isCollapsed()) {
      tx.deleteSelection();
    }
    var snippet = pasteDoc.get(SNIPPET_ID);
    if (snippet.getLength() > 0) {
      var first = snippet.getChildAt(0);
      if (first.isText()) {
        _pasteAnnotatedText(tx, pasteDoc);
        
        
        
        snippet.hideAt(0);
      }
      
      if (snippet.getLength() > 0) {
        _pasteDocument(tx, pasteDoc);
      }
    }
    return args
  }


  function _convertPlainTextToDocument(tx, args) {
    var lines = args.text.split(/\s*\n\s*\n/);
    var pasteDoc = tx.getDocument().newInstance();
    var defaultTextType = pasteDoc.getSchema().getDefaultTextType();
    var container = pasteDoc.create({
      type: 'container',
      id: SNIPPET_ID,
      nodes: []
    });
    var node;
    if (lines.length === 1) {
      node = pasteDoc.create({
        id: TEXT_SNIPPET_ID,
        type: defaultTextType,
        content: lines[0]
      });
      container.show(node.id);
    } else {
      for (var i = 0; i < lines.length; i++) {
        node = pasteDoc.create({
          id: uuid(defaultTextType),
          type: defaultTextType,
          content: lines[i]
        });
        container.show(node.id);
      }
    }
    return pasteDoc
  }

  function _pasteAnnotatedText(tx, copy) {
    var sel = tx.selection;
    var nodes = copy.get(SNIPPET_ID).nodes;
    var firstId = nodes[0];
    var first = copy.get(firstId);
    var textPath = first.getPath();
    var text = copy.get(textPath);
    var annotations = copy.getIndex('annotations').get(textPath);
    
    var path = sel.start.path;
    var offset = sel.start.offset;
    tx.insertText(text);
    
    forEach(annotations, function(anno) {
      var data = anno.toJSON();
      data.start.path = path.slice(0);
      data.start.offset += offset;
      data.end.offset += offset;
      
      if (tx.get(data.id)) { data.id = uuid(data.type); }
      tx.create(data);
    });
  }

  function _pasteDocument(tx, pasteDoc) {
    var sel = tx.selection;
    var containerId = sel.containerId;
    var container = tx.get(containerId);
    var insertPos;
    if (sel.isPropertySelection()) {
      var startPath = sel.start.path;
      var nodeId = sel.start.getNodeId();
      var startPos = container.getPosition(nodeId, 'strict');
      var text = tx.get(startPath);
      
      
      if (text.length === 0) {
        insertPos = startPos;
        container.hide(nodeId);
        deleteNode(tx, tx.get(nodeId));
      } else if ( text.length === sel.start.offset ) {
        insertPos = startPos + 1;
      } else {
        tx.break();
        insertPos = startPos + 1;
      }
    } else if (sel.isNodeSelection()) {
      var nodePos = container.getPosition(sel.getNodeId(), 'strict');
      if (sel.isBefore()) {
        insertPos = nodePos;
      } else if (sel.isAfter()) {
        insertPos = nodePos+1;
      } else {
        throw new Error('Illegal state: the selection should be collapsed.')
      }
    }
    
    var nodeIds = pasteDoc.get(SNIPPET_ID).nodes;
    var insertedNodes = [];
    var visited = {};
    for (var i = 0; i < nodeIds.length; i++) {
      var node = pasteDoc.get(nodeIds[i]);
      
      
      
      
      
      var newId = _transferWithDisambiguatedIds(node.getDocument(), tx, node.id, visited);
      
      node = tx.get(newId);
      container.showAt(insertPos++, newId);
      insertedNodes.push(node);
    }

    if (insertedNodes.length > 0) {
      var lastNode = last$1(insertedNodes);
      setCursor(tx, lastNode, containerId, 'after');
    }
  }






  function _transferWithDisambiguatedIds(sourceDoc, targetDoc, id, visited) {
    if (visited[id]) { throw new Error('FIXME: dont call me twice') }
    var node = sourceDoc.get(id, 'strict');
    var oldId = node.id;
    var newId;
    if (targetDoc.contains(node.id)) {
      
      newId = uuid(node.type);
      node.id = newId;
    }
    visited[id] = node.id;
    var annotationIndex = sourceDoc.getIndex('annotations');
    var nodeSchema = node.getSchema();
    
    var annos = [];
    
    
    
    for (var key in nodeSchema) {
      if (key === 'id' || key === 'type' || !nodeSchema.hasOwnProperty(key)) { continue }
      var prop = nodeSchema[key];
      var name = prop.name;
      
      if ((prop.isReference() && prop.isOwned()) || (prop.type === 'file')) {
        
        
        var val = node[prop.name];
        if (prop.isArray()) {
          _transferArrayOfReferences(sourceDoc, targetDoc, val, visited);
        } else {
          var id$1 = val;
          if (!visited[id$1]) {
            node[name] = _transferWithDisambiguatedIds(sourceDoc, targetDoc, id$1, visited);
          }
        }
      }
      
      else if (prop.isText()) {
        
        
        
        
        
        var _annos = annotationIndex.get([oldId, prop.name]);
        for (var i = 0; i < _annos.length; i++) {
          var anno = _annos[i];
          if (anno.start.path[0] === oldId && newId) {
            anno.start.path[0] = newId;
          }
          if (anno.end.path[0] === oldId && newId) {
            anno.end.path[0] = newId;
          }
          annos.push(anno);
        }
      }
    }
    targetDoc.create(node);
    for (var i$1 = 0; i$1 < annos.length; i$1++) {
      _transferWithDisambiguatedIds(sourceDoc, targetDoc, annos[i$1].id, visited);
    }
    return node.id
  }

  function _transferArrayOfReferences(sourceDoc, targetDoc, arr, visited) {
    for (var i = 0; i < arr.length; i++) {
      var val = arr[i];
      
      if (isArray(val)) {
        _transferArrayOfReferences(sourceDoc, targetDoc, val, visited);
      } else {
        var id = val;
        if (id && !visited[id]) {
          arr[i] = _transferWithDisambiguatedIds(sourceDoc, targetDoc, id, visited);
        }
      }
    }
  }

  var Editing = function Editing () {};

  Editing.prototype.annotate = function annotate (tx, annotation) {
    var sel = tx.selection;
    var schema = tx.getSchema();
    var AnnotationClass = schema.getNodeClass(annotation.type);
    if (!AnnotationClass) { throw new Error('Unknown annotation type', annotation) }
    var start = sel.start;
    var end = sel.end;
    var containerId = sel.containerId;
    var nodeData = { start: start, end: end, containerId: containerId };
    
    
    if (sel.isPropertySelection()) {
      if (!AnnotationClass.prototype._isAnnotation) {
        throw new Error('Annotation can not be created for a selection.')
      }
    } else if (sel.isContainerSelection()) {
      if (AnnotationClass.prototype._isPropertyAnnotation) {
        console.warn('NOT SUPPORTED YET: creating property annotations for a non collapsed container selection.');
      }
    }
    Object.assign(nodeData, annotation);
    return tx.create(nodeData)
  };

  Editing.prototype.break = function break$1 (tx) {
    var sel = tx.selection;
    if (sel.isNodeSelection()) {
      var containerId = sel.containerId;
      var container = tx.get(containerId);
      var nodeId = sel.getNodeId();
      var nodePos = container.getPosition(nodeId, 'strict');
      var textNode = this.createTextNode(tx, container);
      if (sel.isBefore()) {
        container.showAt(nodePos, textNode.id);
        
      } else {
        container.showAt(nodePos+1, textNode.id);
        setCursor(tx, textNode, containerId, 'before');
      }
    }
    else if (sel.isCustomSelection()) ;
    else if (sel.isCollapsed() || sel.isPropertySelection()) {
      var containerId$1 = sel.containerId;
      if (!sel.isCollapsed()) {
        
        this._deletePropertySelection(tx, sel);
        tx.setSelection(sel.collapse('left'));
      }
      
      if (containerId$1) {
        var container$1 = tx.get(containerId$1);
        var nodeId$1 = sel.start.path[0];
        var node = tx.get(nodeId$1);
        this._breakNode(tx, node, sel.start, container$1);
      }
    }
    else if (sel.isContainerSelection()) {
      var start = sel.start;
      var containerId$2 = sel.containerId;
      var container$2 = tx.get(containerId$2);
      var startNodeId = start.path[0];
      var nodePos$1 = container$2.getPosition(startNodeId, 'strict');
      this._deleteContainerSelection(tx, sel, {noMerge: true });
      setCursor(tx, container$2.getNodeAt(nodePos$1+1), containerId$2, 'before');
    }
  };

  Editing.prototype.createTextNode = function createTextNode (tx, container, text) { 
    
    
    return tx.createDefaultTextNode(text)
  };

  Editing.prototype.createListNode = function createListNode (tx, container, params) {
      if ( params === void 0 ) params = {};
   
    
    
    return tx.create({ type: "list", items: [], listType: params.listType || 'bullet' })
  };

  Editing.prototype.delete = function delete$1 (tx, direction) {
    var sel = tx.selection;
    
    
    
    
    if (sel.isNodeSelection()) {
      this._deleteNodeSelection(tx, sel, direction);
    }
    
    else if (sel.isCustomSelection()) ; 
    
    
    else if (sel.isCollapsed()) {
      
      
      
      var path = sel.start.path;
      var node = tx.get(path[0]);
      var text = tx.get(path);
      var offset = sel.start.offset;
      var needsMerge = (sel.containerId && (
        (offset === 0 && direction === 'left') ||
        (offset === text.length && direction === 'right')
      ));
      if (needsMerge) {
        
        
        
        
        
        
        
        var root = node.getContainerRoot();
        if (root.isList() && offset === 0 && direction === 'left') {
          return this.toggleList(tx)
        } else {
          var container = tx.get(sel.containerId);
          this._merge(tx, root, sel.start, direction, container);
        }
      } else {
        
        if ((offset === 0 && direction === 'left') ||
          (offset === text.length && direction === 'right')) {
          return
        }
        var startOffset = (direction === 'left') ? offset-1 : offset;
        var endOffset = startOffset+1;
        var start = { path: path, offset: startOffset };
        var end = { path: path, offset: endOffset };
        deleteTextRange(tx, start, end);
        tx.setSelection({
          type: 'property',
          path: path,
          startOffset: startOffset,
          containerId: sel.containerId
        });
      }
    }
    
    else if (sel.isPropertySelection()) {
      deleteTextRange(tx, sel.start, sel.end);
      tx.setSelection(sel.collapse('left'));
    }
    
    else if (sel.isContainerSelection()) {
      this._deleteContainerSelection(tx, sel);
    }
    else {
      console.warn('Unsupported case: tx.delete(%)', direction, sel);
    }
  };

  Editing.prototype._deleteNodeSelection = function _deleteNodeSelection (tx, sel, direction) {
    var nodeId = sel.getNodeId();
    var container = tx.get(sel.containerId);
    var nodePos = container.getPosition(nodeId, 'strict');
    if (sel.isFull() ||
        sel.isBefore() && direction === 'right' ||
        sel.isAfter() && direction === 'left' ) {
      
      container.hideAt(nodePos);
      deleteNode(tx, tx.get(nodeId));
      var newNode = this.createTextNode(tx, container);
      container.showAt(nodePos, newNode.id);
      tx.setSelection({
        type: 'property',
        path: newNode.getPath(),
        startOffset: 0,
        containerId: container.id,
      });
    } else {
      
      if (sel.isBefore() && direction === 'left') {
        if (nodePos > 0) {
          var previous = container.getNodeAt(nodePos-1);
          if (previous.isText()) {
            tx.setSelection({
              type: 'property',
              path: previous.getPath(),
              startOffset: previous.getLength()
            });
            this.delete(tx, direction);
          } else {
            tx.setSelection({
              type: 'node',
              nodeId: previous.id,
              containerId: container.id
            });
          }
        }
      } else if (sel.isAfter() && direction === 'right') {
        if (nodePos < container.getLength()-1) {
          var next = container.getNodeAt(nodePos+1);
          if (next.isText()) {
            tx.setSelection({
              type: 'property',
              path: next.getPath(),
              startOffset: 0
            });
            this.delete(tx, direction);
          } else {
            tx.setSelection({
              type: 'node',
              nodeId: next.id,
              containerId: container.id
            });
          }
        }
      } else {
        console.warn('Unsupported case: delete(%s)', direction, sel);
      }
    }
  };

  Editing.prototype._deletePropertySelection = function _deletePropertySelection (tx, sel) {
    var path = sel.start.path;
    var start = sel.start.offset;
    var end = sel.end.offset;
    tx.update(path, { type: 'delete', start: start, end: end });
    annotationHelpers.deletedText(tx, path, start, end);
  };


  Editing.prototype._deleteContainerSelection = function _deleteContainerSelection (tx, sel, options) {
      if ( options === void 0 ) options = {};

    var containerId = sel.containerId;
    var container = tx.get(containerId);
    var start = sel.start;
    var end = sel.end;
    var startId = start.getNodeId();
    var endId = end.getNodeId();
    var startPos = container.getPosition(startId, 'strict');
    var endPos = container.getPosition(endId, 'strict');

    
    if (startPos === endPos) {
      
      var node = tx.get(startId).getContainerRoot();
      
      if (node.isText()) {
        deleteTextRange(tx, start, end);
      } else if (node.isList()) {
        deleteListRange(tx, node, start, end);
      } else {
        throw new Error('Not supported yet.')
      }
      tx.setSelection(sel.collapse('left'));
      return
    }

    

    var firstNode = tx.get(start.getNodeId());
    var lastNode = tx.get(end.getNodeId());
    var firstEntirelySelected = isEntirelySelected(tx, firstNode, start, null);
    var lastEntirelySelected = isEntirelySelected(tx, lastNode, null, end);

    
    if (lastEntirelySelected) {
      container.hideAt(endPos);
      deleteNode(tx, lastNode);
    } else {
      
      var node$1 = lastNode.getContainerRoot();
      
      if (node$1.isText()) {
        deleteTextRange(tx, null, end);
      } else if (node$1.isList()) {
        deleteListRange(tx, node$1, null, end);
      }
    }

    
    for (var i = endPos-1; i > startPos; i--) {
      var nodeId = container.getNodeIdAt(i);
      container.hideAt(i);
      deleteNode(tx, tx.get(nodeId));
    }

    
    if (firstEntirelySelected) {
      container.hideAt(startPos);
      deleteNode(tx, firstNode);
    } else {
      
      var node$2 = firstNode.getContainerRoot();
      
      if (node$2.isText()) {
        deleteTextRange(tx, start, null);
      } else if (node$2.isList()) {
        deleteListRange(tx, node$2, start, null);
      }
    }

    
    if (firstEntirelySelected && lastEntirelySelected) {
      
      var textNode = this.createTextNode(tx, container);
      container.showAt(startPos, textNode.id);
      tx.setSelection({
        type: 'property',
        path: textNode.getPath(),
        startOffset: 0,
        containerId: containerId
      });
    } else if (!firstEntirelySelected && !lastEntirelySelected) {
      if (!options.noMerge) {
        this._merge(tx, firstNode, sel.start, 'right', container);
      }
      tx.setSelection(sel.collapse('left'));
    } else if (firstEntirelySelected) {
      setCursor(tx, lastNode, container.id, 'before');
    } else {
      setCursor(tx, firstNode, container.id, 'after');
    }
  };

  Editing.prototype.insertInlineNode = function insertInlineNode (tx, nodeData) {
    var sel = tx.selection;
    var text = "\uFEFF";
    this.insertText(tx, text);
    sel = tx.selection;
    var endOffset = tx.selection.end.offset;
    var startOffset = endOffset - text.length;
    nodeData = Object.assign({}, nodeData, {
      start: {
        path: sel.path,
        offset: startOffset
      },
      end: {
        path: sel.path,
        offset: endOffset
      }
    });
    return tx.create(nodeData)
  };

  Editing.prototype.insertBlockNode = function insertBlockNode (tx, nodeData) {
    var sel = tx.selection;
    
    var blockNode;
    if (!nodeData._isNode || !tx.get(nodeData.id)) {
      blockNode = tx.create(nodeData);
    } else {
      blockNode = tx.get(nodeData.id);
    }
    
    if (sel.isNodeSelection()) {
      var containerId = sel.containerId;
      var container = tx.get(containerId);
      var nodeId = sel.getNodeId();
      var nodePos = container.getPosition(nodeId, 'strict');
      
      if (sel.isBefore()) {
        container.showAt(nodePos, blockNode.id);
      }
      
      else if (sel.isAfter()) {
        container.showAt(nodePos+1, blockNode.id);
        tx.setSelection({
          type: 'node',
          containerId: containerId,
          nodeId: blockNode.id,
          mode: 'after'
        });
      } else {
        container.hideAt(nodePos);
        deleteNode(tx, tx.get(nodeId));
        container.showAt(nodePos, blockNode.id);
        tx.setSelection({
          type: 'node',
          containerId: containerId,
          nodeId: blockNode.id,
          mode: 'after'
        });
      }
    } else if (sel.isPropertySelection()) {
      
      if (!sel.containerId) { throw new Error('insertBlockNode can only be used within a container.') }
      var container$1 = tx.get(sel.containerId);
      if (!sel.isCollapsed()) {
        this._deletePropertySelection(tx, sel);
        tx.setSelection(sel.collapse('left'));
      }
      var node = tx.get(sel.path[0]);
      
      if (!node) { throw new Error('Invalid selection.') }
      var nodePos$1 = container$1.getPosition(node.id, 'strict');
      
      if (node.isText()) {
        var text = node.getText();
        
        if (text.length === 0) {
          container$1.hideAt(nodePos$1);
          deleteNode(tx, node);
          container$1.showAt(nodePos$1, blockNode.id);
          setCursor(tx, blockNode, container$1.id, 'after');
        }
        
        else if (sel.start.offset === 0) {
          container$1.showAt(nodePos$1, blockNode.id);
        }
        
        else if (sel.start.offset === text.length) {
          container$1.showAt(nodePos$1+1, blockNode.id);
          setCursor(tx, blockNode, container$1.id, 'before');
        }
        
        else {
          this.break(tx);
          container$1.showAt(nodePos$1+1, blockNode.id);
          setCursor(tx, blockNode, container$1.id, 'after');
        }
      } else {
        console.error('Not supported: insertBlockNode() on a custom node');
      }
    } else if (sel.isContainerSelection()) {
      if (sel.isCollapsed()) {
        var start = sel.start;
        
        if (start.isPropertyCoordinate()) {
          tx.setSelection({
            type: 'property',
            path: start.path,
            startOffset: start.offset,
            containerId: sel.containerId,
          });
        } else if (start.isNodeCoordinate()) {
          tx.setSelection({
            type: 'node',
            containerId: sel.containerId,
            nodeId: start.path[0],
            mode: start.offset === 0 ? 'before' : 'after',
          });
        } else {
          throw new Error('Unsupported selection for insertBlockNode')
        }
        return this.insertBlockNode(tx, blockNode)
      } else {
        this.break(tx);
        return this.insertBlockNode(tx, blockNode)
      }
    }
    return blockNode
  };

  Editing.prototype.insertText = function insertText (tx, text) {
    var sel = tx.selection;
    
    
    
    if (sel.isNodeSelection()) {
      var containerId = sel.containerId;
      var container = tx.get(containerId);
      var nodeId = sel.getNodeId();
      var nodePos = container.getPosition(nodeId, 'strict');
      var textNode = this.createTextNode(tx, container, text);
      if (sel.isBefore()) {
        container.showAt(nodePos, textNode);
      } else if (sel.isAfter()) {
        container.showAt(nodePos+1, textNode);
      } else {
        container.hide(nodeId);
        deleteNode(tx, tx.get(nodeId));
        container.showAt(nodePos, textNode);
      }
      setCursor(tx, textNode, sel.containerId, 'after');
    } else if (sel.isCustomSelection()) ; else if (sel.isCollapsed() || sel.isPropertySelection()) {
      
      this._insertText(tx, sel, text);
      
    } else if (sel.isContainerSelection()) {
      this._deleteContainerSelection(tx, sel);
      this.insertText(tx, text);
    }
  };

  Editing.prototype.paste = function paste$1 (tx, content) {
    if (!content) { return }
    
    if (isString(content)) {
      paste(tx, {text: content});
    } else if (content._isDocument) {
      paste(tx, {doc: content});
    } else {
      throw new Error('Illegal content for paste.')
    }
  };


  Editing.prototype.switchTextType = function switchTextType (tx, data) {
    var sel = tx.selection;
    
    if (!sel.isPropertySelection()) {
      throw new Error("Selection must be a PropertySelection.")
    }
    var containerId = sel.containerId;
    
    if (!containerId) {
      throw new Error("Selection must be within a container.")
    }
    var path = sel.path;
    var nodeId = path[0];
    var node = tx.get(nodeId);
    
    if (!(node.isInstanceOf('text'))) {
      throw new Error('Trying to use switchTextType on a non text node.')
    }
    var newId = uuid(data.type);
    
    var oldPath = node.getPath();
    console.assert(oldPath.length === 2, "Currently we assume that TextNodes store the plain-text on the first level");
    var textProp = oldPath[1];
    var newPath = [newId, textProp];
    
    var newNodeData = Object.assign({
      id: newId,
      type: data.type,
      direction: node.direction
    }, data);
    newNodeData[textProp] = node.getText();

    var newNode = tx.create(newNodeData);
    annotationHelpers.transferAnnotations(tx, path, 0, newPath, 0);

    
    var container = tx.get(sel.containerId);
    var pos = container.getPosition(nodeId, 'strict');
    container.hide(nodeId);
    deleteNode(tx, node);
    container.showAt(pos, newNode.id);

    tx.setSelection({
      type: 'property',
      path: newPath,
      startOffset: sel.start.offset,
      endOffset: sel.end.offset,
      containerId: containerId
    });

    return newNode
  };

  Editing.prototype.toggleList = function toggleList (tx, params) {
    var sel = tx.selection;
    var container = tx.get(sel.containerId);
    
    if (!container) {
      throw new Error("Selection must be within a container.")
    }
    if (sel.isPropertySelection()) {
      var nodeId = sel.start.path[0];
      
      var node = tx.get(nodeId).getContainerRoot();
      var nodePos = container.getPosition(node.id, 'strict');
      
      if (node.isText()) {
        container.hideAt(nodePos);
        var newList = this.createListNode(tx, container, params);
        var newItem = newList.createListItem(node.getText());
        annotationHelpers.transferAnnotations(tx, node.getPath(), 0, newItem.getPath(), 0);
        newList.appendItem(newItem);
        deleteNode(tx, node);
        container.showAt(nodePos, newList.id);
        tx.setSelection({
          type: 'property',
          path: newItem.getPath(),
          startOffset: sel.start.offset,
          containerId: sel.containerId
        });
      } else if (node.isList()) {
        var itemId = sel.start.path[0];
        var item = tx.get(itemId);
        var itemPos = node.getItemPosition(item);
        var newTextNode = this.createTextNode(tx, container, item.getText());
        annotationHelpers.transferAnnotations(tx, item.getPath(), 0, newTextNode.getPath(), 0);
        
        node.removeItemAt(itemPos);
        if (node.isEmpty()) {
          container.hideAt(nodePos);
          deleteNode(tx, node);
          container.showAt(nodePos, newTextNode.id);
        } else if (itemPos === 0) {
          container.showAt(nodePos, newTextNode.id);
        } else if (node.getLength() <= itemPos){
          container.showAt(nodePos+1, newTextNode.id);
        } else {
          
          var tail = [];
          var items = node.getItems();
          var L = items.length;
          for (var i = L-1; i >= itemPos; i--) {
            tail.unshift(items[i]);
            node.removeItemAt(i);
          }
          var newList$1 = this.createListNode(tx, container, node);
          for (var i$1 = 0; i$1 < tail.length; i$1++) {
            newList$1.appendItem(tail[i$1]);
          }
          container.showAt(nodePos+1, newTextNode.id);
          container.showAt(nodePos+2, newList$1.id);
        }
        tx.setSelection({
          type: 'property',
          path: newTextNode.getPath(),
          startOffset: sel.start.offset,
          containerId: sel.containerId
        });
      }
    } else if (sel.isContainerSelection()) {
      console.error('TODO: support toggleList with ContainerSelection');
    }
  };

  Editing.prototype.indent = function indent (tx) {
    var sel = tx.selection;
    if (sel.isPropertySelection()) {
      var nodeId = sel.start.getNodeId();
      
      var node = tx.get(nodeId).getContainerRoot();
      if (node.isList()) {
        var itemId = sel.start.path[0];
        var item = tx.get(itemId);
        var level = item.getLevel();
        
        if (item && level<3) {
          item.setLevel(item.level+1);
          
          tx.set([node.id, '_itemsChanged'], true);
        }
      }
    } else if (sel.isContainerSelection()) {
      console.error('TODO: support toggleList with ContainerSelection');
    }
  };

  Editing.prototype.dedent = function dedent (tx) {
    var sel = tx.selection;
    if (sel.isPropertySelection()) {
      var nodeId = sel.start.getNodeId();
      
      var node = tx.get(nodeId).getContainerRoot();
      if (node.isList()) {
        var itemId = sel.start.path[0];
        var item = tx.get(itemId);
        var level = item.getLevel();
        if (item) {
          if (level > 1) {
            item.setLevel(item.level-1);
            
            tx.set([node.id, '_itemsChanged'], true);
          }
          
          
          
          
          
        }
      }
    } else if (sel.isContainerSelection()) {
      console.error('TODO: support toggleList with ContainerSelection');
    }
  };


  Editing.prototype._insertText = function _insertText (tx, sel, text) {
    var start = sel.start;
    var end = sel.end;
    
    if (!isArrayEqual(start.path, end.path)) {
      throw new Error('Unsupported state: range should be on one property')
    }
    var path = start.path;
    var startOffset = start.offset;
    var endOffset = end.offset;
    var typeover = !sel.isCollapsed();
    var L = text.length;
    
    if (typeover) {
      tx.update(path, { type: 'delete', start: startOffset, end: endOffset });
    }
    
    tx.update(path, { type: 'insert', start: startOffset, text: text });
    
    var annos = tx.getAnnotations(path);
    annos.forEach(function(anno) {
      var annoStart = anno.start.offset;
      var annoEnd = anno.end.offset;

      
      
      if (annoEnd<startOffset) {
        return
      }
      
      else if (annoStart>=endOffset) {
        tx.update([anno.id, 'start'], { type: 'shift', value: startOffset-endOffset+L });
        tx.update([anno.id, 'end'], { type: 'shift', value: startOffset-endOffset+L });
      }
      
      
      
      
      else if (
        (annoStart>=startOffset && annoEnd<endOffset) ||
        (anno._isInlineNode && annoStart>=startOffset && annoEnd<=endOffset)
      ) {
        tx.delete(anno.id);
      }
      
      else if (annoStart>=startOffset && annoEnd>=endOffset) {
        
        if (annoStart>startOffset || !typeover) {
          tx.update([anno.id, 'start'], { type: 'shift', value: startOffset-annoStart+L });
        }
        tx.update([anno.id, 'end'], { type: 'shift', value: startOffset-endOffset+L });
      }
      
      else if (annoStart<startOffset && annoEnd<endOffset) {
        
        tx.update([anno.id, 'end'], { type: 'shift', value: startOffset-annoEnd+L });
      }
      
      else if (annoEnd === startOffset && !anno.constructor.autoExpandRight) ;
      
      else if (annoStart<startOffset && annoEnd>=endOffset) {
        if (anno._isInlineNode) ; else {
          tx.update([anno.id, 'end'], { type: 'shift', value: startOffset-endOffset+L });
        }
      }
      else {
        console.warn('TODO: handle annotation update case.');
      }
    });
    var offset = startOffset + text.length;
    tx.setSelection({
      type: 'property',
      path: start.path,
      startOffset: offset,
      containerId: sel.containerId,
      surfaceId: sel.surfaceId
    });
  };

  Editing.prototype._breakNode = function _breakNode (tx, node, coor, container) {
    
    node = node.getContainerRoot();
    
    if (node.isText()) {
      this._breakTextNode(tx, node, coor, container);
    } else if (node.isList()) {
      this._breakListNode(tx, node, coor, container);
    } else {
      throw new Error('Not supported')
    }
  };

  Editing.prototype._breakTextNode = function _breakTextNode (tx, node, coor, container) {
    var path = coor.path;
    var offset = coor.offset;
    var nodePos = container.getPosition(node.id, 'strict');
    var text = node.getText();

    
    
    if (offset === 0) {
      var newNode = tx.create({
        type: node.type,
        content: ""
      });
      
      container.showAt(nodePos, newNode.id);
      tx.setSelection({
        type: 'property',
        path: path,
        startOffset: 0,
        containerId: container.id
      });
    }
    
    else {
      var textPath = node.getPath();
      var textProp = textPath[1];
      var newId = uuid(node.type);
      var newNodeData = node.toJSON();
      newNodeData.id = newId;
      newNodeData[textProp] = text.substring(offset);
      
      if (offset === text.length) {
        newNodeData.type = tx.getSchema().getDefaultTextType();
      }
      var newNode$1 = tx.create(newNodeData);
      
      if (offset < text.length) {
        
        annotationHelpers.transferAnnotations(tx, path, offset, newNode$1.getPath(), 0);
        
        tx.update(path, { type: 'delete', start: offset, end: text.length });
      }
      
      container.showAt(nodePos+1, newNode$1.id);
      
      tx.setSelection({
        type: 'property',
        path: newNode$1.getPath(),
        startOffset: 0,
        containerId: container.id
      });
    }
  };

  Editing.prototype._breakListNode = function _breakListNode (tx, node, coor, container) {
    var path = coor.path;
    var offset = coor.offset;
    var listItem = tx.get(path[0]);

    var L = node.length;
    var itemPos = node.getItemPosition(listItem);
    var text = listItem.getText();
    var textProp = listItem.getPath()[1];
    var newItemData = listItem.toJSON();
    delete newItemData.id;
    if (offset === 0) {
      
      if (!text) {
        
        
        var nodePos = container.getPosition(node.id, 'strict');
        var newTextNode = this.createTextNode(tx, container);
        
        if (L < 2) {
          container.hide(node.id);
          deleteNode(tx, node);
          container.showAt(nodePos, newTextNode.id);
        }
        
        else if (itemPos === 0) {
          node.removeItem(listItem);
          deleteNode(tx, listItem);
          container.showAt(nodePos, newTextNode.id);
        }
        
        else if (itemPos >= L-1) {
          node.removeItem(listItem);
          deleteNode(tx, listItem);
          container.showAt(nodePos+1, newTextNode.id);
        }
        
        else {
          var tail = [];
          var items = node.getItems().slice();
          for (var i = L-1; i > itemPos; i--) {
            tail.unshift(items[i]);
            node.removeItem(items[i]);
          }
          node.removeItem(items[itemPos]);
          var newList = this.createListNode(tx, container, node);
          for (var i$1 = 0; i$1 < tail.length; i$1++) {
            newList.appendItem(tail[i$1]);
          }
          container.showAt(nodePos+1, newTextNode.id);
          container.showAt(nodePos+2, newList.id);
        }
        tx.setSelection({
          type: 'property',
          path: newTextNode.getPath(),
          startOffset: 0
        });
      }
      
      else {
        newItemData[textProp] = "";
        var newItem = tx.create(newItemData);
        node.insertItemAt(itemPos, newItem);
        tx.setSelection({
          type: 'property',
          path: listItem.getPath(),
          startOffset: 0
        });
      }
    }
    
    else {
      newItemData[textProp] = text.substring(offset);
      var newItem$1 = tx.create(newItemData);
      
      if (offset < text.length) {
        
        annotationHelpers.transferAnnotations(tx, path, offset, newItem$1.getPath(), 0);
        
        tx.update(path, { type: 'delete', start: offset, end: text.length });
      }
      node.insertItemAt(itemPos+1, newItem$1);
      tx.setSelection({
        type: 'property',
        path: newItem$1.getPath(),
        startOffset: 0
      });
    }
  };

  Editing.prototype._merge = function _merge (tx, node, coor, direction, container) {
    
    
    if (node.isList()) {
      var list = node;
      var itemId = coor.path[0];
      var item = tx.get(itemId);
      var itemPos = list.getItemPosition(item);
      var withinListNode = (
        (direction === 'left' && itemPos > 0) ||
        (direction === 'right' && itemPos<list.items.length-1)
      );
      if (withinListNode) {
        itemPos = (direction === 'left') ? itemPos-1 : itemPos;
        var target = list.getItemAt(itemPos);
        var targetLength = target.getLength();
        mergeListItems(tx, list.id, itemPos);
        tx.setSelection({
          type: 'property',
          path: target.getPath(),
          startOffset: targetLength,
          containerId: container.id
        });
        return
      }
    }
    
    var nodePos = container.getPosition(node, 'strict');
    if (direction === 'left' && nodePos > 0) {
      this._mergeNodes(tx, container, nodePos-1, direction);
    } else if (direction === 'right' && nodePos<container.getLength()-1) {
      this._mergeNodes(tx, container, nodePos, direction);
    }
  };

  Editing.prototype._mergeNodes = function _mergeNodes (tx, container, pos, direction) {
    var first = container.getChildAt(pos);
    var second = container.getChildAt(pos+1);
    if (first.isText()) {
      
      if (first.isEmpty()) {
        container.hide(first.id);
        deleteNode(tx, first);
        
        
        setCursor(tx, second, container.id, 'before');
        return
      }
      var target = first;
      var targetPath = target.getPath();
      var targetLength = target.getLength();
      if (second.isText()) {
        var source = second;
        var sourcePath = source.getPath();
        container.hide(source.id);
        
        tx.update(targetPath, { type: 'insert', start: targetLength, text: source.getText() });
        
        annotationHelpers.transferAnnotations(tx, sourcePath, 0, targetPath, targetLength);
        deleteNode(tx, source);
        tx.setSelection({
          type: 'property',
          path: targetPath,
          startOffset: targetLength,
          containerId: container.id
        });
      } else if (second.isList()) {
        var list = second;
        if (!second.isEmpty()) {
          var source$1 = list.getFirstItem();
          var sourcePath$1 = source$1.getPath();
          
          list.removeItemAt(0);
          
          tx.update(targetPath, { type: 'insert', start: targetLength, text: source$1.getText() });
          
          annotationHelpers.transferAnnotations(tx, sourcePath$1, 0, targetPath, targetLength);
          
          deleteNode(tx, source$1);
        }
        if (list.isEmpty()) {
          container.hide(list.id);
          deleteNode(tx, list);
        }
        tx.setSelection({
          type: 'property',
          path: targetPath,
          startOffset: targetLength,
          containerId: container.id
        });
      } else {
        selectNode(tx, direction === 'left' ? first.id : second.id, container.id);
      }
    } else if (first.isList()) {
      if (second.isText()) {
        var target$1 = first.getLastItem();
        var targetPath$1 = target$1.getPath();
        var targetLength$1 = target$1.getLength();
        var third = (container.length > pos+2) ? container.getChildAt(pos+2) : null;
        if (second.getLength() === 0) {
          container.hide(second.id);
          deleteNode(tx, second);
        } else {
          var source$2 = second;
          var sourcePath$2 = source$2.getPath();
          container.hide(source$2.id);
          tx.update(targetPath$1, { type: 'insert', start: targetLength$1, text: source$2.getText() });
          annotationHelpers.transferAnnotations(tx, sourcePath$2, 0, targetPath$1, targetLength$1);
          deleteNode(tx, source$2);
        }
        
        if (third && third.type === first.type) {
          this._mergeTwoLists(tx, container, first, third);
        }
        tx.setSelection({
          type: 'property',
          path: target$1.getPath(),
          startOffset: targetLength$1,
          containerId: container.id
        });
      } else if (second.isList()) {
        
        if (direction !== 'right') {
          
          
          throw new Error('Illegal state')
        }
        var item = first.getLastItem();
        this._mergeTwoLists(tx, container, first, second);
        tx.setSelection({
          type: 'property',
          path: item.getPath(),
          startOffset: item.getLength(),
          containerId: container.id
        });
      } else {
        selectNode(tx, direction === 'left' ? first.id : second.id, container.id);
      }
    } else {
      if (second.isText() && second.isEmpty()) {
        container.hide(second.id);
        deleteNode(tx, second);
        setCursor(tx, first, container.id, 'after');
      } else {
        selectNode(tx, direction === 'left' ? first.id : second.id, container.id);
      }
    }
  };

  Editing.prototype._mergeTwoLists = function _mergeTwoLists (tx, container, first, second) {
    container.hide(second.id);
    var secondItems = second.getItems().slice();
    for (var i=0; i<secondItems.length;i++) {
      second.removeItemAt(0);
      first.appendItem(secondItems[i]);
    }
    deleteNode(tx, second);
  };

  var EditingInterface = function EditingInterface(doc, options) {
    if ( options === void 0 ) options = {};

    this._document = doc;
    this._selection = null;
    
    this._impl = options.editing || new Editing();
    this._direction = null;
  };

  var prototypeAccessors$7 = { selection: { configurable: true },textDirection: { configurable: true } };

  EditingInterface.prototype.dispose = function dispose () {};

  EditingInterface.prototype.getDocument = function getDocument () {
    return this._document
  };



  EditingInterface.prototype.get = function get () {
      var ref;

      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];
    return (ref = this._document).get.apply(ref, args)
  };

  EditingInterface.prototype.contains = function contains (id) {
    return this._document.contains(id)
  };

  EditingInterface.prototype.create = function create (nodeData) {
    return this._document.create(nodeData)
  };

  EditingInterface.prototype.createDefaultTextNode = function createDefaultTextNode (content) {
    return this._document.createDefaultTextNode(content, this._direction)
  };

  EditingInterface.prototype.delete = function delete$1 (nodeId) {
    return this._document.delete(nodeId)
  };

  EditingInterface.prototype.set = function set (path, value) {
    return this._document.set(path, value)
  };

  EditingInterface.prototype.update = function update (path, diffOp) {
    return this._document.update(path, diffOp)
  };

  EditingInterface.prototype.updateNode = function updateNode (id, newProps) {
    return this._document.updateNode(id, newProps)
  };



  EditingInterface.prototype.createSelection = function createSelection (selData) {
    
    
    
    
    
    
    
    selData = augmentSelection(selData, this._selection);
    return this._document.createSelection(selData)
  };

  EditingInterface.prototype.setSelection = function setSelection (sel) {
    if (!sel) {
      sel = Selection.nullSelection;
    } else if (isPlainObject(sel)) {
      sel = this.createSelection(sel);
    } else {
      sel = augmentSelection(sel, this._selection);
    }
    this._selection = sel;
    return sel
  };

  EditingInterface.prototype.getSelection = function getSelection () {
    return this._selection
  };

  prototypeAccessors$7.selection.get = function () {
    return this._selection
  };

  prototypeAccessors$7.selection.set = function (sel) {
    this.setSelection(sel);
  };


  prototypeAccessors$7.textDirection.get = function () {
    return this._direction
  };

  prototypeAccessors$7.textDirection.set = function (dir) {
    this._direction = dir;
  };



  EditingInterface.prototype.annotate = function annotate (annotationData) {
    var sel = this._selection;
    if (sel && (sel.isPropertySelection() || sel.isContainerSelection())) {
      return this._impl.annotate(this, annotationData)
    }
  };

  EditingInterface.prototype.break = function break$1 () {
    if (this._selection && !this._selection.isNull()) {
      this._impl.break(this);
    }
  };

  EditingInterface.prototype.copySelection = function copySelection$1 () {
    var sel = this._selection;
    if (sel && !sel.isNull() && !sel.isCollapsed()) {
      return copySelection(this.getDocument(), this._selection)
    }
  };

  EditingInterface.prototype.deleteSelection = function deleteSelection (options) {
    var sel = this._selection;
    if (sel && !sel.isNull() && !sel.isCollapsed()) {
      this._impl.delete(this, 'right', options);
    }
  };

  EditingInterface.prototype.deleteCharacter = function deleteCharacter (direction) {
    var sel = this._selection;
    if (!sel || sel.isNull()) ; else if (!sel.isCollapsed()) {
      this.deleteSelection();
    } else {
      this._impl.delete(this, direction);
    }
  };

  EditingInterface.prototype.insertText = function insertText (text) {
    var sel = this._selection;
    if (sel && !sel.isNull()) {
      this._impl.insertText(this, text);
    }
  };


  EditingInterface.prototype.insertInlineNode = function insertInlineNode (inlineNode) {
    var sel = this._selection;
    if (sel && !sel.isNull() && sel.isPropertySelection()) {
      return this._impl.insertInlineNode(this, inlineNode)
    }
  };

  EditingInterface.prototype.insertBlockNode = function insertBlockNode (blockNode) {
    var sel = this._selection;
    if (sel && !sel.isNull()) {
      return this._impl.insertBlockNode(this, blockNode)
    }
  };

  EditingInterface.prototype.paste = function paste (content) {
    var sel = this._selection;
    if (sel && !sel.isNull() && !sel.isCustomSelection()) {
      return this._impl.paste(this, content)
    }
  };

  EditingInterface.prototype.switchTextType = function switchTextType (nodeData) {
    var sel = this._selection;
    if (sel && !sel.isNull()) {
      return this._impl.switchTextType(this, nodeData)
    }
  };

  EditingInterface.prototype.toggleList = function toggleList (params) {
    var sel = this._selection;
    if (sel && !sel.isNull()) {
      return this._impl.toggleList(this, params)
    }
  };

  EditingInterface.prototype.indent = function indent () {
    var sel = this._selection;
    if (sel && !sel.isNull()) {
      return this._impl.indent(this)
    }
  };

  EditingInterface.prototype.dedent = function dedent () {
    var sel = this._selection;
    if (sel && !sel.isNull()) {
      return this._impl.dedent(this)
    }
  };



  EditingInterface.prototype.getIndex = function getIndex () {
      var ref;

      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];
    return (ref = this._document).getIndex.apply(ref, args)
  };

  EditingInterface.prototype.getAnnotations = function getAnnotations () {
      var ref;

      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];
    return (ref = this._document).getAnnotations.apply(ref, args)
  };

  EditingInterface.prototype.getSchema = function getSchema () {
    return this._document.getSchema()
  };

  EditingInterface.prototype.createSnippet = function createSnippet () {
    return this._document.createSnippet()
  };

  Object.defineProperties( EditingInterface.prototype, prototypeAccessors$7 );

  var JSONConverter = function JSONConverter () {};

  JSONConverter.prototype.importDocument = function importDocument (doc, json) {
    if (!json.nodes) {
      throw new Error('Invalid JSON format.')
    }
    var schema = doc.getSchema();
    if (json.schema && schema.name !== json.schema.name) {
      throw new Error('Incompatible schema.')
    }
    
    var nodes = json.nodes;
    
    
    
    doc.import(function(tx) {
      forEach(nodes, function(node) {
        
        if (tx.get(node.id)) {
          tx.delete(node.id);
        }
        tx.create(node);
      });
    });
    return doc
  };

  JSONConverter.prototype.exportDocument = function exportDocument (doc) {
    var schema = doc.getSchema();
    var json = {
      schema: {
        name: schema.name
      },
      nodes: {}
    };
    forEach(doc.getNodes(), function(node) {
      if (node._isDocumentNode) {
        json.nodes[node.id] = node.toJSON();
      }
    });
    return json
  };

  var ParentNodeHook = function ParentNodeHook(doc) {
    this.doc = doc;
    this.table = {};
    doc.data.on('operation:applied', this._onOperationApplied, this);
  };

  ParentNodeHook.prototype._onOperationApplied = function _onOperationApplied (op) {
    var doc = this.doc;
    var table = this.table;
    var node = doc.get(op.path[0]);
    
    
    switch(op.type) {
      case 'create': {
        switch (node.type) {
          case 'list':
            _setParent(node, node.items);
            break
          case 'list-item': {
            _setRegisteredParent(node);
            break
          }
          case 'table':
            _setParent(node, node.cells);
            break
          case 'table-cell': {
            _setRegisteredParent(node);
            break
          }
          default:
            
        }
        break
      }
      case 'update': {
        
        
        var update = op.diff;
        switch(node.type) {
          case 'list':
            if (op.path[1] === 'items') {
              if (update.isInsert()) {
                _setParent(node, update.getValue());
              }
            }
            break
          case 'table':
            if (op.path[1] === 'cells') {
              if (update.isInsert()) {
                _setParent(node, update.getValue());
              }
            }
            break
          default:
            
        }
        break
      }
      case 'set': {
        switch(node.type) {
          case 'list':
            if (op.path[1] === 'items') {
              _setParent(node, op.getValue());
            }
            break
          case 'table':
            if (op.path[1] === 'cells') {
              _setParent(node, op.getValue());
            }
            break
          default:
            
        }
        break
      }
      default:
        
    }

    function _setParent(parent, ids) {
      if (ids) {
        if (isArray(ids)) {
          ids.forEach(_set);
        } else {
          _set(ids);
        }
      }
      function _set(id) {
        
        
        
        table[id] = parent;
        var child = doc.get(id);
        if (child) {
          child.parent = parent;
        }
      }
    }
    function _setRegisteredParent(child) {
      var parent = table[child.id];
      if (parent) {
        child.parent = parent;
      }
    }
  };

  ParentNodeHook.register = function(doc) {
    return new ParentNodeHook(doc)
  };

  var converter = new JSONConverter();



  var Document = (function (EventEmitter$$1) {
    function Document(schema) {
      var ref;

      var args = [], len = arguments.length - 1;
      while ( len-- > 0 ) args[ len ] = arguments[ len + 1 ];
      EventEmitter$$1.call(this);

      this.schema = schema;
      
      if (!schema) {
        throw new Error('A document needs a schema for reflection.')
      }

      
      this._ops = []

      (ref = this)._initialize.apply(ref, args);
    }

    if ( EventEmitter$$1 ) Document.__proto__ = EventEmitter$$1;
    Document.prototype = Object.create( EventEmitter$$1 && EventEmitter$$1.prototype );
    Document.prototype.constructor = Document;

    var prototypeAccessors = { id: { configurable: true } };

    Document.prototype._initialize = function _initialize () {
      this.__id__ = uuid();
      this.nodeFactory = new DocumentNodeFactory(this);
      this.data = new IncrementalData(this.schema, this.nodeFactory);
      
      this.addIndex('type', new PropertyIndex('type'));
      
      this.addIndex('annotations', new AnnotationIndex());
      
      
      
      this.addIndex('container-annotations', new ContainerAnnotationIndex());
      
      
      ParentNodeHook.register(this);
    };

    Document.prototype.dispose = function dispose () {
      this.off();
      this.data.off();
    };

    prototypeAccessors.id.get = function () {
      return this.__id__
    };

    
    Document.prototype.getSchema = function getSchema () {
      return this.schema
    };

    
    Document.prototype.contains = function contains (id) {
      return this.data.contains(id)
    };

    
    Document.prototype.get = function get (path, strict) {
      return this.data.get(path, strict)
    };

    
    Document.prototype.getNodes = function getNodes$$1 () {
      return this.data.getNodes()
    };

    Document.prototype.getAnnotations = function getAnnotations (path) {
      return this.getIndex('annotations').get(path)
    };

    
    Document.prototype.import = function import$1 (importer) {
      try {
        this.data._stopIndexing();
        importer(this);
        this.data._startIndexing();
      } finally {
        this.data.queue = [];
        this.data._startIndexing();
      }
    };

    
    Document.prototype.create = function create (nodeData) {
      if (!nodeData.id) {
        nodeData.id = uuid(nodeData.type);
      }
      if (!nodeData.type) {
        throw new Error('No node type provided')
      }
      var op = this._create(nodeData);
      if (op) {
        this._ops.push(op);
        if (!this._isTransactionDocument) {
          this._emitChange(op);
        }
        return this.get(nodeData.id)
      }
    };

    Document.prototype.createDefaultTextNode = function createDefaultTextNode (text, dir) {
      return this.create({
        type: this.getSchema().getDefaultTextType(),
        content: text || '',
        direction: dir
      })
    };

    
    Document.prototype.delete = function delete$1 (nodeId) {
      var node = this.get(nodeId);
      var op = this._delete(nodeId);
      if (op) {
        this._ops.push(op);
        if (!this._isTransactionDocument) {
          this._emitChange(op);
        }
      }
      return node
    };

    
    Document.prototype.set = function set (path, value) {
      var oldValue = this.get(path);
      var op = this._set(path, value);
      if (op) {
        this._ops.push(op);
        if (!this._isTransactionDocument) {
          this._emitChange(op);
        }
      }
      return oldValue
    };

    
    Document.prototype.update = function update (path, diff) {
      var op = this._update(path, diff);
      if (op) {
        this._ops.push(op);
        if (!this._isTransactionDocument) {
          this._emitChange(op);
        }
      }
      return op
    };

    
    Document.prototype.updateNode = function updateNode (id, newProps) {
      var this$1 = this;

      var node = this.get(id);
      forEach(newProps, function (value, key) {
        if (!isEqual(node[key], newProps[key])) {
          this$1.set([id, key], value);
        }
      });
    };

    
    Document.prototype.addIndex = function addIndex (name, index) {
      return this.data.addIndex(name, index)
    };

    
    Document.prototype.getIndex = function getIndex (name) {
      return this.data.getIndex(name)
    };

    
    Document.prototype.createSelection = function createSelection (data) {
      var assign, assign$1;

      var sel;
      if (isNil(data)) { return Selection.nullSelection }
      if (arguments.length !== 1 || !isPlainObject(data)) {
        throw new Error('Illegal argument: call createSelection({ type: ... }')
      } else {
        switch (data.type) {
          case 'property': {
            if (isNil(data.endOffset)) {
              data.endOffset = data.startOffset;
            }
            if (!data.hasOwnProperty('reverse')) {
              if (data.startOffset>data.endOffset) {
                (assign = [data.endOffset, data.startOffset], data.startOffset = assign[0], data.endOffset = assign[1]);
                data.reverse = !data.reverse;
              }
            }
            
            var text = this.get(data.path, 'strict');
            if (data.startOffset < 0 || data.startOffset > text.length) {
              throw new Error('Invalid startOffset: target property has length '+text.length+', given startOffset is ' + data.startOffset)
            }
            if (data.endOffset < 0 || data.endOffset > text.length) {
              throw new Error('Invalid startOffset: target property has length '+text.length+', given endOffset is ' + data.endOffset)
            }
            sel = new PropertySelection(data);
            break
          }
          case 'container': {
            var container = this.get(data.containerId, 'strict');
            if (!container) { throw new Error('Can not create ContainerSelection: container "'+data.containerId+'" does not exist.') }
            var start = this._normalizeCoor({ path: data.startPath, offset: data.startOffset});
            var end = this._normalizeCoor({ path: data.endPath, offset: data.endOffset});
            var startAddress = container.getAddress(start);
            var endAddress = container.getAddress(end);
            if (!startAddress) {
              throw new Error('Invalid arguments for ContainerSelection: ', start.toString())
            }
            if (!endAddress) {
              throw new Error('Invalid arguments for ContainerSelection: ', end.toString())
            }
            if (!data.hasOwnProperty('reverse')) {
              if (endAddress.isBefore(startAddress, 'strict')) {
                (assign$1 = [end, start], start = assign$1[0], end = assign$1[1]);
                data.reverse = true;
              }
            }
            sel = new ContainerSelection(container.id, start.path, start.offset, end.path, end.offset, data.reverse, data.surfaceId);
            break
          }
          case 'node': {
            sel = createNodeSelection({
              doc: this,
              nodeId: data.nodeId,
              mode: data.mode,
              containerId: data.containerId,
              reverse: data.reverse,
              surfaceId: data.surfaceId
            });
            break
          }
          case 'custom': {
            sel = CustomSelection.fromJSON(data);
            break
          }
          default:
            throw new Error('Illegal selection type', data)
        }
      }
      if (!sel.isNull()) {
        sel.attach(this);
      }
      return sel
    };

    Document.prototype.newInstance = function newInstance () {
      var DocumentClass = this.constructor;
      return new DocumentClass(this.schema)
    };

    
    Document.prototype.createSnippet = function createSnippet () {
      var snippet = this.newInstance();
      var snippetContainer = snippet.create({
        type: 'container',
        id: SNIPPET_ID
      });
      snippet.getContainer = function() {
        return snippetContainer
      };
      return snippet
    };

    Document.prototype.createFromDocument = function createFromDocument (doc) {
      var this$1 = this;

      
      this.clear();

      var nodes = doc.getNodes();
      var annotations = [];
      var contentNodes = [];
      var containers = [];
      forEach(nodes, function (node) {
        if (node.isAnnotation()) {
          annotations.push(node);
        } else if (node.isContainer()) {
          containers.push(node);
        } else {
          contentNodes.push(node);
        }
      });
      contentNodes.concat(annotations).concat(containers).forEach(function (n){
        this$1.create(n);
      });

      return this
    };

    
    Document.prototype.toJSON = function toJSON () {
      return converter.exportDocument(this)
    };

    Document.prototype.clone = function clone () {
      var copy = this.newInstance();
      copy.createFromDocument(this);
      return copy
    };

    Document.prototype.clear = function clear () {
      this.data.clear();
      this._ops.length = 0;
    };

    
    Document.prototype.createEditingInterface = function createEditingInterface () {
      return new EditingInterface(this)
    };

    Document.prototype.invert = function invert (change) {
      return change.invert()
    };

    Document.prototype._apply = function _apply (documentChange) {
      var this$1 = this;

      forEach(documentChange.ops, function (op) {
        this$1._applyOp(op);
      });
      
      documentChange._extractInformation(this);
    };

    Document.prototype._applyOp = function _applyOp (op) {
      this.data.apply(op);
      this.emit('operation:applied', op);
    };

    Document.prototype._create = function _create (nodeData) {
      return this.data.create(nodeData)
    };

    Document.prototype._delete = function _delete (nodeId) {
      return this.data.delete(nodeId)
    };

    Document.prototype._set = function _set (path, value) {
      return this.data.set(path, value)
    };

    Document.prototype._update = function _update (path, diff) {
      return this.data.update(path, diff)
    };


    Document.prototype._emitChange = function _emitChange (op) {
      var change = new DocumentChange([op], {}, {});
      change._extractInformation(this);
      this._notifyChangeListeners(change, { hidden: true });
    };

    Document.prototype._notifyChangeListeners = function _notifyChangeListeners (change, info) {
      info = info || {};
      this.emit('document:changed', change, info, this);
    };

    
    Document.prototype._createSelectionFromRange = function _createSelectionFromRange (range) {
      if (!range) { return Selection.nullSelection }
      var inOneNode = isEqual(range.start.path, range.end.path);
      if (inOneNode) {
        if (range.start.isNodeCoordinate()) {
          
          
          return new NodeSelection(range.containerId, range.start.getNodeId(), 'full', range.reverse, range.surfaceId)
        } else {
          return this.createSelection({
            type: 'property',
            path: range.start.path,
            startOffset: range.start.offset,
            endOffset: range.end.offset,
            reverse: range.reverse,
            containerId: range.containerId,
            surfaceId: range.surfaceId
          })
        }
      } else {
        return this.createSelection({
          type: 'container',
          startPath: range.start.path,
          startOffset: range.start.offset,
          endPath: range.end.path,
          endOffset: range.end.offset,
          reverse: range.reverse,
          containerId: range.containerId,
          surfaceId: range.surfaceId
        })
      }
    };

    Document.prototype._normalizeCoor = function _normalizeCoor (ref) {
      var path = ref.path;
      var offset = ref.offset;

      
      if (path.length === 1) {
        var node = this.get(path[0]).getContainerRoot();
        if (node.isText()) {
          
          return new Coordinate(node.getPath(), offset === 0 ? 0 : node.getLength())
        } else if (node.isList()) {
          
          if (offset === 0) {
            var item = node.getItemAt(0);
            return new Coordinate(item.getPath(), 0)
          } else {
            var item$1 = this.get(last$1(node.items));
            return new Coordinate(item$1.getPath(), item$1.getLength())
          }
        }
      }
      return new Coordinate(path, offset)
    };

    Object.defineProperties( Document.prototype, prototypeAccessors );

    return Document;
  }(EventEmitter));

  Document.prototype._isDocument = true;

  var ParentNodeHook$1 = function ParentNodeHook(doc) {
    this.doc = doc;
    
    this.parents = {};
    doc.data.on('operation:applied', this._onOperationApplied, this);
  };

  ParentNodeHook$1.prototype._onOperationApplied = function _onOperationApplied (op) {
    var doc = this.doc;
    var parents = this.parents;
    var node = doc.get(op.path[0]);
    switch(op.type) {
      case 'create': {
        if (node._childNodes) {
          _setParent(node, node._childNodes);
        }
        _setRegisteredParent(node);
        break
      }
      case 'update': {
        
        
        var update = op.diff;
        if (op.path[1] === '_childNodes') {
          if (update.isInsert()) {
            _setParent(node, update.getValue());
          } else if (update.isDelete()) {
            _setParent(null, update.getValue());
          }
        }
        break
      }
      case 'set': {
        if (op.path[1] === '_childNodes') {
          _setParent(null, op.getOldValue());
          _setParent(node, op.getValue());
        }
        break
      }
      default:
        
    }

    function _setParent(parent, ids) {
      if (ids) {
        if (isArray(ids)) {
          ids.forEach(_set);
        } else {
          _set(ids);
        }
      }
      function _set(id) {
        
        
        
        parents[id] = parent;
        var child = doc.get(id);
        if (child) {
          child.parentNode = parent;
        }
      }
    }
    function _setRegisteredParent(child) {
      var parent = parents[child.id];
      if (parent) {
        child.parentNode = parent;
      }
    }
  };

  ParentNodeHook$1.register = function(doc) {
    return new ParentNodeHook$1(doc)
  };

  var XMLEditingInterface = (function (EditingInterface$$1) {
    function XMLEditingInterface () {
      EditingInterface$$1.apply(this, arguments);
    }

    if ( EditingInterface$$1 ) XMLEditingInterface.__proto__ = EditingInterface$$1;
    XMLEditingInterface.prototype = Object.create( EditingInterface$$1 && EditingInterface$$1.prototype );
    XMLEditingInterface.prototype.constructor = XMLEditingInterface;

    XMLEditingInterface.prototype.find = function find (cssSelector) {
      return this.getDocument().find(cssSelector)
    };

    XMLEditingInterface.prototype.findAll = function findAll (cssSelector) {
      return this.getDocument().findAll(cssSelector)
    };

    XMLEditingInterface.prototype.createElement = function createElement () {
      var ref;

      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];
      return (ref = this.getDocument()).createElement.apply(ref, args)
    };

    return XMLEditingInterface;
  }(EditingInterface));

  var XMLDocument = (function (Document$$1) {
    function XMLDocument () {
      Document$$1.apply(this, arguments);
    }

    if ( Document$$1 ) XMLDocument.__proto__ = Document$$1;
    XMLDocument.prototype = Object.create( Document$$1 && Document$$1.prototype );
    XMLDocument.prototype.constructor = XMLDocument;

    XMLDocument.prototype._initialize = function _initialize () {
      this.nodeFactory = new DocumentNodeFactory(this);
      this.data = new IncrementalData(this.schema, this.nodeFactory);
      
      this.addIndex('type', new PropertyIndex('type'));
      
      this.addIndex('annotations', new AnnotationIndex());
      ParentNodeHook$1.register(this);
    };

    XMLDocument.prototype.toXML = function toXML () {
      var dom = DefaultDOMElement.createDocument('xml');
      dom.setDoctype.apply(dom, this.getDocTypeParams());
      var xml = dom.createProcessingInstruction('xml', 'version="1.0" encoding="UTF-8"');
      dom.insertAt(0, xml);
      var rootElement = this.getRootNode().toXML();
      dom.append(rootElement);
      return dom
    };

    XMLDocument.prototype.getDocTypeParams = function getDocTypeParams () {
      
      throw new Error('This method is abstract')
    };

    XMLDocument.prototype.getXMLSchema = function getXMLSchema () {
      
      throw new Error('This method is abstract')
    };

    XMLDocument.prototype.getRootNode = function getRootNode () {
      
      throw new Error('This method is abstract')
    };

    
    XMLDocument.prototype.getDocTypeAsString = function getDocTypeAsString () {
      return new Error('This method is abstract')
    };

    XMLDocument.prototype.createEditingInterface = function createEditingInterface () {
      return new XMLEditingInterface(this)
    };

    XMLDocument.prototype.find = function find (cssSelector) {
      return this.getRootNode().find(cssSelector)
    };

    XMLDocument.prototype.findAll = function findAll (cssSelector) {
      return this.getRootNode().findAll(cssSelector)
    };

    XMLDocument.prototype.createElement = function createElement (tagName, data) {
      var node = this.create(Object.assign({
        id: uuid(tagName),
        type: tagName
      }, data));
      return node
    };

    XMLDocument.prototype.getElementSchema = function getElementSchema (type) {
      return this.getXMLSchema().getElementSchema(type)
    };

    XMLDocument.prototype._validateChange = function _validateChange (change) {
      var this$1 = this;

      var changed = {};
      var deleted = [];
      change.ops.forEach(function (op) {
        switch (op.type) {
          case "delete": {
            deleted.push(op.val.id);
            break
          }
          case "create": {
            changed[op.val.id] = true;
            break
          }
          default: {
            if (op.path) {
              changed[op.path[0]] = true;
            }
          }
        }
      });
      
      deleted.forEach(function (id) { return delete changed[id]; });

      var xmlSchema = this.getXMLSchema();
      var errors = [];
      Object.keys(changed).forEach(function (id) {
        var node = this$1.get(id);
        var res = xmlSchema.validateElement(node);
        if (!res.ok) {
          errors = errors.concat(res.errors);
        }
      });
      return {
        ok: errors.length === 0,
        errors: errors
      }
    };

    
    XMLDocument.prototype._apply = function _apply (documentChange) {
      var this$1 = this;

      
      if (!documentChange.ops) { return }

      var created = {};
      var deleted = {};
      var updated = {};

      var doc = this;

      function _recordUpdate(id, op) {
        var record = updated[id];
        if (!record) {
          record = updated[id] = { ops: [] };
        }
        if (op) { record.ops.push(op); }
        return record
      }

      
      function _checkAnnotation(op) {
        
        switch (op.type) {
          case "create":
          case "delete": {
            var annoData = op.val;
            if (annoData.hasOwnProperty('start')) {
              updated[annoData.start.path] = true;
              var node = doc.get(annoData.start.path[0]);
              if (node) {
                _recordUpdate(node.id, op);
              }
            }
            if (annoData.hasOwnProperty('end')) {
              updated[annoData.end.path] = true;
              var node$1 = doc.get(annoData.start.path[0]);
              if (node$1) {
                _recordUpdate(node$1.id, op);
              }
            }
            break
          }
          case "update":
          case "set": {
            var anno = doc.get(op.path[0]);
            if (anno) {
              if (anno.isPropertyAnnotation()) {
                updated[anno.start.path] = true;
                var node$2 = doc.get(anno.start.path[0]);
                if (node$2) {
                  _recordUpdate(node$2.id, op);
                }
              } else if (anno.isContainerAnnotation()) ;
            }
            break
          }
          default:
            
        }
      }

      documentChange.ops.forEach(function (op) {
        
        
        
        switch(op.type) {
          case 'create': {
            created[op.path[0]] = op.val;
            break
          }
          case 'delete': {
            var node = this$1.get(op.path[0]);
            deleted[node.id] = node;
            break
          }
          case 'update':
          case 'set': {
            updated[op.path] = true;
            var node$1 = this$1.get(op.path[0]);
            _recordUpdate(node$1.id, op);
            break
          }
          default:
            
        }
        
        _checkAnnotation(op);

        this$1._applyOp(op);
      });
      
      Object.keys(deleted).forEach(function (id) {
        delete updated[id];
      });
      documentChange.created = created;
      documentChange.updated = updated;
      documentChange.deleted = deleted;
      documentChange._extracted = true;
    };

    return XMLDocument;
  }(Document));

  XMLDocument.prototype._isXMLDocument = true;

  var ManifestDocument = (function (XMLDocument$$1) {
    function ManifestDocument () {
      XMLDocument$$1.apply(this, arguments);
    }

    if ( XMLDocument$$1 ) ManifestDocument.__proto__ = XMLDocument$$1;
    ManifestDocument.prototype = Object.create( XMLDocument$$1 && XMLDocument$$1.prototype );
    ManifestDocument.prototype.constructor = ManifestDocument;

    ManifestDocument.prototype.getRootNode = function getRootNode () {
      var this$1 = this;

      if (!this.root) {
        var nodes = this.getNodes();
        var ids = Object.keys(nodes);
        for (var i = 0; i < ids.length; i++) {
          var node = nodes[ids[i]];
          if (node.type === 'dar') {
            this$1.root = node;
          }
        }
      }
      return this.root
    };

    ManifestDocument.prototype.getDocTypeParams = function getDocTypeParams () {
      return ManifestSchema.getDocTypeParams()
    };

    ManifestDocument.prototype.getXMLSchema = function getXMLSchema () {
      return ManifestSchema
    };

    ManifestDocument.prototype.getDocumentNodes = function getDocumentNodes () {
      return this.findAll('documents > document')
    };

    ManifestDocument.prototype.getAssetNodes = function getAssetNodes () {
      return this.findAll('assets > asset')
    };

    ManifestDocument.prototype.getDocumentEntries = function getDocumentEntries () {
      var documents = this.findAll('documents > document');
      return documents.map(_getEntryFromDocumentNode)
    };

    ManifestDocument.prototype.getDocumentEntry = function getDocumentEntry (id) {
      var entryNode = this.get(id);
      if (entryNode && entryNode.type === 'document') {
        return _getEntryFromDocumentNode(entryNode)
      }
    };

    return ManifestDocument;
  }(XMLDocument));

  function _getEntryFromDocumentNode(documentNode) {
    return Object.assign({ id: documentNode.id }, documentNode.getAttributes())
  }

  function loadManifest(xmlStr) {
    var configurator = new Configurator();
    registerSchema(configurator, ManifestSchema, ManifestDocument, {
      ImporterClass: XMLDocumentImporter
    });
    var importer = configurator.createImporter(ManifestSchema.getName());
    var manifest = importer.importDocument(xmlStr);
    var editorSession = new EditorSession(manifest, { configurator: configurator });
    return editorSession
  }

  var path = {
    join: function join(p1, p2) { return p1+'/'+p2 }
  };

  var VfsLoader = function VfsLoader(vfs, loaders) {
    this.vfs = vfs;
    this.loaders = loaders;
  };

  VfsLoader.prototype.load = function load (rdcUri) {
      var this$1 = this;

    var vfs = this.vfs;
    return new Promise(function (resolve, reject) {
      var manifestPath = path.join(rdcUri, 'manifest.xml');
      var manifestXml;
      try {
        manifestXml = vfs.readFileSync(manifestPath);
      } catch (err) {
        return reject(err)
      }
      var manifest, manifestEditorSession;
      try {
        var ref = loadManifest(manifestXml);
          var editorSession = ref.editorSession;
        manifest = editorSession.getDocument();
        manifestEditorSession = editorSession;
      } catch (err) {
        return reject(err)
      }
      var sessions = {};
      sessions['manifest'] = manifestEditorSession;
      manifest.findAll('container > documents > document').forEach(function (el) {
        var session;
        session = this$1._loadDocument(rdcUri, el);
        sessions[el.id] = session;
      });
      var dc = new DocumentArchive(sessions);
      resolve(dc);
    })
  };

  VfsLoader.prototype._loadDocument = function _loadDocument (rdcUri, el) {
    var vfs = this.vfs;
    var type = el.attr('type');
    var relPath = el.attr('path');
    var uri = path.join(rdcUri, relPath);
    var content = vfs.readFileSync(uri);
    var loader = this.loaders[type];
    return loader.load(content)
  };

  var ManifestLoader = {
    load: function load(manifestXml) {
      return loadManifest(manifestXml)
    }
  }

  var PersistedDocumentArchive = (function (EventEmitter$$1) {
    function PersistedDocumentArchive(storage, buffer) {
      EventEmitter$$1.call(this);
      this.storage = storage;
      this.buffer = buffer;

      this._archiveId = null;
      this._upstreamArchive = null;
      this._sessions = null;
      this._pendingFiles = {};
    }

    if ( EventEmitter$$1 ) PersistedDocumentArchive.__proto__ = EventEmitter$$1;
    PersistedDocumentArchive.prototype = Object.create( EventEmitter$$1 && EventEmitter$$1.prototype );
    PersistedDocumentArchive.prototype.constructor = PersistedDocumentArchive;

    PersistedDocumentArchive.prototype.hasPendingChanges = function hasPendingChanges () {
      return this.buffer.hasPendingChanges()
    };

    PersistedDocumentArchive.prototype.createFile = function createFile (file) {
      var assetId = uuid();
      var fileExtension = last$1(file.name.split('.'));
      var filePath = assetId + "." + fileExtension;
      this._sessions.manifest.transaction(function (tx) {
        var assets = tx.find('assets');
        var asset = tx.createElement('asset', { id: assetId }).attr({
          path: filePath,
          type: file.type
        });
        assets.appendChild(asset);
      });
      this.buffer.addBlob(assetId, {
        id: assetId,
        path: filePath,
        blob: file
      });
      this._pendingFiles[filePath] = URL.createObjectURL(file);
      return filePath
    };

    
    PersistedDocumentArchive.prototype._addDocumentRecord = function _addDocumentRecord (documentId, type, name, path) {
      this._sessions.manifest.transaction(function (tx) {
        var documents = tx.find('documents');
        var docEntry = tx.createElement('document', { id: documentId }).attr({
          name: name,
          path: path,
          type: type
        });
        documents.appendChild(docEntry);
      });
    };

    PersistedDocumentArchive.prototype.addDocument = function addDocument (type, name, xml) {
      var documentId = uuid();
      var sessions = this._sessions;
      var session = this._loadDocument(type, { data: xml }, sessions);
      sessions[documentId] = session;

      this._registerForSessionChanges(session, documentId);

      this._addDocumentRecord(documentId, type, name, documentId+'.xml');

      return documentId
    };

    PersistedDocumentArchive.prototype.removeDocument = function removeDocument (documentId) {
      var session = this._sessions[documentId];
      this._unregisterFromSession(session);
      this._sessions.manifest.transaction(function (tx) {
        var documents = tx.find('documents');
        var docEntry = tx.find(("#" + documentId));
        documents.removeChild(docEntry);
      });
    };

    PersistedDocumentArchive.prototype.renameDocument = function renameDocument (documentId, name) {
      this._sessions.manifest.transaction(function (tx) {
        var docEntry = tx.find(("#" + documentId));
        docEntry.attr({name: name});
      });
    };

    PersistedDocumentArchive.prototype.getDocumentEntries = function getDocumentEntries () {
      return this.getEditorSession('manifest').getDocument().getDocumentEntries()
    };

    PersistedDocumentArchive.prototype.resolveUrl = function resolveUrl (path) {
      var blobUrl = this._pendingFiles[path];
      if (blobUrl) {
        return blobUrl
      } else {
        var fileRecord = this._upstreamArchive.resources[path];
        if (fileRecord && fileRecord.encoding === 'url') {
          return fileRecord.data
        }
      }
    };

    PersistedDocumentArchive.prototype.load = function load (archiveId) {
      var this$1 = this;

      var storage = this.storage;
      var buffer = this.buffer;

      var upstreamArchive;
      return Promise.resolve()
      .then(function () {
        return storage.read(archiveId)
      })
      .then(function (res) {
        upstreamArchive = res;
        return buffer.load()
      })
      .then(function () {
        
        
        
        
        if (!buffer.hasPendingChanges()) {
          var localVersion = buffer.getVersion();
          var upstreamVersion = upstreamArchive.version;
          if (localVersion && upstreamVersion && localVersion !== upstreamVersion) {
            
            
            console.error('Upstream document has changed. Discarding local changes');
            this$1.buffer.reset(upstreamVersion);
          } else {
            buffer.reset(upstreamVersion);
          }
        }
      })
      .then(function () {
        
        var sessions = this$1._ingest(upstreamArchive);
        
        if (!sessions['manifest']) {
          throw new Error('There must be a manifest session.')
        }
        
        if (!buffer.hasPendingChanges()) ; else {
          buffer.reset(upstreamArchive.version);
        }
        
        this$1._registerForAllChanges(sessions);

        this$1._archiveId = archiveId;
        this$1._upstreamArchive = upstreamArchive;
        this$1._sessions = sessions;

        
        this$1._repair();
        return this$1
      })
    };

    PersistedDocumentArchive.prototype._repair = function _repair () {
      
    };

    PersistedDocumentArchive.prototype.save = function save () {
      if (!this.buffer.hasPendingChanges()) {
        console.info('Save: no pending changes.');
        return Promise.resolve()
      }
      return this._save(this._archiveId)
    };

    
    PersistedDocumentArchive.prototype.saveAs = function saveAs (newArchiveId) {
      var this$1 = this;

      return this.storage.clone(this._archiveId, newArchiveId).then(function () {
        return this$1._save(newArchiveId)
      })
    };

    PersistedDocumentArchive.prototype.getEditorSession = function getEditorSession (docId) {
      return this._sessions[docId]
    };

    PersistedDocumentArchive.prototype._loadManifest = function _loadManifest (record) {
      if (!record) {
        throw new Error('manifest.xml is missing')
      }
      return ManifestLoader.load(record.data)
    };

    PersistedDocumentArchive.prototype._registerForAllChanges = function _registerForAllChanges (sessions) {
      var this$1 = this;

      forEach(sessions, function (session, docId) {
        this$1._registerForSessionChanges(session, docId);
      });
    };

    PersistedDocumentArchive.prototype._registerForSessionChanges = function _registerForSessionChanges (session, docId) {
      var this$1 = this;

      session.onUpdate('document', function (change) {
        this$1.buffer.addChange(docId, change);
        
        this$1.emit('archive:changed');
      }, this);
    };

    PersistedDocumentArchive.prototype._unregisterFromSession = function _unregisterFromSession (session) {
      session.off(this);
    };

    
    PersistedDocumentArchive.prototype._save = function _save (archiveId) {
      var this$1 = this;

      var buffer = this.buffer;
      var storage = this.storage;
      var sessions = this._sessions;

      var rawArchive = this._exportChanges(sessions, buffer);

      
      
      
      
      
      
      return storage.write(archiveId, rawArchive).then(function (res) {
        
        
        res = JSON.parse(res);
        
        buffer.reset(res.version);

        
        this$1._archiveId = archiveId;
      }).catch(function (err) {
        console.error('Saving failed.', err);
      })
    };

    PersistedDocumentArchive.prototype._exportAssets = function _exportAssets (sessions, buffer, rawArchive) {
      var manifest = sessions.manifest.getDocument();
      var assetNodes = manifest.getAssetNodes();
      assetNodes.forEach(function (node) {
        var id = node.attr('id');
        if (!buffer.hasBlob(id)) { return }
        var path = node.attr('path') || id;
        var blobRecord = buffer.getBlob(id);
        rawArchive.resources[path] = {
          id: id,
          data: blobRecord.blob,
          encoding: 'blob',
          createdAt: Date.now(),
          updatedAt: Date.now()
        };
      });
    };

    
    PersistedDocumentArchive.prototype._exportChanges = function _exportChanges (sessions, buffer) {
      var rawArchive = {
        version: buffer.getVersion(),
        diff: buffer.getChanges(),
        resources: {}
      };
      this._exportManifest(sessions, buffer, rawArchive);
      this._exportDocuments(sessions, buffer, rawArchive);
      this._exportAssets(sessions, buffer, rawArchive);
      return rawArchive
    };

    return PersistedDocumentArchive;
  }(EventEmitter));

  function sendRequest(params, cb) {
    return new Promise(function(resolve, reject) {
      var method = (params.method || 'GET').toUpperCase();
      var url = params.url;
      if (['GET', 'POST', 'PUT', 'DELETE'].indexOf(method) < 0) {
        throw new Error("Parameter 'method' must be 'GET', 'POST', 'PUT', or 'DELETE'.")
      }
      if (!url) {
        throw new Error("Parameter 'url' is required.")
      }
      var xmlhttp = new XMLHttpRequest();
      xmlhttp.onreadystatechange = function() {
        
        
        if (xmlhttp.readyState === 4) { return _done() }
      };
      xmlhttp.open(method, url, true);
      if (params.header) {
        forEach(params.header, function(val, key) {
          xmlhttp.setRequestHeader(key, val);
        });
      }
      if (params.data instanceof FormData) { 
        xmlhttp.send(params.data);
      } else if (params.data) {
        xmlhttp.send(JSON.stringify(params.data));
      } else {
        xmlhttp.send();
      }

      function _done() {
        if (xmlhttp.status === 200) {
          var response = xmlhttp.responseText;
          if (cb) { cb(null, response); }
          resolve(response);
        } else {
          console.error(xmlhttp.statusText);
          var err = new Error(xmlhttp.statusText);
          if (cb) { cb(err); }
          reject(err);
        }
      }
    })
  }

  var HttpStorageClient = function HttpStorageClient(apiUrl) {
    this.apiUrl = apiUrl;
  };


  HttpStorageClient.prototype.read = function read (archiveId) {
    var url = this.apiUrl;
    if (archiveId) {
      url = url + '/' + archiveId;
    }
    return sendRequest({
      method: 'GET',
      url: url
    }).then(function (response) {
      return JSON.parse(response)
    })
  };

  HttpStorageClient.prototype.write = function write (archiveId, data) {
    var form = new FormData();
    forEach(data.resources, function (record, filePath) {
      if (record.encoding === 'blob') {
        
        form.append(record.id, record.data, filePath);
        delete record.data;
      }
    });
    form.append('_archive', JSON.stringify(data));
    var url = this.apiUrl;
    if (archiveId) {
      url = url + '/' + archiveId;
    }
    return sendRequest({
      method: 'PUT',
      url: url,
      data: form
    })
  };

  var InMemoryDarBuffer = function InMemoryDarBuffer() {
    this._version = null;
    this._changes = [];
    this._isDirty = {};
    this._blobs = {};
  };

  InMemoryDarBuffer.prototype.getVersion = function getVersion () {
    return this._version
  };

  InMemoryDarBuffer.prototype.load = function load (archiveId) { 
    return Promise.resolve()
  };

  InMemoryDarBuffer.prototype.addChange = function addChange (docId, change) {
    
    if (change.ops.length === 0) { return }
    
    this._isDirty[docId] = true;
    this._changes.push({
      docId: docId, change: change
    });
  };

  InMemoryDarBuffer.prototype.hasPendingChanges = function hasPendingChanges () {
    return this._changes.length > 0
  };

  InMemoryDarBuffer.prototype.getChanges = function getChanges () {
    return this._changes.slice()
  };

  InMemoryDarBuffer.prototype.hasResourceChanged = function hasResourceChanged (docId) {
    return this._isDirty[docId]
  };

  InMemoryDarBuffer.prototype.hasBlob = function hasBlob (id) {
    return this._blobs[id]
  };

  InMemoryDarBuffer.prototype.addBlob = function addBlob (id, blob) {
    this._blobs[id] = blob;
  };

  InMemoryDarBuffer.prototype.getBlob = function getBlob (id) {
    return this._blobs[id]
  };

  InMemoryDarBuffer.prototype.reset = function reset (version) {
    this._version = version;
    this._changes = [];
    this._blobs = {};
    this._isDirty = {};
  };

  var VfsStorageClient = function VfsStorageClient(vfs, baseUrl) {
    this.vfs = vfs;

    
    this.baseUrl = baseUrl;
  };

  VfsStorageClient.prototype.read = function read (archiveId) {
    var rawArchive = _readRawArchive(this.vfs, archiveId, this.baseUrl);
    return Promise.resolve(rawArchive)
  };

  VfsStorageClient.prototype.write = function write (archiveId, data) { 
    console.error('Can not write on virtual file system');
    console.info('This would have been written:', data);
    return Promise.resolve(false)
  };

  function _readRawArchive(fs, archiveId, baseUrl) {
    if ( baseUrl === void 0 ) baseUrl = '';

    var manifestXML = fs.readFileSync((archiveId + "/manifest.xml"));
    var manifestSession = ManifestLoader.load(manifestXML);
    var manifest = manifestSession.getDocument();
    var docs = manifest.findAll('documents > document');
    var assets = manifest.findAll('assets > asset');
    var rawArchive = {
      version: "0",
      resources: {
        'manifest.xml': {
          encoding: 'utf8',
          data: manifestXML
        }
      }
    };

    docs.forEach(function (entry) {
      var path = entry.attr('path');
      if (fs.existsSync((archiveId + "/" + (entry.path)))) {
        var content = fs.readFileSync((archiveId + "/" + (entry.path)));
        rawArchive.resources[path] = {
          encoding: 'utf8',
          data: content
        };
      } else {
        console.warn((archiveId + "/" + (entry.path) + " not found in vfs"));
      }
    });
    assets.forEach(function (asset) {
      var path = asset.attr('path');
      
      rawArchive.resources[path] = {
        encoding: 'url',
        data: baseUrl+archiveId+'/'+path
      };
    });
    return rawArchive
  }

  function findParentDOMElement(nativeEl) {
    while(nativeEl) {
      var el = DefaultDOMElement.unwrap(nativeEl);
      if (el) { return el }
      nativeEl = nativeEl.parentNode;
    }
  }

  function findParent(el, selector) {
    while(el) {
      if (el.is(selector)) { return el }
      el = el.getParent();
    }
  }

  function stop(event) {
    event.stopPropagation();
  }

  function stopAndPrevent(event) {
    event.stopPropagation();
    event.preventDefault();
  }

  function walk(el, cb) {
    _walk(el, cb, 0);
  }

  function _walk(el, cb, level) {
    cb(el, level);
    if (el.getChildCount() > 0) {
      var it = el.getChildNodeIterator();
      while(it.hasNext()) {
        _walk(it.next(), cb, level+1);
      }
    }
  }

  function isRightButton(event) {
    var isRightButton = false;
    if ('which' in event) {
      isRightButton = (event.which === 3);
    } else if ('button' in event) {
      isRightButton = (event.button === 2);
    }
    return isRightButton
  }

  function getBoundingRect(el) {
    var _rect = el.getNativeElement().getBoundingClientRect();
    return {
      top: _rect.top,
      left: _rect.left,
      height: _rect.height,
      width: _rect.width
    }
  }

  function getBoundingRectForRects() {
    var rects = [], len = arguments.length;
    while ( len-- ) rects[ len ] = arguments[ len ];

    var top, left, bottom, right;
    if (rects.length > 0) {
      var first = rects[0];
      top = first.top;
      left = first.left;
      bottom = top + first.height;
      right = left + first.width;
      for (var i = 1; i < rects.length; i++) {
        var r = rects[i];
        top = Math.min(top, r.top);
        left = Math.min(left, r.left);
        bottom = Math.max(bottom, r.top+r.height);
        right = Math.max(right, r.left+r.width);
      }
    }
    return {
      top: top,
      left: left,
      right: right,
      bottom: bottom,
      height: bottom-top,
      width: right-left
    }
  }

  function isXInside(x, rect) {
    return x >= rect.left && x <= rect.left+rect.width
  }

  function isYInside(y, rect) {
    return y >= rect.top && y <= rect.top+rect.height
  }

  var domHelpers = /*#__PURE__*/Object.freeze({
    findParentDOMElement: findParentDOMElement,
    findParent: findParent,
    stop: stop,
    stopAndPrevent: stopAndPrevent,
    walk: walk,
    isRightButton: isRightButton,
    getBoundingRect: getBoundingRect,
    getBoundingRectForRects: getBoundingRectForRects,
    isXInside: isXInside,
    isYInside: isYInside
  });

  var BlockNode = (function (DocumentNode$$1) {
  	function BlockNode () {
  		DocumentNode$$1.apply(this, arguments);
  	}if ( DocumentNode$$1 ) BlockNode.__proto__ = DocumentNode$$1;
  	BlockNode.prototype = Object.create( DocumentNode$$1 && DocumentNode$$1.prototype );
  	BlockNode.prototype.constructor = BlockNode;

  	

  	return BlockNode;
  }(DocumentNode));

  BlockNode.isBlock = true;

  var ChangeRecorder = (function (EditingInterface$$1) {
    function ChangeRecorder(doc) {
      EditingInterface$$1.call(this, doc.clone());
    }

    if ( EditingInterface$$1 ) ChangeRecorder.__proto__ = EditingInterface$$1;
    ChangeRecorder.prototype = Object.create( EditingInterface$$1 && EditingInterface$$1.prototype );
    ChangeRecorder.prototype.constructor = ChangeRecorder;

    ChangeRecorder.prototype.generateChange = function generateChange () {
      var doc = this.getDocument();
      var ops = doc._ops.slice();
      doc._ops.length = 0;
      var change = new DocumentChange(ops, {}, {});
      change._extractInformation(doc);
      return change
    };

    return ChangeRecorder;
  }(EditingInterface));

  function encodeXMLEntities(str) {
    return String(str)
      .replace(/&/g, '&amp;')
      .replace(/</g, '&lt;')
      .replace(/>/g, '&gt;')
      .replace(/"/g, '&quot;')
  }

  var DOMExporter = function DOMExporter(config, context) {
    this.context = context || {};
    if (!config.converters) {
      throw new Error('config.converters is mandatory')
    }
    if (!config.converters._isRegistry) {
      this.converters = new Registry();
      config.converters.forEach(function(Converter) {
        var converter = isFunction$1(Converter) ? new Converter() : Converter;
        if (!converter.type) {
          console.error('Converter must provide the type of the associated node.', converter);
          return
        }
        this.converters.add(converter.type, converter);
      }.bind(this));
    } else {
      this.converters = config.converters;
    }

    this.state = {
      doc: null
    };
    this.config = config;
    
    
    this._elementFactory = config.elementFactory;
    if (!this._elementFactory) {
      throw new Error("'elementFactory' is mandatory")
    }
    this.$$ = this.createElement.bind(this);
  };

  DOMExporter.prototype.exportDocument = function exportDocument (doc) {
    
    
    
    
    return this.convertDocument(doc)
  };


  DOMExporter.prototype.convertDocument = function convertDocument (doc) { 
    throw new Error('This method is abstract')
  };

  DOMExporter.prototype.convertContainer = function convertContainer (container) {
      var this$1 = this;

    if (!container) {
      throw new Error('Illegal arguments: container is mandatory.')
    }
    var doc = container.getDocument();
    this.state.doc = doc;
    var elements = [];
    container.getContent().forEach(function (id) {
      var node = doc.get(id);
      var nodeEl = this$1.convertNode(node);
      elements.push(nodeEl);
    });
    return elements
  };

  DOMExporter.prototype.convertNode = function convertNode (node) {
    if (isString(node)) {
      
      node = this.state.doc.get(node);
    } else {
      this.state.doc = node.getDocument();
    }
    var converter = this.getNodeConverter(node);
    
    
    if (node.isPropertyAnnotation() && (!converter || !converter.export)) {
      return this._convertPropertyAnnotation(node)
    }
    if (!converter) {
      converter = this.getDefaultBlockConverter();
    }
    var el;
    if (converter.tagName) {
      el = this.$$(converter.tagName);
    } else {
      el = this.$$('div');
    }
    el.attr(this.config.idAttribute, node.id);
    if (converter.export) {
      el = converter.export(node, el, this) || el;
    } else {
      el = this.getDefaultBlockConverter().export(node, el, this) || el;
    }
    return el
  };

  DOMExporter.prototype.convertProperty = function convertProperty (doc, path, options) {
    this.initialize(doc, options);
    var wrapper = this.$$('div')
      .append(this.annotatedText(path));
    return wrapper.innerHTML
  };

  DOMExporter.prototype.annotatedText = function annotatedText (path) {
    var doc = this.state.doc;
    var text = doc.get(path);
    var annotations = doc.getIndex('annotations').get(path);
    return this._annotatedText(text, annotations)
  };

  DOMExporter.prototype.getNodeConverter = function getNodeConverter (node) {
    return this.converters.get(node.type)
  };

  DOMExporter.prototype.getDefaultBlockConverter = function getDefaultBlockConverter () {
    throw new Error('This method is abstract.')
  };

  DOMExporter.prototype.getDefaultPropertyAnnotationConverter = function getDefaultPropertyAnnotationConverter () {
    throw new Error('This method is abstract.')
  };

  DOMExporter.prototype.getDocument = function getDocument () {
    return this.state.doc
  };

  DOMExporter.prototype.createElement = function createElement (str) {
    return this._elementFactory.createElement(str)
  };

  DOMExporter.prototype._annotatedText = function _annotatedText (text, annotations) {
    var self = this;

    var annotator = new Fragmenter();
    annotator.onText = function(context, text) {
      context.children.push(encodeXMLEntities(text));
    };
    annotator.onEnter = function(fragment) {
      var anno = fragment.node;
      return {
        annotation: anno,
        children: []
      }
    };
    annotator.onExit = function(fragment, context, parentContext) {
      var anno = context.annotation;
      var converter = self.getNodeConverter(anno);
      if (!converter) {
        converter = self.getDefaultPropertyAnnotationConverter();
      }
      var el;
      if (converter.tagName) {
        el = this.$$(converter.tagName);
      } else {
        el = this.$$('span');
      }
      el.attr(this.config.idAttribute, anno.id);
      el.append(context.children);
      if (converter.export) {
        el = converter.export(anno, el, self) || el;
      }
      parentContext.children.push(el);
    }.bind(this);
    var wrapper = { children: [] };
    annotator.start(wrapper, text, annotations);
    return wrapper.children
  };


  DOMExporter.prototype._convertPropertyAnnotation = function _convertPropertyAnnotation (anno) {
    
    var wrapper = this.$$('div').append(this.annotatedText(anno.path));
    var el = wrapper.find('['+this.config.idAttribute+'="'+anno.id+'"]');
    return el
  };

  var HTMLExporter = (function (DOMExporter$$1) {
    function HTMLExporter(config, context) {
      DOMExporter$$1.call(this, _defaultConfig(config), context);
    }

    if ( DOMExporter$$1 ) HTMLExporter.__proto__ = DOMExporter$$1;
    HTMLExporter.prototype = Object.create( DOMExporter$$1 && DOMExporter$$1.prototype );
    HTMLExporter.prototype.constructor = HTMLExporter;

    HTMLExporter.prototype.exportDocument = function exportDocument (doc) {
      var htmlEl = DefaultDOMElement.parseHTML('<html><head></head><body></body></html>');
      return this.convertDocument(doc, htmlEl)
    };

    HTMLExporter.prototype.getDefaultBlockConverter = function getDefaultBlockConverter () {
      return defaultBlockConverter 
    };

    HTMLExporter.prototype.getDefaultPropertyAnnotationConverter = function getDefaultPropertyAnnotationConverter () {
      return defaultAnnotationConverter 
    };

    return HTMLExporter;
  }(DOMExporter));

  function _defaultConfig(config) {
    config = Object.assign({
      idAttribute: 'data-id'
    }, config);
    if (!config.elementFactory) {
      config.elementFactory = DefaultDOMElement.createDocument('html');
    }
    return config
  }


  var defaultAnnotationConverter = {
    tagName: 'span',
    export: function(node, el) {
      el.tagName = 'span';
      el.attr('data-type', node.type);
      var properties = node.toJSON();
      forEach(properties, function(value, name) {
        if (name === 'id' || name === 'type') { return }
        if (isString(value) || isNumber(value) || isBoolean(value)) {
          el.attr('data-'+name, value);
        }
      });
    }
  };

  var defaultBlockConverter = {
    export: function(node, el, converter) {
      el.attr('data-type', node.type);
      var properties = node.toJSON();
      forEach(properties, function(value, name) {
        if (name === 'id' || name === 'type') {
          return
        }
        var prop = converter.$$('div').attr('property', name);
        if (node.getPropertyType(name) === 'string' && value) {
          prop.append(converter.annotatedText([node.id, name]));
        } else {
          prop.text(value);
        }
        el.append(prop);
      });
    }
  };

  var ClipboardExporter = (function (HTMLExporter$$1) {
    function ClipboardExporter () {
      HTMLExporter$$1.apply(this, arguments);
    }

    if ( HTMLExporter$$1 ) ClipboardExporter.__proto__ = HTMLExporter$$1;
    ClipboardExporter.prototype = Object.create( HTMLExporter$$1 && HTMLExporter$$1.prototype );
    ClipboardExporter.prototype.constructor = ClipboardExporter;

    ClipboardExporter.prototype.exportDocument = function exportDocument (doc) {
      this.state.doc = doc;
      var html;
      var elements = this.convertDocument(doc);
      
      if (elements.length === 1 && elements[0].attr('data-id') === TEXT_SNIPPET_ID) {
        html = elements[0].innerHTML;
      } else {
        html = elements.map(function(el) {
          return el.outerHTML
        }).join('');
      }
      var jsonConverter = new JSONConverter();
      var jsonStr = JSON.stringify(jsonConverter.exportDocument(doc));
      var substanceContent = "<script id=\"substance-clipboard\" type=\"application/json\">" + jsonStr + "</script>";
      return '<html><head>' +substanceContent+ '</head><body>' + html + '</body></html>'
    };

    
    ClipboardExporter.prototype.convertDocument = function convertDocument (doc) {
      var content = doc.get(SNIPPET_ID);
      if (!content) {
        throw new Error('Illegal clipboard document: could not find container "' + SNIPPET_ID + '"')
      }
      return this.convertContainer(content)
    };

    return ClipboardExporter;
  }(HTMLExporter));

  var HTMLImporter = (function (DOMImporter$$1) {
    function HTMLImporter(config) {
      DOMImporter$$1.call(this, Object.assign({ idAttribute: 'data-id' }, config));

      
      this._el = DefaultDOMElement.parseHTML('<html></html>');
    }

    if ( DOMImporter$$1 ) HTMLImporter.__proto__ = DOMImporter$$1;
    HTMLImporter.prototype = Object.create( DOMImporter$$1 && DOMImporter$$1.prototype );
    HTMLImporter.prototype.constructor = HTMLImporter;

    HTMLImporter.prototype.importDocument = function importDocument (html) {
      this.reset();
      var parsed = DefaultDOMElement.parseHTML(html);
      this.convertDocument(parsed);
      return this.state.doc
    };

    
    HTMLImporter.prototype.convertDocument = function convertDocument (documentEl) { 
      throw new Error('This method is abstract')
    };

    return HTMLImporter;
  }(DOMImporter));

  var INLINENODES = ['a','b','big','i','small','tt','abbr','acronym','cite','code','dfn','em','kbd','strong','samp','time','var','bdo','br','img','map','object','q','script','span','sub','sup','button','input','label','select','textarea'].reduce(function (m,n){m[n]=true;return m}, {});


  var ClipboardImporter = (function (HTMLImporter$$1) {
    function ClipboardImporter(config) {
      HTMLImporter$$1.call(this, _withCatchAllConverter(config));
      
      this.IGNORE_DEFAULT_WARNINGS = true;

      Object.assign(config, {
        trimWhitespaces: true,
        REMOVE_INNER_WS: true
      });

      this.editorOptions = config.editorOptions;
    }

    if ( HTMLImporter$$1 ) ClipboardImporter.__proto__ = HTMLImporter$$1;
    ClipboardImporter.prototype = Object.create( HTMLImporter$$1 && HTMLImporter$$1.prototype );
    ClipboardImporter.prototype.constructor = ClipboardImporter;

    
    ClipboardImporter.prototype.importDocument = function importDocument (html) {
      if (platform.isWindows) {
        
        
        var match = /<!--StartFragment-->(.*)<!--EndFragment-->/.exec(html);
        if (match) {
          html = match[1];
        }
      }

      
      
      
      if (html.search(/script id=.substance-clipboard./)>=0) {
        var htmlDoc$1 = DefaultDOMElement.parseHTML(html);
        var substanceData = htmlDoc$1.find('#substance-clipboard');
        if (substanceData) {
          var jsonStr = substanceData.textContent;
          try {
            return this.importFromJSON(jsonStr)
          } finally {
            
          }
        }
      }

      if (this.editorOptions && this.editorOptions['forcePlainTextPaste']) {
        return null;
      }

      var htmlDoc = DefaultDOMElement.parseHTML(html);
      var generatorMeta = htmlDoc.find('meta[name="generator"]');
      var xmnlsw = htmlDoc.find('html').getAttribute('xmlns:w');
      if(generatorMeta) {
        var generator = generatorMeta.getAttribute('content');
        if(generator.indexOf('LibreOffice') > -1) { this._isLibreOffice = true; }
      } else if(xmnlsw) {
        if(xmnlsw.indexOf('office:word') > -1) {
          this._isMicrosoftWord = true;
          
          
          
          var match$1 = /<!--StartFragment-->([\s\S]*)<!--EndFragment-->/.exec(html);
          if (match$1) {
            htmlDoc = DefaultDOMElement.parseHTML(match$1[1]);
          }
        }
      } else if(html.indexOf('docs-internal-guid') > -1) {
        this._isGoogleDoc = true;
      }

      var body = htmlDoc.find('body');
      body = this._sanitizeBody(body);
      if (!body) {
        console.warn('Invalid HTML.');
        return null
      }
      this._wrapIntoParagraph(body);
      this.reset();
      this.convertBody(body);
      var doc = this.state.doc;
      return doc
    };

    ClipboardImporter.prototype._sanitizeBody = function _sanitizeBody (body) {
      
      body.findAll('meta').forEach(function (el) { return el.remove(); });

      
      
      if(this._isLibreOffice || this._isMicrosoftWord) {
        var bodyHtml = body.getInnerHTML();
        body.setInnerHTML(bodyHtml.replace(/\r\n|\r|\n/g, ' '));
      } else if (this._isGoogleDoc) {
        body = this._fixupGoogleDocsBody(body);
      }

      return body
    };

    ClipboardImporter.prototype._fixupGoogleDocsBody = function _fixupGoogleDocsBody (body) {
      if (!body) { return }
      
      
      
      
      var bold = body.find('b');
      if (bold && /^docs-internal/.exec(bold.id)) {
        body = bold;
      }

      body.findAll('span').forEach(function (span) {
        
        
        
        
        
        
        
        
        var nodeTypes = [];
        if(span.getStyle('font-weight') === '700') { nodeTypes.push('b'); }
        if(span.getStyle('font-style') === 'italic') { nodeTypes.push('i'); }
        if(span.getStyle('vertical-align') === 'super') { nodeTypes.push('sup'); }
        if(span.getStyle('vertical-align') === 'sub') { nodeTypes.push('sub'); }

        function createInlineNodes(parentEl, isRoot) {
          if(nodeTypes.length > 0) {
            var el = parentEl.createElement(nodeTypes[0]);
            if(nodeTypes.length === 1) { el.append(span.textContent); }

            if(isRoot) {
              parentEl.replaceChild(span, el);
            } else {
              parentEl.appendChild(el);
            }

            nodeTypes.shift();
            createInlineNodes(el);
          }
        }

        createInlineNodes(span.getParent(), true);
      });

      var tags = ['b', 'i', 'sup', 'sub'];

      tags.forEach(function (tag) {
        body.findAll(tag).forEach(function (el) {
          
          
          var previousSiblingEl = el.getPreviousSibling();
          if(previousSiblingEl && el.tagName === previousSiblingEl.tagName) {
            var parentEl = el.getParent();
            var newEl = parentEl.createElement(tag);
            newEl.setInnerHTML(previousSiblingEl.getInnerHTML() + el.getInnerHTML());
            parentEl.replaceChild(el, newEl);
            parentEl.removeChild(previousSiblingEl);
          }

          
          
          
          
          if(previousSiblingEl && previousSiblingEl.tagName && el.getChildCount() > 0 && el.getChildAt(0).tagName === previousSiblingEl.tagName) {
            var parentEl$1 = el.getParent();
            var childEl = el.getChildAt(0);
            var newEl$1 = parentEl$1.createElement(previousSiblingEl.tagName);
            var newChildEl = newEl$1.createElement(tag);
            newChildEl.setTextContent(childEl.textContent);
            newEl$1.appendChild(newChildEl);
            parentEl$1.replaceChild(el, newEl$1);
          }
        });
      });

      return body
    };

    ClipboardImporter.prototype._wrapIntoParagraph = function _wrapIntoParagraph (body) {
      var childNodes = body.getChildNodes();
      var shouldWrap = false;
      for (var i = 0; i < childNodes.length; i++) {
        var c = childNodes[i];
        if (c.isTextNode()) {
          if (!(/^\s+$/.exec(c.textContent))) {
            shouldWrap = true;
            break
          }
        } else if (INLINENODES[c.tagName]) {
          shouldWrap = true;
          break
        }
      }
      if (shouldWrap) {
        var p = body.createElement('p');
        p.append(childNodes);
        body.append(p);
      }
    };

    ClipboardImporter.prototype.importFromJSON = function importFromJSON (jsonStr) {
      this.reset();
      var doc = this.getDocument();
      var jsonData = JSON.parse(jsonStr);
      var converter = new JSONConverter();
      converter.importDocument(doc, jsonData);
      return doc
    };

    
    ClipboardImporter.prototype.convertBody = function convertBody (body) {
      this.convertContainer(body.childNodes, SNIPPET_ID);
    };

    
    ClipboardImporter.prototype._createDocument = function _createDocument () {
      var emptyDoc = HTMLImporter$$1.prototype._createDocument.call(this);
      return emptyDoc.createSnippet()
    };

    return ClipboardImporter;
  }(HTMLImporter));

  function _withCatchAllConverter(config) {
    config = Object.assign({}, config);
    var defaultTextType = config.schema.getDefaultTextType();
    if (defaultTextType) {
      config.converters = config.converters.concat([{
        type: defaultTextType,
        matchElement: function(el) { return el.is('div') },
        import: function(el, node, converter) {
          node.content = converter.annotatedText(el, [node.id, 'content']);
        }
      }]);
    }
    return config
  }

  var ContainerAdapter = (function (Container$$1) {
    function ContainerAdapter(doc, path) {
      Container$$1.call(this, doc, { id: String(path) });
      this.document = doc;
      this.path = path;

      
      
      doc.data.nodes[this.id] = this;
    }

    if ( Container$$1 ) ContainerAdapter.__proto__ = Container$$1;
    ContainerAdapter.prototype = Object.create( Container$$1 && Container$$1.prototype );
    ContainerAdapter.prototype.constructor = ContainerAdapter;

    var prototypeAccessors = { nodes: { configurable: true } };

    ContainerAdapter.prototype.getContentPath = function getContentPath () {
      return this.path
    };

    prototypeAccessors.nodes.get = function () {
      return this.document.get(this.path)
    };

    Object.defineProperties( ContainerAdapter.prototype, prototypeAccessors );

    return ContainerAdapter;
  }(Container));

  ContainerAdapter.prototype._isDocumentNode = false;
  ContainerAdapter.prototype._isContainer = false;

  var CoordinateAdapter = (function (Coordinate$$1) {
    function CoordinateAdapter(owner, pathProperty, offsetProperty) {
      Coordinate$$1.call(this, 'SKIP');

      this._owner = owner;
      this._pathProp = pathProperty;
      this._offsetProp = offsetProperty;
      Object.freeze(this);
    }

    if ( Coordinate$$1 ) CoordinateAdapter.__proto__ = Coordinate$$1;
    CoordinateAdapter.prototype = Object.create( Coordinate$$1 && Coordinate$$1.prototype );
    CoordinateAdapter.prototype.constructor = CoordinateAdapter;

    var prototypeAccessors = { path: { configurable: true },offset: { configurable: true } };

    CoordinateAdapter.prototype.equals = function equals (other) {
      return (other === this ||
        (isArrayEqual(other.path, this.path) && other.offset === this.offset) )
    };

    prototypeAccessors.path.get = function () {
      return this._owner[this._pathProp];
    };

    prototypeAccessors.path.set = function (path) {
      this._owner[this._pathProp] = path;
    };

    prototypeAccessors.offset.get = function () {
      return this._owner[this._offsetProp];
    };

    prototypeAccessors.offset.set = function (offset) {
      this._owner[this._offsetProp] = offset;
    };

    CoordinateAdapter.prototype.toJSON = function toJSON () {
      return {
        path: this.path,
        offset: this.offset
      }
    };

    CoordinateAdapter.prototype.toString = function toString () {
      return "(" + this.path.join('.') + ", " + this.offset + ")"
    };

    Object.defineProperties( CoordinateAdapter.prototype, prototypeAccessors );

    return CoordinateAdapter;
  }(Coordinate));

  function createDocumentFactory(ArticleClass, create) {
    return {
      ArticleClass: ArticleClass,
      createEmptyArticle: function() {
        var doc = new ArticleClass();
        return doc
      },
      createArticle: function() {
        var doc = new ArticleClass();
        create(doc);
        return doc
      },
      createChangeset: function() {
        var doc = new ArticleClass();
        var tx = new ChangeRecorder(doc);
        create(tx);
        var change = tx.generateChange();
        return [change.toJSON()]
      }
    }
  }

  var DefaultChangeCompressor = function DefaultChangeCompressor () {};

  DefaultChangeCompressor.prototype.shouldMerge = function shouldMerge (lastChange, newChange) {
    return false
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    

    
  };


  DefaultChangeCompressor.prototype.merge = function merge (first, second) {
    
    
    var firstOp = first.ops[0];
    var secondOp = second.ops[0];
    var firstDiff = firstOp.diff;
    var secondDiff = secondOp.diff;
    var mergedOp = false;
    if (firstDiff.isInsert()) {
      if (firstDiff.pos+firstDiff.getLength() === secondDiff.pos) {
        mergedOp = firstOp.toJSON();
        mergedOp.diff.str += secondDiff.str;
      }
    }
    else if (firstDiff.isDelete()) {
      
      
      if (firstDiff.pos === secondDiff.pos) {
        mergedOp = firstOp.toJSON();
        mergedOp.diff.str += secondDiff.str;
      } else if (secondDiff.pos+secondDiff.getLength() === firstDiff.pos) {
        mergedOp = firstOp.toJSON();
        mergedOp.diff = secondDiff;
        mergedOp.diff.str += firstDiff.str;
      }
    }
    if (mergedOp) {
      first.ops[0] = ObjectOperation.fromJSON(mergedOp);
      if (first.ops.length > 1) {
        
        
        
        first.ops = first.ops.concat(second.ops.slice(1));
        first.after = second.after;
      }
      return true
    }
    return false
  };

  var FileProxy = function FileProxy(fileNode, context) {
    this.fileNode = fileNode;
    this.context = context;
    fileNode.setProxy(this);
  };

  var prototypeAccessors$8 = { id: { configurable: true } };

  prototypeAccessors$8.id.get = function () {
    return this.fileNode.id
  };


  FileProxy.prototype.triggerUpdate = function triggerUpdate () {
    var fileId = this.fileNode.id;
    this.context.editorSession.transaction(function (tx) {
      tx.set([fileId, '__changed__'], '');
    }, { history: false });
  };

  FileProxy.prototype.getUrl = function getUrl () {
    return ''
  };

  FileProxy.prototype.sync = function sync () {
    return Promise.reject(new Error('sync method not implemented'))
  };

  Object.defineProperties( FileProxy.prototype, prototypeAccessors$8 );

  FileProxy.match = function(fileNode, context) { 
    return false
  };

  var DefaultFileProxy = (function (FileProxy$$1) {
    function DefaultFileProxy(fileNode, context) {
      FileProxy$$1.call(this, fileNode, context);

      
      this.file = fileNode.sourceFile;
      if (this.file) {
        this._fileUrl = URL.createObjectURL(this.file);
      }
      this.url = fileNode.url;
    }

    if ( FileProxy$$1 ) DefaultFileProxy.__proto__ = FileProxy$$1;
    DefaultFileProxy.prototype = Object.create( FileProxy$$1 && FileProxy$$1.prototype );
    DefaultFileProxy.prototype.constructor = DefaultFileProxy;

    DefaultFileProxy.prototype.getUrl = function getUrl () {
      
      if (this.url) {
        return this.url
      }
      
      if (this._fileUrl) {
        return this._fileUrl
      }
      
      return ''
    };

    DefaultFileProxy.prototype.sync = function sync () {
      if (!this.url) {
        console.info('Simulating file upload. Creating blob url instead.', this._fileUrl);
        this.url = this._fileUrl;
      }
      return Promise.resolve()
    };

    return DefaultFileProxy;
  }(FileProxy));

  var FileNode = (function (DocumentNode$$1) {
    function FileNode() {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      DocumentNode$$1.apply(this, args);
    }

    if ( DocumentNode$$1 ) FileNode.__proto__ = DocumentNode$$1;
    FileNode.prototype = Object.create( DocumentNode$$1 && DocumentNode$$1.prototype );
    FileNode.prototype.constructor = FileNode;

    FileNode.prototype.getUrl = function getUrl () {
      if (this.proxy) {
        return this.proxy.getUrl()
      } else {
        
        console.warn('No file proxy attached to ', this.id);
        return ''
      }
    };

    FileNode.prototype.setProxy = function setProxy (proxy) {
      this.proxy = proxy;
    };

    return FileNode;
  }(DocumentNode));

  FileNode.type = 'file';

  FileNode.schema = {
    url: { type: 'string', optional: true },
    fileType: { type: 'string', optional: true },
    mimeType: { type: 'string', optional: true },
    sourceFile: { type: 'object', optional: true }
  };

  FileNode.prototype._isFileNode = true;
  FileNode._isFileNode = true;

  function getChangeFromDocument(doc) {
    var recorder = new ChangeRecorder(doc);
    return recorder.generateChange()
  }

  var InlineNode = (function (PropertyAnnotation$$1) {
  	function InlineNode () {
  		PropertyAnnotation$$1.apply(this, arguments);
  	}if ( PropertyAnnotation$$1 ) InlineNode.__proto__ = PropertyAnnotation$$1;
  	InlineNode.prototype = Object.create( PropertyAnnotation$$1 && PropertyAnnotation$$1.prototype );
  	InlineNode.prototype.constructor = InlineNode;

  	

  	return InlineNode;
  }(PropertyAnnotation));

  InlineNode.prototype._isInlineNode = true;

  InlineNode.isInline = true;

  var ERR_ABSTRACT = 'This method is abstract!';

  function ListMixin(DocumentNode) {

    var AbstractList = (function (DocumentNode) {
      function AbstractList () {
        DocumentNode.apply(this, arguments);
      }

      if ( DocumentNode ) AbstractList.__proto__ = DocumentNode;
      AbstractList.prototype = Object.create( DocumentNode && DocumentNode.prototype );
      AbstractList.prototype.constructor = AbstractList;

      var prototypeAccessors = { length: { configurable: true } };

      AbstractList.prototype.isList = function isList () {
        return true
      };

      AbstractList.prototype.createListItem = function createListItem (text) {
        throw new Error(ERR_ABSTRACT)
      };

      AbstractList.prototype.getItems = function getItems () {
        throw new Error(ERR_ABSTRACT)
      };

      AbstractList.prototype.getItemsPath = function getItemsPath () {
        throw new Error(ERR_ABSTRACT)
      };

      AbstractList.prototype.insertItemAt = function insertItemAt (pos, item) {
        throw new Error(ERR_ABSTRACT)
      };

      AbstractList.prototype.removeItemAt = function removeItemAt (pos) {
        throw new Error(ERR_ABSTRACT)
      };

      

      AbstractList.prototype.getItemAt = function getItemAt (idx) {
        return this.getItems()[idx]
      };

      AbstractList.prototype.getItemPosition = function getItemPosition (item) {
        return this.getItems().indexOf(item)
      };

      AbstractList.prototype.getLength = function getLength () {
        return this.getItems().length
      };

      AbstractList.prototype.getFirstItem = function getFirstItem () {
        return this.getItemAt(0)
      };

      AbstractList.prototype.getLastItem = function getLastItem () {
        return this.getItemAt(this.getLength()-1)
      };

      AbstractList.prototype.appendItem = function appendItem (item) {
        this.insertItemAt(this.items.length, item);
      };

      AbstractList.prototype.removeItem = function removeItem (item) {
        var pos = this.getItemPosition(item);
        if (pos >= 0) {
          this.removeItemAt(pos);
        }
      };

      AbstractList.prototype.isEmpty = function isEmpty () {
        return this.getLength() === 0
      };

      prototypeAccessors.length.get = function () {
        return this.getLength()
      };

      AbstractList.prototype.getListType = function getListType (level) {
        
        var idx = level-1;
        var listTypes = this._getListTypes();
        return listTypes[idx] || 'bullet'
      };

      AbstractList.prototype.setListType = function setListType (level, listType) {
        var idx = level-1;
        var listTypes = this._getListTypes();
        if (listTypes.length < level) {
          for (var i = 0; i < idx; i++) {
            if (!listTypes[i]) { listTypes[i] = 'bullet'; }
          }
        }
        listTypes[idx] = listType;
        this._setListTypes(listTypes);
      };

      AbstractList.prototype._getListTypes = function _getListTypes () {
        var listTypeString = this.getListTypeString();
        return listTypeString ? listTypeString.split(',').map(function (s) { return s.trim(); }) : []
      };

      AbstractList.prototype._setListTypes = function _setListTypes (listTypeString) {
        if (isArray(listTypeString)) {
          listTypeString = listTypeString.join(',');
        }
        var oldListTypeString = this.getListTypeString();
        if (oldListTypeString !== listTypeString) {
          this.setListTypeString(listTypeString);
        }
      };

      Object.defineProperties( AbstractList.prototype, prototypeAccessors );

      return AbstractList;
    }(DocumentNode));

    return AbstractList
  }

  var Range = function Range(start, end, reverse, containerId, surfaceId) {
    
    if (arguments[0] === 'SKIP') { return }
    if (arguments.length === 1 && isPlainObject(arguments[0])) {
      var data = arguments[0];
      this.start = data.start;
      this.end = data.end;
      this.reverse = Boolean(data.reverse);
      this.containerId = data.containerId;
      this.surfaceId = data.surfaceId;
    } else {
      this.start = start;
      this.end = end;
      this.reverse = Boolean(reverse);
      this.containerId = containerId;
      this.surfaceId = surfaceId;
    }
  };

  Range.prototype.isCollapsed = function isCollapsed () {
    return this.start.equals(this.end)
  };

  Range.prototype.equals = function equals (other) {
    if (this === other) { return true }
    else {
      return (
        this.containerId === other.containerId &&
        this.start.equals(other.start) &&
        this.end.equals(other.end)
      )
    }
  };

  Range.prototype.isReverse = function isReverse () {
    return this.reverse
  };

  Range.prototype.toString = function toString () {
    var str = [this.start.toString(), '->', this.end.toString()];
    if (this.isReverse()) {
      str.push('[reverse]');
    }
    if (this.containerId) {
      str.push('[container='+this.containerId+']');
    }
    if (this.surfaceId) {
      str.push('[surface='+this.surfaceId+']');
    }
    return str.join('')
  };

  Range.prototype._isRange = true;

  var TextNode = (function (superclass) {
    function TextNode () {
      superclass.apply(this, arguments);
    }

    if ( superclass ) TextNode.__proto__ = superclass;
    TextNode.prototype = Object.create( superclass && superclass.prototype );
    TextNode.prototype.constructor = TextNode;

    TextNode.prototype.getPath = function getPath () {
      return [this.id, 'content']
    };

    TextNode.prototype.getText = function getText () {
      return this.content
    };

    return TextNode;
  }(TextNodeMixin(DocumentNode)));

  TextNode.isText = true;

  TextNode.schema = {
    type: "text",
    content: "text",
    direction: { type: "string", optional: true },
    textAlign: { type: "string", default: 'left' }
  };

  var TextBlock = (function (TextNode$$1) {
  	function TextBlock () {
  		TextNode$$1.apply(this, arguments);
  	}if ( TextNode$$1 ) TextBlock.__proto__ = TextNode$$1;
  	TextBlock.prototype = Object.create( TextNode$$1 && TextNode$$1.prototype );
  	TextBlock.prototype.constructor = TextBlock;

  	

  	return TextBlock;
  }(TextNode));

  TextBlock.isBlock = true;

  var XMLExporter = (function (DOMExporter$$1) {
    function XMLExporter(config, context) {
      DOMExporter$$1.call(this, _defaultConfig$1(config), context);
    }

    if ( DOMExporter$$1 ) XMLExporter.__proto__ = DOMExporter$$1;
    XMLExporter.prototype = Object.create( DOMExporter$$1 && DOMExporter$$1.prototype );
    XMLExporter.prototype.constructor = XMLExporter;

    XMLExporter.prototype.getDefaultBlockConverter = function getDefaultBlockConverter () {
      return defaultBlockConverter$1 
    };

    XMLExporter.prototype.getDefaultPropertyAnnotationConverter = function getDefaultPropertyAnnotationConverter () {
      return defaultAnnotationConverter$1 
    };

    return XMLExporter;
  }(DOMExporter));

  function _defaultConfig$1(config) {
    config = Object.assign({
      idAttribute: 'id'
    }, config);
    if (!config.elementFactory) {
      config.elementFactory = DefaultDOMElement.createDocument('xml');
    }
    return config
  }

  var defaultAnnotationConverter$1 = {
    tagName: 'annotation',
    export: function(node, el) {
      el.attr('type', node.type);
      var properties = node.toJSON();
      forEach(properties, function(value, name) {
        if (name === 'id' || name === 'type') { return }
        if (isString(value) || isNumber(value) || isBoolean(value)) {
          el.attr(name, value);
        }
      });
    }
  };

  var defaultBlockConverter$1 = {
    tagName: 'block',
    export: function(node, el, converter) {
      el.attr('type', node.type);
      var properties = node.toJSON();
      forEach(properties, function(value, name) {
        if (name === 'id' || name === 'type') {
          return
        }
        var prop = converter.$$(name);
        if (node.getPropertyType(name) === 'string') {
          prop.append(converter.annotatedText([node.id, name]));
        } else {
          prop.text(value);
        }
        el.append(prop);
      });
    }
  };

  var XMLImporter = (function (DOMImporter$$1) {
    function XMLImporter(config, context) {
      DOMImporter$$1.call(this, Object.assign({ idAttribute: 'id' }, config), context);
    }

    if ( DOMImporter$$1 ) XMLImporter.__proto__ = DOMImporter$$1;
    XMLImporter.prototype = Object.create( DOMImporter$$1 && DOMImporter$$1.prototype );
    XMLImporter.prototype.constructor = XMLImporter;

    XMLImporter.prototype.importDocument = function importDocument (xml) {
      this.reset();
      var dom = DefaultDOMElement.parseXML(xml);
      this.convertDocument(dom);
      return this.state.doc
    };

    XMLImporter.prototype.convertDocument = function convertDocument (xmlDocument) {
      var rootNode = xmlDocument.children[0];
      if (!rootNode) { throw new Error('XML Root node could not be found.') }
      this.convertElement(rootNode);
    };

    return XMLImporter;
  }(DOMImporter));

  var AbstractClipboard = function AbstractClipboard(configurator) {
    var schema = configurator.getSchema();
    var converterRegistry = configurator.converterRegistry;
    var htmlConverters = [];
    if (converterRegistry && converterRegistry.contains('html')) {
      htmlConverters = converterRegistry.get('html').values() || [];
    }
    
    
    this._config = {
      schema: schema,
      DocumentClass: schema.getDocumentClass(),
      converters: htmlConverters,
      editorOptions: configurator.getEditorOptions()
    };
  };


  AbstractClipboard.prototype.onCopy = function onCopy (event) {
    
    var clipboardData = this._copy();
    substanceGlobals._clipboardData = event.clipboardData;

    if (event.clipboardData && clipboardData.doc) {
      event.preventDefault();
      
      event.clipboardData.setData('text/plain', clipboardData.text);
      
      if (!platform.isIE && !platform.isEdge) {
        event.clipboardData.setData('text/html', clipboardData.html);
      }
    }
  };


  AbstractClipboard.prototype.onCut = function onCut (event) {
    
    event.preventDefault();
    
    this.onCopy(event);
    this._cut();
  };



  AbstractClipboard.prototype.onPaste = function onPaste (event) {
    var clipboardData = event.clipboardData;

    var types = {};
    for (var i = 0; i < clipboardData.types.length; i++) {
      types[clipboardData.types[i]] = true;
    }
    

    event.preventDefault();
    event.stopPropagation();

    var plainText;
    var html;
    if (types['text/plain']) {
      plainText = clipboardData.getData('text/plain');
    }
    if (types['text/html']) {
      html = clipboardData.getData('text/html');
    }

    
    
    if (platform.isEdge &&
        substanceGlobals.clipboardData &&
        substanceGlobals.clipboardData.text === plainText) {
      html = substanceGlobals.clipboardData.html;
    } else {
      substanceGlobals.clipboardData = {
        text: plainText,
        html: html
      };
    }

    

    
    
    if (platform.isFF && !html) {
      this._pastePlainText(plainText);
      return
    }

    
    
    if (html) {
      if (!this._pasteHtml(html, plainText)) {
        this._pastePlainText(plainText);
      }
    } else {
      this._pastePlainText(plainText);
    }
  };

  AbstractClipboard.prototype._getImporter = function _getImporter () {
    return new ClipboardImporter(this._config)
  };

  AbstractClipboard.prototype._getExporter = function _getExporter () {
    return new ClipboardExporter(this._config)
  };


  AbstractClipboard.prototype._copy = function _copy () {
    throw new Error('This method is abstract.')
  };


  AbstractClipboard.prototype._cut = function _cut () {
    throw new Error('This method is abstract.')
  };


  AbstractClipboard.prototype._pastePlainText = function _pastePlainText (plainText) { 
    throw new Error('This method is abstract.')
  };


  AbstractClipboard.prototype._pasteHtml = function _pasteHtml (html, text) { 
    throw new Error('This method is abstract.')
  };

  var ResourceManager = function ResourceManager(editorSession, context) {
    this.editorSession = editorSession;
    this.context = context;
    this.editorSession.onRender('document', this._onDocumentChange, this);
  };

  ResourceManager.prototype.dispose = function dispose () {
    this.editorSession.off(this);
  };


  ResourceManager.prototype.triggerFetch = function triggerFetch (resource) {
      var this$1 = this;

    resource.fetchPayload(this.context, function (err, props) {
      if (err) {
        this$1._updateNode(resource.id, {
          errorMessage: err.toString()
        });
      } else {
        this$1._updateNode(resource.id, props);
      }
    });
  };

  ResourceManager.prototype._onDocumentChange = function _onDocumentChange (change) {
      var this$1 = this;

    var doc = this.editorSession.getDocument();
    forEach(change.created, function (_, id) {
      var node = doc.get(id);
      if (node.constructor.isResource) {
        setTimeout(function () {
          this$1.triggerFetch(node);
        });
      }
    });
  };


  ResourceManager.prototype._updateNode = function _updateNode (nodeId, props) {
    var editorSession = this.editorSession;
    editorSession.transaction(function (tx) {
      forEach(props, function (val, key) {
        tx.set([nodeId, key], val);
      });
    });
  };

  function levenshtein(a, b){
    var m = [];
    for(var i = 0; i <= b.length; i++) {
      m[i] = [i];
      if(i === 0) { continue; }
      var ib = i-1;
      for(var j = 0; j <= a.length; j++){
        m[0][j] = j;
        if(j === 0) { continue; }
        var jb = j-1;
        m[i][j] = b.charAt(ib) === a.charAt(jb) ? m[ib][jb] : Math.min(
          m[ib][jb]+1,
          m[i][jb]+1,
          m[ib][j]+1
        );
      }
    }
    return m
  }

  function diff(a, b, offset) {
    if (!isString(a) || !isString(b)) {
      throw new Error('Illegal arguments.')
    }
    offset = offset || 0;
    var changes = [];
    if (a || b) {
      if (!a && b) {
        changes.push({ type:'insert', start:offset, text:b });
      } else if (a && !b) {
        changes.push({ type:'delete', start:offset, end:offset+a.length });
      } else {
        var m = levenshtein(a, b);
        changes = _diff(a, b, m, offset);
      }
    }
    return changes
  }

  function _diff(a, b, m, offset) {
    var i = b.length;
    var j = a.length;
    var changes = [];
    var current;
    while (i>0 && j>0) {
      _next();
    }
    _commit();
    return changes

    function _next() {
      var d = m[i][j];
      var ib = i-1;
      var jb = j-1;
      
      if (m[ib][jb]<d) {
        if (current && current.type === 'replace') {
          current.start--;
          current.text.unshift(b.charAt(ib));
        } else {
          _commit();
          current = { type:'replace', start:jb, end:j, text:[b.charAt(ib)] };
        }
        i--;
        j--;
      }
      
      else if (m[ib][j]<d) {
        if (current && current.type === 'insert') {
          current.start--;
          current.text.unshift(b.charAt(ib));
        } else {
          _commit();
          current = { type:'insert', start:jb, text:[b.charAt(ib)] };
        }
        i--;
      }
      
      else if (m[i][jb]<d) {
        if (current && current.type === 'delete') {
          current.start--;
        } else {
          _commit();
          current = { type:'delete', start:jb, end:j };
        }
        j--;
      }
      
      else {
        _commit();
        i--;
        j--;
      }
    }

    function _commit() {
      if (current) {
        switch (current.type) {
          case 'insert':
            current.start += offset;
            current.text = current.text.join('');
            break
          case 'delete':
            current.start += offset;
            current.end += offset;
            break
          case 'replace':
            current.start += offset;
            current.end += offset;
            current.text = current.text.join('');
            break
          default:
            throw new Error('Invalid state')
        }
        changes.push(current);
        current = null;
      }
    }

  }

  var AnnotationComponent = (function (Component$$1) {
    function AnnotationComponent () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) AnnotationComponent.__proto__ = Component$$1;
    AnnotationComponent.prototype = Object.create( Component$$1 && Component$$1.prototype );
    AnnotationComponent.prototype.constructor = AnnotationComponent;

    AnnotationComponent.prototype.didMount = function didMount () {
      var node = this.props.node;
      node.on('highlighted', this.onHighlightedChanged, this);
    };

    
    AnnotationComponent.prototype.dispose = function dispose () {
      var node = this.props.node;
      node.off(this);
    };

    AnnotationComponent.prototype.render = function render ($$) {
      var el = $$(this.getTagName())
        .attr("data-id", this.props.node.id)
        .addClass(this.getClassNames());
      if (this.props.node.highlighted) {
        el.addClass('sm-highlighted');
      }
      el.append(this.props.children);
      return el
    };

    AnnotationComponent.prototype.getClassNames = function getClassNames () {
      return 'sc-'+this.props.node.type
    };

    AnnotationComponent.prototype.onHighlightedChanged = function onHighlightedChanged () {
      if (this.props.node.highlighted) {
        this.el.addClass('sm-highlighted');
      } else {
        this.el.removeClass('sm-highlighted');
      }
    };

    AnnotationComponent.prototype.getTagName = function getTagName () {
      return 'span'
    };

    return AnnotationComponent;
  }(Component));

  var AbstractIsolatedNodeComponent = (function (Component$$1) {
    function AbstractIsolatedNodeComponent() {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      Component$$1.apply(this, args);

      this.name = this.props.node.id;
      this._state = { selectionFragment: null };

      this.handleAction('escape', this.escape);
      this.ContentClass = this._getContentClass(this.props.node);

      
      var useBlocker = platform.isFF || !this.ContentClass.noBlocker;
      this.blockingMode = useBlocker ? 'closed' : 'open';
    }

    if ( Component$$1 ) AbstractIsolatedNodeComponent.__proto__ = Component$$1;
    AbstractIsolatedNodeComponent.prototype = Object.create( Component$$1 && Component$$1.prototype );
    AbstractIsolatedNodeComponent.prototype.constructor = AbstractIsolatedNodeComponent;

    var prototypeAccessors = { id: { configurable: true } };

    AbstractIsolatedNodeComponent.prototype.getInitialState = function getInitialState () {
      var selState = this.context.editorSession.getSelectionState();
      return this._deriveStateFromSelectionState(selState)
    };

    AbstractIsolatedNodeComponent.prototype.getChildContext = function getChildContext () {
      return {
        parentSurfaceId: this.getId(),
        isolatedNodeComponent: this,
        
        
        surface: undefined
      }
    };


    AbstractIsolatedNodeComponent.prototype.didMount = function didMount () {
      Component$$1.prototype.didMount.call(this);

      var editorSession = this.context.editorSession;
      editorSession.onRender('selection', this._onSelectionChanged, this);
    };

    AbstractIsolatedNodeComponent.prototype.dispose = function dispose () {
      Component$$1.prototype.dispose.call(this);

      var editorSession = this.context.editorSession;
      editorSession.off(this);
    };

    AbstractIsolatedNodeComponent.prototype.renderContent = function renderContent ($$, node, options) {
      if ( options === void 0 ) options = {};

      var ComponentClass = this.ContentClass;
      if (!ComponentClass) {
        console.error('Could not resolve a component for type: ' + node.type);
        return $$(this.__elementTag)
      } else {
        var props = Object.assign({
          disabled: this.props.disabled,
          node: node,
          isolatedNodeState: this.state.mode,
          focused: (this.state.mode === 'focused')
        }, options);
        return $$(ComponentClass, props)
      }
    };

    AbstractIsolatedNodeComponent.prototype.getId = function getId () {
      
      
      if (!this._id) {
        this._id = this.context.parentSurfaceId +'/'+ this.name;
      }
      return this._id
    };

    prototypeAccessors.id.get = function () { return this.getId() };

    AbstractIsolatedNodeComponent.prototype.getMode = function getMode () {
      return this.state.mode
    };

    AbstractIsolatedNodeComponent.prototype.isOpen = function isOpen () {
      return this.blockingMode === 'open'
    };

    AbstractIsolatedNodeComponent.prototype.isClosed = function isClosed () {
      return this.blockingMode === 'closed'
    };

    AbstractIsolatedNodeComponent.prototype.isNotSelected = function isNotSelected () {
      return !this.state.mode
    };

    AbstractIsolatedNodeComponent.prototype.isSelected = function isSelected () {
      return this.state.mode === 'selected'
    };

    AbstractIsolatedNodeComponent.prototype.isCoSelected = function isCoSelected () {
      return this.state.mode === 'co-selected'
    };

    AbstractIsolatedNodeComponent.prototype.isFocused = function isFocused () {
      return this.state.mode === 'focused'
    };

    AbstractIsolatedNodeComponent.prototype.isCoFocused = function isCoFocused () {
      return this.state.mode === 'co-focused'
    };

    AbstractIsolatedNodeComponent.prototype.getParentSurface = function getParentSurface () {
      return this.context.surface
    };

    AbstractIsolatedNodeComponent.prototype.escape = function escape () {
      
      this.selectNode();
    };

    AbstractIsolatedNodeComponent.prototype._onSelectionChanged = function _onSelectionChanged () {
      var editorSession = this.context.editorSession;
      var newState = this._deriveStateFromSelectionState(editorSession.getSelectionState());
      if (!newState && this.state.mode) {
        this.extendState({ mode: null });
      } else if (newState && newState.mode !== this.state.mode) {
        this.extendState(newState);
      }
    };

    AbstractIsolatedNodeComponent.prototype.onKeydown = function onKeydown (event) {
      
      
      
      
      if (event.keyCode === keys$1.ESCAPE && this.state.mode === 'focused') {
        event.stopPropagation();
        event.preventDefault();
        this.escape();
      }
    };

    AbstractIsolatedNodeComponent.prototype._getContentClass = function _getContentClass (node) {
      var ComponentClass;
      
      ComponentClass = this.getComponent(node.type, true);
      
      if (!ComponentClass) {
        ComponentClass = Component$$1;
      }
      return ComponentClass
    };

    AbstractIsolatedNodeComponent.prototype._getSurface = function _getSurface (selState) {
      var surface = selState.get('surface');
      if (surface === undefined) {
        var sel = selState.getSelection();
        if (sel && sel.surfaceId) {
          var surfaceManager = this.context.surfaceManager;
          surface = surfaceManager.getSurface(sel.surfaceId);
        } else {
          surface = null;
        }
        selState.set('surface', surface);
      }
      return surface
    };

    
    
    AbstractIsolatedNodeComponent.prototype._getIsolatedNodes = function _getIsolatedNodes (selState) {
      var isolatedNodes = selState.get('isolatedNodes');
      if (!isolatedNodes) {
        var sel = selState.getSelection();
        isolatedNodes = [];
        if (sel && sel.surfaceId) {
          var surfaceManager = this.context.surfaceManager;
          var surface = surfaceManager.getSurface(sel.surfaceId);
          if (surface) {
            isolatedNodes = surface.getComponentPath().filter(function (comp) { return comp._isAbstractIsolatedNodeComponent; });
          }
        }
        selState.set('isolatedNodes', isolatedNodes);
      }
      return isolatedNodes
    };

    AbstractIsolatedNodeComponent.prototype._shouldConsumeEvent = function _shouldConsumeEvent (event) {
      var comp = Component$$1.unwrap(event.target);
      var isolatedNodeComponent = this._getIsolatedNode(comp);
      return (isolatedNodeComponent === this)
    };

    AbstractIsolatedNodeComponent.prototype._getIsolatedNode = function _getIsolatedNode (comp) {
      if (comp._isAbstractIsolatedNodeComponent) {
        return this
      } else if (comp.context.isolatedNodeComponent) {
        return comp.context.isolatedNodeComponent
      } else if (comp.context.surface) {
        return comp.context.surface.context.isolatedNodeComponent
      }
    };

    Object.defineProperties( AbstractIsolatedNodeComponent.prototype, prototypeAccessors );

    return AbstractIsolatedNodeComponent;
  }(Component));

  AbstractIsolatedNodeComponent.prototype._isAbstractIsolatedNodeComponent = true;

  var InlineNodeComponent = (function (AbstractIsolatedNodeComponent$$1) {
    function InlineNodeComponent () {
      AbstractIsolatedNodeComponent$$1.apply(this, arguments);
    }

    if ( AbstractIsolatedNodeComponent$$1 ) InlineNodeComponent.__proto__ = AbstractIsolatedNodeComponent$$1;
    InlineNodeComponent.prototype = Object.create( AbstractIsolatedNodeComponent$$1 && AbstractIsolatedNodeComponent$$1.prototype );
    InlineNodeComponent.prototype.constructor = InlineNodeComponent;

    InlineNodeComponent.prototype.render = function render ($$) {
      var node = this.props.node;
      var ContentClass = this.ContentClass;
      var state = this.state;

      var el = $$('span');
      el.addClass(this.getClassNames())
        .addClass('sc-inline-node')
        .addClass('sm-'+this.props.node.type)
        .attr("data-id", node.id)
        .attr('data-inline', '1');

      var disabled = this.isDisabled();

      if (state.mode) {
        el.addClass('sm-'+state.mode);
      } else {
        el.addClass('sm-not-selected');
      }

      if (!ContentClass.noStyle) {
        el.addClass('sm-default-style');
      }

      
      
      
      el.on('keydown', this.onKeydown);

      el.append(
        this.renderContent($$, node)
          .ref('content')
          .addClass('se-content')
      );

      if (disabled) {
        el.addClass('sm-disabled')
           .attr('contenteditable', false)
           .on('click', this.onClick);
      }

      
      
      
      
      
      
      
      el.attr('draggable', true);

      return el

    };

    InlineNodeComponent.prototype.isDisabled = function isDisabled () {
      return !this.state.mode || ['co-selected', 'cursor'].indexOf(this.state.mode) > -1;
    };

    InlineNodeComponent.prototype.getClassNames = function getClassNames () {
      return ''
    };

    InlineNodeComponent.prototype.onClick = function onClick (event) {
      if (!this._shouldConsumeEvent(event)) {
        return
      }
      this.selectNode();
    };

    InlineNodeComponent.prototype.selectNode = function selectNode () {
      
      var editorSession = this.context.editorSession;
      var surface = this.context.surface;
      var node = this.props.node;
      editorSession.setSelection({
        type: 'property',
        path: node.start.path,
        startOffset: node.start.offset,
        endOffset: node.end.offset,
        containerId: surface.getContainerId(),
        surfaceId: surface.id
      });
    };

    InlineNodeComponent.prototype._getContentClass = function _getContentClass (node) {
      var ComponentClass;
      
      ComponentClass = this.getComponent(node.type, true);
      
      
      if (!ComponentClass) {
        ComponentClass = this.getComponent('unsupported-inline-node', true);
      }
      if (!ComponentClass) {
        console.error(("No component registered for inline node '" + (node.type) + "'."));
        ComponentClass = StubInlineNodeComponent;
      }
      return ComponentClass
    };

    
    
    InlineNodeComponent.prototype._deriveStateFromSelectionState = function _deriveStateFromSelectionState (selState) {
      var surface = this._getSurface(selState);
      if (!surface) { return null }
      
      if (surface === this.context.surface) {
        var sel = selState.getSelection();
        var node = this.props.node;
        if (sel.isPropertySelection() && !sel.isCollapsed() && isEqual(sel.start.path, node.start.path)) {
          var nodeSel = node.getSelection();
          if(nodeSel.equals(sel)) {
            return { mode: 'selected' }
          }
          if (sel.contains(nodeSel)) {
            return { mode: 'co-selected' }
          }
        }
      }
      var isolatedNodeComponent = surface.context.isolatedNodeComponent;
      if (!isolatedNodeComponent) { return null }
      if (isolatedNodeComponent === this) {
        return { mode: 'focused' }
      }
      var isolatedNodes = this._getIsolatedNodes(selState);
      if (isolatedNodes.indexOf(this) > -1) {
        return { mode: 'co-focused' }
      }
      return null
    };

    return InlineNodeComponent;
  }(AbstractIsolatedNodeComponent));

  InlineNodeComponent.prototype._isInlineNodeComponent = true;

  var StubInlineNodeComponent = (function (Component$$1) {
    function StubInlineNodeComponent () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) StubInlineNodeComponent.__proto__ = Component$$1;
    StubInlineNodeComponent.prototype = Object.create( Component$$1 && Component$$1.prototype );
    StubInlineNodeComponent.prototype.constructor = StubInlineNodeComponent;

    StubInlineNodeComponent.prototype.render = function render ($$) {
      var node = this.props.node;
      return $$('span').text('???').attr('data-id', node.id).attr('data-type', node.type)
    };

    return StubInlineNodeComponent;
  }(Component));

  var AnnotatedTextComponent = (function (Component$$1) {
    function AnnotatedTextComponent () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) AnnotatedTextComponent.__proto__ = Component$$1;
    AnnotatedTextComponent.prototype = Object.create( Component$$1 && Component$$1.prototype );
    AnnotatedTextComponent.prototype.constructor = AnnotatedTextComponent;

    AnnotatedTextComponent.prototype.render = function render ($$) {
      var el = this._renderContent($$)
        .addClass('sc-annotated-text')
        .css({ whiteSpace: "pre-wrap" });
      return el
    };

    AnnotatedTextComponent.prototype.getPath = function getPath () {
      return this.props.path
    };

    AnnotatedTextComponent.prototype.getText = function getText () {
      return this.getDocument().get(this.props.path) || ''
    };

    AnnotatedTextComponent.prototype.isEmpty = function isEmpty () {
      return !(this.getText())
    };

    AnnotatedTextComponent.prototype.getAnnotations = function getAnnotations () {
      return this.getDocument().getIndex('annotations').get(this.props.path)
    };

    AnnotatedTextComponent.prototype.getDocument = function getDocument () {
      return this.props.doc || this.context.doc
    };

    AnnotatedTextComponent.prototype._getTagName = function _getTagName () {
      return this.props.tagName
    };

    AnnotatedTextComponent.prototype._onDocumentChange = function _onDocumentChange (update) {
      if (update.change && update.change.updated[this.getPath()]) {
        this.rerender();
      }
    };

    AnnotatedTextComponent.prototype._renderContent = function _renderContent ($$) {
      var text = this.getText();
      var annotations = this.getAnnotations();
      var el = $$(this._getTagName() || 'span');
      if (annotations && annotations.length > 0) {
        var fragmenter = new Fragmenter({
          onText: this._renderTextNode.bind(this),
          onEnter: this._renderFragment.bind(this, $$),
          onExit: this._finishFragment.bind(this)
        });
        fragmenter.start(el, text, annotations);
      } else {
        el.append(text);
      }
      return el
    };

    AnnotatedTextComponent.prototype._renderTextNode = function _renderTextNode (context, text) {
      if (text && text.length > 0) {
        context.append(text);
      }
    };

    AnnotatedTextComponent.prototype._renderFragment = function _renderFragment ($$, fragment) {
      var doc = this.getDocument();
      var componentRegistry = this.getComponentRegistry();
      var node = fragment.node;
      
      if (node.type === "container-annotation-fragment") ; else if (node.type === "container-annotation-anchor") ; else {
        var ComponentClass = componentRegistry.get(node.type) || AnnotationComponent;
        if (node.constructor.isInline &&
            
            !ComponentClass.prototype._isInlineNodeComponent &&
            
            !ComponentClass.isCustom) {
          ComponentClass = InlineNodeComponent;
        }
        var el = $$(ComponentClass, { doc: doc, node: node });
        return el
      }
    };

    AnnotatedTextComponent.prototype._finishFragment = function _finishFragment (fragment, context, parentContext) {
      parentContext.append(context);
    };

    return AnnotatedTextComponent;
  }(Component));

  var CursorComponent = (function (Component$$1) {
    function CursorComponent () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) CursorComponent.__proto__ = Component$$1;
    CursorComponent.prototype = Object.create( Component$$1 && Component$$1.prototype );
    CursorComponent.prototype.constructor = CursorComponent;

    CursorComponent.prototype.render = function render ($$) {
      
      var el = $$('span').addClass('se-cursor');
      
      
      
      el.append("\uFEFF");
      el.append($$('div').addClass('se-cursor-inner'));

      if (this.props.collaborator) {
        var collaboratorIndex = this.props.collaborator.colorIndex;
        el.addClass('sm-collaborator-'+collaboratorIndex);
      } else {
        el.addClass('sm-local-user');
      }

      return el
    };

    return CursorComponent;
  }(Component));

  var SelectionFragmentComponent = (function (Component$$1) {
    function SelectionFragmentComponent () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) SelectionFragmentComponent.__proto__ = Component$$1;
    SelectionFragmentComponent.prototype = Object.create( Component$$1 && Component$$1.prototype );
    SelectionFragmentComponent.prototype.constructor = SelectionFragmentComponent;

    SelectionFragmentComponent.prototype.render = function render ($$) {
      
      var el = $$('span').addClass('se-selection-fragment');
      if (this.props.collaborator) {
        var collaboratorIndex = this.props.collaborator.colorIndex;
        el.addClass('sm-collaborator-'+collaboratorIndex);
      } else {
        el.addClass('sm-local-user');
      }
      el.append(this.props.children);
      return el
    };

    return SelectionFragmentComponent;
  }(Component));

  var TextPropertyComponent = (function (AnnotatedTextComponent$$1) {
    function TextPropertyComponent () {
      AnnotatedTextComponent$$1.apply(this, arguments);
    }

    if ( AnnotatedTextComponent$$1 ) TextPropertyComponent.__proto__ = AnnotatedTextComponent$$1;
    TextPropertyComponent.prototype = Object.create( AnnotatedTextComponent$$1 && AnnotatedTextComponent$$1.prototype );
    TextPropertyComponent.prototype.constructor = TextPropertyComponent;

    TextPropertyComponent.prototype.getInitialState = function getInitialState () {
      var markersManager = this.context.markersManager;
      var path = this.getPath();
      var markers;
      if (markersManager) {
        
        markersManager.register(this);
        markers = markersManager.getMarkers(path, {
          surfaceId: this.getSurfaceId(),
          containerId: this.getContainerId()
        });
      } else {
        var doc = this.getDocument();
        markers = doc.getAnnotations(path);
      }
      return {
        markers: markers
      }
    };

    TextPropertyComponent.prototype.didMount = function didMount () {
      
      
      
      
      
    };

    TextPropertyComponent.prototype.dispose = function dispose () {
      if (this.context.markersManager) {
        this.context.markersManager.deregister(this);
      }
      if (this.domObserver) {
        this.domObserver.disconnect();
      }
    };

    TextPropertyComponent.prototype.render = function render ($$) {
      var path = this.getPath();



      var el = this._renderContent($$)
        .addClass('sc-text-property')
        .attr({
          'data-path': path.join('.')
        })
        .css({
          'white-space': 'pre-wrap'
        });

      if (this.isEmpty()) {
        el.addClass('sm-empty');
        if (this.props.placeholder) {
          el.append(
            $$('span').addClass('se-placeholder').append(
              this.props.placeholder
            )
          );
        }
      }

      if (!this.props.withoutBreak) {
        el.append($$('br'));
      }

      return el
    };

    TextPropertyComponent.prototype.getAnnotations = function getAnnotations () {
      if (this.props.markers) {
        return this.state.markers.concat(this.props.markers)
      } else {
        return this.state.markers
      }
    };

    TextPropertyComponent.prototype._renderFragment = function _renderFragment ($$, fragment) {
      var node = fragment.node;
      var id = node.id;
      var el;
      if (node.type === 'cursor') {
        el = $$(CursorComponent, { collaborator: node.collaborator });
      } else if (node.type === 'selection-fragment') {
        el = $$(SelectionFragmentComponent, { collaborator: node.collaborator });
      } else {
        el = AnnotatedTextComponent$$1.prototype._renderFragment.apply(this, arguments);
        if (id) {
          el.ref(id + '@' + fragment.counter);
        }
      }
      el.attr('data-offset', fragment.pos);
      return el
    };

    TextPropertyComponent.prototype._onDomMutations = function _onDomMutations (mutations) {
      
      
      if (mutations.length === 2 && mutations[0].target === mutations[1].target) {
        var textEl = DefaultDOMElement.unwrap(mutations[0].target);
        if (textEl) {
          this._applyTextMutation(textEl, mutations[0].oldValue);
          return
        }
      }
      
      this.rerender();
    };

    TextPropertyComponent.prototype._applyTextMutation = function _applyTextMutation (textEl, oldText) {
      
      var offset = _getCharPos(textEl, 0);
      var newText = textEl.textContent;
      var changes = diff(oldText, newText, offset);

      var editorSession = this.context.editorSession;
      var path = this.getPath();
      editorSession.transaction(function(tx) {
        changes.forEach(function(change) {
          
          if (change.type === 'replace') {
            tx.update(path, { type: 'delete', start: change.start, end: change.end });
            tx.update(path, { type: 'insert', start: change.start, text: change.text });
          } else {
            tx.update(path, change);
          }
        });
      });
    };

    TextPropertyComponent.prototype.getSurface = function getSurface () {
      return this.props.surface || this.context.surface
    };

    TextPropertyComponent.prototype.getSurfaceId = function getSurfaceId () {
      var surface = this.getSurface();
      return surface ? surface.id : null
    };

    TextPropertyComponent.prototype.getContainerId = function getContainerId () {
      var surface = this.getSurface();
      return surface ? surface.getContainerId() : null
    };

    TextPropertyComponent.prototype.isEditable = function isEditable () {
      var surface = this.getSurface();
      return surface ? surface.isEditable() : false
    };

    TextPropertyComponent.prototype.isReadonly = function isReadonly () {
      var surface = this.getSurface();
      return surface ? surface.isReadonly() : true
    };

    TextPropertyComponent.prototype.getDOMCoordinate = function getDOMCoordinate (charPos) {
      return this._getDOMCoordinate(this.el, charPos)
    };


    TextPropertyComponent.prototype._finishFragment = function _finishFragment (fragment, context, parentContext) {
      context.attr('data-length', fragment.length);
      parentContext.append(context);
    };

    TextPropertyComponent.prototype._getDOMCoordinate = function _getDOMCoordinate (el, charPos) {
      var this$1 = this;

      var l;
      var idx = 0;
      if (charPos === 0) {
        return {
          container: el.getNativeElement(),
          offset: 0
        }
      }
      for (var child = el.getFirstChild(); child; child = child.getNextSibling(), idx++) {
        if (child.isTextNode()) {
          l = child.textContent.length;
          if (l >= charPos) {
            return {
              container: child.getNativeElement(),
              offset: charPos
            }
          } else {
            charPos -= l;
          }
        } else if (child.isElementNode()) {
          var length = child.getAttribute('data-length');
          if (length) {
            l = parseInt(length, 10);
            if (l >= charPos) {
              
              if (child.attr('data-inline')) {
                var nextSibling = child.getNextSibling();
                if (nextSibling && nextSibling.isTextNode()) {
                  return {
                    container: nextSibling.getNativeElement(),
                    offset: 0
                  }
                } else {
                  return {
                    container: el.getNativeElement(),
                    offset: el.getChildIndex(child) + 1
                  }
                }
              }
              return this$1._getDOMCoordinate(child, charPos, idx)
            } else {
              charPos -= l;
            }
          } else {
            console.error('FIXME: Can not map to DOM coordinates.');
            return null
          }
        }
      }
    };

    return TextPropertyComponent;
  }(AnnotatedTextComponent));

  TextPropertyComponent.prototype._isTextPropertyComponent = true;





  TextPropertyComponent.getCoordinate = function(root, el, offset) {
    var context = _getPropertyContext(root, el, offset);
    if (!context) {
      return null
    }
    
    
    var charPos;
    if (el.parentNode && el.parentNode.is('.se-placeholder')) {
      charPos = 0;
    } else {
      
      
      
      charPos = _getCharPos(context.node, context.offset);
    }
    if (isNumber(charPos)) {
      var coor = new Coordinate(context.path, charPos);
      coor._comp = context.comp;
      return coor
    } else {
      return null
    }
  };

  function _getPropertyContext(root, node, offset) {
    var result = {
      comp: null,
      el: null,
      path: null,
      node: node,
      offset: offset
    };
    while (node && node !== root) {
      if (node.isElementNode()) {
        var comp = Component.unwrap(node);
        if (comp && comp._isTextPropertyComponent) {
          result.comp = comp;
          result.el = node;
          result.path = comp.getPath();
          return result;
        }
        
        
        
        if (node.getAttribute('data-inline')) {
          result.node = node;
          if (offset > 0) {
            result.offset = 1;
          }
        }
      }
      node = node.getParent();
    }
    return null
  }

  function _getCharPos(node, offset) {
    var charPos = offset;
    var parent, childIdx;

    

    parent = node.getParent();
    if (node.isTextNode()) {
      
      if (node === parent.firstChild) {
        
        var parentPath = parent.getAttribute('data-path');
        var parentOffset = parent.getAttribute('data-offset');
        if (parentPath) {
          charPos = offset;
        }
        
        else if (parentOffset) {
          charPos = parseInt(parentOffset, 10) + offset;
        }
        
        else {
          charPos = _getCharPos(parent, 0) + offset;
        }
      } else {
        
        childIdx = parent.getChildIndex(node);
        charPos = _getCharPos(parent, childIdx) + offset;
      }
    } else if (node.isElementNode()) {
      var pathStr = node.getAttribute('data-path');
      var offsetStr = node.getAttribute('data-offset');
      
      
      if (pathStr) {
        charPos = _countCharacters(node, offset);
      }
      
      
      else if (offsetStr) {
        childIdx = parent.getChildIndex(node);
        charPos = parseInt(offsetStr, 10) + _countCharacters(node, offset);
      }
      
      
      else {
        childIdx = parent.getChildIndex(node);
        charPos = _getCharPos(parent, childIdx) + _countCharacters(node, offset);
      }
    } else {
      
      return null
    }
    return charPos;
  }

  function _countCharacters(el, maxIdx) {
    var charPos = 0;
    
    if (el.getAttribute('data-inline')) {
      return maxIdx === 0 ? 0 : 1;
    }
    var l = el.getChildCount();
    if (arguments.length === 1) {
      maxIdx = l;
    }
    maxIdx = Math.min(l, maxIdx);
    for (var i=0, child = el.getFirstChild(); i < maxIdx; child = child.getNextSibling(), i++) {
      if (child.isTextNode()) {
        charPos += child.getTextContent().length;
      } else if (child.isElementNode()) {
        var length = child.getAttribute('data-length');
        if (child.getAttribute('data-inline')) {
          charPos += 1;
        } else if (length) {
          charPos += parseInt(length, 10);
        } else {
          charPos += _countCharacters(child);
        }
      }
    }
    return charPos
  }

  var BRACKET = 'X';
  var SLASH = '/'.charCodeAt(0);


  var IsolatedNodeComponent = (function (AbstractIsolatedNodeComponent$$1) {
    function IsolatedNodeComponent() {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      AbstractIsolatedNodeComponent$$1.apply(this, args);
    }

    if ( AbstractIsolatedNodeComponent$$1 ) IsolatedNodeComponent.__proto__ = AbstractIsolatedNodeComponent$$1;
    IsolatedNodeComponent.prototype = Object.create( AbstractIsolatedNodeComponent$$1 && AbstractIsolatedNodeComponent$$1.prototype );
    IsolatedNodeComponent.prototype.constructor = IsolatedNodeComponent;

    IsolatedNodeComponent.prototype.render = function render ($$) {
      var node = this.props.node;
      var ContentClass = this.ContentClass;
      var disabled = this.props.disabled;

      
      var el = $$('div');
      el.addClass(this.getClassNames())
        .addClass('sc-isolated-node')
        .addClass('sm-'+this.props.node.type)
        .attr("data-id", node.id);
      if (disabled) {
        el.addClass('sm-disabled');
      }
      if (this.state.mode) {
        el.addClass('sm-'+this.state.mode);
      }
      if (!ContentClass.noStyle) {
        el.addClass('sm-default-style');
      }
      
      el.on('keydown', this.onKeydown);

      
      var shouldRenderBlocker = (
        this.blockingMode === 'closed' &&
        !this.state.unblocked
      );

      
      
      el.append(
        $$('div').addClass('se-bracket sm-left').ref('left')
          .append(BRACKET)
      );

      var content = this.renderContent($$, node, {
        disabled: this.props.disabled || shouldRenderBlocker
      }).ref('content');
      content.attr('contenteditable', false);

      el.append(content);
      el.append($$(Blocker).ref('blocker'));
      el.append(
        $$('div').addClass('se-bracket sm-right').ref('right')
          .append(BRACKET)
      );

      if (!shouldRenderBlocker) {
        el.addClass('sm-no-blocker');
        el.on('click', this.onClick)
          .on('dblclick', this.onDblClick);
      }
      el.on('mousedown', this._reserveMousedown, this);

      return el
    };

    IsolatedNodeComponent.prototype.getClassNames = function getClassNames () {
      return ''
    };

    IsolatedNodeComponent.prototype.getContent = function getContent () {
      return this.refs.content
    };

    IsolatedNodeComponent.prototype.selectNode = function selectNode () {
      
      var editorSession = this.context.editorSession;
      var surface = this.context.surface;
      var nodeId = this.props.node.id;
      editorSession.setSelection({
        type: 'node',
        nodeId: nodeId,
        containerId: surface.getContainerId(),
        surfaceId: surface.id
      });
    };

    
    
    IsolatedNodeComponent.prototype.onClick = function onClick (event) {
      
      event.stopPropagation();
    };

    IsolatedNodeComponent.prototype.onDblClick = function onDblClick (event) {
      
      event.stopPropagation();
    };

    IsolatedNodeComponent.prototype.grabFocus = function grabFocus (event) {
      var content = this.refs.content;
      if (content.grabFocus) {
        content.grabFocus(event);
        return true
      }
    };

    
    
    IsolatedNodeComponent.prototype._reserveMousedown = function _reserveMousedown (event) {
      if (event.__reserved__) {
        
        return
      } else {
        
        event.__reserved__ = this;
      }
    };

    IsolatedNodeComponent.prototype._deriveStateFromSelectionState = function _deriveStateFromSelectionState (selState) {
      var surface = this._getSurface(selState);
      var newState = { mode: null, unblocked: null};
      if (!surface) { return newState }
      
      if (surface === this.context.surface) {
        var sel = selState.getSelection();
        var nodeId = this.props.node.id;
        if (sel.isNodeSelection() && sel.getNodeId() === nodeId) {
          if (sel.isFull()) {
            newState.mode = 'selected';
            newState.unblocked = true;
          } else if (sel.isBefore()) {
            newState.mode = 'cursor';
            newState.position = 'before';
          } else if (sel.isAfter()) {
            newState.mode = 'cursor';
            newState.position = 'after';
          }
        }
        if (sel.isContainerSelection() && sel.containsNode(nodeId)) {
          newState.mode = 'co-selected';
        }
      } else {
        var sel$1 = selState.getSelection();
        var surfaceId = sel$1.surfaceId;
        var id = this.getId();
        if (id.length < surfaceId.length && surfaceId.startsWith(id) && surfaceId.charCodeAt(id.length) === SLASH) {
          var tail = surfaceId.slice(id.length+1);
          if (tail.indexOf('/') > 0) {
            newState.mode = 'co-focused';
            newState.unblocked = true;
          } else {
            newState.mode = 'focused';
            newState.unblocked = true;
          }
        }
      }
      return newState
    };

    return IsolatedNodeComponent;
  }(AbstractIsolatedNodeComponent));

  IsolatedNodeComponent.prototype._isIsolatedNodeComponent = true;

  IsolatedNodeComponent.prototype._isDisabled = IsolatedNodeComponent.prototype.isDisabled;

  IsolatedNodeComponent.getDOMCoordinate = function(comp, coor) {
    var ref = IsolatedNodeComponent.getDOMCoordinates(comp);
    var start = ref.start;
    var end = ref.end;
    if (coor.offset === 0) { return start }
    else { return end }
  };

  IsolatedNodeComponent.getDOMCoordinates = function(comp) {
    var left = comp.refs.left;
    var right = comp.refs.right;
    return {
      start: {
        container: left.getNativeElement(),
        offset: 0
      },
      end: {
        container: right.getNativeElement(),
        offset: right.getChildCount()
      }
    }
  };

  IsolatedNodeComponent.getCoordinate = function(nodeEl, options) {
    var comp = Component.unwrap(nodeEl, 'strict').context.isolatedNodeComponent;
    var offset = null;
    if (options.direction === 'left' || nodeEl === comp.refs.left.el) {
      offset = 0;
    } else if (options.direction === 'right' || nodeEl === comp.refs.right.el) {
      offset = 1;
    }
    var coor;
    if (offset !== null) {
      coor = new Coordinate([comp.props.node.id], offset);
      coor._comp = comp;
    }
    return coor
  };

  var Blocker = (function (Component$$1) {
    function Blocker () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) Blocker.__proto__ = Component$$1;
    Blocker.prototype = Object.create( Component$$1 && Component$$1.prototype );
    Blocker.prototype.constructor = Blocker;

    Blocker.prototype.render = function render ($$) {
      return $$('div').addClass('sc-isolated-node-blocker')
        .attr('draggable', true)
        .attr('contenteditable', false)
        .on('click', this.onClick)
        .on('dblclick', this.onDblClick)
    };

    Blocker.prototype.onClick = function onClick (event) {
      if (event.target !== this.getNativeElement()) { return }
      
      event.stopPropagation();
      var comp = this._getIsolatedNodeComponent();
      comp.extendState({ mode: 'selected', unblocked: true });
      comp.selectNode();
    };

    Blocker.prototype.onDblClick = function onDblClick (event) {
      
      event.stopPropagation();
    };

    Blocker.prototype._getIsolatedNodeComponent = function _getIsolatedNodeComponent () {
      return this.context.isolatedNodeComponent
    };

    return Blocker;
  }(Component));

  var DEBUG$1 = false;


  var DOMSelection = function DOMSelection(editor) {
    this.editor = editor;
    if (platform.inBrowser) {
      this.wRange = window.document.createRange();
    }
    
    
    this.state = { dom: null, model: null };
  };


  DOMSelection.prototype.getSelection = function getSelection (options) {
    
    if (!platform.inBrowser) { return }
    var range = this.mapDOMSelection(options);
    var doc = this.editor.getDocument();
    
    return doc._createSelectionFromRange(range)
  };

  DOMSelection.prototype.getSelectionForDOMRange = function getSelectionForDOMRange (wrange) {
    var range = this.mapDOMRange(wrange);
    var doc = this.editor.getDocument();
    return doc._createSelectionFromRange(range)
  };


  DOMSelection.prototype.mapDOMSelection = function mapDOMSelection (options) {
    var wSel = window.getSelection();
    var state = this.state;
    var range;
    
    
    if (DEBUG$1) { console.info('DOM->Model: ', wSel.anchorNode, wSel.anchorOffset, wSel.focusNode, wSel.focusOffset); }
    if (wSel.rangeCount === 0) { return _null() }
    var anchorNode = DefaultDOMElement.wrapNativeElement(wSel.anchorNode);
    if (wSel.isCollapsed) {
      var coor = this._getCoordinate(anchorNode, wSel.anchorOffset, options);
      if (!coor) { return _null() }
      range = _createRange({
        start: coor,
        end: coor
      });
    }
    else {
      var focusNode = DefaultDOMElement.wrapNativeElement(wSel.focusNode);
      range = this._getRange(anchorNode, wSel.anchorOffset, focusNode, wSel.focusOffset, options);
    }
    if (DEBUG$1) { console.info('DOM->Model: range ', range ? range.toString() : null); }
    state.model = range;
    return range

    function _null() {
      state.dom = null;
      state.model = null;
      return null
    }
  };


  DOMSelection.prototype.setSelection = function setSelection (sel) {
      var assign;

    
    if (!platform.inBrowser) { return }
    var state = this.state;
    var wSel = window.getSelection();
    var wRange = this.wRange;
    if (!sel || sel.isNull()) { return this.clear() }
    
    var ref = this.mapModelToDOMCoordinates(sel);
      var start = ref.start;
      var end = ref.end;
    if (!start) { return this.clear() }
    if (sel.isReverse()) {
      (assign = [end, start], start = assign[0], end = assign[1]);
    }
    state.dom = {
      anchorNode: start.container,
      anchorOffset: start.offset,
      focusNode: end.container,
      focusOffset: end.offset
    };
    _set(state.dom);

    function _set(ref) {
        var anchorNode = ref.anchorNode;
        var anchorOffset = ref.anchorOffset;
        var focusNode = ref.focusNode;
        var focusOffset = ref.focusOffset;

      wSel.removeAllRanges();
      wRange.setStart(anchorNode, anchorOffset);
      wRange.setEnd(anchorNode, anchorOffset);
      wSel.addRange(wRange);
      if (focusNode !== anchorOffset || focusOffset !== anchorOffset) {
        wSel.extend(focusNode, focusOffset);
      }
    }
  };

  DOMSelection.prototype.mapModelToDOMCoordinates = function mapModelToDOMCoordinates (sel) {
    if (DEBUG$1) { console.info('Model->DOM: sel =', sel.toString()); }
    var rootEl;
    var surface = this.editor.getSurfaceManager().getSurface(sel.surfaceId);
    if (!surface) {
      console.warn('No surface:', sel.surfaceId);
      rootEl = this.editor.getElement();
    } else {
      rootEl = surface.el;
    }
    if (sel.isNull() || sel.isCustomSelection()) {
      return {}
    }

    var start, end;
    if (sel.isPropertySelection() || sel.isContainerSelection()) {
      start = this._getDOMCoordinate(rootEl, sel.start);
      if (!start) {
        console.warn('FIXME: selection seems to be invalid.');
        return {}
      }
      if (sel.isCollapsed()) {
        end = start;
      } else {
        end = this._getDOMCoordinate(rootEl, sel.end);
        if (!end) {
          console.warn('FIXME: selection seems to be invalid.');
          return {}
        }
      }
    } else if (sel.isNodeSelection()) {
      var comp = Component.unwrap(rootEl.find('*[data-id="'+sel.getNodeId()+'"]'));
      if (!comp) {
        console.error('Could not find component with id', sel.getNodeId());
        return {}
      }
      if (comp._isIsolatedNodeComponent) {
        var coors = IsolatedNodeComponent.getDOMCoordinates(comp, sel);
        start = coors.start;
        end = coors.end;
        
        
        
      } else {
        var _nodeEl = comp.el;
        start = {
          container: _nodeEl.getNativeElement(),
          offset: 0
        };
        end = {
          container: _nodeEl.getNativeElement(),
          offset: _nodeEl.getChildCount()
        };
      }
    }
    if (DEBUG$1) { console.info('Model->DOM:', start.container, start.offset, end.container, end.offset, 'isReverse?', sel.isReverse()); }
    return {start: start,end: end}
  };

  DOMSelection.prototype._getDOMCoordinate = function _getDOMCoordinate (rootEl, coor) {
    var comp, domCoor = null;
    if (coor.isNodeCoordinate()) {
      comp = Component.unwrap(rootEl.find('*[data-id="'+coor.getNodeId()+'"]'));
      if (comp) {
        if (comp._isIsolatedNodeComponent) {
          domCoor = IsolatedNodeComponent.getDOMCoordinate(comp, coor);
        } else {
          var domOffset = 0;
          if (coor.offset > 0) {
            domOffset = comp.getChildCount();
          }
          domCoor = {
            container: comp.getNativeElement(),
            offset: domOffset
          };
        }
      }
    } else {
      comp = Component.unwrap(rootEl.find('.sc-text-property[data-path="'+coor.path.join('.')+'"]'));
      if (comp) {
        domCoor = comp.getDOMCoordinate(coor.offset);
      }
    }
    return domCoor
  };


  DOMSelection.prototype.mapDOMRange = function mapDOMRange (wRange, options) {
    return this._getRange(
      DefaultDOMElement.wrapNativeElement(wRange.startContainer),
      wRange.startOffset,
      DefaultDOMElement.wrapNativeElement(wRange.endContainer),
      wRange.endOffset, options)
  };


  DOMSelection.prototype.clear = function clear () {
    window.getSelection().removeAllRanges();
    this.state.dom = null;
    this.state.model = null;
  };

  DOMSelection.prototype.collapse = function collapse (dir) {
    var wSel = window.getSelection();
    var wRange;
    if (wSel.rangeCount > 0) {
      wRange = wSel.getRangeAt(0);
      wRange.collapse(dir === 'left');
      wSel.removeAllRanges();
      wSel.addRange(wRange);
    }
  };

  DOMSelection.prototype.select = function select (el) {
    var wSel = window.getSelection();
    var wRange = window.document.createRange();
    wRange.selectNode(el.getNativeElement());
    wSel.removeAllRanges();
    wSel.addRange(wRange);
  };

  DOMSelection.prototype.extend = function extend (el, offset) {
    var wSel = window.getSelection();
    wSel.extend(el.getNativeElement(), offset);
  };

  DOMSelection.prototype.setCursor = function setCursor (el, offset) {
    var wSel = window.getSelection();
    var wRange = window.document.createRange();
    wRange.setStart(el.getNativeElement(), offset);
    wSel.removeAllRanges();
    wSel.addRange(wRange);
  };


  DOMSelection.prototype._getRange = function _getRange (anchorNode, anchorOffset, focusNode, focusOffset, options) {
      if ( options === void 0 ) options = {};

    var isReverse = DefaultDOMElement.isReverse(anchorNode, anchorOffset, focusNode, focusOffset);
    var isCollapsed = (anchorNode === focusNode && anchorOffset === focusOffset);
    var start, end;
    if (isCollapsed) {
      start = end = this._getCoordinate(anchorNode, anchorOffset, options);
    } else {
      start = this._getCoordinate(anchorNode, anchorOffset, { direction: isReverse ? 'right' : 'left' });
      end = this._getCoordinate(focusNode, focusOffset, options);
    }
    if (start && end) {
      return _createRange({ start: start, end: end, isReverse: isReverse })
    } else {
      return null
    }
  };


  DOMSelection.prototype._getCoordinate = function _getCoordinate (nodeEl, offset, options) {
      if ( options === void 0 ) options={};

    var coor = null;
    
    if (!coor) {
      coor = TextPropertyComponent.getCoordinate(this.editor.getElement(), nodeEl, offset);
    }
    var comp = Component.unwrap(nodeEl);
    if (!coor && comp) {
      
      if (comp.context.isolatedNodeComponent) {
        coor = IsolatedNodeComponent.getCoordinate(nodeEl, options);
      }
    }
    
    
    if (!coor) {
      
      if (comp && comp._isContainerEditor) {
        var childIdx = (offset === 0) ? 0 : offset-1;
        var isBefore = (offset === 0);
        var container = comp.getContainer();
        var childNode = container.getNodeAt(childIdx);
        var childComp = comp.getChildAt(childIdx);
        coor = new Coordinate([childNode.id], isBefore?0:1 );
        coor._comp = childComp;
      }
      
      else if (nodeEl.isElementNode() && nodeEl.getChildCount() > 0) {
        var child = (offset > 0) ? nodeEl.getChildAt(offset-1) : nodeEl.firstChild;
        var prop;
        var childComp$1 = Component.unwrap(child);
        if (childComp$1 && childComp$1._isTextPropertyComponent) {
          prop = child;
        }
        
        if (prop) {
          coor = TextPropertyComponent.getCoordinate(nodeEl, prop, (offset > 0) ? prop.getChildCount() : 0);
        }
      }
    }
    return coor
  };


  function _createRange(ref) {
    var assign;

    var start = ref.start;
    var end = ref.end;
    var isReverse = ref.isReverse;
    if (isReverse) {
      (assign = [end, start], start = assign[0], end = assign[1]);
    }
    if (!start._comp || !end._comp) {
      console.error('FIXME: getCoordinate() should provide a component instance');
      return null
    }
    var surface = start._comp.context.surface;
    if (!surface) {
      console.error('FIXME: Editable components should have their surface in the context');
      return null
    }
    if (surface !== end._comp.context.surface) {
      console.error('Coordinates are within two different surfaces. Can not create a selection.');
      return null
    }
    return new Range(start, end, isReverse, surface.getContainerId(), surface.id)
  }

  var AbstractEditor = (function (Component$$1) {
    function AbstractEditor() {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      Component$$1.apply(this, args);

      this._initialize(this.props);
    }

    if ( Component$$1 ) AbstractEditor.__proto__ = Component$$1;
    AbstractEditor.prototype = Object.create( Component$$1 && Component$$1.prototype );
    AbstractEditor.prototype.constructor = AbstractEditor;

    AbstractEditor.prototype.didMount = function didMount () {
      
      this.getEditorSession().attachEditor(this);
    };

    AbstractEditor.prototype.dispose = function dispose () {
      this._dispose();
    };

    AbstractEditor.prototype._initialize = function _initialize (props) {
      var this$1 = this;

      var editorSession = props.editorSession;
      if (!editorSession) {
        throw new Error('EditorSession instance required');
      }
      this.editorSession = editorSession;
      this.doc = editorSession.getDocument();

      var configurator = editorSession.getConfigurator();
      this.componentRegistry = configurator.getComponentRegistry();
      this.commandGroups = configurator.getCommandGroups();
      this.keyboardShortcuts = configurator.getKeyboardShortcutsByCommand();
      this.tools = configurator.getTools();
      this.labelProvider = configurator.getLabelProvider();
      this.iconProvider = configurator.getIconProvider();

      
      this.surfaceManager = editorSession.surfaceManager;
      this.commandManager = editorSession.commandManager;
      this.dragManager = editorSession.dragManager;
      this.macroManager = editorSession.macroManager;
      this.converterRegistry = editorSession.converterRegistry;
      this.globalEventHandler = editorSession.globalEventHandler;
      this.editingBehavior = editorSession.editingBehavior;
      this.markersManager = editorSession.markersManager;

      this.resourceManager = new ResourceManager(editorSession, this.getChildContext());
      this.domSelection = new DOMSelection(this);

      
      this.labelProvider.setLanguage(editorSession.getLanguage());
      
      
      editorSession.onUpdate('lang', function (lang) {
        this$1.labelProvider.setLanguage(lang);
      }, this);
      editorSession.onRender('lang', this.rerender, this);
    };

    AbstractEditor.prototype.willReceiveProps = function willReceiveProps (nextProps) {
      var newSession = nextProps.editorSession;
      var shouldDispose = newSession && newSession !== this.editorSession;
      if (shouldDispose) {
        this._dispose();
        this._initialize(nextProps);
      }
    };

    AbstractEditor.prototype._dispose = function _dispose () {
      var editorSession = this.getEditorSession();
      editorSession.off(this);
      editorSession.detachEditor(this);
      
      
      this.empty();
      
      
      this.resourceManager.dispose();
    };

    AbstractEditor.prototype.getChildContext = function getChildContext () {
      return {
        editor: this,
        editorSession: this.editorSession,
        doc: this.doc, 
        document: this.doc,
        componentRegistry: this.componentRegistry,
        surfaceManager: this.surfaceManager,
        domSelection: this.domSelection,
        commandManager: this.commandManager,
        markersManager: this.markersManager,
        converterRegistry: this.converterRegistry,
        dragManager: this.dragManager,
        editingBehavior: this.editingBehavior,
        globalEventHandler: this.globalEventHandler,
        iconProvider: this.iconProvider,
        labelProvider: this.labelProvider,
        resourceManager: this.resourceManager,
        commandGroups: this.commandGroups,
        tools: this.tools,
        keyboardShortcuts: this.keyboardShortcuts
      }
    };

    
    AbstractEditor.prototype.onKeyDown = function onKeyDown (event) {
      
      if ( event.key === 'Dead' ) { return }
      
      var custom = this.editorSession.keyboardManager.onKeydown(event);
      return custom
    };

    AbstractEditor.prototype.getDocument = function getDocument () {
      return this.editorSession.getDocument()
    };

    AbstractEditor.prototype.getConfigurator = function getConfigurator () {
      return this.editorSession.getConfigurator()
    };

    AbstractEditor.prototype.getEditorSession = function getEditorSession () {
      return this.editorSession
    };

    AbstractEditor.prototype.getComponentRegistry = function getComponentRegistry () {
      return this.componentRegistry
    };

    AbstractEditor.prototype.getSurfaceManager = function getSurfaceManager () {
      return this.surfaceManager
    };

    AbstractEditor.prototype.getLabelProvider = function getLabelProvider () {
      return this.labelProvider
    };

    return AbstractEditor;
  }(Component));

  function getRelativeMouseBounds(mouseEvent, containerEl) {
    var containerElRect = containerEl.getBoundingClientRect();
    var left = mouseEvent.clientX - containerElRect.left;
    var top = mouseEvent.clientY - containerElRect.top;
    var res = {
      left: left,
      right: containerElRect.width - left,
      top: top,
      bottom: containerElRect.height - top
    };
    return res;
  }

  var AbstractScrollPane = (function (Component$$1) {
    function AbstractScrollPane () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) AbstractScrollPane.__proto__ = Component$$1;
    AbstractScrollPane.prototype = Object.create( Component$$1 && Component$$1.prototype );
    AbstractScrollPane.prototype.constructor = AbstractScrollPane;

    AbstractScrollPane.prototype.getChildContext = function getChildContext () {
      return {
        scrollPane: this
      }
    };

    AbstractScrollPane.prototype.didMount = function didMount () {
      if (platform.inBrowser) {
        this.windowEl = DefaultDOMElement.wrapNativeElement(window);
        this.windowEl.on('resize', this.onSelectionPositioned, this);
      }
    };

    AbstractScrollPane.prototype.dispose = function dispose () {
      if (this.windowEl) {
        this.windowEl.off(this);
      }
    };

    AbstractScrollPane.prototype.getName = function getName () {
      return this.props.name
    };

    
    AbstractScrollPane.prototype.onSelectionPositioned = function onSelectionPositioned () {
      var contentRect = this._getContentRect();
      var selectionRect = this._getSelectionRect();
      if (!selectionRect) { return }
      var hints = {
        contentRect: contentRect,
        selectionRect: selectionRect
      };
      this._emitSelectionPositioned(hints);
      this._scrollSelectionIntoView(selectionRect);
    };

    AbstractScrollPane.prototype._emitSelectionPositioned = function _emitSelectionPositioned (hints) {
      
      
      this.emit('selection:positioned', hints);
      
      this.emit('dom-selection:rendered', hints);
    };

    
    AbstractScrollPane.prototype._onContextMenu = function _onContextMenu (e) {
      e.preventDefault();
      var mouseBounds = this._getMouseBounds(e);
      this.emit('context-menu:opened', {
        mouseBounds: mouseBounds
      });
    };

    AbstractScrollPane.prototype._scrollSelectionIntoView = function _scrollSelectionIntoView (selectionRect) {
      var upperBound = this.getScrollPosition();
      var lowerBound = upperBound + this.getHeight();
      var selTop = selectionRect.top;
      var selBottom = selectionRect.top + selectionRect.height;
      if ((selTop < upperBound && selBottom < upperBound) ||
          (selTop > lowerBound && selBottom > lowerBound)) {
        this.setScrollPosition(selTop);
      }
    };

    
    AbstractScrollPane.prototype.getHeight = function getHeight () {
      throw new Error('Abstract method')
    };

    
    AbstractScrollPane.prototype.getContentHeight = function getContentHeight () {
      throw new Error('Abstract method')
    };

    AbstractScrollPane.prototype.getContentElement = function getContentElement () {
      
      throw new Error('Abstract method')
    };

    
    AbstractScrollPane.prototype.getScrollableElement = function getScrollableElement () {
      throw new Error('Abstract method')
    };

    
    AbstractScrollPane.prototype.getScrollPosition = function getScrollPosition () {
      throw new Error('Abstract method')
    };

    AbstractScrollPane.prototype.setScrollPosition = function setScrollPosition () {
      throw new Error('Abstract method')
    };

    
    AbstractScrollPane.prototype.getPanelOffsetForElement = function getPanelOffsetForElement (el) { 
      throw new Error('Abstract method')
    };

    
    AbstractScrollPane.prototype.scrollTo = function scrollTo (componentId, onlyIfNotVisible) { 
      throw new Error('Abstract method')
    };

    AbstractScrollPane.prototype._getContentRect = function _getContentRect () {
      return this.getContentElement().getNativeElement().getBoundingClientRect()
    };

    
    AbstractScrollPane.prototype._getSelectionRect = function _getSelectionRect () {
      return getSelectionRect(this._getContentRect())
    };

    AbstractScrollPane.prototype._getMouseBounds = function _getMouseBounds (e) {
      return getRelativeMouseBounds(e, this.getContentElement().getNativeElement())
    };

    return AbstractScrollPane;
  }(Component));

  var Command = function Command(config) {
    this.config = config || {};
    this.name = this.config.name;
    if (!this.name) {
      throw new Error("'name' is required");
    }
  };

  var prototypeAccessors$9 = { isAsync: { configurable: true } };

  prototypeAccessors$9.isAsync.get = function () {
    return false
  };


  Command.prototype.getName = function getName () {
    return this.name
  };


  Command.prototype.getCommandState = function getCommandState (params, context) { 
    throw new Error('Command.getCommandState() is abstract.')
  };


  Command.prototype.execute = function execute (params, context) { 
    throw new Error('Command.execute() is abstract.')
  };


  Command.prototype.isAnnotationCommand = function isAnnotationCommand () {
    return false
  };


  Command.prototype.isInsertCommand = function isInsertCommand () {
    return false
  };


  Command.prototype.isSwitchTypeCommand = function isSwitchTypeCommand () {
    return false
  };

  Command.prototype._getEditorSession = function _getEditorSession (params, context) {
    var editorSession = params.editorSession || context.editorSession;
    if (!editorSession) {
      throw new Error("'editorSession' is required.")
    }
    return editorSession
  };

  Command.prototype._getSelection = function _getSelection (params) {
    var sel = params.selection || params.selectionState.getSelection();
    if (!sel) {
      throw new Error("'selection' is required.")
    }
    return sel
  };

  Object.defineProperties( Command.prototype, prototypeAccessors$9 );

  Command.prototype._isCommand = true;

  var AnnotationCommand = (function (Command$$1) {
    function AnnotationCommand() {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      Command$$1.apply(this, args);

      if (!this.config.nodeType) {
        throw new Error("'nodeType' is required")
      }
    }

    if ( Command$$1 ) AnnotationCommand.__proto__ = Command$$1;
    AnnotationCommand.prototype = Object.create( Command$$1 && Command$$1.prototype );
    AnnotationCommand.prototype.constructor = AnnotationCommand;

    
    AnnotationCommand.prototype.getAnnotationType = function getAnnotationType () {
      return this.config.nodeType
    };

    AnnotationCommand.prototype.getType = function getType () {
      return this.getAnnotationType()
    };

    
    AnnotationCommand.prototype.getAnnotationData = function getAnnotationData () {
      return {}
    };

    
    AnnotationCommand.prototype.isDisabled = function isDisabled (sel, params) {
      var selectionState = params.selectionState;
      var isBlurred = params.editorSession.isBlurred();
      
      
      if (isBlurred || !sel || sel.isNull() || !sel.isAttached() || sel.isCustomSelection()||
          sel.isNodeSelection() || sel.isContainerSelection() || selectionState.isInlineNodeSelection()) {
        return true
      }
      return false
    };

    
    AnnotationCommand.prototype.showInContext = function showInContext (sel) {
      return !sel.isCollapsed()
    };

    
    
    AnnotationCommand.prototype.canCreate = function canCreate (annos, sel) {
      return (annos.length === 0 && !sel.isCollapsed())
    };

    
    AnnotationCommand.prototype.canFuse = function canFuse (annos, sel) {
      
      return (annos.length >= 2 && !sel.isCollapsed())
    };

    
    AnnotationCommand.prototype.canDelete = function canDelete (annos, sel) {
      
      if (annos.length !== 1) { return false }
      var annoSel = annos[0].getSelection();
      return sel.isInsideOf(annoSel)
    };

    
    AnnotationCommand.prototype.canExpand = function canExpand (annos, sel) {
      
      if (annos.length !== 1) { return false }
      var annoSel = annos[0].getSelection();
      return sel.overlaps(annoSel) && !sel.isInsideOf(annoSel)
    };

    
    AnnotationCommand.prototype.canTruncate = function canTruncate (annos, sel) {
      if (annos.length !== 1) { return false }
      var annoSel = annos[0].getSelection();

      return (sel.isLeftAlignedWith(annoSel) || sel.isRightAlignedWith(annoSel)) &&
             !sel.contains(annoSel) &&
             !sel.isCollapsed()
    };

    
    AnnotationCommand.prototype.getCommandState = function getCommandState (params) { 
      var sel = this._getSelection(params);
      
      
      
      if (this.isDisabled(sel, params)) {
        return {
          disabled: true
        }
      }
      var annos = this._getAnnotationsForSelection(params);
      var newState = {
        disabled: false,
        active: false,
        mode: null
      };
      if (this.canCreate(annos, sel)) {
        newState.mode = 'create';
      } else if (this.canFuse(annos, sel)) {
        newState.mode = 'fuse';
      } else if (this.canTruncate(annos, sel)) {
        newState.active = true;
        newState.mode = 'truncate';
      } else if (this.canExpand(annos, sel)) {
        newState.mode = 'expand';
      } else if (this.canDelete(annos, sel)) {
        newState.active = true;
        newState.mode = 'delete';
      } else {
        newState.disabled = true;
      }
      newState.showInContext = this.showInContext(sel, params);
      return newState
    };

    
    
    AnnotationCommand.prototype.execute = function execute (params) {
      
      
      
      var commandState = params.commandState;

      if (commandState.disabled) { return false }
      switch(commandState.mode) {
        case 'create':
          return this.executeCreate(params)
        case 'fuse':
          return this.executeFuse(params)
        case 'truncate':
          return this.executeTruncate(params)
        case 'expand':
          return this.executeExpand(params)
        case 'delete':
          return this.executeDelete(params)
        default:
          console.warn('Command.execute(): unknown mode', commandState.mode);
          return false
      }
    };

    AnnotationCommand.prototype.executeCreate = function executeCreate (params) {
      var annos = this._getAnnotationsForSelection(params);
      this._checkPrecondition(params, annos, this.canCreate);
      var editorSession = this._getEditorSession(params);
      var annoData = this.getAnnotationData();
      annoData.type = this.getAnnotationType();
      var anno;
      editorSession.transaction(function (tx) {
        anno = tx.annotate(annoData);
      });
      return {
        mode: 'create',
        anno: anno
      }
    };

    AnnotationCommand.prototype.executeFuse = function executeFuse (params) {
      var annos = this._getAnnotationsForSelection(params);
      this._checkPrecondition(params, annos, this.canFuse);
      this._applyTransform(params, function(tx) {
        annotationHelpers.fuseAnnotation(tx, annos);
      });
      return {
        mode: 'fuse',
        anno: annos[0]
      }
    };

    AnnotationCommand.prototype.executeTruncate = function executeTruncate (params) {
      var annos = this._getAnnotationsForSelection(params);
      var anno = annos[0];
      this._checkPrecondition(params, annos, this.canTruncate);
      this._applyTransform(params, function(tx) {
        annotationHelpers.truncateAnnotation(tx, anno, params.selection);
      });
      return {
        mode: 'truncate',
        anno: anno
      }
    };

    AnnotationCommand.prototype.executeExpand = function executeExpand (params) {
      var annos = this._getAnnotationsForSelection(params);
      var anno = annos[0];
      this._checkPrecondition(params, annos, this.canExpand);
      this._applyTransform(params, function(tx) {
        annotationHelpers.expandAnnotation(tx, anno, params.selection);
      });
      return {
        mode: 'expand',
        anno: anno
      }
    };

    AnnotationCommand.prototype.executeDelete = function executeDelete (params) {
      var annos = this._getAnnotationsForSelection(params);
      var anno = annos[0];
      this._checkPrecondition(params, annos, this.canDelete);
      this._applyTransform(params, function(tx) {
        return tx.delete(anno.id)
      });
      return {
        mode: 'delete',
        annoId: anno.id
      }
    };

    AnnotationCommand.prototype.isAnnotationCommand = function isAnnotationCommand () {
      return true
    };

    AnnotationCommand.prototype._checkPrecondition = function _checkPrecondition (params, annos, checker) {
      var sel = this._getSelection(params);
      if (!checker.call(this, annos, sel)) {
        throw new Error("AnnotationCommand: can't execute command for selection " + sel.toString())
      }
    };

    AnnotationCommand.prototype._getAnnotationsForSelection = function _getAnnotationsForSelection (params) {
      return params.selectionState.getAnnotationsForType(this.getAnnotationType())
    };

    
    AnnotationCommand.prototype._applyTransform = function _applyTransform (params, transformFn) {
      var sel = this._getSelection(params);
      if (sel.isNull()) { return }

      var editorSession = this._getEditorSession(params);
      var result; 
      editorSession.setSelection(sel);
      editorSession.transaction(function(tx) {
        var out = transformFn(tx, params);
        if (out) { result = out.result; }
      });
      return result
    };

    return AnnotationCommand;
  }(Command));

  var NodeComponent = (function (Component$$1) {
    function NodeComponent () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) NodeComponent.__proto__ = Component$$1;
    NodeComponent.prototype = Object.create( Component$$1 && Component$$1.prototype );
    NodeComponent.prototype.constructor = NodeComponent;

    NodeComponent.prototype.didMount = function didMount () {
      this.context.editorSession.onRender('document', this.rerender, this, { path: [this.props.node.id]});
    };

    NodeComponent.prototype.dispose = function dispose () {
      this.context.editorSession.off(this);
    };

    NodeComponent.prototype.render = function render ($$) {
      var tagName = this.getTagName();
      var el = $$(tagName)
        .attr('data-id', this.props.node.id)
        .addClass(this.getClassNames());
      return el
    };

    NodeComponent.prototype.getTagName = function getTagName () {
      return 'div'
    };

    NodeComponent.prototype.getClassNames = function getClassNames () {
      return ''
    };

    NodeComponent.prototype.rerender = function rerender () {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      
      if (this.props.node.isDisposed()) { return }
      Component$$1.prototype.rerender.apply(this, args);
    };

    return NodeComponent;
  }(Component));

  var BlockNodeComponent = (function (NodeComponent$$1) {
    function BlockNodeComponent () {
      NodeComponent$$1.apply(this, arguments);
    }if ( NodeComponent$$1 ) BlockNodeComponent.__proto__ = NodeComponent$$1;
    BlockNodeComponent.prototype = Object.create( NodeComponent$$1 && NodeComponent$$1.prototype );
    BlockNodeComponent.prototype.constructor = BlockNodeComponent;

    

    return BlockNodeComponent;
  }(NodeComponent));

  var Clipboard = (function (AbstractClipboard$$1) {
    function Clipboard(editorSession) {
      AbstractClipboard$$1.call(this, editorSession.getConfigurator());

      this.editorSession = editorSession;
    }

    if ( AbstractClipboard$$1 ) Clipboard.__proto__ = AbstractClipboard$$1;
    Clipboard.prototype = Object.create( AbstractClipboard$$1 && AbstractClipboard$$1.prototype );
    Clipboard.prototype.constructor = Clipboard;

    
    Clipboard.prototype._copy = function _copy () {
      var editorSession = this.getEditorSession();
      var sel = editorSession.getSelection();
      var doc = editorSession.getDocument();
      var clipboardDoc = null;
      var clipboardText = "";
      var clipboardHtml = "";
      var htmlExporter = this._getExporter();
      if (!sel.isCollapsed()) {
        clipboardText = getTextForSelection(doc, sel) || "";
        clipboardDoc = copySelection(doc, sel);
        clipboardHtml = htmlExporter.exportDocument(clipboardDoc);
      }
      return {
        doc: clipboardDoc,
        html: clipboardHtml,
        text: clipboardText
      }
    };

    Clipboard.prototype._cut = function _cut () {
      var editorSession = this.getEditorSession();
      editorSession.transaction(function (tx){
        tx.deleteSelection();
      });
    };

    
    Clipboard.prototype._pastePlainText = function _pastePlainText (plainText) {
      var editorSession = this.getEditorSession();
      editorSession.transaction(function(tx) {
        tx.paste(plainText);
      }, { action: 'paste' });
    };

    
    Clipboard.prototype._pasteHtml = function _pasteHtml (html, text) {
      var htmlImporter = this._getImporter();
      var content = htmlImporter.importDocument(html) || text;
      if (content) {
        var editorSession = this.getEditorSession();
        editorSession.transaction(function (tx) {
          tx.paste(content);
        }, { action: 'paste' });
      }
      return true
    };

    Clipboard.prototype.getEditorSession = function getEditorSession () {
      return this.editorSession
    };

    return Clipboard;
  }(AbstractClipboard));

  function findParentComponent(el) {
    while (el) {
      var comp = Component.unwrap(el);
      if (comp) { return comp }
      el = el.parentNode;
    }
  }

  function startsWith(str, prefix) {
    if (!isString(str)) { return false }
    if (str.startsWith) { return str.startsWith(prefix) }
    if (!isString(prefix)) { prefix = String(prefix); }
    return str.slice(0, prefix.length) === prefix
  }

  var UnsupportedNodeComponent = (function (Component$$1) {
    function UnsupportedNodeComponent () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) UnsupportedNodeComponent.__proto__ = Component$$1;
    UnsupportedNodeComponent.prototype = Object.create( Component$$1 && Component$$1.prototype );
    UnsupportedNodeComponent.prototype.constructor = UnsupportedNodeComponent;

    UnsupportedNodeComponent.prototype.render = function render ($$) {
      return $$('pre')
        .addClass('content-node unsupported')
        .attr({
          'data-id': this.props.node.id,
          contentEditable: false
        })
        .append(
          JSON.stringify(this.props.node.properties, null, 2)
        )
    };

    return UnsupportedNodeComponent;
  }(Component));

  var BROWSER_DELAY = platform.isFF ? 1 : 0;


  var Surface = (function (Component$$1) {
    function Surface() {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      Component$$1.apply(this, args);

      this._initialize();
    }

    if ( Component$$1 ) Surface.__proto__ = Component$$1;
    Surface.prototype = Object.create( Component$$1 && Component$$1.prototype );
    Surface.prototype.constructor = Surface;

    var prototypeAccessors = { id: { configurable: true } };

    Surface.prototype._initialize = function _initialize () {
      
      
      this.editorSession = this.props.editorSession || this.context.editorSession;
      if (!this.editorSession) {
        throw new Error('No EditorSession provided')
      }
      this.name = this.props.name;
      if (!this.name) {
        throw new Error('Surface must have a name.')
      }
      if (this.name.indexOf('/') > -1) {
        
        throw new Error("Surface.name must not contain '/'")
      }
      
      
      this._surfaceId = Surface.createSurfaceId(this);

      this.clipboard = new Clipboard(this.editorSession);

      this.domSelection = this.context.domSelection;
      if (!this.domSelection) { throw new Error('DOMSelection instance must be provided via context.') }

      this._state = {
        
        skipNextFocusEvent: false
      };
    };

    Surface.prototype.getChildContext = function getChildContext () {
      return {
        surface: this,
        parentSurfaceId: this.getId(),
        doc: this.getDocument(),
        
        
        isolatedNodeComponent: null
      }
    };

    Surface.prototype.didMount = function didMount () {
      var editorSession = this.getEditorSession();
      var surfaceManager = this.getSurfaceManager();
      if (surfaceManager) {
        surfaceManager.registerSurface(this);
      }
      editorSession.onRender('selection', this._onSelectionChanged, this);

      var globalEventHandler = editorSession.globalEventHandler;
      globalEventHandler.addEventListener('keydown', this._muteNativeHandlers, this);
    };


    Surface.prototype.dispose = function dispose () {
      var editorSession = this.getEditorSession();
      var surfaceManager = this.getSurfaceManager();
      editorSession.off(this);
      if (surfaceManager) {
        surfaceManager.unregisterSurface(this);
      }

      var globalEventHandler = editorSession.globalEventHandler;
      globalEventHandler.removeEventListener('keydown', this._muteNativeHandlers);
    };

    Surface.prototype.didUpdate = function didUpdate () {
      this._updateContentEditableState();
    };

    Surface.prototype.render = function render ($$) {
      var tagName = this.props.tagName || 'div';
      var el = $$(tagName)
        .addClass('sc-surface')
        .attr('tabindex', 2)
        .attr('data-surface-id', this.id);

      if (!this.isDisabled()) {
        if (this.isEditable()) {
          
          el.on('keydown', this.onKeyDown);
          
          if (!platform.isIE) {
            el.on('compositionstart', this.onCompositionStart);
            el.on('compositionend', this.onCompositionEnd);
          }
          
          
          
          
          if (platform.inBrowser && window.TextEvent && !platform.isIE) {
            el.on('textInput', this.onTextInput);
          } else {
            el.on('keypress', this.onTextInputShim);
          }
          el.on('paste', this._onPaste);
          el.on('cut', this._onCut);
        }
        if (!this.isReadonly()) {
          
          el.on('mousedown', this.onMouseDown);
          el.on('contextmenu', this.onContextMenu);
          
          
          el.on('focus', this.onNativeFocus);
          el.on('blur', this.onNativeBlur);
        }
        el.on('copy', this._onCopy);
      }
      return el
    };

    Surface.prototype.renderNode = function renderNode ($$, node) {
      var componentRegistry = this.getComponentRegistry();
      var ComponentClass = componentRegistry.get(node.type);
      if (!ComponentClass) {
        console.error('Could not resolve a component for type: ' + node.type);
        ComponentClass = UnsupportedNodeComponent;
      }
      return $$(ComponentClass, this._extractNodeProps(node)).ref(node.id)
    };

    Surface.prototype.getComponentRegistry = function getComponentRegistry () {
      return this.context.componentRegistry || this.props.componentRegistry
    };

    Surface.prototype.getName = function getName () {
      return this.name
    };

    Surface.prototype.getId = function getId () {
      return this._surfaceId
    };

    Surface.prototype.getSurfaceId = function getSurfaceId () {
      return this._surfaceId
    };

    Surface.prototype.isDisabled = function isDisabled () {
      return this.props.disabled
    };

    Surface.prototype.isEditable = function isEditable () {
      return (this.props.editing === "full" || this.props.editing === undefined)
    };

    Surface.prototype.isSelectable = function isSelectable () {
      return (this.props.editing === "selection" || this.props.editing === "full")
    };

    Surface.prototype.isReadonly = function isReadonly () {
      return this.props.editing === "readonly"
    };

    Surface.prototype.getElement = function getElement () {
      return this.el
    };

    Surface.prototype.getDocument = function getDocument () {
      return this.getEditorSession().getDocument()
    };

    Surface.prototype.getEditorSession = function getEditorSession () {
      return this.editorSession
    };

    Surface.prototype.getSurfaceManager = function getSurfaceManager () {
      return this.context.surfaceManager
    };

    Surface.prototype.isEnabled = function isEnabled () {
      return !this.state.disabled
    };

    Surface.prototype.isContainerEditor = function isContainerEditor () {
      return false
    };

    Surface.prototype.isCustomEditor = function isCustomEditor () {
      return false
    };

    Surface.prototype.hasNativeSpellcheck = function hasNativeSpellcheck () {
      return this.props.spellcheck === 'native'
    };

    Surface.prototype.getContainerId = function getContainerId () {
      return null
    };

    Surface.prototype.focus = function focus () {
      if (this.getEditorSession().getFocusedSurface() !== this) {
        this.selectFirst();
      }
    };

    Surface.prototype.blur = function blur () {
      if (this.getEditorSession().getFocusedSurface() === this) {
        this.getEditorSession().setSelection(null);
      }
    };

    Surface.prototype.selectFirst = function selectFirst () {
      throw new Error('This method is abstract.')
    };

    
    Surface.prototype.rerenderDOMSelection = function rerenderDOMSelection () {
      if (this.isDisabled()) { return }
      if (platform.inBrowser) {
        
        var sel = this.getEditorSession().getSelection();
        if (sel.surfaceId === this.getId()) {
          this.domSelection.setSelection(sel);
          
          
          
          var scrollPane = this.context.scrollPane;
          if (scrollPane) {
            this.context.scrollPane.onSelectionPositioned();
          }
        }
      }
    };

    Surface.prototype.getDomNodeForId = function getDomNodeForId (nodeId) {
      return this.el.getNativeElement().querySelector('*[data-id="'+nodeId+'"]')
    };

    

    
    Surface.prototype.onKeyDown = function onKeyDown (event) {
      if (!this._shouldConsumeEvent(event)) { return }
      

      
      if ( event.key === 'Dead' ) { return }

      
      var custom = this.getEditorSession().keyboardManager.onKeydown(event);
      if (!custom) {
        
        switch ( event.keyCode ) {
          
          case keys$1.LEFT:
          case keys$1.RIGHT:
            return this._handleLeftOrRightArrowKey(event)
          case keys$1.UP:
          case keys$1.DOWN:
            return this._handleUpOrDownArrowKey(event)
          case keys$1.HOME:
          case keys$1.END:
            return this._handleHomeOrEndKey(event)
          case keys$1.PAGEUP:
          case keys$1.PAGEDOWN:
            return this._handlePageUpOrDownKey(event)
          
          case keys$1.ENTER:
            return this._handleEnterKey(event)
          case keys$1.TAB:
            return this._handleTabKey(event)
          case keys$1.BACKSPACE:
          case keys$1.DELETE:
            return this._handleDeleteKey(event)
          case keys$1.ESCAPE:
            return this._handleEscapeKey(event)
          case keys$1.SPACE:
            return this._handleSpaceKey(event)
          default:
            break
        }
      }
    };

    Surface.prototype.onTextInput = function onTextInput (event) {
      if (!this._shouldConsumeEvent(event)) { return }
      
      event.preventDefault();
      event.stopPropagation();
      if (!event.data) { return }
      var editorSession = this.getEditorSession();
      var text = event.data;
      if (!editorSession.keyboardManager.onTextInput(text)) {
        editorSession.transaction(function (tx) {
          tx.insertText(text);
        }, { action: 'type' });
      }
    };

    
    Surface.prototype.onCompositionStart = function onCompositionStart (event) {
      if (!this._shouldConsumeEvent(event)) { return }
      
      
      
      
      
      
      
      if (event.data) {
        var l = event.data.length;
        var sel = this.getEditorSession().getSelection();
        if (sel.isPropertySelection() && sel.isCollapsed()) {
          
          var offset = sel.start.offset;
          this.getEditorSession().setSelection(sel.createWithNewRange(offset-l, offset));
        }
      }
    };

    Surface.prototype.onCompositionEnd = function onCompositionEnd (event) {
      var this$1 = this;

      if (!this._shouldConsumeEvent(event)) { return }
      
      
      
      if (platform.isFF) {
        event.preventDefault();
        event.stopPropagation();
        if (!event.data) { return }
        this._delayed(function () {
          var text = event.data;
          if (!this$1.getEditorSession().keyboardManager.onTextInput(text)) {
            this$1.getEditorSession().transaction(function (tx) {
              tx.insertText(text);
            }, { action: 'type' });
          }
        });
      }
    };

    
    Surface.prototype.onTextInputShim = function onTextInputShim (event) {
      if (!this._shouldConsumeEvent(event)) { return }
      
      if (
        
        event.which === 0 || event.charCode === 0 ||
        
        event.keyCode === keys$1.TAB || event.keyCode === keys$1.ESCAPE ||
        
        Boolean(event.metaKey) || (Boolean(event.ctrlKey)^Boolean(event.altKey))
      ) {
        return
      }
      var character = String.fromCharCode(event.which);
      if (!event.shiftKey) {
        character = character.toLowerCase();
      }
      event.preventDefault();
      event.stopPropagation();
      if (!this.getEditorSession().keyboardManager.onTextInput(character)) {
        if (character.length>0) {
          this.getEditorSession().transaction(function (tx) {
            tx.insertText(character);
          }, { action: 'type' });
        }
      }
    };

    
    
    
    
    
    Surface.prototype.onMouseDown = function onMouseDown (event) {
      if (!this._shouldConsumeEvent(event)) {
        
        return
      }

      
      
      
      
      
      if (event.__reserved__) {
        
        return
      } else {
        
        event.__reserved__ = this;
      }

      
      
      
      if (this.isEditable()) {
        this.el.setAttribute('contenteditable', true);
      }

      
      if ( event.button !== 0 ) {
        return
      }

      
      if (!(platform.isIE && platform.version<12) && event.detail >= 3) {
        var sel = this.getEditorSession().getSelection();
        if (sel.isPropertySelection()) {
          this._selectProperty(sel.path);
          event.preventDefault();
          event.stopPropagation();
          return
        } else if (sel.isContainerSelection()) {
          this._selectProperty(sel.startPath);
          event.preventDefault();
          event.stopPropagation();
          return
        }
      }
      
      
      
      
      this._state.skipNextFocusEvent = true;

      
      if (platform.inBrowser) {
        var documentEl = DefaultDOMElement.wrapNativeElement(window.document);
        documentEl.on('mouseup', this.onMouseUp, this, { once: true });
      }
    };

    Surface.prototype.onMouseUp = function onMouseUp (e) {
      var this$1 = this;

      
      
      
      
      
      e.stopPropagation();
      
      
      
      
      this._delayed(function () {
        var sel = this$1.domSelection.getSelection();
        this$1._setSelection(sel);
      });
    };

    
    
    
    Surface.prototype.onContextMenu = function onContextMenu (event) {
      if (!this._shouldConsumeEvent(event)) { return }
      var sel = this.domSelection.getSelection();
      this._setSelection(sel);
    };

    Surface.prototype.onNativeBlur = function onNativeBlur () {
      
      var _state = this._state;
      _state.hasNativeFocus = false;
    };

    Surface.prototype.onNativeFocus = function onNativeFocus () {
      
      var _state = this._state;
      _state.hasNativeFocus = true;
    };

    Surface.prototype._onCopy = function _onCopy (e) {
      this.clipboard.onCopy(e);
    };

    Surface.prototype._onCut = function _onCut (e) {
      this.clipboard.onCut(e);
    };

    Surface.prototype._onPaste = function _onPaste (e) {
      this.clipboard.onPaste(e);
    };

    

    Surface.prototype._onSelectionChanged = function _onSelectionChanged (selection) {
      var newMode = this._deriveModeFromSelection(selection);
      if (this.state.mode !== newMode) {
        this.extendState({
          mode: newMode
        });
      }
    };

    
    Surface.prototype._deriveModeFromSelection = function _deriveModeFromSelection (sel) {
      if (!sel) { return null }
      var surfaceId = sel.surfaceId;
      var id = this.getId();
      var mode;
      if (startsWith(surfaceId, id)) {
        if (surfaceId.length === id.length) {
          mode = 'focused';
        } else {
          mode = 'co-focused';
        }
      }
      return mode
    };

    Surface.prototype._updateContentEditableState = function _updateContentEditableState () {
      
      
      
      
      
      var enableContenteditable = false;
      if (this.isEditable() && !this.props.disabled) {
        enableContenteditable = true;
        if (this.state.mode === 'co-focused') {
          var selState = this.context.editorSession.getSelectionState();
          var sel = selState.getSelection();
          var surface = this.getSurfaceManager().getSurface(sel.surfaceId);
          if (surface) {
            var isolatedNodeComponent = surface.context.isolatedNodeComponent;
            if (isolatedNodeComponent) {
              enableContenteditable = isolatedNodeComponent.isOpen();
            }
          }
        }
      }
      if (enableContenteditable) {
        this.el.setAttribute('contenteditable', true);
      } else {
        
        this.el.removeAttribute('contenteditable');
      }
    };

    Surface.prototype._blur = function _blur () {
      if (this.el) {
        this.el.blur();
      }
    };

    Surface.prototype._focus = function _focus () {
      if (this.isDisabled()) { return }
      
      
      
      if (platform.isFF) {
        this.domSelection.clear();
        this.el.getNativeElement().blur();
      }
      this._focusElement();
    };

    Surface.prototype._focusElement = function _focusElement () {
      this._state.hasNativeFocus = true;
      
      
      
      
      if (this.el && !platform.isWebkit) {
        this._state.skipNextFocusEvent = true;
        
        
        this.el.focus();
        this._state.skipNextFocusEvent = false;
      }
    };

    Surface.prototype._handleLeftOrRightArrowKey = function _handleLeftOrRightArrowKey (event) {
      var this$1 = this;

      event.stopPropagation();
      var direction = (event.keyCode === keys$1.LEFT) ? 'left' : 'right';
      
      
      this._delayed(function () {
        this$1._updateModelSelection({direction: direction});
      });
    };

    Surface.prototype._handleUpOrDownArrowKey = function _handleUpOrDownArrowKey (event) {
      var this$1 = this;

      event.stopPropagation();
      
      
      this._delayed(function () {
        var options = {
          direction: (event.keyCode === keys$1.UP) ? 'left' : 'right'
        };
        this$1._updateModelSelection(options);
      });
    };

    Surface.prototype._handleHomeOrEndKey = function _handleHomeOrEndKey (event) {
      var this$1 = this;

      event.stopPropagation();
      
      
      this._delayed(function () {
        var options = {
          direction: (event.keyCode === keys$1.HOME) ? 'left' : 'right'
        };
        this$1._updateModelSelection(options);
      });
    };

    Surface.prototype._handlePageUpOrDownKey = function _handlePageUpOrDownKey (event) {
      var this$1 = this;

      event.stopPropagation();
      
      
      this._delayed(function () {
        var options = {
          direction: (event.keyCode === keys$1.PAGEUP) ? 'left' : 'right'
        };
        this$1._updateModelSelection(options);
      });
    };

    Surface.prototype._handleSpaceKey = function _handleSpaceKey (event) {
      event.stopPropagation();
      event.preventDefault();
      var text = ' ';
      if (!this.getEditorSession().keyboardManager.onTextInput(text)) {
        this.getEditorSession().transaction(function (tx) {
          tx.insertText(text);
        }, { action: 'type' });
      }
    };

    Surface.prototype._handleTabKey = function _handleTabKey (event) {
      event.stopPropagation();
      this.el.emit('tab', {
        altKey: event.altKey,
        ctrlKey: event.ctrlKey,
        metaKey: event.metaKey,
        shiftKey: event.shiftKey,
        code: event.code
      });
      if (this.props.handleTab === false) {
        event.preventDefault();
      } else {
        this.__handleTab(event);
      }
    };

    Surface.prototype.__handleTab = function __handleTab () {
      var this$1 = this;

      
      
      this._delayed(function () {
        this$1._updateModelSelection();
      });
    };

    Surface.prototype._handleEnterKey = function _handleEnterKey (event) {
      event.preventDefault();
      event.stopPropagation();
      this.getEditorSession().transaction(function (tx) {
        tx.break();
      }, { action: 'break' });
    };

    Surface.prototype._handleEscapeKey = function _handleEscapeKey () {};

    Surface.prototype._handleDeleteKey = function _handleDeleteKey (event) {
      event.preventDefault();
      event.stopPropagation();
      var direction = (event.keyCode === keys$1.BACKSPACE) ? 'left' : 'right';
      this.getEditorSession().transaction(function (tx) {
        tx.deleteCharacter(direction);
      }, { action: 'delete' });
    };

    Surface.prototype._hasNativeFocus = function _hasNativeFocus () {
      return Boolean(this._state.hasNativeFocus)
    };

    Surface.prototype._setSelection = function _setSelection (sel) {
      
      
      
      
      
      
      
      
      if (!sel.isNull() && sel.surfaceId === this.id && platform.isFF) {
        this._focusElement();
      }
      this.getEditorSession().setSelection(sel);
    };

    Surface.prototype._updateModelSelection = function _updateModelSelection (options) {
      var sel = this.domSelection.getSelection(options);
      
      
      
      this._setSelection(sel);
    };

    Surface.prototype._selectProperty = function _selectProperty (path) {
      var doc = this.getDocument();
      var text = doc.get(path);
      this._setSelection(doc.createSelection({
        type: 'property',
        path: path,
        startOffset: 0,
        endOffset: text.length
      }));
    };

    
    
    Surface.prototype._renderNode = function _renderNode ($$, nodeId) {
      var doc = this.getDocument();
      var node = doc.get(nodeId);
      var componentRegistry = this.context.componentRegistry || this.props.componentRegistry;
      var ComponentClass = componentRegistry.get(node.type);
      if (!ComponentClass) {
        console.error('Could not resolve a component for type: ' + node.type);
        ComponentClass = UnsupportedNodeComponent;
      }
      return $$(ComponentClass, {
        doc: doc,
        node: node
      })
    };

    Surface.prototype._extractNodeProps = function _extractNodeProps (node) {
      var doc = this.getDocument();
      return {
        placeholder: this.props.placeholder,
        doc: doc,
        node: node
      }
    };

    
    Surface.prototype._shouldConsumeEvent = function _shouldConsumeEvent (event) {
      
      var comp = Component$$1.unwrap(event.target);
      return (comp && (comp === this || comp.context.surface === this))
    };

    
    Surface.prototype.getSelectionFromEvent = function getSelectionFromEvent (event) {
      var domRange = getDOMRangeFromEvent(event);
      var sel = this.domSelection.getSelectionForDOMRange(domRange);
      sel.surfaceId = this.getId();
      return sel;
    };

    Surface.prototype.setSelectionFromEvent = function setSelectionFromEvent (event) {
      var sel = this.getSelectionFromEvent(event);
      if (sel) {
        this._state.skipNextFocusEvent = true;
        this._setSelection(sel);
      } else {
        console.error('Could not create a selection from event.');
      }
    };

    prototypeAccessors.id.get = function () {
      return this._surfaceId
    };

    Surface.prototype._delayed = function _delayed (fn) {
      window.setTimeout(fn, BROWSER_DELAY);
    };

    
    Surface.prototype._muteNativeHandlers = function _muteNativeHandlers (event) {
      var contentEditableShortcuts;

      if (platform.isMac) {
        contentEditableShortcuts = [
          'META+66', 
          'META+73', 
          'META+85' 
        ];
      } else {
        contentEditableShortcuts = [
          'CTRL+66', 
          'CTRL+73', 
          'CTRL+85' 
        ];
      }

      var key = parseKeyEvent(event);
      if(contentEditableShortcuts.indexOf(key) > -1) {
        event.preventDefault();
      }
    };

    Object.defineProperties( Surface.prototype, prototypeAccessors );

    return Surface;
  }(Component));

  Surface.prototype._isSurface = true;


  Surface.createSurfaceId = function(surface) {
    var parentSurfaceId = surface.context.parentSurfaceId;
    if (parentSurfaceId) {
      return parentSurfaceId + '/' + surface.name
    } else {
      return surface.name
    }
  };

  var ContainerEditor = (function (Surface$$1) {
    function ContainerEditor(parent, props, el) {
      
      props.containerId = props.containerId || props.node.id;
      props.name = props.name || props.containerId || props.node.id;

      Surface$$1.call(this, parent, props, el);

      this.containerId = this.props.containerId;
      if (!isString(this.containerId)) {
        throw new Error("Property 'containerId' is mandatory.")
      }
      var doc = this.getDocument();
      this.container = doc.get(this.containerId);
      if (!this.container) {
        throw new Error('Container with id ' + this.containerId + ' does not exist.')
      }

      this.editingBehavior = this.context.editingBehavior || new EditingBehavior();

      this._deriveInternalState(this.props);
    }

    if ( Surface$$1 ) ContainerEditor.__proto__ = Surface$$1;
    ContainerEditor.prototype = Object.create( Surface$$1 && Surface$$1.prototype );
    ContainerEditor.prototype.constructor = ContainerEditor;

    
    ContainerEditor.prototype.shouldRerender = function shouldRerender (newProps) {
      if (newProps.disabled !== this.props.disabled) { return true }
      
      
      return false
    };

    ContainerEditor.prototype.willReceiveProps = function willReceiveProps (newProps) {
      Surface$$1.prototype.willReceiveProps.apply(this, arguments);
      this._deriveInternalState(newProps);
    };

    ContainerEditor.prototype.didMount = function didMount () {
      Surface$$1.prototype.didMount.apply(this, arguments);
      var editorSession = this.getEditorSession();
      editorSession.onUpdate('document', this._onContainerChanged, this, {
        path:  this.container.getContentPath()
      });
      this._attachPlaceholder();
    };

    ContainerEditor.prototype.dispose = function dispose () {
      Surface$$1.prototype.dispose.apply(this, arguments);
      var editorSession = this.getEditorSession();
      editorSession.off(this);
    };

    ContainerEditor.prototype.render = function render ($$) {
      var el = Surface$$1.prototype.render.call(this, $$);

      var doc = this.getDocument();
      var containerId = this.getContainerId();
      var containerNode = doc.get(containerId);
      if (!containerNode) {
        console.warn('No container node found for ', containerId);
      }
      el.addClass('sc-container-editor container-node ' + containerId)
        .attr("data-id", containerId);

      
      el.attr('spellcheck', this.props.spellcheck === 'native');

      containerNode.getNodes().forEach(function(node, index) {
        el.append(this._renderNode($$, node, index));
      }.bind(this));

      
      if (!this.props.disabled && !this.isEmpty()) {
        el.addClass('sm-enabled');
        el.setAttribute('contenteditable', true);
      }

      return el
    };

    ContainerEditor.prototype.selectFirst = function selectFirst () {
      var container = this.getContainer();
      if (container.getLength() > 0) {
        var editorSession = this.getEditorSession();
        var first = container.getChildAt(0);
        setCursor(editorSession, first, container.id, 'before');
      }
    };

    ContainerEditor.prototype._renderNode = function _renderNode ($$, node, nodeIndex) {
      var props = { node: node };
      if (!node) { throw new Error('Illegal argument') }
      if (node.isText()) {
        return this.renderNode($$, node, nodeIndex)
      } else {
        var componentRegistry = this.context.componentRegistry;
        var ComponentClass = componentRegistry.get(node.type);
        if (ComponentClass.prototype._isCustomNodeComponent || ComponentClass.prototype._isIsolatedNodeComponent) {
          return $$(ComponentClass, props).ref(node.id)
        } else {
          return $$(IsolatedNodeComponent, props).ref(node.id)
        }
      }
    };

    ContainerEditor.prototype._extractNodeProps = function _extractNodeProps (node) {
      var doc = this.getDocument();
      return {
        doc: doc,
        node: node
      }
    };

    ContainerEditor.prototype._deriveInternalState = function _deriveInternalState (props) {
      var _state = this._state;
      if (!props.hasOwnProperty('enabled') || props.enabled) {
        _state.enabled = true;
      } else {
        _state.enabled = false;
      }
    };

    ContainerEditor.prototype._selectNextIsolatedNode = function _selectNextIsolatedNode (direction) {
      var selState = this.getEditorSession().getSelectionState();
      var node = (direction === 'left') ? selState.getPreviousNode() : selState.getNextNode();
      var isIsolatedNode = !node.isText() && !node.isList();
      if (!node || !isIsolatedNode) { return false }
      if (
        (direction === 'left' && selState.isFirst()) ||
        (direction === 'right' && selState.isLast())
      ) {
        this.getEditorSession().setSelection({
          type: 'node',
          nodeId: node.id,
          containerId: selState.getContainer().id,
          surfaceId: this.id
        });
        return true
      }
      return false
    };

    ContainerEditor.prototype._handleLeftOrRightArrowKey = function _handleLeftOrRightArrowKey (event) {
      var this$1 = this;

      event.stopPropagation();
      var doc = this.getDocument();
      var sel = this.getEditorSession().getSelection();
      var left = (event.keyCode === keys$1.LEFT);
      var right = !left;
      var direction = left ? 'left' : 'right';

      if (sel && !sel.isNull()) {
        var container = doc.get(sel.containerId, 'strict');

        
        if (sel.isNodeSelection()) {
          var nodePos = container.getPosition(doc.get(sel.getNodeId()));
          if ((left && nodePos === 0) || (right && nodePos === container.length-1)) {
            event.preventDefault();
            return
          }
        }

        if (sel.isNodeSelection() && !event.shiftKey) {
          this.domSelection.collapse(direction);
        }
      }

      this._delayed(function () {
        this$1._updateModelSelection({ direction: direction });
      });
    };

    ContainerEditor.prototype._handleUpOrDownArrowKey = function _handleUpOrDownArrowKey (event) {
      var this$1 = this;

      event.stopPropagation();
      var doc = this.getDocument();
      var sel = this.getEditorSession().getSelection();
      var up = (event.keyCode === keys$1.UP);
      var down = !up;
      var direction = up ? 'left' : 'right';

      if (sel && !sel.isNull()) {
        var container = doc.get(sel.containerId, 'strict');
        
        if (sel.isNodeSelection()) {
          var nodePos = container.getPosition(doc.get(sel.getNodeId()));
          if ((up && nodePos === 0) || (down && nodePos === container.length-1)) {
            event.preventDefault();
            return
          }
          
          
          var editorSession = this.getEditorSession();
          
          
          
          if (!event.shiftKey) {
            event.preventDefault();
            if (up) {
              var prev = container.getChildAt(nodePos-1);
              setCursor(editorSession, prev, sel.containerId, 'after');
              return
            } else {
              var next = container.getChildAt(nodePos+1);
              setCursor(editorSession, next, sel.containerId, 'before');
              return
            }
          }
        }
      }

      this._delayed(function () {
        this$1._updateModelSelection({ direction: direction });
      });
    };

    ContainerEditor.prototype._handleTabKey = function _handleTabKey (event) {
      var editorSession = this.getEditorSession();
      var sel = editorSession.getSelection();
      
      if (sel.isNodeSelection() && sel.isFull()) {
        var comp = this.refs[sel.getNodeId()];
        if (comp && stepIntoIsolatedNode(editorSession, comp)) {
          event.preventDefault();
          event.stopPropagation();
          return
        }
      }
      Surface$$1.prototype._handleTabKey.call(this, event);
    };

    ContainerEditor.prototype.__handleTab = function __handleTab (e) {
      e.preventDefault();
      if (e.shiftKey) {
        this.getEditorSession().transaction(function (tx) {
          tx.dedent();
        }, { action: 'dedent' });
      } else {
        this.getEditorSession().transaction(function (tx) {
          tx.indent();
        }, { action: 'indent' });
      }
    };

    
    ContainerEditor.prototype.isContainerEditor = function isContainerEditor () {
      return true
    };

    
    ContainerEditor.prototype.getContainerId = function getContainerId () {
      return this.containerId
    };

    ContainerEditor.prototype.getContainer = function getContainer () {
      return this.getDocument().get(this.getContainerId())
    };

    ContainerEditor.prototype.isEmpty = function isEmpty () {
      var containerNode = this.getContainer();
      return (containerNode && containerNode.length === 0)
    };

    
    ContainerEditor.prototype._attachPlaceholder = function _attachPlaceholder () {
      var firstNode = this.childNodes[0];
      
      if (this.placeholderNode) {
        this.placeholderNode.extendProps({
          placeholder: undefined
        });
      }

      if (this.childNodes.length === 1 && this.props.placeholder) {
        firstNode.extendProps({
          placeholder: this.props.placeholder
        });
        this.placeholderNode = firstNode;
      }
    };

    ContainerEditor.prototype.isEditable = function isEditable () {
      return Surface$$1.prototype.isEditable.call(this) && !this.isEmpty()
    };

    
    ContainerEditor.prototype._onContainerChanged = function _onContainerChanged (change) {
      var this$1 = this;

      var doc = this.getDocument();
      
      var renderContext = RenderingEngine.createContext(this);
      var $$ = renderContext.$$;
      var container = this.getContainer();
      var path = container.getContentPath();
      for (var i = 0; i < change.ops.length; i++) {
        var op = change.ops[i];
        if (op.type === "update" && op.path[0] === path[0]) {
          var diff = op.diff;
          if (diff.type === "insert") {
            var nodeId = diff.getValue();
            var node = doc.get(nodeId);
            var nodeEl = (void 0);
            if (node) {
              nodeEl = this$1._renderNode($$, node);
            } else {
              
              
              
              nodeEl = $$('div');
            }
            this$1.insertAt(diff.getOffset(), nodeEl);
          } else if (diff.type === "delete") {
            this$1.removeAt(diff.getOffset());
          }
        }
      }
      this._attachPlaceholder();
    };

    return ContainerEditor;
  }(Surface));

  ContainerEditor.prototype._isContainerEditor = true;

  var CustomSurface = (function (Component$$1) {
    function CustomSurface() {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      Component$$1.apply(this, args);

      this._name = this._getCustomResourceId();
      this._surfaceId = this._createSurfaceId();
    }

    if ( Component$$1 ) CustomSurface.__proto__ = Component$$1;
    CustomSurface.prototype = Object.create( Component$$1 && Component$$1.prototype );
    CustomSurface.prototype.constructor = CustomSurface;

    var prototypeAccessors = { name: { configurable: true } };

    CustomSurface.prototype.getChildContext = function getChildContext () {
      return {
        surface: this,
        parentSurfaceId: this.getId()
        
        
        
      }
    };


    CustomSurface.prototype.didMount = function didMount () {
      var surfaceManager = this.context.editorSession.surfaceManager;
      surfaceManager.registerSurface(this);
    };

    CustomSurface.prototype.dispose = function dispose () {
      var surfaceManager = this.context.editorSession.surfaceManager;
      surfaceManager.unregisterSurface(this);
    };

    CustomSurface.prototype.rerenderDOMSelection = function rerenderDOMSelection () {
      
    };

    prototypeAccessors.name.get = function () {
      return this._name
    };

    CustomSurface.prototype.getId = function getId () {
      return this._surfaceId
    };

    CustomSurface.prototype.getSurfaceId = function getSurfaceId () {
      return this.getId()
    };

    CustomSurface.prototype.getContainer = function getContainer () {
      return undefined
    };

    CustomSurface.prototype.getContainerId = function getContainerId () {
      return undefined
    };

    CustomSurface.prototype.isContainerEditor = function isContainerEditor () {
      return false
    };

    CustomSurface.prototype.isCustomEditor = function isCustomEditor () {
      return true
    };

    CustomSurface.prototype.isDisabled = function isDisabled () {
      return Boolean(this.props.disabled)
    };

    CustomSurface.prototype._focus = function _focus () {
      
    };

    CustomSurface.prototype._createSurfaceId = function _createSurfaceId () {
      var isolatedNodeComponent = this.context.isolatedNodeComponent;
      if (isolatedNodeComponent) {
        var parentSurface = isolatedNodeComponent.context.surface;
        return parentSurface.id + '/' + isolatedNodeComponent.props.node.id + '/' + this._name
      } else {
        return this._name
      }
    };

    CustomSurface.prototype._getCustomResourceId = function _getCustomResourceId () {
      throw new Error('This method needs to be implemented by a CustomSurface')
    };

    Object.defineProperties( CustomSurface.prototype, prototypeAccessors );

    return CustomSurface;
  }(Component));

  var EditAnnotationCommand = (function (Command$$1) {
    function EditAnnotationCommand() {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      Command$$1.apply(this, args);

      if (!this.config.nodeType) {
        throw new Error("'nodeType' is required")
      }
    }

    if ( Command$$1 ) EditAnnotationCommand.__proto__ = Command$$1;
    EditAnnotationCommand.prototype = Object.create( Command$$1 && Command$$1.prototype );
    EditAnnotationCommand.prototype.constructor = EditAnnotationCommand;

    
    EditAnnotationCommand.prototype.getCommandState = function getCommandState (params) {
      var sel = this._getSelection(params);
      var annos = this._getAnnotationsForSelection(params);
      var newState = {
        disabled: true,
      };
      if (annos.length === 1 && sel.isPropertySelection() && sel.isCollapsed()) {
        newState.disabled = false;
        newState.showInContext = true;
        newState.nodeId = annos[0].id;
      }
      return newState
    };

    EditAnnotationCommand.prototype.execute = function execute (params) { }; 

    EditAnnotationCommand.prototype._getAnnotationsForSelection = function _getAnnotationsForSelection (params) {
      return params.selectionState.getAnnotationsForType(this.config.nodeType)
    };

    return EditAnnotationCommand;
  }(Command));

  var EditInlineNodeCommand = (function (Command$$1) {
    function EditInlineNodeCommand() {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      Command$$1.apply(this, args);
      if (!this.config.nodeType) {
        throw new Error('Every AnnotationCommand must have a nodeType')
      }
    }

    if ( Command$$1 ) EditInlineNodeCommand.__proto__ = Command$$1;
    EditInlineNodeCommand.prototype = Object.create( Command$$1 && Command$$1.prototype );
    EditInlineNodeCommand.prototype.constructor = EditInlineNodeCommand;

    EditInlineNodeCommand.prototype.getCommandState = function getCommandState (params) {
      var sel = params.selection;
      var newState = {
        disabled: true,
        active: false
      };
      var annos = this._getAnnotationsForSelection(params);
      if (annos.length === 1 && annos[0].getSelection().equals(sel)) {
        newState.disabled = false;
        newState.nodeId = annos[0].id;
      }
      return newState
    };

    EditInlineNodeCommand.prototype.execute = function execute (params) { 

    };

    EditInlineNodeCommand.prototype._getAnnotationsForSelection = function _getAnnotationsForSelection (params) {
      return params.selectionState.getAnnotationsForType(this.config.nodeType)
    };

    return EditInlineNodeCommand;
  }(Command));

  var Highlights = (function (EventEmitter$$1) {
    function Highlights(doc) {
      EventEmitter$$1.call(this);

      this.doc = doc;
      this._highlights = {};
    }

    if ( EventEmitter$$1 ) Highlights.__proto__ = EventEmitter$$1;
    Highlights.prototype = Object.create( EventEmitter$$1 && EventEmitter$$1.prototype );
    Highlights.prototype.constructor = Highlights;

    
    Highlights.prototype.get = function get () {
      return this._highlights
    };

    
    Highlights.prototype.set = function set (highlights) {
      var oldHighlights = this._highlights;
      var doc = this.doc;
      
      forEach(highlights, function(newScopedHighlights, scope) {
        var oldScopedHighlights = oldHighlights[scope] || [];

        
        
        
        var toBeDeleted = without(oldScopedHighlights, newScopedHighlights);
        var toBeAdded = without(newScopedHighlights, oldScopedHighlights);

        
        forEach(toBeDeleted, function(nodeId) {
          var node = doc.get(nodeId);
          
          if (node) {
            if (node.setHighlighted) {
              node.setHighlighted(false, scope);
            } else {
              console.warn('setHighlighted is not defined on target node');
            }
          }
        });

        forEach(toBeAdded, function(nodeId) {
          var node = doc.get(nodeId);
          if (node) {
            if (node.setHighlighted) {
              node.setHighlighted(true, scope);
            } else {
              console.warn('setHighlighted is not defined on target node');
            }
          }
        });
      });

      this._highlights = highlights;

      
      this.emit('highlights:updated', highlights);
    };

    return Highlights;
  }(EventEmitter));

  var InsertInlineNodeCommand = (function (Command$$1) {
    function InsertInlineNodeCommand() {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      Command$$1.apply(this, args);
    }

    if ( Command$$1 ) InsertInlineNodeCommand.__proto__ = Command$$1;
    InsertInlineNodeCommand.prototype = Object.create( Command$$1 && Command$$1.prototype );
    InsertInlineNodeCommand.prototype.constructor = InsertInlineNodeCommand;

    
    InsertInlineNodeCommand.prototype.getCommandState = function getCommandState (params) {
      var sel = params.selection;
      var newState = {
        disabled: this.isDisabled(params),
        active: false,
        showInContext: this.showInContext(sel, params)
      };
      return newState
    };

    
    InsertInlineNodeCommand.prototype.showInContext = function showInContext (sel) {
      return !sel.isCollapsed()
    };

    InsertInlineNodeCommand.prototype.isDisabled = function isDisabled (params) {
      var sel = params.selection;
      var selectionState = params.editorSession.getSelectionState();
      if (!sel.isPropertySelection()) {
        return true
      }

      
      
      if (selectionState.isInlineNodeSelection()) {
        return true
      }
      return false
    };

    
    InsertInlineNodeCommand.prototype.isAnnotationCommand = function isAnnotationCommand () {
      return true
    };

    
    InsertInlineNodeCommand.prototype.execute = function execute (params) {
      var this$1 = this;

      var state = this.getCommandState(params);
      if (state.disabled) { return }
      var editorSession = this._getEditorSession(params);
      editorSession.transaction(function (tx) {
        var nodeData = this$1.createNodeData(tx, params);
        tx.insertInlineNode(nodeData);
      });
    };

    InsertInlineNodeCommand.prototype.createNodeData = function createNodeData (tx) { 
      throw new Error('This method is abstract')
    };

    return InsertInlineNodeCommand;
  }(Command));

  var InsertNodeCommand = (function (Command$$1) {
    function InsertNodeCommand(config) {
      Command$$1.call(this, config);

      
      
      
      if (!this.config.nodeType) {
        console.error("'config.nodeType' should be provided for InsertNodeCommand");
      }
    }

    if ( Command$$1 ) InsertNodeCommand.__proto__ = Command$$1;
    InsertNodeCommand.prototype = Object.create( Command$$1 && Command$$1.prototype );
    InsertNodeCommand.prototype.constructor = InsertNodeCommand;

    InsertNodeCommand.prototype.getType = function getType () {
      return this.config.nodeType
    };

    InsertNodeCommand.prototype.getCommandState = function getCommandState (params) {
      var sel = params.selection;
      var newState = {
        disabled: true,
        active: false
      };
      if (sel && !sel.isNull() && !sel.isCustomSelection() && sel.containerId) {
        newState.disabled = false;
      }
      newState.showInContext = this.showInContext(sel, params);
      return newState
    };

    InsertNodeCommand.prototype.showInContext = function showInContext (sel, params) {
      var selectionState = params.selectionState;
      return sel.isCollapsed() && selectionState.isFirst() && selectionState.isLast()
    };

    InsertNodeCommand.prototype.isInsertCommand = function isInsertCommand () {
      return true
    };

    InsertNodeCommand.prototype.execute = function execute (params, context) {
      var this$1 = this;

      var state = params.commandState;
      if (state.disabled) { return }
      var editorSession = this._getEditorSession(params, context);
      editorSession.transaction(function (tx) {
        var nodeData = this$1.createNodeData(tx, params, context);
        var node = tx.insertBlockNode(nodeData);
        this$1.setSelection(tx, node);
      });
    };

    InsertNodeCommand.prototype.createNodeData = function createNodeData (tx, params, context) {
      var type = params.type;
      if (!type) { throw new Error("'type' is mandatory") }
      var doc = context.editorSession.getDocument();
      var nodeSchema = doc.getSchema().getNodeSchema(type);
      var nodeData = {type: type};
      forEach(nodeSchema, function (key) {
        if (params.hasOwnProperty(key)) {
          nodeData[key] = params[key];
        }
      });
      return nodeData
    };

    InsertNodeCommand.prototype.setSelection = function setSelection (tx, node) {
      if (node.isText()) {
        tx.selection = {
          type: 'property',
          path: node.getPath(),
          startOffset: 0
        };
      }
    };

    return InsertNodeCommand;
  }(Command));

  var MenuItem = (function (Component$$1) {
    function MenuItem () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) MenuItem.__proto__ = Component$$1;
    MenuItem.prototype = Object.create( Component$$1 && Component$$1.prototype );
    MenuItem.prototype.constructor = MenuItem;

    MenuItem.prototype.render = function render ($$) {
      var commandState = this.props.commandState;
      var el = $$('button')
        .addClass('sc-menu-item')
        .append(
          this._renderIcon($$),
          this._renderLabel($$),
          this._renderKeyboardShortcut($$)
        )
        .on('click', this._onClick);

      if (this.props.label) {
        el.append(this.renderLabel($$));
      }
      if (commandState.active) {
        el.addClass('sm-active');
      }
      if (commandState.disabled) {
        
        el.attr('tabindex', -1)
          .attr('disabled', true);
      } else {
        
        el.attr('tabindex', 1);
      }
      return el
    };

    MenuItem.prototype._renderLabel = function _renderLabel ($$) {
      return $$('div').addClass('se-label').append(
        this._getLabel()
      )
    };

    MenuItem.prototype._renderIcon = function _renderIcon ($$) {
      return $$('div').addClass('se-icon').append(
        this.context.iconProvider.renderIcon($$, this.props.name)
      )
    };

    MenuItem.prototype._renderKeyboardShortcut = function _renderKeyboardShortcut ($$) {
      return $$('div').addClass('se-keyboard-shortcut').append(
        this._getKeyboardShortcut()
      )
    };

    MenuItem.prototype._getLabel = function _getLabel () {
      var labelProvider = this.context.labelProvider;
      return labelProvider.getLabel(this.props.name, this.props.commandState)
    };

    MenuItem.prototype._getKeyboardShortcut = function _getKeyboardShortcut () {
      var name = this.props.name;
      var keyboardShortcuts = this.context.keyboardShortcuts;
      if (keyboardShortcuts[name]) {
        return keyboardShortcuts[name]
      } else {
        return ''
      }
    };

    MenuItem.prototype._onClick = function _onClick (e) {
      e.preventDefault();
      e.stopPropagation();
      if (!this.props.commandState.disabled) { this.executeCommand(); }
    };

    
    MenuItem.prototype.executeCommand = function executeCommand (props) {
      
      this.context.commandManager.executeCommand(this.props.name, props);
    };

    return MenuItem;
  }(Component));

  var ToolPanel = (function (Component$$1) {
    function ToolPanel () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) ToolPanel.__proto__ = Component$$1;
    ToolPanel.prototype = Object.create( Component$$1 && Component$$1.prototype );
    ToolPanel.prototype.constructor = ToolPanel;

    ToolPanel.prototype.didMount = function didMount () {
      this.context.editorSession.onRender(this._onCommandStatesChanged, this);
    };

    ToolPanel.prototype.dispose = function dispose () {
      this.context.editorSession.off(this);
    };

    ToolPanel.prototype.render = function render ($$) { 
      throw new Error('This method is abstract')
    };

    
    ToolPanel.prototype.getEntryTypeComponents = function getEntryTypeComponents () {
      return {
        'tool-group': this.getComponent('tool-group'),
        'tool-dropdown': this.getComponent('tool-dropdown'),
        'tool-prompt': this.getComponent('tool-prompt'),
        'tool-separator': this.getComponent('tool-separator')
      }
    };

    ToolPanel.prototype.renderEntries = function renderEntries ($$) {
      var this$1 = this;

      return this.props.toolPanel.map(function (entry) {
        var entryTypeComponents = this$1.getEntryTypeComponents();
        var ComponentClass = entryTypeComponents[entry.type];
        if (!ComponentClass) { throw new Error('Toolpanel entry type not found') }
        var props = Object.assign({}, entry, { theme: this$1.getTheme() });
        var el = $$(ComponentClass, props);
        if (entry.name) { el.ref(entry.name); }
        return el
      })
    };

    ToolPanel.prototype.hasEnabledTools = function hasEnabledTools () {
      var entriesContainer = this.refs.entriesContainer;
      var entries = entriesContainer.childNodes;
      var hasEnabledTools = false;
      entries.forEach(function (entry) {
        if (entry.hasEnabledTools()) {
          hasEnabledTools = true;
        }
      });
      return hasEnabledTools
    };

    ToolPanel.prototype.getActiveToolGroupNames = function getActiveToolGroupNames () {
      throw new Error('Abstract method')
    };

    ToolPanel.prototype.showDisabled = function showDisabled () {
      return false
    };

    
    ToolPanel.prototype.getToolStyle = function getToolStyle () {
      throw new Error('Abstract method')
    };

    ToolPanel.prototype.getTheme = function getTheme () {
      return this.props.theme || 'dark'
    };

    ToolPanel.prototype._onCommandStatesChanged = function _onCommandStatesChanged (editorSession) {
      if (editorSession.hasChanged('commandStates')) {
        this.rerender();
      }
    };

    return ToolPanel;
  }(Component));

  var Overlay = (function (ToolPanel$$1) {
    function Overlay () {
      ToolPanel$$1.apply(this, arguments);
    }

    if ( ToolPanel$$1 ) Overlay.__proto__ = ToolPanel$$1;
    Overlay.prototype = Object.create( ToolPanel$$1 && ToolPanel$$1.prototype );
    Overlay.prototype.constructor = Overlay;

    Overlay.prototype.didMount = function didMount () {
      ToolPanel$$1.prototype.didMount.call(this);
      if (!this.context.scrollPane) {
        throw new Error('Requires scrollPane context')
      }
      this.context.scrollPane.on('selection:positioned', this._onSelectionPositioned, this);
    };

    Overlay.prototype.dispose = function dispose () {
      ToolPanel$$1.prototype.dispose.call(this);
      this.context.scrollPane.off(this);
    };

    Overlay.prototype.render = function render ($$) {
      var el = $$('div').addClass('sc-overlay');
      el.addClass('sm-hidden');
      el.addClass('sm-theme-'+this.getTheme());
      el.append(
        $$('div').addClass('se-active-tools').append(
          this.renderEntries($$)
        ).ref('entriesContainer')
      );
      return el
    };

    Overlay.prototype.show = function show (hints) {
      this.el.removeClass('sm-hidden');
      this._position(hints);
    };

    Overlay.prototype.hide = function hide () {
      this.el.addClass('sm-hidden');
    };

    Overlay.prototype._onSelectionPositioned = function _onSelectionPositioned (hints) {
      if (this.hasEnabledTools()) {
        this.el.removeClass('sm-hidden');
        var overlayWidth = this.el.htmlProp('offsetWidth');
        var selRect = hints.selectionRect;
        var selectionMaxWidth = selRect.width;
        
        this.el.css('top', selRect.top + selRect.height);
        var leftPos = selRect.left + selectionMaxWidth/2 - overlayWidth/2;
        
        leftPos = Math.max(leftPos, 0);
        
        var maxLeftPos = selRect.left + selectionMaxWidth + selRect.right - overlayWidth;
        leftPos = Math.min(leftPos, maxLeftPos);
        this.el.css('left', leftPos);
      } else {
        this.el.addClass('sm-hidden');
      }
    };

    Overlay.prototype.getTheme = function getTheme () {
      return this.props.theme || 'dark'
    };

    return Overlay;
  }(ToolPanel));

  var ResponsiveApplication = (function (Component$$1) {
    function ResponsiveApplication() {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      Component$$1.apply(this, args);

      this.pages = {};

      this.handleActions({
        'navigate': this.navigate
      });
    }

    if ( Component$$1 ) ResponsiveApplication.__proto__ = Component$$1;
    ResponsiveApplication.prototype = Object.create( Component$$1 && Component$$1.prototype );
    ResponsiveApplication.prototype.constructor = ResponsiveApplication;

    ResponsiveApplication.prototype.getInitialState = function getInitialState () {
      return {
        route: undefined,
        mobile: this._isMobile()
      }
    };

    ResponsiveApplication.prototype.didMount = function didMount () {
      if (platform.inBrowser) {
        var _window = DefaultDOMElement.getBrowserWindow();
        _window.on('resize', this._onResize, this);
      }
      this.router = this.getRouter();
      this.router.on('route:changed', this._onRouteChanged, this);
      var route = this.router.readRoute();
      
      
      this.navigate(route, {replace: true});
    };

    ResponsiveApplication.prototype.dispose = function dispose () {
      this.router.off(this);
      this.router.dispose();
    };

    
    ResponsiveApplication.prototype.navigate = function navigate (route, opts) {
      this.extendState({
        route: route
      });
      this.router.writeRoute(route, opts);
    };

    ResponsiveApplication.prototype._onRouteChanged = function _onRouteChanged (route) {
      
      this.navigate(route, {replace: true});
    };

    ResponsiveApplication.prototype._isMobile = function _isMobile () {
      if (platform.inBrowser) {
        return window.innerWidth < 700
      }
    };

    ResponsiveApplication.prototype._onResize = function _onResize () {
      if (this._isMobile()) {
        
        if (!this.state.mobile) {
          this.extendState({
            mobile: true
          });
        }
      } else {
        if (this.state.mobile) {
          this.extendState({
            mobile: false
          });
        }
      }
    };

    ResponsiveApplication.prototype._getPage = function _getPage () {
      return this.state.route.page || this.getDefaultPage()
    };

    ResponsiveApplication.prototype._getPageClass = function _getPageClass () {
      var page = this._getPage();
      return this.pages[page]
    };

    ResponsiveApplication.prototype._getPageProps = function _getPageProps () {
      var props = cloneDeep(this.state.route);
      delete props.page;
      props.mobile = this.state.mobile;
      return props
    };

    ResponsiveApplication.prototype.addPage = function addPage (pageName, PageClass) {
      this.pages[pageName] = PageClass;
    };

    ResponsiveApplication.prototype.renderPage = function renderPage ($$) {
      var PageClass = this._getPageClass();
      var pageName = this._getPage();
      return $$(PageClass, this._getPageProps()).ref(pageName)
    };

    ResponsiveApplication.prototype.render = function render ($$) {
      var el = $$('div').addClass('sc-responsive-application');

      if (this.state.route === undefined) {
        
        return el
      }

      el.append(
        this.renderPage($$)
      );

      return el
    };

    return ResponsiveApplication;
  }(Component));

  var Router = (function (EventEmitter$$1) {
    function Router() {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      EventEmitter$$1.apply(this, args);
      this.__isStarted__ = false;
    }

    if ( EventEmitter$$1 ) Router.__proto__ = EventEmitter$$1;
    Router.prototype = Object.create( EventEmitter$$1 && EventEmitter$$1.prototype );
    Router.prototype.constructor = Router;

    
    Router.prototype.start = function start () {
      var window = DefaultDOMElement.getBrowserWindow();
      window.on('hashchange', this._onHashChange, this);
      this.__isStarted__ = true;
    };

    
    Router.prototype.readRoute = function readRoute () {
      if (!this.__isStarted__) { this.start(); }
      return this.parseRoute(this.getRouteString())
    };

    
    Router.prototype.writeRoute = function writeRoute (route, opts) {
      opts = opts || {};
      var routeString = this.stringifyRoute(route);
      if (!routeString) {
        this.clearRoute(opts);
      } else {
        this._writeRoute(routeString, opts);
      }
    };

    Router.prototype.dispose = function dispose () {
      var window = DefaultDOMElement.getBrowserWindow();
      window.off(this);
    };

    
    Router.prototype.parseRoute = function parseRoute (routeString) {
      return Router.routeStringToObject(routeString)
    };

    
    Router.prototype.stringifyRoute = function stringifyRoute (route) {
      return Router.objectToRouteString(route)
    };

    Router.prototype.getRouteString = function getRouteString () {
      return window.location.hash.slice(1)
    };

    Router.prototype._writeRoute = function _writeRoute (route, opts) {
      this.__isSaving__ = true;
      try {
        if (opts.replace) {
          window.history.replaceState({} , '', '#'+route);
        } else {
          window.history.pushState({} , '', '#'+route);
        }
      } finally {
        this.__isSaving__ = false;
      }
    };

    Router.prototype.clearRoute = function clearRoute (opts) {
      this._writeRoute('', opts);
    };

    Router.prototype._onHashChange = function _onHashChange () {
      
      if (this.__isSaving__) {
        return
      }
      if (this.__isLoading__) {
        console.error('FIXME: router is currently applying a route.');
        return
      }
      this.__isLoading__ = true;
      try {
        var routeString = this.getRouteString();
        var route = this.parseRoute(routeString);
        this.emit('route:changed', route);
      } finally {
        this.__isLoading__ = false;
      }
    };

    return Router;
  }(EventEmitter));

  Router.objectToRouteString = function(obj) {
    var route = [];
    forEach(obj, function(val, key) {
      route.push(key+'='+val);
    });
    return route.join(',')
  };

  Router.routeStringToObject = function(routeStr) {
    var obj = {};
    
    if (!routeStr) { return obj }
    var params = routeStr.split(',');
    params.forEach(function(param) {
      var tuple = param.split('=');
      if (tuple.length !== 2) {
        throw new Error('Illegal route.')
      }
      obj[tuple[0].trim()] = tuple[1].trim();
    });
    return obj
  };

  var SwitchTextTypeCommand = (function (Command$$1) {
    function SwitchTextTypeCommand(config) {
      Command$$1.call(this, config);
      if (!config.spec) {
        throw new Error("'config.spec' is mandatory")
      }
      if (!config.spec.type) {
        throw new Error("'config.spec.type' is mandatory")
      }
    }

    if ( Command$$1 ) SwitchTextTypeCommand.__proto__ = Command$$1;
    SwitchTextTypeCommand.prototype = Object.create( Command$$1 && Command$$1.prototype );
    SwitchTextTypeCommand.prototype.constructor = SwitchTextTypeCommand;

    SwitchTextTypeCommand.prototype.getType = function getType () {
      return this.config.spec.type
    };

    SwitchTextTypeCommand.prototype.getCommandState = function getCommandState (params) {
      var doc = params.editorSession.getDocument();
      var sel = params.selection;
      var isBlurred = params.editorSession.isBlurred();

      var commandState = {
        disabled: false
      };

      if (sel.isPropertySelection() && !isBlurred) {
        var path = sel.getPath();
        var node = doc.get(path[0]);
        if (node && node.isText() && node.isBlock()) {
          commandState.active = isMatch(node, this.config.spec);
          
          
          commandState.showInContext = sel.start.offset === 0 && sel.end.offset === 0;
        } else {
          commandState.disabled = true;
        }
      } else {
        
        commandState.disabled = true;
      }

      return commandState
    };

    
    SwitchTextTypeCommand.prototype.execute = function execute (params) {
      var this$1 = this;

      var surface = params.surface;
      var editorSession = params.editorSession;
      if (!surface) {
        console.warn('No focused surface. Stopping command execution.');
        return
      }
      editorSession.transaction(function (tx) {
        return tx.switchTextType(this$1.config.spec)
      });
    };

    SwitchTextTypeCommand.prototype.isSwitchTypeCommand = function isSwitchTypeCommand () {
      return true
    };

    return SwitchTextTypeCommand;
  }(Command));

  var TextBlockComponent = (function (NodeComponent$$1) {
    function TextBlockComponent () {
      NodeComponent$$1.apply(this, arguments);
    }

    if ( NodeComponent$$1 ) TextBlockComponent.__proto__ = NodeComponent$$1;
    TextBlockComponent.prototype = Object.create( NodeComponent$$1 && NodeComponent$$1.prototype );
    TextBlockComponent.prototype.constructor = TextBlockComponent;

    TextBlockComponent.prototype.didMount = function didMount () {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      NodeComponent$$1.prototype.didMount.apply(this, args);

      var node = this.props.node;
      this.context.editorSession.onRender('document', this.rerender, this, {
        path: [node.id, 'textAlign']
      });
    };

    TextBlockComponent.prototype.dispose = function dispose () {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      NodeComponent$$1.prototype.dispose.apply(this, args);
      this.context.editorSession.off(this);
    };

    TextBlockComponent.prototype.render = function render ($$) {
      var el = NodeComponent$$1.prototype.render.call(this, $$);
      el.addClass('sc-text-block');

      var node = this.props.node;
      
      
      if (node.direction) {
        
        el.attr('dir', node.direction);
      }
      if (node.textAlign) {
        el.addClass('sm-align-'+node.textAlign);
      }
      el.append($$(TextPropertyComponent, {
        placeholder: this.props.placeholder,
        path: node.getPath(),
        direction: node.direction
      }));
      return el
    };

    return TextBlockComponent;
  }(NodeComponent));

  var TextPropertyEditor = (function (Surface$$1) {
    function TextPropertyEditor(parent, props) {
      
      props.name = props.name || props.path.join('.');
      Surface$$1.call(this, parent, props);

      if (!props.path) {
        throw new Error("Property 'path' is mandatory.")
      }
    }

    if ( Surface$$1 ) TextPropertyEditor.__proto__ = Surface$$1;
    TextPropertyEditor.prototype = Object.create( Surface$$1 && Surface$$1.prototype );
    TextPropertyEditor.prototype.constructor = TextPropertyEditor;

    TextPropertyEditor.prototype.render = function render ($$) {
      var el = Surface$$1.prototype.render.apply(this, arguments);
      el.addClass("sc-text-property-editor");

      if (!this.props.disabled) {
        el.addClass('sm-enabled');
        el.attr('contenteditable', true);
        
        el.attr('spellcheck', this.props.spellcheck === 'native');
      }

      el.append(
        $$(TextPropertyComponent, {
          placeholder: this.props.placeholder,
          tagName: this.props.tagName || "div",
          path: this.props.path,
          markers: this.props.markers,
          withoutBreak: this.props.withoutBreak
        }).ref('property')
      );

      return el
    };

    TextPropertyEditor.prototype.selectFirst = function selectFirst () {
      this.editorSession.setSelection({
        type: "property",
        path: this.getPath(),
        startOffset: 0,
        surfaceId: this.id
      });
    };

    TextPropertyEditor.prototype.getPath = function getPath () {
      return this.props.path
    };

    TextPropertyEditor.prototype._handleEnterKey = function _handleEnterKey (event) {
      event.stopPropagation();
      if (this.props.handleEnter === false || !this.props.multiLine) {
        event.preventDefault();
      }
      this.el.emit('enter', {
        altKey: event.altKey,
        ctrlKey: event.ctrlKey,
        metaKey: event.metaKey,
        shiftKey: event.shiftKey,
        code: event.code
      });
    };

    

    TextPropertyEditor.prototype._handleEscapeKey = function _handleEscapeKey (event) {
      this.el.emit('escape', {
        altKey: event.altKey,
        ctrlKey: event.ctrlKey,
        metaKey: event.metaKey,
        shiftKey: event.shiftKey,
        code: event.code
      });
    };

    return TextPropertyEditor;
  }(Surface));

  TextPropertyEditor.prototype._isTextPropertyEditor = true;

  var Tooltip = (function (Component$$1) {
    function Tooltip () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) Tooltip.__proto__ = Component$$1;
    Tooltip.prototype = Object.create( Component$$1 && Component$$1.prototype );
    Tooltip.prototype.constructor = Tooltip;

    Tooltip.prototype.render = function render ($$) {
      var el = $$('div').addClass('sc-tooltip');
      el.append(this.props.text);
      return el
    };

    return Tooltip;
  }(Component));

  var ToggleTool = (function (Component$$1) {
    function ToggleTool () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) ToggleTool.__proto__ = Component$$1;
    ToggleTool.prototype = Object.create( Component$$1 && Component$$1.prototype );
    ToggleTool.prototype.constructor = ToggleTool;

    var prototypeAccessors = { _isTool: { configurable: true } };

    prototypeAccessors._isTool.get = function () {
      return true
    };

    
    ToggleTool.prototype.render = function render ($$) {
      var el = $$('div')
        .addClass('sc-toggle-tool');

      var customClassNames = this.getClassNames();
      if (customClassNames) {
        el.addClass(customClassNames);
      }

      el.append(
        this.renderButton($$)
      );

      
      el.append(
        $$(Tooltip, {
          text: this._getTooltipText()
        })
      );
      return el
    };

    ToggleTool.prototype.renderButton = function renderButton ($$) {
      var commandState = this.props.commandState;
      var Button = this.getComponent('button');
      var btn = $$(Button, {
        icon: this.getIconName(),
        active: commandState.active,
        disabled: commandState.disabled,
        theme: this.props.theme
      }).on('click', this.onClick);
      return btn
    };

    ToggleTool.prototype.getClassNames = function getClassNames () {
      return ''
    };

    ToggleTool.prototype.getCommandName = function getCommandName () {
      return this.getName()
    };

    ToggleTool.prototype.getName = function getName () {
      return this.props.name
    };

    ToggleTool.prototype.getIconName = function getIconName () {
      return this.props.name
    };

    ToggleTool.prototype.onClick = function onClick (e) {
      e.preventDefault();
      e.stopPropagation();
      if (!this.props.disabled) {
        this.executeCommand();
      }
    };

    ToggleTool.prototype._getTooltipText = function _getTooltipText () {
      var name = this.props.name;
      var label = this.context.labelProvider.getLabel(name);
      var keyboardShortcuts = this.context.keyboardShortcuts;
      if (keyboardShortcuts[name]) {
        return [label, ' (', keyboardShortcuts[name], ')'].join('')
      } else {
        return label
      }
    };

    
    ToggleTool.prototype.executeCommand = function executeCommand (props) {
      props = Object.assign({ mode: this.props.mode }, props);
      this.context.commandManager.executeCommand(this.getCommandName(), props);
    };

    Object.defineProperties( ToggleTool.prototype, prototypeAccessors );

    return ToggleTool;
  }(Component));

  var ToolGroup = (function (Component$$1) {
    function ToolGroup () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) ToolGroup.__proto__ = Component$$1;
    ToolGroup.prototype = Object.create( Component$$1 && Component$$1.prototype );
    ToolGroup.prototype.constructor = ToolGroup;

    ToolGroup.prototype.isToolEnabled = function isToolEnabled (commandName, commandState) {
      var enabled = true;
      if (this.props.contextual && !commandState.showInContext) {
        enabled = false;
      }
      if (commandState.disabled) {
        enabled = false;
      }
      return enabled
    };

    
    ToolGroup.prototype.hasEnabledTools = function hasEnabledTools (commandStates) {
      var this$1 = this;

      if (!commandStates) {
        commandStates = this._getCommandStates();
      }
      var hasEnabledTools;
      forEach(commandStates, function (commandState, commandName) {
        if (this$1.isToolEnabled(commandName, commandState)) {
          hasEnabledTools = true;
        }
      });
      return hasEnabledTools
    };

    ToolGroup.prototype.render = function render ($$) {
      var this$1 = this;

      var commandStates = this._getCommandStates();
      var el = $$('div').addClass(this._getClassNames());
      el.addClass('sm-'+this.props.name);
      forEach(commandStates, function (commandState, commandName) {
        if (this$1.isToolEnabled(commandName, commandState) || this$1.props.showDisabled) {
          var ToolClass = this$1._getToolClass(commandName);
          el.append(
            $$(ToolClass, {
              name: commandName,
              commandState: commandState,
              style: this$1.props.style,
              theme: this$1.props.theme
            }).ref(commandName)
          );
        }
      });
      return el
    };

    
    ToolGroup.prototype._getCommandStates = function _getCommandStates () {
      var commandStates = this.context.editorSession.getCommandStates();
      var commandGroups = this.context.commandGroups;
      var filteredCommandStates = {}; 
      this.props.commandGroups.forEach(function (commandGroup) {
        
        
        if (!commandGroups[commandGroup]) {
          
          return
        }
        commandGroups[commandGroup].forEach(function (commandName) {
          
          
          
          
          
          if (commandStates[commandName]) {
            filteredCommandStates[commandName] = commandStates[commandName];
          }
        });
      });
      return filteredCommandStates
    };

    ToolGroup.prototype._getClassNames = function _getClassNames () {
      return 'sc-tool-group'
    };

    ToolGroup.prototype._getToolClass = function _getToolClass (commandName) {
      var tools = this.context.tools;
      var DefaultToolClass;
      if (this.props.style === 'descriptive') {
        DefaultToolClass = MenuItem;
      } else {
        DefaultToolClass = ToggleTool;
      }
      return tools[commandName] || DefaultToolClass
    };

    return ToolGroup;
  }(Component));

  var Menu = (function (Component$$1) {
    function Menu () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) Menu.__proto__ = Component$$1;
    Menu.prototype = Object.create( Component$$1 && Component$$1.prototype );
    Menu.prototype.constructor = Menu;

    Menu.prototype.render = function render ($$) {
      var commandState = this.props.commandStates;
      var el = $$('div').addClass('sc-menu');
      this.props.items.forEach(function (item) {
        if (item.command) {
          el.append(
            $$(MenuItem, {
              name: item.command,
              commandState: commandState[item.command]
            })
          );
        } else if (item.type === 'separator') {
          el.append(
            $$('div').addClass('separator')
          );
        }
      });
      return el
    };

    return Menu;
  }(Component));

  var ToolDropdown = (function (ToolGroup$$1) {
    function ToolDropdown () {
      ToolGroup$$1.apply(this, arguments);
    }

    if ( ToolGroup$$1 ) ToolDropdown.__proto__ = ToolGroup$$1;
    ToolDropdown.prototype = Object.create( ToolGroup$$1 && ToolGroup$$1.prototype );
    ToolDropdown.prototype.constructor = ToolDropdown;

    ToolDropdown.prototype.willReceiveProps = function willReceiveProps () {
      this.setState({showChoices: false});
    };

    ToolDropdown.prototype.render = function render ($$) {
      var Button = this.getComponent('button');
      var commandStates = this._getCommandStates();
      var el = $$('div').addClass('sc-tool-dropdown');
      el.addClass('sm-'+this.props.name);

      
      var toggleName = this._getToggleName(commandStates) || this.props.name;

      
      
      if (this.props.showDisabled || this.hasEnabledTools(commandStates)) {
        var toggleButton;
        if (this.props.style === 'minimal') {
          toggleButton = $$(Button, {
            icon: toggleName,
            dropdown: true,
            active: this.state.showChoices,
            theme: this.props.theme
          }).on('click', this._toggleChoices);
        } else if (this.props.style === 'descriptive') {
          toggleButton = $$(Button, {
            label: toggleName,
            
            commandState: commandStates[toggleName],
            dropdown: true,
            active: this.state.showChoices,
            theme: this.props.theme
          }).on('click', this._toggleChoices);
        } else {
          throw new Error('Style '+this.props.style+' not supported')
        }
        el.append(toggleButton);

        if (this.state.showChoices) {
          el.append(
            $$('div').addClass('se-choices').append(
              $$(Menu, {
                commandStates: commandStates,
                items: this._getMenuItems(commandStates)
              })
            )
          );
        } else if (this.props.style === 'minimal' || toggleName !== this.props.name) {
          
          el.append(
            this._renderToolTip($$)
          );
        }
      }
      return el
    };

    ToolDropdown.prototype._renderToolTip = function _renderToolTip ($$) {
      var labelProvider = this.context.labelProvider;
      return $$(Tooltip, {
        text: labelProvider.getLabel(this.props.name)
      })
    };

    
    ToolDropdown.prototype._getToggleName = function _getToggleName (commandStates) {
      return this._getActiveCommandName(commandStates)
    };

    
    ToolDropdown.prototype._getMenuItems = function _getMenuItems (commandStates) {
      var this$1 = this;

      var showDisabled = this.props.showDisabled;
      var menuItems = [];
      forEach(commandStates, function (commandState, commandName) {
        
        
        if (showDisabled || this$1.isToolEnabled(commandName, commandState)) {
          menuItems.push({
            command: commandName
          });
        }
      });
      return menuItems
    };

    ToolDropdown.prototype._getActiveCommandName = function _getActiveCommandName (commandStates) {
      var activeCommand;

      forEach(commandStates, function (commandState, commandName) {
        if (commandState.active && !activeCommand) {
          activeCommand = commandName;
        }
      });
      return activeCommand
    };

    ToolDropdown.prototype._toggleChoices = function _toggleChoices () {
      this.setState({
        showChoices: !(this.state.showChoices)
      });
    };

    return ToolDropdown;
  }(ToolGroup));

  var MenuGroup = (function (ToolGroup$$1) {
    function MenuGroup () {
      ToolGroup$$1.apply(this, arguments);
    }

    if ( ToolGroup$$1 ) MenuGroup.__proto__ = ToolGroup$$1;
    MenuGroup.prototype = Object.create( ToolGroup$$1 && ToolGroup$$1.prototype );
    MenuGroup.prototype.constructor = MenuGroup;

    MenuGroup.prototype._getToolClass = function _getToolClass (commandName) {
      var tools = this.context.tools;
      return tools[commandName] || MenuItem
    };

    MenuGroup.prototype._getClassNames = function _getClassNames () {
      return 'sc-menu-group'
    };

    return MenuGroup;
  }(ToolGroup));

  var ToolPrompt = (function (ToolGroup$$1) {
    function ToolPrompt () {
      ToolGroup$$1.apply(this, arguments);
    }

    if ( ToolGroup$$1 ) ToolPrompt.__proto__ = ToolGroup$$1;
    ToolPrompt.prototype = Object.create( ToolGroup$$1 && ToolGroup$$1.prototype );
    ToolPrompt.prototype.constructor = ToolPrompt;

    ToolPrompt.prototype._getClassNames = function _getClassNames () {
      return 'sc-tool-prompt'
    };

    return ToolPrompt;
  }(ToolGroup));

  var WorkflowPane = (function (ToolPanel$$1) {
    function WorkflowPane () {
      ToolPanel$$1.apply(this, arguments);
    }

    if ( ToolPanel$$1 ) WorkflowPane.__proto__ = ToolPanel$$1;
    WorkflowPane.prototype = Object.create( ToolPanel$$1 && ToolPanel$$1.prototype );
    WorkflowPane.prototype.constructor = WorkflowPane;

    WorkflowPane.prototype.render = function render ($$) {
      var el = $$('div').addClass('sc-workflow-pane');
      el.append(
        $$('div').addClass('se-active-tools').append(
          this.renderEntries($$)
        ).ref('entriesContainer')
      );
      return el
    };

    WorkflowPane.prototype.getTheme = function getTheme () {
      return this.props.theme || 'light'
    };

    return WorkflowPane;
  }(ToolPanel));

  var Button = (function (Component$$1) {
    function Button () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) Button.__proto__ = Component$$1;
    Button.prototype = Object.create( Component$$1 && Component$$1.prototype );
    Button.prototype.constructor = Button;

    Button.prototype.render = function render ($$) {
      var el = $$('button')
        .addClass('sc-button');

      if (this.props.icon) {
        el.append(this.renderIcon($$));
      }
      if (this.props.label) {
        el.append(this.renderLabel($$));
      }

      if (this.props.dropdown) {
        el.append(this.renderDropdownIcon($$));
      }

      if (this.props.active) {
        el.addClass('sm-active');
      }
      if (this.props.theme) {
        el.addClass('sm-theme-'+this.props.theme);
      }

      if (this.props.disabled) {
        
        el.attr('tabindex', -1)
          .attr('disabled', true);
      } else {
        
        el.attr('tabindex', 1);
      }

      
      el.append(this.props.children);
      return el
    };

    Button.prototype.renderIcon = function renderIcon ($$) {
      var iconEl = this.context.iconProvider.renderIcon($$, this.props.icon);
      return iconEl
    };

    Button.prototype.renderDropdownIcon = function renderDropdownIcon ($$) {
      var iconEl = this.context.iconProvider.renderIcon($$, 'dropdown');
      iconEl.addClass('se-dropdown');
      return iconEl
    };

    Button.prototype.renderLabel = function renderLabel ($$) {
      return $$('span').addClass('se-label').append(
        this.getLabel(this.props.label)
      )
    };

    Button.prototype.getLabel = function getLabel (name) {
      var labelProvider = this.context.labelProvider;
      return labelProvider.getLabel(name, this.props.commandState)
    };

    return Button;
  }(Component));

  var Layout = (function (Component$$1) {
    function Layout () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) Layout.__proto__ = Component$$1;
    Layout.prototype = Object.create( Component$$1 && Component$$1.prototype );
    Layout.prototype.constructor = Layout;

    Layout.prototype.render = function render ($$) {
      var el = $$('div').addClass('sc-layout');
      el.addClass('sm-width-'+this.props.width);
      if (this.props.textAlign) {
        el.addClass('sm-text-align-'+this.props.textAlign);
      }

      if (this.props.noPadding) {
        el.addClass('sm-no-padding');
      }

      el.append(this.props.children);
      return el
    };

    return Layout;
  }(Component));

  function getRelativeBoundingRect(els, containerEl) {
    var nativeCotainerEl;
    if (containerEl._isDOMElement) {
      nativeCotainerEl = containerEl.getNativeElement();
    } else {
      nativeCotainerEl = containerEl;
    }
    if (!isArray(els)) { els = [els]; }
    var elRects = els.map(function (el) {
      var nativeEl;
      if (el._isDOMElement) {
        nativeEl = el.getNativeElement();
      } else {
        nativeEl = el;
      }
      return _getBoundingOffsetsRect(nativeEl, nativeCotainerEl)
    });

    var elsRect = _getBoundingRect(elRects);
    var containerElRect = nativeCotainerEl.getBoundingClientRect();
    return {
      left: elsRect.left,
      top: elsRect.top,
      right: containerElRect.width - elsRect.left - elsRect.width,
      bottom: containerElRect.height - elsRect.top - elsRect.height,
      width: elsRect.width,
      height: elsRect.height
    }
  }


  function _getBoundingRect(rects) {
    var bounds = {
      left: Number.POSITIVE_INFINITY,
      top: Number.POSITIVE_INFINITY,
      right: Number.NEGATIVE_INFINITY,
      bottom: Number.NEGATIVE_INFINITY,
      width: Number.NaN,
      height: Number.NaN
    };

    forEach(rects, function(rect) {
      if (rect.left < bounds.left) {
        bounds.left = rect.left;
      }
      if (rect.top < bounds.top) {
        bounds.top = rect.top;
      }
      if (rect.left + rect.width > bounds.right) {
        bounds.right = rect.left + rect.width;
      }
      if (rect.top + rect.height > bounds.bottom) {
        bounds.bottom = rect.top + rect.height;
      }
    });
    bounds.width = bounds.right - bounds.left;
    bounds.height = bounds.bottom - bounds.top;
    return bounds
  }


  function _getBoundingOffsetsRect(el, relativeParentEl) {
    var relativeParentElRect = relativeParentEl.getBoundingClientRect();
    var elRect = _getBoundingRect(el.getClientRects());

    var left = elRect.left - relativeParentElRect.left;
    var top = elRect.top - relativeParentElRect.top;
    return {
      left: left,
      top: top,
      right: relativeParentElRect.width - left - elRect.width,
      bottom: relativeParentElRect.height - top - elRect.height,
      width: elRect.width,
      height: elRect.height
    }
  }

  var Scrollbar = (function (Component$$1) {
    function Scrollbar () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) Scrollbar.__proto__ = Component$$1;
    Scrollbar.prototype = Object.create( Component$$1 && Component$$1.prototype );
    Scrollbar.prototype.constructor = Scrollbar;

    Scrollbar.prototype.didMount = function didMount () {
      
      DefaultDOMElement.getBrowserWindow().on('resize', this.onResize, this);
      
      this.props.scrollPane.on('scroll', this.onScroll, this);
      
      setTimeout(function() {
        this.updatePositions();
      }.bind(this));
    };

    Scrollbar.prototype.dispose = function dispose () {
      DefaultDOMElement.getBrowserWindow().off(this);
      this.props.scrollPane.off(this);
    };

    Scrollbar.prototype.didUpdate = function didUpdate () {
      this.updatePositions();
    };

    Scrollbar.prototype.render = function render ($$) {
      var el = $$('div')
        .addClass('sc-scrollbar')
        .on('mousedown', this.onMouseDown);

      if (this.props.highlights) {
        var highlightEls = [];

        forEach(this.props.highlights, function(highlights, scope) {
          forEach(highlights, function(h) {
            highlightEls.push(
              $$('div').ref(h).addClass('se-highlight sm-'+scope)
            );
          });
        });

        el.append(
          $$('div').ref('highlights')
            .addClass('se-highlights')
            .append(highlightEls)
        );
      }
      el.append($$('div').ref('thumb').addClass('se-thumb'));
      return el
    };

    Scrollbar.prototype.updatePositions = function updatePositions () {
      var scrollPane = this.props.scrollPane;
      var scrollableEl = scrollPane.getScrollableElement();
      var contentHeight = scrollPane.getContentHeight();
      var scrollPaneHeight = scrollPane.getHeight();
      var scrollTop = scrollPane.getScrollPosition();
      var contentEl = scrollPane.getContentElement();

      
      this.factor = (contentHeight / scrollPaneHeight);

      if (this.factor <= 1) {
        this.el.addClass('sm-hide-thumb');
      } else {
        this.el.removeClass('sm-hide-thumb');
      }

      this.refs.thumb.css({
        top: scrollTop / this.factor,
        height: scrollPaneHeight / this.factor
      });

      
      if (this.props.highlights) {
        
        forEach(this.props.highlights,function(highlights) {
          forEach(highlights, function(nodeId) {
            var nodeEl = scrollableEl.find('*[data-id="'+nodeId+'"]');

            if (!nodeEl) { return }

            
            var rect = getRelativeBoundingRect(nodeEl, contentEl);
            var top = rect.top / this.factor;
            var height = rect.height / this.factor;

            
            if (height < Scrollbar.overlayMinHeight) {
              height = Scrollbar.overlayMinHeight;
            }

            var highlightEl = this.refs[nodeId];
            if (highlightEl) {
              this.refs[nodeId].css({
                top: top,
                height: height
              });
            } else {
              console.warn('no ref found for highlight', nodeId);
            }
          }.bind(this));
        }.bind(this));
      }
    };

    Scrollbar.prototype.getScrollableElement = function getScrollableElement () {
      return this.props.scrollPane.getScrollableElement()
    };

    Scrollbar.prototype.onResize = function onResize () {
      this.rerender();
    };

    Scrollbar.prototype.onScroll = function onScroll () {
      this.updatePositions();
    };

    Scrollbar.prototype.onMouseDown = function onMouseDown (e) {
      e.stopPropagation();
      e.preventDefault();
      this._mouseDown = true;

      
      
      var _window = DefaultDOMElement.getBrowserWindow();
      _window.on('mousemove', this.onMouseMove, this);
      _window.on('mouseup', this.onMouseUp, this);

      var scrollBarOffset = this.el.getOffset().top;
      var y = e.pageY - scrollBarOffset;
      var thumbEl = this.refs.thumb.el;
      if (e.target !== thumbEl.getNativeElement()) {
        
        this.offset = thumbEl.height / 2;
        this.onMouseMove(e);
      } else {
        this.offset = y - thumbEl.getPosition().top;
      }
    };

    
    Scrollbar.prototype.onMouseUp = function onMouseUp () {
      this._mouseDown = false;
      var _window = DefaultDOMElement.getBrowserWindow();
      _window.off('mousemove', this.onMouseMove, this);
      _window.off('mouseup', this.onMouseUp, this);
    };

    Scrollbar.prototype.onMouseMove = function onMouseMove (e) {
      if (this._mouseDown) {
        var scrollPane = this.props.scrollPane;
        var scrollableEl = scrollPane.getScrollableElement();
        var scrollBarOffset = this.el.getOffset().top;
        var y = e.pageY - scrollBarOffset;

        
        var scroll = (y-this.offset)*this.factor;
        scrollableEl.setProperty('scrollTop', scroll);
      }
    };

    return Scrollbar;
  }(Component));

  Scrollbar.overlayMinHeight = 2;

  var ScrollPane = (function (AbstractScrollPane$$1) {
    function ScrollPane () {
      AbstractScrollPane$$1.apply(this, arguments);
    }

    if ( AbstractScrollPane$$1 ) ScrollPane.__proto__ = AbstractScrollPane$$1;
    ScrollPane.prototype = Object.create( AbstractScrollPane$$1 && AbstractScrollPane$$1.prototype );
    ScrollPane.prototype.constructor = ScrollPane;

    ScrollPane.prototype.didMount = function didMount () {
      AbstractScrollPane$$1.prototype.didMount.call(this);
      if (this.refs.scrollbar && this.props.highlights) {
        this.props.highlights.on('highlights:updated', this.onHighlightsUpdated, this);
      }
      if (this.refs.scrollbar) {
        if (platform.inBrowser) {
          this.domObserver = new window.MutationObserver(this._onContentChanged.bind(this));
          this.domObserver.observe(this.el.getNativeElement(), {
            subtree: true,
            attributes: true,
            characterData: true,
            childList: true,
          });
        }
        this.context.editorSession.onPosition(this._onPosition, this);
      }
    };

    ScrollPane.prototype.dispose = function dispose () {
      AbstractScrollPane$$1.prototype.dispose.call(this);
      if (this.props.highlights) {
        this.props.highlights.off(this);
      }
      this.context.editorSession.off(this);
      this.context.dragManager.off(this);
      if (this.domObserver) {
        this.domObserver.disconnect();
      }
    };

    ScrollPane.prototype.render = function render ($$) {
      var el = $$('div')
        .addClass('sc-scroll-pane');

      if (platform.isFF) {
        el.addClass('sm-firefox');
      }

      
      
      if (!this.props.noStyle) {
        el.addClass('sm-default-style');
      }

      
      if (this.props.scrollbarType === 'substance') {
        el.addClass('sm-substance-scrollbar');
        el.addClass('sm-scrollbar-position-' + this.props.scrollbarPosition);

        el.append(
          
          
          $$(Scrollbar, {
            scrollPane: this
          }).ref('scrollbar')
            .attr('id', 'content-scrollbar')
        );

        
        el.append(
          $$('div').ref("scanline").addClass('se-scanline')
        );
      }

      el.append(
        $$('div').ref('scrollable').addClass('se-scrollable').append(
          this.renderContent($$)
        ).on('scroll', this.onScroll)
      );
      return el
    };

    ScrollPane.prototype.renderContent = function renderContent ($$) {
      var contentEl = $$('div').ref('content').addClass('se-content');
      contentEl.append(this.props.children);
      if (this.props.contextMenu === 'custom') {
        contentEl.on('contextmenu', this._onContextMenu);
      }
      return contentEl
    };

    ScrollPane.prototype._onContentChanged = function _onContentChanged () {
      this._contentChanged = true;
    };

    ScrollPane.prototype._onPosition = function _onPosition () {
      if (this.refs.scrollbar && this._contentChanged) {
        this._contentChanged = false;
        this._updateScrollbar();
      }
    };

    ScrollPane.prototype._updateScrollbar = function _updateScrollbar () {
      if (this.refs.scrollbar) {
        this.refs.scrollbar.updatePositions();
      }
    };

    ScrollPane.prototype.onHighlightsUpdated = function onHighlightsUpdated (highlights) {
      this.refs.scrollbar.extendProps({
        highlights: highlights
      });
    };

    ScrollPane.prototype.onScroll = function onScroll () {
      var scrollPos = this.getScrollPosition();
      var scrollable = this.refs.scrollable;
      if (this.props.onScroll) {
        this.props.onScroll(scrollPos, scrollable);
      }
      
      if (this.props.tocProvider) {
        this.props.tocProvider.markActiveEntry(this);
      }
      this.emit('scroll', scrollPos, scrollable);
    };

    
    ScrollPane.prototype.getHeight = function getHeight () {
      var scrollableEl = this.getScrollableElement();
      return scrollableEl.height
    };

    
    ScrollPane.prototype.getContentHeight = function getContentHeight () {
      var contentEl = this.refs.content.el.getNativeElement();
      
      
      
      return contentEl.scrollHeight
    };

    
    ScrollPane.prototype.getContentElement = function getContentElement () {
      return this.refs.content.el
    };

    
    ScrollPane.prototype.getScrollableElement = function getScrollableElement () {
      return this.refs.scrollable.el
    };

    
    ScrollPane.prototype.getScrollPosition = function getScrollPosition () {
      var scrollableEl = this.getScrollableElement();
      return scrollableEl.getProperty('scrollTop')
    };

    ScrollPane.prototype.setScrollPosition = function setScrollPosition (scrollPos) {
      var scrollableEl = this.getScrollableElement();
      scrollableEl.setProperty('scrollTop', scrollPos);
    };

    
    ScrollPane.prototype.getPanelOffsetForElement = function getPanelOffsetForElement (el) {
      var contentContainerEl = this.refs.content.el;
      var rect = getRelativeBoundingRect(el, contentContainerEl);
      return rect.top
    };

    
    ScrollPane.prototype.scrollTo = function scrollTo (selector, onlyIfNotVisible) {
      var scrollableEl = this.getScrollableElement();
      var targetNode = scrollableEl.find(selector);
      if (targetNode) {
        var offset = this.getPanelOffsetForElement(targetNode);
        var shouldScroll = true;
        if (onlyIfNotVisible) {
          var height = scrollableEl.height;
          var oldOffset = scrollableEl.getProperty('scrollTop');
          shouldScroll = (offset < oldOffset || oldOffset+height<offset);
        }
        if (shouldScroll) {
          this.setScrollPosition(offset);
        }
      } else {
        console.warn(("No match found for selector '" + selector + "' in scrollable container"));
      }
    };

    
    ScrollPane.prototype.onSelectionPositioned = function onSelectionPositioned () {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      AbstractScrollPane$$1.prototype.onSelectionPositioned.apply(this, args);
      this._updateScrollbar();
    };

    ScrollPane.prototype._onContextMenu = function _onContextMenu (e) {
      AbstractScrollPane$$1.prototype._onContextMenu.call(this, e);
      this._updateScrollbar();
    };

    return ScrollPane;
  }(AbstractScrollPane));

  var SplitPane = (function (Component$$1) {
    function SplitPane () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) SplitPane.__proto__ = Component$$1;
    SplitPane.prototype = Object.create( Component$$1 && Component$$1.prototype );
    SplitPane.prototype.constructor = SplitPane;

    SplitPane.prototype.render = function render ($$) {
      if (this.props.children.length !== 2) {
        throw new Error('SplitPane only works with exactly two child elements')
      }

      var el = $$('div').addClass('sc-split-pane');
      if (this.props.splitType === 'horizontal') {
        el.addClass('sm-horizontal');
      } else {
        el.addClass('sm-vertical');
      }

      var paneA = this.props.children[0];
      var paneB = this.props.children[1];

      
      if (this.props.sizeB) {
        paneB.addClass('se-pane sm-sized');
        paneB.css(this.getSizedStyle(this.props.sizeB));
        paneA.addClass('se-pane sm-auto-fill');
      } else {
        paneA.addClass('se-pane sm-sized');
        paneA.css(this.getSizedStyle(this.props.sizeA));
        paneB.addClass('se-pane sm-auto-fill');
      }

      el.append(
        paneA,
        paneB
      );
      return el
    };

    
    SplitPane.prototype.getSizedStyle = function getSizedStyle (size) {
      if (!size || size === 'inherit') { return {} }
      if (this.props.splitType === 'horizontal') {
        return {'height': size}
      } else {
        return {'width': size}
      }
    };

    return SplitPane;
  }(Component));

  var Toolbar = (function (ToolPanel$$1) {
    function Toolbar () {
      ToolPanel$$1.apply(this, arguments);
    }

    if ( ToolPanel$$1 ) Toolbar.__proto__ = ToolPanel$$1;
    Toolbar.prototype = Object.create( ToolPanel$$1 && ToolPanel$$1.prototype );
    Toolbar.prototype.constructor = Toolbar;

    Toolbar.prototype.render = function render ($$) {
      var el = $$('div').addClass('sc-toolbar');
      el.append(
        $$('div').addClass('se-active-tools').append(
          this.renderEntries($$)
        ).ref('entriesContainer')
      );
      return el
    };

    Toolbar.prototype.getTheme = function getTheme () {
      return this.props.theme || 'light'
    };

    return Toolbar;
  }(ToolPanel));

  function series(tasks, cb, i) {
    i = i || 0;
    tasks[i](function(err) {
      
      if (err) { return cb(err) }
      if (i === tasks.length-1) {
        cb.apply(void 0, arguments); 
      } else {
        series(tasks, cb, i + 1);
      }
    });
  }

  var async = /*#__PURE__*/Object.freeze({
    series: series
  });

  function array2table(keys) {
    return keys.reduce(function (obj, key) {
      obj[key] = true;
      return obj
    }, {})
  }

  var Factory = (function (Registry$$1) {
    function Factory () {
      Registry$$1.apply(this, arguments);
    }

    if ( Registry$$1 ) Factory.__proto__ = Registry$$1;
    Factory.prototype = Object.create( Registry$$1 && Registry$$1.prototype );
    Factory.prototype.constructor = Factory;

    Factory.prototype.create = function create (name) {
      var clazz = this.get(name);
      if (!clazz) {
        throw new Error( 'No class registered by that name: ' + name )
      }
      
      var args = Array.prototype.slice.call( arguments, 1 );
      var obj = Object.create( clazz.prototype );
      clazz.apply( obj, args );
      return obj
    };

    return Factory;
  }(Registry));

  function getRangeFromMatrix(cells, startRow, startCol, endRow, endCol, force2D) {
    if (!force2D) {
      if (startRow === endRow && startCol === endCol) {
        var row = cells[startCol];
        if (row) { return row[endCol] }
        else { return undefined }
      }
      if (startRow === endRow) {
        var row$1 = cells[startRow];
        if (row$1) { return row$1.slice(startCol, endCol+1) }
        else { return [] }
      }
      if (startCol === endCol) {
        var res$1 = [];
        for (var i$1 = startRow; i$1 <= endRow; i$1++) {
          var row$2 = cells[i$1];
          if (row$2) { res$1.push(row$2[startCol]); }
        }
        return res$1
      }
    }
    var res = [];
    for (var i = startRow; i < endRow+1; i++) {
      var row$3 = cells[i];
      if (row$3) { res.push(row$3.slice(startCol, endCol+1)); }
    }
    return res
  }

  function includes(arr, val) {
    if (!arr) { return false }
    return (arr.indexOf(val) >= 0)
  }

  function merge$1(a, b, options) {
    options = options || {};
    var _with = null;
    if (options.array === 'replace') {
      _with = _replaceArrays;
    } else if (options.array === 'concat') {
      _with = _concatArrays;
    }
    if (_with) {
      return mergeWith(a, b, _with)
    } else {
      return merge(a, b)
    }
  }

  function _concatArrays(objValue, srcValue) {
    if (isArray(objValue)) {
      return objValue.concat(srcValue)
    } else {
      return null
    }
  }

  function _replaceArrays(objValue, srcValue) {
    if (isArray(objValue)) {
      return srcValue
    } else {
      return null
    }
  }

  function percentage(ratio) {
    return String(Math.floor(ratio*100*100)/100) + ' %'
  }

  function pluck(collection, prop) {
    return map(collection, function(item) { return item[prop] })
  }

  function printStacktrace() {
    try {
      throw new Error();
    } catch (err) {
      console.error(err.stack);
    }
  }

  function renderListNode(listNode, $$) {
    var items = listNode.getItems();
    var stack = [$$(_getTagName(1))];
    for (var i = 0; i < items.length; i++) {
      var item = items[i];
      
      var level = item.getLevel();
      if (level < stack.length) {
        for (var j = stack.length; j > level; j--) {
          stack.pop();
        }
      } else if (level > stack.length) {
        for (var j$1 = stack.length; j$1 < level; j$1++) {
          var list = stack[j$1-1];
          var childCount = list.getChildCount();
          var item$1 = (void 0);
          if (childCount === 0) {
            item$1 = $$('li');
            list.append(item$1);
          } else {
            item$1 = list.getChildAt(childCount-1);
          }
          var sublist = $$(_getTagName(j$1+1));
          item$1.append(sublist);
          stack.push(sublist);
        }
      }
      console.assert(level === stack.length, 'item.level should now be the same as stack.length');
      
      last$1(stack).append(
        $$(item)
      );
    }
    for(var j$2=stack.length; j$2>1;j$2--) {
      stack.pop();
    }

    return stack[0]

    function _getTagName(level) {
      var listType = listNode.getListType(level);
      return listType === 'order' ? 'ol' : 'ul'
    }
  }

  function times$2(num, fn, ctx) {
    for (var i=0; i<num; i++) {
      fn.call(ctx);
    }
  }

  function checkSchema(xmlSchema) {
    var tagNames = xmlSchema.getTagNames();
    var issues = [];
    for (var i = 0; i < tagNames.length; i++) {
      var name = tagNames[i];
      var elementSchema = xmlSchema.getElementSchema(name);
      switch(elementSchema.type) {
        case 'text': {
          issues = issues.concat(_checkTextElement(elementSchema));
          break
        }
        case 'element':
        case 'container': {
          issues = issues.concat(_checkElement(elementSchema));
          break
        }
        case 'hybrid': {
          issues = issues.concat(_checkHybridElement(elementSchema));
          break
        }
        case 'annotation': {
          issues = issues.concat(_checkAnnotation(elementSchema));
          break
        }
        case 'inline-element': {
          issues = issues.concat(_checkInlineElement(elementSchema));
          break
        }
        case 'anchor': {
          issues = issues.concat(_checkAnchor(elementSchema));
          break
        }
        case 'external': {
          break
        }
        default:
          throw new Error('Unsupported element classification.')
      }
    }
    return issues
  }

  function _checkTextElement(elementSchema) {
    var issues = [];
    
    if (_usedInline(elementSchema)) {
      issues.push(("[1]: text element <" + (elementSchema.name) + "> is used inline by " + (_usedInlineBy(elementSchema).join(','))));
    }
    
    if (!elementSchema.isText) {
      issues.push(("[4]: text element <" + (elementSchema.name) + "> does not allow text content"));
    }
    
    if (elementSchema.isStructured) {
      issues.push(("[5]: text element <" + (elementSchema.name) + "> must not allow structured content"));
    }
    return issues
  }

  function _checkElement(elementSchema) {
    var issues = [];
    if (_usedInline(elementSchema)) {
      issues.push(("[2]: element <" + (elementSchema.name) + "> is used inline by " + (_usedInlineBy(elementSchema).join(','))));
    }
    
    if (elementSchema.isText) {
      issues.push(("[6]: element <" + (elementSchema.name) + "> must not allow text content"));
    }
    return issues
  }

  function _checkHybridElement(elementSchema) {
    var issues = [];
    if (_usedInlineOnly(elementSchema)) {
      issues.push(("[2.2]: hybrid element <" + (elementSchema.name) + "> is used inline by " + (_usedInlineBy(elementSchema).join(','))));
    }
    
    if (!elementSchema.isText || !elementSchema.isStructured) {
      issues.push(("[7]: hybrid element <" + (elementSchema.name) + "> must allow text and structured content"));
    }
    return issues
  }

  function _checkAnnotation(elementSchema) {
    var issues = [];
    
    if (!_usedInlineOnly(elementSchema)) {
      issues.push(("[3]: annotation <" + (elementSchema.name) + "> is used in structured content by " + (_usedStructuredBy(elementSchema).join(','))));
    }
    if (!elementSchema.isText) {
      issues.push(("[8]: annotation <" + (elementSchema.name) + "> does not allow text content"));
    }
    
    
    
    
    
    
    return issues
  }

  function _checkInlineElement(elementSchema) {
    var issues = [];
    if (!_usedInlineOnly(elementSchema)) {
      issues.push(("[3]: inline-element <" + (elementSchema.name) + "> is used in structured content by " + (_usedStructuredBy(elementSchema).join(','))));
    }
    return issues
  }

  function _checkAnchor(elementSchema) {
    var issues = [];
    if (!_usedInlineOnly(elementSchema)) {
      issues.push(("[3]: anchor <" + (elementSchema.name) + "> is used in structured content by " + (_usedStructuredBy(elementSchema).join(','))));
    }
    return issues
  }

  function _usedInline(elementSchema) {
    var usedInlineBy = elementSchema.usedInlineBy || {};
    return Object.keys(usedInlineBy).length > 0
  }

  function _usedInlineBy(elementSchema) {
    var usedInlineBy = elementSchema.usedInlineBy || {};
    return Object.keys(usedInlineBy)
  }


  function _usedInlineOnly(elementSchema) {
    var usedStructuredBy = _usedStructuredBy(elementSchema);
    return (usedStructuredBy.length === 0)
  }

  function _usedStructuredBy(elementSchema) {
    var usedInlineBy = elementSchema.usedInlineBy || {};
    var parents = Object.keys(elementSchema.parents);
    var usedStructuredBy = parents.filter(function (name) {
      return !(usedInlineBy[name])
    });
    return usedStructuredBy
  }

  function _lookupRNG(fs, searchDirs, file) {
    for (var i = 0; i < searchDirs.length; i++) {
      var absPath = searchDirs[i] + '/' + file;
      if (fs.existsSync(absPath)) {
        return absPath
      }
    }
  }

  function _expandIncludes(fs, searchDirs, root) {
    var includes = root.findAll('include');
    if (includes.length === 0) { return false }

    includes.forEach(function (include) {
      var parent = include.parentNode;
      var href = include.attr('href');
      var rngPath = _lookupRNG(fs, searchDirs, href);
      if (!rngPath) { throw new Error(("Could not find " + href)) }
      var rngStr = fs.readFileSync(rngPath, 'utf8');
      var rng = DefaultDOMElement.parseXML(rngStr, 'full-doc');
      var grammar = rng.find('grammar');
      if (!grammar) { throw new Error('No grammar element found') }
      grammar.children.forEach(function (child) {
        parent.insertBefore(child, include);
      });
      include.remove();
    });
    return true
  }

  function _loadRNG(fs, searchDirs, entry) {
    if (!isArray(searchDirs)) { searchDirs = [searchDirs]; }
    var rngPath = _lookupRNG(fs, searchDirs, entry);
    var rngStr = fs.readFileSync(rngPath, 'utf8');
    var rng = DefaultDOMElement.parseXML(rngStr, 'full-doc');
    
    while(_expandIncludes(fs, searchDirs, rng)) {  }
    return rng
  }

  var TEXT$5 = DFA.TEXT;

  function analyze(elementSchemas) {
    forEach(elementSchemas, function (elementSchema) {
      Object.assign(elementSchema, {
        children: {},
        parents: {},
        siblings: {},
        usedInlineBy: {},
        usedStructuredBy: {}
      });
    });
    forEach(elementSchemas, function (elementSchema) {
      _analyzeElementSchema(elementSchema, elementSchemas);
    });
  }


  function _analyzeElementSchema(elementSchema, elementSchemas) {
    var expr = elementSchema.expr;
    var name = elementSchema.name;
    if (!expr) { return }
    var _siblings = [];
    if (expr instanceof DFAExpr) {
      if (expr.dfa) {
        _siblings = expr.dfa._tokensByPath();
      }
    } else if (expr instanceof InterleaveExpr) {
      expr.dfas.forEach(function (dfa) {
        if (dfa) {
          _siblings = _siblings.concat(dfa._tokensByPath());
        }
      });
    }
    if (_siblings.length === 0) {
      
      
      if (elementSchema.type === 'implicit') { elementSchema.type = 'element'; }
    }

    var hasText = false;
    var hasElements = false;
    _siblings.forEach(function (tagNames) {
      
      var _hasText = tagNames.indexOf(TEXT$5) >= 0;
      var _hasElements = (!_hasText && tagNames.length > 0);
      if (_hasText) {
        hasText = true;
      }
      if (_hasElements) {
        hasElements = true;
      }
      tagNames.forEach(function (tagName) {
        var childSchema = elementSchemas[tagName];
        if (!childSchema) { return }
        childSchema.parents[name] = true;
        elementSchema.children[tagName] = true;
        
        elementSchema.siblings[name] = tagNames;
        if (_hasElements) { childSchema.usedStructuredBy[name] = true; }
        if (_hasText) { childSchema.usedInlineBy[name] = true; }
      });
    });
    
    if (hasElements) {
      elementSchema.isStructured = true;
    }
    if (hasText) {
      elementSchema.isText = true;
    }
    if (elementSchema.type === 'implicit') {
      if (hasText) {
        elementSchema.type = 'text';
      } else {
        elementSchema.type = 'element';
      }
    }
  }

  var TEXT$6 = DFA.TEXT;


  function compileRNG(fs, searchDirs, entry) {
    var rng;
    
    if (arguments.length === 1 && isString(arguments[0])) {
      rng = DefaultDOMElement.parseXML(arguments[0]);
    } else {
      rng = _loadRNG(fs, searchDirs, entry);
    }

    var grammar = rng.find('grammar');
    if (!grammar) { throw new Error('<grammar> not found.') }

    
    _registerDefinitions(grammar);

    
    var transformedGrammar = _transformRNG(grammar);

    

    var xmlSchema = _compile(transformedGrammar);

    return xmlSchema
  }




  function _registerDefinitions(grammar) {
    var defs = {};
    
    grammar.children.forEach(function (child) {
      var tagName = nameWithoutNS$1(child.tagName);
      if (tagName === 'define') {
        _processDefine(child, defs);
      }
    });
    grammar.defs = defs;
  }

  function _processDefine(el, defs) {
    var name = el.attr('name');
    var combine = el.attr('combine');
    if (combine === 'interleave') {
      if (defs[name]) {
        defs[name].append(el.children);
      } else {
        defs[name] = el;
      }
    } else {
      if (defs[name]) {
        console.info(("Overwriting definition " + name));
      }
      defs[name] = el;
    }
  }



  function _transformRNG(grammar) { 
    
    var elements = {};
    var defs = grammar.defs;
    var elementDefinitions = grammar.findAll('define > element');
    var doc = DefaultDOMElement.createDocument('xml');
    var newGrammar = doc.createElement('grammar');
    elementDefinitions.forEach(function (el) {
      var name = el.attr('name');
      if (!name) { throw new Error("'name' is mandatory.") }
      var transformed = _transformElementDefinition(doc, name, el, defs);
      elements[name] = transformed;
      newGrammar.appendChild(transformed);
    });

    
    var elementTypes = grammar.findAll('elementType');
    elementTypes.forEach(function (typeEl) {
      var name = typeEl.attr('name');
      var type = typeEl.attr('s:type') || typeEl.attr('type');
      if (!name || !type) { throw new Error('Attributes name and type are mandatory.') }
      var element = elements[name];
      if (!element) { throw new Error(("Unknown element " + name + ".")) }
      element.attr('type', type);
    });

    
    var startElement = _extractStart(grammar);
    if (!startElement) { throw new Error('<start> is mandatory.') }
    newGrammar.appendChild(doc.createElement('start').attr('name', startElement));

    return newGrammar
  }

  function _transformElementDefinition(doc, name, orig, defs) {
    var el = doc.createElement('element').attr('name', name);
    
    
    el.attr('type', 'implicit');
    
    
    var attributes = doc.createElement('attributes');
    var children = doc.createElement('children');
    orig.children.forEach(function (child) {
      var block = _transformBlock(doc, child, defs, {});
      block.forEach(function (el) {
        if (el.find('attribute') || el.is('attribute')) {
          attributes.appendChild(el);
        } else {
          children.appendChild(el);
        }
      });
    });
    el.appendChild(attributes);
    el.appendChild(children);

    

    var hasPruned = true;
    var loop = function () {
      
      var nestedChoice = children.find('choice > choice');
      if (nestedChoice) {
        
        var parentChoice = nestedChoice.parentNode;
        
        var children$1 = nestedChoice.children;
        children$1.forEach(function (child) {
          parentChoice.insertBefore(child, nestedChoice);
        });
        parentChoice.removeChild(nestedChoice);
        return
      }
      
      var choices = children.findAll('choice');
      for (var i = 0; i < choices.length; i++) {
        var choice = choices[i];
        var children$2 = choice.children;
        if (children$2.length === 1) {
          choice.parentNode.replaceChild(choice, children$2[0]);
        }
        continue
      }
      hasPruned = false;
    };

    while (hasPruned) loop();

    return el
  }

  function _transformBlock(doc, block, defs, visiting) {
    if ( visiting === void 0 ) visiting={};

    
    
    var tagName = block.tagName;
    switch (tagName) {
      case 'element': {
        return [doc.createElement('element').attr('name',block.attr('name'))]
      }
      case 'ref': {
        return _expandRef(doc, block, defs, visiting)
      }
      case 'empty':
      case 'notAllowed': {
        return []
      }
      default: {
        var clone = block.clone(false);
        block.children.forEach(function (child) {
          clone.append(_transformBlock(doc, child, defs, visiting));
        });
        return [clone]
      }
    }
  }

  function _expandRef(doc, ref, defs, visiting) {
    if ( visiting === void 0 ) visiting={};

    var name = ref.attr('name');
    
    if (visiting[name]) {
      throw new Error('Cyclic references are not supported.')
    }
    visiting[name] = true;

    var def = defs[name];
    if (!def) { throw new Error(("Unknown definition " + name)) }

    var expanded = [];
    var children = def.children;
    children.forEach(function (child) {
      var transformed = _transformBlock(doc, child, defs, visiting);
      expanded = expanded.concat(transformed);
    });

    
    delete visiting[name];
    return expanded
  }

  function _extractStart(grammar) {
    
    
    var start = grammar.find('start');
    if (!start) {
      throw new Error('<grammar> must have a <start> element')
    }
    
    
    var startRef = start.find('ref');
    if (!startRef) {
      throw new Error('Expecting one <ref> inside of <start>.')
    }
    var name = startRef.attr('name');
    return name
  }

  function _compile(grammar) {
    var schemas = {};
    var elements = grammar.children.filter(function (el){ return el.tagName==='element'; });
    elements.forEach(function (element) {
      var name = element.attr('name');
      var attributes = _collectAttributes(element.find('attributes'));
      var children = element.find('children');
      var type = element.attr('type');
      var block = _processChildren(children, grammar);
      var expr = createExpression(name, block);
      var schema = { name: name, type: type, attributes: attributes, expr: expr };
      schemas[name] = schema;
    });

    
    analyze(schemas);

    var start = grammar.find('start');
    if (!start) {
      throw new Error('<start> is mandatory')
    }
    var startElement = start.attr('name');
    if (!startElement) {
      throw new Error('<start> must have "name" set')
    }
    return new XMLSchema(schemas, startElement)
  }

  function _processChildren(el, grammar) {
    var blocks = _processBlocks(el.children, grammar);
    if (blocks.length === 1) {
      return blocks[0]
    } else {
      return new Sequence(blocks)
    }
  }

  function _processBlocks(children, grammar) {
    var blocks = [];
    for (var i = 0; i < children.length; i++) {
      var child = children[i];
      
      switch(child.tagName) {
        
        case 'attribute':
        case 'empty':
        case 'notAllowed': {
          break
        }
        case 'element': {
          var elName = child.attr('name');
          blocks.push(new Token(elName));
          break
        }
        case 'text': {
          blocks.push(new Token(TEXT$6));
          break
        }
        case 'ref': {
          var block = _processReference(child, grammar);
          blocks.push(block);
          break
        }
        case 'group': {
          blocks.push(_processSequence(child, grammar));
          break
        }
        case 'choice': {
          var block$1 = _processChoice(child, grammar);
          blocks.push(block$1);
          break
        }
        case 'optional': {
          var block$2 = new Optional(_processChildren(child, grammar));
          blocks.push(block$2);
          break
        }
        case 'oneOrMore': {
          var block$3 = new Plus(_processChildren(child, grammar));
          blocks.push(block$3);
          break
        }
        case 'zeroOrMore': {
          var block$4 = new Kleene(_processChildren(child, grammar));
          blocks.push(block$4);
          break
        }
        case 'interleave': {
          var block$5 = new Interleave(_processBlocks(child.children, grammar));
          blocks.push(block$5);
          break
        }
        default:
          throw new Error('Not supported yet: ' + child.tagName)
      }
    }
    return blocks
  }

  function _processSequence(el, grammar) {
    if (el.expr) { return el.expr.copy() }
    var blocks = _processBlocks(el.children, grammar);
    el.expr = new Sequence(blocks);
    return el.expr
  }

  function _processChoice(el, grammar) {
    if (el.expr) { return el.expr.copy() }
    var blocks = _processBlocks(el.children, grammar);
    el.expr = new Choice(blocks);
    return el.expr
  }

  function _processReference(ref, grammar) {
    var name = ref.attr('name');
    var def = grammar.defs[name];
    if (!def) { throw new Error(("Illegal ref: " + name + " is not defined.")) }
    if (def.expr) { return def.expr.copy() }
    
    
    if (grammar._visiting[name]) {
      throw new Error('Cyclic references are not supported yet')
    }
    grammar._visiting[name] = true;
    var block = _processChildren(def, grammar);
    def.expr = block;
    delete grammar._visiting[name];
    return def.expr
  }

  function _collectAttributes(el, grammar, attributes) {
    if ( attributes === void 0 ) attributes = {};

    
    
    var children = el.children;
    for (var i = 0; i < children.length; i++) {
      var child = children[i];
      switch (child.tagName) {
        case 'attribute': {
          var attr = _transformAttribute(child);
          attributes[attr.name] = attr;
          break
        }
        case 'group':
        case 'choice':
        case 'optional':
        case 'oneOrMore':
        case 'zeroOrMore': {
          _collectAttributes(child, grammar, attributes);
          break
        }
        default:
          
      }
    }
    return attributes
  }

  function _transformAttribute(el) {
    var name = el.attr('name');
    
    return {
      name: name
    }
  }

  var DISABLED = Object.freeze({
    disabled: true
  });


  var SchemaDrivenCommandManager = (function (CommandManager$$1) {
    function SchemaDrivenCommandManager () {
      CommandManager$$1.apply(this, arguments);
    }

    if ( CommandManager$$1 ) SchemaDrivenCommandManager.__proto__ = CommandManager$$1;
    SchemaDrivenCommandManager.prototype = Object.create( CommandManager$$1 && CommandManager$$1.prototype );
    SchemaDrivenCommandManager.prototype.constructor = SchemaDrivenCommandManager;

    SchemaDrivenCommandManager.prototype._initialize = function _initialize () {
      
      
      var annotationCommands = new Map();
      var insertCommands = new Map();
      var switchTypeCommands = new Map();
      var otherCommands = new Map();
      this.commands.forEach(function (command) {
        var name = command.getName();
        if (command.isAnnotationCommand()) {
          annotationCommands.set(name, command);
        } else if (command.isInsertCommand()) {
          insertCommands.set(name, command);
        } else if (command.isSwitchTypeCommand()) {
          switchTypeCommands.set(name, command);
        } else {
          otherCommands.set(name, command);
        }
      });
      this.annotationCommands = annotationCommands;
      this.insertCommands = insertCommands;
      this.switchTypeCommands = switchTypeCommands;
      this.otherCommands = otherCommands;
    };

    SchemaDrivenCommandManager.prototype._getCommand = function _getCommand (commandName) {
      var cmd = this.annotationCommands.get(commandName) ||
        this.insertCommands.get(commandName) ||
        this.switchTypeCommands.get(commandName) ||
        this.otherCommands.get(commandName);
      return cmd
    };

    SchemaDrivenCommandManager.prototype._updateCommandStates = function _updateCommandStates (editorSession) {
      var commandStates = {};
      var annotationCommands = this.annotationCommands;
      var insertCommands = this.insertCommands;
      var switchTypeCommands = this.switchTypeCommands;
      var commandContext = this._getCommandContext();
      var params = this._getCommandParams();
      var doc = editorSession.getDocument();
      var selectionState = params.selectionState;
      var sel = params.selection;
      var isBlurred = params.editorSession.isBlurred();
      var noSelection = !sel || sel.isNull() || !sel.isAttached();

      
      
      
      
      if (isBlurred || noSelection || sel.isCustomSelection()) {
        _disableEditingCommands();
      } else {
        var path = sel.start.path;
        var node = doc.get(path[0]);

        
        
        
        
        if (!node) {
          
          throw new Error('FIXME: explain when this happens')
        }

        var isInsideText = node.isText();
        var xmlSchema = doc.getXMLSchema();

        
        
        if (isInsideText && sel.isPropertySelection() && !selectionState.isInlineNodeSelection()) {
          var elementSchema = xmlSchema.getElementSchema(node.type);
          _evaluateTyped(this.annotationCommands, elementSchema);
        } else {
          _disable(this.annotationCommands);
        }

        
        var parentNode = node.parentNode;
        if (!parentNode || !parentNode.isContainer()) {
          _disable(this.insertCommands);
        } else {
          
          
          var elementSchema$1 = xmlSchema.getElementSchema(parentNode.type);
          _evaluateTyped(this.insertCommands, elementSchema$1);
        }

        
        
        if (!parentNode || !sel.containerId || !isInsideText || !parentNode.isContainer()) {
          _disable(this.switchTypeCommands);
        } else {
          
          
          var elementSchema$2 = xmlSchema.getElementSchema(parentNode.type);
          _evaluateTyped(this.switchTypeCommands, elementSchema$2);
        }
      }

      
      _evaluateUntyped(this.otherCommands);

      
      
      editorSession.setCommandStates(commandStates);

      function _disableEditingCommands() {
        _disable(annotationCommands);
        _disable(insertCommands);
        _disable(switchTypeCommands);
      }

      function _disable(commands) {
        commands.forEach(function (cmd, name) {
          commandStates[name] = DISABLED;
        });
      }
      function _evaluateTyped(commands, elementSchema) {
        commands.forEach(function (cmd, name) {
          var type = cmd.getType();
          if (elementSchema.isAllowed(type)) {
            commandStates[name] = cmd.getCommandState(params, commandContext);
          } else {
            commandStates[name] = DISABLED;
          }
        });
      }
      function _evaluateUntyped(commands) {
        commands.forEach(function (cmd, name) {
          commandStates[name] = cmd.getCommandState(params, commandContext);
        });
      }

    };

    return SchemaDrivenCommandManager;
  }(CommandManager));

  function validateXML(xmlSchema, dom) {
    var root = dom.find(xmlSchema.getStartElement());
    if (!root) {
      return {
        errors: [{
          msg: 'Start element is missing.',
          el: dom
        }]
      }
    } else {
      return validateElement(xmlSchema, root)
    }
  }

  function validateElement(xmlSchema, el) {
    var errors = [];
    var valid = true;
    var q = [el];
    while(q.length>0) {
      var next = q.shift();
      var res = xmlSchema.validateElement(next);
      if (!res.ok) {
        errors = errors.concat(res.errors);
        valid = false;
      }
      if (next.isElementNode()) {
        q = q.concat(next.getChildren());
      }
    }
    return {
      errors: errors,
      ok: valid
    }
  }

  function prettyPrintXML(xml) {
    var dom;
    if (isString(xml)) {
      dom = DefaultDOMElement.parseXML(xml);
    } else {
      dom = xml;
    }
    var result = [];
    dom.childNodes.forEach(function (el) {
      _prettyPrint(result, el, 0);
    });
    return result.join('\n')
  }

  function _prettyPrint(result, el, level) {
    var indent = new Array(level*2).fill(' ').join('');
    if (el.isElementNode()) {
      var isMixed = _isMixed(el);
      var containsCDATA = _containsCDATA(el);
      if (isMixed || containsCDATA) {
        result.push(indent + el.outerHTML);
      } else {
        var children = el.children;
        var tagName = el.tagName;
        var tagStr = [("<" + tagName)];
        el.getAttributes().forEach(function (val, name) {
          tagStr.push((name + "=\"" + val + "\""));
        });
        if (children.length > 0) {
          result.push(indent + tagStr.join(' ') + '>');
          el.children.forEach(function (child) {
            _prettyPrint(result, child, level+1);
          });
          result.push(indent + "</" + tagName + ">");
        } else {
          result.push(indent + tagStr.join(' ') + ' />');
        }
      }
    } else {
      result.push(indent + el.outerHTML);
    }
  }

  function _isMixed(el) {
    var childNodes = el.childNodes;
    for (var i = 0; i < childNodes.length; i++) {
      var child = childNodes[i];
      if (child.isTextNode() && !_isTextNodeEmpty(child)) {
        return true
      }
    }
  }

  function _containsCDATA(el) {
    var childNodes = el.childNodes;
    for (var i = 0; i < childNodes.length; i++) {
      var child = childNodes[i];
      if (child.getNodeType() === 'cdata') {
        return true
      }
    }
  }

  var ButtonPackage = {
    name: 'button',
    configure: function(config) {
      config.addComponent('button', Button);
      config.addIcon('dropdown', { 'fontawesome': 'fa-angle-down' });
    },
    Button: Button
  }

  var ContextMenu = (function (ToolPanel$$1) {
    function ContextMenu () {
      ToolPanel$$1.apply(this, arguments);
    }

    if ( ToolPanel$$1 ) ContextMenu.__proto__ = ToolPanel$$1;
    ContextMenu.prototype = Object.create( ToolPanel$$1 && ToolPanel$$1.prototype );
    ContextMenu.prototype.constructor = ContextMenu;

    ContextMenu.prototype.didMount = function didMount () {
      ToolPanel$$1.prototype.didMount.call(this);
      if (!this.context.scrollPane) {
        throw new Error('Requires a scrollPane context')
      }
      this.context.scrollPane.on('context-menu:opened', this._onContextMenuOpened, this);
    };

    ContextMenu.prototype.dispose = function dispose () {
      ToolPanel$$1.prototype.dispose.call(this);
      this.context.scrollPane.off(this);
    };

    ContextMenu.prototype.render = function render ($$) {
      var el = $$('div').addClass('sc-context-menu sm-hidden');
      el.addClass('sm-theme-'+this.getTheme());
      el.append(
        $$('div').addClass('se-active-tools').append(
          this.renderEntries($$)
        ).ref('entriesContainer')
      );
      return el
    };

    ContextMenu.prototype.show = function show (hints) {
      this.el.removeClass('sm-hidden');
      this._position(hints);
    };

    ContextMenu.prototype.hide = function hide () {
      this.el.addClass('sm-hidden');
    };

    
    ContextMenu.prototype._onContextMenuOpened = function _onContextMenuOpened (hints) {
      var mouseBounds = hints.mouseBounds;
      this.el.removeClass('sm-hidden');
      var contextMenuWidth = this.el.htmlProp('offsetWidth');

      
      this.el.css('top', mouseBounds.top);
      var leftPos = mouseBounds.left;
      
      leftPos = Math.max(leftPos, 0);
      
      var maxLeftPos = mouseBounds.left + mouseBounds.right - contextMenuWidth;
      leftPos = Math.min(leftPos, maxLeftPos);
      this.el.css('left', leftPos);
    };

    ContextMenu.prototype.getTheme = function getTheme () {
      return this.props.theme || 'dark'
    };

    return ContextMenu;
  }(ToolPanel));

  var ContextMenuPackage = {
    name: 'context-menu',
    configure: function(config) {
      config.addComponent('context-menu', ContextMenu);
    }
  }

  var Grid = (function (Component$$1) {
    function Grid () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) Grid.__proto__ = Component$$1;
    Grid.prototype = Object.create( Component$$1 && Component$$1.prototype );
    Grid.prototype.constructor = Grid;

    Grid.prototype.render = function render ($$) {
      var el = $$('div').addClass('sc-grid');
      if (this.props.mobile) {
        el.addClass('sm-mobile');
      }
      el.append(this.props.children);
      return el
    };

    return Grid;
  }(Component));


  var Row = (function (Component$$1) {
    function Row () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) Row.__proto__ = Component$$1;
    Row.prototype = Object.create( Component$$1 && Component$$1.prototype );
    Row.prototype.constructor = Row;

    Row.prototype.render = function render ($$) {
      var el = $$('div').addClass('se-row');
      el.append(this.props.children);
      return el
    };

    return Row;
  }(Component));


  var Cell = (function (Component$$1) {
    function Cell () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) Cell.__proto__ = Component$$1;
    Cell.prototype = Object.create( Component$$1 && Component$$1.prototype );
    Cell.prototype.constructor = Cell;

    Cell.prototype.render = function render ($$) {
      var el = $$('div').addClass('se-cell');
      el.addClass('sm-column-'+this.props.columns);
      el.append(this.props.children);
      return el
    };

    return Cell;
  }(Component));

  Grid.Row = Row;
  Grid.Cell = Cell;

  var GridPackage = {
    name: 'grid',
    configure: function(config) {
      config.addComponent('grid', Grid);
    },
    Grid: Grid
  }

  var Gutter = (function (ToolPanel$$1) {
    function Gutter () {
      ToolPanel$$1.apply(this, arguments);
    }

    if ( ToolPanel$$1 ) Gutter.__proto__ = ToolPanel$$1;
    Gutter.prototype = Object.create( ToolPanel$$1 && ToolPanel$$1.prototype );
    Gutter.prototype.constructor = Gutter;

    Gutter.prototype.render = function render ($$) {
      var this$1 = this;

      var el = $$('div').addClass(this.getClassNames());
      el.addClass('sm-hidden');
      el.addClass('sm-theme-'+this.getTheme());
      var activeToolGroups = this.state.activeToolGroups;
      var activeToolsEl = $$('div').addClass('se-active-tools');

      activeToolGroups.forEach(function (toolGroup) {
        var toolGroupProps = Object.assign({}, toolGroup, {
          toolStyle: this$1.getToolStyle(),
          showIcons: true
        });
        activeToolsEl.append(
          $$(toolGroup.Class, toolGroupProps)
        );
      });

      el.append(activeToolsEl);
      return el
    };

    
    Gutter.prototype.getToolStyle = function getToolStyle () {
      return 'outline-dark'
    };

    Gutter.prototype.show = function show (hints) {
      this.el.removeClass('sm-hidden');
      this._position(hints);
    };

    Gutter.prototype.hide = function hide () {
      this.el.addClass('sm-hidden');
    };

    Gutter.prototype._position = function _position (hints) {
      if (hints) {
        
        this.el.css('top', hints.rectangle.top + hints.rectangle.height - hints.rectangle.height / 2);
        this.el.css('left', 0);
      }
    };

    Gutter.prototype.getClassNames = function getClassNames () {
      return 'sc-gutter'
    };

    Gutter.prototype.getTheme = function getTheme () {
      return 'dark'
    };

    Gutter.prototype.getActiveToolGroupNames = function getActiveToolGroupNames () {
      return ['gutter']
    };

    return Gutter;
  }(ToolPanel));

  var GutterPackage = {
    name: 'gutter',
    configure: function(config) {
      config.addComponent('gutter', Gutter);
    }
  }

  var Input = (function (Component$$1) {
    function Input () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) Input.__proto__ = Component$$1;
    Input.prototype = Object.create( Component$$1 && Component$$1.prototype );
    Input.prototype.constructor = Input;

    Input.prototype.render = function render ($$) {
      var val = this._getDocumentValue();

      var el = $$('input').attr({
        value: val,
        type: this.props.type,
        placeholder: this.props.placeholder
      }).addClass('sc-input')
        .on('keydown', this._onKeydown);

      if (this.props.path) {
        el.on('change', this._onChange);
      }

      if (this.props.centered) {
        el.addClass('sm-centered');
      }

      return el
    };

    Input.prototype._onChange = function _onChange () {
      var editorSession = this.context.editorSession;
      var path = this.props.path;
      var newVal = this.el.val();
      var oldVal = this._getDocumentValue();
      if (newVal !== oldVal) {
        editorSession.transaction(function(tx) {
          tx.set(path, newVal);
        });
        if (this.props.retainFocus) {
          
          
          this.el.getNativeElement().focus();
        }
      }
    };

    Input.prototype._getDocumentValue = function _getDocumentValue () {
      if (this.props.val) {
        return this.props.val
      } else {
        var editorSession = this.context.editorSession;
        var path = this.props.path;
        return editorSession.getDocument().get(path)
      }
    };

    Input.prototype._onKeydown = function _onKeydown (event) {
      if (event.keyCode === keys$1.ESCAPE) {
        event.stopPropagation();
        event.preventDefault();
        this.el.val(this._getDocumentValue());
      }
    };

    return Input;
  }(Component));

  var InputPackage = {
    name: 'input',
    configure: function(config) {
      config.addComponent('input', Input);
    },
    Input: Input
  }

  var LayoutPackage = {
    name: 'layout',
    configure: function(config) {
      config.addComponent('layout', Layout);
    },
    Layout: Layout
  }

  var Modal = (function (Component$$1) {
    function Modal () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) Modal.__proto__ = Component$$1;
    Modal.prototype = Object.create( Component$$1 && Component$$1.prototype );
    Modal.prototype.constructor = Modal;

    Modal.prototype.render = function render ($$) {
      var el = $$('div').addClass('sc-modal');

      
      
      el.on('click', this._closeModal);

      if (this.props.width) {
        el.addClass('sm-width-'+this.props.width);
      }

      el.append(
        $$('div').addClass('se-body').append(
          this.props.children
        )
      );
      return el
    };

    Modal.prototype._closeModal = function _closeModal (e) {
      var closeSurfaceClick = e.target.classList.contains('sc-modal');
      if (closeSurfaceClick) {
        this.send('closeModal');
      }
    };

    return Modal;
  }(Component));

  var ModalPackage = {
    name: 'modal',
    configure: function(config) {
      config.addComponent('modal', Modal);
    },
    Modal: Modal
  }

  var OverlayPackage = {
    name: 'overlay',
    configure: function(config) {
      config.addComponent('overlay', Overlay);
    }
  }

  var Dropzones = (function (Component$$1) {
    function Dropzones () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) Dropzones.__proto__ = Component$$1;
    Dropzones.prototype = Object.create( Component$$1 && Component$$1.prototype );
    Dropzones.prototype.constructor = Dropzones;

    Dropzones.prototype.didMount = function didMount () {
      this.context.dragManager.on('drag:started', this.onDragStarted, this);
      this.context.dragManager.on('drag:finished', this.onDragFinished, this);
    };

    Dropzones.prototype.render = function render ($$) {
      var this$1 = this;

      var el = $$('div').addClass('sc-dropzones');

      if (this.state.dropzones) {
        el.on('dragenter', this.onDrag)
          .on('dragover', this.onDrag);

        
        forEach(this.state.dropzones, function (dropzones, surfaceId) {
          dropzones.forEach(function (dropzone, index) {
            var dropType = dropzone.type;
            var dropzoneEl;
            if (dropType === 'place') {
              dropzoneEl = $$('div').addClass('se-dropzone')
                .attr({
                  'data-dropzone-index': index,
                  'data-dropzone-surface': surfaceId
                }).append(
                  $$('div').addClass('se-drop-teaser').css({
                    top: dropzone.teaserPos
                  })
                );
            } else if (dropType === 'custom') {
              dropzoneEl = $$('div').addClass('se-custom-dropzone').attr({
                'data-dropzone-index': index,
                'data-dropzone-surface': surfaceId
              }).append(
                
                
                $$('div').addClass('se-message').append(dropzone.message)
              );
            }
            if (dropzoneEl) {
              var shield = $$('div').addClass('se-drop-shield')
                .on('dragenter', this$1.onDragEnter)
                .on('dragleave', this$1.onDragLeave)
                .on('drop', this$1.onDrop)
                .on('mouseenter', this$1.onDragEnter)
                .on('mouseleave', this$1.onDragLeave)
                .on('mouseup', this$1.onDrop);
              dropzoneEl.append(shield);
              dropzoneEl.css({
                position: 'absolute',
                top: dropzone.top,
                left: dropzone.left,
                width: dropzone.width,
                height: dropzone.height
              });
              el.append(dropzoneEl);
            }
          });
        });
      } else {
        el.addClass('sm-hidden');
      }
      return el
    };

    
    Dropzones.prototype.onDragStarted = function onDragStarted (dragState) {
      var this$1 = this;

      var dropzones = this._computeDropzones(dragState);
      setTimeout(function () {
        this$1.setState({
          dropzones: dropzones
        });
      }, 250);
    };

    
    Dropzones.prototype.onDragFinished = function onDragFinished () {
      this.setState({});
    };

    Dropzones.prototype.onDragEnter = function onDragEnter (e) {
      
      e.target.parentNode.classList.add('sm-over');
    };

    Dropzones.prototype.onDragLeave = function onDragLeave (e) {
      
      e.target.parentNode.classList.remove('sm-over');
    };

    
    Dropzones.prototype.onDrag = function onDrag (e) { 
      
      e.preventDefault();
    };

    Dropzones.prototype.onDrop = function onDrop (e) {
      
      
      e.__reserved__ = true;
      e.preventDefault();
      e.stopPropagation();
      var dropzoneIndex = e.target.parentNode.dataset.dropzoneIndex;
      var dropzoneSurface = e.target.parentNode.dataset.dropzoneSurface;
      var dropzone = this.state.dropzones[dropzoneSurface][dropzoneIndex];
      var dropParams = dropzone.dropParams;
      var dropType = dropzone.type;
      
      var targetSurface = this.context.surfaceManager.getSurface(dropzoneSurface);
      
      var component = dropzone.component;
      var dropzoneComponent = dropzone.dropzoneComponent;
      
      var dragManager = this.context.dragManager;
      dragManager.extendDragState({
        targetSurface: targetSurface,
        dropType: dropType,
        dropParams: dropParams,
        component: component,
        dropzoneComponent: dropzoneComponent
      });
      dragManager._onDragEnd(e);
    };

    
    Dropzones.prototype._getBoundingRect = function _getBoundingRect (comp) {
      var scrollPane = comp.context.scrollPane;
      var contentElement = scrollPane.getContentElement();
      var rect = getRelativeBoundingRect(comp.el, contentElement);
      return rect
    };

    Dropzones.prototype._computeDropzones = function _computeDropzones (dragState) {
      var this$1 = this;

      var scrollPaneName = this.context.scrollPane.getName();
      var surfaces = dragState.scrollPanes[scrollPaneName].surfaces;
      var scopedDropzones = {};

      forEach(surfaces, function (surface) {
        var components = surface.childNodes;

        
        var numDropzones = components.length + 1;
        var dropzones = [];

        for (var i = 0; i < numDropzones; i++) {
          if (i === 0) {
            
            var firstComp = this$1._getBoundingRect(components[0]);
            dropzones.push({
              type: 'place',
              left: firstComp.left,
              top: firstComp.top,
              width: firstComp.width,
              height: firstComp.height / 2,
              teaserPos: 0,
              dropParams: {
                insertPos: i
              }
            });
          } else if (i === numDropzones - 1) {
            
            var lastComp = this$1._getBoundingRect(components[i - 1]);
            dropzones.push({
              type: 'place',
              left: lastComp.left,
              top: lastComp.top + lastComp.height / 2,
              width: lastComp.width,
              height: lastComp.height / 2,
              teaserPos: lastComp.height / 2,
              dropParams: {
                insertPos: i
              }
            });
          } else {
            
            var upperComp = this$1._getBoundingRect(components[i-1]);
            var lowerComp = this$1._getBoundingRect(components[i]);
            var topBound = upperComp.top + upperComp.height / 2;
            var bottomBound = lowerComp.top + lowerComp.height / 2;

            dropzones.push({
              type: 'place',
              left: upperComp.left,
              top: topBound,
              width: upperComp.width,
              height: bottomBound - topBound,
              teaserPos: (upperComp.top + upperComp.height + lowerComp.top) / 2 - topBound,
              dropParams: {
                insertPos: i
              }
            });
          }

          if (i < numDropzones - 1) {
            var comp = components[i];
            
            if (comp._isIsolatedNodeComponent) {
              comp = comp.getContent();
            }
            
            if (comp.getDropzoneSpecs) {
              var dropzoneSpecs = comp.getDropzoneSpecs();
              dropzoneSpecs.forEach(function (dropzoneSpec) {
                var dropzoneComp = dropzoneSpec.component;
                var rect = this$1._getBoundingRect(dropzoneComp);
                dropzones.push({
                  type: 'custom',
                  component: comp,
                  dropzoneComponent: dropzoneComp,
                  left: rect.left,
                  top: rect.top,
                  width: rect.width,
                  height: rect.height,
                  message: dropzoneSpec.message,
                  dropParams: dropzoneSpec.dropParams
                });
              });
            }
          }
        }
        scopedDropzones[surface.getName()] = dropzones;
      });
      return scopedDropzones
    };

    Dropzones.prototype._renderDropTeaser = function _renderDropTeaser (hints) {
      if (hints.visible) {
        this.el.removeClass('sm-hidden');
        this.el.css('top', hints.rect.top);
        this.el.css('left', hints.rect.left);
        this.el.css('right', hints.rect.right);
      } else {
        this.el.addClass('sm-hidden');
      }
    };

    return Dropzones;
  }(Component));

  var DropzonesPackage = {
    name: 'dropzones',
    configure: function(config) {
      config.addComponent('dropzones', Dropzones);
    },
    Dropzones: Dropzones
  }

  var ScrollbarPackage = {
    name: 'scrollbar',
    configure: function(config) {
      config.addComponent('scrollbar', Scrollbar);
    },
    Scrollbar: Scrollbar
  }

  var ScrollPanePackage = {
    name: 'scroll-pane',
    configure: function(config) {
      config.addComponent('scroll-pane', ScrollPane);
    },
    ScrollPane: ScrollPane
  }

  var BodyScrollPane = (function (AbstractScrollPane$$1) {
    function BodyScrollPane () {
      AbstractScrollPane$$1.apply(this, arguments);
    }

    if ( AbstractScrollPane$$1 ) BodyScrollPane.__proto__ = AbstractScrollPane$$1;
    BodyScrollPane.prototype = Object.create( AbstractScrollPane$$1 && AbstractScrollPane$$1.prototype );
    BodyScrollPane.prototype.constructor = BodyScrollPane;

    BodyScrollPane.prototype.getChildContext = function getChildContext () {
      return {
        scrollPane: this
      }
    };

    BodyScrollPane.prototype.getName = function getName () {
      return 'body'
    };

    BodyScrollPane.prototype.render = function render ($$) {
      var el = $$('div');
      if (this.props.contextMenu === 'custom') {
        el.on('contextmenu', this._onContextMenu);
      }
      el.append(this.props.children);
      return el
    };

    
    BodyScrollPane.prototype.getHeight = function getHeight () {
      if (platform.inBrowser) {
        return window.innerHeight
      } else {
        return 0
      }
    };

    
    BodyScrollPane.prototype.getContentHeight = function getContentHeight () {
      if (platform.inBrowser) {
        return document.body.scrollHeight
      } else {
        return 0
      }
    };

    BodyScrollPane.prototype.getContentElement = function getContentElement () {
      if (platform.inBrowser) {
        return DefaultDOMElement.wrapNativeElement(window.document.body)
      } else {
        return null
      }
    };

    
    
    
    BodyScrollPane.prototype.getScrollableElement = function getScrollableElement () {
      if (platform.inBrowser) {
        return document.body
      } else {
        return null
      }
    };

    
    BodyScrollPane.prototype.getScrollPosition = function getScrollPosition () {
      if (platform.inBrowser) {
        return document.body.scrollTop
      } else {
        return 0
      }
    };

    BodyScrollPane.prototype.setScrollPosition = function setScrollPosition (scrollPos) {
      if (platform.inBrowser) {
        document.body.scrollTop = scrollPos;
      }
    };

    
    BodyScrollPane.prototype.getPanelOffsetForElement = function getPanelOffsetForElement (el) { 
      console.warn('TODO: implement');
    };

    
    BodyScrollPane.prototype.scrollTo = function scrollTo (componentId, onlyIfNotVisible) { 
      console.warn('TODO: implement');
    };

    return BodyScrollPane;
  }(AbstractScrollPane));

  var BodyScrollPanePackage = {
    name: 'body-scroll-pane',
    configure: function(config) {
      config.addComponent('body-scroll-pane', BodyScrollPane);
    },
    BodyScrollPane: BodyScrollPane
  }

  var SplitPanePackage = {
    name: 'split-pane',
    configure: function(config) {
      config.addComponent('split-pane', SplitPane);
    },
    SplitPane: SplitPane
  }

  var TabbedPane = (function (Component$$1) {
    function TabbedPane () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) TabbedPane.__proto__ = Component$$1;
    TabbedPane.prototype = Object.create( Component$$1 && Component$$1.prototype );
    TabbedPane.prototype.constructor = TabbedPane;

    TabbedPane.prototype.render = function render ($$) {
      var el = $$('div').addClass('sc-tabbed-pane');
      var tabsEl = $$('div').addClass('se-tabs');
      forEach(this.props.tabs, function(tab) {
        var tabEl = $$('a')
          .addClass("se-tab")
          .attr({
            href: "#",
            "data-id": tab.id,
          })
          .on('click', this.onTabClicked);
        if (tab.id === this.props.activeTab) {
          tabEl.addClass("sm-active");
        }
        tabEl.append(
          $$('span').addClass('label').append(tab.name)
        );
        tabsEl.append(tabEl);
      }.bind(this));

      el.append(tabsEl);
      
      el.append(
        $$('div').addClass('se-tab-content').ref('tabContent').append(
          this.props.children
        )
      );
      return el
    };

    TabbedPane.prototype.onTabClicked = function onTabClicked (e) {
      e.preventDefault();
      var tabId = e.currentTarget.dataset.id;
      this.send('switchTab', tabId);
    };

    return TabbedPane;
  }(Component));

  var TabbedPanePackage = {
    name: 'tabbed-pane',
    configure: function(config) {
      config.addComponent('tabbed-pane', TabbedPane);
    },
    TabbedPane: TabbedPane
  }

  var FilePackage = {
    name: 'file',
    configure: function(config) {
      config.addNode(FileNode);
    }
  }

  var Undo = (function (Command$$1) {
    function Undo () {
      Command$$1.apply(this, arguments);
    }

    if ( Command$$1 ) Undo.__proto__ = Command$$1;
    Undo.prototype = Object.create( Command$$1 && Command$$1.prototype );
    Undo.prototype.constructor = Undo;

    Undo.prototype.getCommandState = function getCommandState (params) {
      var editorSession = params.editorSession;
      return {
        disabled: !editorSession.canUndo(),
        active: false
      }
    };

    Undo.prototype.execute = function execute (params) {
      var editorSession = params.editorSession;
      if (editorSession.canUndo()) {
        editorSession.undo();
      }
      return true
    };

    return Undo;
  }(Command));

  var Redo = (function (Command$$1) {
    function Redo () {
      Command$$1.apply(this, arguments);
    }

    if ( Command$$1 ) Redo.__proto__ = Command$$1;
    Redo.prototype = Object.create( Command$$1 && Command$$1.prototype );
    Redo.prototype.constructor = Redo;

    Redo.prototype.getCommandState = function getCommandState (params) {
      var editorSession = params.editorSession;
      return {
        disabled: !editorSession.canRedo(),
        active: false
      }
    };

    Redo.prototype.execute = function execute (params) {
      var editorSession = params.editorSession;
      if (editorSession.canRedo()) {
        editorSession.redo();
        return true
      } else {
        return false
      }
    };

    return Redo;
  }(Command));

  var SelectAll = (function (Command$$1) {
    function SelectAll () {
      Command$$1.apply(this, arguments);
    }

    if ( Command$$1 ) SelectAll.__proto__ = Command$$1;
    SelectAll.prototype = Object.create( Command$$1 && Command$$1.prototype );
    SelectAll.prototype.constructor = SelectAll;

    SelectAll.prototype.getCommandState = function getCommandState (params) {
      var editorSession = params.editorSession;
      var isBlurred = editorSession.isBlurred();
      var surface = params.surface || editorSession.getFocusedSurface();
      var sel = editorSession.getSelection();
      
      var disabled = (
        isBlurred ||
        !sel || sel.isNull() ||
        !surface ||
        !(surface._isContainerEditor || surface._isTextPropertyEditor)
      );
      return { disabled: disabled }
    };

    SelectAll.prototype.execute = function execute (params) {
      var editorSession = params.editorSession;
      var doc = editorSession.getDocument();
      var surface = params.surface || editorSession.getFocusedSurface();
      if (surface) {
        var sel;
        
        if (surface._isContainerEditor) {
          var container = surface.getContainer();
          var nodeIds = container.getContent();
          if (nodeIds.length === 0) { return false }
          var firstNodeId = nodeIds[0];
          var lastNodeId = last$1(nodeIds);
          sel = editorSession.createSelection({
            type: 'container',
            startPath: [firstNodeId],
            startOffset: 0,
            endPath: [lastNodeId],
            endOffset: 1,
            containerId: container.id,
            surfaceId: surface.id
          });
        } else if (surface._isTextPropertyEditor) {
          var path = surface.getPath();
          var text = doc.get(path);
          sel = editorSession.createSelection({
            type: 'property',
            path: path,
            startOffset: 0,
            endOffset: text.length,
            surfaceId: surface.id
          });
        }
        editorSession.setSelection(sel);
        return true
      }
      return false
    };

    return SelectAll;
  }(Command));

  var ToolSeparator = (function (Component$$1) {
    function ToolSeparator () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) ToolSeparator.__proto__ = Component$$1;
    ToolSeparator.prototype = Object.create( Component$$1 && Component$$1.prototype );
    ToolSeparator.prototype.constructor = ToolSeparator;

    ToolSeparator.prototype.render = function render ($$) {
      var el = $$('div').addClass('sc-tool-separator');
      return el
    };

    ToolSeparator.prototype.hasEnabledTools = function hasEnabledTools () {
      return false
    };

    return ToolSeparator;
  }(Component));

  var ToolPanelPackage = {
    name: 'tool-panel',
    configure: function configure(config) {
      config.addComponent('tool-panel', ToolPanel);
      config.addComponent('tool-dropdown', ToolDropdown);
      config.addComponent('tool-group', ToolGroup);
      config.addComponent('menu-group', MenuGroup);
      config.addComponent('tool-prompt', ToolPrompt);
      config.addComponent('tool-separator', ToolSeparator);
    }
  }

  var BasePackage = {
    name: 'base',
    configure: function(config) {
      config.import(FilePackage);
      config.import(ScrollPanePackage);
      config.import(BodyScrollPanePackage);
      config.import(SplitPanePackage);
      config.import(TabbedPanePackage);
      config.import(ScrollbarPackage);
      config.import(GridPackage);
      config.import(ModalPackage);
      config.import(InputPackage);
      config.import(ButtonPackage);
      config.import(LayoutPackage);
      config.import(ContextMenuPackage);
      config.import(OverlayPackage);
      config.import(DropzonesPackage);
      config.import(GutterPackage);
      config.import(ToolPanelPackage);

      
      config.addCommand('undo', Undo, { commandGroup: 'undo-redo' });
      config.addCommand('redo', Redo, { commandGroup: 'undo-redo' });
      config.addCommand('select-all', SelectAll, { commandGroup: 'selection' });

      
      config.addIcon('insert', { 'fontawesome': 'fa-plus' });
      config.addIcon('undo', { 'fontawesome': 'fa-undo' });
      config.addIcon('redo', { 'fontawesome': 'fa-repeat' });
      config.addIcon('edit', { 'fontawesome': 'fa-cog' });
      config.addIcon('delete', { 'fontawesome': 'fa-times' });
      config.addIcon('expand', { 'fontawesome': 'fa-arrows-h' });
      config.addIcon('truncate', { 'fontawesome': 'fa-arrows-h' });

      
      config.addLabel('undo', {
        en: 'Undo',
        de: 'Rückgängig'
      });
      config.addLabel('redo', {
        en: 'Redo',
        de: 'Wiederherstellen'
      });
      config.addLabel('select-all', {
        en: 'Select All',
        de: 'Alles Auswählen'
      });
      config.addLabel('text-types', {
        en: 'Text Type',
        de: 'Texttyp'
      });
      config.addLabel('container-selection', {
        en: 'Container',
        de: 'Container'
      });
      config.addLabel('container', {
        en: 'Container',
        de: 'Container'
      });
      config.addLabel('insert', {
        en: 'Insert',
        de: 'Einfügen'
      });
      config.addLabel('insert-container', {
        en: 'Insert Container',
        de: 'Container einfügen'
      });

      config.addKeyboardShortcut('CommandOrControl+Z', { command: 'undo' });
      config.addKeyboardShortcut('CommandOrControl+Shift+Z', { command: 'redo' });
      config.addKeyboardShortcut('CommandOrControl+A', { command: 'select-all' });
    },
    UndoCommand: Undo,
    RedoCommand: Redo,
    SelectAllCommand: SelectAll
  }

  var Blockquote = (function (TextBlock$$1) {
  	function Blockquote () {
  		TextBlock$$1.apply(this, arguments);
  	}if ( TextBlock$$1 ) Blockquote.__proto__ = TextBlock$$1;
  	Blockquote.prototype = Object.create( TextBlock$$1 && TextBlock$$1.prototype );
  	Blockquote.prototype.constructor = Blockquote;

  	

  	return Blockquote;
  }(TextBlock));

  Blockquote.type = "blockquote";

  var BlockquoteComponent = (function (TextBlockComponent$$1) {
    function BlockquoteComponent () {
      TextBlockComponent$$1.apply(this, arguments);
    }

    if ( TextBlockComponent$$1 ) BlockquoteComponent.__proto__ = TextBlockComponent$$1;
    BlockquoteComponent.prototype = Object.create( TextBlockComponent$$1 && TextBlockComponent$$1.prototype );
    BlockquoteComponent.prototype.constructor = BlockquoteComponent;

    BlockquoteComponent.prototype.render = function render ($$) {
      var el = TextBlockComponent$$1.prototype.render.call(this, $$);
      return el.addClass('sc-blockquote')
    };

    return BlockquoteComponent;
  }(TextBlockComponent));

  var BlockquoteHTMLConverter = {

    type: 'blockquote',
    tagName: 'blockquote',

    import: function(el, node, converter) {
      node.content = converter.annotatedText(el, [node.id, 'content']);
    },

    export: function(node, el, converter) {
      el.append(
        converter.annotatedText([node.id, 'content'])
      );
    }

  }

  var BlockquotePackage = {
    name: 'blockquote',
    configure: function(config) {
      config.addNode(Blockquote);
      config.addComponent(Blockquote.type, BlockquoteComponent);
      config.addConverter('html', BlockquoteHTMLConverter);
      config.addConverter('xml', BlockquoteHTMLConverter);
      config.addCommand('blockquote', SwitchTextTypeCommand, {
        spec: { type: 'blockquote' },
        commandGroup: 'text-types'
      });
      config.addIcon('blockquote', { 'fontawesome': 'fa-quote-right' });
      config.addLabel('blockquote', {
        en: 'Blockquote',
        de: 'Blockzitat'
      });
      config.addKeyboardShortcut('CommandOrControl+Alt+B', { command: 'blockquote' });
    },
    Blockquote: Blockquote,
    BlockquoteComponent: BlockquoteComponent,
    BlockquoteHTMLConverter: BlockquoteHTMLConverter
  }

  var Code = (function (PropertyAnnotation$$1) {
  	function Code () {
  		PropertyAnnotation$$1.apply(this, arguments);
  	}if ( PropertyAnnotation$$1 ) Code.__proto__ = PropertyAnnotation$$1;
  	Code.prototype = Object.create( PropertyAnnotation$$1 && PropertyAnnotation$$1.prototype );
  	Code.prototype.constructor = Code;

  	

  	return Code;
  }(PropertyAnnotation));

  Code.type = 'code';

  var CodeHTMLConverter = {
    type: 'code',
    tagName: 'code'
  }

  var CodePackage = {
    name: 'code',
    configure: function(config) {
      config.addNode(Code);
      config.addConverter('html', CodeHTMLConverter);
      config.addConverter('xml', CodeHTMLConverter);
      config.addComponent('code', AnnotationComponent);
      config.addCommand('code', AnnotationCommand, {
        nodeType: Code.type,
        commandGroup: 'annotations'
      });
      config.addIcon('code', { 'fontawesome': 'fa-code' });
      config.addLabel('code', {
        en: 'Code',
        de: 'Code'
      });
    },
    Code: Code,
    CodeHTMLConverter: CodeHTMLConverter
  }

  var Codeblock = (function (TextBlock$$1) {
  	function Codeblock () {
  		TextBlock$$1.apply(this, arguments);
  	}if ( TextBlock$$1 ) Codeblock.__proto__ = TextBlock$$1;
  	Codeblock.prototype = Object.create( TextBlock$$1 && TextBlock$$1.prototype );
  	Codeblock.prototype.constructor = Codeblock;

  	

  	return Codeblock;
  }(TextBlock));

  Codeblock.type = "codeblock";

  var CodeblockComponent = (function (TextBlockComponent$$1) {
    function CodeblockComponent () {
      TextBlockComponent$$1.apply(this, arguments);
    }

    if ( TextBlockComponent$$1 ) CodeblockComponent.__proto__ = TextBlockComponent$$1;
    CodeblockComponent.prototype = Object.create( TextBlockComponent$$1 && TextBlockComponent$$1.prototype );
    CodeblockComponent.prototype.constructor = CodeblockComponent;

    CodeblockComponent.prototype.render = function render ($$) {
      var el = TextBlockComponent$$1.prototype.render.call(this, $$);
      return el.addClass('sc-codeblock')
    };

    return CodeblockComponent;
  }(TextBlockComponent));

  var CodeblockHTMLConverter = {

    type: 'codeblock',
    tagName: 'pre',

    import: function(el, node, converter) {
      var codeEl = el.find('code');
      if (codeEl) {
        node.content = converter.annotatedText(codeEl, [node.id, 'content'], { preserveWhitespace: true });
      }
    },

    export: function(node, el, converter) {
      var $$ = converter.$$;
      el.append(
        $$('code').append(
          converter.annotatedText([node.id, 'content'])
        )
      );
    }
  }

  var CodeblockPackage = {
    name: 'codeblock',
    configure: function(config) {
      config.addNode(Codeblock);
      config.addComponent('codeblock', CodeblockComponent);
      config.addConverter('html', CodeblockHTMLConverter);
      config.addConverter('xml', CodeblockHTMLConverter);
      config.addCommand('codeblock', SwitchTextTypeCommand, {
        spec: { type: 'codeblock' },
        commandGroup: 'text-types'
      });
      config.addIcon('codeblock', { 'fontawesome': 'fa-quote-right' });
      config.addLabel('codeblock', {
        en: 'Codeblock',
        de: 'Codeblock'
      });
      config.addKeyboardShortcut('CommandOrControl+Alt+C', { command: 'codeblock' });
    },
    Codeblock: Codeblock,
    CodeblockComponent: CodeblockComponent,
    CodeblockHTMLConverter: CodeblockHTMLConverter
  }

  var Emphasis = (function (PropertyAnnotation$$1) {
  	function Emphasis () {
  		PropertyAnnotation$$1.apply(this, arguments);
  	}if ( PropertyAnnotation$$1 ) Emphasis.__proto__ = PropertyAnnotation$$1;
  	Emphasis.prototype = Object.create( PropertyAnnotation$$1 && PropertyAnnotation$$1.prototype );
  	Emphasis.prototype.constructor = Emphasis;

  	

  	return Emphasis;
  }(PropertyAnnotation));

  Emphasis.type = "emphasis";


  Emphasis.fragmentation = Fragmenter.ANY;

  var EmphasisHTMLConverter = {

    type: 'emphasis',
    tagName: 'em',

    matchElement: function(el) {
      return el.is('em, i')
    }

  }

  var EmphasisComponent = (function (AnnotationComponent$$1) {
    function EmphasisComponent () {
      AnnotationComponent$$1.apply(this, arguments);
    }

    if ( AnnotationComponent$$1 ) EmphasisComponent.__proto__ = AnnotationComponent$$1;
    EmphasisComponent.prototype = Object.create( AnnotationComponent$$1 && AnnotationComponent$$1.prototype );
    EmphasisComponent.prototype.constructor = EmphasisComponent;

    EmphasisComponent.prototype.getTagName = function getTagName () {
      return 'em'
    };

    return EmphasisComponent;
  }(AnnotationComponent));

  var EmphasisPackage = {
    name: 'emphasis',
    configure: function(config) {
      config.addNode(Emphasis);
      config.addConverter('html', EmphasisHTMLConverter);
      config.addConverter('xml', EmphasisHTMLConverter);
      config.addComponent('emphasis', EmphasisComponent);
      config.addCommand('emphasis', AnnotationCommand, {
        nodeType: Emphasis.type,
        commandGroup: 'annotations'
      });
      config.addIcon('emphasis', { 'fontawesome': 'fa-italic' });
      config.addLabel('emphasis', {
        en: 'Emphasis',
        de: 'Betonung'
      });
      config.addKeyboardShortcut('CommandOrControl+I', { command: 'emphasis' });
    },
    Emphasis: Emphasis,
    EmphasisComponent: EmphasisComponent,
    EmphasisHTMLConverter: EmphasisHTMLConverter
  }

  var FindAndReplaceCommand = (function (Command$$1) {
    function FindAndReplaceCommand () {
      Command$$1.apply(this, arguments);
    }

    if ( Command$$1 ) FindAndReplaceCommand.__proto__ = Command$$1;
    FindAndReplaceCommand.prototype = Object.create( Command$$1 && Command$$1.prototype );
    FindAndReplaceCommand.prototype.constructor = FindAndReplaceCommand;

    FindAndReplaceCommand.prototype.getCommandState = function getCommandState (ref) {
      var editorSession = ref.editorSession;

      var findAndReplaceManager = editorSession.getManager('find-and-replace');
      var findAndReplaceState = findAndReplaceManager.getCommandState();
      return findAndReplaceState
    };

    FindAndReplaceCommand.prototype.execute = function execute () {
      
    };

    return FindAndReplaceCommand;
  }(Command));

  var ToggleFindAndReplaceCommand = (function (Command$$1) {
    function ToggleFindAndReplaceCommand () {
      Command$$1.apply(this, arguments);
    }

    if ( Command$$1 ) ToggleFindAndReplaceCommand.__proto__ = Command$$1;
    ToggleFindAndReplaceCommand.prototype = Object.create( Command$$1 && Command$$1.prototype );
    ToggleFindAndReplaceCommand.prototype.constructor = ToggleFindAndReplaceCommand;

    ToggleFindAndReplaceCommand.prototype.getCommandState = function getCommandState () {
      return {
        disabled: false
      }
    };

    ToggleFindAndReplaceCommand.prototype.execute = function execute (ref) {
      var editorSession = ref.editorSession;

      var findAndReplaceManager = editorSession.getManager('find-and-replace');
      var surface = editorSession.getFocusedSurface();
      if (surface) {
        findAndReplaceManager.enable();
      }
    };

    return ToggleFindAndReplaceCommand;
  }(Command));

  var ToggleFindAndReplaceCommand$1 = (function (Command$$1) {
    function ToggleFindAndReplaceCommand () {
      Command$$1.apply(this, arguments);
    }

    if ( Command$$1 ) ToggleFindAndReplaceCommand.__proto__ = Command$$1;
    ToggleFindAndReplaceCommand.prototype = Object.create( Command$$1 && Command$$1.prototype );
    ToggleFindAndReplaceCommand.prototype.constructor = ToggleFindAndReplaceCommand;

    ToggleFindAndReplaceCommand.prototype.getCommandState = function getCommandState (ref) {
      var editorSession = ref.editorSession;

      var findAndReplaceManager = editorSession.getManager('find-and-replace');
      var state = findAndReplaceManager.getCommandState();
      return {
        disabled: state.disabled
      }
    };

    ToggleFindAndReplaceCommand.prototype.execute = function execute (ref) {
      var editorSession = ref.editorSession;

      var findAndReplaceManager = editorSession.getManager('find-and-replace');
      var findAndReplaceState = findAndReplaceManager.getCommandState();
      if (!findAndReplaceState.disabled) {
        findAndReplaceManager.disable();
      }
    };

    return ToggleFindAndReplaceCommand;
  }(Command));

  var FindNextCommand = (function (Command$$1) {
    function FindNextCommand () {
      Command$$1.apply(this, arguments);
    }

    if ( Command$$1 ) FindNextCommand.__proto__ = Command$$1;
    FindNextCommand.prototype = Object.create( Command$$1 && Command$$1.prototype );
    FindNextCommand.prototype.constructor = FindNextCommand;

    FindNextCommand.prototype.getCommandState = function getCommandState () {
      return {
        disabled: false
      }
    };

    FindNextCommand.prototype.execute = function execute (ref) {
      var editorSession = ref.editorSession;

      var findAndReplaceManager = editorSession.getManager('find-and-replace');
      findAndReplaceManager.findNext();
    };

    return FindNextCommand;
  }(Command));

  var FindPreviousCommand = (function (Command$$1) {
    function FindPreviousCommand () {
      Command$$1.apply(this, arguments);
    }

    if ( Command$$1 ) FindPreviousCommand.__proto__ = Command$$1;
    FindPreviousCommand.prototype = Object.create( Command$$1 && Command$$1.prototype );
    FindPreviousCommand.prototype.constructor = FindPreviousCommand;

    FindPreviousCommand.prototype.getCommandState = function getCommandState (ref) {
      var editorSession = ref.editorSession;

      var findAndReplaceManager = editorSession.getManager('find-and-replace');
      var findAndReplaceState = findAndReplaceManager.getCommandState();
      return findAndReplaceState
    };

    FindPreviousCommand.prototype.execute = function execute (ref) {
      var editorSession = ref.editorSession;

      var findAndReplaceManager = editorSession.getManager('find-and-replace');
      findAndReplaceManager.findPrevious();
    };

    return FindPreviousCommand;
  }(Command));

  var ReplaceNextCommand = (function (Command$$1) {
    function ReplaceNextCommand () {
      Command$$1.apply(this, arguments);
    }

    if ( Command$$1 ) ReplaceNextCommand.__proto__ = Command$$1;
    ReplaceNextCommand.prototype = Object.create( Command$$1 && Command$$1.prototype );
    ReplaceNextCommand.prototype.constructor = ReplaceNextCommand;

    ReplaceNextCommand.prototype.getCommandState = function getCommandState (ref) {
      var editorSession = ref.editorSession;

      var findAndReplaceManager = editorSession.getManager('find-and-replace');
      var findAndReplaceState = findAndReplaceManager.getCommandState();
      return findAndReplaceState
    };

    ReplaceNextCommand.prototype.execute = function execute (ref) {
      var editorSession = ref.editorSession;

      var findAndReplaceManager = editorSession.getManager('find-and-replace');
      findAndReplaceManager.replaceNext();
    };

    return ReplaceNextCommand;
  }(Command));

  var ReplaceAllCommand = (function (Command$$1) {
    function ReplaceAllCommand () {
      Command$$1.apply(this, arguments);
    }

    if ( Command$$1 ) ReplaceAllCommand.__proto__ = Command$$1;
    ReplaceAllCommand.prototype = Object.create( Command$$1 && Command$$1.prototype );
    ReplaceAllCommand.prototype.constructor = ReplaceAllCommand;

    ReplaceAllCommand.prototype.getCommandState = function getCommandState (ref) {
      var editorSession = ref.editorSession;

      var findAndReplaceManager = editorSession.getManager('find-and-replace');
      var findAndReplaceState = findAndReplaceManager.getCommandState();
      return findAndReplaceState
    };

    ReplaceAllCommand.prototype.execute = function execute (ref) {
      var editorSession = ref.editorSession;

      var findAndReplaceManager = editorSession.getManager('find-and-replace');
      findAndReplaceManager.replaceAll();
    };

    return ReplaceAllCommand;
  }(Command));

  var FindAndReplaceTool = (function (ToggleTool$$1) {
    function FindAndReplaceTool () {
      ToggleTool$$1.apply(this, arguments);
    }

    if ( ToggleTool$$1 ) FindAndReplaceTool.__proto__ = ToggleTool$$1;
    FindAndReplaceTool.prototype = Object.create( ToggleTool$$1 && ToggleTool$$1.prototype );
    FindAndReplaceTool.prototype.constructor = FindAndReplaceTool;

    FindAndReplaceTool.prototype.didMount = function didMount () {
      this.context.editorSession.onPosition(this._onPosition, this);
    };

    FindAndReplaceTool.prototype.dispose = function dispose () {
      this.context.editorSession.off(this);
    };

    FindAndReplaceTool.prototype._onPosition = function _onPosition () {
      var this$1 = this;

      var findAndReplaceManager = this.context.editorSession.getManager('find-and-replace');
      
      
      
      if (findAndReplaceManager._requestLookupMatch) {
        this._scrollToSelectedMatch();
        findAndReplaceManager._requestLookupMatch = false;
      }

      if (findAndReplaceManager._requestFocusSearchString) {
        setTimeout(function () {
          this$1.refs.findString.el.focus();
          this$1.refs.findString.el.select();
          findAndReplaceManager._requestFocusSearchString = false;
        },0);
      }
    };

    FindAndReplaceTool.prototype.render = function render ($$) {
      var commandState = this.props.commandState;
      var el = $$('div').addClass('sc-find-and-replace-tool');

      el.append(
        this._renderStatusDescription($$),
        
        
        
        
        
        
        
        
        
        
        
        $$('div').addClass('se-section').append(
          $$('div')
            .addClass('se-section-item se-find-input')
            .addClass('sm-flex')
            .append(
              $$('input')
                .ref('findString')
                .attr('type', 'text')
                .attr('placeholder', 'Find in body')
                .attr('tabindex', 500)
                .val(commandState.findString)
                .on('input', debounce(this._startFind, 300))
                .on('keypress', this._triggerFindNext)
                .on('focus', this._onFocus)
                .on('blur', this._onBlur),
              this._renderStatusCounter($$)
            ),
          $$('div')
            .addClass('se-section-item se-replace-input')
            .append(
              $$('button')
                .append('Find')
                .attr('tabindex', 502)
                .on('click', this._findNext)
            ),
          $$('div')
            .addClass('se-section-item')
            
            
            
        ),
        $$('div').addClass('se-section').append(
          $$('div')
            .addClass('se-section-item')
            .addClass('sm-flex')
            .append(
              $$('input')
                .ref('replaceString')
                .val(commandState.replaceString)
                .attr('type', 'text')
                .attr('tabindex', 501)
                .attr('placeholder', 'Replace in body')
                .on('focus', this._onFocus)
                .on('blur', this._onBlur)
                .on('input', this._setReplaceString)
                .on('keypress', this._triggerReplace)
            ),
          $$('div')
            .addClass('se-section-item')
            .append(
              $$('button')
                .append('Replace')
                .attr('tabindex', 503)
                .on('click', this._replaceNext)
            ),
          $$('div')
            .addClass('se-section-item')
            .append(
              $$('button')
                .append('Replace All')
                .attr('tabindex', 504)
                .on('click', this._replaceAll)
            )
        )
      );
      return el
    };

    FindAndReplaceTool.prototype._renderStatusDescription = function _renderStatusDescription ($$) {
      var commandState = this.props.commandState;
      var statusDescriptionEl = $$('div').addClass('se-status').append(
        $$('div').addClass('se-status-title').append(
          this.getLabel('find-and-replace-title')
        )
      );

      if (commandState.totalMatches > 0) {
        statusDescriptionEl.append(
          $$('div').addClass('se-status-description').append(
            commandState.totalMatches,
            ' results found for ',
            '"'+ commandState.findString +'"'
          )
        );
      } else if (commandState.findString !== '') {
        statusDescriptionEl.append(
          $$('div').addClass('se-status-description').append(
            'No results found for ',
            '"'+ commandState.findString +'"'
          )
        );
      } else {
        statusDescriptionEl.append(
          $$('div').addClass('se-status-description').append(
            'Close this panel with ESC key'
          )
        );
      }
      return statusDescriptionEl
    };

    
    FindAndReplaceTool.prototype._onFocus = function _onFocus () {
      var editorSession = this.context.editorSession;
      editorSession.setBlurred(true);
      
      editorSession.startFlow();
    };

    
    FindAndReplaceTool.prototype._onBlur = function _onBlur () {
      var editorSession = this.context.editorSession;
      editorSession.setBlurred(false);
    };

    FindAndReplaceTool.prototype._renderStatusCounter = function _renderStatusCounter ($$) {
      var commandState = this.props.commandState;
      var statusCounterEl;

      if (commandState.totalMatches > 0) {
        statusCounterEl = $$('span').addClass('se-status-counter').append(
          [commandState.selectedMatch, commandState.totalMatches].join(' of ')
        );
      }

      return statusCounterEl
    };

    FindAndReplaceTool.prototype._findNext = function _findNext () {
      var findAndReplaceManager = this.context.editorSession.getManager('find-and-replace');
      findAndReplaceManager.findNext();
    };

    FindAndReplaceTool.prototype._replaceNext = function _replaceNext () {
      var findAndReplaceManager = this.context.editorSession.getManager('find-and-replace');
      findAndReplaceManager.replaceNext();
    };

    FindAndReplaceTool.prototype._replaceAll = function _replaceAll () {
      var findAndReplaceManager = this.context.editorSession.getManager('find-and-replace');
      findAndReplaceManager.replaceAll();
    };


    

    FindAndReplaceTool.prototype.findStringHasChanged = function findStringHasChanged () {
      var findString = this.refs.findString.val();
      var previousFindString = this._previousFindString;
      this._previousFindString = findString;
      return findString !== previousFindString
    };

    
    FindAndReplaceTool.prototype._startFind = function _startFind () {
      var editorSession = this.context.editorSession;
      var findString = this.refs.findString.val();
      var findAndReplaceManager = editorSession.getManager('find-and-replace');
      findAndReplaceManager.startFind(findString);
    };

    FindAndReplaceTool.prototype._scrollToSelectedMatch = function _scrollToSelectedMatch () {
      var editorSession = this.context.editorSession;
      var surface = editorSession.getFocusedSurface();
      surface.context.scrollPane.scrollTo('.sc-selected-match', 'onlyIfNotVisible');
    };

    FindAndReplaceTool.prototype._triggerFindNext = function _triggerFindNext (e) {
      if (e.keyCode === 13) {
        this._findNext();
      }
    };

    FindAndReplaceTool.prototype._triggerReplace = function _triggerReplace (e) {
      if (e.keyCode === 13) {
        this._replaceNext();
      }
    };

    
    FindAndReplaceTool.prototype._setReplaceString = function _setReplaceString () {
      var replaceString = this.refs.replaceString.val();
      var findAndReplaceManager = this.context.editorSession.getManager('find-and-replace');
      findAndReplaceManager.setReplaceString(replaceString);
    };

    return FindAndReplaceTool;
  }(ToggleTool));

  var FindAndReplaceManager = function FindAndReplaceManager(context) {
    if (!context.editorSession) {
      throw new Error('EditorSession required.')
    }

    this.editorSession = context.editorSession;
    this.editorSession.onRender('document', this._onDocumentChanged, this);

    this.doc = this.editorSession.getDocument();
    this.context = Object.assign({}, context, {
      
      doc: this.doc
    });

    this._state = {
      disabled: true,
      findString: '',
      replaceString: '',
      
      matches: [],
      selectedMatch: undefined
    };

    
    this._requestLookupMatch = false;
    
    this._requestFocusSearchString = false;
  };

  FindAndReplaceManager.prototype.dispose = function dispose () {
    this.editorSession.off(this);
  };


  FindAndReplaceManager.prototype._resetState = function _resetState () {
    this._state.disabled = true;
    this._state.matches = [];
    this._state.selectedMatch = undefined;
  };


  FindAndReplaceManager.prototype.getCommandState = function getCommandState () {
    var state = this._state;
    var commandState = {
      disabled: state.disabled,
      findString: state.findString,
      replaceString: state.replaceString,
      
      totalMatches: state.matches.length,
      selectedMatch: state.selectedMatch + 1
    };
    return commandState
  };

  FindAndReplaceManager.prototype.enable = function enable () {
    this._state.disabled = false;
    this._requestFocusSearchString = true;
    
    this.startFind(this._state.findString);
  };

  FindAndReplaceManager.prototype.disable = function disable () {
    this._state.disabled = true;
    this._resetState();
    this._propagateUpdate();
  };

  FindAndReplaceManager.prototype._onDocumentChanged = function _onDocumentChanged () {
    if (!this._state.disabled) {
      this._computeMatches();
      this._state.selectedMatch = 0;
      this._updateMarkers();
    }
  };


  FindAndReplaceManager.prototype.startFind = function startFind (findString) {
    this._state.findString = findString;
    this._computeMatches();
    var closestMatch = this._getClosestMatch();
    this._state.selectedMatch = closestMatch > 0 ? closestMatch : 0;
    this._requestLookupMatch = true;
    this._setSelection();
    this._propagateUpdate();

  };

  FindAndReplaceManager.prototype.setReplaceString = function setReplaceString (replaceString) {
    
    this._state.replaceString = replaceString;
  };


  FindAndReplaceManager.prototype.findNext = function findNext () {
    var index = this._state.selectedMatch;
    var totalMatches = this._state.matches.length;
    if (totalMatches === 0) { return }
    this._state.selectedMatch = (index + 1) % totalMatches;
    this._requestLookupMatch = true;
    this._setSelection();
    this._propagateUpdate();
  };


  FindAndReplaceManager.prototype.findPrevious = function findPrevious () {
    var index = this._state.selectedMatch;
    var totalMatches = this._state.matches.length;
    if (totalMatches === 0) { return }
    this._state.selectedMatch = index > 0 ? index - 1 : totalMatches - 1;
    this._requestLookupMatch = true;
    this._setSelection();
    this._propagateUpdate();
  };

  FindAndReplaceManager.prototype._setSelection = function _setSelection () {
    var match = this._state.matches[this._state.selectedMatch];
    if (!match) { return }
    
    
    var sel = match.getSelection();
    this.editorSession.setSelection(sel, 'skipFlow');
  };


  FindAndReplaceManager.prototype.replaceNext = function replaceNext () {
      var this$1 = this;

    var index = this._state.selectedMatch;
    var match = this._state.matches[index];
    var totalMatches = this._state.matches.length;
    if(match !== undefined) {
      this.editorSession.transaction(function (tx, args) {
        tx.setSelection(match.getSelection());
        tx.insertText(this$1._state.replaceString);
        return args
      });
      this._computeMatches();
      if(index + 1 < totalMatches) {
        this._state.selectedMatch = index;
      }
      this._requestLookupMatch = true;
      this._setSelection();
      this._propagateUpdate();
    }
  };


  FindAndReplaceManager.prototype.replaceAll = function replaceAll () {
      var this$1 = this;

    
    
    var matches = this._state.matches.reverse();

    this.editorSession.transaction(function (tx, args) {
      matches.forEach(function (match) {
        tx.setSelection(match.getSelection());
        tx.insertText(this$1._state.replaceString);
      });
      return args
    });

    this._computeMatches();
  };


  FindAndReplaceManager.prototype._getClosestMatch = function _getClosestMatch () {
    var doc = this.editorSession.getDocument();
    var nodeIds = Object.keys(doc.getNodes());
    var sel = this.editorSession.getSelection();
    var closest = 0;

    if(!sel.isNull()) {
      var startOffset = sel.start.offset;
      var selStartNode = sel.start.path[0];
      var selStartNodePos = nodeIds.indexOf(selStartNode);
      var matches = this._state.matches;

      closest = matches.findIndex(function (match) {
        var markerSel = match.getSelection();
        var markerStartOffset = markerSel.start.offset;
        var markerStartNode = markerSel.start.path[0];
        var markerStartNodePos = nodeIds.indexOf(markerStartNode);
        if(selStartNodePos > markerStartNodePos) {
          return false
        } else if (selStartNodePos < markerStartNodePos) {
          return true
        } else {
          if(startOffset <= markerStartOffset) {
            return true
          } else {
            return false
          }
        }
      });
    }

    return closest
  };

  FindAndReplaceManager.prototype._computeMatches = function _computeMatches () {
    var currentMatches = this._state.matches;
    var currentTotal = currentMatches === undefined ? 0 : currentMatches.length;

    this._state.matches = this._findAllMatches();

    
    
    
    var newMatches = this._state.matches;

    if(newMatches.length !== currentTotal) {
      this._state.selectedMatch = newMatches.length > 0 ? 0 : undefined;
    }
  };



  FindAndReplaceManager.prototype._findAllMatches = function _findAllMatches () {
      var this$1 = this;

    var textProperties = this._getAllAffectedTextPropertiesInOrder();
    var pattern = this._state.findString;
    var matches = [];
    if (pattern) {
      textProperties.forEach(function (textPropertyPath) {
        var found = this$1._findInTextProperty({
          path: textPropertyPath,
          findString: pattern
        });
        matches = matches.concat(found);
      });
    }
    return matches
  };


  FindAndReplaceManager.prototype._getAllAffectedTextPropertiesInOrder = function _getAllAffectedTextPropertiesInOrder () {
    var textProperties = [];
    var editor = this.editorSession.getEditor();
    
    var rootElement = editor.find(this.getRootElementSelector()) || editor;
    textProperties = rootElement.findAll('.sc-text-property').map(function (tpc) {
      return tpc.props.path
    });
    return textProperties
  };


  FindAndReplaceManager.prototype._findInTextProperty = function _findInTextProperty (ref) {
      var path = ref.path;
      var findString = ref.findString;

    var doc = this.doc;
    var text = doc.get(path);

    
    var matcher = new RegExp(findString, 'ig');
    var matches = [];
    var match;

    while ((match = matcher.exec(text))) {
      var marker = new Marker(doc, {
        type: 'match',
        start: {
          path: path,
          offset: match.index
        },
        end: {
          offset: matcher.lastIndex
        }
      });
      matches.push(marker);
    }
    return matches
  };

  FindAndReplaceManager.prototype._propagateUpdate = function _propagateUpdate () {
    
    this._updateMarkers();
    this.editorSession._setDirty('commandStates');
    this.editorSession.startFlow();
  };

  FindAndReplaceManager.prototype._updateMarkers = function _updateMarkers () {
    var state = this._state;
    var editorSession = this.editorSession;
    var markersManager = editorSession.markersManager;
    state.matches.forEach(function (m, idx) {
      m.type = (idx === state.selectedMatch) ? 'selected-match' : 'match';
    });
    
    markersManager.setMarkers('find-and-replace', state.matches);
  };


  FindAndReplaceManager.prototype.getRootElementSelector = function getRootElementSelector () {
    var configurator = this.editorSession.getConfigurator();
    var config = configurator.getFindAndReplaceConfig();
    return config.rootElement
  };

  var FindAndReplacePackage = {
    name: 'find-and-replace',
    configure: function(config, userConfig) {
      config.addCommand('find-and-replace', FindAndReplaceCommand, {
        commandGroup: 'workflows'
      });
      config.addCommand('toggle-find-and-replace', ToggleFindAndReplaceCommand, {
        commandGroup: 'find-and-replace'
      });
      config.addCommand('close-find-and-replace', ToggleFindAndReplaceCommand$1, {
        commandGroup: 'find-and-replace'
      });
      config.addCommand('find-next', FindNextCommand, {
        commandGroup: 'find-and-replace'
      });
      config.addCommand('find-previous', FindPreviousCommand, {
        commandGroup: 'find-and-replace'
      });
      config.addCommand('replace-next', ReplaceNextCommand, {
        commandGroup: 'find-and-replace'
      });
      config.addCommand('replace-all', ReplaceAllCommand, {
        commandGroup: 'find-and-replace'
      });
      config.addTool('find-and-replace', FindAndReplaceTool);
      config.addKeyboardShortcut('CommandOrControl+F', { command: 'toggle-find-and-replace' });
      config.addKeyboardShortcut('CommandOrControl+Alt+F', { command: 'toggle-find-and-replace' });
      config.addKeyboardShortcut('CommandOrControl+G', { command: 'find-next' });
      config.addKeyboardShortcut('CommandOrControl+Shift+G', { command: 'find-previous' });
      config.addKeyboardShortcut('CommandOrControl+Alt+E', { command: 'replace-next' });
      
      config.addKeyboardShortcut('esc', { command: 'close-find-and-replace' });
      config.addManager('find-and-replace', FindAndReplaceManager);
      config.addLabel('find-and-replace-title', {
        en: 'Find and replace',
        de: 'Suchen und Ersetzen'
      });
      config.setFindAndReplaceConfig(userConfig);
    },
    FindAndReplaceCommand: FindAndReplaceCommand,
    ToggleFindAndReplaceCommand: ToggleFindAndReplaceCommand,
    FindNextCommand: FindNextCommand,
    FindPreviousCommand: FindPreviousCommand,
    ReplaceNextCommand: ReplaceNextCommand,
    ReplaceAllCommand: ReplaceAllCommand,
    FindAndReplaceTool: FindAndReplaceTool,
    FindAndReplaceManager: FindAndReplaceManager
  }

  var Heading = (function (TextBlock$$1) {
    function Heading () {
      TextBlock$$1.apply(this, arguments);
    }

    if ( TextBlock$$1 ) Heading.__proto__ = TextBlock$$1;
    Heading.prototype = Object.create( TextBlock$$1 && TextBlock$$1.prototype );
    Heading.prototype.constructor = Heading;

    Heading.prototype.isHeading = function isHeading () {
      return true
    };

    Heading.prototype.getLevel = function getLevel () {
      return this.level
    };

    return Heading;
  }(TextBlock));

  Heading.schema = {
    type: "heading",
    level: { type: "number", default: 1 }
  };

  var HeadingComponent = (function (TextBlockComponent$$1) {
    function HeadingComponent () {
      TextBlockComponent$$1.apply(this, arguments);
    }

    if ( TextBlockComponent$$1 ) HeadingComponent.__proto__ = TextBlockComponent$$1;
    HeadingComponent.prototype = Object.create( TextBlockComponent$$1 && TextBlockComponent$$1.prototype );
    HeadingComponent.prototype.constructor = HeadingComponent;

    HeadingComponent.prototype.render = function render ($$) {
      var el = TextBlockComponent$$1.prototype.render.call(this, $$);
      return el.addClass("sc-heading sm-level-"+this.props.node.level)
    };

    HeadingComponent.prototype.getTagName = function getTagName () {
      return 'h'+this.props.node.level
    };

    return HeadingComponent;
  }(TextBlockComponent));

  var HeadingMacro = {

    appliesTo: ['paragraph'],

    execute: function(props) {
      if (this.appliesTo.indexOf(props.node.type) === -1) {
        return false
      }
      if (props.action !== 'type') {
        return false
      }
      var match = /^#\s/.exec(props.text);
      if (match) {
        
        var editorSession = props.editorSession;
        var sel = editorSession.getSelection();
        var path = sel.start.path;
        var startOffset = sel.start.offset;
        editorSession.transaction(function(tx) {
          tx.setSelection({
            type: 'property',
            path: path,
            startOffset: 0,
            endOffset: match[0].length
          });
          tx.deleteSelection();
          var node = tx.switchTextType({
            type: 'heading',
            level: 1
          });
          tx.setSelection({
            type: 'property',
            path: node.getPath(),
            startOffset: startOffset - match[0].length
          });
        });
        return true
      }
    }
  }

  var HeadingHTMLConverter = {

    type: "heading",

    matchElement: function(el) {
      return /^h\d$/.exec(el.tagName)
    },

    import: function(el, node, converter) {
      node.level = Number(el.tagName[1]);
      var textAlign = el.attr('data-text-align');
      node.textAlign = textAlign || 'left';
      node.content = converter.annotatedText(el, [node.id, 'content']);
    },

    export: function(node, el, converter) {
      el.tagName = 'h'+node.level;
      
      if (node.textAlign !== 'left') {
        el.attr('data-text-align', node.textAlign);
      }
      el.append(
        converter.annotatedText([node.id, 'content'])
      );
    }

  }

  var HeadingPackage = {
    name: 'heading',
    configure: function(config) {
      config.addNode(Heading);
      config.addComponent(Heading.type, HeadingComponent);
      config.addConverter('html', HeadingHTMLConverter);
      config.addConverter('xml', HeadingHTMLConverter);

      config.addCommand('heading1', SwitchTextTypeCommand, {
        spec: { type: 'heading', level: 1 },
        commandGroup: 'text-types'
      });
      config.addCommand('heading2', SwitchTextTypeCommand, {
        spec: { type: 'heading', level: 2 },
        commandGroup: 'text-types'
      });
      config.addCommand('heading3', SwitchTextTypeCommand, {
        spec: { type: 'heading', level: 3 },
        commandGroup: 'text-types'
      });
      config.addKeyboardShortcut('CommandOrControl+Alt+1', { command: 'heading1' });
      config.addKeyboardShortcut('CommandOrControl+Alt+2', { command: 'heading2' });
      config.addKeyboardShortcut('CommandOrControl+Alt+3', { command: 'heading3' });

      config.addLabel('heading1', {
        en: 'Heading 1',
        de: 'Überschrift 1'
      });
      config.addLabel('heading2', {
        en: 'Heading 2',
        de: 'Überschrift 2'
      });
      config.addLabel('heading3', {
        en: 'Heading 3',
        de: 'Überschrift 3'
      });
      config.addIcon('heading1', { 'fontawesome': 'fa-header' });
      config.addIcon('heading2', { 'fontawesome': 'fa-header' });
      config.addIcon('heading3', { 'fontawesome': 'fa-header' });
    },
    Heading: Heading,
    HeadingComponent: HeadingComponent,
    HeadingHTMLConverter: HeadingHTMLConverter,
    HeadingMacro: HeadingMacro
  }

  var ImageNode = (function (DocumentNode$$1) {
    function ImageNode () {
      DocumentNode$$1.apply(this, arguments);
    }

    if ( DocumentNode$$1 ) ImageNode.__proto__ = DocumentNode$$1;
    ImageNode.prototype = Object.create( DocumentNode$$1 && DocumentNode$$1.prototype );
    ImageNode.prototype.constructor = ImageNode;

    ImageNode.prototype.getImageFile = function getImageFile () {
      if (this.imageFile) {
        return this.document.get(this.imageFile)
      }
    };

    ImageNode.prototype.getUrl = function getUrl () {
      var imageFile = this.getImageFile();
      if (imageFile) {
        return imageFile.getUrl()
      }
    };

    return ImageNode;
  }(DocumentNode));

  ImageNode.schema = {
    type: 'image',
    imageFile: { type: 'file' }
  };

  var ImageComponent = (function (NodeComponent$$1) {
    function ImageComponent () {
      NodeComponent$$1.apply(this, arguments);
    }

    if ( NodeComponent$$1 ) ImageComponent.__proto__ = NodeComponent$$1;
    ImageComponent.prototype = Object.create( NodeComponent$$1 && NodeComponent$$1.prototype );
    ImageComponent.prototype.constructor = ImageComponent;

    ImageComponent.prototype.didMount = function didMount () {
      NodeComponent$$1.prototype.didMount.call(this);
      this.context.editorSession.onRender('document', this._onDocumentChange, this);
    };

    ImageComponent.prototype.dispose = function dispose () {
      NodeComponent$$1.prototype.dispose.call(this);
      this.context.editorSession.off(this);
    };

    
    ImageComponent.prototype._onDocumentChange = function _onDocumentChange (change) {
      if (change.hasUpdated(this.props.node.id) ||
        change.hasUpdated(this.props.node.imageFile)) {
        this.rerender();
      }
    };

    ImageComponent.prototype.render = function render ($$) {
      var el = NodeComponent$$1.prototype.render.call(this, $$);
      el.addClass('sc-image');
      el.append(
        $$('img').attr({
          src: this.props.node.getUrl(),
        }).ref('image')
      );
      return el
    };

    
    ImageComponent.prototype.getDropzoneSpecs = function getDropzoneSpecs () {
      return [
        {
          component: this.refs['image'],
          message: 'Replace Image',
          dropParams: {
            action: 'replace-image',
            nodeId: this.props.node.id,
          }
        }
      ]
    };

    ImageComponent.prototype.handleDrop = function handleDrop (tx, dragState) {
      var newImageFile = dragState.data.files[0];
      if (dragState.external) {
        var imageFile = tx.create({
          type: 'file',
          fileType: 'image',
          mimeType: newImageFile.type,
          url: URL.createObjectURL(newImageFile)
        });
        
        
        tx.set([this.props.node.id, 'imageFile'], imageFile.id);
      } else {
        var nodeId = dragState.sourceSelection.nodeId;
        var node = tx.get(nodeId);
        if (node.type === 'image') {
          
          tx.set([this.props.node.id, 'imageFile'], node.imageFile);
        }
      }


    };

    return ImageComponent;
  }(NodeComponent));

  var ImageHTMLConverter = {

    type: 'image',
    tagName: 'img',

    import: function(el, node, converter) {
      var imageFile = converter.getDocument().create({
        id: 'file-'+node.id,
        type: 'file',
        fileType: 'image',
        url: el.attr('src')
      });
      node.imageFile = imageFile.id;
    },

    export: function(node, el) {
      var imageFile = node.document.get(node.imageFile);
      el.attr('src', imageFile.getUrl());
    }
  }

  function insertImage(tx, file) {
    
    var imageFile = tx.create({
      type: 'file',
      fileType: 'image',
      mimeType: file.type,
      sourceFile: file
    });
    
    tx.insertBlockNode({
      type: 'image',
      imageFile: imageFile.id
    });
  }

  var InsertImageCommand = (function (InsertNodeCommand$$1) {
    function InsertImageCommand () {
      InsertNodeCommand$$1.apply(this, arguments);
    }

    if ( InsertNodeCommand$$1 ) InsertImageCommand.__proto__ = InsertNodeCommand$$1;
    InsertImageCommand.prototype = Object.create( InsertNodeCommand$$1 && InsertNodeCommand$$1.prototype );
    InsertImageCommand.prototype.constructor = InsertImageCommand;

    InsertImageCommand.prototype.execute = function execute (params) {
      var editorSession = params.editorSession;

      editorSession.transaction(function (tx) {
        params.files.forEach(function (file) {
          insertImage(tx, file);
        });
      });
    };

    return InsertImageCommand;
  }(InsertNodeCommand));

  var InsertImageTool = (function (ToggleTool$$1) {
    function InsertImageTool () {
      ToggleTool$$1.apply(this, arguments);
    }

    if ( ToggleTool$$1 ) InsertImageTool.__proto__ = ToggleTool$$1;
    InsertImageTool.prototype = Object.create( ToggleTool$$1 && ToggleTool$$1.prototype );
    InsertImageTool.prototype.constructor = InsertImageTool;

    InsertImageTool.prototype.getClassNames = function getClassNames () {
      return 'sc-insert-image-tool'
    };

    InsertImageTool.prototype.renderButton = function renderButton ($$) {
      var button = ToggleTool$$1.prototype.renderButton.call(this, $$);
      var input = $$('input').attr('type', 'file').ref('input')
        .on('change', this.onFileSelect);
      return [button, input]
    };

    InsertImageTool.prototype.onClick = function onClick () {
      this.refs.input.click();
    };

    InsertImageTool.prototype.onFileSelect = function onFileSelect (e) {
      var files = e.currentTarget.files;
      this.executeCommand({
        files: Array.prototype.slice.call(files)
      });
    };

    return InsertImageTool;
  }(ToggleTool));

  var DropImage = {
    type: 'drop-asset',
    match: function match(params) {
      
      var isImage = params.file.type.indexOf('image/') === 0;
      return params.type === 'file' && isImage
    },
    drop: function drop(tx, params) {
      insertImage(tx, params.file);
    }
  }

  var ImageProxy = (function (DefaultFileProxy$$1) {
    function ImageProxy () {
      DefaultFileProxy$$1.apply(this, arguments);
    }if ( DefaultFileProxy$$1 ) ImageProxy.__proto__ = DefaultFileProxy$$1;
    ImageProxy.prototype = Object.create( DefaultFileProxy$$1 && DefaultFileProxy$$1.prototype );
    ImageProxy.prototype.constructor = ImageProxy;

    

    return ImageProxy;
  }(DefaultFileProxy));


  ImageProxy.match = function(fileNode) {
    return fileNode.fileType === 'image'
  };

  var ImagePackage = {
    name: 'image',
    configure: function(config) {
      config.addNode(ImageNode);
      config.addComponent('image', ImageComponent);
      config.addConverter('html', ImageHTMLConverter);
      config.addConverter('xml', ImageHTMLConverter);
      config.addCommand('insert-image', InsertImageCommand, {
        nodeType: 'image',
        commandGroup: 'insert'
      });
      config.addTool('insert-image', InsertImageTool);
      config.addIcon('insert-image', { 'fontawesome': 'fa-image' });
      config.addLabel('image', {
        en: 'Image',
        de: 'Bild'
      });
      config.addLabel('insert-image', {
        en: 'Insert image',
        de: 'Bild einfügen'
      });
      config.addDropHandler(DropImage);
      config.addFileProxy(ImageProxy);
    },
    ImageNode: ImageNode,
    ImageComponent: ImageComponent,
    ImageHTMLConverter: ImageHTMLConverter,
    InsertImageCommand: InsertImageCommand,
    InsertImageTool: InsertImageTool,
    DropImage: DropImage
  }

  var InlineWrapper = (function (InlineNode$$1) {
    function InlineWrapper () {
      InlineNode$$1.apply(this, arguments);
    }

    if ( InlineNode$$1 ) InlineWrapper.__proto__ = InlineNode$$1;
    InlineWrapper.prototype = Object.create( InlineNode$$1 && InlineNode$$1.prototype );
    InlineWrapper.prototype.constructor = InlineWrapper;

    InlineWrapper.prototype.getWrappedNode = function getWrappedNode () {
      return this.getDocument().get(this.wrappedNode)
    };

    return InlineWrapper;
  }(InlineNode));

  InlineWrapper.schema = {
    type: 'inline-wrapper',
    wrappedNode: 'id'
  };

  var InlineWrapperComponent = (function (InlineNodeComponent$$1) {
    function InlineWrapperComponent () {
      InlineNodeComponent$$1.apply(this, arguments);
    }

    if ( InlineNodeComponent$$1 ) InlineWrapperComponent.__proto__ = InlineNodeComponent$$1;
    InlineWrapperComponent.prototype = Object.create( InlineNodeComponent$$1 && InlineNodeComponent$$1.prototype );
    InlineWrapperComponent.prototype.constructor = InlineWrapperComponent;

    InlineWrapperComponent.prototype.getClassNames = function getClassNames () {
      
      return 'sc-inline-wrapper sc-inline-node'
    };

    InlineWrapperComponent.prototype.renderContent = function renderContent ($$) {
      var node = this.props.node;
      var doc = node.getDocument();

      var wrappedNode = doc.get(node.wrappedNode);
      var el = $$('span').addClass('sc-inline-wrapper');
      if (wrappedNode) {
        var componentRegistry = this.context.componentRegistry;
        var ComponentClass = componentRegistry.get(wrappedNode.type);
        if (ComponentClass) {
          el.append($$(ComponentClass, {
            disabled: this.props.disabled,
            node: wrappedNode,
          }).ref('wrappedNode'));
        } else {
          console.error('No component registered for node type' + wrappedNode.type);
        }
      } else {
        console.error('Could not find wrapped node: ' + node.wrappedNode);
      }
      return el
    };

    return InlineWrapperComponent;
  }(InlineNodeComponent));

  var InlineWrapperConverter = {
    type: 'inline-wrapper',

    matchElement: function(el, converter) {
      var blockConverter = converter._getConverterForElement(el, 'block');
      return Boolean(blockConverter)
    },

    import: function(el, node, converter) {
      
      node.id = converter.nextId('inline-wrapper');
      var state = converter.state;
      state.popElementContext();
      state.pushElementContext(state.getCurrentElementContext().tagName);
      node.wrappedNode = converter.convertElement(el).id;
    },

    export: function(node, el, converter) {
      return converter.convertNode(node.wrappedNode)
    }
  }

  var InlineWrapperPackage = {
    name: 'inline-wrapper',
    configure: function(config, options) {
      config.addNode(InlineWrapper);
      config.addComponent(InlineWrapper.type, InlineWrapperComponent);
      if (options.converters) {
        options.converters.forEach(function(name) {
          config.addConverter(name, InlineWrapperConverter);
        });
      }
    },
    InlineWrapper: InlineWrapper,
    InlineWrapperComponent: InlineWrapperComponent,
    InlineWrapperConverter: InlineWrapperConverter
  }

  var Link = (function (PropertyAnnotation$$1) {
    function Link () {
      PropertyAnnotation$$1.apply(this, arguments);
    }if ( PropertyAnnotation$$1 ) Link.__proto__ = PropertyAnnotation$$1;
    Link.prototype = Object.create( PropertyAnnotation$$1 && PropertyAnnotation$$1.prototype );
    Link.prototype.constructor = Link;

    

    return Link;
  }(PropertyAnnotation));

  Link.schema = {
    type: "link",
    title: { type: 'string', optional: true },
    url: { type: 'string', 'default': ''}
  };


  Link.fragmentation = Fragmenter.SHOULD_NOT_SPLIT;

  Link.autoExpandRight = false;

  var LinkComponent = (function (AnnotationComponent$$1) {
    function LinkComponent () {
      AnnotationComponent$$1.apply(this, arguments);
    }

    if ( AnnotationComponent$$1 ) LinkComponent.__proto__ = AnnotationComponent$$1;
    LinkComponent.prototype = Object.create( AnnotationComponent$$1 && AnnotationComponent$$1.prototype );
    LinkComponent.prototype.constructor = LinkComponent;

    LinkComponent.prototype.didMount = function didMount () {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      AnnotationComponent$$1.prototype.didMount.apply(this, args);

      var node = this.props.node;
      this.context.editorSession.onRender('document', this.rerender, this, {
        path: [node.id, 'url']
      });
    };

    LinkComponent.prototype.dispose = function dispose () {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      AnnotationComponent$$1.prototype.dispose.apply(this, args);

      this.context.editorSession.off(this);
    };

    LinkComponent.prototype.render = function render ($$) { 
      var el = AnnotationComponent$$1.prototype.render.call(this, $$);

      el.tagName = 'a';
      el.attr('href', this.props.node.url);

      var titleComps = [this.props.node.url];
      if (this.props.node.title) {
        titleComps.push(this.props.node.title);
      }

      return el.attr("title", titleComps.join(' | '))
    };

    return LinkComponent;
  }(AnnotationComponent));

  var LinkCommand = (function (AnnotationCommand$$1) {
    function LinkCommand () {
      AnnotationCommand$$1.apply(this, arguments);
    }

    if ( AnnotationCommand$$1 ) LinkCommand.__proto__ = AnnotationCommand$$1;
    LinkCommand.prototype = Object.create( AnnotationCommand$$1 && AnnotationCommand$$1.prototype );
    LinkCommand.prototype.constructor = LinkCommand;

    LinkCommand.prototype.canFuse = function canFuse () { return false };

    
    LinkCommand.prototype.executeCreate = function executeCreate (params) {
      var result = AnnotationCommand$$1.prototype.executeCreate.call(this, params);
      var editorSession = this._getEditorSession(params);
      editorSession.transaction(function (tx) {
        tx.setSelection(tx.selection.collapse());
      });
      return result
    };

    return LinkCommand;
  }(AnnotationCommand));

  var LinkHTMLConverter = {

    type: "link",
    tagName: 'a',

    import: function(el, node) {
      node.url = el.attr('href');
      node.title = el.attr('title');
    },

    export: function(link, el) {
      el.attr({
        href: link.url,
        title: link.title
      });
    }

  }

  var EditLinkTool = (function (ToggleTool$$1) {
    function EditLinkTool () {
      ToggleTool$$1.apply(this, arguments);
    }

    if ( ToggleTool$$1 ) EditLinkTool.__proto__ = ToggleTool$$1;
    EditLinkTool.prototype = Object.create( ToggleTool$$1 && ToggleTool$$1.prototype );
    EditLinkTool.prototype.constructor = EditLinkTool;

    EditLinkTool.prototype.getUrlPath = function getUrlPath () {
      var propPath = this.constructor.urlPropertyPath;
      return [this.getNodeId()].concat(propPath)
    };

    EditLinkTool.prototype.getNodeId = function getNodeId () {
      return this.props.commandState.nodeId
    };

    EditLinkTool.prototype._openLink = function _openLink () {
      var doc = this.context.editorSession.getDocument();
      window.open(doc.get(this.getUrlPath()), '_blank');
    };

    EditLinkTool.prototype.render = function render ($$) {
      var Input = this.getComponent('input');
      var Button = this.getComponent('button');
      var commandState = this.props.commandState;
      var el = $$('div').addClass('sc-edit-link-tool');

      
      if (commandState.disabled) {
        console.warn('Tried to render EditLinkTool while disabled.');
        return el
      }

      var urlPath = this.getUrlPath();

      el.append(
        $$(Input, {
          type: 'url',
          path: urlPath,
          placeholder: 'Paste or type a link url'
        }).ref('input'),
        $$(Button, {
          icon: 'open-link',
          theme: this.props.theme,
        }).attr('title', this.getLabel('open-link'))
          .on('click', this._openLink),

        $$(Button, {
          icon: 'delete',
          theme: this.props.theme,
        }).attr('title', this.getLabel('delete-link'))
          .on('click', this.onDelete)
      );
      return el
    };

    EditLinkTool.prototype.onDelete = function onDelete (e) {
      e.preventDefault();
      var nodeId = this.getNodeId();
      var sm = this.context.surfaceManager;
      var surface = sm.getFocusedSurface();
      if (!surface) {
        console.warn('No focused surface. Stopping command execution.');
        return
      }
      var editorSession = this.context.editorSession;
      editorSession.transaction(function(tx, args) {
        tx.delete(nodeId);
        return args
      });
    };

    return EditLinkTool;
  }(ToggleTool));

  EditLinkTool.urlPropertyPath = ['url'];

  var LinkPackage = {
    name: 'link',
    configure: function(config) {
      config.addNode(Link);
      config.addComponent('link', LinkComponent);
      config.addConverter('html', LinkHTMLConverter);
      config.addConverter('xml', LinkHTMLConverter);
      config.addCommand('link', LinkCommand, {
        nodeType: 'link',
        commandGroup: 'annotations'
      });
      config.addCommand('edit-link', EditAnnotationCommand, {
        nodeType: 'link',
        commandGroup: 'prompt'
      });
      config.addTool('edit-link', EditLinkTool);
      config.addIcon('link', { 'fontawesome': 'fa-link'});
      config.addIcon('open-link', { 'fontawesome': 'fa-external-link' });
      config.addLabel('link', {
        en: 'Link',
        de: 'Link'
      });
      config.addLabel('open-link', {
        en: 'Open Link',
        de: 'Link öffnen'
      });
      config.addLabel('delete-link', {
        en: 'Remove Link',
        de: 'Link löschen'
      });
      config.addKeyboardShortcut('CommandOrControl+K', { command: 'link' });
    },
    Link: Link,
    LinkComponent: LinkComponent,
    LinkCommand: LinkCommand,
    LinkHTMLConverter: LinkHTMLConverter,
    EditLinkTool: EditLinkTool
  }

  var ListNode = (function (superclass) {
    function ListNode () {
      superclass.apply(this, arguments);
    }

    if ( superclass ) ListNode.__proto__ = superclass;
    ListNode.prototype = Object.create( superclass && superclass.prototype );
    ListNode.prototype.constructor = ListNode;

    ListNode.prototype.createListItem = function createListItem (text) {
      return this.getDocument().create( { type: 'list-item', content: text })
    };

    ListNode.prototype.getItems = function getItems () {
      var doc = this.getDocument();
      return this.items.map(function (id) {
        return doc.get(id)
      })
    };

    ListNode.prototype.getItemsPath = function getItemsPath () {
      return [this.id, 'items']
    };

    ListNode.prototype.insertItemAt = function insertItemAt (pos, item) {
      var doc = this.getDocument();
      var id = item.id;
      doc.update(this.getItemsPath(), { type: 'insert', pos: pos, value: id });
    };

    ListNode.prototype.removeItemAt = function removeItemAt (pos) {
      var doc = this.getDocument();
      doc.update(this.getItemsPath(), { type: 'delete', pos: pos });
    };

    

    ListNode.prototype.getItemAt = function getItemAt (idx) {
      return this.getDocument().get(this.items[idx])
    };

    ListNode.prototype.getItemPosition = function getItemPosition (item) {
      var id = item.id;
      var pos = this.items.indexOf(id);
      if (pos < 0) { throw new Error('Item is not within this list: ' + id) }
      return pos
    };

    ListNode.prototype.getLength = function getLength () {
      return this.items.length
    };

    ListNode.prototype.getListTypeString = function getListTypeString () {
      return this.listType
    };

    ListNode.prototype.setListTypeString = function setListTypeString (levelTypeStr) {
      this.getDocument().set([this.id, 'listType'], levelTypeStr);
    };

    return ListNode;
  }(ListMixin(DocumentNode)));

  ListNode.type = 'list';

  ListNode.schema = {
    
    
    
    items: { type: [ 'array', 'id' ], default: [], owned: true },
    listType: { type: 'string', default: 'bullet' }
  };

  var ListItem = (function (TextNode$$1) {
    function ListItem () {
      TextNode$$1.apply(this, arguments);
    }

    if ( TextNode$$1 ) ListItem.__proto__ = TextNode$$1;
    ListItem.prototype = Object.create( TextNode$$1 && TextNode$$1.prototype );
    ListItem.prototype.constructor = ListItem;

    ListItem.prototype.isListItem = function isListItem () {
      return true
    };

    ListItem.prototype.getLevel = function getLevel () {
      return this.level
    };

    ListItem.prototype.setLevel = function setLevel (newLevel) {
      if (this.level !== newLevel) {
        this.getDocument().set([this.id, 'level'], newLevel);
      }
    };

    return ListItem;
  }(TextNode));

  ListItem.type = 'list-item';

  ListItem.schema = {
    level: { type: "number", default: 1 }
  };

  var ListItemComponent = (function (Component$$1) {
    function ListItemComponent () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) ListItemComponent.__proto__ = Component$$1;
    ListItemComponent.prototype = Object.create( Component$$1 && Component$$1.prototype );
    ListItemComponent.prototype.constructor = ListItemComponent;

    ListItemComponent.prototype.render = function render ($$) {
      var el = $$('li').addClass('sc-list-item');
      el.append(
        $$(TextPropertyComponent, { path: this.props.path}).ref('text')
      );
      
      if (this.props.children) {
        el.append(this.props.children);
      }
      return el
    };

    return ListItemComponent;
  }(Component));

  var ListComponent = (function (NodeComponent$$1) {
    function ListComponent () {
      NodeComponent$$1.apply(this, arguments);
    }

    if ( NodeComponent$$1 ) ListComponent.__proto__ = NodeComponent$$1;
    ListComponent.prototype = Object.create( NodeComponent$$1 && NodeComponent$$1.prototype );
    ListComponent.prototype.constructor = ListComponent;

    ListComponent.prototype.render = function render ($$) {
      var node = this.props.node;
      var el = renderListNode(node, function (item) {
        
        if (isString(item)) {
          return $$(item)
        } else if(item.type === 'list-item') {
          var path = item.getPath();
          return $$(ListItemComponent, {
            path: path,
            node: item,
            tagName: 'li'
          }).ref(item.id)
        }
      });
      el.addClass('sc-list').attr('data-id', node.id);
      return el
    };

    return ListComponent;
  }(NodeComponent));


  ListComponent.prototype._isCustomNodeComponent = true;

  var ListHTMLConverter = function ListHTMLConverter () {};

  var prototypeAccessors$10 = { type: { configurable: true } };

  prototypeAccessors$10.type.get = function () { return "list" };

  ListHTMLConverter.prototype.matchElement = function matchElement (el) {
    return el.is('ul') || el.is('ol')
  };

  ListHTMLConverter.prototype.import = function import$1 (el, node, converter) {
    this._santizeNestedLists(el);

    var items = [];
    var config = [];
    walk(el, (function (el, level) {
      if (!el.isElementNode()) { return }
      if (el.is('li')) {
        items.push({ el: el, level: level });
      } else if (!config[level]) {
        if (el.is('ul')) { config[level] = 'bullet'; }
        else if (el.is('ol')) { config[level] = 'order'; }
      }
    }));
    this._createItems(converter, node, items, config);
  };


  ListHTMLConverter.prototype._createItems = function _createItems (converter, node, items, levelTypes) {
    node.items = items.map(function (d) {
      var listItem = converter.convertElement(d.el);
      listItem.level = d.level;
      return listItem.id
    });
    node.listType = levelTypes.join(',');
  };

  ListHTMLConverter.prototype.export = function export$1 (node, el, converter) {
    var $$ = converter.$$;
    var _createElement = function(arg) {
      if (isString(arg)) {
        return $$(arg)
      } else {
        var item = arg;
        var path = item.getPath();
        return $$('li').append(converter.annotatedText(path))
      }
    };
    var _el = renderListNode(node, _createElement);
    el.tagName = _el.tagName;
    el.attr(_el.getAttributes());
    el.append(_el.getChildNodes());
    return el
  };

  ListHTMLConverter.prototype._santizeNestedLists = function _santizeNestedLists (root) {
    
    
    var nestedLists = root.findAll('ol,ul');
    nestedLists.forEach(function (el){
      while (!el.parentNode.is('ol,ul')) {
        var parent = el.parentNode;
        var grandParent = parent.parentNode;
        var pos = grandParent.getChildIndex(parent);
        grandParent.insertAt(pos+1, el);
      }
    });
  };

  Object.defineProperties( ListHTMLConverter.prototype, prototypeAccessors$10 );

  var ListItemHTMLConverter = {

    type: "list-item",

    matchElement: function(el) {
      return el.is('li')
    },

    import: function(el, node, converter) {
      node.content = converter.annotatedText(el, [node.id, 'content']);
    },

    export: function(node, el, converter) {
      el.append(converter.annotatedText(node.getPath()));
    }
  }

  var ToggleListCommand = (function (Command$$1) {
    function ToggleListCommand () {
      Command$$1.apply(this, arguments);
    }

    if ( Command$$1 ) ToggleListCommand.__proto__ = Command$$1;
    ToggleListCommand.prototype = Object.create( Command$$1 && Command$$1.prototype );
    ToggleListCommand.prototype.constructor = ToggleListCommand;

    ToggleListCommand.prototype.getCommandState = function getCommandState (params) {
      var editorSession = this._getEditorSession(params);
      var doc = editorSession.getDocument();
      var sel = this._getSelection(params);
      if (sel && sel.isPropertySelection()) {
        var path = sel.path;
        var node = doc.get(path[0]);
        if (node) {
          if (node.isListItem()) {
            var level = node.getLevel();
            var list = node.getParent();
            var listType = list.getListType(level);
            var active = listType === this.config.spec.listType;
            var action = active ? 'toggleList' : 'setListType';
            var listId = list.id;
            return {
              disabled: false, active: active,
              action: action, listId: listId, level: level
            }
          } else if (node.isText() && node.isBlock) {
            return {
              disabled: false,
              action: 'switchTextType'
            }
          }
        }
      }
      return { disabled: true }
    };

    ToggleListCommand.prototype.execute = function execute (params) {
      var this$1 = this;

      var commandState = params.commandState;
      var disabled = commandState.disabled;
      var action = commandState.action;
      if (disabled) { return }

      var editorSession = params.editorSession;
      switch(action) {
        case 'toggleList': {
          editorSession.transaction(function (tx) {
            tx.toggleList();
          }, { action: 'toggleList' });
          break
        }
        case 'setListType': {
          var listId = commandState.listId;
          var level = commandState.level;
          editorSession.transaction(function (tx) {
            var list = tx.get(listId);
            list.setListType(level, this$1.config.spec.listType);
          }, { action: 'setListType' });
          break
        }
        case 'switchTextType': {
          editorSession.transaction(function (tx) {
            tx.toggleList({ listType: this$1.config.spec.listType });
          }, { action: 'toggleList' });
          break
        }
        default:
          
      }
    };

    return ToggleListCommand;
  }(Command));

  var IndentListCommand = (function (Command$$1) {
    function IndentListCommand () {
      Command$$1.apply(this, arguments);
    }

    if ( Command$$1 ) IndentListCommand.__proto__ = Command$$1;
    IndentListCommand.prototype = Object.create( Command$$1 && Command$$1.prototype );
    IndentListCommand.prototype.constructor = IndentListCommand;

    IndentListCommand.prototype.getCommandState = function getCommandState (params) {
      var editorSession = this._getEditorSession(params);
      var doc = editorSession.getDocument();
      var sel = this._getSelection(params);
      if (sel && sel.isPropertySelection()) {
        var path = sel.path;
        var node = doc.get(path[0]);
        if (node) {
          if (node.isListItem()) {
            return {
              disabled: false
            }
          }
        }
      }
      return { disabled: true }
    };

    IndentListCommand.prototype.execute = function execute (params) {
      var commandState = params.commandState;
      var disabled = commandState.disabled;

      if (disabled) { return }

      var editorSession = params.editorSession;
      var action = this.config.spec.action;
      switch(action) {
        case 'indent': {
          editorSession.transaction(function (tx) {
            tx.indent();
          }, { action: 'indent' });
          break
        }
        case 'dedent': {
          editorSession.transaction(function (tx) {
            tx.dedent();
          }, { action: 'dedent' });
          break
        }
        default:
          
      }
    };

    return IndentListCommand;
  }(Command));

  var ListPackage = {
    name: 'list',
    configure: function(config) {
      config.addNode(ListNode);
      config.addNode(ListItem);
      config.addComponent('list', ListComponent);
      config.addConverter('html', ListHTMLConverter);
      config.addConverter('html', ListItemHTMLConverter);
      config.addCommand('toggle-unordered-list', ToggleListCommand, {
        spec: { listType: 'bullet' },
        commandGroup: 'list'
      });
      config.addLabel('toggle-unordered-list', {
        en: 'Toggle list',
        de: 'Liste entfernen'
      });
      config.addIcon('toggle-unordered-list', { 'fontawesome': 'fa-list-ul' });

      config.addCommand('toggle-ordered-list', ToggleListCommand, {
        spec: { listType: 'order' },
        commandGroup: 'list'
      });
      config.addLabel('toggle-ordered-list', {
        en: 'Toggle list',
        de: 'Aufzählung entfernen'
      });
      config.addIcon('toggle-ordered-list', { 'fontawesome': 'fa-list-ol' });

      config.addCommand('indent-list', IndentListCommand, {
        spec: { action: 'indent' },
        commandGroup: 'list'
      });
      config.addLabel('indent-list', {
        en: 'Increase indentation',
        de: 'Einrückung vergrößern'
      });
      config.addIcon('indent-list', { 'fontawesome': 'fa-indent' });

      config.addCommand('dedent-list', IndentListCommand, {
        spec: { action: 'dedent' },
        commandGroup: 'list'
      });
      config.addLabel('dedent-list', {
        en: 'Decrease indentation',
        de: 'Einrückung verringern'
      });
      config.addIcon('dedent-list', { 'fontawesome': 'fa-dedent' });

    },
    ListNode: ListNode,
    ListItemNode: ListItem,
    ListComponent: ListComponent,
    ListHTMLConverter: ListHTMLConverter,
    ListItemHTMLConverter: ListItemHTMLConverter,
    ToggleListCommand: ToggleListCommand,
    IndentListCommand: IndentListCommand
  }

  var MultiSelect = (function (Component$$1) {
    function MultiSelect () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) MultiSelect.__proto__ = Component$$1;
    MultiSelect.prototype = Object.create( Component$$1 && Component$$1.prototype );
    MultiSelect.prototype.constructor = MultiSelect;

    MultiSelect.prototype.didMount = function didMount () {
      this._recomputeOrder();
    };

    MultiSelect.prototype.getInitialState = function getInitialState () {
      return {
        options: this.props.options,
        selectedOptions: this.props.selectedOptions,
        collapsed: true
      }
    };

    MultiSelect.prototype.render = function render ($$) {
      var this$1 = this;

      var options = this.state.options;
      var maxItems = this.props.maxItems;
      var selectedOptions = this.state.selectedOptions;
      var collapsed = this.state.collapsed;
      var limit = selectedOptions.length > maxItems ? selectedOptions.length : maxItems;

      var el = $$('div').addClass('sc-multi-select');

      var optionsEl = options.map(function (option, index) {
        if(!collapsed || index <= limit - 1) {
          var optionEl = $$('div').addClass('se-select-option');
          var icon = selectedOptions.indexOf(option.id) > -1 ? 'selected-option' : 'unselected-option';
          optionEl.append(
            this$1.renderIcon($$, icon).addClass('se-option-icon'),
            $$('span').addClass('se-option-label').append(option.label)
          ).on('click', this$1._onToggleCheckbox.bind(this$1, option.id));

          return optionEl
        } else {
          return ''
        }
      });

      el.append(optionsEl);

      var collapseLabelEl = $$('div').addClass('se-multi-collapse-label');

      if(options.length > limit) {
        if(collapsed) {
          var leftItems = options.length - limit;
          collapseLabelEl.append(
            this.getLabel('expand-options') + ' (' + leftItems + ')'
          ).on('click', this._onToggleExpand);
        } else {
          collapseLabelEl.append(
            this.getLabel('collapse-options')
          ).on('click', this._onToggleExpand);
        }
      }

      el.append(collapseLabelEl);

      return el
    };

    MultiSelect.prototype.renderIcon = function renderIcon ($$, icon) {
      var iconEl = this.context.iconProvider.renderIcon($$, icon);
      return iconEl
    };

    MultiSelect.prototype.getSelectedOptions = function getSelectedOptions () {
      return this.state.selectedOptions
    };

    MultiSelect.prototype._recomputeOrder = function _recomputeOrder () {
      var options = this.state.options;
      var selectedOptions = this.state.selectedOptions;

      var sortedOptions = orderBy(options, function (option) {
        return selectedOptions.indexOf(option.id)
      }, 'desc');

      this.extendState({options: sortedOptions});
    };

    MultiSelect.prototype._onToggleCheckbox = function _onToggleCheckbox (id) {
      var selectedOptions = this.state.selectedOptions;
      var index = selectedOptions.indexOf(id);

      if(index > -1) {
        selectedOptions.splice(index, 1);
      } else {
        selectedOptions.push(id);
      }

      this.extendState({selectedOptions: selectedOptions});
      this.el.emit('change');
    };

    MultiSelect.prototype._onToggleExpand = function _onToggleExpand () {
      var collapsed = this.state.collapsed;

      this._recomputeOrder();
      this.extendState({collapsed: !collapsed});
    };

    return MultiSelect;
  }(Component));

  var MultiSelectPackage = {
    name: 'multi-select',
    configure: function(config) {
      config.addComponent('multi-select', MultiSelect);
      config.addIcon('selected-option', { 'fontawesome': 'fa-check-square-o' });
      config.addIcon('unselected-option', { 'fontawesome': 'fa-square-o' });
      config.addLabel('expand-options', {
        en: 'Show more',
        de: 'Zeig mehr'
      });
      config.addLabel('collapse-options', {
        en: 'Show less',
        de: 'Zeige weniger'
      });
    }
  }

  var Paragraph = (function (TextBlock$$1) {
  	function Paragraph () {
  		TextBlock$$1.apply(this, arguments);
  	}if ( TextBlock$$1 ) Paragraph.__proto__ = TextBlock$$1;
  	Paragraph.prototype = Object.create( TextBlock$$1 && TextBlock$$1.prototype );
  	Paragraph.prototype.constructor = Paragraph;

  	

  	return Paragraph;
  }(TextBlock));

  Paragraph.type = "paragraph";

  var ParagraphComponent = (function (TextBlockComponent$$1) {
    function ParagraphComponent () {
      TextBlockComponent$$1.apply(this, arguments);
    }

    if ( TextBlockComponent$$1 ) ParagraphComponent.__proto__ = TextBlockComponent$$1;
    ParagraphComponent.prototype = Object.create( TextBlockComponent$$1 && TextBlockComponent$$1.prototype );
    ParagraphComponent.prototype.constructor = ParagraphComponent;

    ParagraphComponent.prototype.render = function render ($$) {
      var el = TextBlockComponent$$1.prototype.render.call(this, $$);
      return el.addClass('sc-paragraph')
    };

    ParagraphComponent.prototype.getTagName = function getTagName () {
      return 'p'
    };

    return ParagraphComponent;
  }(TextBlockComponent));

  var ParagraphHTMLConverter = {

    type: 'paragraph',
    tagName: 'p',

    import: function(el, node, converter) {
      var textAlign = el.attr('data-text-align');
      node.textAlign = textAlign || 'left';
      node.content = converter.annotatedText(el, [node.id, 'content']);
    },

    export: function(node, el, converter) {
      
      if (node.textAlign !== 'left') {
        el.attr('data-text-align', node.textAlign);
      }
      el.append(converter.annotatedText([node.id, 'content']));
    }

  }

  var ParagraphPackage = {
    name: 'paragraph',
    configure: function(config) {
      config.addNode(Paragraph);
      config.addComponent(Paragraph.type, ParagraphComponent);
      config.addConverter('html', ParagraphHTMLConverter);
      config.addConverter('xml', ParagraphHTMLConverter);
      config.addCommand('paragraph', SwitchTextTypeCommand, {
        spec: { type: 'paragraph' },
        commandGroup: 'text-types'
      });
      config.addIcon('paragraph', { 'fontawesome': 'fa-paragraph' });
      config.addLabel('paragraph', {
        en: 'Paragraph',
        de: 'Paragraph'
      });
      config.addKeyboardShortcut('CommandOrControl+Alt+0', { command: 'paragraph' });
    },
    Paragraph: Paragraph,
    ParagraphComponent: ParagraphComponent,
    ParagraphHTMLConverter: ParagraphHTMLConverter
  }

  var SaveCommand = (function (Command$$1) {
    function SaveCommand() {
      Command$$1.call(this, { name: 'save' });
    }

    if ( Command$$1 ) SaveCommand.__proto__ = Command$$1;
    SaveCommand.prototype = Object.create( Command$$1 && Command$$1.prototype );
    SaveCommand.prototype.constructor = SaveCommand;

    SaveCommand.prototype.getCommandState = function getCommandState (params) {
      var dirty = params.editorSession.hasUnsavedChanges();
      return {
        disabled: !dirty,
        active: false
      }
    };

    SaveCommand.prototype.execute = function execute (params) {
      var editorSession = params.editorSession;
      editorSession.save();
      return {
        status: 'saving-process-started'
      }
    };

    return SaveCommand;
  }(Command));

  var PersistencePackage = {
    name: 'persistence',
    configure: function(config) {
      config.addCommand('save', SaveCommand, {
        commandGroup: 'persistence'
      });
      config.addIcon('save', { 'fontawesome': 'fa-save' });
      config.addLabel('save', {
        en: 'Save',
        de: 'Speichern'
      });
      config.addKeyboardShortcut('CommandOrControl+S', { command: 'save' });

    },
    SaveCommand: SaveCommand
  }

  var Strong = (function (PropertyAnnotation$$1) {
  	function Strong () {
  		PropertyAnnotation$$1.apply(this, arguments);
  	}if ( PropertyAnnotation$$1 ) Strong.__proto__ = PropertyAnnotation$$1;
  	Strong.prototype = Object.create( PropertyAnnotation$$1 && PropertyAnnotation$$1.prototype );
  	Strong.prototype.constructor = Strong;

  	

  	return Strong;
  }(PropertyAnnotation));

  Strong.type = "strong";
  Strong.fragmentation = Fragmenter.ANY;

  var StrongHTMLConverter = {

    type: "strong",
    tagName: "strong",

    matchElement: function(el) {
      return el.is("strong, b")
    }

  }

  var StrongComponent = (function (AnnotationComponent$$1) {
    function StrongComponent () {
      AnnotationComponent$$1.apply(this, arguments);
    }

    if ( AnnotationComponent$$1 ) StrongComponent.__proto__ = AnnotationComponent$$1;
    StrongComponent.prototype = Object.create( AnnotationComponent$$1 && AnnotationComponent$$1.prototype );
    StrongComponent.prototype.constructor = StrongComponent;

    StrongComponent.prototype.getTagName = function getTagName () {
      return 'strong'
    };

    return StrongComponent;
  }(AnnotationComponent));

  var StrongPackage = {
    name: 'strong',
    configure: function(config) {
      config.addNode(Strong);
      config.addConverter('html', StrongHTMLConverter);
      config.addConverter('xml', StrongHTMLConverter);
      config.addComponent('strong', StrongComponent);
      config.addCommand('strong', AnnotationCommand, {
        nodeType: 'strong',
        commandGroup: 'annotations'
      });
      config.addIcon('strong', { 'fontawesome': 'fa-bold' });
      config.addLabel('strong', {
        en: 'Strong',
        de: 'Fett'
      });
      config.addKeyboardShortcut('CommandOrControl+B', { command: 'strong' });
    },
    Strong: Strong,
    StrongComponent: StrongComponent,
    StrongHTMLConverter: StrongHTMLConverter
  }

  var Subscript = (function (PropertyAnnotation$$1) {
  	function Subscript () {
  		PropertyAnnotation$$1.apply(this, arguments);
  	}if ( PropertyAnnotation$$1 ) Subscript.__proto__ = PropertyAnnotation$$1;
  	Subscript.prototype = Object.create( PropertyAnnotation$$1 && PropertyAnnotation$$1.prototype );
  	Subscript.prototype.constructor = Subscript;

  	

  	return Subscript;
  }(PropertyAnnotation));

  Subscript.type = 'subscript';


  Subscript.fragmentation = Fragmenter.ANY;

  var SubscriptHTMLConverter = {
    type: 'subscript',
    tagName: 'sub'
  }

  var SubscriptPackage = {
    name: 'subscript',
    configure: function(config) {
      config.addNode(Subscript);
      config.addConverter('html', SubscriptHTMLConverter);
      config.addConverter('xml', SubscriptHTMLConverter);
      config.addComponent('subscript', AnnotationComponent);
      config.addCommand('subscript', AnnotationCommand, {
        nodeType: 'subscript',
        commandGroup: 'annotations'
      });
      config.addIcon('subscript', { 'fontawesome': 'fa-subscript' });
      config.addLabel('subscript', {
        en: 'Subscript',
        de: 'Tiefgestellt'
      });
    },
    Subscript: Subscript,
    SubscriptHTMLConverter: SubscriptHTMLConverter
  }

  var Superscript = (function (PropertyAnnotation$$1) {
  	function Superscript () {
  		PropertyAnnotation$$1.apply(this, arguments);
  	}if ( PropertyAnnotation$$1 ) Superscript.__proto__ = PropertyAnnotation$$1;
  	Superscript.prototype = Object.create( PropertyAnnotation$$1 && PropertyAnnotation$$1.prototype );
  	Superscript.prototype.constructor = Superscript;

  	

  	return Superscript;
  }(PropertyAnnotation));

  Superscript.type = 'superscript';


  Superscript.fragmentation = Fragmenter.ANY;

  var SuperscriptHTMLConverter = {
    type: 'superscript',
    tagName: 'sup'
  }

  var SuperscriptPackage = {
    name: 'superscript',
    configure: function(config) {
      config.addNode(Superscript);
      config.addConverter('html', SuperscriptHTMLConverter);
      config.addConverter('xml', SuperscriptHTMLConverter);
      config.addComponent('superscript', AnnotationComponent);
      config.addCommand('superscript', AnnotationCommand, {
        nodeType: 'superscript',
        commandGroup: 'annotations'
      });
      config.addIcon('superscript', { 'fontawesome': 'fa-superscript' });
      config.addLabel('superscript', {
        en: 'Superscript',
        de: 'Hochgestellt'
      });
    },
    Superscript: Superscript,
    SuperscriptHTMLConverter: SuperscriptHTMLConverter
  }

  var LEFT_QUOTE = "\u201C";
  var RIGHT_QUOTE = "\u201D";

  var InsertQuoteMarkCommand = (function (Command$$1) {
    function InsertQuoteMarkCommand () {
      Command$$1.apply(this, arguments);
    }

    if ( Command$$1 ) InsertQuoteMarkCommand.__proto__ = Command$$1;
    InsertQuoteMarkCommand.prototype = Object.create( Command$$1 && Command$$1.prototype );
    InsertQuoteMarkCommand.prototype.constructor = InsertQuoteMarkCommand;

    InsertQuoteMarkCommand.prototype.getCommandState = function getCommandState (params, context) { 
      
      
      return {
        disabled: false
      }
    };

    InsertQuoteMarkCommand.prototype.execute = function execute (params, context) { 
      var editorSession = params.editorSession;
      var sel = editorSession.getSelection();
      var doc = editorSession.getDocument();
      if (sel.isPropertySelection()) {
        var nodeId = sel.start.getNodeId();
        var node = doc.get(nodeId);
        if (node.isText()) {
          var text = node.getText();
          var offset = sel.start.offset;
          var mark;
          if (offset === 0 || /\s/.exec(text.slice(offset-1, offset))) {
            mark = LEFT_QUOTE;
          } else {
            mark = RIGHT_QUOTE;
          }
          editorSession.transaction(function (tx) {
            tx.insertText(mark);
          });
          return true
        }
      }
      return false
    };

    return InsertQuoteMarkCommand;
  }(Command));

  var QuoteMarksPackage = {

    name: 'quote-marks',

    configure: function(config) {
      config.addCommand('insert-quote-marks', InsertQuoteMarkCommand, {
        commandGroup: 'text-macros'
      });
      config.addKeyboardShortcut('"', { type: 'textinput', command: 'insert-quote-marks' });
    }

  }

  var Table = (function (BlockNode$$1) {
    function Table () {
      BlockNode$$1.apply(this, arguments);
    }

    if ( BlockNode$$1 ) Table.__proto__ = BlockNode$$1;
    Table.prototype = Object.create( BlockNode$$1 && BlockNode$$1.prototype );
    Table.prototype.constructor = Table;

    Table.prototype.getRowCount = function getRowCount () {
      return this.cells.length
    };

    Table.prototype.getColCount = function getColCount () {
      if (this.cells.length > 0) {
        return this.cells[0].length
      } else {
        return 0
      }
    };

    Table.prototype.getCellAt = function getCellAt (row, col) {
      var cellId = this.cells[row][col];
      if (cellId) {
        return this.document.get(cellId)
      }
    };

    return Table;
  }(BlockNode));

  Table.schema = {
    type: 'table',
    cells: { type: ['array', 'array', 'id'], default: [], owned: true }
  };

  var TableCell = (function (TextNode$$1) {
    function TableCell () {
      TextNode$$1.apply(this, arguments);
    }if ( TextNode$$1 ) TableCell.__proto__ = TextNode$$1;
    TableCell.prototype = Object.create( TextNode$$1 && TextNode$$1.prototype );
    TableCell.prototype.constructor = TableCell;

    

    return TableCell;
  }(TextNode));

  TableCell.schema = {
    type: 'table-cell',
    rowspan: { type: 'number', default: 0 },
    colspan: { type: 'number', default: 0 }
  };

  var TableCellComponent = (function (Component$$1) {
    function TableCellComponent () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) TableCellComponent.__proto__ = Component$$1;
    TableCellComponent.prototype = Object.create( Component$$1 && Component$$1.prototype );
    TableCellComponent.prototype.constructor = TableCellComponent;

    TableCellComponent.prototype.render = function render ($$) {
      var node = this.props.node;
      var el = $$('td').addClass('sc-table-cell');
      el.append(
        $$(TextPropertyEditor, {
          path: node.getPath(),
          disabled: this.props.disabled
        }).ref('editor')
      );
      if (node.rowspan > 0) {
        el.attr('rowspan', node.rowspan);
      }
      if (node.colspan > 0) {
        el.attr('colspan', node.colspan);
      }
      return el
    };

    TableCellComponent.prototype.grabFocus = function grabFocus () {
      var node = this.props.node;
      this.context.editorSession.setSelection({
        type: 'property',
        path: node.getPath(),
        startOffset: node.getLength(),
        surfaceId: this.refs.editor.id
      });
    };

    return TableCellComponent;
  }(Component));

  TableCellComponent.prototype._isTableCellComponent = true;

  var TableComponent = (function (Component$$1) {
    function TableComponent () {
      Component$$1.apply(this, arguments);
    }

    if ( Component$$1 ) TableComponent.__proto__ = Component$$1;
    TableComponent.prototype = Object.create( Component$$1 && Component$$1.prototype );
    TableComponent.prototype.constructor = TableComponent;

    TableComponent.prototype.render = function render ($$) {
      var this$1 = this;

      var el = $$('table').addClass('sc-table');
      var node = this.props.node;
      var doc = this.props.node.getDocument();
      var cells = this.props.node.cells;
      var rowCount = node.getRowCount();
      var colCount = node.getColCount();
      for (var i = 0; i < rowCount; i++) {
        var rowEl = $$('tr');
        for (var j = 0; j < colCount; j++) {
          var cellId = cells[i][j];
          
          if (cellId) {
            var cellNode = doc.get(cellId);
            var cellEl = $$(TableCellComponent, {
              node: cellNode,
              disabled: this$1.props.disabled
            }).ref(cellNode.id);
            rowEl.append(cellEl);
          }
        }
        el.append(rowEl);
      }
      el.on('click', this.onClick);
      el.on('dblclick', this.onDblClick);
      return el
    };

    TableComponent.prototype.onClick = function onClick (event) {
      event.stopPropagation();
      
    };

    
    TableComponent.prototype.onDblClick = function onDblClick (event) {
      event.stopPropagation();
      

      
      
      var comp = Component$$1.unwrap(event.target);
      if (comp) {
        var cellComp;
        if (comp._isTableCellComponent) {
          cellComp = comp;
        } else if (comp._isTextPropertyEditor) {
          cellComp = comp.getParent();
        } else if (comp._isTextPropertyComponent) {
          cellComp = comp.getParent().getParent();
        } else {
          console.warn('TODO: find the right cell');
        }
        if (cellComp) {
          cellComp.grabFocus();
        }
      }
    };

    TableComponent.prototype.grabFocus = function grabFocus () {
      var cellId = this.props.node.cells[0][0];
      if (cellId) {
        var comp = this.refs[cellId];
        comp.grabFocus();
      }
    };

    return TableComponent;
  }(Component));

  TableComponent.hasDropzones = true;

  var TableHTMLConverter = {

    type: 'table',
    tagName: 'table',

    
    import: function(el, node, converter) {
      var trs = el.find('tbody').getChildren();
      var colCount = 0;
      var cells = [];
      var rowspans = []; 
      var loop = function ( i ) {
        var tds = trs[i].getChildren();
        var row = [];
        colCount = Math.max(tds.length, colCount);
        for (var j = 0; j < tds.length; j++) {
          var td = tds[j];
          
          if (rowspans[j] > 1) {
            row.push(null);
            rowspans[j] -= 1; 
          }
          var tableCell = converter.convertElement(td);
          row.push(tableCell.id);
          if (tableCell.rowspan > 1) {
            rowspans[j] = tableCell.rowspan;
          }
          if (tableCell.colspan > 1) {
            
            times$2(tableCell.colspan - 1, function () {
              row.push(null);
            });
          }
        }
        cells.push(row);
      };

      for (var i = 0; i < trs.length; i++) loop( i );
      node.cells = cells;
    },

    export: function(node, el, converter) {
      var $$ = converter.$$;
      var rowCount = node.getRowCount();
      var colCount = node.getColCount();
      for (var i = 0; i < rowCount; i++) {
        var rowEl = $$('tr');
        for (var j = 0; j < colCount; j++) {
          var cellId = node.cells[i][j];
          
          if (cellId) {
            var cellEl = converter.convertNode(cellId);
            rowEl.append(cellEl);
          }
        }
        el.append(rowEl);
      }
      return el
    }
  }

  var TableCellHTMLConverter = {

    type: 'table-cell',
    tagName: 'td',

    import: function(el, node, converter) {
      node.content = converter.annotatedText(el, [node.id, 'content']);
      var colspan = el.attr('colspan');
      var rowspan = el.attr('rowspan');
      if (colspan) {
        node.colspan = Number(colspan);
      }
      if (rowspan) {
        node.rowspan = Number(rowspan);
      }
    },

    export: function(node, el, converter) {
      el.append(converter.annotatedText([node.id, 'content']));
      if (node.rowspan > 0) {
        el.attr('rowspan', node.rowspan);
      }
      if (node.colspan > 0) {
        el.attr('colspan', node.colspan);
      }
      return el
    }
  }

  var InsertTableCommand = (function (InsertNodeCommand$$1) {
    function InsertTableCommand () {
      InsertNodeCommand$$1.apply(this, arguments);
    }

    if ( InsertNodeCommand$$1 ) InsertTableCommand.__proto__ = InsertNodeCommand$$1;
    InsertTableCommand.prototype = Object.create( InsertNodeCommand$$1 && InsertNodeCommand$$1.prototype );
    InsertTableCommand.prototype.constructor = InsertTableCommand;

    InsertTableCommand.prototype.createNodeData = function createNodeData (tx) {

      
      var a1 = tx.create({ id: uuid('table-cell'), type: 'table-cell', content: "A1" });
      var b1 = tx.create({ id: uuid('table-cell'), type: 'table-cell', content: "B1" });
      
      var a2 = tx.create({ id: uuid('table-cell'), type: 'table-cell', content: "A2" });
      var b2 = tx.create({ id: uuid('table-cell'), type: 'table-cell', content: "B2" });

      return {
        id: uuid('table'),
        type: 'table',
        
        cells: [
          [a1.id, b1.id ],
          [a2.id, b2.id]
        ]
      }

    };

    return InsertTableCommand;
  }(InsertNodeCommand));

  var TablePackage = {
    name: 'table',
    configure: function(config) {
      config.addNode(Table);
      config.addNode(TableCell);
      config.addComponent('table', TableComponent);
      config.addConverter('html', TableHTMLConverter);
      config.addConverter('html', TableCellHTMLConverter);
      config.addConverter('xml', TableHTMLConverter);
      config.addConverter('xml', TableCellHTMLConverter);
      config.addCommand('insert-table', InsertTableCommand, {
        nodeType: 'table',
        commandGroup: 'insert'
      });
      config.addIcon('insert-table', { 'fontawesome': 'fa-table' });
      config.addLabel('insert-table', {
        en: 'Table',
        de: 'Tabelle'
      });
      config.addLabel('table', {
        en: 'Table',
        de: 'Tabelle'
      });
      config.addLabel('table-cell.content', {
        en: 'Cell',
        de: 'Zelle'
      });
    },
    Table: Table,
    TableCell: TableCell
  }

  var ProseArticle = (function (Document$$1) {
    function ProseArticle () {
      Document$$1.apply(this, arguments);
    }

    if ( Document$$1 ) ProseArticle.__proto__ = Document$$1;
    ProseArticle.prototype = Object.create( Document$$1 && Document$$1.prototype );
    ProseArticle.prototype.constructor = ProseArticle;

    ProseArticle.prototype._initialize = function _initialize () {
      Document$$1.prototype._initialize.call(this);

      this.create({
        type: 'container',
        id: 'body',
        nodes: []
      });
    };

    return ProseArticle;
  }(Document));

  var ProseEditor = (function (AbstractEditor$$1) {
    function ProseEditor () {
      AbstractEditor$$1.apply(this, arguments);
    }

    if ( AbstractEditor$$1 ) ProseEditor.__proto__ = AbstractEditor$$1;
    ProseEditor.prototype = Object.create( AbstractEditor$$1 && AbstractEditor$$1.prototype );
    ProseEditor.prototype.constructor = ProseEditor;

    ProseEditor.prototype.render = function render ($$) {
      var SplitPane = this.componentRegistry.get('split-pane');
      var el = $$('div').addClass('sc-prose-editor');
      var toolbar = this._renderToolbar($$);
      var editor = this._renderEditor($$);
      var configurator = this.getConfigurator();
      var ScrollPane = this.componentRegistry.get('scroll-pane');
      var Overlay = this.componentRegistry.get('overlay');
      var ContextMenu = this.componentRegistry.get('context-menu');
      var Dropzones = this.componentRegistry.get('dropzones');

      var contentPanel = $$(ScrollPane, {
        name: 'contentPanel',
        contextMenu: this.props.contextMenu || 'native',
        scrollbarPosition: 'right',
        scrollbarType: this.props.scrollbarType
      }).append(
        editor,
        $$(Overlay, {
          toolPanel: configurator.getToolPanel('main-overlay'),
          theme: 'dark'
        }),
        $$(ContextMenu),
        $$(Dropzones)
      ).ref('contentPanel');

      el.append(
        $$(SplitPane, {splitType: 'horizontal'}).append(
          toolbar,
          $$(SplitPane, {splitType: 'horizontal', sizeB: 'inherit'}).append(
            contentPanel,
            $$(WorkflowPane, {
              toolPanel: configurator.getToolPanel('workflow')
            })
          )
        )
      );
      return el
    };

    ProseEditor.prototype._renderToolbar = function _renderToolbar ($$) {
      var configurator = this.getConfigurator();
      return $$('div').addClass('se-toolbar-wrapper').append(
        $$(Toolbar, {
          toolPanel: configurator.getToolPanel('toolbar')
        }).ref('toolbar')
      )
    };

    ProseEditor.prototype._renderEditor = function _renderEditor ($$) {
      var configurator = this.getConfigurator();
      return $$(ContainerEditor, {
        disabled: this.props.disabled,
        editorSession: this.editorSession,
        node: this.doc.get('body'),
        commands: configurator.getSurfaceCommandNames()
      }).ref('body')
    };

    return ProseEditor;
  }(AbstractEditor));

  var ProseArticleImporter = (function (HTMLImporter$$1) {
    function ProseArticleImporter () {
      HTMLImporter$$1.apply(this, arguments);
    }

    if ( HTMLImporter$$1 ) ProseArticleImporter.__proto__ = HTMLImporter$$1;
    ProseArticleImporter.prototype = Object.create( HTMLImporter$$1 && HTMLImporter$$1.prototype );
    ProseArticleImporter.prototype.constructor = ProseArticleImporter;

    ProseArticleImporter.prototype.convertDocument = function convertDocument (documentEl) {
      var body = documentEl.find('body');
      this.convertContainer(body.children, 'body');
    };

    return ProseArticleImporter;
  }(HTMLImporter));

  var ProseEditorPackage = {
    name: 'prose-editor',
    configure: function(config) {
      config.defineSchema({
        name: 'prose-article',
        DocumentClass: ProseArticle,
        defaultTextType: 'paragraph',
        version: '1.0'
      });
      config.addImporter('html', ProseArticleImporter);

      
      config.import(BasePackage);
      config.import(BlockquotePackage);
      config.import(ParagraphPackage);
      config.import(HeadingPackage);
      config.import(CodeblockPackage);
      config.import(CodePackage);
      config.import(EmphasisPackage);
      config.import(FindAndReplacePackage, {
        targetSurfaces: ['body']
      });
      config.import(StrongPackage);
      config.import(SubscriptPackage);
      config.import(SuperscriptPackage);
      config.import(LinkPackage);
      config.import(QuoteMarksPackage);
      config.import(ListPackage);
      config.import(TablePackage);

      
      config.addToolPanel('main-overlay', [
        {
          name: 'prompt',
          type: 'tool-group',
          commandGroups: ['prompt']
        }
      ]);

      config.addToolPanel('workflow', [
        {
          name: 'workflow',
          type: 'tool-group',
          commandGroups: ['workflows']
        }
      ]);

      
      config.addToolPanel('toolbar', [
        {
          name: 'text-types',
          type: 'tool-dropdown',
          showDisabled: true,
          style: 'descriptive',
          commandGroups: ['text-types']
        },
        {
          name: 'annotations',
          type: 'tool-group',
          showDisabled: true,
          style: 'minimal',
          commandGroups: ['annotations']
        },
        {
          name: 'insert',
          type: 'tool-dropdown',
          showDisabled: true,
          style: 'descriptive',
          commandGroups: ['insert']
        }
      ]);
    },
    ProseEditor: ProseEditor,
    ProseArticle: ProseArticle
  }

  var CorrectionTool = (function (Tool) {
    function CorrectionTool () {
      Tool.apply(this, arguments);
    }

    if ( Tool ) CorrectionTool.__proto__ = Tool;
    CorrectionTool.prototype = Object.create( Tool && Tool.prototype );
    CorrectionTool.prototype.constructor = CorrectionTool;

    CorrectionTool.prototype.render = function render ($$) {
      var this$1 = this;

      var node = this.props.node;
      var Button = this.getComponent('button');
      var el = $$('div').addClass('sc-correction-tool');

      if (node && node.suggestions.length > 0) {
        node.suggestions.forEach(function (s) {
          el.append(
            $$(Button, {
              label: s,
              style: this$1.props.style
            }).attr('title', this$1.getLabel('open-link'))
              .attr('data-correction', s)
              .on('click', this$1._applyCorrection.bind(this$1, s))
          );
        });
      } else {
        el.append(
          $$(Button, {
            label: 'No suggestions',
            style: this.props.style,
            disabled: true
          })
        );
      }
      return el
    };

    CorrectionTool.prototype._applyCorrection = function _applyCorrection (suggestion) {
      var editorSession = this.context.editorSession;
      var node = this.props.node;
      editorSession.transaction(function (tx) {
        var sel = tx.getSelection();
        tx.setSelection({
          type: 'property',
          path: node.start.path,
          startOffset: node.start.offset,
          endOffset: node.end.offset,
          containerId: sel.containerId
        });
        tx.insertText(suggestion);
      });
    };

    return CorrectionTool;
  }(ToggleTool));

  var SpellCheckCommand = (function (Command$$1) {
    function SpellCheckCommand () {
      Command$$1.apply(this, arguments);
    }

    if ( Command$$1 ) SpellCheckCommand.__proto__ = Command$$1;
    SpellCheckCommand.prototype = Object.create( Command$$1 && Command$$1.prototype );
    SpellCheckCommand.prototype.constructor = SpellCheckCommand;

    SpellCheckCommand.prototype.getCommandState = function getCommandState (params) {

      var state = params.selectionState;
      var markers = state.getMarkers();
      if (markers.length === 0) {
        return {
          disabled: true
        }
      }
      markers = markers.filter(function(m) {
        return m.type === 'spell-error'
      });

      if (markers.length > 0) {
        return {
          disabled: false,
          active: false,
          mode: null,
          node: markers[0]
        }
      } else {
        return {
          disabled: true
        }
      }
    };

    return SpellCheckCommand;
  }(Command));

  var DEFAULT_API_URL = 'http://localhost:4777/api/check';

  var SpellCheckManager = function SpellCheckManager(editorSession, options) {
    options = options || {};
    var wait = options.wait || 750;

    this.editorSession = editorSession;
    this.apiURL = options.apiURL || DEFAULT_API_URL;

    
    this.textPropertyManager = editorSession.markersManager;
    this.markersManager = editorSession.markersManager;

    this._schedule = {};
    this._scheduleCheck = debounce(this._runSpellCheck.bind(this), wait);

    editorSession.onFinalize('document', this._onDocumentChange, this);
  };

  SpellCheckManager.prototype.dispose = function dispose () {
    this.editorSession.off(this);
  };

  SpellCheckManager.prototype.check = function check (path) {
    this._runSpellCheck(String(path));
  };

  SpellCheckManager.prototype.runGlobalCheck = function runGlobalCheck () {
      var this$1 = this;

    var paths = Object.keys(this.textPropertyManager._textProperties);
    paths.forEach(function (p) {
      this$1._runSpellCheck(p);
    });
  };

  SpellCheckManager.prototype._onDocumentChange = function _onDocumentChange (change, info) {
      var this$1 = this;

    if (info.spellcheck) { return }
    
    
    
    var textProperties = this.textPropertyManager._textProperties;
    Object.keys(change.updated).forEach(function (pathStr) {
      if (textProperties[pathStr]) { this$1._scheduleCheck(pathStr); }
    });
  };

  SpellCheckManager.prototype._runSpellCheck = function _runSpellCheck (pathStr) {
      var this$1 = this;

    
    var path = pathStr.split(',');
    var text = this.editorSession.getDocument().get(path);
    var lang = this.editorSession.getLanguage();
    if (!text || !isString(text)) { return }
    sendRequest({
      method: 'POST',
      url: this.apiURL,
      header: {
        'Content-Type': 'application/json; charset=UTF-8',
      },
      data: {
        text: text,
        lang: lang
      }
    }).then(function (data) {
      data = JSON.parse(data);
      this$1._addSpellErrors(path, data);
    }).catch(function(err) {
      console.error(err);
    });
  };


  SpellCheckManager.prototype._addSpellErrors = function _addSpellErrors (path, data) {
    var editorSession = this.editorSession;
    var markersManager = editorSession.markersManager;
    
    
    var key = 'spell-error:'+path.join('.');
    var markers = data.map(function (m) {
      return {
        type: 'spell-error',
        start: {
          path: path,
          offset: m.start
        },
        end: {
          offset: m.end
        },
        suggestions: m.suggestions
      }
    });
    markersManager.setMarkers(key, markers);
    editorSession.startFlow();
  };

  var SpellCheckPackage = {
    name: 'spell-check',
    configure: function(config) {
      config.addCommand('correction', SpellCheckCommand, {
        commandGroup: 'spell-check'
      });
      config.addTool('correction', CorrectionTool);
    },
    SpellCheckManager: SpellCheckManager
  }

  var UndoCommand = BasePackage.UndoCommand;
  var RedoCommand = BasePackage.RedoCommand;
  var SelectAllCommand = BasePackage.SelectAllCommand;




  var TextInput = (function (AbstractEditor$$1) {
    function TextInput(parent, props) {
      if ( props === void 0 ) props = {};

      AbstractEditor$$1.call(this, parent, _createEditorSession(props));

      this.doc = this.editorSession.getDocument();
    }

    if ( AbstractEditor$$1 ) TextInput.__proto__ = AbstractEditor$$1;
    TextInput.prototype = Object.create( AbstractEditor$$1 && AbstractEditor$$1.prototype );
    TextInput.prototype.constructor = TextInput;

    TextInput.prototype.render = function render ($$) {
      var el = $$(this._getTagName()).addClass('sc-text-input');
      el.append(
        $$(TextInputEditor, {
          path: ['input', 'content']
        }).ref('input')
          .on('enter', this._onEnter)
          .on('escape', this._onEscape)
      );
      return el
    };

    TextInput.prototype.didMount = function didMount () {
      
      this.refs.input.selectLast();
    };

    TextInput.prototype.dispose = function dispose () {
      AbstractEditor$$1.prototype.dispose.call(this);
      this.doc.dispose();
      this.editorSession.dispose();
    };

    
    TextInput.prototype.shouldRerender = function shouldRerender () {
      return false
    };

    TextInput.prototype.getContent = function getContent () {
      return this.getDocument().getContent()
    };

    TextInput.prototype._getDocument = function _getDocument () {
      return this.context.editorSession.getDocument()
    };

    TextInput.prototype._getTagName = function _getTagName () {
      return this.props.tagName || 'div'
    };

    TextInput.prototype._onEnter = function _onEnter (event) {
      event.stopPropagation();
      this.el.emit('confirm');
    };

    TextInput.prototype._onEscape = function _onEscape (event) {
      event.stopPropagation();
      this.el.emit('cancel');
    };

    return TextInput;
  }(AbstractEditor));

  function _createEditorSession(props) {
    var config = new Configurator();
    config.addNode(TextNode);
    config.addCommand('undo', UndoCommand);
    config.addCommand('redo', RedoCommand);
    config.addCommand('select-all', SelectAllCommand);
    config.addKeyboardShortcut('CommandOrControl+Z', { command: 'undo' });
    config.addKeyboardShortcut('CommandOrControl+Shift+Z', { command: 'redo' });
    config.addKeyboardShortcut('CommandOrControl+A', { command: 'select-all' });

    config.defineSchema({
      name: 'text-input',
      
      
      defaultTextType: 'text',
      
      
      ArticleClass: TextInputDocument,
    });
    if (props.package) {
      config.import(props.package);
    }
    var doc = config.createArticle();
    if (props.content) {
      doc.set(['input', 'content'], props.content);
    }
    var editorSession = new EditorSession(doc, {
      configurator: config
    });
    return {
      editorSession: editorSession
    }
  }

  var TextInputDocument = (function (Document$$1) {
    function TextInputDocument() {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      Document$$1.apply(this, args);

      this.create({
        type: 'text',
        id: 'input',
        content: ''
      });
    }

    if ( Document$$1 ) TextInputDocument.__proto__ = Document$$1;
    TextInputDocument.prototype = Object.create( Document$$1 && Document$$1.prototype );
    TextInputDocument.prototype.constructor = TextInputDocument;
    TextInputDocument.prototype.getContentNode = function getContentNode () {
      return this.get('input')
    };
    TextInputDocument.prototype.getContent = function getContent () {
      return this.getContentNode().getText()
    };

    return TextInputDocument;
  }(Document));



  var TextInputEditor = (function (TextPropertyEditor$$1) {
    function TextInputEditor () {
      TextPropertyEditor$$1.apply(this, arguments);
    }

    if ( TextPropertyEditor$$1 ) TextInputEditor.__proto__ = TextPropertyEditor$$1;
    TextInputEditor.prototype = Object.create( TextPropertyEditor$$1 && TextPropertyEditor$$1.prototype );
    TextInputEditor.prototype.constructor = TextInputEditor;

    TextInputEditor.prototype.onKeyDown = function onKeyDown (event) {
      var handled = false;
      if (event.keyCode === 27) {
        handled = true;
        this.el.emit('escape');
      }
      if (handled) {
        event.stopPropagation();
        event.preventDefault();
      } else {
        TextPropertyEditor$$1.prototype.onKeyDown.call(this, event);
      }
    };

    TextInputEditor.prototype.selectLast = function selectLast () {
      var doc = this.getDocument();
      var input = doc.getContentNode();
      this.editorSession.setSelection({
        type: 'property',
        path: input.getPath(),
        startOffset: input.getLength(),
        surfaceId: this.id
      });
    };

    TextInputEditor.prototype.selectAll = function selectAll () {
      var doc = this.getDocument();
      var input = doc.getContentNode();
      this.editorSession.setSelection({
        type: 'property',
        path: input.getPath(),
        startOffset: 0,
        endffset: input.getLength(),
        surfaceId: this.id
      });
    };

    TextInputEditor.prototype._handleEnterKey = function _handleEnterKey () {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      TextPropertyEditor$$1.prototype._handleEnterKey.apply(this, args);
      this.el.emit('enter');
    };

    return TextInputEditor;
  }(TextPropertyEditor));

  var TextInputPackage = {
    name: 'text-input',
    configure: function configure(config) {
      config.addComponent('text-input', TextInput);
    },
    TextInput: TextInput
  }

  var ToolbarPackage = {
    name: 'toolbar',
    configure: function configure(config) {
      config.addComponent('toolbar', Toolbar);
    },
    Toolbar: Toolbar
  }

  var TextAlignCommand = (function (Command$$1) {
    function TextAlignCommand () {
      Command$$1.apply(this, arguments);
    }

    if ( Command$$1 ) TextAlignCommand.__proto__ = Command$$1;
    TextAlignCommand.prototype = Object.create( Command$$1 && Command$$1.prototype );
    TextAlignCommand.prototype.constructor = TextAlignCommand;

    TextAlignCommand.prototype.getCommandState = function getCommandState (params) {
      var sel = this._getSelection(params);
      var selectionState = params.editorSession.getSelectionState();
      var doc = params.editorSession.getDocument();
      var commandState = { disabled: true };

      if (sel.isPropertySelection() && !selectionState.isInlineNodeSelection()) {
        var path = sel.getPath();
        var node = doc.get(path[0]);
        if (node && node.isText() && node.isBlock()) {
          commandState.nodeId = node.id;
          commandState.disabled = false;
          if (node.textAlign === this.config.textAlign) {
            commandState.active = true;
          }
          
          
          commandState.showInContext = sel.start.offset === 0 && sel.end.offset === 0;
        }
      }
      return commandState
    };

    TextAlignCommand.prototype.execute = function execute (params) {
      var this$1 = this;

      var nodeId = params.commandState.nodeId;
      var editorSession = params.editorSession;
      editorSession.transaction(function (tx) {
        tx.set([nodeId, 'textAlign'], this$1.config.textAlign);
      });
    };

    return TextAlignCommand;
  }(Command));

  var TextAlignPackage = {
    name: 'text-align',
    configure: function(config) {
      config.addCommand('align-left', TextAlignCommand, {
        textAlign: 'left',
        commandGroup: 'text-align'
      });
      config.addCommand('align-center', TextAlignCommand, {
        textAlign: 'center',
        commandGroup: 'text-align'
      });
      config.addCommand('align-right', TextAlignCommand, {
        textAlign: 'right',
        commandGroup: 'text-align'
      });

      config.addKeyboardShortcut('CommandOrControl+Shift+L', { command: 'align-left' });
      config.addKeyboardShortcut('CommandOrControl+Shift+E', { command: 'align-center' });
      config.addKeyboardShortcut('CommandOrControl+Shift+R', { command: 'align-right' });

      config.addIcon('align-left', { 'fontawesome': 'fa-align-left' });
      config.addIcon('align-center', { 'fontawesome': 'fa-align-center' });
      config.addIcon('align-right', { 'fontawesome': 'fa-align-right' });

      config.addLabel('align-left', 'Left');
      config.addLabel('align-center', 'Center');
      config.addLabel('align-right', 'Right');

      config.addLabel('text-align', 'Text Align');
    }
  }

  exports.cssSelect = cssSelect;
  exports.DomUtils = domUtils;
  exports.ChangeStore = ChangeStore;
  exports.CollabServer = CollabServer;
  exports.DocumentClient = DocumentClient;
  exports.ServerRequest = ServerRequest;
  exports.WebSocketConnection = WebSocketConnection;
  exports.ClientConnection = ClientConnection;
  exports.CollabServerConfigurator = CollabServerConfigurator;
  exports.DocumentEngine = DocumentEngine;
  exports.ServerResponse = ServerResponse;
  exports.computeSnapshot = computeSnapshot;
  exports.CollabClient = CollabClient;
  exports.CollabServerPackage = CollabServerPackage;
  exports.DocumentServer = DocumentServer;
  exports.SnapshotEngine = SnapshotEngine;
  exports.CollabEngine = CollabEngine;
  exports.CollabSession = CollabSession;
  exports.Server = Server;
  exports.SnapshotStore = SnapshotStore;
  exports.DocumentArchive = DocumentArchive;
  exports.VfsLoader = VfsLoader;
  exports.ManifestLoader = ManifestLoader;
  exports.PersistedDocumentArchive = PersistedDocumentArchive;
  exports.HttpStorageClient = HttpStorageClient;
  exports.InMemoryDarBuffer = InMemoryDarBuffer;
  exports.VfsStorageClient = VfsStorageClient;
  exports.DOMElement = DOMElement;
  exports.DOMEventListener = DOMEventListener;
  exports.DefaultDOMElement = DefaultDOMElement;
  exports.BrowserDOMElement = BrowserDOMElement;
  exports.MemoryDOMElement = MemoryDOMElement;
  exports.domHelpers = domHelpers;
  exports.documentHelpers = documentHelpers;
  exports.operationHelpers = operationHelpers;
  exports.selectionHelpers = selectionHelpers;
  exports.annotationHelpers = annotationHelpers;
  exports.AnnotationIndex = AnnotationIndex;
  exports.AnnotationMixin = AnnotationMixin;
  exports.ArrayOperation = ArrayOperation;
  exports.BlockNode = BlockNode;
  exports.ChangeHistory = ChangeHistory;
  exports.ChangeRecorder = ChangeRecorder;
  exports.ClipboardExporter = ClipboardExporter;
  exports.ClipboardImporter = ClipboardImporter;
  exports.Conflict = Conflict;
  exports.Container = Container;
  exports.ContainerAdapter = ContainerAdapter;
  exports.ContainerAddress = ContainerAddress;
  exports.ContainerAnnotation = ContainerAnnotation;
  exports.ContainerAnnotationIndex = ContainerAnnotationIndex;
  exports.ContainerMixin = ContainerMixin;
  exports.ContainerSelection = ContainerSelection;
  exports.Coordinate = Coordinate;
  exports.CoordinateAdapter = CoordinateAdapter;
  exports.CoordinateOperation = CoordinateOperation;
  exports.copySelection = copySelection;
  exports.createDocumentFactory = createDocumentFactory;
  exports.CustomSelection = CustomSelection;
  exports.Data = Data;
  exports.DefaultChangeCompressor = DefaultChangeCompressor;
  exports.DefaultFileProxy = DefaultFileProxy;
  exports.Document = Document;
  exports.DocumentChange = DocumentChange;
  exports.DocumentIndex = DocumentIndex;
  exports.DocumentNode = DocumentNode;
  exports.DocumentNodeFactory = DocumentNodeFactory;
  exports.DocumentSchema = DocumentSchema;
  exports.DOMExporter = DOMExporter;
  exports.DOMImporter = DOMImporter;
  exports.Editing = Editing;
  exports.EditingBehavior = EditingBehavior;
  exports.EditingInterface = EditingInterface;
  exports.FileNode = FileNode;
  exports.FileProxy = FileProxy;
  exports.Fragmenter = Fragmenter;
  exports.getChangeFromDocument = getChangeFromDocument;
  exports.HTMLExporter = HTMLExporter;
  exports.HTMLImporter = HTMLImporter;
  exports.IncrementalData = IncrementalData;
  exports.InlineNode = InlineNode;
  exports.JSONConverter = JSONConverter;
  exports.ListMixin = ListMixin;
  exports.Marker = Marker;
  exports.Node = Node;
  exports.NodeIndex = NodeIndex;
  exports.NodeRegistry = NodeRegistry;
  exports.NodeSelection = NodeSelection;
  exports.ObjectOperation = ObjectOperation;
  exports.OperationSerializer = OperationSerializer;
  exports.ParentNodeHook = ParentNodeHook;
  exports.paste = paste;
  exports.Property = Property;
  exports.PropertyAnnotation = PropertyAnnotation;
  exports.PropertyIndex = PropertyIndex;
  exports.PropertySelection = PropertySelection;
  exports.Range = Range;
  exports.Schema = Schema;
  exports.Selection = Selection;
  exports.SelectionState = SelectionState;
  exports.TextBlock = TextBlock;
  exports.TextNode = TextNode;
  exports.TextNodeMixin = TextNodeMixin;
  exports.TextOperation = TextOperation;
  exports.Transaction = Transaction;
  exports.XMLExporter = XMLExporter;
  exports.XMLImporter = XMLImporter;
  exports.AbstractClipboard = AbstractClipboard;
  exports.AbstractEditor = AbstractEditor;
  exports.AbstractIsolatedNodeComponent = AbstractIsolatedNodeComponent;
  exports.AbstractScrollPane = AbstractScrollPane;
  exports.AnnotatedTextComponent = AnnotatedTextComponent;
  exports.AnnotationCommand = AnnotationCommand;
  exports.AnnotationComponent = AnnotationComponent;
  exports.BlockNodeComponent = BlockNodeComponent;
  exports.Clipboard = Clipboard;
  exports.Command = Command;
  exports.CommandManager = CommandManager;
  exports.Component = Component;
  exports.ComponentRegistry = ComponentRegistry;
  exports.Configurator = Configurator;
  exports.ContainerEditor = ContainerEditor;
  exports.CursorComponent = CursorComponent;
  exports.CustomSurface = CustomSurface;
  exports.DefaultLabelProvider = DefaultLabelProvider;
  exports.DOMSelection = DOMSelection;
  exports.DragAndDropHandler = DragAndDropHandler;
  exports.DragManager = DragManager;
  exports.EditAnnotationCommand = EditAnnotationCommand;
  exports.EditInlineNodeCommand = EditInlineNodeCommand;
  exports.EditorSession = EditorSession;
  exports.ExecuteCommandHandler = ExecuteCommandHandler;
  exports.FileManager = FileManager;
  exports.FontAwesomeIcon = FontAwesomeIcon;
  exports.FontAwesomeIconProvider = FontAwesomeIconProvider;
  exports.GlobalEventHandler = GlobalEventHandler;
  exports.Highlights = Highlights;
  exports.InlineNodeComponent = InlineNodeComponent;
  exports.InsertInlineNodeCommand = InsertInlineNodeCommand;
  exports.InsertNodeCommand = InsertNodeCommand;
  exports.IsolatedNodeComponent = IsolatedNodeComponent;
  exports.KeyboardManager = KeyboardManager;
  exports.MacroManager = MacroManager;
  exports.MarkersManager = MarkersManager;
  exports.MenuItem = MenuItem;
  exports.NodeComponent = NodeComponent;
  exports.Overlay = Overlay;
  exports.RenderingEngine = RenderingEngine;
  exports.ResourceManager = ResourceManager;
  exports.ResponsiveApplication = ResponsiveApplication;
  exports.Router = Router;
  exports.SelectionFragmentComponent = SelectionFragmentComponent;
  exports.Surface = Surface;
  exports.SurfaceManager = SurfaceManager;
  exports.SwitchTextTypeCommand = SwitchTextTypeCommand;
  exports.TextBlockComponent = TextBlockComponent;
  exports.TextPropertyComponent = TextPropertyComponent;
  exports.TextPropertyEditor = TextPropertyEditor;
  exports.ToggleTool = ToggleTool;
  exports.ToolDropdown = ToolDropdown;
  exports.ToolGroup = ToolGroup;
  exports.MenuGroup = MenuGroup;
  exports.ToolPrompt = ToolPrompt;
  exports.ToolPanel = ToolPanel;
  exports.Tooltip = Tooltip;
  exports.UnsupportedNodeComponent = UnsupportedNodeComponent;
  exports.VirtualElement = VirtualElement;
  exports.WorkflowPane = WorkflowPane;
  exports.Button = Button;
  exports.Layout = Layout;
  exports.ScrollPane = ScrollPane;
  exports.SplitPane = SplitPane;
  exports.Toolbar = Toolbar;
  exports.Tool = ToggleTool;
  exports.findParentComponent = findParentComponent;
  exports.async = async;
  exports.ArrayIterator = ArrayIterator;
  exports.ArrayTree = ArrayTree;
  exports.array2table = array2table;
  exports.camelCase = camelCase;
  exports.capitalize = capitalize;
  exports.clone = clone;
  exports.cloneDeep = cloneDeep;
  exports.createCountingIdGenerator = createCountingIdGenerator;
  exports.debounce = debounce;
  exports.deleteFromArray = deleteFromArray;
  exports.diff = diff;
  exports.encodeXMLEntities = encodeXMLEntities;
  exports.EventEmitter = EventEmitter;
  exports.extend = extend;
  exports.Factory = Factory;
  exports.filter = filter;
  exports.find = find;
  exports.findIndex = findIndex$1;
  exports.flatten = flatten;
  exports.flattenOften = flattenOften;
  exports.forEach = forEach;
  exports.getRangeFromMatrix = getRangeFromMatrix;
  exports.getRelativeBoundingRect = getRelativeBoundingRect;
  exports.getRelativeMouseBounds = getRelativeMouseBounds;
  exports.inBrowser = inBrowser;
  exports.includes = includes;
  exports.isArray = isArray;
  exports.isArrayEqual = isArrayEqual;
  exports.isBoolean = isBoolean;
  exports.isEqual = isEqual;
  exports.isFunction = isFunction$1;
  exports.isMatch = isMatch;
  exports.isNil = isNil;
  exports.isNumber = isNumber;
  exports.isObject = isObject;
  exports.isPlainObject = isPlainObject;
  exports.isString = isString;
  exports.last = last$1;
  exports.levenshtein = levenshtein;
  exports.makeMap = array2table;
  exports.map = map;
  exports.merge = merge$1;
  exports.PathObject = PathObject;
  exports.percentage = percentage;
  exports.platform = platform;
  exports.pluck = pluck;
  exports.printStacktrace = printStacktrace;
  exports.Registry = Registry;
  exports.request = request;
  exports.renderListNode = renderListNode;
  exports.sendRequest = sendRequest;
  exports.startsWith = startsWith;
  exports.SubstanceError = SubstanceError;
  exports.substanceGlobals = substanceGlobals;
  exports.times = times$2;
  exports.TreeIndex = TreeIndex;
  exports.uniq = uniq;
  exports.uuid = uuid;
  exports.without = without;
  exports.orderBy = orderBy;
  exports.keys = keys$1;
  exports.parseKeyEvent = parseKeyEvent;
  exports.getDOMRangeFromEvent = getDOMRangeFromEvent;
  exports.getSelectionRect = getSelectionRect;
  exports.getRelativeRect = getRelativeRect;
  exports.isMouseInsideDOMSelection = isMouseInsideDOMSelection;
  exports.setDOMSelection = setDOMSelection;
  exports.getQueryStringParam = getQueryStringParam;
  exports.XMLAnchorNode = XMLAnchorNode;
  exports.XMLAnnotationNode = XMLAnnotationNode;
  exports.XMLContainerNode = XMLContainerNode;
  exports.XMLDocument = XMLDocument;
  exports.XMLDocumentImporter = XMLDocumentImporter;
  exports.XMLEditingInterface = XMLEditingInterface;
  exports.XMLElementNode = XMLElementNode;
  exports.XMLElementNodeConverter = ElementNodeConverter;
  exports.XMLExternalNode = XMLExternalNode;
  exports.XMLExternalNodeConverter = ExternalNodeConverter;
  exports.XMLSchema = XMLSchema;
  exports.XMLTextElement = XMLTextElement;
  exports.XMLTextElementConverter = XMLTextElementConverter;
  exports.checkSchema = checkSchema;
  exports.compileRNG = compileRNG;
  exports.registerSchema = registerSchema;
  exports.SchemaDrivenCommandManager = SchemaDrivenCommandManager;
  exports.loadRNG = _loadRNG;
  exports.validateXMLSchema = validateXML;
  exports.ValidatingChildNodeIterator = ValidatingChildNodeIterator;
  exports.prettyPrintXML = prettyPrintXML;
  exports.BasePackage = BasePackage;
  exports.BlockquotePackage = BlockquotePackage;
  exports.BodyScrollPanePackage = BodyScrollPanePackage;
  exports.ButtonPackage = ButtonPackage;
  exports.CodePackage = CodePackage;
  exports.CodeblockPackage = CodeblockPackage;
  exports.ContextMenuPackage = ContextMenuPackage;
  exports.DropzonesPackage = DropzonesPackage;
  exports.EmphasisPackage = EmphasisPackage;
  exports.FilePackage = FilePackage;
  exports.FindAndReplacePackage = FindAndReplacePackage;
  exports.GridPackage = GridPackage;
  exports.GutterPackage = GutterPackage;
  exports.HeadingPackage = HeadingPackage;
  exports.ImagePackage = ImagePackage;
  exports.InlineWrapperPackage = InlineWrapperPackage;
  exports.InputPackage = InputPackage;
  exports.LayoutPackage = LayoutPackage;
  exports.LinkPackage = LinkPackage;
  exports.ListPackage = ListPackage;
  exports.ModalPackage = ModalPackage;
  exports.MultiSelectPackage = MultiSelectPackage;
  exports.OverlayPackage = OverlayPackage;
  exports.ParagraphPackage = ParagraphPackage;
  exports.PersistencePackage = PersistencePackage;
  exports.ProseEditorPackage = ProseEditorPackage;
  exports.QuoteMarksPackage = QuoteMarksPackage;
  exports.ScrollPanePackage = ScrollPanePackage;
  exports.ScrollbarPackage = ScrollbarPackage;
  exports.SpellCheckPackage = SpellCheckPackage;
  exports.SplitPanePackage = SplitPanePackage;
  exports.StrongPackage = StrongPackage;
  exports.SubscriptPackage = SubscriptPackage;
  exports.SuperscriptPackage = SuperscriptPackage;
  exports.TabbedPanePackage = TabbedPanePackage;
  exports.TablePackage = TablePackage;
  exports.TextInputPackage = TextInputPackage;
  exports.ToolbarPackage = ToolbarPackage;
  exports.TextAlignPackage = TextAlignPackage;

  Object.defineProperty(exports, '__esModule', { value: true });

})));

//# sourceMappingURL=./substance.es5.js.map